package charite.christo;
import charite.christo.interfaces.*;
import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.List;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**
   @author Christoph Gille
*/
public final class Archive {
    private static class Entry implements Comparable {
        long time;//=System.currentTimeMillis();
        int from, to, hashCd;
        String name;
        //@Override public String toString() { return name+" {"+from+","+to+"}";}
        public int compareTo(Object o) {
            final Entry e=(Entry)o;
            return e.from<from ? 1 : e.from==from ? 0 : -1;
        }
    }
    private List<Entry> vEntries;
    private final File file, fileHeader;
    private RandomAccessFile _fp;
    private String _fpMode;
    public Archive(File f,File fHeader) {
        fileHeader=fHeader;
        mkParentDrs(file=f);
    }
    /*
    private RandomAccessFile ________raf(boolean write) throws IOException {
        RandomAccessFile raf=this._fp;
        if (raf!=null && write && this._fpMode=="r") {
            raf.close();
            raf=null;
        }
        if (!write && sze(file)==0) return null;
        if (raf==null) this._fp=raf=Insecure.randomAccessFile(this.file,this._fpMode=write ? "rw" : "r");
        return raf;
    }
    */
    private RandomAccessFile raf(boolean write) throws IOException {
        RandomAccessFile raf=this._fp;
        if (raf==null) this._fp=raf=Insecure.randomAccessFile(this.file, "rw");
        return raf;
    }

    private List<Entry> vEntries() {
        if (vEntries==null) {
            vEntries=new java.util.ArrayList();
            final ByteArray ba=readBytes(fileHeader,(ByteArray)null);
            if (ba!=null) {
                final long len=file.length();
                final byte[] bb=ba.getBytes();
                final int ends[]=ba.getLineEnds();
                for(int i=0;i<ends.length;i++) {
                    final int begin= i==0? 0 : ends[i-1]+1, end=ends[i];
                    int tab1=-1, tab2=-1, tab3=-1, tab4=-1;
                    for(int pos=begin;pos<end;pos++) {
                        if (bb[pos]=='\t') {
                            if (tab1<0) tab1=pos;
                            else if (tab2<0) tab2=pos;
                            else if (tab3<0) tab3=pos;
                            else if (tab4<0) tab4=pos;
                        }
                    }
                    if (tab3<0) {  continue;}
                    final int to=atoi(bb,tab3+1);
                    if (len<to) continue;
                    final Entry e=new Entry();
                    e.name=new String(bb, begin,tab1-begin);
                    e.time=atol(bb,tab1+1,MAX_INT);
                    e.from=atoi(bb,tab2+1);
                    e.to=atoi(bb,tab3+1);
                    if (tab4>0) e.hashCd=atoi(bb,tab4+1);
                    vEntries().add(e);
                }
            }
        }
        return vEntries;
    }
    private Entry getEntry(String entryName) {
        final String name=entryName.replace('\n','_');
        final List<Entry> v=vEntries();
        final int L=name.length();
        for(int i=v.size(); --i>=0;) {
            final Entry e=v.get(i);
            if (L==e.name.length() && name.equals(e.name)) return e;
        }
        return null;
    }
    public boolean contains(String s) { return getEntry(s)!=null;}
    public long getLastModified(String s) {
        Entry e=getEntry(s);
        return e!=null ? e.time : 0;
    }
    public ByteArray get(String name, ByteArray buffer) {
        name=name.replace('\n','_');
        final Entry e=getEntry(name);
        if (e==null) return null;
        final ByteArray ba=buffer!=null ? buffer : new ByteArray(0);
        final int len=e.to-e.from;//len();
        byte bb[]=ba.getBytes();
        if (len>bb.length) bb=new byte[len];
        try {
            final RandomAccessFile fp=raf(false);
            if (fp==null) return null;
            fp.seek(e.from);
            fp.readFully(bb,0,len);
            ba.set(bb,0,len);
            //if (isMyComputer()) putln ("e.hashCd="+e.hashCd+" hashCd="+hashCd(bb,0,len,false));
            if (e.hashCd!=0 && e.hashCd!=hashCd(bb,0,len,false)) {
                putln(RED_WARNING+"Archive: wrong hashcode", file);
                return null;
            }
        }catch(IOException ex){stckTrc(ex); return null;}
        return ba;
    }
    /* <<< Get  <<< */
    /* ---------------------------------------- */
    /* >>> Write  >>> */
    private boolean modified;
    public void write(String name,ByteArray ba) {
        final byte[] T=ba.getBytes();
        final int B=ba.getBegin(), E=ba.getEnd();
        name=name.replace('\n','_');
        if (vEntries!=null) vEntries.remove(getEntry(name));
        assertNotNull(name);
        if (ba==null || name==null) return;
        modified=true;
        final int len=E-B;
        sortArry(vEntries());
        final int nE=vEntries().size();
        int insertAt=-1, insertGap=MAX_INT, lastTo=0,maxTo=0;
        for(int i=0; i<nE; i++) {
            final Entry e=vEntries().get(i);
            final int gap=e.from-lastTo;
            if (gap>=len && gap<insertGap) {
                insertAt=lastTo;
                insertGap=gap;
            }
            lastTo=e.to;
            if (maxTo<lastTo) maxTo=lastTo;
        }
        try {
            Entry newE=getEntry(name);
            final boolean contains=newE!=null;
            if (!contains) (newE=new Entry()).name=name;
            final RandomAccessFile fp=raf(true);
            newE.from=insertAt<0 ? maxTo : insertAt;
            newE.to=newE.from+len;
            newE.time=System.currentTimeMillis();
            newE.hashCd=hashCd(T,B,E, false);
            fp.seek(newE.from);
            fp.write(T, B, E);
            if (!contains) vEntries().add(newE);
        } catch(IOException ex){stckTrc(ex);}
    }
    private boolean dirty;
    void deleteBecauseError() {
        putln(RED_ERROR+" Archive.deleteBecauseError()  ",file);
        dirty=true;
        delFile(file);
        delFile(fileHeader);
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Close >>> */
    void close() {
        final RandomAccessFile fp=this._fp;
        if (fp!=null) try { fp.close();} catch(IOException e){}
    }
    void dispose() {
        if (modified && !dirty) {
            final List<Entry> v=vEntries();
            final int nE=v.size();
            int size=0;
            ByteArray sb=null;
            while(true) {
                for(int i=0; i<nE; i++) {
                    final Entry e;
                    try {e=v.get(i);} catch(Exception idxOutOfBounds){ break;}
                    if (sb!=null) sb.append(e.name).append('\t').append(e.time).append('\t')
                                      .append(e.from).append('\t').append(e.to).append('\t')
                                      .append(e.hashCd).append('\n');
                    else size+=e.name.length()+(4+30);
                }
                if (sb==null) sb=new ByteArray(size); else break;
            }
            wrte(fileHeader,sb);
            if (isMyComputer()) {
                putln("Archive save ",fileHeader+"  "+sze(fileHeader));
            }

        }
    }
    @Override protected void finalize() throws Throwable {
        //logFinalize(ANSI_RED+ANSI_FG_WHITE+"Archive","");
        dispose();
        close();
        super.finalize();
    }
}
