package charite.christo;
import java.util.*;
import charite.christo.interfaces.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import java.io.*;
/**
   Similar to java.lang.StringBuffer but more efficient.
   Consumes only 8 bit and does not throw OutOfMemory as aften as StringBuffer.
   Contains an 8bit  byte array to store text.
   Line ends are returned as an array of text positions.
*/
public final class ByteArray implements CharSequence, Comparable, ChRunnable, HasModiCount {
    private final static Object SYNC_getLineEnds=new Object();
    public final static Object CLEAR=new Object();
    private final static int CACHE_TO_STRG=0, CACHE_ENDS=1, CACHE_SB_TMP=2;

    private byte bytes[];
    private int begin,end,  modiCount, reserve=99, hashCode;
    private boolean isLC;

    private Object refCached;
    private static Object bufEnds;
       /* <<< Fields <<< */
    /* ---------------------------------------- */
    /* >>> Constructor >>> */

    {  assertInstAllowed(); }
    public ByteArray(byte[] bb) { set(bb,0,sze(bb));}
    public ByteArray(byte[] bb, int from, int to) { set(bb, from, to); }
    public ByteArray(int l) { bytes=l>0 ? new byte[l] : NO_BYTE; }
    public ByteArray(CharSequence cs) {
        if (cs instanceof ByteArray) {
            ByteArray ba=(ByteArray)cs;
            set(ba.getBytes(), ba.getBegin(), ba.getEnd());
        } else if (cs!=null && cs!="") {
            set(strg2bytes(toStrg(cs)),0,cs.length());
        }
    }
    /* <<< Constructor <<< */
    /* ---------------------------------------- */
    /* >>> Setters >>> */
    public ByteArray setBegin(int index){
        begin=index;
        txtOrRangeChanged();
        return this;
    }
    private void txtOrRangeChanged() {
        modiCount++; refCached=null; hashCode=0;
    }
    public ByteArray setEnd(int index){
        final int e=maxi(0,index);
        if (end<e) escapesL=escapesEndText=0;
        end=e;
        txtOrRangeChanged();
        final byte bb[]=getBytes();
        return this;
    }
    public ByteArray set(byte[] bb, int b, int e) {
        if (!equlsBB(bb, 0, mini(e,this.end), this.bytes, 0)) escapesL=escapesEndText=0;
        bytes=bb; begin=b;
        return setEnd(e);
    }
    /* <<< Setters <<< */
    /* ---------------------------------------- */
    /* >>> Basic Getters >>> */
    /** First text position +1 */
    public int getBegin(){return begin;}
    /** Last text position +1 */
    public int getEnd() { return mini(getBytes().length, end);}
    /** The byte array used internally to store the text; */
    public byte[] getBytes() { final byte bb[]=this.bytes; return bb!=null ? bb : NO_BYTE; }

    /* <<< Basic Getters <<< */
    /* ---------------------------------------- */
    /* >>> Line ends >>> */
   /**
       Indices of the '\n'.
       If trailing  '\n' is missing inlcude also text-length+1 .
    */
    public int[] getLineEnds(){
        final Object[] cached=cached();
        int ends[]=(int[]) cached[CACHE_ENDS];
        if (ends==null) {
            synchronized(SYNC_getLineEnds) {
                int[] buf=(int[])deref(bufEnds);
                if (buf==null) bufEnds=newSoftRef(buf=new int[32*1024]);
                final byte bb[]=getBytes();
                final int e=mini(bb.length,getEnd());
                int count=0;
                for(int i=getBegin(); i<e; i++) {
                    if (bb[i]=='\n') {
                        if (buf.length<=count+1) buf=chSze(buf,count+1+count/2);
                        buf[count++]=i;
                    }
                }
                if (e==0 || bb[e-1]!='\n') {
                    buf[count++]=e;
                }
                System.arraycopy(buf, 0, ends=new int[count], 0, count);
                cached[CACHE_ENDS]=ends;
            }
        }
        return ends;
    }
    /* <<< Line Ends <<< */
    /* ---------------------------------------- */
    /* >>> Ansi >>> */
    private long[] escapes;
    private int escapesL, escapesEndText;
    public int countAnsiEscapes() {
        getAnsiEscapes();
        return escapesL;
    }
    public long[] getAnsiEscapes() {
        if (getEnd()>escapesEndText) {
            final long escape[]=findEscapes();
            removeEscapes(escapes, this.escapesL);
        }
       final long[] ee=this.escapes;
       return ee!=null ? ee : NO_LONG;
    }
    public long[] findEscapes() {
        final byte[] T=getBytes();
        final int B=maxi(this.escapesEndText, getBegin(),0),  E=mini(getEnd(),sze(T));
        long ee[]=this.escapes;
        int count=this.escapesL;
        for(int pos=B; pos+2<E; pos++) {
            if (T[pos]!='\u001B' || T[pos+1]!='[') continue;
            int begin=pos, i=pos+2;
            for(int end; (end=AnsiEscape.escapeEndsAt(i, T, E))>0; ) {
                if (ee==null || count>=ee.length) ee=chSze(ee,count+10+count/2);
                ee[count++]=AnsiEscape.newEscape(atoi(T,i), begin, (end-begin));
                i=begin=end;
            }
        }
        this.escapesL=count;
        this.escapesEndText=E;
        return this.escapes=ee;
    }
    private int removeEscapes(long[] escape, int numEscape) {
        final byte[] T=getBytes();
        final int E=mini(getEnd(),sze(T));
        final int N=mini(numEscape,sze(escape));
        int sum=0;
        for(int i=0; i<N; i++) {
             long e=escape[i];
            final int len= (int)((e>>AnsiEscape.SHIFT_LEN)&255);
            if (len==0) continue;
            final int pos=(int)(e&AnsiEscape.MASK_POS);
            final int from=pos+len, to=i+1<N ? (int)(escape[i+1]&AnsiEscape.MASK_POS) : E;
            final int dest=pos-sum;
            try {
                System.arraycopy(T,from, T,dest, to-from);
            } catch (Exception ex) {
                putln(RED_ERROR+" from="+from+" dest="+dest+" N="+(to-from));
                continue;
            }

            e&=~ (255L<<AnsiEscape.SHIFT_LEN);
            e=(e& AnsiEscape.MASK_POS_INV) |  (pos-sum);
            escape[i]=e;
            sum+=len;
        }
        if (sum>0) {
            this.end=maxi(0,E-sum);
            this.escapesEndText-=sum;
            refCached=null;
        }
        return sum;
    }

    /* <<< Ansi <<< */
    /* ---------------------------------------- */
    /* >>> pre-tags >>> */
    public void insidePreTags() {
        delBlanksL();
        final int b=getBegin(), e=getEnd();
        final byte[] T=getBytes();
        if (e-b>20 &&   T[b]=='<' &&  (equls(STRSTR_IC,"<!DOCTYPE",T,b) || equls(STRSTR_IC,"<html",T,b) || equls(STRSTR_IC,"<?xml ",T,b) || equls("<p>Multiple sequence alignment ",T,b)  )) {
            /* Ausnahme: http://www.ensembl.org//Homo_sapiens/Component/Gene/Web/ComparaTree/Align?g=ENSG00000008853 */
            final int pre=strstr(STRSTR_IC|STRSTR_AFTER,"<pre>",T,b+10,MAX_INT);
            final int erp=pre>0 ? strstr(STRSTR_IC,"</pre>",T,pre,MAX_INT) : -1;
            if (erp>0) {
                setBegin(pre);
                setEnd(erp);
                delBlanksL();
                { /* For files from Kegg */
                    while(equls("<!-- ",T,getBegin()) ) {
                        final int commentE= strstr(STRSTR_AFTER," -->",T,getBegin(),erp);
                        if (commentE<0) break;
                        setBegin(commentE);
                        delBlanksL();
                    }
                }

            }

        }
    }

    public ByteArray delBlanksL() {
        final int b=getBegin(), e=getEnd();
        setBegin(nxtE(NO_SPC,getBytes(),b,e));
        return this;
    }
    public ByteArray delBlanksR() {
        final int b=getBegin(), e=getEnd();
        setEnd(prevE(NO_SPC,getBytes(),e-1,b-1)+1);
        return this;
    }

    /**
        Optimizes the size of the byte array,
        if (getEnd()-getBegin()) is less than getBytes().length then the size of the byte array.
    */
    public static byte[] trimSize(ByteArray ba) {
        if (ba==null) return null;
        final byte bb[]=ba.getBytes();
        if (bb==null) return null;
        if (bb.length>ba.end-ba.begin) {
            final byte[] bbNew=new byte[ba.end-ba.begin];
            System.arraycopy(bb,ba.begin,bbNew,0,ba.end-ba.begin);
            return bbNew;
        }
        return bb;
    }
    /** returns the text between getBegin() and getEnd() */
    public byte[] newByteArray(final int b, final int e) {
        final int B=maxi(0,b);
        final int E=mini(e,getEnd());
        final byte[] T=getBytes();
        if (E<=B) return NO_BYTE;
        final byte bb[]=new byte[e-B];
        System.arraycopy(T,B,bb,0,bb.length);
        return bb;
    }

    //   public String toStringIP(int from, int to) {  return toStrgIP(getBytes(),from,to); }

    public String newString() { return newString(getBegin(),getEnd()); }
    public String newString(int from, int to) { return toString(from,to, (long)0);}
    public String newStringTrim(int from, int to) {
        return toString(nxtE(NO_SPC,getBytes(),from,to),to, (long)0);
    }

    public String newStringToGt(int from) {
        final int gt=from<0 ? -1:strchr('<',getBytes(), from, getEnd());
        return gt<0?null:newString(from,gt);
    }
    public ByteArray newBaToGt(int from) {
        final int gt=from<0 ? -1 : strchr('<',getBytes(), from ,getEnd());
        return gt<0?null: new ByteArray(getBytes(),from,gt);
    }
    public final static long  STRING_POOL=1L<<30, STRING_TRIM=1L<<31;

    /* Options can be combined with ChUtils.mapColumn1Column2(... ) */
    public String toString(int from, int to, long options) {
        final byte[] bb=getBytes();
        int f=maxi(0,from), t=mini(sze(bb),to,bb.length);
        if ( (options&STRING_TRIM)!=0) {
            f=nxtE(NO_SPC,bb,f,t);
            t=prevE(NO_SPC,bb,t-1,f-1)+1;
        }
        return
            f>=t ?  "" :
            (options&STRING_POOL)!=0 ?  toStrgIP(bb,from,to) :
            bytes2strg(bb,f,t);
    }
    public int idxToRow(final int position){
        try {
            final int[] ends=getLineEnds();
            if (position<0) return -1;
            if (position>ends[ends.length-1]) return ends.length-1;
            int lastStep=-1, lastRow=-1, step=ends.length/2, row=step;
            for(;;) {
                if (lastStep==step && lastRow==row) { stckTrc(); return -1; } // should never happen
                lastStep=step;
                lastRow=row;
                if (row<0) row=0;
                if (row>=ends.length) row=ends.length-1;
                step=(step+1)/2;
                final int begin=row==0?0:ends[row-1]+1;
                if (begin<=position && position<=ends[row]) return row;
                row+=begin<position ? step : -step;
            }
        } catch(Exception e){ stckTrc(e);}
        return -1;
    }

    public int getModificationCount() { return modiCount;}

    public void setTextSegment(javax.swing.text.Segment s) {
        if (s==null) return;
        final int sBegin=s.getBeginIndex();
        final int sEnd=s.getEndIndex();
        final byte[] bb=sze(bytes)>sEnd-sBegin ? bytes : new byte[sEnd-sBegin+99];
        final char cc[]=s.array;
        int iB=0;
        for(int iS=sBegin; iS<sEnd;iS++) {
            if (cc[iS]!='\r' && cc[iS]!=0) bb[iB++]=(byte)cc[iS];
        }
        set(bb,0,iB);
    }
    public boolean onlyWhiteSpace() {
        final byte bb[]=getBytes();
        final int end=getEnd();
        for(int i=getBegin(); i<end; i++) {
            final int c=bb[i];
            if (c!=' ' && c!='\n' && c!='\r' && c!='\t' && c!=0) return false;
        }
        return true;
    }

    public ByteArray enlargeCapacityBy(int reserve) { this.reserve=reserve; return this;}


    public ByteArray append(long num) { return append(num,0);}
    public ByteArray append(char ch) {
        final int end=getEnd();
        minBufferSize(end+1)[end]=(byte)ch;
        setEnd(end+1);
        return this;
    }
    public ByteArray append(char ch, int times) {
        if (times>0) {
        final int end=getEnd();
        final byte txt[]=minBufferSize(end+times);
        for(int i=times;--i>=0;) txt[end+i]=(byte)ch;
        setEnd(end+times);
        }
        return this;
    }

    public ByteArray appendSomeBytes(long n,int num) {
        final int e=getEnd();
        final byte txt[]=minBufferSize(e+num);
        setEnd(e+num);
        for(int i=num; --i>=0;) {
            byte b=(byte) ((n>>>(i*8)) & 127);
            txt[e+i]=b==0?32:b;
        }
        return this;
    }

    public ByteArray append(double f) {
        final Object[] cached=cached();
        StringBuffer sb=(StringBuffer)cached[CACHE_SB_TMP];
        if (sb==null) cached[CACHE_SB_TMP]=sb=new StringBuffer(); else sb.setLength(0);
        return append(sb.append(f));
    }

    public ByteArray appendFloat(double f, int precision) {
        if (precision>0 && javaVsn()>14) {
            final String s=forJavaVersion(15).run(RunIf15orHigher.RUN_BIG_DECI, new Object[]{Double.toString(f),intObjct(precision)}).toString();
            final int e=s.indexOf('E'), nulls=(e>0 ? e : s.length())-1;
            if (chrAt(nulls,s)=='0' && s.indexOf('.')>0) {
                int nul=nulls;
                while( chrAt(nul-1,s)=='0') nul--;
                if(chrAt(nul-1,s)=='.') nul--;
                return append(s,0,nul).append(s,nulls+1,MAX_INT);
            } else return append(s);
        }
        else {
            final Object[] cached=cached();
            StringBuffer sb=(StringBuffer)cached[CACHE_SB_TMP];
            if (sb==null) cached[CACHE_SB_TMP]=sb=new StringBuffer(); else sb.setLength(0);
            if (f>0 ? f<1E-44 : -f<1E-44) sb.append(f); else sb.append((float)f);
            final int dot=strchr(STRSTR_E, '.', sb);
            return append(sb);
        }
    }
    public ByteArray join(Object oo[], String glue) {
        final int L=sze(oo);
        boolean first=true;
        for(int i=0;i<L;i++) {
            if (oo[i]==null) continue;
            if (!first) append(glue);
            first=true;
            append(oo[i]);
        }
        return this;
    }

    public ByteArray appendln(Object s) { return append(s).append('\n'); }
    public ByteArray append(Object s) {
        if (s==CLEAR) setEnd(0); else append(s,0,MAX_INT);
        return this;
    }

    public ByteArray append(final Object o, final int from, final int to) {
        if (o instanceof ByteArray) {
            final ByteArray ba=(ByteArray)o;
            return append(ba.getBytes(),  maxi(from,ba.getBegin()), mini(to,ba.getEnd()));
        }
        final byte bb[]=o instanceof byte[] ? (byte[])o : null;
        final CharSequence s=bb!=null ? null : o instanceof CharSequence ? (CharSequence)o : toStrg(o);
        final int t=mini(to, bb!=null?bb.length:sze(s)), f=maxi(0,from), len=t-f;
        if (len>0) {
            final int end=getEnd();
            final byte bytes[]=minBufferSize(end+len);
            if (bb!=null) System.arraycopy(bb,f, bytes, end,len);
            else for(int i=len; --i>=0;) bytes[end+i]=(byte)s.charAt(i+f);
            setEnd(end+len);
        }
        return this;
    }

    public ByteArray filter(int type, Object o, int from ,int to, char toUpper) {
       if (o instanceof ByteArray) {
            final ByteArray ba=(ByteArray)o;
            filter(type, ba.getBytes(),  maxi(from,ba.getBegin()), mini(to,ba.getEnd()), toUpper);
       } else {
           final byte bb[]=o instanceof byte[] ? (byte[])o : null;
           final CharSequence s=bb!=null ? null : o instanceof CharSequence ? (CharSequence)o : toStrg(o);
           final int t=mini(to,bb!=null ? bb.length : sze(s)), f=maxi(0,from), E=getEnd();
           if (t>f) {
               final byte bytes[]=minBufferSize(E+t-f);
               int count=0;
               for(int i=f; i<t; i++) {
                   final byte c=s!=null ? (byte)s.charAt(i) : bb[i];
                   if (is(type,c)) bytes[E+count++]=  c;
                   else if (toUpper=='_' || toUpper=='-')  bytes[E+count++]= (byte)toUpper;
               }
               setEnd(E+count);
               if (toUpper=='U') for(int i=E+count;  --i>=E;) if ( 'a'<=bytes[i] && bytes[i]<='z')  bytes[i]&=~32;
               if (toUpper=='L') for(int i=E+count;  --i>=E;) if ( 'A'<=bytes[i] && bytes[i]<='Z')  bytes[i]|=32;
           }
       }
       return this;

    }

    public ByteArray append(long num,int width) {
        final int strSize=stringSizeOfInt(num);
        final int size=width>strSize?width:strSize;
        final int end=getEnd();
        final byte[] bb=minBufferSize(end+size);
        for(int idx=end +size-strSize; --idx>=end;) bb[idx]=32;
        itoa(num,bb,end +size-strSize);
        setEnd(end+size);
        return this;
    }

    public ByteArray append(double number,int wInt,int wFrac) {
        final int end=getEnd();
        minBufferSize(end+100);
        ftoa(number,wInt,wFrac,getBytes(),end);
        setEnd(end+wInt+1+wFrac);
        return this;
    }

    private final byte[] minBufferSize(int n) {
        final byte bb[]=this.bytes;
        return this.bytes= bb==null? new byte[n] : n>bb.length ? chSze(bb, n+reserve) : bb;
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */

    public ByteArray replace(ChTextReplacement r) {
        if (r!=null) replace(r.getNeedle(), r.getReplacement());
        return this;
    }
    public ByteArray replace(char c,char replacement) {
        final int b=getBegin(), e=getEnd();
        boolean changed=false;
        final byte[] t=getBytes();
        for(int i=b;i<e;i++) if (t[i]==c){  t[i]=(byte)replacement; changed=true;}
        if (changed)modiCount++;
        return this;
    }
public ByteArray replace(char needle,  char replacement, int from, int to) {
        final byte[] T=getBytes();
        final byte rplc=(byte)replacement;
        final int f=maxi(0,from);
        final int t=mini(to,getEnd());
        boolean changed=false;
        for(int i=f; i<t; i++) {
            if (T[i]==needle) {  T[i]=rplc;  changed=true;  }
        }
        if (changed)modiCount++;
        return this;
    }

    public ByteArray replace(Object needle,  Object replacement) { return replace(needle,replacement,0,MAX_INT); }

    public ByteArray replace(Object oNeedle,  Object replacement, int from, int to) {
        final byte needle[]=oNeedle instanceof byte[] ? (byte[])oNeedle:null;
        final CharSequence sNeedle=needle==null ? toCharSeq(oNeedle): null;
        assertIsText(oNeedle);
        final int L=sze(needle!=null ? needle:sNeedle);
        if (L>0) {
            final byte n0=needle!=null ? needle[0] : (byte)sNeedle.charAt(0);
            final int B=maxi(from,getBegin());
            byte[] T=getBytes();
            for(int i=mini(to,getEnd())-L+1; --i>=B;) {
                if (T[i]==n0 && needle!=null ? equls(needle,T,i) : equls(sNeedle,T,i) ) {
                    replace(i,i+L,replacement);
                    T=getBytes();
                }
            }
        }
        return this;
    }

    /*
    public ByteArray replaceQuotedString(String needle,  String replacement, int fromPos, int to) {
        final int N=needle.length(), R=replacement.length(),
            from=maxi(1,fromPos), B=maxi(from,getBegin());
        final byte[] T=getBytes();
        boolean failed=false;
        for(int i=mini(to, T.length)-N-2; --i>=from;) {
            putln("i="+i+" from="+from);
            if (T[i-1]=='"' && T[i+N]=='"' && equls(needle,T,i)) {
                final int noSpc=nxt(NO_SPC,T,i+N+1, to);
                putln("noSpc="+noSpc);
                if (noSpc-i>R+1) {
                    for(int r=R; --r>=0;) T[r+i]=(byte)replacement.charAt(r);
                    T[i+R]='"';
                    for(int r=i+R+1; r<noSpc; r++) T[r]=' ';
                } else failed=true;
            }
        }
        return this;
        //  return failed ? replace(needle,replacement, from,to) : this;
    }
    */
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */

    public ByteArray foldText(byte[] bb,int from,int to, int lettersPerLine) {
        final int f=maxi(0,from), t=mini(to, bb!=null?bb.length:0), end=getEnd();
        if (t<=f) return this;
        final byte txt[]=minBufferSize(end+to-from+(to-from)/lettersPerLine+2);
        int dest=end;
        for(int i=f, c=0; i<t && dest<txt.length; ) {
            final byte ch=bb[i++];
            if (ch=='\n') c=0;
            else  if (c++>=lettersPerLine) { txt[dest++]=(byte)'\n'; c=0; }
            txt[dest++]=ch;
        }
        setEnd(dest);
        return this;
    }

   public ByteArray onlyLetters(byte[] bb,int from,int to) {
        final int f=maxi(0,from), t=mini(to, bb!=null?bb.length:0), end=getEnd();
        if (t<=f) return this;
        final byte txt[]=minBufferSize(end+to-from+(to-from));
        int dest=end;
        for(int i=f; i<t && dest<txt.length; ) {
            final byte ch=bb[i++];
            if (is(LETTR,ch)) txt[dest++]=ch;
        }
        setEnd(dest);
        return this;
    }

    public ByteArray urlEncode(Object text) { return urlEncode(text,0,MAX_INT);}
    public ByteArray urlEncode(Object text,int from,int to) {
        final int f=maxi(0,from), t=mini(to, sze(text)), end=getEnd();
        if (t<=f || text==null) return this;
        final byte bb[]=text instanceof byte[] ? (byte[])text:null;
        final CharSequence str=text instanceof CharSequence ? (CharSequence)text : null;
        assertNotNull(text);
        byte txt[]=null;
        while(true) {
            int dest=end;
            for(int i=f; i<t;  i++) {
                final byte c=bb!=null ? bb[i] : (byte)str.charAt(i);
                if (is(LETTR_DIGT_US,c)  || c=='-' || c=='.' || c=='*') {
                    if (txt!=null) txt[dest]=c;
                    dest++;
                } else if (c==' ') {
                    if (txt!=null) txt[dest]=(byte)'+';
                    dest++;
                } else {
                    if (txt!=null) { txt[dest]=(byte)'%'; txt[dest+1]= hexDigit(c>>>4) ; txt[dest+2]=hexDigit(c&15); }
                    dest+=3;
                }
            }
            if (txt==null) txt=minBufferSize(dest); else { setEnd(dest); break;}
        }
        return this;
    }

    private static byte hexDigit(int c0) {
        final byte c=(byte)(c0&15);
        return (byte)(c+ (c<10 ? '0' : 'A'-10));
    }

    public ByteArray delSuffix(final String s) {
        if (s!=null) {
            final int e=getEnd(), L=sze(s);
            if (equls(s,getBytes(),e-L)) setEnd(e-L);
        }
        return this;
    }
    public void write(OutputStream os) throws IOException {
        final int b=getBegin(), e=getEnd();
        if (e-b>0) os.write(getBytes(), b, e-b);
    }

    @Override public String toString() {
        final Object[] cached=cached();
        String s=(String)cached[CACHE_TO_STRG];
        if (s==null) cached[CACHE_TO_STRG]=s=newString();
        return s;
    }

    private Object[] cached() {
        Object oo[]=(Object[])deref(refCached);
        if (oo==null) refCached=newSoftRef(oo=new Object[5]);
        return oo;
    }

    public void setLowercaseForcharAt(boolean b) { isLC=b;}
    public char charAt(int i) {
        final int b=get(i+getBegin(),getBytes());
        return (char)(isLC ? b|32 : b);
    }
    public int length() { return maxi(0,getEnd()-getBegin());}
    public ByteArray subSequence(int start, int end) {
        final int b=getBegin(); return new ByteArray(getBytes(), b+start, b+end);
    }

    public byte[] subSequenceAsBytes(int start, int end) {
        final int b=maxi(0,start), e=mini(getEnd(),end);
        if (e-b<=0) return NO_BYTE;
        final byte[] T=getBytes(), bb=new byte[e-b];
        System.arraycopy(T,b, bb,0, e-b);
        return bb;

    }
    public boolean replace(final int from, final  int to, Object oReplacement) {
        final int E=getEnd();
        if (from<0||from>to||to>E || oReplacement==null) return false;
        final byte replacement[]=oReplacement instanceof byte[] ? (byte[])oReplacement:null;
        final CharSequence sReplacement=replacement==null ? toCharSeq(oReplacement): null;
        assertNotNull(oReplacement);
        final int L=sze(replacement!=null ? replacement:sReplacement);
        final int increase=L- to+from;
        final byte txt[]=minBufferSize(getEnd()+increase);
        System.arraycopy(txt,to,txt,to+increase,E-to);
        if (L>0) {
            if (replacement!=null)  System.arraycopy(replacement,0,txt,from,L);
            else if (sReplacement instanceof ByteArray) System.arraycopy( ((ByteArray)sReplacement).getBytes(), ((ByteArray)sReplacement).getBegin(), txt, from, L);
            else {
                for(int i=L; --i>=0;) txt[from+i]=(byte)sReplacement.charAt(i);
            }
        }
        setEnd(E+increase);
        return true;
    }

    @Override public boolean equals(Object o) {
        if (!(o instanceof CharSequence)) return false;
        final CharSequence cs=(CharSequence)o;
        final int L=cs.length();
        final byte[] T=getBytes();
        final int b=maxi(0,getBegin()), e=mini(T.length, getEnd());
        if (L!=e-b) return false;
        for(int i=L; --i>=0; ) {
            if (cs.charAt(i)!=T[b+i]) return false;
        }
        if (o instanceof ByteArray) assert toString().equals(o.toString());
        return true;
    }

    public int compareTo(Object anotherString) {
        if (!(anotherString instanceof CharSequence)) return 0;
        final CharSequence anotherStr=(CharSequence)anotherString;
        int len1 = length();
        int len2 = anotherStr.length();
        int n = mini(len1, len2);
        byte v1[] = getBytes();
        byte v2[] = anotherStr instanceof ByteArray ? ((ByteArray)anotherStr).getBytes() : null;
        int i = getBegin();
        int j = v2!=null ? ((ByteArray)anotherStr).getBegin() : 0;
        if (i == j) {
            int k = i;
            int lim = n + i;
            while (k < lim) {
                byte c1 = v1[k];
                byte c2 = v2!=null ? v2[k] : (byte)anotherStr.charAt(k);
                if (c1 != c2) return c1 - c2;
                k++;
            }
        } else {
            while (n-- != 0) {
                byte c1 = v1[i++];
                byte c2 = v2!=null ? v2[j++] : (byte)anotherStr.charAt(j++);
                if (c1 != c2) return c1 - c2;
            }
        }
        return len1 - len2;
    }
    @Override public int hashCode() {
        if (hashCode==0) hashCode=hashCd(getBytes(), getBegin(), getEnd(),false);
        return hashCode;
    }
    public ByteArray removeLeadingSpace() {
        final byte[] T=getBytes();
        final int B=getBegin(), ends[]=getLineEnds();
        int indent=MAX_INT;
        for(boolean process : FALSE_TRUE) {
            for(int iL=0; iL<ends.length; iL++) {
                final int b=iL==0 ? B : ends[iL-1]+1, e=prev(NO_SPC,T,ends[iL]-1,b-1)+1;
                if (e<=b) continue;
                if (process) {
                    System.arraycopy(T,b+indent,T,b,e-b-indent);
                    for(int i=e-indent;i<e;i++) T[i]=(byte)' ';
                } else {
                    final int noSpc=strchr(STRCHR_NOT,' ',T,b,e);
                    if (indent>noSpc-b) indent=noSpc-b;
                    if (indent==0) return this;
                }
            }
        }
        return this;

    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> ChRunnable >>> */
    public Object run(String id, Object arg) {
        if (id==ChRunnable.APPEND) append(arg);
        return null;
    }

}
