package charite.christo;
import charite.christo.interfaces.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.AttributeSet;
import javax.swing.text.StyledDocument;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
import  javax.swing.text.StyleConstants;
/*
   @author Christoph Gille
*/
public class ChCodeViewer {
    private final static int CLASS=1,  COMMENT=2, KEYWORD=3, STRING=4;
    private final static String KEY_WORDS[]={"package","abstract","do","inner","public","var","boolean","continue","for(","int","return","void","break","else","interface","short","volatile","byvalue","extends","long","static","while","case","final","naive","super","transient","cast","float","new","rest","catch","for","null","synchronized","char","finally","operator","this","class","generic","outer","switch","const","goto","package","throw","double","if","private","true","default","import","protected","try","import"};
      final static String LEFT_CLASS_DELIM=" >\t\n\r()=+,*/&|!;>{}:",  CLASS_DELIM=LEFT_CLASS_DELIM+".[]";
    private ChCodeViewer(){}

    private static SimpleAttributeSet _as[];
    private final static SimpleAttributeSet as(int t) {
        if (_as==null) _as=new SimpleAttributeSet[5];
        for(int i=5; --i>=0;) {
            final SimpleAttributeSet a=_as[i]=new SimpleAttributeSet();
            StyleConstants.setBackground(a, C(i==COMMENT ? 0xFF :  0xFFffFF));
            StyleConstants.setForeground(a, C(i==CLASS ? 0xAA : i==COMMENT ? 0xFFffFF : i==KEYWORD ? 0xaa00aa : 0x007c00));
            if (i==KEYWORD) StyleConstants.setBold(a,true);
            if (i==CLASS) StyleConstants.setUnderline(a,true);
        }
        return _as[t];
    }
    private static void s(StyledDocument doc,StringBuffer sb,int from,int to,int lTxt,AttributeSet as){
        if (doc!=null && as!=null)
            try{
                doc.setCharacterAttributes(from,to-from,as,true);
            }catch(Exception e){}
        if (sb!=null) {
            final int longer=sb.length()-lTxt;
            Object ins1="<font>", ins2="</font>";
            if (as==as(KEYWORD)) { ins1="<b>"; ins2="</b>"; }
            if (as==as(COMMENT))   ins1="<font color=\"#4444AA\">";
            if (as==as(STRING))    ins1="<font color=\"#00AA00\">";
            sb.insert(to+longer,ins2);
            sb.insert(from+longer,ins1);
        }
    }
    public static StringBuffer highlightSyntax(ByteArray txt,String[] vPackages) {
        StringBuffer sb=new StringBuffer(9999);
        appnd(txt.getBytes(),0,txt.getEnd(),sb);
        highlightSyntax(txt,null,sb,vPackages);
        return sb;
    }
    public static void highlightSyntax(ByteArray ba,StyledDocument doc) {
        highlightSyntax(ba,doc,null,null);
    }
    private static void highlightSyntax(final ByteArray ba,final StyledDocument doc,final StringBuffer sb, final String[] packages) {
        final HashSet<String> setNoClass=new HashSet();
        final Map<String,String> mapFullClassName=new HashMap();
        final byte txt[]=ba.getBytes();
        final int lTxt=ba.getEnd();
        final String[] imported;
        {
            final String pp[]=getImportedPackages(ba);
            if (packages!=null) {
                imported=new String[pp.length+packages.length];
                System.arraycopy(pp,0,imported,0,pp.length);
                System.arraycopy(packages,0,imported,pp.length,packages.length);
            } else imported=pp;
        }
        /* --- multi line comment --- */
        int multiComment=-1,singleComment=-1,string=-1,aClass=-1,possibleKeyword=-1,classWithPackage=-1,countSlash=0;
        for(int i=-1;i<lTxt-1;i++) {
            final int c0= i<0 ? '\n' : txt[i],c1=txt[i+1];
            if (multiComment>=0 && c0=='*' && txt[i+1]=='/') {
                s(doc,sb,multiComment,i+2,lTxt,as(COMMENT));
                multiComment=-1;
            }
            else if (singleComment>=0 && c0=='\n') {
                s(doc,sb,singleComment,i,lTxt,as(COMMENT));
                singleComment=-1;
            }
            else if (string>=0 && c0=='"') {
                if ( (countSlash&1)==0) {
                    s(doc,sb,string,i+1,lTxt,as(STRING));
                    string=-1;
                    continue;
                }
            }
            else if (aClass>=0 && " =()[.\"/ ;".indexOf(c0)>=0) {
                final int start=classWithPackage>=0 ? classWithPackage : aClass;
                final String sClass=new String(txt,start+1,i  -start-1);
                if (!setNoClass.contains(sClass)) {
                    String fullClassName=mapFullClassName.get(sClass);
                    if (fullClassName==null) mapFullClassName.put(sClass, fullClassName=fullClassName(sClass,imported));
                    if (fullClassName==null) setNoClass.add(sClass);
                    else {
                        if (sb!=null) {
                            final CharSequence javadoc=ChJavadoc.javadocUrl(fullClassName);
                            if (javadoc==null) debug(" fullClassName="+fullClassName);
                            final int d=sb.length()-lTxt;
                            sb.delete(d+start+1,d+i);
                            sb.insert(d+start+1,javadoc!=null ? "<a href=\""+javadoc+"\">"+sClass+"</a>" : "<font color=\"#AA0000\">"+sClass+"</font>");
                        }
                        s(doc,null,start+1,i,lTxt,as(CLASS));
                    }
                }
                aClass=-1;
            }
            else if (possibleKeyword>=0 && " =().\"/ ;,[]".indexOf(c0)>=0) {
                if (c0==' ' || c0=='(' || c0=='{') {
                    final String kw=new String(txt,possibleKeyword,i-possibleKeyword);
                    if (c0==' ' || c0=='(' && ( "if".equals(kw) || "for".equals(kw) || "while".equals(kw) ) || c0=='{' && "try".equals(kw)) {
                        for (String KW : KEY_WORDS) {
                            if (KW.equals(kw)) {
                                s(doc,sb,possibleKeyword,i,lTxt,as(KEYWORD));
                                break;
                            }
                        }
                    }
                }
                possibleKeyword=-1;
            }
            if (aClass>=0 && " =()\"/ ;".indexOf(c0)>=0) classWithPackage=-1;
            if (sb!=null) {
                final String r=c0=='<' ? "&lt;" : c0=='>' ? "&gt;" : c0=='&' ? "&amp;" : null;
                if (r!=null) {
                    final int d=sb.length()-lTxt;
                    sb.delete(d+i,d+i+1);
                    sb.insert(d+i,r);
                }
            }
            if (multiComment<0 && singleComment<0 && string<0) {
                if (" =()".indexOf(c0)>=0 && ('A'<=c1 && c1<='Z' || 'a'<=c1 && c1<='z')) classWithPackage=i;
                if ("\n =()".indexOf(c0)>=0 && 'a'<=c1 && c1<='z') possibleKeyword=i+1;
                if (c0=='/' && c1=='*') {multiComment=i; continue;}
                if (c0=='/' && c1=='/') {singleComment=i; continue;}
                if (c0=='"' && (countSlash&1)==0) { string=i; continue;}
                if (" =().".indexOf(c0)>=0 && 'A'<=c1 && c1<='Z') {
                    aClass=i;
                    continue;
                }
            }
            if (c0=='\\') countSlash++; else countSlash=0;
        }
    }

    public static String[] getImportedPackages(ByteArray ba) {
        String pp[]=null;
        while(true) {
            final int  ends[]=ba.getLineEnds();
            final byte[] txt=ba.getBytes();
            int count=0;
            nextLine:
            for(int iL=0; iL<ends.length; iL++) {
                final int b= iL==0 ? 0 : ends[iL-1]+1;
                final int e=ends[iL];
                if (e-b<6) continue;
                final int c0=txt[b];
                if (c0!='i' && c0!='p') {
                    if (c0!=' ' && c0!='\t')  break nextLine;
                    continue;
                }
                final int semicolon=strchr(';', txt,b,e);
                if (semicolon<0) continue;
                if (c0=='p' && strstr("package ",txt,b,e)==b) {
                    int next=b+8; while(next<e && ( txt[next]==' ' || txt[next]=='\t')) next++;
                    if (semicolon>next) {
                        if (pp!=null) pp[count]=new String(txt,next,semicolon-next)+'.';
                        count++;
                    }
                }
                if (c0=='i' && strstr("import ",txt,b,e)==b) {
                    int next=b+7; while(next<e && txt[next]==' ') next++;
                    if (strstr("static ",txt,next,e)!=next && semicolon>next) {
                        final int to=txt[semicolon-1]=='*' ? semicolon-1 : semicolon;
                        if (to>next) {
                            if (pp!=null) pp[count]=new String(txt,next,to-next);
                            count++;
                        }
                    }
                }
            }
            if (pp==null) pp=new String[count]; else break;
        }
        return pp;
    }

    public static void main(String... argv) {
        for(String arg :argv) {
            final File f=file(arg);
            final ByteArray src=readBytes(f);
            if (src==null) { return; }
            final StringBuffer sb=ChCodeViewer.highlightSyntax(src,(String[])null);
            sb.append("\n</pre></body></html>").insert(0,"<html><title>"+f.getName()+"</title>\n<body>\n<pre>\n");
            wrte(file(arg+".html"),sb);
        }
    }

}
