package charite.christo;
import charite.christo.interfaces.*;
import java.io.*;
import java.awt.*;
import javax.swing.JComponent;
import java.util.List;
import java.util.*;
import java.lang.ref.Reference;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**
   execute external programs
   Several alternative command lines can be set.
   They are probed until one succeeds.
   The trying of command lines stops at an empty command line.
   The method error(Exception) is called when non of the alternative command lines succeeds.
   @author Christoph Gille
*/
public final class ChExec implements Runnable,HasControlPanel,java.awt.event.ActionListener,ChRunnable {
    public final static long STDOUT=1<<0, STDERR=1<<1,IGNORE_ERROR=1<<2, CYGWIN_DLL=1<<3, SHOW_STREAMS=1<<4,
        LOG=1<<5, STDOUT_IN_TEXT_BOX=1<<6, PLEASE_INSTALL_MANUALLY=1<<7, EXEC_WITHOUT_ASKING=1<<8;
      final static String WINDOWS_addBinToPath=
        "export PATH=/usr/local/bin:/usr/bin:/bin:/usr/X11R6/bin:/usr/local/bin:/usr/bin:/bin:/usr/X11R6/bin; "+
        "export DISPLAY=localhost:0;\n";

    public final static String ACTION_ERROR="ACTION_ERROR", ACTION_TERMINATED="ACTION_TERMINATED";
    private final static int OUT=0, ERR=1, STATUS=2, KILL=3;
    private final static int OUTandERR=3;

    private String icon, label="Status";
    /* >>> Instance >>> */
    private Object refProcess;
    private boolean hasFinished,couldLaunch;
    private long mode;
    public ChExec(long mode) {setOptions(mode);}
    public ChExec setOptions(long m) {
        this.mode=m;
        if ( 0!=(mode&STDOUT_IN_TEXT_BOX)) m|=STDOUT;
        if ( (m&STDERR)!=0) vBytes[ERR]=WANT_DATA;
        if ( (m&STDOUT)!=0) vBytes[OUT]=WANT_DATA;
        if ( (m&SHOW_STREAMS)!=0) showStdoutAndStderr(null);
        return this;
    }
    public boolean hasFinished(){return hasFinished;}
    public boolean couldNotLaunch(){return hasFinished && !couldLaunch;}
    private int exitValue=Integer.MIN_VALUE;
    public int exitValue() { return  exitValue;}
    @Override protected void finalize() throws Throwable{
        logFinalize(ANSI_GREEN+" ChExec ","");
        super.finalize();
    }
    public void kill() {
        final Process p=(Process)deref(refProcess);
        if (p!=null) try {
            p.destroy();
            handleActEvt(this,ACTION_TERMINATED,0);
        } catch(Exception e){ }
    }
    private boolean[] isRunning;
    public boolean isRunning() { return get(0,isRunning); }
    public void dispose() {
        closeStrm(bufStdin);
        kill();
        bufStdin=null;
        userDir=null;
        sbStatus.setEnd(0);
        vBytes[0]=vBytes[1]=null;
        ChDelay.invokeAndWait(this,"DISPOSE",null);
    }
    /* <<< Instance <<< */
    /* ---------------------------------------- */
    /* >>> Log  >>> */
    private static Writer fwLog;
    public ChExec log(Object... oo0) {
        Object[] oo=oo0;
        showStdoutAndStderr(null);
        final AsciiView ta=get(OUTandERR,refTA, AsciiView.class);
        if (sze(oo)>0) {
            for (int i=oo.length; --i>=0;) {
                if (oo[i] instanceof Customize) {
                    this.customize=(Customize)oo[i];
                    (oo=oo.clone())[i]=null;
                }
            }
            //log("\nSETTINGS: ",CYAN,this.customize,"\n",joinLines(this.customize.getSettings()),"\n");
            for(Object o : oo) if (o!=null) ta.appendS(toCharSeq(o));
        } else ta.setText("");
        return this;
    }
    private final ByteArray sbStatus=new ByteArray(222);
    private void status(AsciiView ta, CharSequence... oo) {
        if (ta!=null) ta.appendSS(oo);
        status(false,oo);
    }

    private void status(boolean toStdout, CharSequence... oo) {
        if (sze(oo)==0) return;
        for(CharSequence o : oo) this.sbStatus.append(o);
        if (toStdout) putln(oo);
        final AsciiView ta=get(STATUS,refTA, AsciiView.class);
        if (ta!=null) ta.repaint();
    }
    private AsciiView hardReferenceOutAndErr;
    public ChExec showStdoutAndStderr(ChFrame f) {
        if (f!=null) fOutErr=f;
        AsciiView ta=get(OUTandERR,refTA, AsciiView.class);
        if (ta==null) {
            ta=this.hardReferenceOutAndErr=new AsciiView(999);

            this.refTA[OUTandERR]=weakRef(ta);
        }
        mode=mode|STDOUT|STDERR;
        return this;
    }
    /* <<< Log <<< */
    /* ---------------------------------------- */
    /* >>> Environment  >>> */
    private List<String> vEnvironement;
    public ChExec addToEnvironement(Object o) {
        if (o!=null) {
            List<String> v=vEnvironement;
            if(v==null) v=vEnvironement=new ArrayList();
            if (o instanceof Map) {
                Map m=(Map)o;
                for(Map.Entry e : entryArry(m)) {
                    if (e.getValue()!=null) adUniq(e.getKey()+"="+e.getValue(),v);
                }
            } else adUniq(toStrg(o), v);
        }
        return this;
    }
    /* <<< Environment <<< */
    /* ---------------------------------------- */
    /* >>> Command Lines >>> */
    private String commandLinesV[][];
    private Customize customize;
    public ChExec setCommandLine(Object s) { return setCommandLines(new String[]{toStrg(s)}); }
    public ChExec setCommandLines(String[] lines) {
        this.commandLinesV=new String[lines.length][];
        for(int i=0;i<lines.length;i++) {
            final String ss[]=splitTokns(0L, lines[i]);
            for(int j=ss.length; --j>=0;) ss[j]=rplcToStrg("%20"," ",ss[j]);
            if (ss.length>0) ss[0]=rplcEnvInPath(ss[0]);
            this.commandLinesV[i]=ss;
        }
        return this;
    }
    public ChExec setCommandsAndArgument(String[] commands,Object argument0) {
        Object argument=argument0;
        if (argument==null) return setCommandLines(commands);
        final String lines[]=new String[commands.length];
        for(int i=0;i<commands.length;i++) {
            final String s=rplcEnvInPath(commands[i]);
            if (sze(s)==0) continue;
            if (s.startsWith("CYGWINSH") && argument instanceof File) argument=Microsoft.toCygwinPath(argument);
            String sArgument=toStrg(argument);
            if (sArgument!=null && sArgument.startsWith("file:/") && !sArgument.startsWith("file://")) sArgument="file://"+sArgument.substring(5);
            lines[i]=s.indexOf('*')>=0 ?  rplcToStrg("*",sArgument,rplcToStrg("$*",sArgument,s)) : s+" "+sArgument;
        }
        return setCommandLines(lines);
    }
     /**
       null is skipped, for List and Arrays the elements are inserted.
       If arg[0] is   "CYGWINSH" all files are written in the style /cygdrive/c/...
    */
    public ChExec setCommandLineV(Object... objects) {
        final List v=new ArrayList(7);
        for(Object o : objects) {
            if (o==null) continue;
            if (o instanceof Collection) v.addAll((Collection)o);
            else if (o instanceof Object[]) adAll( (Object[])o, v);
            else {
                if (o instanceof File && "CYGWINSH".equals(objects[0])) o=Microsoft.toCygwinPath(o);
                v.add(o);
            }
        }
        commandLinesV=new String[][]{toStringArray(v.toArray())};
        return this;
    }
    /* <<< Command Lines <<< */
    /* ---------------------------------------- */
    /* >>> Streams >>> */
    private final static byte NL[]={(byte)'\n'};
    private final static List<byte[]> WANT_DATA=new ArrayList(0);
    private final List<byte[]> vBytes[]=new ArrayList[2];
    private ChRunnable printlnStdout;
    private final boolean stdoutDone[]={false};
    private String inString;
    private BufferedOutputStream bufStdin;
    public ChExec printlnStdout(ChRunnable a) { printlnStdout=a;return this;}
    private static void readStream(InputStream is,boolean closed[], boolean isRunning[],
                                    ArrayList<byte[]> vBytes[], Reference<AsciiView> refLog[], final ChRunnable lines,
                                    Reference<ChButton> refButs[], int outOrErr) {
        if (is==null) return;
        final ChInStream br=lines!=null ? new ChInStream(is)  : null;
        final ByteArray LINE=lines!=null ? new ByteArray(99) : null;
        stop:
        while(true) {
            try{
                final String line;
                final byte[] appnd;
                if (lines!=null) {
                    appnd=null;
                    if (!br.readLine(LINE.setEnd(0))) break stop;
                    lines.run(ChRunnable.INTERPRET_LINE,line=LINE.newString());
                } else {
                    line=null;
                    int available=0;
                    while ((available=is.available())==0) {
                        if (!isRunning[0] && is.available()==0) break stop;
                        Thread.sleep(22);
                    }
                    final byte bb[]=new byte[available];
                    final int nRead=is.read(bb,0,available);
                    if (nRead==0) {
                        try { Thread.sleep(22); } catch(Exception e) {}
                        continue;
                    }
                    if (strstr(0L,"failed to open display",bb,0,nRead)>0) {
                        error("There might be problem running an external program.<br>"+
                              "It may help to deactivate display control by typing into a command shell"+
                              "<pre class=\"terminal\"> xhost +</pre>");
                    }
                    appnd=nRead==bb.length ? bb:chSze(bb,nRead);
                }
                if (appnd!=null && appnd.length>0 || line!=null) {
                    ChDelay.setEnabled(get(outOrErr,refButs,JComponent.class), true,333);
                    if (vBytes[outOrErr]==WANT_DATA) vBytes[outOrErr]=new ArrayList(2);
                    final List<byte[]> vB=vBytes[outOrErr];
                    if (vB!=null) {
                        if(appnd!=null) vB.add(appnd);
                        if (line!=null) { vB.add(line.getBytes()); vB.add(NL); }
                    }
                    for(int i=refLog.length; --i>=0;) {
                        final AsciiView l=get(i,refLog, AsciiView.class);
                        if (l!=null && (i==OUTandERR || i==outOrErr)) {
                            if (appnd!=null) l.appendBB(appnd);
                            if (line!=null) l.appendS(line).appendS("\n");
                        }
                    }
                }
            } catch(Exception ex) {
                putln(RED_CAUGHT_IN+"ChExec#readStream() ",ex);
                stckTrc(ex);
                if (!isRunning[0])  break stop;
            }
            try { Thread.sleep(22); } catch(Exception e) { break stop; }
        }
        if (closed!=null) closed[0]=true;
        closeStrm(is);
    }
    public ChExec setStdin(String s) { inString=s; return this; }
    public Object run(String id,Object arg) {
        final Object argv[]=arg instanceof Object[] ? (Object[])arg : null;
        if (id=="SHW_F") {
            final AsciiView ta=this.hardReferenceOutAndErr;
            if (ta!=null) {
                ChFrame f=get(OUTandERR,refFrames, ChFrame.class);
                if (f==null) {
                    if (this.fOutErr!=null) { (f=this.fOutErr).getContentPane().removeAll();}
                    else addActLi(this, f=new ChFrame("Execute external program"));
                    refFrames[OUTandERR]=weakRef(f);
                    f.ad(newCNSEW(scrllpn(0L,ta), customize!=null ? pnl(customize.newButton()) : null, getControlPanel(),null,null)).shw(CLOSE_ON_CtrlW|ChFrame.STAGGER);
                }
                f.size(600,300).shw(0);
            }
        }
        if (id=="SAY_FAILED") {
            status(false,ANSI_RED+"\nFailed"+ANSI_RESET+"\n");
            disableKill("Failed",IC_TRAFFICLIGHTRED);
            handleActEvt(this,ACTION_ERROR,0);
        }

        if (id=="SHOW_OUT") {
            final ByteArray ba=getStdoutAsByteArray();
            if (sze(ba)>0) new AsciiView(ba).showInFrame("stdout");
        }
        if (id=="DISPOSE") {
            for(Object r:refFrames) dispos(0L, r);
        }
        if (id=="STREAM") {
            Process pp[]=(Process[])argv[0];
            boolean[] closed=(boolean[])argv[1];
            boolean[] isRunning=(boolean[])argv[2];
            ArrayList[] vBytes=(ArrayList[]) argv[3];
            final Reference<AsciiView>[] refLogs=(Reference[]) argv[4];
            ChRunnable lines=(ChRunnable)argv[5];
            Reference<ChButton> refB[]=(Reference[])argv[6];
            int O_or_E="O"==argv[7] ? OUT : ERR;

            while(pp[0]==null) {
                try { Thread.sleep(99); } catch(Exception e) {}
                if (couldNotLaunch()) return null;
            }
            Process p=pp[0];
            if (p!=null) {
                InputStream is=O_or_E==OUT ? p.getInputStream() : p.getErrorStream();
                if (!(is instanceof BufferedInputStream)) is=new BufferedInputStream(is);
                readStream(is,closed,isRunning,vBytes,refLogs,lines, refB, O_or_E);
            }
        }
        if (id=="WRITE") threadWriteToProcess((String)argv[0],(boolean[])argv[1]);
        return null;
    }
    private void threadWriteToProcess(String inString, boolean isRunning[]) {
        Process p=null;
        while(p==null) {
            try {  Thread.sleep(99); } catch(Exception e) {}
            if (!isRunning[0]) return;
            p=(Process)deref(refProcess);
        }
        if (inString==null || p==null) return;
        final BufferedOutputStream os=new BufferedOutputStream(p.getOutputStream());
        try{
            Thread.sleep(99);
            os.write(inString.getBytes());
        } catch(Exception e){ putln("caught ChExec#writeToProcess ", e);}
        closeStrm(os);
    }
    public ChExec write(String s) throws IOException {
        BufferedOutputStream os=bufStdin;
        final Process p=(Process)deref(refProcess);
        if (p!=null && isRunning()) {
            if (os==null) os=bufStdin=new BufferedOutputStream(p.getOutputStream());
            os.write(s.getBytes());
            os.flush();
        }
        return this;
    }
    public byte[] getStderr() { return bytesToBytes(this.vBytes[ERR]);}
    public byte[] getStdout() {
        if (couldNotLaunch()) return null;
        List v= waitForStdout();
        return couldNotLaunch() ? null : bytesToBytes(v);
    }
    public ByteArray getStdoutAsByteArray() {
        final byte[] stdout=getStdout();
        return stdout!=null ? new ByteArray(getStdout()) : null;
    }
    private List waitForStdout() {
        if (vBytes[OUT]==null) return null;
        try {
            while(!stdoutDone[0] && !couldNotLaunch()) { Thread.sleep(55); }
        } catch(InterruptedException e){}
        return vBytes[OUT];
    }
    /* <<< Streams <<< */
    /* ---------------------------------------- */
    /* >>> Files >>> */
    private File userDir;
    public ChExec dir(File f) { userDir=f; return this;}
    /* <<< Files <<< */
    /* ---------------------------------------- */
    /* >>> Thread >>> */
    public void run() {
        if (hardReferenceOutAndErr!=null) ChDelay.invokeLater(this,"SHW_F",null);
        couldLaunch=hasFinished=false;
        final File dir=userDir!=null ? userDir : dirWorking();
        /*
        if (dir!=null) {
            status(false,"\n  cd "+dir+"\n");
            if (isWindows()) status(false,"  cd ",Microsoft.toCygwinPath(dir),"\n");
        }
        if (isWindows() && (mode&(CYGWIN_DLL))!=0) addToEnvironement("PATH="+file(dirCygwin()+"\\bin"));
        for(String s : strgArry(vEnvironement)) {
            final int eq=s!=null ? s.indexOf('=')+1 : 0;
            if (eq>0) {
                final String key=s.substring(0,eq).intern(),value=s.substring(eq).intern();
                status(false," export ",key,"\"",value,"\"\n");
                if (isWindows()) status(false,"  set ",key,"\"",value,"\"\n");
            }
        }
        */
        boolean ok=false;
        for(int iTry=0; !ok && iTry<sze(commandLinesV); iTry++) {
            if (ok=runLine(commandLinesV[iTry],dir)) handleActEvt(this,ACTION_TERMINATED,0);
            else try { Thread.sleep(222); } catch(Exception e) {}
        }
        if (!ok) ChDelay.invokeLater(this,"SAY_FAILED",null);
        if (0!=(mode&STDOUT_IN_TEXT_BOX)) ChDelay.invokeLater(this,"SHOW_OUT",null);
    }
    /* <<< Thread <<< */
    /* ---------------------------------------- */
    /* >>> Exec >>> */
    private boolean runLine(String execv0[],File dir) {
        final boolean LOG= (mode&ChExec.LOG)!=0;
        String[] execv=execv0;
        if (execv==null || execv.length==0 || chrAt(0,execv[0])=='#') return false;
        if ("CYGWINSH".equals(execv[0])) {
            final String[] execvNew={dirCygwin()+"\\bin\\sh.exe","-c",""};
            final StringBuffer sb=new StringBuffer(111).append(WINDOWS_addBinToPath);
            for(int i=1;i<execv.length;i++) sb.append(execv[i]).append(" ");
            execvNew[2]=sb.toString();
            execv=execvNew;
        }
        if ("SH".equals(execv[0])) {
            String[] execvNew={isWindows() ? Microsoft.cmd() : "sh" , isWindows() ? "/c" : "-c", ""};
            StringBuffer sb=new StringBuffer(111);
            for(int i=1;i<execv.length;i++) sb.append(execv[i]).append(' ');
            execvNew[2]=sb.toString();
            execv=execvNew;
        }
        final AsciiView ta=this.hardReferenceOutAndErr;
        try {

            int commandLength=0;
            for(int i=0;i<execv.length;i++) {
                String s=execv[i];
                //if (replacementForAsterisk!=null) s=rplc("*",replacementForAsterisk,s);
                s=rplcToStrg("STRAP_COLOR_DEPTH",String.valueOf(Toolkit.getDefaultToolkit().getColorModel().getPixelSize()),s);
                if (isWindows()) {
                    if (s.startsWith("?:")) s=processFileName(s);
                    final File dc=dirCygwin();
                    if (dc!=null) s=rplcToStrg("DIR_CYGWIN",toStrg(dc),s);
                }
                execv[i]=s;
                commandLength+=s.length();
            }
            if (commandLength==0) return false;
            try {
                for(int i=2; --i>=0;) {
                    if ((i==0 ? "ECLIPSE":"J_JAVA_EDITOR").equals(execv[0])) {
                        boolean success=false;
                        for(int iF=execv.length;--iF>=1;) success=success||(i==0 ? new FileEditorEclipse() : new J_JavaEditor()).editFile(file(execv[iF]),-1);
                        return success;
                    }
                }
            } catch(Exception e){ stckTrc(e);}
            final String env[]=strgArry(vEnvironement);
            //if (ta!=null) ChDelay.invokeLater(this,"SHW_F",null);

            final StringBuffer sb=new StringBuffer(999);
            join(execv,", ",sb.append("\n[")).append("]\n  Dir: ").append(filepathUnixStyle(dir)).append("\n\n");
            if (sze(env)>0) join(env,", ",sb.append("]\n  Environment: [")).append("]\n");
            final String info=sb.toString();
            if (ta!=null) ta.appendSS(info);
            Writer fw=fwLog;
            if (fw==null) {
                try {
                    fw=fwLog=fWriter(dirLog()+"/exec.log");
                    fw.write("Programs started with  java.lang.Runtime exec(String[] )\n\n");
                } catch(IOException ioex) {stckTrc(ioex);}
            }
            if (fw!=null) {  fw.write(info);  fw.flush(); }
            if (LOG) putln("ChExec  cd ",dir);
            final String home=systProprty(SYSP_USER_HOME);
            for(int i=0; i<execv.length; i++) {
                String s=execv[i],tilde="";
                final boolean needsQuotes=nxt(NO_FILEP,s)>=0;
                if (LOG) puts(needsQuotes ? " \"": " ");
                if (LOG) puts(s);
                if (!isWindows() && s.startsWith(home)) s=filepathUnixStyle(s);
                if (needsQuotes && LOG) puts("\"");
                status(false,needsQuotes ? "\"" : "",tilde,s, needsQuotes ? "\"  " : "  ");
            }

            status(LOG,"\n");
            final Process pp[]={null};
            if (this.isRunning!=null) this.isRunning[0]=false;
            isRunning=new boolean[]{false};
            ChDelay.newThread(this,"STREAM",new Object[]{pp,null           ,isRunning, vBytes, this.refTA, null,         this.refButtons, "E"}).start();
            ChDelay.newThread(this,"STREAM",new Object[]{pp,this.stdoutDone,isRunning, vBytes, this.refTA, printlnStdout,this.refButtons, "O"}).start();
            if (inString!=null) { ChDelay.newThread(this,"WRITE",new Object[]{inString,isRunning}).start(); inString=null;}
            if (LOG && debug(ANSI_MAGENTA+"!"+ANSI_RESET+"Going to call exec ")) putln(execv);

            final String envNull[]=env.length>0?env:null;
            final String explain=ChExecSecurity.describeExec(execv,envNull,dir, null);
            putln(explain);
            status(false, explain);

            final boolean ask=isGraphicsAvailable() && 0==(mode&ChExec.EXEC_WITHOUT_ASKING);
            final Process process=pp[0]=
                ask ? ChExecSecurity.runtimeExecS(execv,envNull,dir) :
                Insecure.runtimeExec(execv,envNull,dir);
            refProcess=weakRef(process);
            isRunning[0]=couldLaunch=true;

            if (process==null) return false;

            status(LOG, " # Running ...\n");
            this.icon=IC_TRAFFICLIGHTGREEN;
            final ChButton butStatus=get(STATUS,refButtons,ChButton.class);
            if (butStatus!=null) butStatus.i(IC_TRAFFICLIGHTGREEN);
            try { process.waitFor(); } catch(Exception e) { putln("caught ChExec waitFor() ",e);}
            isRunning[0]=false;
            exitValue=process.exitValue();

            rmFromParent(deref(refCenter));
            if (fw!=null)  try {fw.write(" Exit value" +exitValue);} catch(IOException ioex){}
            status(ta," # WIKI:Exit_status ="+exitValue+"\n");
            if (exitValue!=0) {
                if (! (execv.length>1 && (execv[0]!=null && execv[0].endsWith("dialign-t.exe")||execv[1]!=null && execv[0].endsWith("dialign-t.exe") ))) return false;
                disableKill("terminated != 0","");
                if (isWindows() && (mode&(CYGWIN_DLL))!=0) {
                    final File f=file(dirCygwin()+"\\bin");
                    if (!f.isDirectory()) {
                        if (msgCygwin==null) { msgCygwin=vBhB("To run this external program, you need to install cygwin!","WIKI:Cygwin for details ",Microsoft.newCygwinButton(null));
                            error(0L,"installCygwin",msgCygwin);
                        }
                    }
                }
            }
            disableKill("success",IC_HAPPY);
            return true;
        } catch(IOException e){
            final String nr="Failed to run the program ";
            status(ta,  this.couldLaunch ? nr : "\nterminated\n"+e+"\n");
            if (0!=(mode&PLEASE_INSTALL_MANUALLY)) error(nr+get(0,execv)+"<br>Please install manually.");
            return false;
        }
        finally {
            final boolean isR[]=isRunning;
            if (isR!=null) isR[0]=false;
            hasFinished=true;
            closeStrm(bufStdin);
        }
    }
    public ChExec waitFor(){
        try {
            while(!isRunning() && !hasFinished() && !couldNotLaunch()) { Thread.sleep(99); }
        } catch(InterruptedException ex){}
        return this;
    }
    /* <<< Exec <<< */
    /* ---------------------------------------- */
    /* >>> GUI >>> */
    private final Object[] refTA=new Reference[4], refButtons=new Reference[4],  refFrames=new Reference[4];
    private ChFrame fOutErr;
    private static JComponent msgCygwin;
    private Object refCenter;
    private void disableKill(String status,String icon) {
        final ChButton bK=get(KILL,refButtons,ChButton.class);
        final ChButton butStatus=get(STATUS,refButtons,ChButton.class);
        if (bK!=null) {
            rmFromParent( derefArray(gcp(KEY_CLONES,bK), ChButton.class));
            rmFromParent(bK);
            bK.setEnabled(false);
        }
        this.label=status;
        this.icon=icon;
        if (butStatus!=null) butStatus.t(this.label).i(this.icon);
    }
    /* <<< GUI <<< */
    /* ---------------------------------------- */
    /* >>> Event >>> */
    public void actionPerformed(java.awt.event.ActionEvent ev) {
        for(int i=0; i<4;i++) {
            if (refButtons[i]==null || ev.getSource()!=get(i,refButtons)) continue;
            if (i==KILL) kill();
            else if (i==OUT || i==ERR || i==STATUS) {
                AsciiView ta=(AsciiView)get(i,refTA);
                if (ta==null) {
                    ta=i==STATUS ? new AsciiView(sbStatus) : new AsciiView(999);
                    refTA[i]=weakRef(ta);
                    if (i==OUT || i==ERR)  {
                        final List<byte[]> v=this.vBytes[i];
                        for(int j=0; j<sze(v); j++) {
                            ta.appendBB(v.get(j));
                        }
                    }
                }

                final Container ctrlPnl=(Container)deref(refCtrl);
                final Window window= parentWndw(ctrlPnl);
                if (isRunning() && window!=null) {
                    if (window.getHeight()<500) window.setSize(window.getWidth(),500);
                    window.show();
                    JComponent center=deref(refCenter,JComponent.class);
                    if (center!=null) ctrlPnl.remove(center);
                    refCenter=weakRef(center=ta.tools().scrollPane());
                    ctrlPnl.add(center,BorderLayout.CENTER);
                    center.revalidate();
                    repaintOnMouseEnterAndExit(false,center);
                    center.paintImmediately(0,0,99999,99999);
                } else {
                    ChFrame f=(ChFrame)get(i,refFrames);
                    if (f==null) {
                        f=new ChFrame(i==OUT ? "Stdout" : i==ERR ? "Stderr" : "Status");
                        this.refFrames[i]=weakRef(f);
                        addActLi(this, f);
                        f.ad(ta.tools().scrollPane()).shw(ChFrame.DISPOSE_ON_CLOSE|ChFrame.AT_CLICK|ChFrame.SIZE_REGARD_RESOLUTION);
                    }
                    f.shw(ChFrame.AT_CLICK);
                }
            }
        }
    }
    /* <<< Event <<< */
    /* ---------------------------------------- */
    /* >>> Control Panel >>> */
    private Object refCtrl;
    public Container getControlPanel() {
        Container ctrl=(Container)deref(refCtrl);
        if (ctrl==null) {
            final ChButton bb[]=new ChButton[4];
            for(int i=0;i<4;i++) {
                if (i==KILL && hasFinished() || (i==OUT || i==ERR) && vBytes[i]==null) continue;
                bb[i]=get(i,refButtons,ChButton.class);
                if (bb[i]==null) {
                    final ChButton b=bb[i]=new ChButton().li(this).cp(KEY_AVOID_GARBAGE_COLLECT,this);
                    refButtons[i]=weakRef(b);
                    if (i==STATUS) b.t(label).i(icon).tt("Status of the process");
                    else if (i==KILL) b.t("kill").tt("Kill the process");
                    else if (i==OUT)  b.t("stdout").tt("Standard output");
                    else if (i==ERR)  b.t("stderr").tt("Standard error output");
                    if (i==OUT || i==ERR) b.setEnabled(0<sze(vBytes[i]));
                } else bb[i]=bb[i].cln();
            }
            ctrl=newCNSEW(null,pnl(bb),null,null,null);
            refCtrl=weakRef(ctrl);
        }
        return ctrl;
    }
    /* <<< Control Panel <<< */
    /* ---------------------------------------- */
    /* >>> Static utils >>> */
    private static byte[] bytesToBytes(List v) {
        if (v==null) return null;
        if (v==WANT_DATA) return NO_BYTE;
        byte BB[]=null;
        final Object oo[]=v.toArray();
        if (oo.length==1 && oo[0] instanceof byte[]) return (byte[])oo[0];
        while(true) {
            int count=0;
            for(Object o : oo) {
                if (o instanceof byte[]) {
                    final byte[] bb=(byte[])o;
                    if (BB!=null) System.arraycopy(bb,0,BB,count,bb.length);
                    count+=bb.length;
                }
            }
            if (BB!=null) break; else BB=new byte[count];
        }
        return BB;
    }

}
