package charite.christo;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.awt.image.RGBImageFilter;
import java.awt.image.ImageObserver;
import java.awt.image.FilteredImageSource;
import java.net.URL;
import java.util.Map;
import javax.swing.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
import static javax.swing.plaf.metal.MetalIconFactory.*;

/**HELP

Image icons are a visual aid to recognize proteins and residue
annotations. They are associated to proteins or annotations by WIKI:Drag_and_drop
(See  http://www.bioinformatics.org/strap/dragProteinLink.html).
Supported formats are gif, png, bmp and jpg.

Images in Web pages within the Web browser can only be dragged if they
do not serve as a hyper-link.

Obtaining icons:

<ul>
<li>http://images.google.com/images?imgsz=icon&amp;q=insects</li>
<li>http://icone.goldenweb.it?l=en</li>
</ul>

<br><br><b>Species Icons: </b>
Some protein files have a record of the species name.
These species names can be  used to set a suitable icon for the selected proteins:
<i>BUTTON:charite.christo.strap.StrapAlign#button(CMD_SPECIES_ICONS)!</i>.

<br><br><b>Structure Icons: </b>
To download the 3d-structure  thumb-nails from the PDB and use them as icons for the selected proteins,
 press this button:
<i>BUTTON:charite.christo.strap.StrapAlign#button(CMD_PDB_ICONS)!</i>.

<br><br><b>To delete</b> the icons of all selected proteins press this button:
<i>BUTTON:charite.christo.strap.StrapAlign#button(CMD_DEL_ICONS)!</i>.

<br><br><b>Mapping of species and icons: </b> A customizable table  that maps species to images can be inspected:
<i>CUSTOMIZE:speciesIcons!</i>
To deactivate an icon type, associated "NONE".
<br><br>
To see a list of currently loaded images press
<i>BUTTON:charite.christo.strap.StrapAlign#button(CMD_TABLE_ICONS)!</i>.

*/

public final class ChIcon extends RGBImageFilter {
    {  assertInstAllowed(); }
    public final static int FILTER_GRAY=0, FILTER_BRIGHTER_4=1, FILTER_WHITE_TO_TRANSPARENT=2;
    private static ChIcon instances[]=new ChIcon[3];

    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Filter  >>> */
    private int filterType;
    public static RGBImageFilter getFilter(int id) {
        if (instances[id]==null) (instances[id]=new ChIcon()).filterType=id;
        return instances[id];
    }

    @Override public int filterRGB(int x, int y, int rgb) {
         int
            b=(rgb>>0)  &255,
            g=(rgb>>8)  &255,
            r=(rgb>>16) &255,
            a=(rgb>>24) &255;
         if (filterType==FILTER_WHITE_TO_TRANSPARENT) {
             return a>0xDE && g>0xDE && b>0xDE ? 0 : rgb;
         }

         if (filterType==FILTER_GRAY) {
             final int v=(r+g+b)/3;
             return v + (v<<8) +(v<<16) +(v<<24);
         }

         if (filterType==FILTER_BRIGHTER_4) {
             float fBrighter=4;
             b=(int) (b*fBrighter); if (b>255) b=255;
             g=(int) (g*fBrighter); if (g>255) g=255;
             r=(int) (r*fBrighter); if (r>255) r=255;
             return b + (r<<8) +(r<<16) +(a<<24);
         }

         return 0;
    }
    public static ImageIcon newFiltered(int filter,ImageIcon icn) {
        final Image i=icn!=null ? newFiltered(filter,icn.getImage()) : null;
        return i!=null ? new ImageIcon(i) : null;
    }
    public static Image newFiltered(int filterType, Image i) {
           return i==null || !isGraphicsAvailable() ? null :
           getDefaultTK().createImage(new FilteredImageSource(i.getSource(),getFilter(filterType)));
    }

    /* <<< Filter  <<< */
    /* ---------------------------------------- */
    /* >>> Transform >>> */
    private static Image flip(Image im,char direction) {
        final ImageObserver obs=imageObserver();
        int w=im.getWidth(obs); int h=im.getHeight(obs);
        Image imFlip=new BufferedImage(w,h,BufferedImage.TYPE_INT_ARGB);
        Graphics g=imFlip.getGraphics();
        if (direction=='H') g.drawImage(im,w,0,0,h,  0,0,w,h, obs);
        else g.drawImage(im,0,h,w,0,  0,0,w,h, obs);
        //int dx1, int dy1, dx2, dy2,
        //int sx1, int sy1, sx2, sy2,
        return imFlip;
    }
    /* <<< Transform <<< */
    /* ---------------------------------------- */
    /* >>> special Image >>> */
    private static Image crossOut(Image im) {
        final ImageObserver obs=imageObserver();
        int w=im.getWidth(obs); int h=im.getHeight(obs);
        Image imNew=new BufferedImage(w,h,BufferedImage.TYPE_INT_ARGB);
        Graphics g=imNew.getGraphics();

        g.drawImage(im,0,0,w,h, obs);
        g.setColor(C(0xFF0000));
        for(int j=5; --j>=0;) {
            g.drawLine(0+j,0,w+j,h);
            g.drawLine(0+j,h,w+j,0);
        }
        return imNew;
   }    public static Image text2image(String str, Font font, Color fg, Color bg) {
        final String ss[]=str.split("\n");
        final boolean isCursor=font==null;
        int w=0,h=0,fnSize=10;
        Font f=font;
        Dimension d=null;
        for(int iTry=4; --iTry>=0;) {

            f=font!=null ? font : getFnt(fnSize,true,Font.BOLD);
            w=8;
            for(String s: ss) w=maxi(w,4+strgWidth(f,s));
            h=charH(f)*ss.length+1;
            if (!isCursor) {
                final int hw=mini(ICON_HEIGHT, maxi(h,w));
                d=new Dimension(hw,hw);
                break;
            }
            d=getDefaultTK().getBestCursorSize(w,h);
            if (w<=d.width && h<=d.height) break;
            else {
                final float scale1=d.width/(float)w, scale2=d.height/(float)h;
                fnSize=(int)(fnSize* (scale1<scale2 ? scale1:scale2));
            }
        }
        final int W=d.width, H=d.height;
        final int top= maxi(0,  (H-h)/(isCursor?1:2));
        final BufferedImage image=new BufferedImage(W,H,BufferedImage.TYPE_INT_ARGB);
        final Graphics g= image.getGraphics();
        g.setColor(C(0xff0000,0));
        if (bg!=null) {
            g.setColor(bg);
            if (str.length()==1) {
                g.fillOval(0 ,0, W,H);
            }
        }
        g.setFont(f);
        if (isCursor) {
            g.setColor(C(0));
            g.fillRoundRect(0,top,w,charH(f)*ss.length, 6,6);
            g.drawLine(0,0,3,8);
            g.drawLine(0,0,2,8);
        }
        g.setColor(fg);
        for(int i=ss.length; --i>=0;) g.drawString(ss[i],1,top+i*(charH(f)-1)+charA(f));
        return image;

    }

    private static Image icon2image(Icon ic) {
        if (ic==null||!isGraphicsAvailable()) return null;
        if (forIcon2image==null) forIcon2image=new JComponent[]{new JButton(),new JMenuItem()};
        final BufferedImage im=new BufferedImage(ic.getIconWidth(),ic.getIconHeight(),  BufferedImage.TYPE_INT_ARGB);
        for(int i=0; i<forIcon2image.length; i++) {
            try {
                ic.paintIcon(forIcon2image[i],im.getGraphics(),0,0);
                break;
            } catch(Exception e){}
        }
        return im;
    }
    private static JComponent[] forIcon2image;
    public static Image getScaledInstance(Image im, int w, int h, String fileNameExt,Color bg) {
        if (im==null) return null;
        final Component obs=imageObserver();
        final BufferedImage scaled=new BufferedImage(w,h, fileNameExt!=".jpg" && fileNameExt!=".jpeg" ?  BufferedImage.TYPE_INT_ARGB : BufferedImage.TYPE_INT_RGB);
        final Graphics g=scaled.getGraphics();
        g.setColor(bg!=null ? bg: C(0,0));
        g.fillRect(0,0,w,h);
        g.drawImage(im, 0,0, w,h, 0,0,  im.getWidth(obs), im.getHeight(obs),obs);

        return scaled;
    }
    /* <<< special Images <<< */
    /* ---------------------------------------- */
    /* >>> Factory for standard icons  >>> */
    public final static String TRANSPARENT="CI$$T";
    static ImageIcon findIcon(String name) {
      final String s=delSfx(TRANSPARENT,name);
        if (s==null||"".equals(s)||!isGraphicsAvailable()) return null;
        if (s==IC_HIDE) {
            final ImageIcon ic=iicon(IC_SHOW);
            return ic==null ? null : new ImageIcon(newFiltered(FILTER_GRAY,crossOut(ic.getImage())));
        }
        {
            final Icon ic=
                //s==IC_DETACH ?  getInternalFrameMaximizeIcon(32) :
                s==IC_MAXIMIZE16 ? getInternalFrameMaximizeIcon(16) :
                //s==IC_ATTACH ?  getInternalFrameMinimizeIcon(32) :
                null;

            if (ic!=null) return new ImageIcon(icon2image(ic));
        }
        final int L=s.length();
        ImageIcon im=null;
        final boolean slash=s.indexOf('/')>0;
        for(int iExt=3; --iExt>=0;) {
            final String ext=iExt==2? ".gif":iExt==1?".png":".jpg";
            //final URL u=ChIcon.class.getResource("myIcons/"+ s+ext);
            final StringBuffer SB=new StringBuffer(99);
            if (!slash) SB.append("charite/christo/myIcons/");
            SB.append(s).append(ext);
            final URL u=resourceAsURL(0L, SB.toString());
            if (u==null) continue;
            try {im=new ImageIcon(u); } catch(Exception ex){}
            if (im!=null) break;
        }

        if (im==null) {
            final char cz=s.charAt(L-1);
            final int iH= prev(NO_DIGT,s,L- (cz=='|'?2:1) ,-1)+1;
            int width=-1,height=-1;
            String s0=s;
            if (chrAt(iH-1,s)=='x' && (height=atoi(s,iH))>0) {
                final int iW=prev(NO_DIGT,s,iH-2,-1)+1;
                s0=s.substring(0,iW);
                if(iW>=0) width=atoi(s,iW);
            }

            if (s0.length()==0) {
                im=width>0 && height>0?new ImageIcon(new BufferedImage(width ,height,BufferedImage.TYPE_INT_ARGB_PRE)):null;
            }  else if (s0.length()<L) {
                final ImageIcon ic0=findIcon(s0);
                if (ic0!=null) {
                    Image i0=ic0.getImage();
                    if (cz=='|') i0=flip(i0,'H');
                    final boolean jpg=ic0.getDescription().endsWith(".jpg");
                    if (width>0 && height>0) {
                        i0=getScaledInstance(i0,width,height,jpg?".jpg":"",jpg?C(0xFFffFF):null);
                    }
                    im=new ImageIcon(i0);
                }
            }
        }
        if (s==IC_SUPERIMPOSE || s==IC_JAVA || s==IC_STRUCTURE || s==IC_JMOL || s==IC_PDB || s==IC_NCBI ||
            name.endsWith(TRANSPARENT)) {
            im=newFiltered(FILTER_WHITE_TO_TRANSPARENT,im);
        }
        return im;
    }
    // ----------------------------------------
    public static void drawImage(Graphics g,Image im,int x, int y,int w,int h, AlphaComposite com,ImageObserver obs) {
        if (g==null) return;
        final Graphics2D g2d=(Graphics2D)g;
        final Composite originalComposite = g2d.getComposite();
        g2d.setComposite(com);
        g2d.drawImage(im,x,y,w,h,obs);
        g2d.setComposite(originalComposite);
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Cursor >>> */
    private static Image grayBall;
    private static AlphaComposite _ac[];
    public static AlphaComposite getAlphaComposite(int percent) {
        if (_ac==null) _ac=new AlphaComposite[101];
        return _ac[percent]==null ?  _ac[percent]=AlphaComposite.getInstance(AlphaComposite.SRC_OVER, percent*.01f) : _ac[percent];
    }
    public static void drawCursor(Graphics g,int x, int y,int w, int h, ImageObserver obs) {
        final ImageIcon ic=iicon(IC_CURSOR16);
        if (ic!=null) drawImage(g,ic.getImage(),x,y,w,h, getAlphaComposite(70),obs);
    }
     public static Image coloredBall(Color color,Component obs) {
        if (grayBall==null){
            ImageIcon icCursor=iicon(IC_CURSOR16);
            if (icCursor==null) return null;
            grayBall=newFiltered(FILTER_GRAY,icCursor).getImage();
        }
        if (color==null) return null;
        Image i= obs.createImage(new FilteredImageSource(grayBall.getSource(),new ColorFilter(color)));
        ChIcon.waitFor(i,obs);
        return i;
    }
    //
    private static Map<String,Cursor> mapCursor;
    public static Cursor getCursor(final String str, Component obs) {
        if (str==null||!isGraphicsAvailable()) return null;
        if (mapCursor==null) mapCursor=new java.util.HashMap();
        Cursor c=mapCursor.get(str);
        if (c==null) {
            try{
                final ImageIcon icon=iicon(str);
                final Image image;
                final Graphics g;
                if (icon!=null) {
                    final int h=icon.getIconHeight(), w=icon.getIconWidth();
                    if (h<1 && w<1) return null;
                    final Dimension d=getDefaultTK().getBestCursorSize(w,h);
                    image=new BufferedImage(d.width,d.height,BufferedImage.TYPE_INT_ARGB);
                    g=image.getGraphics();
                    g.drawImage(icon.getImage(),4,0,obs);
                } else {
                    image=text2image(str, null, C(0xFFffFF),C(0));
                }
                c=getDefaultTK().createCustomCursor(image, str==IC_EDIT ? new Point(5,isWindows() ? 20+8 : isWindows() ? 20+8 : 20) :  new Point(0,0),"");
            } catch(Exception e){ c=getCursr('H'); stckTrc(e); }
            mapCursor.put(str,c);
        }
        return c;
    }
    /* <<< Cursor  <<< */
    /* ---------------------------------------- */
    /* >>> Utils >>> */
    private static Component imageObserver_;
    public static Component imageObserver() {
        if (imageObserver_==null) imageObserver_=isGraphicsAvailable() ? new Frame() : new JButton();
        return imageObserver_;
    }
    public static void setImageObserver(Component c) { imageObserver_=c; }

    public static int getHeight(Image im) { return im==null ? 0 : im.getHeight(imageObserver()); }
    public static int getWidth( Image im) { return im==null ? 0 : im.getWidth( imageObserver()); }

    private static void waitFor(Image im,Component comp) {
        if (im==null) return;
        MediaTracker mt=new MediaTracker(comp);
        mt.addImage(im,0);
        try { mt.waitForAll(); } catch(Exception e) {}
    }

}

// --- Check ---
//   A=`cat $HOME/java/charite/christo/icon_java | fgrep public | fgrep -v fgrep | fgrep 'return gif("' | sed 's|^.*("||1;s|").*$|.gif|1'`
//
