package charite.christo;
import charite.christo.interfaces.*;
import java.awt.*;
import javax.swing.*;
import static javax.swing.SwingUtilities.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
/**
   @author Christoph Gille
*/
public final class ChSprite implements Runnable {
    public final static int SPEED_SINUS=1<<1;
    private final static int TEXT_MARGIN=4;
    private final Point  pDest;
    private final Object image;
    private final Font font=getFnt(12, true, 0);
    private final int width, height, options, screenX0, screenY0;
    private final float mseconds;
    private int lastScrX=MIN_INT, lastScrY,  screenX, screenY;
    public ChSprite(final long options, Component cSrc, Point pSrc, Component cDest, Point pDest, Object image, float mseconds) {
        this._canvasDest=weakRef(cDest);
        this.pDest=pDest;
        this.mseconds=mseconds;
        this.image=image;
        this.options=(int)options;
        if (image instanceof String) {
            height=2*TEXT_MARGIN+charH(font);
            width=2*TEXT_MARGIN+strgWidth(font, (String)image);
        } else width=height=10;
        final Point p=new Point(pSrc.x, pSrc.y);
        convertPointToScreen(p, cSrc);
        screenX0=p.x;
        screenY0=p.y;
    }
    /* <<< Constructor <<< */
    /* ---------------------------------------- */
    /* >>> canvas >>> */
    private final Object _canvasDest;
    private Component _lastDeep;
    private Component canvas() {
        final Component
            c0=deref(_canvasDest, Component.class),
            cRoot=c0==null?null : getRootPane(c0),
            cDeep=cRoot==null?null : getDeepestComponentAt(cRoot, this.screenX,this.screenY);
        final Component c=cDeep instanceof JComponent || cDeep==null ? cRoot : cDeep;
        if (c==null) return null;
        pNow.move(screenX, screenY);
        convertPointFromScreen(pNow,c);
        if (lastScrX==MIN_INT) pLast.move(screenX, screenY); else pLast.move(lastScrX, lastScrY);
        convertPointFromScreen(pLast, c);
        if (_lastDeep!=null && _lastDeep!=cDeep) ChDelay.repaint(_lastDeep,99);
        _lastDeep=cDeep;
        return c;
    }
    /* <<<  canvas <<< */
    /* ---------------------------------------- */
    /* >>>  offImage >>> */
    private Image offImage;
    private static Object offImage2;
    private int offX, offY, imW, imH;
    private final Point pLast=new Point(), pNow=new Point();

    private Image createOffImage() {
        final Component c=canvas();
        if (c==null) return null;
        Image im=offImage;
        final int
            x=pNow.x,
            y=pNow.y,
            lX=pLast.x,
            lY=pLast.y,
            xMi=mini(x,lX),
            yMi=mini(y,lY),
            xMa=maxi(x,lX)+width,
            yMa=maxi(y,lY)+height;

        if (im==null || xMi<offX || xMa>offX+imW ||  yMi<offY ||   yMa>offY+imH) {
            if (im==null) {
              offImage=im=c.createImage(imW=xMa-xMi+width*4, imH=yMa-yMi+height*4);
                //offImage=im=c.createImage(imW=xMa-xMi+33, imH=yMa-yMi+33);
            }
            if (im==null) return null;
            offX=x+ (lX==x ? (width -imW)/2 : lX<x?0 :  width-imW);
            offY=y+ (lY==y ? (height-imH)/2 : lY<y?0 : height-imH);

            offX=x+ (width -imW)/2;
            offY=y+ (height-imH)/2;
            //putln(" offX="+offX+"  "+offY);
            final Graphics g=im.getGraphics();
            g.translate(-offX, -offY);
            c.paint(g);
            //putln(count++ +" paint "+screenY);
            g.translate(offX, offY);
        }
        return im;
    }
    /* <<< offImage <<< */
    /* ---------------------------------------- */
    /* >>>  offImage >>> */
 private void stroke() {
        final Component c=canvas();
        final Graphics g=c==null?null:c.getGraphics();
        if (g==null) return;
        final Image im=createOffImage();
        final int
            x=pNow.x,
            y=pNow.y,
            lX=pLast.x,
            lY=pLast.y,
            xMi=mini(x,lX),
            yMi=mini(y,lY),
            xMa=maxi(x,lX)+width,
            yMa=maxi(y,lY)+height,
            dx=xMa-xMi,
            dy=yMa-yMi;
        if (im==null) return;
        // putln(" x="+offX+" y="+offY);
        Image off2=deref(this.offImage2, Image.class);
        if (off2==null || off2.getWidth(c)<dx || off2.getHeight(c)<dy) {
            off2=c.createImage(dx+33, dy+33);
            this.offImage2=newSoftRef(off2);
        }
        final Graphics g2=off2.getGraphics();

        g2.drawImage(im,  0,0,dx,dy,    xMi-offX,yMi-offY,xMa-offX,yMa-offY,   c);
        /*
        if (isMyComputer()) {
          g2.setColor(Color.RED);
          g2.drawRect(0,0,dx,dy);
        }
        */

        if (this.image instanceof String) {

            final int ax=screenX0-screenX, ay=screenY0-screenY, transp=32+(int)(2*Math.sqrt(ax*ax+ay*ay));
            g2.setColor(transp>255 ? C(0xFFffFF): C(0xFFffFF,transp));
            g2.fillRoundRect(x-xMi+1, y-yMi+1, width-2, height-2, TEXT_MARGIN/2, TEXT_MARGIN/2);
            g2.setColor(C(0));
            g2.setFont(font);
            g2.drawString((String)image, x-xMi+TEXT_MARGIN, y-yMi+TEXT_MARGIN+charA(font));
            //g.drawImage(off2, xMi, yMi,c);
            g.drawImage(off2,   xMi,yMi,xMa,yMa,  0,0, dx,dy,     c);

        }

    }

    private final Point pDestS=new Point();
    public void run() {
        if (isEDT()) {
            stroke();
        } else {
            final Component cDest=deref(_canvasDest, Component.class);
            final long time0=System.currentTimeMillis();
            for(;;) {
                final long time=System.currentTimeMillis();
                pDestS.move(this.pDest.x, this.pDest.y);
                convertPointToScreen(pDestS, cDest);

                final double dx=pDestS.x-screenX0, dy=pDestS.y-screenY0, dist=Math.sqrt(  dx*dx+dy*dy);
                final long time1=(long)(time0+dist*mseconds);
                if (time>time1) break;
                double part=(time-time0)/(time1-(double)time0);
                if (0!=(options&SPEED_SINUS)) {
                    final double sin=Math.sin(Math.PI*(part-0.5));
                    part=0.5+sin/2;
                    part*=part;
                }
                this.screenX=(int)(screenX0+  dx*part)-width/2;
                this.screenY=(int)(screenY0+  dy*part)-height/2;
                inEDT(this);
                this.lastScrX=screenX;
                this.lastScrY=screenY;
                sleep(1);
            }
            ChDelay.repaint(canvas(),2222);
        }
    }

}