package charite.christo;
import charite.christo.interfaces.*;
import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.JComponent;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP
<h3>Exceptions and the streams stdout and stderr </h3>
The standard streams  WIKI:Standard_streams  stdout  and  stderr are
pre-connected output channels of computer program.

Stdout and stderr are collected in files residing in
"./strapTmp" unless STRAP was started with
the command line parameter "-stderr" and "-stdout".

They can be viewed in the menu "errors" of the option menu.

Alternatively, the program <i>STRING:ChUtils#jControl()</i> in
<i>STRING:ChUtils#dirJavaws()</i> can be used to switch the
so-called Java-console on.

In case of runtime errors like division by zero
the Java machine throws  so-called <I>exceptions</I> in <I>stderr</I>.

   @author Christoph Gille
*/
public class ChStdout implements ChRunnable,ActionListener {
    private final static byte[] OUTOFMEMORYERROR=OutOfMemoryError.class.getName().getBytes();
    private static ChButton cbSend;
    private static ChStdout inst;
    private static ChStdout instance() {  return inst==null ?  inst=new ChStdout() : inst; }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> PrintStream >>> */
    private static PrintStream _printStream;
    private static PrintStream printStream() {
        if(_printStream==null)  {
            try {
            _printStream=new PrintStream(fileOutStream(logFile('R'))) {
                    @Override public void write(byte[] bb) throws IOException {      super.write(bb); ChStdout.write(bb, 0, bb.length);}
                    @Override public void write(byte[] bb,int off,int len) {      super.write(bb,off,len); ChStdout.write(bb, off, len);}
                };
            } catch(Exception ex){}
        }
        return _printStream;
    }

    public static void write(byte[] bb,int off,int len) {
        if (strstr(OUTOFMEMORYERROR,bb,off,off+len)>=0) {
            error(OutOfMemoryError.class+": <br>See web site how to increase the Java Heap size.");
        }
        if (!haveAlreadyREportedError && strstr(_CC,bb,off,off+len)>=0) {
            haveAlreadyREportedError=true;
            ChDelay.runAfterMS(EDT,instance(),"didTheErrorOccurInMyCode",99);
            //            didTheErrorOccurInMyCode();
        }
        if (strstr("CDataTransferer_dragQueryFile",bb,off,off+len)>=0) {
            handleEvt(ChStdout.class,EVT_CDataTransferer_dragQueryFile);
        }
    }
    /* <<< PrintStream <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */

    public final static ActionEvent EVT_CDataTransferer_dragQueryFile=new ActionEvent("",ActionEvent.ACTION_PERFORMED,"",0);
    private static PrintStream stdout;
    private static File logStdoutFile_, logStderrFile_;
    private static File logFile(char O_or_R) {
        File f= O_or_R=='O' ? logStdoutFile_ : logStderrFile_;
            if (f==null) {
                f=file(dirLog()+ (O_or_R=='O' ? "/stdout.log" :  "/stderr.log"));
                if (O_or_R=='O') logStdoutFile_=f; else logStderrFile_=f;
                mkParentDrs(f);
            }

        return f;
    }
    private static boolean haveAlreadyREportedError, isStdoutRedirected,isStderrRedirected;

    public static void redirect(char O_or_R){
        try {
            mkParentDrs(logFile('O'));
            putln("Going to redirect");
            if (O_or_R=='O') {
                stdout=new PrintStream(fileOutStream(logFile('O')));
                putln("Stdout ", logFile('O'));
                System.setOut(stdout);
                isStdoutRedirected=true;
            } else if (O_or_R=='R') {
                putln("Stderr ", logFile('R'));
                System.setErr(printStream());
                isStderrRedirected=true;
            } else assertTrue(false);
        } catch(Exception e){ putln("Caught in ChStdout#redirect ",e);}
    }
    private static void didTheErrorOccurInMyCode() {
        printStream().flush();
        BufferedReader r=null;
        try{
            r=new BufferedReader(new FileReader(logFile('R')));
            String line;
            boolean trigger=false,sendReport=false;
            while( (line=r.readLine())!=null) {
                if (!is(SPC,line,0))  { trigger=line.indexOf("Exception")>=0 || line.indexOf("OutOfMemoryError")>=0; continue;}
                if (trigger) {
                    if (line.indexOf("jPlugs")>0 || line.indexOf("astex.MoleculeRenderer")>0)  trigger=false;
                    else if (line.indexOf("charite.christo")>0) {
                        sendReport=haveAlreadyREportedError=true;
                        break;
                    }
                }
            }
            if (sendReport) uploadErrorReport(readStrg(logFile('R')),false);
        } catch(Exception e){putln("caught in ChStdout#didTheErrorOccurInMyCode",e);}
        finally {closeStrm(r);}
    }
    public static void accordingToCommandLineOptions(String[] argv) {
        boolean redirectStdout=true,redirectStderr=true;
        for(String arg: argv) {
            if (OPTION_stdout.equals(arg) || isSystProprty(IS_WEBSTARTED)) redirectStdout=false;
            if (OPTION_stderr.equals(arg)) redirectStderr=false;
        }
        if (redirectStderr) redirect('R');
        if (redirectStdout) redirect('O');
    }
    public static ChButton newButton(char t) {
        final ActionListener ali=instance();
        if (t=='e') return new ChButton("Simulate an exception ").r(ChDelay.newThread(instance(),"SIMUL",null));
        if (t=='E') return new ChButton("SIMUL").t("Simulate an exception in thread").li(ali);
        if (t=='F') return new ChButton("DEL_F").t("Illegal file access").li(ali);
        if (t=='A') return new ChButton("A").t("Throw assertion error").li(ali);
        if (t=='O' && !isStdoutRedirected || t=='R' && !isStderrRedirected) return null;
        return new ChButton( (t=='O' ? "Stdout" : "Stderr")).actionViewFile(logFile(t))
            .tt("Observing standard  error stream<br>Is inactive when STRAP is started with the options <I>-stderr</I><br>"+
                "Exceptions marked <I>\"caught\"</I> are harmless<br><br>");

    }

    private static ChTextArea taException;
    private static Class clas;
    private static int lineNumber=0;
    private static StringBuffer getSourceLine(String s) {
        final int idx=s.indexOf("at charite.christo")+3;
        final StringBuffer sbOut=new StringBuffer(5555);
        if (idx<3) return sbOut.append("lineNumber: -1");
        final int idxDollar=s.indexOf('$',idx), idxOpen=s.indexOf('(',idx);
        int idxEndClassName=-1;
        if (idxDollar>0 && idxDollar<idxOpen) idxEndClassName=idxDollar;
        else
            for(idxEndClassName=idxOpen;  idxEndClassName>idx && chrAt(idxEndClassName,s)!='.'; idxEndClassName--);
        if (idxEndClassName<0) return null;
        final String name=s.substring(idx,idxEndClassName);
        clas=null;
        try { clas=Class.forName(name); }catch(Throwable e){putln("caught in ChStdout#getSourceLine ",e); return sbOut;}
        sbOut.append(ANSI_RED+"CLASS"+ANSI_RESET+"=").append(clas).append("\n");
        final int idxColon=s.indexOf(':',idx);
        if (idxColon<0) return sbOut;
        lineNumber=atoi(s,idxColon+1)-1;

        sbOut.append("line number=").append(lineNumber).append(" \njava version=").append(systProprty(SYSP_JAVA_VERSION)).append("\ncompiled=").append(dateOfCompilation()).append("\n");
        final ByteArray ba=getJavaSource(clas.getName());
        if (ba!=null) {
            final int ends[]=ba.getLineEnds();
            if (lineNumber>0 && lineNumber<ends.length) {
                appnd(ba.getBytes(),ends[lineNumber-1],ends[lineNumber],sbOut).append("\n");
            }
        }
        return sbOut;
    }
    public static void uploadErrorReport(String txt, boolean isModal) {
        if (txt==null) return;
        final StringBuffer sb=new StringBuffer(9999).append("\nUpload this text or mail it to christophgil@googlemail.com !\n\n").append(txt);
        final StringBuffer srcLine=getSourceLine(txt);
        sb.append(day(0)+" OS_ARCH="+systProprty(SYSP_OS_ARCH)+" OS_NAME="+systProprty(SYSP_OS_NAME)+" \nDateOfCompilation="+dateOfCompilation()+"\n"+srcLine+"\n\n");
        if (taException==null) {
            taException=new ChTextArea(7,70);
            cbSend=toggl("CB").t("Yes, send  this bug report").li(instance()).s(true);
        }
        sb.append("getPrgParameters()= "+joinAsStrg(getPrgParameters(),"  ")+"\n");
        taException.t(sb);
        final File f=file(dirLog()+"/bugReport.log");
        wrte(f,sb);
        final JComponent panUpload=vBhB(
                                  "<h2>Transmitting the error protokoll</h2>"+
                                  "An error has occured!<br>"+
                                  "Please upload this bugreport.<br>"+
                                  "Then save the alignment and restart the program.<br>"+
                                  "Bugs are usually fixed within  48 h.<br>"+
                                  "<br>"+
                                  "The following and only the following text is transmitted to the STRAP server:",
                                  scrllpn(0L,taException),
                                  hBL(cbSend.cb(), isModal ? null : new ChButton("OK").li(instance()))
                                  );
        if(isModal) {
            ChMsg.msgDialog(0L, panUpload);
            if (cbSend.s()) thread_uploadBug(toStrg(taException)).start();
        } else addActLi(instance(), new ChFrame("error").ad(panUpload).size(800,500).shw(0));
    }

    public static Thread thread_uploadBug(CharSequence txt) {
        return threadM("uploadBug",ChStdout.class, new Object[]{txt});
    }
  private static boolean uploading;
    public static void uploadBug(CharSequence txt) {

        Web.getServerResponse(0L,URL_STRAP+"error.php"+DO_NOT_ASK,new Object[][]{{"error",txt}});

    }

    public Object run(String id, Object arg) {
        if (id=="SIMUL") {
            try {
                int idx=4711;Object oo[]=new Object[4]; for(Object o:oo) oo[idx]=null;
            } catch(Exception ex) { ex.printStackTrace();}
        }

        if (id=="didTheErrorOccurInMyCode") didTheErrorOccurInMyCode();
        return null;
    }

    public void actionPerformed(ActionEvent ev) {
        final String cmd=ev.getActionCommand();
            if (cmd=="A") { assertTrue(false); }
        if (cmd=="SIMUL") {
            try {
                int idx=4711;Object oo[]=new Object[4]; for(Object o:oo) oo[idx]=null;
            } catch(Exception ex) { ex.printStackTrace();}
        }

        if (cmd=="CB") taException.setEnabled(cbSend.s());
        if ((cmd=="OK" || cmd==ACTION_WINDOW_CLOSING) && taException!=null) {
            if (cbSend.s()) thread_uploadBug(toStrg(taException)).start();
            parentWndw(ev.getSource()).dispose();
        }
        if (cmd=="DEL_F") {
            if (tf_del==null) tf_del=new ChTextField("",COMPLETION_FILES_IN_WORKING_DIR).cols(80,true,false).saveInFile("ChStdout_del");
            ChMsg.msgDialog(0L, vBhB(
                                "Here you can challenge the security system.",
                                "It should only be poosible to create files in the working and the settings directory.",
                                "Creating/Deleting other files should lead to an Error",
                                "Enter a file to be deleted and replaced:",
                                tf_del)
                                );
            final String path=tf_del.toString();
            if (sze(path)>0) ChMsg.msgDialog(0L,  wrte(file(path),"checkIllegalFileAccess") ? "Success" : "Failed");
        }
    }
    private static ChTextField tf_del;
}
