package charite.christo;
import charite.christo.interfaces.*;
import java.awt.*;
import java.util.*;
import java.util.List;
import javax.swing.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
import static java.awt.event.MouseEvent.*;
/**
   A reimplementation of tabbed pane since JTabbedPane does not work very well
   @author Christoph Gille
*/
public class ChTabPane extends ChPanel implements ChProcessEvent,PaintHook,PreferredSize,Comparator  {
    public final static String ACTION_REMOVED="TP_A_RM", ACTION_ADDED="TP_A_ADD", ACTION_SELECTION_CHANGED="TP_A_SC",
        KEY_SORT="CTB$$S", KEY_CONTAINING_CTP="CTP$$P", KEY_UNIQUE_ID="CTP$$U";
    public final static int LEFT=1<<1, RIGHT=1<<2, COLLAPSE=1<<3;
    private final Object KEY_WHEN=new Object(), KEY_TAB=new Object();
    private final CollapseComponent collapseComponents;
    private final ArrayList<JComponent> componentsV=new ArrayList();
    private final JComponent panTabs,pTabs;
    private JComponent completePanTabs;
    private transient Object ref_pCenter;
    private final boolean leftSide;
    private final List<Component> focusedV=new ArrayList();
    private JList jListHistory;
    private final ChEventAdapter LI=new ChEventAdapter(this);
    private ChButton bSelected;
    private final ChButton butPrevious=new ChButton("\u25C4", ChButton.DISABLED|ChButton.MAC_TYPE_ICON|ChButton.HIDE_IF_DISABLED)
        .li(LI)
        .tt("selects the previously focused dialog");

    public int compare(Object o1,Object o2) {
        final String k1=gcps(KEY_SORT,o1), k2=gcps(KEY_SORT,o2);
        final int w1=(atoi(gcp(KEY_WHEN,o1))<<16) | componentsV.indexOf(o1), w2=(atoi(gcp(KEY_WHEN,o2))<<16) | componentsV.indexOf(o2);
        final int comp=k1!=null && k2!=null  ? k1.compareTo(k2) : 0;
        return comp!=0 ? comp : w1<w2 ? -1 : 1;
    }
    public ChTabPane(long options) {
        this.leftSide=0!=(options&LEFT);
        this.insets=new Insets(0,leftSide?0:EX*3/2,0,leftSide?EX*3/2:0);
        setPrefSize(this, panTabs=pnl(OPAQUE_FALSE));
        pTabs=pnl(OPAQUE_FALSE);
        pTabs.setMinimumSize(dim(0,0));
        pTabs.add(panTabs);
        completePanTabs=pTabs;
        if (leftSide) {
            completePanTabs=pnl(OPAQUE_TRUE);
            addPaintHook(this,completePanTabs);
            enlargeFont(completePanTabs,2);
            completePanTabs.setLayout(new BorderLayout());
            completePanTabs.add(pTabs,BorderLayout.CENTER);
            setMinimumSize(dim(3,3));
        }
        if (leftSide) completePanTabs.add(pnl(OPAQUE_FALSE,butPrevious),BorderLayout.NORTH);
        cnsew(this,null,null,null, leftSide?null:completePanTabs, leftSide?completePanTabs:null);
        collapseComponents=0==(options&COLLAPSE) ? null: new CollapseComponent(completePanTabs,"EW").parent(ChTabPane.this).child(panTabs);
    }
    public void setSelectedComponent(JComponent o) {  addTab(o);}
    public JComponent getSelectedComponent() {  return (JComponent)deref(ref_pCenter); }

    public JComponent[] getTabComponents() { return toArry(componentsV,JComponent.class);}
    private final static int xx[]=new int[4], yy[]=new int[4];
    private static Paint gradPaint;
    public boolean paintHook(JComponent c,Graphics g1, boolean after) {
        final Graphics2D g=(Graphics2D)g1;
        final int h=c.getHeight(),w=c.getWidth();
        if ("TAB"==gcp(KEY_ACTION_COMMAND,c)) {
            if (c==bSelected && after) {
                    final int hi=h/3, xTip=leftSide?w:0, xBase=leftSide?w-EX/2:EX/2;
                    xx[0]=xBase; yy[0]=(h-hi)/2;
                    xx[1]=xBase; yy[1]=(h-hi)/2+hi;
                    xx[2]=xTip;  yy[2]=h/2;
                    xx[3]=xx[0]; yy[3]=xx[0];
                    g.setColor(C(0));
                    g.fillPolygon(xx,yy,3);
                }
            if (!after) {
                final Color color=colr(gcp(ChButton.KEY_COLOR_BOX_L,c));
                g.setColor(color!=null?color:C(0));
                final int rx=leftSide?0:EX*3/2, rw=w-EX*3/2;
                if (c==bSelected) g.drawRect(rx,0,rw,h-1);
            }
        }
        if (after && c==completePanTabs) {
            if (gradPaint==null) gradPaint=new GradientPaint(w,h,C(DEFAULT_BACKGROUND),0,0,C(0x888888));
            g.setPaint(gradPaint);
            g.fillRect(0,0,w-2,h);
        }

        return true;
    }
    public void removeAllTabs() {
        componentsV.clear();
        final Component pC=deref(ref_pCenter,Component.class);
        if (pC!=null && pC.getParent()==this) remove(pC);
        updateGUI();
        revalidate();
        repaint();
        handleActEvt(this,ACTION_REMOVED,0);

    }

    private void updateGUI() {
        panTabs.removeAll();
        panTabs.setLayout(new GridLayout(sze(componentsV),1));
        for(JComponent jc: getTabComponents())  {
            final ChButton tog=gcp(KEY_TAB,jc,ChButton.class);
            panTabs.add(tog);
        }
        ChDelay.revalidate(jListHistory,222,(long[])null);
        for(int i=sze(focusedV);--i>=0;) {
            if (!componentsV.contains(focusedV.get(i))) focusedV.remove(i);
        }
        bSelected=gcp(KEY_TAB,getSelectedComponent(), ChButton.class);

        setEnabld(sze(focusedV)>4, butPrevious);
    }
    public boolean removeTab(JComponent... cc) {
        JComponent c0=null;
        boolean success=false;
        for(JComponent c:cc) {
            if (c!=null && componentsV.remove(c)) {
                pcp(KEY_CONTAINING_CTP,null,c);
                pcp(KEY_TAB,null,c);
                success=true;
            } else if (c0==null) {
            }

        }
        if (success) {
            previousTab(c0,true);
            handleActEvt(this,ACTION_REMOVED,0);
        }
        return success;
    }
    private void previousTab(Object c,boolean shouldRemove) {
        if (shouldRemove) componentsV.remove(c);
        if (sze(focusedV)>0 && (c==null || focusedV.get(0)==c)) {
            focusedV.remove(0);
        }
        if (c==null || c==getSelectedComponent()) {
            final JComponent newC=get(0, sze(focusedV)>0 ? focusedV : componentsV,JComponent.class);
            privateSelectedComponent(newC);
        }
        updateGUI();
        revalidate();
        repaint();
        butPrevious.repaint();
    }

    public void addTab(HasPanel hp) {  /* Koennte man verallgemeinern addTab(Object) */
        final Component c=hp.getPanel(HasPanel.NEW_PANEL);
        TabItemTipIcon.set(TabItemTipIcon.tab(hp),null,null,null,  c);
        addTab(c);

    }

    public void addTab(Component c) {

        if (c==null || c==getSelectedComponent()) {
            handleActEvt(this,ACTION_SELECTION_CHANGED,0);
            return;
        }
        if (gcp(KEY_WHEN,c)==null) pcp(KEY_WHEN,intObjct((int)(System.currentTimeMillis()&0x7fFFffFF)),c);
        final Object noDuplicates=gcp(KEY_UNIQUE_ID,c);
        if (noDuplicates!=null) {
            for(int i=sze(componentsV); --i>=0;) {
                final JComponent jc=get(i,componentsV,JComponent.class);
                if (noDuplicates.equals(gcp(KEY_UNIQUE_ID,jc))) {
                    privateSelectedComponent(jc);
                    return;
                }
            }
        }
        handleActEvt(this,ACTION_ADDED,0);
        adUniq(c,componentsV);
        if (gcp(KEY_TAB,c)==null) pcp(KEY_TAB,newTabButton(c),c);
        pcp(KEY_CONTAINING_CTP,weakRef(this),c);
        privateSelectedComponent(c);
        addMoliRefreshTip(c);
        final JComponent sorted[]=getTabComponents();
        sortArry(sorted,this);
        componentsV.clear();
        adAllUniq(sorted,componentsV);
        updateGUI();
        doLayout();
        revalidate();
        repaint();
        ChDelay.updateUI(this,2222,durationUpdateUI);
    }
    private final long durationUpdateUI[]=new long[ARRAY_DURATION];
    private void privateSelectedComponent(Component c) {
        final JComponent center=getSelectedComponent();
        if (center!=null) remove(center);
        ref_pCenter=weakRef(c);
        if (c==null) return;
        add(c,BorderLayout.CENTER);
        if (center!=c)  handleActEvt(this,ACTION_SELECTION_CHANGED, 0);

        /*
           IllegalArgumentException:
           adding a container to a container on a different GraphicsDevicecause: nulljava.lang.IllegalArgumentException: adding a container to a container on a different GraphicsDevice
        */
        bSelected=gcp(KEY_TAB,c,ChButton.class);
        focusedV.remove(c);
        focusedV.add(0,c);
        revalidate();
        repaint();
        butPrevious.repaint();
    }
    private final Insets insets;
    private ChButton newTabButton(Component comp) {
        final String tab=TabItemTipIcon.tab(comp);
        final Icon ic=TabItemTipIcon.icn(comp);
        final ChButton tog=new ChButton("TAB", ChButton.NO_FILL| ChButton.NO_BORDER).li(LI)
            .t(tab).i(leftSide?ic:null)
            .tt(TabItemTipIcon.tip(comp))
            .cp(JComponent.class,weakRef(comp));

        Colored colored=deref(comp,Colored.class);
        if (colored==null) colored=deref(gcp(KEY_MODEL,comp),Colored.class);
        tog.cp(leftSide ? ChButton.KEY_COLOR_BOX_L : ChButton.KEY_COLOR_BOX_L, colored);
        if (comp instanceof Disposable) ChRenderer.setSmallButtonX(ChRenderer.SMALL_CLOSE, leftSide?0:-ChRenderer.BUT_WIDTH, tog);

        LI.addTo("m",tog);
        tog.setHorizontalAlignment(leftSide?AbstractButton.TRAILING:AbstractButton.LEFT);
        if (leftSide) tog.setHorizontalTextPosition(AbstractButton.LEFT);

        tog.setMargin(insets);
        addPaintHook(this,tog);
        if (collapseComponents!=null) collapseComponents.preventedBy(this);
        setTabUndocksPanel(tog,comp);
        return tog;
    }

    public void updateTabLabels() {
        for(JComponent c:getTabComponents()) {
            final ChButton tog=gcp(KEY_TAB,c, ChButton.class);
            if (tog!=null) {
                tog.t(TabItemTipIcon.tab(c));
            }
        }
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> ActionListener >>> */

    public AWTEvent processEvent(JComponent c, AWTEvent ev) {
        final Object q=ev.getSource();
        final String cmd=actionCommand(ev);
        final int id=ev.getID();
        if (cmd=="TAB" && q!=bSelected) {
            final JComponent pc=getSelectedComponent();
            if (pc!=null) remove(pc);
            bSelected=(ChButton)q;
            addTab(gcp(JComponent.class,q,JComponent.class));
        }
        if (q==butPrevious) {
            if (isCtrl(ev)) {
                if (jListHistory==null) jListHistory=new ChJList(focusedV,ChJTable.ICON_ROW_HEIGHT|ChJTable.CLASS_RENDERER);
                ChFrame.frame("History",jListHistory,ChFrame.SCROLLPANE).shw(0);
            } else previousTab(null,false);
            ChDelay.revalidate(jListHistory,333,(long[])null);
        }
        if (id==MOUSE_PRESSED && "TAB"==gcp(KEY_ACTION_COMMAND,c) && ChRenderer.inSmallButClose(q,x(ev), y(ev))) {
            dispos(isCtrl(ev)?0L:DISPOSE_ASK_CTRL, gcp(JComponent.class,c,Disposable.class));
        }

        return ev;
    }
    /* <<< ActionListener <<< */
    /* ---------------------------------------- */
    /* >>> getPreferredSize >>> */

    public Dimension getPreferredSize(JComponent component) {
        int height=0;
        final Component cc[]=component.getComponents();
        final int width=maxi(ICON_HEIGHT, prefWidth(cc));

        if (component==panTabs) height=mini(pTabs.getHeight()-ICON_HEIGHT,sze(componentsV)*ICON_HEIGHT);
        if (component==pTabs)   height=mini(cc.length*ICON_HEIGHT,getHeight()-butPrevious.getHeight());
        return dim(width,height);
    }

}
