package charite.christo;
import charite.christo.interfaces.*;
import java.awt.*;
import java.util.*;
import javax.swing.*;
import static java.awt.event.MouseEvent.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
public final class ChTextArea extends JTextArea implements ChRunnable, HasWeakRef {
    public final Object WEAK_REF=newWeakRef(this);
    public Object getWeakRef() { return WEAK_REF;}

    /* >>> Constructor >>> */
    private boolean assertEDT;
    public ChTextArea(CharSequence s, long options) {
        super(sze(s)==0 ? "" : toStrg(s));
        if (options!=0) tools().underlineRefs(options);
        setWrapStyleWord(true);
        monospc(this);
        assertEDT=true;
        setBG(0xFFffFF, this);
    }
    public ChTextArea(CharSequence s) {
        this(s, 0L);
    }
    public ChTextArea(int rows,int cols) {this("",0L); setColumns(cols); setRows(rows);}
    @Override public String toString() { tools().loadSavedText();   return getText();    }
    /* <<< Constructor <<< */
    /* ---------------------------------------- */
    /* >>> set >>> */
    public ChTextArea t(Object s) {
        if (s!=null) {
            tools().setTextTS(toStrg(s));
        }
        tools().setChanged();
        return this;
    }
    private long duration[];
    public long[] getWhenPainted() {
        if (duration==null) duration=new long[ARRAY_DURATION];
        return duration;
    }
    @Override public void append(String s) {
        assertEDT();
        tools().underlineRefs(UNDERLINE_REFS_DISABLE_UNTIL_MOUSE);super.append(s);
    }
    @Override public void setText(String s) {
        if (assertEDT) assertEDT();
        super.setText(s);
    }
    @Override public String getText() { assertEDT(); return super.getText(); }
    public ChTextArea ro(){ setEditable(false);return this;}
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Paint >>> */
    private boolean isPainting;
    @Override public boolean isOpaque() { return isPainting ? false : super.isOpaque(); }

    private boolean isPainted;
    public void paintComponent(Graphics g) {
        final ChTextComponents tools=tools();
        final long time=System.currentTimeMillis();
        final int w=getWidth(), h=getHeight();
        final Color bg=getBackground();
        if (bg!=null && isOpaque()) {
            g.setColor(bg);
            fillBigRect(g,0,0,w,h);
        }

        if (tools.paintHook(this,g,false)) {
            try {
                isPainting=true;
                super.paintComponent(g);
                isPainting=false;
            } catch(Throwable e){}
        }
        if (!isPainted) {
            isPainted=true;
            if (this.textAppender!=null) this.textAppender.run();
        }
        tools.paintHook(this,g,true);
        final long d[]=this.duration;
        final Rectangle clip=g.getClipBounds();
        if (d!=null && (clip==null || clip.width>20)) d[DURATION]=(d[WHEN]=System.currentTimeMillis())-time;
    }
    private final static Map<Color,javax.swing.text.Highlighter.HighlightPainter> MAP_HighlightPainter=new HashMap();
    public static javax.swing.text.Highlighter.HighlightPainter highlightPainter(Color c) {
        javax.swing.text.Highlighter.HighlightPainter hp=MAP_HighlightPainter.get(c);
        if (hp==null) MAP_HighlightPainter.put(c,hp=new javax.swing.text.DefaultHighlighter.DefaultHighlightPainter(c));
        return hp;
    }

    /* <<< Paint <<< */
    /* ---------------------------------------- */
    /* >>>  Append >>> */
    public ChTextArea appnd(Object o) {
        if (o==null || !(o instanceof Color) && sze(o)==0) return this;
        if (o instanceof Object[]) return appnd((Object[])o);
        if (isEDT() && isVisible()) {
            appendNow(o);
        } else {
            getTextAppender().appnd(o);
        }
        tools().setChanged();
        return this;
    }
    public ChTextArea appnd(Object... oo) {
        if (countNotNull(oo)==0) return this;
        if (isEDT() && isVisible()) {
            for(Object o :oo) appendNow(o);
        } else {
            getTextAppender().appnd(oo);
        }
        tools().setChanged();
        return this;
    }
    public ChTextArea appndS(String s) {
        if (s!=null) append(s);
        return this;
    }
    private transient TextAppender textAppender;
    public void setMaxTextLength(int max) { getTextAppender().setMaxLength(max);}
    private int lastStart;
    private Color lastColor;
    public void appendNow(Object o0) {
        if (o0==null) return;
        assertEDT();
        Object o=o0;
        final Color color=o instanceof Color ? (Color)o : null;
        if (o instanceof byte[]) o=new String((byte[])o);
        else if (o instanceof ByteArray) o=((ByteArray)o).newString();
        else if (color!=null) {
            final int len=getDocument().getLength();
            if (color!=lastColor) {
                if (lastColor!=null && lastColor!=Color.WHITE) {
                    if (vColors==null) vColors=new ArrayList();
                    synchronized(vColors) {
                        vColors.add(new Object[]{lastColor,new int[]{lastStart,len}});
                    }
                }
                lastColor=color;
                lastStart=len;
            }
            ChDelay.runAfterMS(EDT,THREAD_NOW,999);
        }
        if (o!=null && color==null) appendNowStrg(o==TextAppender.OVERFLOW ? "\n!" : o.toString());
    }
    final static java.util.List vSplit=new ArrayList();
    private void appendNowStrg(String s) {
        if (sze(s)==0) return;
        vSplit.clear();
        AnsiEscape.extractColors(s,vSplit);
        final int N=sze(vSplit);
        for(int i=0;i<N; i++) {
            final Object o=vSplit.get(i);
            if (o instanceof String) append((String)o);
            else appendNow(o);
        }
        if (gcp(KEY_SCROLL_TO_END,this)!=null && isVisible())  ChDelay.scrollToEnd(this,333);
    }
    public boolean shouldAppend() {return isPainted;}
    private synchronized TextAppender getTextAppender() {
        if (this.textAppender==null) this.textAppender=new TextAppender(this);
        return this.textAppender;
    }
    /* <<< Append  <<< */
    /* ---------------------------------------- */
    /* >>> Threading  >>> */
    public Object run(final String id, Object arg) {
        if (id==ChRunnable.APPEND && arg!=null) appnd(arg);
        if (id=="showInFrame") showInFrame(toStrg(arg),0);
        if (id=="NOW") {
            if (!isEDT()) {
                SwingUtilities.invokeLater(THREAD_NOW);
                return null;
            }
            if (vColors!=null) synchronized(vColors) {
                    final int N=sze(vColors);
                    for(int i=0;i<N;i++) {
                        final Object[] oo=get(i,vColors,Object[].class);
                        final int[] fromTo=(int[]) oo[1];
                        try {
                            final Color color=(Color)oo[0];
                            getHighlighter().addHighlight(fromTo[0],fromTo[1],highlightPainter(color));
                        } catch(Exception e){}
                    }
                    vColors.clear();
                }
        }
        return null;
    }
    /* <<< Threading <<< */
    /* ---------------------------------------- */
    /* >>> AWTEvent >>> */
    { this.enableEvents(ENABLE_EVT_MASK);}
    @Override public void processEvent(AWTEvent ev) {
        final AWTEvent newEv=tools().processEvent(this,ev);
        if (newEv!=null) try { super.processEvent(newEv); } catch(Throwable e){}
    }
    private final transient ChTextComponents tools=new ChTextComponents(this);
    public ChTextComponents tools() { return tools; }
    private java.util.List<Object[]>vColors;
    private final transient Thread THREAD_NOW=ChDelay.newThread(this,"NOW",null);
    /* <<< AWTEvent <<< */
    /* ---------------------------------------- */
    /* >>> Layout  >>> */
    @Override public boolean getScrollableTracksViewportWidth() {
        return gcp(KOPT_TRACKS_VIEWPORT_WIDTH,this)!=null ? true : super.getScrollableTracksViewportWidth();
    }
    public int getFixedRowHeight() { return getFontMetrics(getFont()).getHeight();}
    public ChTextArea showInFrame(String title, long options) {
        if (!isEDT()) ChDelay.invokeAndWait(this,"showInFrame",title);
        else ChFrame.frame(title,this,ChFrame.SCROLLPANE|ChFrame.AT_CLICK|ChFrame.CLEAR_BUTTON|CLOSE_ON_CtrlW|ChFrame.SIZE_REGARD_RESOLUTION).shw(options);
        return this;
    }
    @Override public String getToolTipText(java.awt.event.MouseEvent mev) {
        final String tt=tools().getToolTipText(mev);
        return sze(tt)>0 ? tt :  super.getToolTipText(mev);
    }
    public ChTextArea saveInFile(String fileName) {tools().saveInFile(fileName); return this;}
    @Override public Dimension getPreferredSize() {
        final Dimension ps=getPrefSize(this);
        return  ps!=null ? ps : super.getPreferredSize();
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>>  Get rid of Background. setOpaque not working >>> */
    /*
    @Override public void updateUI() { try { setUI(this.ui=new ChTaUI()); } catch(Exception e){}  }
    private ChTaUI ui;
    private static class ChTaUI extends javax.swing.plaf.basic.BasicTextAreaUI {
        @Override protected void paintBackground(Graphics g) {}
    }
    */
  public java.util.List getDndDateien() { return getDndFilesV(this); }
}
