package charite.christo;
import charite.christo.interfaces.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import java.util.List;
import javax.swing.*;
import javax.swing.event.DocumentEvent;
import javax.swing.text.JTextComponent;
import javax.swing.text.Segment;
import javax.swing.text.Document;
import javax.swing.undo.UndoManager;
import javax.swing.event.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
import static java.awt.event.KeyEvent.*;
import static java.awt.event.MouseEvent.*;
/**
   Utilities for TextComponents
   @author Christoph Gille
*/
public final class ChTextComponents implements  DocumentListener, ChRunnable, PaintHook, UndoableEditListener, ChProcessEvent, HasModiCount  {
    {  assertInstAllowed(); }
    public final static byte[] WRONG_CHARS_IN_FILEN={(byte)'x'};
    public final static String
        KEY_CLASS_AT_CURSOR="CTCs$$_2",
        ACTION_INSERTED_TEXT_LONGER_ONE="CTCs$$I",
        RANGE_SUFX_XML="_VIEW_XML_",
        FILE_RANGE_SUFX[]={"_VIEW_TEXT_", RANGE_SUFX_XML};
    private final static String
        imageRefsS[]={
        "MOVIE:","ARCHSCHEMA", "PRODOM:", "PfamA:", "PFAM:", "GOOGLE{", "JENA:", "PDB_SUM:", "BRENDAec{", "MOL_TOOLKIT:",
        "WIKI{",              "WIKI:",      "PMID",  "PUBMED:",   "big_pubmed",  "PUBMED{",
        "PDB:", "BIONITY:", "WWONLINE:",
        "/", "~", "file:/", "C:", "D:", "c:", "d:", "0","1","2","3","4","5","6","7","8","9", "http://", "ftp://"},
        imageRefsI[]={
            "movie","archschema", "prodom", "pfam", "pfam", null, "big_jenaLib", "big_pdbsum", "brenda_big", "MOL_TOOLKIT",
            "big_wikipedia"+ChIcon.TRANSPARENT, "big_wikipedia"+ChIcon.TRANSPARENT, "big_pubmed", "big_pubmed", "big_pubmed", "big_pubmed",
            "big_pdb", "bionity", "wordwebonline"
        },
        MSG_NO_EVT="Mouse-over events in text-fields can be suppressed either by holding the shift while moving the mouse or by hitting the ESC key";
    public final static Color COLOR_U_GRAY=new Color(0,0,0,144), COLOR_HYPERREFS[]={new Color(32,32,255,128), COLOR_U_GRAY};
    final static String PDF4PMID="&PDF4PMID", DOCID_TMP="ChTextComponents_tmp";
    private boolean noIcons, needsSaving, alreadyPainted, linksDisabled,
        byteArrayValid,segmentValid, alreadyTyped, alreadUlRefs, pmidDownload,
        hlAuto=true,hlMouseEnter=false, ulWebColors,
        changedSinceGainedFocus,
        eventsEnabled=true, doingDnd, isDragging;
    private final ChEventAdapter LI=new ChEventAdapter(this);
    private static ChTextComponents textAreaForAbstract;
    public static void setTextAreaForAbstract(ChTextComponents t) { textAreaForAbstract=t;}
    private DialogFindInText dialogFindInText;
    private static boolean escKey;
    private static int modifiers;
    /* ---------------------------------------- */
    /* >>> Instance >>> */
    private final  Object ref;
    private static ChTextComponents inst;
    private static ChTextComponents instance() {if (inst==null) inst=new ChTextComponents(null); return inst;}
    public ChTextComponents(JComponent jc) {
        ref=weakRef(jc);
        final JTextComponent tc=deref(jc, JTextComponent.class);
        if (jc!=null) {
            if (tc!=null) {
                tc.getDocument().addDocumentListener(this);
                tc.setSelectionColor(C(0xFF, 0x55));
            }
            scrollByWheel(jc);
        }
        if (jc instanceof ChTextArea || jc instanceof AsciiView) addActLi(LI,ChPubmed.class);
        shiftTooltip(jc);
    }

    public JComponent jComponent() { return (JComponent)deref(ref);}
    /* <<< Instance <<< */
    /* ---------------------------------------- */
    /* >>> Undo/Redo >>> */
    private Object undoM;
    private UndoManager undoManager(boolean instantiate) {
        UndoManager m=(UndoManager) deref(this.undoM);
        if (m==null && instantiate) { m=new UndoManager(); this.undoM=undoSoft ? newSoftRef(m) : m; }
        return m;
    }
    private boolean undoSoft;
    public void undoableEditHappened(javax.swing.event.UndoableEditEvent e) {
        undoManager(true).addEdit(e.getEdit());
        // http://java.sun.com/docs/books/tutorial/uiswing/components/generaltext.html#undo
    }
    public ChTextComponents enableUndo(boolean soft) {
        final JTextComponent tc=deref(jComponent(), JTextComponent.class);
        if (tc!=null) tc.getDocument().addUndoableEditListener(this);
        return this;
    }
    /* <<< Undo/Redo <<< */
    /* ---------------------------------------- */
    /* >>> Tab-Completion >>> */
    public final static int COMPL_TAB=1<<1, COMPL_IGN_CASE=1<<2;// COMPL_DEBUG=1<<3;
    private Set<String> compl_offered;
    private Object vComplFirst, vComplLast;
    private Object[] compl_dirListing;
    private long compl_options=0;
    private int compl_lDelim=TAB_COMPL_DELIM, compl_rDelim=TAB_COMPL_DELIM;
    private int compl_lastCaret, compl_nxtIdx[]=NO_INT;
    private FileFilter compl_fileFilter;
    public void setWordCompletionSetFirst(Object collection) { vComplFirst=collection;}

   //  static boolean isXXX(Object o) {
//         if (o instanceof HashMap) return true;
//         if (o instanceof Object[]) for(Object q : oo(o)) if (isXXX(q)) return true;
//         if (o instanceof Collection) for(Object q : (Collection)o) if (isXXX(q)) return true;
//         return false;
//     }

    public ChTextComponents enableWordCompletion(Object collection) {
        //assert !isXXX(collection);
        return enableWordCompletion(collection,0,0, COMPL_TAB);
    }
    public ChTextComponents enableWordCompletion(Object collection, int lDelim, int rDelim, long options) {
        if (collection==null) return this;
        if (collection instanceof FileFilter) compl_fileFilter=(FileFilter)collection;
        else vComplLast=collection;
        compl_lDelim=lDelim!=0 ? lDelim : TAB_COMPL_DELIM;
        compl_rDelim=rDelim!=0 ? rDelim : TAB_COMPL_DELIM;
        compl_options=options;
        final JComponent jc=jComponent();
        if (jc!=null && (compl_options&COMPL_TAB)!=0)  jc.setFocusTraversalKeys( KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS, Collections.EMPTY_SET);
        return this;
    }
    private void compl_clear() {
        if (compl_offered!=null) compl_offered.clear();
        compl_dirListing=null;
        Arrays.fill(compl_nxtIdx,0);
        compl_lastCaret=0;
    }
    private static Object toComplObject(Object o) {
        Object c=null!=tools(o) ? tools(o).byteArray() : o;
        while(!(c instanceof ByteArray) && !(c instanceof List) && c instanceof ChRunnable)  c=((ChRunnable)c).run(PROVIDE_WORD_COMPLETION_LIST, null);
        if (c instanceof HasName) c=((HasName)c).getName();
        else if (c instanceof File) c=((File)c).getName();
        else if (c instanceof JTextComponent) c=((JTextComponent)c).getText();
        return c;
    }
    private boolean completeRecursive(final Object collection0, final int recursion, long option) {
        if (collection0==null) return false;
        final boolean eventsEnabledSave=this.eventsEnabled;
        setTextChangeEventEnabled(false);
        try {
            final boolean ignCase= (option&COMPL_IGN_CASE)!=0;
            final long strstr_opt=ignCase?STRSTR_IC :0L;
            if (compl_offered==null) compl_offered=new HashSet();
            final Object collection=toComplObject(collection0);
            if (sze(compl_nxtIdx)<=recursion) compl_nxtIdx=chSze(compl_nxtIdx,recursion+5);
            final JTextComponent tc=(JTextComponent)jComponent();
            if (compl_lastCaret>0) {
                try {
                    tc.getDocument().remove(compl_lastCaret, tc.getCaretPosition()-compl_lastCaret);
                } catch (Exception e){ putln("Caught in complete()",e);}
            }
            final ByteArray byteArray=byteArray();
            final byte T[]=byteArray.getBytes();
            final int caret=tc.getCaretPosition();
            final int wordStart=prevE(compl_lDelim,T,caret-1,-1)+1;
            final int insOffset=(caret-wordStart);
            String ins=null;
            final String strg=collection instanceof String ? (String)collection : null;
            if (collection instanceof ByteArray || collection instanceof byte[]) {
                final byte[] colT;
                final int B,E;
                if (collection instanceof ByteArray) {
                    final ByteArray ba=(ByteArray)collection;
                    colT=ba.getBytes();
                    E=ba.getEnd();
                    B=ba.getBegin();
                } else { B=0; E=(colT=(byte[])collection).length; }
                for(int jT=maxi(B,compl_nxtIdx[recursion]); jT<E; jT++) {
                    final int iCol=collection==byteArray && jT<caret ? caret-jT-1 : jT;
                    if (wordStart!=iCol && !is(compl_lDelim,colT,iCol) && (iCol==0 || is(compl_lDelim,colT,iCol-1)) && equls(strstr_opt,colT, iCol, iCol+insOffset,  T, wordStart)) {
                        final int wordEnd=nxtE(compl_rDelim, colT,iCol+insOffset,E);
                        if (wordEnd==iCol+insOffset) continue;
                        final String sIns=toStrgIP(colT,iCol+insOffset,wordEnd);
                        if (!compl_offered.contains(sIns)) {
                            compl_offered.add(sIns);
                            ins=sIns;
                            compl_nxtIdx[recursion]=jT+1;
                            break;
                        }
                    }
                }
            } else if (strg!=null || collection instanceof List || collection instanceof Object[]) {
                //runUpdateHooks(collection);
                final int N=strg!=null ? 1 : sze(collection);
                for(int i=compl_nxtIdx[recursion]; i<N; i++) {
                    Object o0=strg!=null ? strg : toComplObject(deref(get(i,collection)));
                    final Object o=collection==compl_dirListing && (tc instanceof JTextComponent) ? rplcToStrg(" ","%20",toStrg(o0)) : o0;
                    if (o==null) continue;
                    else if (o instanceof List || o instanceof ByteArray || o instanceof byte[] || o instanceof Object[]) {
                        if (recursion>10)  putln(RED_WARNING+" word completion: Recursion="+recursion);
                        else if (completeRecursive(o, recursion+1, option)) return true;
                    } else if (o instanceof String) {
                        final int L=sze(o);
                        if (L>insOffset && equls(strstr_opt, T,wordStart,caret, o,0)) {
                            final String s=toStrg(o, insOffset, L);
                            if (!compl_offered.contains(s)) {
                                compl_offered.add(s);
                                ins=s;
                                compl_nxtIdx[recursion]=i+1;
                                break;
                            }
                        }
                    } else {
                        putln(RED_WARNING+" ChTextComponents unsupported word completion collection ",shrtClasNam(o));
                    }
                }
            }
            if (ins!=null) {
                assertEDT();
                try { tc.getDocument().insertString(caret, ins,null); } catch (Exception e){ putln("Caught in complete() ",e);}
                compl_lastCaret=caret;
                return true;
            }
            return false;
        } finally {
            setTextChangeEventEnabled(eventsEnabledSave);
        }
    }
    /* <<< Tab-Completion <<<  */
    /* ---------------------------------------- */
    /* >>> Packages >>> */
    private String[] vPackages;
    void setImportedPackages(String[] pp) { vPackages=pp;}
    private static boolean isCertainlyClassName(byte T[], int from, int to) {
        if (equls(_CC,T,from)) return true;
        if (!is(UPPR,T[from])) return false;
        boolean upper2=false;
        for(int i=from+1; i<to; i++) {
            if (is(UPPR,T[i])) upper2=true;
            if (!is(LETTR_DIGT_US,T[i])&& T[i]!='.') return false;
        }
        return upper2;
    }
    /* <<< Packages <<< */
    /* ---------------------------------------- */
    /* >>> Word under Mouse >>> */
    private static String protDBColonId, classUnderMouse, urlUnderMouse, pluginUnderMouse,wordUnderMouse,
        mMoved_lastWordUnderMouse, idUnderMouse;
    private int wordUnderMouseFom;
    private static File fileUnderMouse;
    private static Range _range;
    private static Range range() { return _range==null?_range=new Range() : _range;}
    private int mouseTextPos=-99999, lastClassFrom=-99999, lastWordFrom=-99999, lastWordTo=-99999, mMoved_lastModi;
    private void setMousePosition(int iPosition) {
        final String lastW=this.wordUnderMouse;
        final JComponent jc=jComponent();
        final ByteArray ba=byteArray();
        final byte[] T=ba.getBytes();
        /* --- java class --- */
        final Range rangeClass=getWordAt(iPosition,ba,ChCodeViewer.LEFT_CLASS_DELIM,ChCodeViewer.CLASS_DELIM,range());
        if (rangeClass!=null) {
            final int from=rangeClass.getFrom(), to=rangeClass.getTo();

            if (this.lastClassFrom!=from || this.classUnderMouse==null) {
                this.lastClassFrom=from;
                boolean dotCapital=false;
                for(int i=from+1; i<to; i++) dotCapital=dotCapital || T[i-1]=='.' && is(UPPR,T[i]);
                if ( (dotCapital&&is(LOWR,T[from])||is(UPPR,T[from])) && (isCertainlyClassName(T,from,to) || gcp(KEY_CLASS_AT_CURSOR,jc)!=null)) {
                    if (vPackages==null) vPackages=ChCodeViewer.getImportedPackages(ba);
                    this.classUnderMouse=fullClassName(ba.newString(from,to),vPackages);
                } else this.classUnderMouse=null;
            }
        } else this.classUnderMouse=null;
        final Range r=getWordAt(iPosition,ba," \t\n\r\"("," \t\n\r\")",range());
        if (r!=null) {
            final int from=r.getFrom(), to0=r.getTo();
            final int to=(get(to0,T)==')' && equls("WIKI",T, from)) ? to0+1 : to0;
            final boolean semicolonSpace=from>8 && T[from-1]==' ' && T[from-2]==';';
            if (this.lastWordFrom!=from || this.lastWordTo!=to || lastW==null || lastW.indexOf(':')>0 || semicolonSpace) {
                final String w=ba.newString(from,to);
                final int colon=w.indexOf(':');
                final boolean looksLikeMail=colon<0 && looks(LIKE_MAILADDR,w);
                //this.ecUnderMouse=looks(LIKE_EC,w, colon+1,MAX_INT) ? w.substring(colon+1, nxtE(NO_DIGT_DOT,w,colon+1,MAX_INT)) : null;
                if (looksLikeMail) { this.urlUnderMouse="mailto:"+w; this.wordUnderMouse=w; this.wordUnderMouseFom=from; }
                else if (!w.equals(lastW) || colon>0 || semicolonSpace) {
                    this.wordUnderMouse=w;
                    this.wordUnderMouseFom=from;
                    this.urlUnderMouse=this.protDBColonId=this.pluginUnderMouse=null;
                    this.fileUnderMouse=null;
                    if (sze(w)>2) {
                        {
                            final int fFrom=T[from]=='[' ? from+1 : from;
                            final int fTo=fileEndsAt(T,fFrom,ba.getEnd());
                            if (fTo>0) this.fileUnderMouse=myFile(ba.newString(fFrom,fTo));
                        }

                        this.idUnderMouse=
                            looks(LIKE_EC,w, colon+1,MAX_INT) ? w.substring(colon+1, nxtE(NO_DIGT_DOT,w,colon+1,MAX_INT)) :
                            w.startsWith("PDB:") || w.startsWith("UNIPROT:") || guessProteinDbFromId(w)=="UNIPROT:" ? w :
                            null;
                        String emblDB=null;
                        String url=Hyperrefs.toUrlString(delSfx('.',w),Hyperrefs.WEB_LINK|Hyperrefs.PLAIN_TEXT|Hyperrefs.PROTEIN_FILE);
                        if (url!=null && url.endsWith("?plugins")) this.pluginUnderMouse=url;
                        EMBL_links:
                        if (semicolonSpace) { /* --- EMBL; U13634; AAB03370.1;  --- */
                            final int from0=maxi(0,from-72);
                            for(int semicolon=from-1; --semicolon>from0;) {
                                if (T[semicolon]!=';' || T[semicolon+1]!=' ') continue;
                                for(int space=semicolon; --space>from0;) {
                                    if ( T[space]!=' ') continue;
                                    final int l=semicolon-space;
                                    nextKey:
                                    for(String key : Hyperrefs.getDatabases(Hyperrefs.WEB_LINK|Hyperrefs.PLAIN_TEXT|Hyperrefs.PROTEIN_FILE)[0]) {
                                        if (sze(key)!=l || !key.endsWith(":")) continue;
                                        for(int i=l-1; --i>=0;) if (key.charAt(i)!=T[space+1+i]) continue nextKey;
                                        emblDB=key+delSfx('.',delSfx(';',w));
                                        url=Hyperrefs.toUrlString(emblDB,Hyperrefs.WEB_LINK|Hyperrefs.PLAIN_TEXT|Hyperrefs.PROTEIN_FILE);
                                        break EMBL_links;
                                    }
                                }
                            }
                        }
                        final int middle=emblDB!=null ? sze(w)/2 : colon;
                        this.urlUnderMouse=url;
                        final String e_or_w=emblDB!=null ? emblDB : w;
                        if (middle>=0
                            && iPosition-from>middle
                            && gcp(ACTION_DATABASE_CLICKED,jc)!=null
                            && w.indexOf("://")<0
                            && ( Hyperrefs.toUrlString(e_or_w,Hyperrefs.PROTEIN_FILE)!=null || e_or_w.startsWith("PDB:"))
                            ) { this.protDBColonId= emblDB!=null ? emblDB : w;}
                        this.protDBColonId=rplcToStrg("DIP:DIP:","DIP:",this.protDBColonId);
                    }
                }
            }
        } else {
            this.wordUnderMouse=this.urlUnderMouse=this.protDBColonId=this.pluginUnderMouse=this.idUnderMouse=null;
            this.fileUnderMouse=null;
            this.lastWordFrom=this.lastWordTo=-99999;
        }
    }

    public String getWordUnderMouse(Range r) {
        final String w=this.wordUnderMouse;
        if (r!=null) r.set(w!=null ? this.lastWordFrom:0, w!=null?this.lastWordTo:0);
        return w;
    }
    /* <<< Word under Mouse <<< */
    /* ---------------------------------------- */
    /* >>> Word at position >>> */
    private static Rectangle RECTf=new Rectangle(), RECTt=new Rectangle();
    private static Range getWordAt(Point p, JComponent tc,String lDelimiters,String rDelimiters,Range range) {
        final int iCaret=viewToModel(p,tc);
        final ChTextComponents tools=tools(tc);
        if (tools==null) return null;
        getWordAt(iCaret, tools.byteArray(),lDelimiters,rDelimiters,range);
        final int f=range==null ? 0:range.getFrom(), t=range==null ? 0:range.getTo();
        if (t-f==0) return null;
        modelToView(f,tc, RECTf);
        modelToView(t,tc, RECTt);
        if (RECTf.height>0 && RECTt.height>0) {
            final boolean contains=  RECTf.x<RECTt.x ?
                contains(p, RECTf.x,RECTf.y, RECTt.x+RECTt.width,  RECTt.y+RECTt.height-1) :
                contains(p, RECTf.x, RECTf.y, 99999, RECTf.y+RECTf.height-2) ||
                contains(p, 0,RECTt.y, RECTt.x, RECTt.y+RECTt.height-2);
            return !contains ? null : range;
        }
        return null;
    }
    public static String getWordAt(Point p, JComponent tc,String lDelimiters, String rDelimiters) {
        final Range r=getWordAt(p,tc,lDelimiters,rDelimiters,range());
        final ChTextComponents tools=tools(tc);
        if (r==null || tools==null) return null;
        final int f=r.getFrom(), t=r.getTo();
        final ByteArray ba=tools.byteArray();
        final byte[] bb=ba.getBytes();
        return f>=0 && f<t && t<bb.length ?new String(bb,f,t-f) : null;
    }
    public static String getWordAt(final int iCaret, ByteArray ba,String lDelimiters,String rDelimiters) {
        final Range r=getWordAt(iCaret,ba,lDelimiters,rDelimiters,range());
        if (r==null) return null;
        final int f=r.getFrom(), t=r.getTo();
        final byte[] bb=ba.getBytes();
        return f>=0 && f<t && t<bb.length ?new String(bb,f,t-f) : null;
    }
    public static Range getWordAt(final int iCaret, ByteArray ba,String lDelimiters,String rDelimiters, Range range) {
        if (ba==null || iCaret<0) return null;
        final String lD= lDelimiters==null || lDelimiters=="" ? " \t\n\r(" : lDelimiters;
        final String rD= rDelimiters==null || rDelimiters=="" ? " \t\n\r)" : rDelimiters;
        final byte[] bb=ba.getBytes();
        final int len=ba.getEnd();
        int l,r;
        for(l=iCaret; l> 0 && l<=len && lD.indexOf((char)bb[l-1])<0; l--);
        for(r=iCaret; r>=0 && r< len && rD.indexOf((char)bb[r  ])<0; r++);
        l=mini(maxi(0,l),len);
        r=mini(maxi(0,r),len);
        final int bracket=strchr('{',bb,l,r);
        if (bracket>0) {
            for(String key: Hyperrefs.getKeysWithBrace()) {
                if ( equls(key,bb,l)) {
                    r=strchr('}',bb,l,len);
                    if (r<0) return null;
                    break;
                }
            }
        }
        if (l>=r || r>len || l<0) return null;
        range.set(l,r);
        return range;
    }
    /* <<< Word at position <<< */
    /* ---------------------------------------- */
    /* >>> Cursor >>> */
    private Cursor origCursor;
    public Cursor getCursor() {
        final JComponent jc=jComponent();
        if (jc==null) return getCursr(Cursor.DEFAULT_CURSOR);
        if (origCursor==null) origCursor=jc.getCursor();
        final String w=this.wordUnderMouse;
        Cursor curs=null;
        final int m=this.modifiers;
        final boolean shift=(m&SHIFT_MASK)!=0, ctrl=(m&CTRL_MASK)!=0, alt=(m&ALT_MASK)!=0;
        if (w!=null) {
            if (w.startsWith("PMID")) curs=getCursr(Cursor.DEFAULT_CURSOR);
            else {
                final Object c=
                    linksDisabled?null:
                    this.classUnderMouse!=null ? (
                                                  !(ctrl||alt) ? "Source\nJavadoc" :
                                                  shift ? "Edit" :
                                                  ctrl ? "Source" : "Javadoc" ):

                    fileUnderMouse!=null ? ( isExeFileName(fileUnderMouse) ? "run" : systProprty(SYSP_ICON_FILE_BROWSER)):
                    looks(LIKE_MAILADDR, w) ? "Mail" :
                    protDBColonId!=null ? gcp(ACTION_DATABASE_CLICKED,jc) :
                    pluginUnderMouse!=null ? "Load" :
                    urlUnderMouse!=null || idUnderMouse!=null || equls("MOVIE:",w) ? IC_FIREFOX16 :
                    null;
                curs=ChIcon.getCursor(toStrg(c),jc);
            }
        }
        return curs!=null ? curs : origCursor;
    }
    /* <<< Cursor <<< */
    /* ---------------------------------------- */
    /* >>> Contextmenu >>> */
    private String textForContextMenu;
    private JPopupMenu contextMenu;
    private Object[] getContextMenuObjects(String title) {
        final Object oo[]={
            title,
            new ChButton("W_CPY").li(LI).t("Copy word to clip-board (Paste: Ctrl-V)"),
            new ChButton("W_WEB").li(LI).t("Web-Search word under mouse").i(IC_WWW),
            "-",
            new ChButton("W_HL").li(LI).t("Highlight word under mouse"),
            new ChButton("W_RMHL").li(LI).t("Remove Highlight"),
            "-",
            "&Rc",new ChButton("EXEC").li(LI).t("Execute shell command").i(IC_BATCH),
            "For opening the control panel, click with ctrl key pressed",
            "-",
            new ChButton("H").t("Help on text components").li(LI)
        };
        return oo;
    }

    public static String[] webLinks(String id, String[] customSettings) {
        final String cc[]=customSettings, ss[]=new String[cc.length];
        for(int i=cc.length; --i>=0;) {
            final String c=cc[i].trim();
            if (c.indexOf('*')>0) ss[i]=rplcToStrg("*", id,c);
            else if (sze(c)>0) ss[i]=c+id;
        }
        return ss;
    }

    private final static ChTextArea taWebLinks[]=new ChTextArea[Customize.MAX_ID+1];
    public static void showFrameWeblinks(int type, CharSequence txt, Object pSouth) {

        ChTextArea ta=taWebLinks[type];
        if (ta==null) {

            Object pS=pSouth=="B" ? Customize.customize(type).newButton().t("Customize").i(null) : pSouth;
            if (pSouth instanceof Customize[]) {
                pS=pnl();
                for(Customize c : (Customize[])pSouth) {
                    if (c!=null) ((JComponent)pS).add(c.newButton().t("Customize").i(null));
                }
            }

            ta=taWebLinks[type]=new ChTextArea("");
            pcp("PNL",newCNSEW(ta,null, newCNSEW(null,null,null,pS,null),null,null), ta);
        }
        tools(ta.t("\n"+txt+"\n")).underlineRefs(UNDERLINE_REFS|UNDERLINE_REFS_NO_ICON);
        final JComponent pnl=gcp("PNL",ta,JComponent.class);
        ChFrame.frame("Links",pnl,ChFrame.ALWAYS_ON_TOP).shw(ChFrame.PACK|ChFrame.AT_CLICK|ChFrame.TO_FRONT);
        setDragMovesWindow(true, pnl);
    }
    /* <<< Context <<< */
    /* ---------------------------------------- */
    /* >>> Tip >>> */
    private static String tt_lastW, tt_lastTT;
    public String getToolTipText(MouseEvent ev) {
        final String w=this.wordUnderMouse;
        if (tt_lastW!=w) {
            tt_lastW=w;
            String tt=null;
            if (equls("MOVIE:",w)) tt="Open movie in Browser";
            else if (equls("PMID",w)) {
                final int pmid=atoi(w,4);
                tt="<html><body>"+
                    ChPubmed.pubmedTitle(pmid)+
                    ("<br>Left-Click: Abstract in Browser<br>Right-click: PDF<br>"+
                     "Colors: blue indicates presence of Abstract, red presence of PDF<br>")+
                    (pmidDownload?"Alt-mouse-hover: view PDF as plain text" : "")+
                    "</body></html>";
            }
            else if (urlUnderMouse!=null) tt=urlUnderMouse;
            else tt=toStrg(appendTooltips(jComponent(),ev,null));
            if (tt==null && w!=null) {
                if (w.startsWith("$CTIME=")||w.startsWith("$MTIME=")) {
                    final long t=atol(w,7,99);
                    tt=day(t<MAX_INT ?t*1000 :t);
                }
            }
            this.tt_lastTT=tt;
        }
        return "null".equals(this.tt_lastTT) ? null : this.tt_lastTT;
    }
    /* <<< Tip <<< */
    /* ---------------------------------------- */
    /* >>> AWTEvent >>> */
    private static List<ActionListener> vLiDb;
    public static void addLiDb(ActionListener li) { adUniq(li, vLiDb==null?vLiDb=new ArrayList() : vLiDb); }

    private void he(AWTEvent ev) {
        handleEvt(this,ev);
        handleEvt(jComponent(),ev);
    }
    private Cursor lastCursor;
    private AWTEvent actEvt;
    private MouseEvent mousePressEvent;
    public AWTEvent processEvent(final JComponent jComponent, final AWTEvent ev) {
        final Object q=ev.getSource();
        final String cmd=actionCommand(ev);
        final JComponent jc=jComponent();
        final int id=ev.getID();
        final int modi=this.modifiers=modifiers(ev), iButton=buttn(ev);        
        if (cmd!=null) {
            final int iB=idxOf(q,_of_buttons);
            final String path=iB<0?null:getTxt(_of_lab);
            if (sze(path)>0) {
                final File f=file(path);
                //putln("path="+path+"    "+url(path));
                if (iB==OF_BUT_F && f!=null) viewFile(f, modi);
                if (iB==OF_BUT_f && f!=null) viewFile(file(f.getParent()), modi);
                if (iB==OF_BUT_U) visitURL(path, modi);
                if (iB==OF_BUT_u) visitURL(withoutLastComponent(path,'/'), modi);
            }
        }        
        if (!shouldProcessEvt(jc,ev)) return null;

        final JTextComponent tc=jc instanceof JTextComponent ? (JTextComponent)jc : null;
        final AsciiView ascii=jc instanceof AsciiView ? (AsciiView)jc:null;
        final ChTextArea ta=jc instanceof ChTextArea ? (ChTextArea)jc : null;
        final ChTextField tf=jc instanceof ChTextField ? (ChTextField)jc : null;
        final MouseEvent mev=ev instanceof MouseEvent ? (MouseEvent)ev : null;
        final ByteArray ba=byteArray();
        final byte[] T=ba.getBytes();
           final boolean shift=(modi&SHIFT_MASK)!=0, ctrl=isCtrl(ev), alt=(modi&ALT_MASK)!=0,
            shortCut=ctrl||0!=(modi&shortCutMask());
        final boolean isHtml=tc!=null && tc.getDocument() instanceof javax.swing.text.html.HTMLDocument;
        /*
        if (id==MOUSE_PRESSED) {
            final int msc=getDefaultTK().getMenuShortcutKeyMask();
            putln("\n"+ev+"\nMOUSE_CLICKED ctrl="+(0!=(modi&CTRL_MASK))+
                  " shift="+shift+
                  "   alt="+alt+
                  " shortCutMas="+(0!=(modi&shortCutMask()))+
                  " togglesel="+(0!=(modi&toggleSelectionMask()))+
                  " msc="+(0!=(modi&msc))+
                  " buttn="+iButton+" modi="+Integer.toBinaryString(modi)
                  );
        }
        */

        final int scaleUpDown=scaleUpDown(ev);
        if (scaleUpDown!=0) {
            final double f=isHtml ? 1.05 : 1.1;
            enlargeFont(jc,scale*= (scaleUpDown<0 ? 1/f : f));
            return null;
        }
        if (id==MouseEvent.MOUSE_DRAGGED && !shift) {
            if (!doingDnd && !isDragging) {
                final File file=this.fileUnderMouse;
                final java.net.URL url=url(this.urlUnderMouse);
                final int pos=this.mouseTextPos;
                Object export=(pos-this.wordUnderMouseFom<3 || this.wordUnderMouseFom+sze(this.wordUnderMouse)-pos<3) ? null : file!=null ? file : url;
                if (export!=null) {
                    doingDnd=true;
                    getDndFilesV(jc).clear();
                    getDndFilesV(jc).add(export);
                    exportDrg(jc, mev, TransferHandler.COPY);
                }
            }
            if (doingDnd) return null;
            isDragging=true;
        } else doingDnd=isDragging=false;
        final int kcode=keyCode(ev);
        if (id==KEY_PRESSED) {
            alreadyTyped=true;
            if (ChFrame.setAotEv(ev)) return null;
            changedSinceGainedFocus=true;
            final char kchar=keyChar(ev);
            if ( (kcode==VK_V || kcode==VK_X) && shortCut ) changed(null);
            if (kcode==VK_ESCAPE) escKey=true;
            if (ctrl && !shift && !alt &&  isMacintoshLaf() &&
                (kcode==VK_C || kcode==VK_X || kcode==VK_V || kcode==VK_A)) {
                try {
                    return new KeyEvent(tc,id,System.currentTimeMillis(),shortCutMask(),kcode,kchar);
                } catch(IllegalArgumentException e) {}
            }
            if (!ctrl && !shift && kcode==VK_F7 && !jOrtho && tc!=null && (jOrtho_prepared || ChMsg.yesNo("Start spell Check?")) ) {
                jOrtho_prepared=true;
                jOrtho=true;
                new ChJOrthoPROXY().register(tc);
            }
            if (!ctrl && !shift && kcode==VK_F6) {
                for(int i : new int[]{Customize.webLinks, Customize.webBrowser, Customize.fileBrowsers}) {
                    Customize.addDialog(Customize.customize(i));
                }
            }
            if (kcode==VK_F1) {
                if(shift && tc!=null) {
                    final File f=createTempFile("");
                    wrte(f,toStrg( get(0,gcp(ChTextViewer.KEY_TEXT_CSS_JS,jc))));
                    edFile(-1, f, modi);
                } else showHelp(HyperTextPane.class);
                return null;
            }
            if (undoM!=null && shortCut) {
                final UndoManager um= undoManager(false);
                if (um!=null) {
                    try {
                        if (kcode==VK_Y) um.redo();
                        if (kcode==VK_Z) if (shift) um.redo(); else um.undo();
                    } catch(Exception e) {}
                }
            }
            if (((compl_options&COMPL_TAB)!=0 && kcode==VK_TAB || (kcode==VK_ENTER && ctrl && !alt && !shift) || kcode==VK_SLASH && alt) && tc!=null) {
                File dir_=vComplLast instanceof File ? (File) vComplLast : null;
                if (dir_==null && (vComplLast instanceof List || vComplLast instanceof Object[])) {
                    for(int i=sze(vComplLast); --i>=0;) if ( (dir_=get(i,vComplLast,File.class))!=null) break;
                }
                final File
                    dir0=vComplLast instanceof JFileChooser ? ((JFileChooser)vComplLast).getCurrentDirectory() : null,
                    dir1=dir_!=null ? dir_ : jc instanceof JTextField && dir0!=null  ? file(dir0+"/"+toStrg(jc).trim()) : dir0,
                    currentDir=dir1!=null && dir1.exists() ? dir1 : dir0;
                final int caret=tc.getCaretPosition();
                final int from=prevE(SPC, T,caret-1,-1)+1;
                final boolean isPath=looks(LIKE_FILEPATH_EVALVAR,T,from);
                Object ff[]=compl_dirListing;
                if (ff==null) {
                    if (currentDir!=null || isPath) {
                        final int to=nxtE(SPC, T,caret, ba.getEnd());
                        final int slash=prevE(SLASH, T, to,from);
                        final int endDir=slash>0?slash : to;
                        final String sDir=from==endDir ? "/" : endDir-from==1 && T[from]=='.' ? "./" : ba.newString(from,endDir);
                        final File dir=isPath ? file(sDir) : currentDir;
                        if (dir!=null && dir.isDirectory()) {
                            ff=compl_dirListing= compl_fileFilter!=null ? dir.listFiles(compl_fileFilter) : dir.list();
                            sortArry(ff, comparator(COMPARE_ALPHABET));
                        }
                    }
                }
                final long ignCase=compl_options&COMPL_IGN_CASE;
                final boolean notOnlyDir=!isPath || sze(compl_dirListing)==0;
                if (!(completeRecursive(compl_dirListing,0, 0)  ||
                      notOnlyDir && (
                                     completeRecursive(Hyperrefs.getKeysToHighlight(), 1, 0)  ||
                                     completeRecursive(vComplFirst,2,ignCase) ||
                                     (tf==null && completeRecursive(ba,3, ignCase)) ||
                                     vComplFirst!=vComplLast && completeRecursive(vComplLast,4,ignCase) ||
                                     completeRecursive(ChEnvironment.getNames(ChEnvironment.WITH_DOLLAR),5, 0)  )
                      )
                    ) compl_clear();
                return null;
            } else if (kcode!=VK_SPACE)  compl_clear();
            if (ctrl)  {
                switch(kcode) {
                case VK_P:
                    final String clazz=gcps(KEY_CLASS_NAME, jc);
                    File fOut=clazz!=null ? HtmlDoc.main1(clazz) : null;
                    if (jc instanceof AsciiView) fOut=((AsciiView)jc).getFile();
                    if (fOut==null) {
                        fOut=createTempFile(isHtml ? ".html":".txt");
                        if (isHtml) {
                            try {
                                final Object t_css_js=gcp(ChTextViewer.KEY_TEXT_CSS_JS,jc);
                                final StringBuffer txt=toStringBuffer(get(0,t_css_js));
                                rplcSb(0L,"<CAPTION><BR>","<CAPTION>",txt);
                                final int posCSS=atoi(get(1,t_css_js)), posJS=atoi(get(2,t_css_js));
                                if (posJS >0) txt.insert(posJS,  HtmlDoc.getCssOrJS("js",  DOCTYPE_HTML,true));
                                if (posCSS>0) txt.insert(posCSS, HtmlDoc.getCssOrJS("css", DOCTYPE_HTML,true));
                                final String[] imported=HtmlDoc.getImportedPackages(gcps(KEY_ORIGINAL_TEXT,jc), clazz);
                                HtmlDoc.process(txt,   DOCID_TMP, DOCTYPE_HTML, imported);
                                HtmlDoc.rplcWIKI(txt);
                                addHtmlTags(txt);
                                wrte(fOut, txt);
                                writePng(null,null);
                            } catch(Throwable t){stckTrc(t);}
                        } else wrte(fOut, ba);
                    }
                    if (sze(fOut)>0) {
                        if (isHtml) visitURL(fOut,0);
                        else edFile(-1, fOut, 0);
                    }
                    return null;
                case VK_A:
                    if (ascii!=null) {
                        final int ft[]=ascii.region().fromTo;
                        ft[0]=0;
                        ft[1]=MAX_INT;
                        ascii.repaint();
                        toClipbd(ba.newString());
                        return null;
                    }
                    return ev;
                case VK_U:
                    if (jc!=null) jc.updateUI();
                    if (ascii!=null && ascii.getFile()!=null) ascii.setFile(ascii.getFile());
                    underlineRefs(UNDERLINE_REFS);
                    return null;
                }
            }
            if (!shift && (kcode==VK_F && shortCut || kcode==VK_F3 && !ctrl)) {
                if (!(jc instanceof JTextField)) {
                    (dialogFindInText==null ? dialogFindInText=new DialogFindInText(jc) : dialogFindInText).showInFrame(ChFrame.AT_CLICK);
                }
                return null;
            }
            if (ctrl && kcode==VK_L || !ctrl && kcode==VK_F10) {
                if (jc instanceof HyperTextPane) ((HyperTextPane)jc).toggleLineWrap();
                if (ta!=null) ta.setLineWrap(!ta.getLineWrap());
            }
        }
        {
            if (id==MOUSE_PRESSED && shift && !ctrl) {
                this.mousePressEvent=mev;
                return null;
            }
            final MouseEvent le=this.mousePressEvent;
            this.mousePressEvent=null;
            if (id==MOUSE_RELEASED && shift && le!=null) return le;
        }
        if (id==MOUSE_MOVED)  { evLstnr(MOMOLI_DRAG_WITH_ALT).mouseMoved(mev); }
        if (id==MOUSE_DRAGGED && (iButton==2 || shift)) {
            evLstnr(MOMOLI_DRAG_WITH_ALT).mouseDragged(mev);
            return null;
        }

        final int x=x(ev), y=y(ev);
        final int iP=x>=0 && y>=0 ?  viewToModel(new Point(x,y),jc) : -1;

        final boolean isNewCharacter;
        if (id==MOUSE_MOVED || id==MOUSE_PRESSED || id==MOUSE_CLICKED)  {
            modelToView(iP+1,jc,RECTf);
            final int iPosition=RECTf.height==0 || RECTf.x<x ? -1 : iP;
            if (isNewCharacter=this.mouseTextPos!=iPosition) setMousePosition(this.mouseTextPos=iPosition);
        } else isNewCharacter=false;
        final String w=this.wordUnderMouse;
        final int colon=w!=null ? w.indexOf(':') : -1;
        if ((id==MOUSE_MOVED || id==MOUSE_PRESSED || id==MOUSE_CLICKED) && isNewCharacter && !linksDisabled && !escKey) {
            if (w!=null && pmidDownload  && !shift && !(w.equals(this.mMoved_lastWordUnderMouse)&&modi==this.mMoved_lastModi)) {
                final String d= colon<0 && guessProteinDbFromId(w)=="UNIPROT:" ? "UNIPROT:"+w : w;
                final boolean isPM=d.startsWith("PUBMED:") || d.startsWith("PMID");
                if (isPM || d.startsWith("SWISS:") || d.startsWith("UNIPROT") || d.startsWith("ENSG")) {
                    final File f=ChPubmed.maybeDownload(d);
                    if (f!=null) {
                        final ChTextComponents t=textAreaForAbstract;
                        if (t!=null) ChPubmed.showAbstract(f,t, (modi&ALT_MASK)!=0 ? ChPubmed.SHOWABSTRACT_FULL_TEXT : 0L);
                        ChMsg.savedMsg(MSG_NO_EVT, "MSG_NO_EVT");
                    }
                    if (isPM && d.indexOf('{')<0) ChPubmed.queueForDownload(d,true);
                }
            }
            this.mMoved_lastWordUnderMouse=w;
            this.mMoved_lastModi=modi;
        }

        if (id==MOUSE_PRESSED) {
            this.textForContextMenu=w;
            {
                if (isPopupTrigger(ev) || isClassName(this.classUnderMouse)) {
                    if (w!=null && (w.startsWith("PMID")||w.startsWith("PUBMED:"))) {
                        ChPubmed.showPubmedPopup(w,jc, x,y);
                    }
                    else {
                        final ChRunnable contextMenuProvider=gcp(PROVIDE_JPopupMenu, jc, ChRunnable.class);
                        JPopupMenu menu=
                            classUnderMouse!=null ? JavaclassPopup.menu(classUnderMouse,(String[])null) :
                            !shift&&!alt&&!ctrl && contextMenuProvider!=null ? (JPopupMenu) contextMenuProvider.run(PROVIDE_JPopupMenu,ev) : null;
                        if (menu==null) {
                            final String msg="This context menu acts on a text region selected with the mouse or on the word under the mouse pointer";
                            menu= contextMenu==null ?  contextMenu=jPopupMenu(0L, "Clicked",getContextMenuObjects(msg)) : contextMenu;
                        } else {
                            final String key="TC$$gm";
                            JMenu m=gcp(key,menu,JMenu.class);
                            if (m==null) pcp(key, m=jMenu(0L, getContextMenuObjects("General menu"),null), menu);
                            menu.add(m);
                        }
                        menu.show(jc.isShowing() ? jc:ChFrame.getMainFrame(), x,y);
                    }
                } else if (!linksDisabled) {
                    final String urlMouse0=this.urlUnderMouse, idMouse=this.idUnderMouse, dbE=this.protDBColonId;
                    final int pmidStart=strstr(STRSTR_AFTER,PDF4PMID,urlMouse0), iPmid=atoi(urlMouse0, pmidStart);
                    final String urlMouse=iPmid>0 ? rplcToStrg(".pdf+html", ".pdf",withoutLastComponent(urlMouse0,'&')) : urlMouse0;
                    final boolean
                        dbClicked=urlMouse!=null && dbE!=null && (iPmid<=0 || gcp(ACTION_DATABASE_CLICKED,jc)!=null && !dbE.startsWith("PUBMED") && !equls(STRSTR_IC,"PMID",dbE)),
                        isAfterColon=colon>0 && iP-range().getFrom()>colon,
                        isPDB=idMouse!=null && idMouse.startsWith("PDB:");
                    //puts(" dbClicked="+dbClicked+"    "+urlMouse);
                    final int db=
                        idMouse==null || isAfterColon ? 0 :
                        isPDB ? Customize.pdbLinks :
                        idMouse.startsWith("UNIPROT:") || guessProteinDbFromId(idMouse)=="UNIPROT:" ? Customize.uniprotLinks :
                        looks(LIKE_EC,idMouse) ? Customize.ecClassLinks : 0;
                    if (db!=0 && !shift && !ctrl && !alt) {
                        String id_noDB=delPfx("PDB:",delPfx("UNIPROT:",idMouse));
                        if (isPDB) id_noDB=withoutLastComponent(withoutLastComponent(id_noDB,':'),'_');
                        showFrameWeblinks(db,joinLines(webLinks(id_noDB, getCustomSettings(db))),"B");
                    } else if (pluginUnderMouse!=null) {
                        InteractiveDownload.downloadFilesAndUnzip(new java.net.URL[]{url(pluginUnderMouse)},dirPlugins());
                        he(new ActionEvent(this,ActionEvent.ACTION_PERFORMED, ACTION_DOWNLOAD_PLUGIN,0));
                    } else if (urlMouse!=null) {
                        final int iQu=strchr('?',urlMouse);
                        String pattern=null;
                        if (dbClicked){
                            final String command=ACTION_DATABASE_CLICKED+delSfx(',',delSfx('.',dbE));
                            handleEvt(vLiDb, new ActionEvent(jc, ActionEvent.ACTION_PERFORMED,command,0));
                        } else if (idxOfStrg(urlMouse, getCustomSettings(Customize.testProxySelector))>=0) {
                            Web.testConnectionForURL(urlMouse);
                        } else if (equls(pattern="http://localhost:", urlMouse) && iQu>0) {
                            Web.thread_sendDataToPort(atoi(urlMouse, sze(pattern)), urlDecode(urlMouse.substring(iQu+1)), wordUnderMouse).start();
                        } else {
                            dialogOpenFile(0L, delSfx('.',urlMouse));
                            ChPubmed.maybeAssociatePdf(iPmid, false);
                        }
                    } else {
                        for(ChTextArea taL : taWebLinks) if (jc!=null && jc==taL) ChFrame.frame("",taL,0).setVisible(false);
                    }
                    {
                        final File f=this.fileUnderMouse;
                        if (f!=null&&w!=null) {
                            boolean success=false, isXML=false;
                            int viewR=-1;
                            for(String V : FILE_RANGE_SUFX) {
                                if (viewR<0) { viewR=strstr(STRSTR_AFTER,V,w); isXML=V.indexOf("XML")>0; }
                            }
                            if (viewR>0) {
                                final int from=atoi(w,viewR), to=mini(atoi(w, w.indexOf('-',viewR)+1), sze(f));
                                if (from>0 && to>from) {
                                    InputStream is=null;
                                    try {
                                        is=new FileInputStream(f);
                                        is.skip(from);
                                        final byte[] buf=new byte[to-from];
                                        final int read=is.read(buf);
                                        final ByteArray byteArray=new ByteArray(buf,0,read);
                                        if (sze(byteArray)>0) {
                                            final AsciiView v=new AsciiView(byteArray);
                                            if (isXML) v.highlightXML(byteArray);
                                            v.showInFrame(filepathUnixStyle(f),ChFrame.STAGGER);
                                            success=true;
                                        }
                                    } catch(IOException ex) { stckTrc(ex); }
                                    closeStrm(is);
                                }
                            }
                            if (!success) dialogOpenFile(0L,toStrg(f));  //viewFile(f, modi);
                        }
                    }
                    if (iButton==2)  changed(null);
                    if (dialogFindInText!=null) dialogFindInText.setPosition(iP);
                    if (iButton==2) ChDelay.runAfterMS(EDT,run('M'), 333);
                }
            }
        }
        if (id==MOUSE_ENTERED && q==jc) {
            escKey=false;
            this.wordUnderMouse=/*this.classUnderMouse=*/this.urlUnderMouse=this.idUnderMouse=this.protDBColonId=this.pluginUnderMouse=null;
            this.fileUnderMouse=null;

            if (0!=(ul_options&UNDERLINE_REFS_DISABLE_UNTIL_MOUSE) && !hlAuto) { hlAuto=true;  run('H').run(); }
            if (hlMouseEnter) run('H').run();
        }
        if (id==FocusEvent.FOCUS_GAINED) changedSinceGainedFocus=false;
        if (id==FocusEvent.FOCUS_LOST || id==MOUSE_EXITED || (tf!=null && id==ActionEvent.ACTION_PERFORMED) ) {
            if (needsSaving)  {
                if (saveInFile_loaded) saveInFile(true);
                final Object contentType= tf!=null ? tf.getContentType() : null;
                if (contentType==KEY_EMAIL) setPrpty(ChUtils.class, (String)contentType,jc.toString());
            }
            if (gcp(KEY_IF_EMPTY,jc)!=null) jc.repaint();
        }
        final String act=
            alreadyTyped && id==FocusEvent.FOCUS_LOST && changedSinceGainedFocus ? ACTION_FOCUS_LOST :
            gcp(ACTION_FOCUS_GAINED,jc)!=null && id==FocusEvent.FOCUS_GAINED ? ACTION_FOCUS_GAINED :
            tf!=null && id==KEY_PRESSED && keyChar(ev)=='\n' ? ACTION_ENTER : null;
        if (act!=null && System.currentTimeMillis()-timeEnter>99)  {
            he(new ActionEvent(jc,ActionEvent.ACTION_PERFORMED,act,0));
            timeEnter=System.currentTimeMillis();
        }
        if (tf!=null && act!=null) {
            final Object ct=tf.getContentType();
            if (ct==Integer.class) setTextTS(String.valueOf(atoi(tf)));
            if (ct==Float.class) setTextTS(String.valueOf(atof(tf)));
        }
        if ((id==MOUSE_MOVED || id==KEY_PRESSED || id==KEY_RELEASED) && jc!=null && lastCursor!=getCursor()) jc.setCursor(lastCursor=getCursor());
        if (id==MOUSE_CLICKED && tc!=null) tc.getCaret().setVisible(true);
        if (cmd=="H") showHelp(HyperTextPane.class);
        if (cmd==ChPubmed.ACTION_FULL_TEXT_ASSOCIATED) {
            rmAllHighlights();
            underlineRefs(UNDERLINE_REFS|HIGHLIGHT_REPAINT);
        }

        if (cmd=="W_WEB" || cmd=="W_HL" || cmd=="W_RMHL" || cmd=="W_CPY"  || cmd=="EXEC" ||  kcode==VK_R && shortCut) {
            final Range r=new Range();
            final String wrd=getSelection(r)? byteArray().newString(r.getFrom(), r.getTo()) : this.textForContextMenu;
            if (wrd!=null) {
                if (cmd=="EXEC" || kcode==VK_R && shortCut) {
                    ChFrame.frame("Execute shell script",ExecByRegex.getInstance(0).setText(wrd).getPanel(),ChFrame.PACK)
                        .shw(ChFrame.TO_FRONT|ChFrame.AT_CLICK).i(IC_BATCH);
                }
                if (cmd=="W_CPY") toClipbd(wrd);
                if (cmd=="W_WEB") showFrameWeblinks(Customize.webSearches,joinLines(webLinks(wrd, getCustomSettings(Customize.webSearches))),"B");
                if (cmd=="W_RMHL" || cmd=="W_HL") {
                    for(Highlight h:toArry(HIGHLIGHTS[1], Highlight.class)) {
                        if (wrd.equals(h.needle)) { removeHighlight(h); jc.repaint();}
                    }
                    if (cmd=="W_HL")  highlightOccurrence(wrd,(boolean[])null,(boolean[])null,HIGHLIGHT_REPAINT|HIGHLIGHT_UPDATE_IF_TEXT_CHANGES, C(0xFF00ff));
                }
            }
        }

        if (cmd==CMD_CLEAR && tc!=null) tc.setText("");
        if (cmd=="SAVE_AND_EDIT") {
            if (ascii!=null) {
                final File f=ascii.getFile();
                if (f!=null) edFile(-1, f, modifiers(ev));
            } else {
                File f=file(gcp("F",q));
                if (f==null) f=createTempFile("");
                wrte(f,byteArray());
                edFile(-1, f, modifiers(ev));
            }
        }
        return ev;
    }

    /* <<< AWTEvent <<< */
    /* ---------------------------------------- */
    /* >>> Run >>> */
    private ChDelay run[];
    private ChDelay run(char type) {
        if (run==null) run=new ChDelay['Z'];
        if (run[type]==null) run[type]=ChDelay.newThread(this,("RUN_"+type).intern(),null, new long[ARRAY_DURATION]);
        return run[type];
    }
    private long timeEnter;
    public Object run(String id, Object arg) {
        final Object argv[]=arg instanceof Object[] ? (Object[])arg : null;
        final JComponent jC=jComponent();
        if (id=="ADD_OPT2H") {
            final Highlight h=(Highlight)arg;
            removeHighlight(h);
            h.options|=HIGHLIGHT_REMOVE_IF_TEXT_CHANGES;
            addHighlight(h);
        }
        if (id=="underlineRefs") underlineRefs(atol(arg));
        if (id=="getTextAsSegment") ((Segment[])arg)[0]=getTextAsSegment();
        if (id=="getText") ((String[])arg)[0]=((JTextComponent)jC).getText();
        if (id=="setTextTS") setTextTS(toStrg(argv[0]),argv[1]!=null);
        if (id=="RUN_U") underlineRefs(UNDERLINE_REFS);
        if (id=="RUN_R") {
            final ChTextReplacement rplcMnt[]=this.replacements;
            final ByteArray ba=byteArray();
            if (sze(rplcMnt)>0 && sze(ba)>0) {
                final int mc=ba.getModificationCount();
                for(ChTextReplacement r : rplcMnt) ba.replace(r);
                if (mc!=ba.getModificationCount())  setTextTS(ba,true);
            }
        }
        if (id=="RUN_H") {
            if (sze(HIGHLIGHTS[0])>0) HIGHLIGHTS[0].clear();
            underlineRefs(UNDERLINE_REFS);
            for(List<Highlight> vH : HIGHLIGHTS) {
                for(int iH=sze(vH);--iH>=0;) {
                    final Highlight h=vH.get(iH);
                    if (h==null) continue;
                    if (h!=null && h.needle!=null) {
                        if ((h.options&HIGHLIGHT_UPDATE_IF_TEXT_CHANGES)!=0) {
                            hlo(h.needle,h.delimiterL,h.delimiterR, h.options, color(h),h);
                        }
                    }
                }
            }
            if (ulWebColors) highlightRGB();
            if (RECT_REPAINT.height>0 && jC!=null)  jC.repaint(RECT_REPAINT);
            RECT_REPAINT.setBounds(0,0,0,0);
        }
        if (id=="RUN_M") he(new ActionEvent(jC,ActionEvent.ACTION_PERFORMED, CMD_MIDDLE_MOUSE_BUTTON,0));
        if (id=="RUN_I") he(new ActionEvent(jC,ActionEvent.ACTION_PERFORMED, ACTION_INSERTED_TEXT_LONGER_ONE,0));
        if (id=="RUN_C") he(actEvt==null ? actEvt=new java.awt.event.ActionEvent(jC,ActionEvent.ACTION_PERFORMED, ACTION_TEXT_CHANGED,0) : actEvt);
        if (id=="REDUCE_SIZE") reduceHeight((JFrame)argv[0],(JTextComponent)argv[1]);
        return null;
    }
    /* <<< Run <<< */
    /* ---------------------------------------- */
    /* >>> Buttons >>> */
    public ChButton newClearButton() { return new ChButton(CMD_CLEAR).t("Clear").li(LI);}

    public ChButton butSave(File f) {
        return new ChButton("SAVE_AND_EDIT").t("save").li(LI).cp("F",f);
    }
    /* <<< Buttons <<< */
    /* ---------------------------------------- */
    /* >>> References >>> */
    final static int L_imageRefs=/*18+19*/ imageRefsS.length, STYLE_PDF=L_imageRefs, STYLE_ABSTRACT=L_imageRefs+1, STYLE_FILE=L_imageRefs+2, STYLE_REF=L_imageRefs+3;
    private final static int[] imageRefsStrLen=new int[L_imageRefs-9];
    static {
        assertTrue(L_imageRefs==imageRefsS.length && L_imageRefs==imageRefsI.length+19 && imageRefsS[imageRefsI.length]=="/");
        for(int i=imageRefsStrLen.length; --i>=0;) imageRefsStrLen[i]=imageRefsS[i].length();
    }
    private static int[] sharedFromTo=new int[100], sharedStyle=new int[50];
    private static byte[] lookupTable[][];
    private static int lookupTableMC;
    private static long ul_options;
    private static boolean[] refsDelimL, refsDelimR;
    public boolean underlineRefs(long mode) {
        ul_options=mode;
        if (!isEDT()) {
            ChDelay.invokeAndWait(this,"underlineRefs",longObjct(mode));
            return true;
        }
        if (refsDelimL==null) refsDelimL=charClass("\00"+" \t\n\r>,(){}[]\"");
        if (refsDelimR==null) refsDelimR=charClass("\00"+" \t\n\r>,(){}[]\"");
        alreadUlRefs=true;
        if ( (mode&UNDERLINE_REFS_PMID_DOWNLOAD)!=0) pmidDownload=true;
        if ( (mode&UNDERLINE_REFS_NOT_CLICKABLE)!=0) linksDisabled=true;
        if ( (mode&UNDERLINE_REFS_NO_ICON)!=0) noIcons=true;
        if ( (mode&UNDERLINE_WEB_COLORS)!=0) { ulWebColors=true; highlightRGB(); }
        if ( (mode&UNDERLINE_REFS_DISABLE_UNTIL_MOUSE)!=0) { hlAuto=false; return false; }
        if ( (mode&UNDERLINE_REFS_BG)!=0) { alreadUlRefs=true; ChDelay.newThread(this,"RUN_U",null).start(); return true;}
        else if ( mode==0 || (mode&(UNDERLINE_REFS|UNDERLINE_REFS_NOT_CLICKABLE))!=0) {
            final JComponent jc=jComponent();
            final int transparentDependingOnJC=jc instanceof HyperTextPane ? 44 : 255;

                final int found[]={-1,-1};
                final ByteArray ba=byteArray();
                final byte T[]=ba!=null ? ba.getBytes() : null;
                final int E=T!=null ? mini(T.length,ba.getEnd()) : 0;
                if (E==0) return false;
                final int mc=Hyperrefs.getModificationCount();
                byte[][] lookup[]=this.lookupTable;
                if (lookup==null || this.lookupTableMC!=mc) {
                    lookup=this.lookupTable=OccurInText.lookupTableFirstLetter(new String[][]{imageRefsS, Hyperrefs.getKeysToHighlight()});
                    this.lookupTableMC=mc;
                }
                int fromTo[]=this.sharedFromTo, style[]=sharedStyle, count=0, pos=ba.getBegin()-1;
                while( pos<E-1 && (pos=OccurInText.idxOfNeedles(lookup, refsDelimL, T, pos, E, found, 0L)) >=0 ) {
                    final byte hit[]=lookup[found[0]][found[1]];
                    if (hit==null || hit.length==0) { assertTrue(false); continue; }
                    final byte h0=hit[0], h1=hit.length>1 ? hit[1] : 0;
                    final boolean isEC;
                    if ('1'<=h0 && h0<='9') {
                        if (pos>E-6) continue;
                        final byte p1=T[pos+1], p2=T[pos+2];
                        isEC=('0'<=p1 && p1<='9' && p2=='.' || p1=='.' && '0'<=p2 && p2<='9' ) &&   looks(LIKE_EC,T,pos,E);
                        if (!isEC) continue;
                    } else isEC=false;
                    int stle=0;
                    final boolean
                        isMovie=h0=='M' && equls("MOVIE:",T,pos),
                        looksLikeURL=(h0=='f' || h0=='h') && looks(LIKE_EXTURL,T,pos),
                        looksLikeFile=(h1==':' || h0=='/' || h0=='\\' || h0=='~' || h0=='f'&&h1=='i') && looks(LIKE_FILEPATH_EVALVAR,T,pos),
                        isHyperref=Hyperrefs.isHyperref(T,pos,E);
                    if (isHyperref || looksLikeFile || looksLikeURL || isEC || isMovie) {
                        int toIdx=pos+hit.length;
                        while(toIdx<E && !refsDelimR[127&T[toIdx]]) toIdx++;
                        if (looksLikeFile) {
                            final int toFile=fileEndsAt(T,pos,toIdx);
                            if (toFile>0)  stle=STYLE_FILE;
                            else continue;
                        }
                        if (hit[hit.length-1]=='{') {
                            toIdx--;
                            while(  toIdx+1<E && T[toIdx+1]!='}' && T[toIdx+1]!='\n') toIdx++;
                        }
                        stle=idxOfStrg(hit, imageRefsS,0,imageRefsStrLen.length, imageRefsStrLen);
                        if (stle>=0 && !isEC && !looksLikeFile) {
                            final int
                                numStart,
                                pmid=
                                E>pos+6 && hit.length==4 && h0=='P'&&h1=='M'&&hit[2]=='I'&&hit[3]=='D' &&
                                is(DIGT,T,numStart=pos+ (T[pos+4]==':'?5:4)) ? atoi(T,numStart) : 0;
                            if (pmid>0) {
                                final String sPmid=toStrg(pmid);
                                stle=
                                    ChPubmed.hasPdf(sPmid) ? STYLE_PDF :
                                    sze(ChPubmed.getFile(sPmid, ChPubmed.ABSTRACT))>0 ? STYLE_ABSTRACT : stle;
                            }
                            toIdx=pos+hit.length;
                        } else stle=STYLE_REF;
                        if (2*count>=fromTo.length) {
                            this.sharedStyle=style=chSze(style, count+33);
                            this.sharedFromTo=fromTo=chSze(fromTo, 2*style.length);
                        }
                        style[count]=stle;
                        fromTo[2*count]=pos;
                        fromTo[2*count+1]=toIdx;
                        count++;
                    }
                }
                if (count>0) {
                    for(int iStyle= STYLE_REF+1; --iStyle>=0;) {
                        int cnt=0;
                        for(int i=count; --i>=0;) if (style[i]==iStyle) cnt++;
                        int[] fromTo1=new int[2*cnt];
                        cnt=0;
                        for(int i=0; i<count; i++) {
                            if (style[i]==iStyle) {
                                fromTo1[2*cnt]=fromTo[2*i];
                                fromTo1[2*cnt+1]=fromTo[2*i+1];
                                cnt++;
                            }
                        }
                        if (cnt>0) {
                            final Object color=
                                iStyle==STYLE_FILE || iStyle==STYLE_REF ? null  :
                                iStyle==STYLE_PDF ? C(0xff0000,transparentDependingOnJC) :
                                iStyle==STYLE_ABSTRACT ? C(0x8888ff,transparentDependingOnJC) :
                                (String)get(iStyle,imageRefsI),
                                color1=color!=null ? color : COLOR_HYPERREFS[linksDisabled?1:0];
                            addHighlight(newHighlight(fromTo1, color1, HIGHLIGHT_NOT_IN_SCROLLBAR|HIGHLIGHT_REMOVE_IF_TEXT_CHANGES));
                        }
                    }
                }

            if ( (mode&HIGHLIGHT_REPAINT)!=0) {  ChDelay.repaint(jc,99); ChDelay.repaint(vSB,99); }
            return true;
        }
        return false;
    }
    /* <<< References <<< */
    /* ---------------------------------------- */
    /* >>> Scroll >>> */
    private ChScrollBar vSB;
    public ChScrollBar verticalScrollBar() {
        if (vSB==null) addPaintHook(this, vSB=new ChScrollBar(JScrollBar.VERTICAL));
        return vSB;
    }
    private JScrollPane scrollPane;
    public JScrollPane scrollPane() {
        if (scrollPane==null) {
            scrollPane=new JScrollPane(jComponent());
            final JScrollBar vSB=verticalScrollBar();
            scrollPane.setVerticalScrollBar(vSB);

        }
        return scrollPane;
    }

    public static Thread thread_setScrollPositionV(int pos, JComponent tc) {
        final Object[] para={intObjct(pos),tc};
        return threadM("setScrollPositionV",ChTextComponents.class, para);
    }
    public static void setScrollPositionV(int pos, JComponent tc) {
        final JScrollBar sb=getScrollBar(tc,'V');
        if (sb!=null) {
            try {
                if (tc instanceof JTextComponent) ((JTextComponent)tc).setCaretPosition(pos);
            }catch(Exception ex){}
            final Rectangle r=new Rectangle();
            modelToView(pos, tc, r);
            tc.scrollRectToVisible(r);
        }
    }
    /* <<< Scroll <<< */
    /* ---------------------------------------- */
    /* >>> Paint >>> */
    private boolean wasEmpty;
    private static Point pointTmp=new Point();
    public void repaintLater(int delay) {
        final JComponent jC=jComponent(), vSB=this.vSB;
        if (delay==0) {
            if (jC!=null) jC.repaint();
            if (vSB!=null) vSB.repaint();
        } else {
            ChDelay.repaint(jC,delay);
            ChDelay.repaint(vSB,delay);
        }
    }
    public boolean isAlreadyPainted() { return alreadyPainted; }
    public boolean paintHook(JComponent component, Graphics g, boolean after) {
        antiAliasing(g);
        final JComponent jc=jComponent();
        if (jc==null) return false;
        final Rectangle clip=g.getClipBounds();
        if (!after && component==vSB) {
        paintScrollbar:
            for(List<Highlight> vH : HIGHLIGHTS) {
                for(int iH=sze(vH);--iH>=0;) {
                    final Highlight h=vH.get(iH);
                    if (h!=null && sze(yy(jc,h))>0 && (h.options&HIGHLIGHT_NOT_IN_SCROLLBAR)==0) {
                        vSB.setHasSelection(true);
                        break paintScrollbar;
                    }
                }
            }
        }
        if (after && component==vSB) {
            final int rowH=getFixedRowHeight(jc);
            Rectangle track=null;
            final int jcHeight=jc.getHeight();
            for(List<Highlight> vH : HIGHLIGHTS) {
                for(int iH=sze(vH);--iH>=0;) {
                    final Highlight h=vH.get(iH);
                    if (h==null || (h.options&HIGHLIGHT_NOT_IN_SCROLLBAR)!=0) continue;
                    g.setColor(color(h));
                    final int[] yy=yy(jc,h);
                    for(int i=yy.length; --i>=0;) {
                        if (track==null) track=vSB.getTrack();
                        final int y=track.y+yy[i]*track.height/jcHeight;
                        fillBigRect(g, 1,y,track.width-3,2);
                    }
                }
            }
            //vSB.setBlockIncrement(maxi(1,vSB.getVisibleAmount()-rowH));
        }

        if (component==jc) {
            final int pos1,pos2;
            final int w=jc.getWidth(), h=jc.getHeight();
            if (clip!=null) {
                pointTmp.move(clip.x,clip.y);
                pos1=viewToModel(pointTmp,jc);
                pointTmp.move(clip.x+clip.width, clip.y+clip.height);
                pos2=viewToModel(pointTmp,jc);
            } else {pos1=0; pos2=MAX_INT;}
            if (!after) {
                loadSavedText();
                if (!paintHooks(jc, g, false)) return false;
                if (jc.isOpaque()) {
                    g.setColor(jc.getBackground());
                    fillBigRect(g,0,0,w,h);
                }
                paintHighlights(pos1-1,pos2,false,g, clip);
            } else {
                final String txtEmpty=toStrg(gcpa(KEY_IF_EMPTY,jc));
                final boolean isE=txtEmpty!=null && nxt(NO_SPC, byteArray())<0 && !jc.hasFocus();
                if (wasEmpty !=isE) jc.repaint();
                wasEmpty=isE;
                if (isE) {
                    g.setClip(0,0,w, h);
                    {
                        g.setColor(jc.getBackground());
                        fillBigRect(g, 0,0,w,h);
                        g.setColor(C(0x888888));
                        int y= charA(g)+ jc.getInsets().top;
                        for(String s:splitStrg(txtEmpty,'\n')) { g.drawString(s,EM+jc.getInsets().left,y); y+=charH(g)+1;}
                    }
                } else {

                    paintHighlights(pos1-1,pos2,true,g,clip);

                    paintHooks(jc, g, true);
                    if (!alreadUlRefs && hlAuto) {
                        alreadUlRefs=true;
                        RECT_REPAINT.setBounds(0,0,w, h);
                        ChDelay.runAfterMS(EDT,run('H'),333);
                    }
                }

            }
            alreadyPainted=true;
        }
        return true;
    }
    private void shiftHighlights(int from, int len) {
        for(List<Highlight> vH : HIGHLIGHTS) {
            for(int iH=sze(vH);--iH>=0;) {
                final Highlight h=vH.get(iH);
                final int[] fromTo=h!=null ? h.fromTo : null;
                if (fromTo==null || fromTo.length==0) continue;
                final int binSearch=Arrays.binarySearch(fromTo,from); // (-(insertion point) - 1)
                final int insPoint=binSearch<0 ? -(binSearch+1) : binSearch;
                for(int i=insPoint; i<fromTo.length; i++)  fromTo[i]+=len;
            }
        }
    }
    private void paintHighlights(int from, int to, boolean after, Graphics g, Rectangle clip) {
        final JComponent jc=jComponent();
        final JTextComponent tc=jc instanceof JTextComponent ? (JTextComponent)jc : null;
        final int clipX0=clip!=null ? clip.x : 0;
        final int clipY0=clip!=null ? clip.y : 0;
        final int clipX1=clip!=null ? clipX0+clip.width : MAX_INT;
        final int charW=charW(jComponent());
        final boolean isEditableFocus=tc!=null && tc.isEditable() && jc.hasFocus();
        final boolean isHtmlPane=tc!=null && tc.getDocument() instanceof javax.swing.text.html.HTMLDocument;
        int countPainted=0, countAll=0;
        for(List<Highlight> vH : HIGHLIGHTS) {
            for(int iH=sze(vH);--iH>=0;) {
                final Highlight h=vH.get(iH);
                final int[] fromTo=h!=null ? h.fromTo : null;
                if (fromTo==null || fromTo.length==0) continue;
                countAll++;
                final Object c_or_i=h.colorOrIcon;
                final ImageIcon ic=c_or_i instanceof ImageIcon && !noIcons && !isEditableFocus ? (ImageIcon)c_or_i : null;
                final Image im=ic!=null ? ic.getImage() : null;
                final Color col=c_or_i instanceof Color ? (Color)c_or_i : im==null ?  COLOR_HYPERREFS[linksDisabled?1:0] : null;
                final boolean line=col==COLOR_HYPERREFS[0] || col==COLOR_HYPERREFS[1];
                if (after!=(im!=null || line || isHtmlPane)) continue;
                final int binSearch=Arrays.binarySearch(fromTo,from); // (-(insertion point) - 1)
                final int insPoint=binSearch<0 ? -(binSearch+1) : binSearch;
                if (insPoint>=fromTo.length) continue;

                for(int i=insPoint&~1; i<fromTo.length; i+=2) {
                    final int iFrom=fromTo[i], iTo=fromTo[i+1];
                    if (iFrom>to) break;
                    modelToView(iFrom,jc,RECTf);
                    modelToView(iTo-1,jc,RECTt);
                    final Rectangle r0=RECTf, r1=RECTt;
                    final int r1Width=r1.width, w1=r1Width<2 ? charW : r1Width;
                    final int y0= r0.y,  x0=r0.x,  x1=r1.x, y1=r1.y, w0=r0.width, h0=r0.height, h1=r1.height;

                    if (x0+w0<clipX0 && x1+w1<clipX0 || x0>clipX1 && x1>clipX1) continue;
                    countPainted++;

                    if (y1>=0 && y0>=0 && col!=null) {
                        g.setColor(col);
                        if (y0==y1) {

                            if (line) g.drawLine(x0,y1+h1,  x1+w1,y1+h1 );
                            else {
                                fillBigRect(g, x0, maxi(clipY0, y0),  x1+w1-x0,y1+h1-y0 );
                            }
                        } else {
                            if (line) {
                                g.drawLine(x0, r0.y+h0,  99999, r0.y+h0);
                                g.drawLine(0,y1+h1, x1+w1,y1+h1 );
                            } else {
                                fillBigRect(g, x0,y0+1,jc.getWidth()-x0,h0-2);
                                if (y1>y0+h0) fillBigRect(g, 0,y0+h0,jc.getWidth(),y1-y0-h0); g.setColor(col);
                                fillBigRect(g, 0,y1+1,x1+w1,h1-2);
                            }
                        }
                    } else if (im!=null && x0<x1) {

                        final int x11=x1;
                        g.setColor(jc.getBackground());
                        fillBigRect(g, x0,y0,x11-x0,y0+h0-y0);
                        //g.clearRect(x0,y0,x11-x0,y0+h0-y0);
                        final int icw=ic.getIconWidth(), ich=ic.getIconHeight();
                        if (icw*h0<ich*(x11-x0)) {
                            final int d=icw*h0/ich;
                            final int destX=maxi(x0,x0 + (x11-x0- d)*2/3);
                            g.drawImage(im, destX, y0 , destX+d, y0+h0, 0, 0, icw, ich,jc);
                        } else {
                            final int d=ich*(x11-x0)/icw;
                            final int destY=maxi(y0,y0 + (y0+h0-y0- d)/2);
                            g.drawImage(im, x0, destY , x11, destY+d, 0, 0, icw, ich,jc);
                        }
                    }
                }
            }
        }
        //if (jc instanceof AsciiView)
        // debug(" PaintH="+countPainted+"/"+countAll+" "+jc.isOpaque());
    }
    /* <<< Paint <<< */
    /* ---------------------------------------- */
    /* >>> Dimension >>> */
    private float scale=1;
    public static int getFixedRowHeight(Object o) {
        return
            o instanceof ChTextArea && !((JTextArea)o).getLineWrap() ? ((ChTextArea)o).getFixedRowHeight() :
            o instanceof AsciiView ? ((AsciiView)o).getFixedRowHeight() :
            -1;
    }
    public static void reduceHeight(JFrame f, JTextComponent tc) {
        if (f==null || tc==null) return;
        if (!isEDT()) { ChDelay.invokeLater(instance(),"REDUCE_SIZE",new Object[]{f,tc}); return;}
        modelToView(tc.getDocument().getLength(), tc, RECTf);
        final Rectangle screenBounds=f.getGraphicsConfiguration().getBounds();
        if (RECTf.height>0) {
            final int h=RECTf.y+RECTf.height+3*ICON_HEIGHT;
            if (h<screenBounds.height-ICON_HEIGHT) {
                final Dimension d=f.getSize();
                d.height=h;
                final JScrollBar sb=getScrollBar(tc,'H');
                if(sb!=null && sb.isVisible()) d.height+=ICON_HEIGHT;
                f.setSize(d);
            }
        }
    }
    /* <<< Run <<< */
    /* ---------------------------------------- */
    /* >>> Highlight Class >>> */
    public List<Highlight> HIGHLIGHTS[]=new ArrayList[3];
    public static class Highlight {
        private boolean[] delimiterL, delimiterR;
        private Object needle,colorOrIcon;
        private long options;
        private int fromTo[], yy[];//, rgb=-1;
        int[] fromTo() { return fromTo;}
    }
    private static int[] yy(JComponent jc, Highlight h){
        int[] yy=h.yy;
        if (yy==null) {
            final ChTextComponents tools=tools(jc);
            if (tools==null) return NO_INT;
            final ByteArray ba=tools.byteArray();
            final int rowH=getFixedRowHeight(jc);
            final int fromTo[]=h.fromTo;
            yy=h.yy=new int[fromTo.length/2];
            for(int i=yy.length; --i>=0;) {
                final int f=fromTo[2*i];
                if (rowH>=0) yy[i]=rowH*ba.idxToRow(f);
                else {
                    modelToView(f,jc,RECTf);
                    yy[i]=RECTf.y;
                }
            }
        }
        return yy;
    }
    public static Highlight newHighlight(int[] fromTo, Object c_or_i, long options) {
        final Highlight h=new Highlight();
        h.colorOrIcon=
            c_or_i instanceof String  ? iicon((String)c_or_i) :
            c_or_i instanceof Color || (c_or_i instanceof ImageIcon) ? (Color)c_or_i :
            null;
        h.fromTo=fromTo;
        h.options=options;
        return h;
    }
    private static Color color(Highlight h) {
        return h.colorOrIcon instanceof Color ? (Color)h.colorOrIcon : COLOR_HYPERREFS[0];
    }
    /* <<< Highlight Class <<< */
    /* ---------------------------------------- */
    /* >>> Highlight >>> */
    public void addHighlight(Highlight h) {
        final int i=(h.options&HIGHLIGHT_REMOVE_IF_TEXT_CHANGES)!=0 ? 0 : (h.options&HIGHLIGHT_NEVER_REMOVE)!=0? 2 : 1;
        adUniq(h, (HIGHLIGHTS[i]==null ?  HIGHLIGHTS[i]=new ArrayList() : HIGHLIGHTS[i]));
    }
    public void rmAllHighlights() { for(int i=2; --i>=0;) if (sze(HIGHLIGHTS[i])>0) HIGHLIGHTS[i].clear();}
    public boolean removeHighlight(Highlight h){
        boolean success=false;
        for(int i=0; i<2; i++)  success|=sze(HIGHLIGHTS[i])>0 && HIGHLIGHTS[i].remove(h);
        return success;
    }
    private void highlightRGB() {
        final List<Highlight> v=HIGHLIGHTS[0]==null ? HIGHLIGHTS[0]=new ArrayList() : HIGHLIGHTS[0];
        final ByteArray ba=byteArray();
        final byte[] T=ba.getBytes();
        final int end=ba.getEnd();
        nextCharacter:
        for(int i=ba.getBegin(); i<end; i++) {
            final byte c=T[i];
            if (c!='#' || !is(HEX_DIGT, T,i+1) || !is(HEX_DIGT, T,i+2)) continue;
            for(int iTransp=2; --iTransp>=0;) {
                if (!is(LETTR_DIGT_US,T,i+7+iTransp*2)) {
                    final long rgb=HtmlUtil.hexToInt(T,i+1+iTransp*2);
                    if (rgb<0) continue;
                    v.add(ChTextComponents.newHighlight(new int[]{i,i+1}, C((int)rgb),HIGHLIGHT_NOT_IN_SCROLLBAR|HIGHLIGHT_REMOVE_IF_TEXT_CHANGES));
                }
            }
        }
    }
    /* <<< Highlight <<< */
    /* ---------------------------------------- */
    /* >>> highlightOccurrence >>> */
    public Highlight highlightOccurrence(Object needle, boolean[] delimiterL,boolean[] delimiterR, long mode, Color color) {
        return hlo(needle,delimiterL,delimiterR,mode,color,null);
    }
    private synchronized Highlight hlo(Object oNeedle, boolean[] delimL,boolean[] delimR, long mode, Color color,Highlight highlight) {
        final Object provided= !(oNeedle instanceof ByteArray) && oNeedle instanceof ChRunnable ?  ((ChRunnable)oNeedle).run(PROVIDE_HIGHLIGHTS, jComponent()) : null;
        final int[] buffer=highlight!=null ? highlight.fromTo :null;
        final int[] fromTo;
        final byte[] lookup[][];
        if (oNeedle instanceof Object[] || provided!=null) {
            lookup=
                provided!=null ? (byte[][][]) provided :
                oNeedle instanceof byte[][][] ? (byte[][][]) oNeedle:
                OccurInText.lookupTableFirstLetter((Object[])oNeedle);
            fromTo=OccurInText.findLU(mode,lookup,byteArray(),delimL,delimR,(IsEnabled)null,buffer);
        } else {
            final byte[] needle= oNeedle instanceof byte[] ? (byte[])oNeedle : toByts(toStrg(oNeedle));
            fromTo=OccurInText.find(mode, needle,byteArray(),delimL, delimR, (IsEnabled)null,  buffer);
            lookup=null;
        }
        if ((mode&HIGHLIGHT_UPDATE_ON_MOUSE_ENTER)!=0) { hlMouseEnter=true; return null; }
        final boolean update=(mode&HIGHLIGHT_UPDATE_IF_TEXT_CHANGES)!=0;
        hlAuto|=update;
        final Highlight h=highlight==null && (fromTo.length>0|| update) ? newHighlight(fromTo,color, mode) :highlight;
        if (h!=null) {
            h.needle= provided==null&&lookup!=null ? lookup : oNeedle;
            h.delimiterL=delimL;
            h.delimiterR=delimR;
            h.fromTo=fromTo;
            if (h!=highlight) addHighlight(h);
            if ( (mode&HIGHLIGHT_REPAINT)!=0){
                ChDelay.repaint(jComponent(),99);
                ChDelay.repaint(vSB,999);
            }
        }
        return h;
    }
    public void highlightBlanks() {
        highlightOccurrence(" ",null,null,HIGHLIGHT_UPDATE_IF_TEXT_CHANGES, C(DEFAULT_BACKGROUND));
        highlightOccurrence("\t",null,null,HIGHLIGHT_UPDATE_IF_TEXT_CHANGES,C(0xFF00FF));
    }
    /* <<< highlightOccurrence <<< */
    /* ---------------------------------------- */
    /* >>> Model View >>> */
    public static void modelToView(int m,Object o, Rectangle r) {
        if (o instanceof JTextComponent) {
            final JTextComponent tc=(JTextComponent)o;
            try {
                r.setBounds(tc.modelToView(m));
            } catch(Exception e){  r.setBounds(m<0?0:tc.getWidth(),m<0?0:tc.getHeight(),0,0); }
        } else if (o instanceof AsciiView) ((AsciiView)o).modelToView(m,r);
    }
    /* <<< Model View <<< */
    /* ---------------------------------------- */
    /* >>> Tools >>> */
    public static ChTextComponents tools(Object o0) {
        Object o=deref(o0);
        return
            o instanceof ChTextViewer ? tools(((ChTextViewer)o).getTextComponent()) :
            o instanceof ChTextArea ? ((ChTextArea)o).tools() :
            o instanceof ChTextField ? ((ChTextField)o).tools() :
            o instanceof HyperTextPane ? ((HyperTextPane)o).tools() :
            o instanceof AsciiView ? ((AsciiView)o).tools() :
            null;
    }
    private boolean jOrtho;
    private static boolean jOrtho_prepared;
    /* <<< Tools <<< */
    /* ---------------------------------------- */
    /* >>> EC-Class >>> */
    public static int nxtEC(byte T[], int from, int end) {
        final int dgt=nxt(DIGT,T,from, end);
        if ( (dgt==0 || dgt>0 && is(SPC,T[dgt-1])) && looks(LIKE_EC,T,dgt,end)) return dgt;
        return -1;
    }
    /* <<< EC-Class <<< */
    /* ---------------------------------------- */
    /* >>> Save >>> */
    public final static String DIR_SAVED_TEXT="~/@/savedText/";
    private File saveInFile;
    private boolean saveInFile_loaded;
    public void saveInFile(String fileName) {
        if (fileName!=null) saveInFile(file(DIR_SAVED_TEXT+rplcChars(charClass(NO_FILEP), '_', fileName)+".txt"));
    }
    public void saveInFile(File file) {saveInFile=file;  }

    public static int viewToModel(Point p,Object o) {
        if (o instanceof JTextComponent && p!=null) {
            try { return ((JTextComponent)o).viewToModel(p); } catch(Exception e){}
        } else  if (o instanceof AsciiView) return ((AsciiView)o).viewToModel(p.x,p.y);
        return -1;
    }
    private static File myFile(String n) { return sze(n)==0 ? null : file(isWindows() && n.startsWith("~/.StrapAlign") ? rplcToStrg("~/.StrapAlign", dirSettings().toString(), n) : n); }
    public void loadSavedText() {
        final JComponent jc=jComponent();
        if (jc instanceof JTextComponent && saveInFile!=null && !saveInFile_loaded) {
            saveInFile_loaded=true;
            setTextTS(readBytes(saveInFile));
            needsSaving=false;
        }
    }
    public void saveInFile(boolean always) {
        if (needsSaving || always) wrte(saveInFile,byteArray());
        needsSaving=false;
    }
    /* <<< Save <<< */
    /* ---------------------------------------- */
    /* >>> Text >>> */
    private transient Object refByteArray, refSegment;
    private static int fileEndsAt(final byte[] T,final int pos, final int end) {
        if (pos+3>end || !looks(LIKE_FILEPATH_EVALVAR,T,pos)) return -1;
        final int to;
        final boolean cc[]=charClass(FILENM);
        {
            int i=pos+3;
            nextI:
            while(i<end) {
                final byte c=T[i];
                if (c<=0) break nextI;
                if (!(cc[c] || c=='%'||c=='/'||c=='$'||c=='\\')) break nextI;
                for(String V : FILE_RANGE_SUFX)  if (equls(V,T,i)) break nextI;
                i++;
            }
            to=i;
        }
        final File f=myFile(toStrg(T,pos,to));
        return (f!=null && f.exists()) ? to : -1;
    }
    private transient Object ref_getText;
    public String getText() {
        String s=(String)deref(ref_getText);
        if (s==null) {
            if (isEDT()) s=((JTextComponent)jComponent()).getText();
            else {
                final String ss[]=new String[1];
                inEDT(ChDelay.newThread(this, "getText", ss));
                s=ss[0];
            }
            ref_getText=newSoftRef(s);
        }
        return s;
    }
    //    public static final static long SETTEXT_KEEP_CARET=1<<1, SETTEXT_KEEP_SCROLL=1<<2;
    public void setTextTS(CharSequence txt) { setTextTS(txt,false);}
    public void setTextTS(CharSequence txt, boolean keepCaretPosition) {
        final String s=toStrg(txt);
        if (s!=null) {
            if (!isEDT()) inEDT(ChDelay.newThread(this,"setTextTS",new Object[]{s,keepCaretPosition?"":null}));
            else {
                final JTextComponent tc=(JTextComponent)jComponent();
                final JScrollBar sb=keepCaretPosition&&scrollPane!=null?scrollPane.getVerticalScrollBar() : null;
                final int scrollV=sb!=null?sb.getValue(): -1;
                final int caret=keepCaretPosition ? tc.getCaretPosition() : -1;
                ref_getText=newSoftRef(s);
                final boolean save=eventsEnabled;
                setTextChangeEventEnabled(false);
                alreadUlRefs=false;
                tc.setText(s);
                try {
                    if (caret>=0) tc.setCaretPosition(caret);
                } catch(Exception ex){}
                if (scrollV>=0)  sb.setValue(scrollV);

                setTextChangeEventEnabled(save);
            }
            needsSaving=true;
        }
    }
    public ByteArray byteArray() {
        loadSavedText();
        final JComponent jc=jComponent();
        final AsciiView ascii=jc instanceof AsciiView ? (AsciiView)jc : null;
        if (ascii!=null) return ascii.byteArray();
        ByteArray ba=(ByteArray)deref(refByteArray);
        if (ba==null) { refByteArray=newSoftRef(ba=new ByteArray(NO_BYTE));byteArrayValid=false;}
        if (!byteArrayValid) {
            ba.setTextSegment(getTextAsSegment());
            byteArrayValid=true;
        }

        return ba;
    }
    public Segment getTextAsSegment() {
        if (!isEDT()) {
            final Segment[] ss={null};
            inEDT(ChDelay.newThread(this,"getTextAsSegment",ss));
            return ss[0];
        }
        final JComponent tc=jComponent();
        if (tc instanceof JTextComponent) {
            Segment segm=(Segment)deref(refSegment);
            if (segm==null) { refSegment=newSoftRef(segm=new Segment()); segmentValid=false;}
            if (!segmentValid) {
                segmentValid=true;
                final Document doc=((JTextComponent)tc).getDocument();
                try {
                    segm.array=null;
                    doc.getText(0,doc.getLength(),segm);
                    segmentValid=true;
                } catch(Exception e){ putln("caught in ChTextComponents:  doc.getText(int,int,Segment)", e);}
            }
            return segm;
        }
        return null;
    }
    /* <<< Text<<< */
    /* ---------------------------------------- */
    /* >>> Text Change >>> */
    private ChTextReplacement[] replacements;
    public void setReplaceWhenTextChanges(ChTextReplacement...replacements) { this.replacements=replacements; }
    private int modificationCount;
    public int getModificationCount() { return modificationCount; }
    public void setTextChangeEventEnabled(boolean b) { eventsEnabled=b; }
    public void setChanged(){byteArrayValid=false; }
    public void insertUpdate(DocumentEvent ev){ changed(ev);}
    public void removeUpdate(DocumentEvent ev){ changed(ev);}
    public void changedUpdate(DocumentEvent ev){}

    private void changed(DocumentEvent ev) {
        modificationCount++;
        ref_getText=null;
        this.byteArrayValid=this.segmentValid=false;
        if (alreadyPainted) {
            final JComponent jc=jComponent();
            boolean doRunHL=sze(HIGHLIGHTS[0])>0 || hlAuto;
            for(int i=sze(HIGHLIGHTS[1]); --i>=0 && !doRunHL;) {
                final Highlight h=HIGHLIGHTS[1].get(i);
                if (h!=null && (h.options&HIGHLIGHT_UPDATE_IF_TEXT_CHANGES)!=0 ) doRunHL=true;
            }
            boolean immediately=false;
            this.needsSaving=true;
            if (doRunHL && hlAuto && ev!=null) {
                boolean runThread=true;
                final boolean
                    ins=ev.getType()==DocumentEvent.EventType.INSERT,
                    del=ev.getType()==DocumentEvent.EventType.REMOVE;
                if (!ins && !del || ev.getLength()>1) {
                    RECT_REPAINT.add(0,0);
                    RECT_REPAINT.add(jc.getWidth(),jc.getHeight());
                } else {
                    shiftHighlights(ev.getOffset(), ev.getLength()*(ins?1:-1));
                    if (immediately) {
                        runThread=false;
                        run('H').run();
                    } else {
                        modelToView(ev.getOffset(),jc,RECTf);
                        final int charH=charH(jc), y=RECTf.y;
                        if (RECT_REPAINT.height==0) RECT_REPAINT.setBounds(0,y,0,0);
                        RECT_REPAINT.add(0,y-charH);
                        RECT_REPAINT.add(jc.getWidth(),y+2*charH);
                    }
                }
                if (runThread) ChDelay.runAfterMS(EDT,run('H'),333+10*run('H').LAST[DURATION]);
            }
            if (gcp(KEY_NOT_DISPATCHING_EV,jc)==null && eventsEnabled && alreadyTyped && ev!=null) {
                /* --- Event dispatching needs delay because the caret position is adjusted later --- */
                if (ev.getType()==DocumentEvent.EventType.INSERT && ev.getLength()>2) ChDelay.runAfterMS(EDT,run('I'),333);
                final long delay=run('C').LAST[DURATION]*2+111;
                ChDelay.runAfterMS(EDT,run('C'), delay);
            }

            if (sze(this.replacements)>0) ChDelay.runAfterMS(EDT,run('R'),333);
        }
    }
    private final Rectangle RECT_REPAINT=new Rectangle();
    /* <<< Text  Change<<< */
    /* ---------------------------------------- */
    /* >>> Selection >>> */
    public void marchingAnts(int id, int from, int to) {
        final JComponent jc=jComponent();
        modelToView(from,jc,RECTf);
        modelToView(to,jc,RECTt);
        if (RECTf.height>0 && RECTt.height>0) {
            final Rectangle r=new Rectangle(RECTf.x, RECTf.y,  RECTt.x+RECTt.width-RECTf.x,  RECTt.y+RECTt.height-RECTf.y);
            jc.scrollRectToVisible(new Rectangle(r.x-8, r.y-8, r.width+16, r.height+16));
            setMarchingAnt(ANTS_SEARCH, jc,r, C(0xFF00ff), C(0xFF0000));
        }
    }
    public boolean getSelection(Range range) {
        final JComponent jc=jComponent();
        final int f,t;
        if (jc instanceof JTextComponent) { f=((JTextComponent)jc).getSelectionStart(); t=((JTextComponent)jc).getSelectionEnd(); }
        else if (jc instanceof AsciiView) {
            final int region[]=((AsciiView)jc).region().fromTo;
            f=region[0];
            t=region[1];
        } else f=t=-1;
        range.set(f,t);
        return f>=0 && f<t;
    }
    /* <<< Selection <<< */
    /* ---------------------------------------- */
    /* >>> static utility >>> */
    private static boolean contains(Point p, int x0, int y0, int x1, int y1) {
        return p.x>x0 && p.x<x1 && p.y>y0 && p.y<y1;
    }
    public static long INSERT_IF_NOT_CONTAINS=1<<0;
    public static void insertAtCaret(long options, String s, JComponent jc) {
        final JTextComponent tc=deref(jc,JTextComponent.class);
        if (s==null || tc==null) return;
        final String t=tc.getText();
        if ( (options&INSERT_IF_NOT_CONTAINS)!=0 && t.indexOf(s)>=0) return;
        final int cp=maxi(0,tc.getCaretPosition()), after=cp+s.length();
        try {
            final int caret=tc.getCaretPosition();
            tc.getDocument().insertString(cp, s, null);
        } catch(Exception e){ return; }
        tc.setCaretPosition(after);
        tc.requestFocus();
        final ChTextComponents tools=tools(jc);
        if (tools!=null) {
            final Highlight h=newHighlight(new int[]{cp,after}, C(0xFF00,99), HIGHLIGHT_NOT_IN_SCROLLBAR);
            tools.addHighlight(h);
            ChDelay.runAfterMS(EDT, ChDelay.newThread(tools,"ADD_OPT2H", h),444);
        }

    }

    private static String _of_path;
    private static JComponent _of_dialog;
    private static ChButton _of_buttons[];
    private final static int OF_BUT_F=0, OF_BUT_f=1, OF_BUT_U=2, OF_BUT_u=3;
    private static JTextField _of_lab;
    private static void dialogOpenFile(long options, String path) {
        if (nxt(NO_SPC,path)<0) return;
        if (_of_dialog==null) {
            final ChButton bb[]=_of_buttons=new ChButton[4];
            final String ss[]={"Open file","Parent directory in file browser","In web browser","Parent in web browser"};
            _of_dialog=vB(_of_lab=new JTextField());
            for(int i=0;i<ss.length;i++) {
                _of_dialog.add(hBL(bb[i]=new ChButton(ss[i]).li(instance().LI)));
                pcp(KOPT_HIDE_IF_DISABLED,"",bb[i]);
            }
            if (Insecure.EXEC_ALLOWED) _of_dialog.add(hBL("For details, hold ctrl while pressing the button."));
            setDragMovesWindow(true, _of_dialog);
            closeOnCtrlW(true, _of_dialog, CLOSE_PARENT_WINDOW);
        }
        _of_lab.setText(path.trim());
        final File f=path.indexOf(':')>=0?null:file(path);
        final File fParent=f==null?null:file(f.getParent());
        final Object u=url(path);
        final ChButton bb[]=_of_buttons;
        setEnabld(Insecure.EXEC_ALLOWED && f!=null && f.exists(), bb[OF_BUT_F]);
        setEnabld(Insecure.EXEC_ALLOWED && f!=null && fParent.exists(), bb[OF_BUT_f]);
        setEnabld(u!=null, bb[OF_BUT_U]);
        setEnabld(u!=null && path.indexOf('/',8)>0, bb[OF_BUT_u]);
        ChFrame.frame("Open file", _of_dialog, ChFrame.ALWAYS_ON_TOP).shw(ChFrame.AT_CLICK|ChFrame.PACK);
    }

}