package charite.christo;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import java.awt.*;
import charite.christo.interfaces.*;
import javax.swing.JTextField;
import javax.swing.text.Document;
import java.awt.event.*;
import java.io.File;
import java.util.Map;

/**
   @author Christoph Gille

*/
public final class ChTextField extends JTextField {
    public ChTextField() { this("",0L);}
    public ChTextField(Object defaultText) { this(defaultText,0L);}

    public ChTextField(Object defaultText, long options) {
        assert defaultText instanceof Document || defaultText instanceof JTextField || defaultText instanceof CharSequence || defaultText==null;
        final Document doc=defaultText instanceof Document ? (Document)defaultText :
            defaultText instanceof JTextField ? ((JTextField)defaultText).getDocument() :null;
        if (doc!=null) setDocument(doc);
        else {
            this.defaultText=toStrg(defaultText);
            if (defaultText!=null) t(this.defaultText);
        }
        monospc(this);
        if (options!=0) {
            final ChTextComponents tools=tools();
            final boolean  complFiles= (options&COMPLETION_FILES_IN_WORKING_DIR)!=0;
            if ( (options&COMPLETION_DIRECTORY_PATH)!=0 || complFiles) {
                tools.enableWordCompletion(complFiles ? dirWorking() : new ChFileFilter().setRules(".D"),0,0, ChTextComponents.COMPL_TAB);
                tools.enableUndo(true);
                tools.highlightOccurrence(ChTextComponents.WRONG_CHARS_IN_FILEN, null,null, HIGHLIGHT_UPDATE_IF_TEXT_CHANGES, C(0xFF0000));
                tools.underlineRefs(UNDERLINE_REFS_NOT_CLICKABLE);
            }
            tools.underlineRefs(options);
        }
    }
    public ChTextField cp(Object key, Object value) { pcp(key,value,this); return this;}
    @Override public void processEvent(AWTEvent ev) {
        if (!wasPainted) return;
        final AWTEvent newEv=tools().processEvent(this,ev);
        if (newEv!=null) try { super.processEvent(newEv);} catch(Throwable ex){ if (isMyComputer()) stckTrc(ex);}
    }
    { this.enableEvents(MouseEvent.MOUSE_EVENT_MASK|KeyEvent.KEY_EVENT_MASK|MouseEvent.MOUSE_WHEEL_EVENT_MASK|FocusEvent.FOCUS_EVENT_MASK);}
    private Object contentType;
    public Object getContentType() { return contentType;}
    public ChTextField ct(Object c) {  this.contentType=c;  return this; }
    private boolean wasPainted;
    @Override public String getText() { assert !isDisplayable() || isEDT(); return super.getText();}
    @Override public void setText(String txt) {
        if (isDisplayable()) assertEDT();
        super.setText(txt);
    }

    public ChTextField t(Object o) {
        tools().setTextChangeEventEnabled(false);
        if (o!=null) tools().setTextTS(o.toString());
        tools().setTextChangeEventEnabled(true);
        return this;
    }
    public ChTextField cols(int columns,boolean minimumSize, boolean maximize){
        setColumns(columns);
        if (maximize) setMaximumSize(getPreferredSize());
        if (minimumSize) setMinimumSize(getPreferredSize());
        return this;
    }

    @Override public Dimension getPreferredSize() {
        final Dimension ps=getPrefSize(this);
        return  ps!=null ? ps : super.getPreferredSize();
    }
    public ChTextField tt(CharSequence toolTip) {
        rtt(this);
        this.tip=toStrg(toolTip);
        return this;
    }

    @Override public String getToolTipText(MouseEvent mev) {
        final String tt=tools().getToolTipText(mev);
        final String tt1= sze(tt)>0 ? tt :  super.getToolTipText(mev);
        final String tt2;

        if (this.tip!=null) {
            final StringBuffer sb=strBuf(SB_TIP).append(this.tip);
            if ( defaultText!=null) sb.append("<br>").append(defaultText);
            addHtmlTags(sb);
            tt2=sb.toString();
        } else tt2=tt1;
        return tt2;
    }

    private boolean isPainting;
     @Override public boolean isOpaque() { return isPainting ? false : super.isOpaque(); }

    @Override public void paintComponent(Graphics g) {
       wasPainted=true;
       final int w=getWidth(), h=getHeight();
        final Color bg=getBackground();
        if (bg!=null && isOpaque()) {
            g.setColor(bg);
            fillBigRect(g,0,0,w,h);
        }
        if (tools().paintHook(this,g,false)) {
            isPainting=true;
            super.paintComponent(g);
            isPainting=false;
        }

        if (contentType!=null) {
            final String txt=toString();
            Color colorValid=C(0x00b200);
            String textValid=null;
            if (contentType==File.class && ""!=txt)   {
                final File d=file(txt);
                if (d.length()>0 && !d.isDirectory()) { textValid="exists";}
                else if (d.isDirectory()) { textValid="directory"; colorValid=C(0xFF00FF);}
                else { textValid="no such file"; colorValid=C(0xFF0000);}
            }
            {
                Object o=gcpa(KEY_IF_EMPTY,this);
                if (""==txt && o!=null) { textValid=o.toString(); colorValid=C(0x808080);}
            }

            if (textValid!=null) {
                g.setColor(colorValid);
                g.drawString(textValid,w- strgWidth(g,textValid),charA(g));
            }
        }
        tools().paintHook(this,g,true);

    }

    private String defaultText,tip;
    public ChTextField ro(){ setEditable(false);return this;}
    private final static Map<String,Document> map_property_doc=new java.util.HashMap();
    /*
    public static ChTextField newTextField(final String ct) {
        final String tt=
            ct==KEY_EMAIL ? "email address" :
            null;
        final ChTextField tf=new ChTextField("",UNDERLINE_REFS_NOT_CLICKABLE).tt(tt);

        tf.contentType=ct;
        tf.t(getPrpty(ChUtils.class,ct, ""));
        final Document doc=map_property_doc.get(ct);
        if (doc!=null) {
            tf.setDocument(doc);
            doc.addDocumentListener(tf.tools());
        } else map_property_doc.put(ct, tf.getDocument());
        return tf.cols(50,true,false);
    }
    */
    public ChTextField li(ActionListener l) { addActLi(l,this); return this;}
    @Override public String toString() {
        tools().loadSavedText();
        String s=null;
        if (isEDT()) s=getText();
        else s=tools().getText();
        if (s==null) return "";
        s=s.trim();
        return "".equals(s) ? "" : s;
    }
    private ChTextComponents tools;
    public synchronized ChTextComponents tools() {
        if (tools==null) tools=new ChTextComponents(this); return tools;
    }
    public ChTextField saveInFile(String fileName) {tools().saveInFile(fileName); return this;}

  public java.util.List getDndDateien() { return getDndFilesV(this); }
}
