package charite.christo;
import java.util.List;
import java.util.*;
import javax.swing.*;
import charite.christo.interfaces.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**
   Periodically calls Runnable objects.
   Runs in one One universal thread
*/
public final class ChThread implements Runnable {
    public final static int WEAK_REFERENCE=2;     static { assert EDT!=WEAK_REFERENCE;}

    /* --- Begin Job Instance ---- */
    private ChThread(Object runnable, String name) { this.runnable=runnable; this.name=name; }
    private final Object runnable;
    private final String name;
    private int mode;
    private long lastCall, everyMs, cumulTime;
    /* --- End Job Instance ---- */

    private final static List<ChThread> vRunnable=new ArrayList();
    private static boolean isRunning;
    private static ChThread inst;

    public static ChThread callEvery(int everyMs, Runnable runnable,String name,int mode) {
        ChThread thread=null;
        for(int i=vRunnable.size(); --i>=0;) {
            final ChThread ri=get(i,vRunnable,ChThread.class);
            if (ri==null) continue;
            final Runnable r=deref(ri.runnable,Runnable.class);
            if (everyMs<0) {
                if (r==runnable || r==null) vRunnable.remove(ri);
            } else {
                if (r==runnable) {
                    thread=ri;
                    break;
                }
            }
        }
        if (everyMs>0) {
            if (thread==null) {
                thread=new ChThread(  (mode&WEAK_REFERENCE)!=0 ? weakRef(runnable): runnable, name);
                vRunnable.add(thread);
            }
            thread.mode=mode;
            thread.everyMs=everyMs;
            if (!isRunning) {
                isRunning=true;
                if (inst==null) inst=new ChThread(null,null);
                final Thread t=new Thread(inst);
                t.setDaemon(true);
                t.start();
            }
        }
        return thread;
    }
    public void setEveryMs(long ms) { if (ms<0) vRunnable.remove(this); everyMs=ms;}
    public void run() {
        final int SLEEP=40;
        while(true) {
            for(int i=vRunnable.size(); --i>=0;) {
                final ChThread ri=get(i,vRunnable,ChThread.class);
                if (ri==null) continue;
                final Runnable runnable=deref(ri.runnable,Runnable.class);
                if (runnable==null) { vRunnable.remove(ri); continue;}
                if (ri.lastCall<0) {
                    if ((ri.mode&EDT)!=0) inEDT(runnable);
                    else {
                        final long time=System.currentTimeMillis();
                        runnable.run();
                        ri.cumulTime+=System.currentTimeMillis()-time;
                    }
                    ri.lastCall+=ri.everyMs;
                }
                ri.lastCall-=SLEEP;
            }
            sleep(SLEEP);
        }
    }
    private static StringBuffer debugList(ThreadGroup g, int stufe, StringBuffer sb) {
        if (g!=null) {
            final Thread tt[]=new Thread[g.activeCount()];
            final ThreadGroup gg[]=new ThreadGroup[g.activeGroupCount()];
            final int nT=g.enumerate(tt,false);
            final int nG=g.enumerate(gg,false);
            for(int i=stufe; --i>=0;) sb.append(' ');
            sb.append("\n*** Group  ").append(g.getName())
                .append(" #groups=").append(nG)
                .append(" #threads=").append(nT)
                .append(" ***\n");
            for(int i=0;i<nT;i++) {
                for(int j=stufe; --j>=0;) sb.append(' ');
                sb.append("Thread ").append(tt[i] instanceof ChDelay ? tt[i].toString() : tt[i].getName()).append('\n');
            }
            for(int i=0;i<nG;i++)  debugList(gg[i],stufe+1,sb).append('\n');
        }
        return sb;
    }
    public static StringBuffer debugListThreads(StringBuffer sb) {
        sb.append("\n *** Threads in ChThread:\n");
        for(int i=vRunnable.size(); --i>=0;) {
            ChThread ri=get(i,vRunnable,ChThread.class);
            if (ri!=null) sb.append(ri.name).append("  cumul time=").append(ri.cumulTime).append(" ms\n");
        }
        ThreadGroup g=Thread.currentThread().getThreadGroup(),g2;
        while ( (g2=g.getParent())!=null) g=g2;
        return debugList(g,0, sb);
    }

    public static void simulateLongComputation() {
        final long time=System.currentTimeMillis();
        double a=1.00003;
        while(true) {
            for(int i=0; i<0xFFffFFf; i++) a=a*a;
            puts("/");
            if (System.currentTimeMillis()-time>1000*20) break;
        }
    }

}
