package charite.christo;
import java.util.*;
import java.io.*;
import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;
import java.net.URL;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**
   to extract jar packages
   @author Christoph Gille
*/
public class ChZip  {
    public final static long PACK_IF_NEWER=1<<1;

    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Class File >>> */
    private static class ClassFile {
        private ZipFile jarFile;
        private String classFileEntry;
    }
    private final List<String> vCN=new ArrayList();
    String fullCN(String shortOrLongClassName, String packages[]) {
        final ClassFile ju=classFile(shortOrLongClassName,packages);
        return ju!=null && ju.classFileEntry!=null ? delSfx(".class",ju.classFileEntry).replace('/','.') : null;
    }
    private ClassFile classFile(String shortOrLongClassName, String packages[]) {
        if (shortOrLongClassName==null) return null;
        final String classFile=shortOrLongClassName.replace('.','/')+".class";
        vCN.clear();
        vCN.add(classFile);
        if (CLASSLOADER.getResource(classFile)!=null) {
            ClassFile ret=new ClassFile();
            ret.classFileEntry=classFile;
            return ret;
        }
        vCN.add(classFile);
        if (packages!=null) {
            for(String p : packages) {
                if (p==null) continue;
                if (p.endsWith(shortOrLongClassName) && p.endsWith("."+shortOrLongClassName)) {
                    ClassFile ret=new ClassFile();
                    ret.classFileEntry=p;
                    return ret;
                }
                p=p.replace('.','/');
                final String full = p+classFile;
                vCN.add( full);
                if (CLASSLOADER.getResource(full)!=null) {
                    ClassFile ret=new ClassFile();
                    ret.classFileEntry=full;
                    return ret;
                }
            }
        }
        for(ZipFile jf : jarFilesJF()) {
            for(Object o:vCN.toArray()) {
                if (o==null) continue;
                String classFileEntry=o.toString();
                final ZipEntry je=jf.getEntry(classFileEntry);
                if (je!=null) {
                    final ClassFile ret=new ClassFile();
                    ret.classFileEntry=classFileEntry;
                    ret.jarFile=jf;
                    return ret;
                }
            }
        }
        vCN.clear();
        return null;
    }

    Object javaSource(String shortOrLongClassName, String packages[],boolean returnStream) {
        final ClassFile ju=classFile(shortOrLongClassName,packages);
        if (ju==null) return null;
        final String javaFile=delSfx(".class",ju.classFileEntry)+".java";
        if (ju.jarFile==null) return returnStream ? CLASSLOADER.getResourceAsStream(javaFile) :
            ((CLASSLOADER.getResource(javaFile)!=null) ? Boolean.TRUE:null);
        final ZipEntry je=ju.jarFile.getEntry(javaFile);
        if (je!=null) {
            try {
                return ju.jarFile.getInputStream(je);
            } catch(Exception e){}
        }
        return null;
    }

    /* <<< Class-Files <<< */
    /* ---------------------------------------- */
    /* >>> Jar-Files >>> */
    private long lastFileHash;
    private File[] _jarFiles;
    private static URL[] _jarFilesURL;
    private int _jarFilesmodificationCount;
    public int jarFiles_modificationCount() { return _jarFilesmodificationCount;}
    public File[] jarFiles() {
        long sumModi=0;
        final File dir=parent;
        sumModi+=dir.lastModified();
        if (isWindows()) {
            final File ff[]=dir.listFiles();
            if (ff!=null) {
                for(File f : ff) {
                    final String s=f.toString();
                    if (s.endsWith(".jar"))  sumModi+=f.lastModified()+ ( (long)s.hashCode() << 32);
                }
            }
        }
        if (lastFileHash!=sumModi) {
            lastFileHash=sumModi;
            final List v=new ArrayList();
            final File ff[]=dir.listFiles();
            if (ff!=null) for(File f : ff) if (f.toString().endsWith(".jar")) v.add(f);
            _jarFiles=toArry(v,File.class);
            _jarFilesmodificationCount++;
            _jarFilesJF=null;
            _jarFilesURL=null;
        }
        if (_jarFiles!=null) sortArry(_jarFiles,comparator(COMPARE_NAME));
        return _jarFiles!=null ? _jarFiles : NO_FILE;
    }

    private ZipFile[] _jarFilesJF;
    public ZipFile[] jarFilesJF() {
        final File ff[]=jarFiles();
        if (_jarFilesJF==null) {
            ArrayList<ZipFile> v=null;
            for(File f : ff) {
                if (v==null) v=new ArrayList(ff.length);
                try{ v.add(new ZipFile(f)); } catch(Exception e){}
            }
            _jarFilesJF=toArry(v, ZipFile.class);
        }
        return _jarFilesJF;
    }
    /* <<< Jar-Files <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */
    public String[] joinFiles(String fn, Map<String, ZipFile> map) {
        final List<String> v=new ArrayList(99);
        for(ZipFile jf : jarFilesJF()) {
            String lines[]=readLines(inputStream(fn,jf));
            if (lines==null) continue;
            for(String s : lines) {
                final String line=withoutLastComponent(s,'#').trim();
                v.add(line);
                if (map!=null) map.put(line,jf);
            }
        }
        return strgArry(v);
    }

    public URL resource(String s) {
        if (s==null) return null;
        final ZipFile jj[]=jarFilesJF();
        for(ZipFile zf:jj) {
            final ZipEntry e=zf.getEntry(s);
            if (e!=null) return  url("jar:file://"+zf.getName()+"!/"+s);
        }
        return null;
    }
    public InputStream resourceAsStream(String s) {
        if (s==null) return null;
        final ZipFile jj[]=jarFilesJF();
        for(ZipFile zf:jj) {
            final InputStream is=inputStream(s,zf);
            if (is!=null) return is;
        }
        return null;
    }

    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> related to a directory containing Jars  >>> */
    private final File parent;
    private ChZip(File parent){this.parent=parent;}
    private static Map<String,ChZip> mapInst=new HashMap();
    public static ChZip getInstance(File parent) {
        final String key=toStrg(parent);
        ChZip z=mapInst.get(key);
        if (z==null) {
            mapInst.put(key, z=new ChZip(parent));
        }
        return z;
    }

    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> pack200 >>> */
    // pack200 -Hfalse tt.pack.gz S.jar; time unpack200 tt.pack.gz t.jar; ls -lh t.jar
    private static Set<String> setAlreadyTried=new HashSet();
    public static boolean supportForPack200() { return javaVsn()!=14;}
    public static File unpack200(File fPack, long options) {
        if (fPack==null) return null;
        if (fPack.isDirectory()) {
            StringBuffer sbError=null;
            for(File f:fPack.listFiles()) {
                final String fn=f.getAbsolutePath();
                if (!fn.endsWith(".jar.pack.gz") && !fn.endsWith(".jar.pack.bz2")) continue;
                if (new File(delSfx(".pack.gz",fn)).length()>0) continue;
                final String key=f.getAbsolutePath()+f.lastModified();
                if (setAlreadyTried.contains(key)) continue;
                setAlreadyTried.add(key);
                final File fJar=unpack200(f, options);
                if (fJar==null) {
                    if (sbError==null) sbError=new StringBuffer(999);
                    sbError.append("unpack200 "+f+" failed\n");
                }
            }
            if (sbError!=null) error(sbError);
            return null;
        }
        final String fnPack=fPack.getAbsolutePath();
        final File fJar=new File(delSfx(".pack.gz",fnPack));
        try {
            java.util.jar.JarOutputStream jostream = new java.util.jar.JarOutputStream(fileOutStream(fJar));
            java.util.jar.Pack200.newUnpacker().unpack(deflaterAccordingToExtension(new FileInputStream(fPack),fnPack), jostream);
            // Must explicitly close the output.
            jostream.close();
            return fJar;
        } catch (IOException ioe) {
            stckTrc(ioe);
            return null;
        }
    }
    /* <<< pack200  <<< */
    /* ---------------------------------------- */
    /* >>> static Utils >>> */
    public static boolean extractAll(File jf,File destDirectory) {
        try {
            return extractAll(new ZipFile(jf),destDirectory);
        } catch(Exception e) { putln(RED_CAUGHT_IN+"ChZip#extractAll ",e); return false;}
    }
    public static boolean extractAll(ZipFile jf,File destDirectory) {
        if (destDirectory==null || !destDirectory.isDirectory()) {
            error("Error in ChZip.extractAll: "+destDirectory+" is not a directory");
            return false;
        }
        try {
            final List<ZipEntry> vE=new ArrayList(1000);
            final Enumeration entries=jf.entries();
            while(entries.hasMoreElements()) vE.add((ZipEntry)entries.nextElement());
            String lastDir=null;
            long lastTime=0;
            int count=0;
            final int nE=sze(vE);
            for(int iE=0; iE<nE; iE++) {
                final ZipEntry e=vE.get(iE);
                final File f=file(destDirectory,e.getName());
                if (e.isDirectory()) {
                    mkdrs(f);
                } else {
                    final File dir=f.getParentFile();
                    final String dirS=dir.toString();
                    if (lastDir==null || !lastDir.equals(dirS)) mkdrs(dir);
                    lastDir=dirS;
                    delFile(f);
                    cpy(jf.getInputStream(e),f);
                    final long time=System.currentTimeMillis();
                    if (time-lastTime>999) {
                        sayDownloading(null,ANSI_MAGENTA+"Extracted "+jf.getName()+" #"+count+"/"+nE);
                        lastTime=time;
                    }
                    count++;
                }
            }
            sayDownloading(null,ANSI_GREEN+"Extracted "+jf.getName()+" #"+count+" done");
        } catch(Exception e) {  putln("caught  ChZip#extractAll ", e); return false;}
        return true;
    }
    public static String[] extractAllFilesWithEnding(ZipFile zf, String sRoot, String fileExtension[], File dirDest,boolean createSubdirs) {
        final ArrayList<String> vEntries=new ArrayList();
        final Enumeration entries=zf.entries();
        while(entries.hasMoreElements()) {
            final ZipEntry e=(ZipEntry)entries.nextElement();
            final String sE=e.getName();
            if (sze(sRoot)==0 || sE.startsWith(sRoot)) {
                for(String ext:fileExtension) {
                    if(ext==null || sE.endsWith(ext)) {
                        adUniq(sE,vEntries);
                        final File f= file(dirDest+"/"+(createSubdirs ? sE : sE.substring(sE.lastIndexOf('/')+1)));
                        mkdrs(f.getParentFile());
                        if (f.lastModified()<e.getTime() || f.length()!=e.getSize()) {
                            InputStream is=null;
                            try {
                                cpy(zf.getInputStream(e),f);
                            }catch(Exception ex){ putln("caught#extractAllFilesWithEnding ", ex);}
                            closeStrm(is);
                        }
                    }
                }
            }
        }
        mkdrsErr(dirDest);
        return toArry(vEntries,String.class);
    }

    public static boolean checkIntegrityOfZipFile(File f) {
        final byte[] buffer=new byte[4096];
        try {
            final ZipFile zf=new ZipFile(f);
            final Enumeration entries=zf.entries();
            while(entries.hasMoreElements()) {
                final ZipEntry e=(ZipEntry)entries.nextElement();
                BufferedInputStream is=null;
                try {
                    is=new BufferedInputStream(zf.getInputStream(e));
                    while(is.read(buffer)!=-1);
                } catch(IOException ioex) { putln("ChZip.checkIntegrity ", e); return false;}
                closeStrm(is);
            }
        } catch(Exception e){putln("ChZip.checkIntegrity ",e); return false;}
        return true;
    }

    public static InputStream inputStream(String entry, ZipFile zf) {
        final ZipEntry je=zf==null || entry==null ? null : zf.getEntry(entry);
        try { if (je!=null) return zf.getInputStream(je); } catch(Exception ex){}
        return null;
    }

    public static boolean copyZip(File fFrom, File fTo,Collection<String> skip,boolean compressed) {
        final java.util.zip.CRC32 crc32 =new java.util.zip.CRC32();
        final ZipFile zf;
        try{
            zf=new ZipFile(fFrom);
        } catch(Exception e){putln(" caught in copyZip ",fFrom); return false;}
        final ZipOutputStream zOut;
        try{
            zOut=new ZipOutputStream(fileOutStream(fTo));
        } catch(Exception e){ stckTrc(e); return false;}

        final Enumeration<? extends ZipEntry> entries=zf.entries();
        ByteArray BUFFER=new ByteArray(0);
        while(entries.hasMoreElements()) {
            final ZipEntry eIn=entries.nextElement();
            final String name=eIn.getName();
            if (skip==null || !skip.contains(name)) {
                try {
                    ByteArray ba=readBytes(new BufferedInputStream(zf.getInputStream(eIn)),BUFFER);
                    final ZipEntry e=new ZipEntry(name);
                    if (!compressed) {
                        crc32.reset();
                        crc32.update(ba.getBytes(),0,ba.getEnd());
                        e.setMethod(ZipEntry.STORED);
                        e.setCrc(crc32.getValue());
                        e.setSize(ba.getEnd());
                    }
                    zOut.putNextEntry(e);
                    wrte(zOut,ba);
                } catch(Exception ex){stckTrc(ex);}
            }
        }
        closeStrm(zOut);
        return true;
    }
    public static void main(String argv[])  {
        if ("-unpack200".equals(get(0,argv))) {
            for(int i=1; i<argv.length; i++) {
                final File f=file(argv[i]);
                if (argv[i].endsWith(".pack.gz")) {
                    ChZip.unpack200(f,PACK_IF_NEWER);
                }
            }
        }
    }
}
