package charite.christo;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static java.awt.event.MouseEvent.*;

/**HELP
   Tool-bars and menu-bars waste space on the screen.
   To save screen space they can be hidden when the mouse pointer is away.
   This is particularly useful on laptops with small screen size.
   <i>JCOMPONENT:ChUtils#buttn(TOG_COLLAPSE_COMPONENTS)!</i>
   <br><br>
   Graphical components do not collapse if the SHIFT key is held while the mouse pointer moves.

   @author Christoph Gille
*/
public class CollapseComponent extends MouseAdapter implements MouseMotionListener, AdjustmentListener, charite.christo.interfaces.ChRunnable {
    private final static Object PREVENT_COLLAPSE=new Object(), IS_COLLAPSED=new Object();
    private static CollapseComponent inst;
    private Color bgColor;
    public static void preventedBy(JComponent c) {
        if (c==null) return;
        pcp(PREVENT_COLLAPSE,PREVENT_COLLAPSE,c);
        if (c instanceof JScrollPane) {
            JScrollPane sp=(JScrollPane)c;
            if (inst==null) inst=new CollapseComponent(null,"");
            sp.getVerticalScrollBar().addAdjustmentListener(inst);
            sp.getHorizontalScrollBar().addAdjustmentListener(inst);
        }
        if (inst==null) inst=new CollapseComponent(null,"");
        c.addMouseListener(inst);
    }
    private boolean fixedSize;
    public CollapseComponent fixedSize() {fixedSize=true; return this; }
    public static void doNotCollapseNow() { doNotCollapseNow=true;}
    public void adjustmentValueChanged(AdjustmentEvent ev){ doNotCollapseNow=true;}
    @Override public void mousePressed(MouseEvent ev){
        if (gcp(PREVENT_COLLAPSE,ev.getSource())!=null) doNotCollapseNow=true;
    }
    private String where="";
    public CollapseComponent collapseNSEW(String where) { this.where=where; return this;}
    public CollapseComponent parent(JComponent parenT) { refParent=weakRef(parenT); return this;}
    private java.util.ArrayList vChilds;
    public CollapseComponent child(JComponent chilD) {
        if (chilD==null) return this;
        if (vChilds==null) vChilds=new java.util.ArrayList();
        if (!vChilds.contains(weakRef(chilD))) {
            chilD.addMouseListener(this);
            chilD.addMouseMotionListener(this);
            vChilds.add(weakRef(chilD));
        }
        return this;
    }
    public CollapseComponent(JComponent c,String where) {
        this.where=where;
        refComponent=weakRef(c);
        if (c!=null) {
            c.addMouseListener(this);
            c.addMouseMotionListener(this);
        }
    }
    private Object refParent=weakRef(null), refComponent=weakRef(null);
    @Override public void mouseDragged(MouseEvent ev) {}
    @Override public void mouseEntered(MouseEvent ev) {
        final JComponent component=(JComponent)deref(refComponent);
        if (component==null) return;
        final int modi=ev.getModifiers();
        if (0==(modi&(SHIFT_MASK|CTRL_MASK|BUTTON1_MASK|BUTTON2_MASK|BUTTON3_MASK))) {
            final Dimension dimOriginal=fixedSize ? gcp(IS_COLLAPSED,component,Dimension.class) : null;
            pcp(IS_COLLAPSED,null,component);
            if (dimOriginal!=null) { component.setPreferredSize(dimOriginal); if (bgColor!=null) component.setBackground(bgColor);}
            for(int i=sze(vChilds); --i>=0;)   pcp(IS_COLLAPSED,null, get(i,vChilds));
            mouseExited=false;
            component.repaint();
            if (buttn(TOG_COLLAPSE_COMPONENTS).s()) ChDelay.runAfterMS(EDT,this,"UPDATE",111);
        }
    }
    private boolean mouseExited;
    @Override public void mouseExited(MouseEvent ev) {
        final JComponent component=(JComponent)deref(refComponent);

        if (component==null) return;
        if ((ev.getSource()!=component || isShift(ev))  ||  buttn(ev)==1) return;
        if (!buttn(TOG_COLLAPSE_COMPONENTS).s()) return;

        final int y=ev.getY(),x=ev.getX();
        if (
            where.indexOf('S')>=0 && y>=component.getHeight() ||
            where.indexOf('E')>=0 && x>=component.getWidth() ||
            where.indexOf('N')>=0 && y<=0 ||
            where.indexOf('W')>=0 && x<=0) {
            mouseExited=true;
            if (!isRunning) {
                ChDelay.newThread(this,"COLLAPSE",null).start();
            }
        }
    }
    @Override public void mouseMoved(MouseEvent ev) { if (!isShift(ev)) mouseExited=false; }
    private boolean isRunning;
    private static boolean doNotCollapseNow;
    public Object run(String id, Object arg) {
        if (id=="COLLAPSE") {
            final JComponent component=(JComponent)deref(refComponent);
            if (component==null) return null;
            doNotCollapseNow=false;
            final  Color COLOR=C(0x646464,10);
            try {
                isRunning=true;
                final long time=System.currentTimeMillis();
                while(System.currentTimeMillis()-time<1111) {
                    sleep(33);
                    final Graphics g=component.getGraphics();
                    if (g==null) return null;
                    g.setColor(COLOR);
                    g.fillRect(0,0,99999,9999);
                    if (!mouseExited)  return null;
                    if (doNotCollapseNow) {
                        ChDelay.repaint(component,333);
                        return null;
                    }
                }
                pcp(IS_COLLAPSED,component.getPreferredSize(),component);
                for(int i=sze(vChilds); --i>=0;) {
                    final JComponent c=get(i,vChilds,JComponent.class);
                    if (c!=null) pcp(IS_COLLAPSED,c.getPreferredSize(),c);
                }
                if (fixedSize) { bgColor=component.getBackground(); component.setPreferredSize(dim(1,1)); setBG(0xFF0000,component);}
                ChDelay.runAfterMS(EDT,this,"UPDATE",111);
            } finally{ isRunning=false;}
        }
        if (id=="UPDATE") {
            final JComponent component=(JComponent)deref(refComponent), parent=(JComponent)deref(refParent);
            if (component!=null) {
                final JComponent componentParent=parnt(component,JComponent.class);
                if (componentParent!=null) componentParent.revalidate();
                component.revalidate();
                if (parent!=null) { parent.revalidate();}
            }
        }
        return null;
    }

    public static boolean isCollapsed(JComponent c) { return gcp(IS_COLLAPSED,c)!=null && buttn(TOG_COLLAPSE_COMPONENTS).s(); }
    public static void unCollapse(JComponent c) {
        if (isCollapsed(c)) {
            pcp(IS_COLLAPSED,null,c);
            Container parent=parnt(c,Container.class);
            if (parent instanceof JComponent) ((JComponent)parent).revalidate();
        }
    }
    public static boolean paintComponent(JComponent c,Graphics g) {
        if (isCollapsed(c) && buttn(TOG_COLLAPSE_COMPONENTS).s()) { g.setColor(C(0xFF0000)); g.fillRect(0,0,9999,9999); return true;}
        return false;
    }

}
