package charite.christo;
import charite.christo.interfaces.*;
import java.awt.event.*;
import java.io.File;
import java.util.*;
import javax.swing.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
/**HELP
   The customize-dialogs are used to set program options such as
   the Web-browser and the  text viewer.

   They are opened with  customize buttons
   <i>ICON:IC_CUSTOM</i> at the left top or in the option-menu.

   <br>
   Data-types:
   <ul>

   <li>Single WIKI:Executable: Can be the absolute Path of an executable
   file such as <I>/opt/acrobat5/acroread</I>.
   Only the first line is evaluated, the other lines
   can be used for notes.
   A file-path can start with an environment variable.
   In Windows a shell variable can be written with a leading dollar like $ProgramFiles  or with percent signs like %ProgramFiles%.
   </li>

   <li> List of shell commands: The lines contain alternative shell
   commands to start programs for one particular task.   There are usually several
   programs. For example there are many different pdf-viewers around.

   All lines are successively executed one after the other until one
   line succeeds because the software designated in the line is
   installed on the machine.

   This is tested by STRAP using the  WIKI:Exit_status.

   <b>Note for Windows-Users:</b> The prefix "<b>start</b>" which is
   used to start programs under Windows does not return proper exit
   values.  As a consequence, STRAP will always assume that the respective program was successfully started.
   <br>
   Comments are preceded by "#".

   </li>

   <li>Single text string</li>
   <li>Lists of text strings, one item per line</li>
   </ul>
   <br> <br>

   The settings are saved in text files in the directory <b>~/.StrapAlign</b>.

   <br><br><b>Windows:</b>
   The %PATH%-variable does not contain all WIKI:Executable files and
   therefore full file paths are often needed.
   Adding <I>START</I> at the beginning of a command line may avoid the need to enter absolute paths.
   <br>

   <br><br><b>Using a new shell</b>

   For using complicated shell commands  with pipes, redirection, loops etc. the
   expression should be started with "SH " ("CYGWINSH " on Windows).

   <br><br><b>White space in file names:</b> According to the
   WIKI:Url_encoding, white space in directory and file names must be
   encoded by "%20".  Nevertheless, on standard computer systems, white space
   is generally avoided in file names, and this will only be the case on MS-Windows.

   <br><br><b>Setting values from file:</b> With specific command-line options,
   further lines can be added. These lines are added to the default
   settings or if changed by the user to the text entered by the
   user. The command line option is formed by appending the name of the
   customize dialog to the prefix "customizeAdd". For example to add
   script lines you add the command line parameters
   -customizeAddScriptByRegex <i>file</i>. File is either a file or an
   URL. If the program is started from an URL, then
   "&customizeAddScriptByRegex=<i>file</i>" is added to the URL. Note that
   URL-encoding needs to be applied to file.

   <hr>

   <i>JCOMPONENT:Customize#docuView()</i>

   @author Christoph Gille
*/
public final class Customize implements ActionListener, HasPanel, ChRunnable, HasModiCount {

    public static JComponent docuView() {
        return Customize.customize(webBrowser).getPanel(0);
    }
    public final static int
    databases=Hyperrefs.PLAIN_TEXT,
        proteinDatabases=Hyperrefs.PROTEIN_FILE,
        webLinks=Hyperrefs.WEB_LINK,

        C_compiler=11,
        CplusPlus_compiler=12,
        latexEditors=13,
        dviPsConverter=14,
        dviViewer=15,
        fileBrowsers=16,
        file_directories=17,
        fortran_compiler=18,
        htmlEditors=19,
        javaSourceEditor=20,
        laf=21,
        pdfViewer=22,
        proteinFileExtensions=23,
        psPoster=24,
        psViewer=25,
        textEditors=27,
        watchLogFile=28,
        webBrowser=29,
        error=33,
        pdbSite=34,
        pdflatex=35,
        latex=37,
        openFileWith=38,
        scriptByRegex=39,
        execAminoSeq=40,
        execResidueSelection=41,
        ecClassLinks=42,
        webSearches=43,
        pdbLinks=44,
        uniprotLinks=45,
        testProxySelector=46,
        keggOrganisms=47,
        speciesIcons=48,
        MAX_ID=50,
        EXECUTABLE=1001, EXECUTABLES=1002,
        LATEXBEGIN=1003,
        LATEXEND=1004, LIST=1005, LIST_OF_CLASSES=1006, FILE=1007, FILES=1008, TEXT_LINES=1009, LIST_OF_DIRECTORIES=1010,
        SHELL_SCRIPT=1011, CLASSPATH=1012, MAP_TO_1_FROM_N=1013, URL=1014, LIST_INTERN=1015,
        MAX_VARIABLE=1111,
        FLAG_REPLACE=0x8000, FLAG_NOTIFY_IMMEDIATELY=0x4000;

    private final int type;
    private final String name,defaultSetting[],  help;
    private final boolean notifyImmediately;
    private final static Customize CUSTOMIZE[]=new Customize[MAX_ID+1];
    private static volatile String[] names;
    private int modi, modiHidden;
    public final static Customize[] NONE={};
    public int getModificationCount() {return modi;}
    @Override public String toString() { return name+"."+get(type,getFinalStaticInts(Customize.class,MAX_VARIABLE));}
    private List<String> vHidden, vSettings;
    /* ---------------------------------------- */
    /* >>> Instance >>> */
    final static Map<String,Customize> mapInstances=new HashMap();
    public Customize(String name, String[] defaultSetting, String help, int type) {
        this.notifyImmediately= (type&FLAG_NOTIFY_IMMEDIATELY)!=0;
        this.name=name;
        this.type=type&0xffff;
        this.help=help;
        if ((type&FLAG_REPLACE)!=0) { addHiddenEntries(false, defaultSetting); this.defaultSetting=NO_STRING; }
        else this.defaultSetting=defaultSetting;
        mapInstances.put(name,this);
        downloadFiles(new String[]{name});
    }

    private boolean needUpdate;
    private void myUpdate() {
        getSettingsV();
        handleActEvt(this, ACTION_CUSTOMIZE_CHANGED,0);
    }

    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Parsing text >>> */
  public final static String toCommandLineOption(String name) { return "customizeAdd"+capitalize(name); }
    private final static Set<String>  downloadFiles_done=new HashSet();
    public static void downloadFiles(String namesOfCustomize[]) {
        final String names[]=namesOfCustomize!=null ? namesOfCustomize : getFinalStaticInts(Customize.class, MAX_VARIABLE);
        final String argv[]=getPrgParameters();
        if (argv==null) putln(RED_ERROR," Forgot setPrgParameters()");
        if (names==null) {stckTrc(); return; }
        for(String name : names) {
            if (downloadFiles_done.contains(name)) continue;
            downloadFiles_done.add(name);
            for(String param : getParamValues(toCommandLineOption(name), argv)) {
                if (looks(LIKE_EXTURL,param)) {
                    final File f=file(param);
                    final java.net.URL url=url(param);
                    if (f!=null && url!=null && 0==isUpToDate(f,url)) {
                        f.delete();
                        newThrd(thread_download(url,0), thread_updateAfterDL(name)).start();
                    }
                }
            }
        }
    }

    public static Thread thread_updateAfterDL(String name) { return threadM("updateAfterDL", Customize.class,new Object[]{name});}
    public static void updateAfterDL(String name) {
        final Customize c=mapInstances.get(name);
        putln(ANSI_FG_RED+"Customize updateAfterDL "+c+ANSI_RESET);
        if (c!=null) {
            c.modiHidden++;
            handleActEvt(c, ACTION_CUSTOMIZE_CHANGED,0);

        }
    }

    private String[] getLinesFromParams() {
        final String argv[]=getPrgParameters();
        if (argv==null) putln(RED_ERROR," Forgot setPrgParameters()");
        List<String> vLinesFromParam=null;
        for(String param : getParamValues(toCommandLineOption(this.name), argv)) {
            final String[] lines=readLines(file(param));
            if (sze(lines)>0) {
                if (vLinesFromParam==null) vLinesFromParam=new ArrayList(lines.length);
                adAll(lines, vLinesFromParam);
            }
        }

        return strgArry(vLinesFromParam);
    }
    private ByteArray readTxt() { return readBytes(file("~/@/customize/"+this.name));}
    public synchronized List<String> getSettingsV() {
        final ByteArray txt= this.ta!=null ? this.ta.tools().byteArray() : null;
        final int mc=this.modiHidden+modiCount(txt);
        List<String> v=this.vSettings;
        if (v==null || this.modi!=mc) {
            this.modi=mc;
            this.txt4completion=null;
            this.highlights=null;
            settings=null;
            final ByteArray ta_or_saved=txt!=null ? txt : readTxt();
            final String[] add1=strgArry(vHidden);
            final String[] add2=getLinesFromParams();
            final String[] add3=ta_or_saved!=null ? splitLnes(SPLIT_TRIM,ta_or_saved) : defaultSetting;

            if (v==null) this.vSettings=v=new ArrayList(sze(add1)+sze(add2)+sze(add3)); else v.clear();

            for(int iTxt=3;--iTxt>=0; ) {
                final String[] lines=iTxt==0 ? add1 : iTxt==1 ? add2 : add3;
                if (sze(lines)==0) continue;
                for(String s : lines) {
                    if (s==null) continue;
                    s=s.trim();
                    if (type==LIST_INTERN) s=s.intern();
                    if (sze(s)>0) v.add(s);
                }
            }
        }
        return v;
    }
    public void addHiddenEntries(boolean uniq, String... ss) {
        if (vHidden==null) vHidden=new ArrayList(ss.length);
        if (adAll(uniq, ss,vHidden)) modiHidden++;
    }
    /* <<< Parsing text <<< */
    /* ---------------------------------------- */
    /* >>> Settings >>> */
    private String[] settings;
    public String[] getSettings() {
        final List v=getSettingsV();
        if (settings==null) settings=strgArry(v);
        return settings;
    }
    public String getSetting() {
        for(String s:getSettings()) if (sze(s)>0) return s;
        return "";
    }

    /* <<< static <<< */
    /* ---------------------------------------- */
    /* >>> GUI >>> */
    private JComponent pMain;
    private ChTextArea ta;
    private JFileChooser fs;
    private ChButton butSave,butReset;
    public JComponent getPanel(int options) {
        if (pMain==null && options!=NO_NEW_PANEL || options==NEW_PANEL) {

            this.ta=new ChTextArea("");
            pcp(ACTION_FOCUS_GAINED,"",ta);
            ta.setLineWrap(false);
            addActLi(this,ta.tools().enableWordCompletion("").enableUndo(true));

            final ByteArray saved=readTxt();
            //if (saved!=null) ta.t(saved.newString());
            ta.t(saved!=null ? saved.newString() : ChUtils.joinLines(this.defaultSetting));
            ta.tools().underlineRefs(UNDERLINE_REFS_NOT_CLICKABLE|UNDERLINE_WEB_COLORS);
            final int t=this.type;
            final JComponent
                pButtons=hB(
                            new ChButton("H").li(this).t("Help"),
                            new ChButton("DS").li(this).t("Default settings"),
                            helpBut(Customize.class).t("More help"),
                            butReset=new ChButton("R").li(this).t("Reset ..."),
                            butSave=new ChButton("Save").li(this).tt("write to <b>home</b>-directory"),
                            t==EXECUTABLE||t==EXECUTABLES||t==FILE||t==FILES ?
                            new ChButton("SEAERCH").t("search on hard-disk").li(this).tt("Search file on HD") : null
                            );
            butSave.setEnabled(false);
            pMain=newCNSEW(new JScrollPane(ta),"Type: "+get(t,getFinalStaticInts(Customize.class,MAX_VARIABLE)), pButtons,null,null);
            addActLi(this,ta);
            TabItemTipIcon.set(name,name,help,null,pMain);
        }
        return pMain;
    }
    public static ChButton newButtonAll(String id) {
        return
            id=="W" ?   newButton(Customize.databases, Customize.webLinks, Customize.proteinDatabases, Customize.webSearches, Customize.uniprotLinks, Customize.ecClassLinks).t("Web resources") :
            id=="B" ? newButton(Customize.webBrowser).t("Select a Web browser ...").i(IC_WWW) :
            newButton(webBrowser, fileBrowsers, pdfViewer, textEditors,proteinFileExtensions, watchLogFile,javaSourceEditor).t("Utilities").i(IC_CUSTOM);
    }
    public static ChButton newButton(int... IDs) {
        Customize cc[]=new Customize[IDs.length];
        for(int i=cc.length; --i>=0;) cc[i]=customize(IDs[i]);
        return newButton(cc);
    }

    public ChButton newButton() { return newButton(new Customize[]{this}); }
    public static ChButton newButton(Customize... cc) {
        if (inst==null) inst=new Customize();
        final ChButton b=new ChButton("DIALOG").t("Customize").li(inst).cp("CC",cc);
        final StringBuffer sb=new StringBuffer("Customize dialog for<ul>");
        for(int i=0; i<sze(cc);i++) {
            if (cc[i]!=null) sb.append("<li>").append(cc[i].name).append("</li>");
        }
        return b.tt(sb.append("</ul>")).i(IC_CUSTOM);
    }
    public ChCombo newComboBox() {
        ChCombo c=new ChCombo(this.type==LIST_OF_CLASSES ? ChJTable.CLASS_RENDERER:0, getSettingsV() );
        UIupdateList.add(c,UIupdateList.SETTINGS_CHANGED);
        return c;
    }
    /* <<< GUI <<< */
    /* ---------------------------------------- */
    /* >>> static GUI >>> */
    private static JComponent panel;
    private static ChTabPane parentTabbedPane, tabbedPane;
    private static ActionListener li;
    private Customize(){ type=0; name=help=""; defaultSetting=null; notifyImmediately=false;}
    private static Customize inst;
    public static void setParent(ChTabPane tp,ActionListener alLi) { parentTabbedPane=tp; li=alLi;  }
    public static void addDialog(Customize cus) {
        if (cus==null) return;
        if (panel==null) {
            JComponent pTop=hB("#","<h2>Settings/Preferences</h2>","#",panHS(Customize.class));
            tabbedPane=new ChTabPane(ChTabPane.RIGHT);
            addActLi(li,tabbedPane);
            panel=newCNSEW(tabbedPane,pTop,null,null,null);

            if (parentTabbedPane!=null) pTop.add(new ChButton().actionClose(panel));
            TabItemTipIcon.set(null,null,null,IC_CUSTOM,panel);
        }
        if (parentTabbedPane==null) {
            ChFrame.frame("Customize",panel,0).size(600,400).shw(0);
        }  else parentTabbedPane.addTab(panel);
        tabbedPane.addTab(cus.getPanel(0));
    }
    static Map mapSettings() { return mapSettings;}
    private final static Map<Object,Object> mapSettings=new WeakHashMap();
    private final static List<Customize> v=new ArrayList();
    public static Customize[] cc(Object o) {
        if (o==null) return NONE;
        v.clear();
        Object settings=mapSettings().get(o);
        if (settings==null && !(o instanceof Class)) settings=mapSettings().get(o.getClass());
        for(int i=sze(settings); --i>=0;) {
            final Customize c=settings instanceof Customize[] ? ((Customize[])settings)[i] : settings instanceof int[] ? customize(((int[])settings)[i]) : null;
            if (c!=null) v.add(c);
        }
        return toArry(v,Customize.class);
    }
    /* <<< static GUI <<< */
    /* ---------------------------------------- */
    /* >>> ActionEvent >>> */
    public void actionPerformed(ActionEvent ev) {
        final String cmd=ev.getActionCommand();
        final Object q=ev.getSource();
        final File f=file("~/@/customize/"+name);
        if (cmd=="H") {
            ChTextComponents.tools(new ChTextViewer("<h2>Help for this specific entry</h2>\n"+this.help).shw(0).getTextComponent()).underlineRefs(UNDERLINE_REFS_NO_ICON);
        }
        if (cmd=="DIALOG") {
            Customize cc[]=(Customize[]) gcp("CC",q);
            if (cc!=null) for(Customize c:cc) Customize.addDialog(c);
            return;
        }
        if (ta!=null) {
            if (cmd=="DS") {
                final StringBuffer sb=join(this.defaultSetting,"\n",new StringBuffer());
                if (sze(this.defaultSetting)==0 && this.vHidden!=null) for(Object o:vHidden.toArray()) sb.append(o).append('\n');
                new AsciiView(sb).showInFrame("Default settings").tools().underlineRefs(UNDERLINE_REFS_NOT_CLICKABLE);
            }
            if (cmd=="R") {
                if (!ChMsg.yesNo("Do you really want to revert to original settings and discard your changes?")) return;
                ta.t(joinLines(defaultSetting).append('\n'));
                delFile(f);
                butSave.setEnabled(false);
                butReset.setEnabled(false);
                myUpdate();
                if (isCtrl(ev)) {
                    final ByteArray txt=ta.tools().byteArray();
                    final byte[] bb=txt.getBytes();
                    final int end=txt.getEnd();
                    int i=ta.getCaretPosition();
                    while( --i>=0 && bb[i]!='\n');
                    int to=i;
                    while( ++to<end && bb[to]!='\n' );
                    final String command=new String(bb,i+1,to-i-2);
                    new Thread(new ChExec(0).setCommandLine(command).showStdoutAndStderr(null)).start();
                }
            }
            if (cmd=="SEAERCH") {
                if (fs==null) fs=new ChFileChooser(Customize.class);
                fs.setMultiSelectionEnabled(false);
                if (fs.showOpenDialog(SwingUtilities.getRootPane(getPanel(0)))!=JFileChooser.APPROVE_OPTION) return;
                ChTextComponents.insertAtCaret(0L, rplcToStrg(" ","%20",toStrg(fs.getSelectedFile())), ta);
            }
            if (q==ta) {
                if (butSave!=null) butSave.setEnabled(true);
                if (butReset!=null) butReset.setEnabled(true);
                if (cmd==ACTION_TEXT_CHANGED) needUpdate=true;
                if (needUpdate) {
                    if (notifyImmediately  || (cmd==ACTION_FOCUS_LOST || cmd==ACTION_FOCUS_GAINED)) {
                        needUpdate=false;
                        myUpdate();
                    }
                }
            }
            if (q==butSave) {
                wrte(f, this.ta.tools().byteArray());
                butSave.setEnabled(false);
            }
        }
    }
    /* <<< ActionListener <<< */
    /* ---------------------------------------- */
    /* >>> Create defaults >>> */
    public static Customize customize(final int id) {
        if (id<0) return null;
        if (CUSTOMIZE[id]!=null) return CUSTOMIZE[id];
        final boolean isWin=isSystProprty(IS_WINDOWS), isMac=isMacintosh(), isWin95=isSystProprty(IS_WINDOWS_95_FAMILY);
        final String
            notepadPP="%ProgramFiles%/Notepad++/Notepad++\n",
            INFO_J="<br>J_JAVA_EDITOR denotes an editor with syntax highlighting which will be automatically installed.<br>ECLIPSE denotes the Eclipse IDE.",
            J_ED="J_JAVA_EDITOR\n",
            HELP_COMPILERS=
            "A WIKI:Compiler produces the WIKI:Executable (dot exe file) from the program's WIKI:Source_code.<br>\n"+
            "If the compiler is installed then typing\n"+
            "<pre>cc or c++ or gcc or gcc++</pre>\n"+
            "should prompt:<br>\n"+
            "<pre> no input files </pre>\n"+
            "The switch -lm  includes the math-library.<br>\n"+
            "The switch -O2 means optimization level 2.<br>\n"+
            "If the command f77 for compiling Fortran is not available try f95\n";
        final String h;
        final Object ss;
        final int type;
        final String ms_fph="\nrundll32 url.dll,FileProtocolHandler\n", ms_start=isWin95 ? "\nstart " : "cmd.exe /c start ";
        if (id==openFileWith|| id==execResidueSelection||id==execAminoSeq) {
            ss=ad("SH echo $* > /tmp/t;xterm -e less /tmp/t",customize(textEditors).getSettings(),0);

            final String what= id==execAminoSeq ? "amino acid sequences" : id==execResidueSelection ? "residue selections" : " files";
            h="Shell programs to process "+what+". Used for context menu of "+what+".";
            type=LIST;
        } else if (id==scriptByRegex) {
            h="Two columns: <ol>"+
                "<li>Regular expression, that can match or not matchShell programs to process </li>"+
                "<li>Shell command which can be executed if regular expression matches.</li>"+
                "</ol>";

            ss=
                "#Tab-separated Columns (Type %09 for tab):\n"+
                "#  1st Column: Display name *\n"+
                "#  2nd Column: A regular expression *\n"+
                "#  3rd Column: A shell script *\n\n"+
                "demo\\Example: Matches any%09.*%09 echo Hello, this matches any *\n"+
                "demo\\Example google%09.*%09http://www.google.com/search?q=* \n"+
                "demo\\Example: Matches starting with M%09^M.*$%09 echo starts with letter M *\n"+
                "demo\\Example: Matches numbers%09^.*[0-9]$%09 echo ends with a digit *\n"+
                "util\\GEO_Datasets%09^ENSG[0-9][0-9].*%09 fgrep 'ARGUMENT' ~/testMetannogenExpr/GEO_Datasets.txt\n";
            type=LIST|FLAG_REPLACE;
        } else if (id==pdflatex || id==latex) {
            final String P=id==latex ? "latex" : "pdflatex", NS=" --interaction=nonstopmode\n";
            ss=isWin ? P+NS+"CYGWINSH "+P+NS :
                isMac ?
                "/Applications/texmaker.app\n"+
                "/Applications/tex/LaTeXiT.app\n"+
                "/Applications/tex/TeXShop.app\n"+
                "/Applications/TeXShop.app/Contents/MacOS/TeXShop\n"+
                "/usr/local/teTeX/bin/powerpc-apple-darwin-current/"+P+"\n"+
                "/usr/texbin/"+P+NS :
                P+NS;
            type=EXECUTABLES;
            h="<b>The command "+P+" takes  a latex-file (ending .tex) as input and produces a "+(id==latex ?"PostScript":"PDF")+" file. <br>WIKI:LaTeX home page: http://www.latex-project.org";
        } else if (id==latexEditors || id==textEditors) {
            ss=isWin ? notepadPP+ms_start+"wordpad\n"+ms_start+"winword\n"+ms_start+"edit.com\n"+J_ED :
                isMac ? "/Applications/TextEdit.app/Contents/MacOS/TextEdit\n"+J_ED :
                "gnuclient\nkwrite\nkedit\ngedit\nemacs\nxemacs\nnedit\n# xterm -e vi\n# xterm -e pico\n"+J_ED;
            h="Text editor. <br>"+INFO_J;
            type=EXECUTABLES;
        } else if (id==databases || id==proteinDatabases || id==webLinks) {
            type=LIST|FLAG_REPLACE;
            h="Expressions of the form DATABASE:identifier will be translated into URLS.<br>"+
                "E.g. PDB:1RYP is translated into a link for the PDB-database.<br>"+
                "If an asterisk is present, then it is replaced by the ID otherwise the ID is appended at the end.\n"+
                "The user can enter new URLs or override the default settings.\n";
            ss=readLines(rscAsStream(0L, _CC,get(id,getFinalStaticInts(Customize.class,MAX_VARIABLE))+".rsc"));
        }  else if (id==speciesIcons) {
            type=LIST|FLAG_REPLACE;
            h=
                "Mapping of species names to icon files.<br>"+
                "The text contains lines with base URLs.<br>"+
                "The other lines contain two columns separated by space: "+
                "<ol><li>scientific-species-name</li><li>icon file</li></ol>\n";
            ss=readLines(rscAsStream(0L, _CC,"speciesIcons.rsc"));
        } else switch(id) {
            case keggOrganisms:
                type=LIST|FLAG_REPLACE;
                h="Kegg Organisms";
                ss=readLines(rscAsStream(0L, _CCM,"keggOrganisms.rsc"));
                break;
            case psPoster:
                ss="poster -mA4";
                h="Scale and tile a postscript image to print on multiple pages\nDownload poster at http://www.geocities.com/SiliconValley/5682/poster.html";
                type=EXECUTABLES;
                break;
            case pdfViewer:
                ss=isWin ? ms_fph+"cmd /C\"start c:\\document.doc\"\ncommand /C\"start c:\\document.doc\"\nacroread" :
                    isMac ? "open" : "evince\nxpdf\nacroread";
                h="PDF-viewer.";
                type=EXECUTABLES;
                break;
            case dviViewer:
                ss=isWin ? ms_fph : "xdvi\nkdvi";
                h="DVI is a document format generated by latex and  converted  to PostScript with dvips.";
                type=EXECUTABLES;
                break;
            case psViewer:
                ss=isWin ? "CYGWINSH gv\n"+ms_fph+ms_start+"psview\n" : "gv\nkgv\nggv\nkghostview\nevince\n" ;
                h="PostScript is printable document format.\n";
                type=EXECUTABLES;
                break;
            case dviPsConverter:
                ss="dvips $* -o $*.ps\n"+(isWin ? "CYGWINSH dvips $* -o $*.ps" : "");
                h="Conversion dvi to PostScript ";
                type=EXECUTABLE;
                break;
            case fileBrowsers:
                ss=isWin ? ms_start+"explorer \"$*\"" :
                    isMac ? "open -a Finder" :
                    "nautilus\npcmanfm\nthunar\nkonqueror\nkbear\nxfe\ngcmd\nemacs\ngnuclient\n"+
                    "sh -c xterm%20-title%20*%20-e%20'cd%20*;ls%20-l;/bin/sh'\n"+
                    "xfce4-terminal --title * --working-directory=*\n"+
                    "gnome-terminal --title * --working-directory=*\n"+
                    "konsole -T * --workdir *\n";

                h="File browser ";
                type=EXECUTABLES;
                break;
            case htmlEditors:
                ss=isWin ? ms_start+"winword\n"+ms_start:
                isMac ? "open -a Microsoft%20Word\nopen -e \n# open -a neooffice \n# open -a openoffice \n# open -a pages" :
                    "ooffice3\nooffice2\nooffice\nopenoffice\nsoffice\nabiword";
                h="Word-Processor like Openoffice or MS-Word";
                type=EXECUTABLES;
                break;
            case javaSourceEditor:
                ss=isMac||isWin ? notepadPP+"#ECLIPSE\n"+J_ED :
                "#ECLIPSE\ngnuclient\nkwrite\ngedit\nkedit\n"+J_ED;
                h="Editor for  Java source text.<br>"+INFO_J;
                type=EXECUTABLES;
                break;
            case webBrowser:
                final String expl="EXPLORER.EXE \"$*\"\n";
                ss=isWin ? (isWin95?expl:ms_fph)+(isWin95?ms_fph:expl)+ms_start+"netscape \"$*\"\n"+ms_start+"mozilla \"$*\"":
                    isMac ? "open\nopen -a Safari":
                    "firefox\niceweasel\nicecat\ngnome-www-browser\nkonqueror\nmozilla\nopera\nnetscape\nepiphany";
                h="A web browser.";
                type=EXECUTABLES;
                break;
            case file_directories:
                ss="./  (working directory)\n/pdb/\n/local/pdb/";
                h=
                    "Assume the first line contains <li>/pdb/</li> and the 2nd line contains /home/heinrich/myProteinFiles/<br>"+
                    "A protein file e.g. pdb1ryp.ent will searched in the following locations:\n<ol>"+
                    "<li>The project directory,</li>"+
                    "<li>/pdb/</li>"+
                    "<li>/pdb/ry/</li>"+
                    "<li>/home/heinrich/myProteinFiles/</li>"+
                    "</ol>";
                type=LIST_OF_DIRECTORIES;
                break;
            case laf:
                final UIManager.LookAndFeelInfo lafs[]=UIManager.getInstalledLookAndFeels();
                final StringBuffer sb=new StringBuffer(333);
                for(int i=lafs.length; --i>=0;) sb.append(lafs[i].getClassName()).append('\n');
                ss=sb.toString();
                h="Look and Feel";
                type=LIST_OF_CLASSES;
                break;
            case proteinFileExtensions:
                ss=
                    "all # all files (may cause exception when there are too many)\n"+
                    ".pdb .ent # structure file\n"+
                    ".dssp # secondary structure file\n"+
                    ".swp .swiss # Swissprot sequence file\n"+
                    ".swissprot .sw # Swissprot sequence file\n"+
                    ".embl # EMBL sequence file\n"+
                    ".genbank .gb # Genbank (see menu 'translate dna ..')\n"+
                    ".seq # sequence file\n"+
                    ".dna # nucleotides a,c,t,g\n"+
                    ".fa .fasta .mfa # fasta-format\n"+
                    ".ascii .txt # ascii text\n"+
                    ".msf .clustalw .aln # multiple sequence file containing more than one sequence\n"+
                    "other # any other extension";
                h="File extensions do not influence the way <br>the proteins are parsed !!";
                type=LIST;
                break;
            case watchLogFile:
                ss=isWin ? ms_start+"Firefox \"$*\"\n"+ms_fph+ms_start+"EXPLORER \"$*\"" :
                    isMac ? "open -e " :
                    "SH (xterm -title less_$* -e less $*&);(xterm -title tail_-f_$* -e tail -f $*)" ;
                h="shell command to watch a log file";
                type=EXECUTABLES;
                break;
            case C_compiler:
                ss="cc -lm -O2";
                h=HELP_COMPILERS;
                type=SHELL_SCRIPT;
                break;
            case CplusPlus_compiler:
                ss="c++ -lm -O2";
                h=HELP_COMPILERS;
                type=SHELL_SCRIPT;
                break;
            case fortran_compiler:
                ss=(!isWin && cmdExists("gfortran") ? "gfortran" : !isWin && cmdExists("f95") ? "f95" : "f77")+" -O2";
                h=HELP_COMPILERS;
                type=SHELL_SCRIPT;
                break;
            case pdbSite:
                final String dir= isWin ? "\\myPDB\\pdb\\??\\pdb????.ent" : "/pdb/??/pdb????.ent";
                ss=
                    "http://www.rcsb.org/pdb/files/????.pdb.gz\n"+
                    "ftp://pdb.protein.osaka-u.ac.jp/v3/pub/pdb/data/structures/divided/pdb/??/pdb????.ent.gz\n"+
                    "ftp://ftp.wwpdb.org/pub/pdb/data/structures/divided/pdb/??/pdb????.ent.gz\n"+
                    "ftp://ftp.ebi.ac.uk/pub/databases/rcsb/pdb-remediated/data/structures/divided/pdb/??/pdb????.ent.gz\n"+
                    dir+"\n"+
                    dir+".Z";
                h="the location of the pdb-file collections.<br>Files and URLs are allowed.";
                type=LIST;
                break;
            case ecClassLinks:
                h="A list of URLs for creation external links by EC number. Asterisk is place-holder for EC number.";
                ss="STRAPec{*}\nMETACYCec{*}\nNISTec{*}\nBRENDAec{*}\nREACTOMEec{*}\nEXPASYec{*}\nKEGGec{*}\nIUBMBec{*}\nUMBBDec{*}\nUNIPROT{*}\nENSG{*}\nSABIORKec{*}";
                type=LIST;
                break;
            case webSearches:
                h="A list of URLs for looking up a word or phrase. Used in context menu of text components.";
                ss=" WIKIPEDIA{*}\n GOOGLE{*}\n PUBMED{*}\n UNIPROT{*}\n TDB{*}\n ENSG{*}\n ISI{*}\n CITEXPLORE{*}\n METACYC{*}\n";
                type=LIST;
                break;
            case pdbLinks:
                h="A list of PDB links. Asterisk is place-holder for PDB id.";
                ss="PDBj:*\nPDB_SUM:*\nPDB_RCSB:*\nJENA:*\nPDBe:*\nVIPERDB:*";
                type=LIST;
                break;
            case uniprotLinks:
                h="A listUniprot links. Asterisk is place-holder for UNIPROT id.";
                ss="UNIPROT:* \n EXPASY:* \n BLAST{UNIPROT:*} \n LOC{UNIPROT:*} \nPFAM_PROTEIN:* \n DASTY:* ";
                type=LIST;
                break;
            case testProxySelector:
                h="A list of Web-addresses to test whether the correct WIKI:Proxy_server is selected for Web-addresses. ";
                ss=
                    "ftp://ftp.embl-heidelberg.de/pub/databases/\n"+
                    "http://www.google.de\n"+
                    "http://www.charite.de\n"+
                    "http://www-intern.charite.de\n"+
                    "http://olli.charite.de\n"+
                    "http://141.42.56.58\n"+
                    "http://localhost/strap/\n"+
                    "http://host.mydomain.com\n"+
                    "ftp://ftp.genome.jp/pub/kegg/ligand\n";
                type=LIST;
                break;

            default:
                if (CUSTOMIZE[error]==null) CUSTOMIZE[error]=new Customize("error",NO_STRING,"",LIST);
                return CUSTOMIZE[error];
            }
        if (names==null) names=getFinalStaticInts(Customize.class, MAX_VARIABLE);

        final String ss0[]=ss instanceof String[] ? (String[])ss : splitTokns(0L, ss,"\n");

        final Customize c=CUSTOMIZE[id]=new Customize(names[id],ss0,h,type);
        if (id==webLinks) {
            final String dir=" http://www.bioinformatics.org/strap/swf/";
            c.addHiddenEntries(false,
                               "Spice-Browser* http://www.efamily.org.uk/software/dasclients/spice/",
                               "CE_CL* http://cl.sdsc.edu/",
                               "Vast* http://www.ncbi.nlm.nih.gov/Structure/VAST/vasthelp.html",
                               "GangstaPlus*  http://agknapp.chemie.fu-berlin.de/gplus/",
                               "BioDAS*  http://www.biodas.org/",
                               "Dali* http://ekhidna.biocenter.helsinki.fi/dali",
                               "Asymmetric_unit* http://www.rcsb.org/robohelp/data_download/biological_unit/asymmetric_unit.htm",
                               "Launch_STRAP* http://www.bioinformatics.org/strap/strap.jnlp",
                               MOVIE_Drag_to_another_STRAP+dir+"dnd_other_strap.swf",
                               MOVIE_Load_Proteins+dir+"load_proteins.swf",
                               MOVIE_Load_Proteins_web_pdb+dir+"load_pdb.swf",
                               MOVIE_Load_Proteins_web_uniprot+dir+"load_uniprot.swf",
                               MOVIE_Export_Proteins+dir+"export_proteins.swf",
                               MOVIE_Sequence_Features_in_3D+dir+"selecting_sequence_features_in_3D.html",
                               MOVIE_Context_Menu+dir+"context_menu.swf",
                               MOVIE_Drag_protein_to_3D+dir+"drag_protein_into_3D.swf"
                               );
        }
        return c;
    }
    /* <<< Create defaults <<< */
    /* ---------------------------------------- */
    /* >>> Provide services >>> */
    private byte[][][] highlights;
    private String[][] txt4completion;
    public Object run(String id, Object arg) {
        final String ss[]=getSettings();
        if (id==PROVIDE_WORD_COMPLETION_LIST) {
            String tokens[][]=txt4completion;
            if (tokens==null) {
                tokens=txt4completion=new String[ss.length][];
                for(int i=ss.length; --i>=0;) {
                    tokens[i]=splitTokns(0L, ss[i]);
                }
            }
            return tokens;
        }
        if (id==PROVIDE_HIGHLIGHTS) {
            if (highlights==null) highlights=OccurInText.lookupTableFirstLetter(ss);
            return highlights;
        }
        return null;
    }

    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Class specific using reflection  >>> */
    private final static Map<String,Customize> mapClass2Customize=new HashMap();
    public static Customize getCustomizableForClass(Class clazz, String id, String help) {
        final String key=clazz.getName()+" "+id;
        Customize c=mapClass2Customize.get(key);
        if (c==null) {
            String lines=deref(invokeMthd("getCustomizable"+id, clazz), String.class);
            if (lines!=null) {
                c=new Customize(id+"_"+shrtClasNam(clazz),splitStrg((String)lines,'\n') , help, Customize.LIST);
                mapClass2Customize.put(key,c);
            }
        }
        return c;
    }
    public final static int  CLASS_Example=1001, CLASS_ExecPath=10002, CLASS_InitCommands=10003, CLASS_LateCommands=10004;
    public static Customize getCustomizableForClass(Class c,int id) {
        switch(id) {
        case CLASS_Example: return getCustomizableForClass(c, "Examples", "Example commands");
        case CLASS_ExecPath: return  Customize.getCustomizableForClass(c, "ExecPath", "file path of Executable");
        case CLASS_InitCommands: return Customize.getCustomizableForClass(c, "InitCommands", "Renering commands executed at program start");
        case CLASS_LateCommands: return Customize.getCustomizableForClass(c, "LateCommands", "Renering commands executed after all proteins loaded.");
        default: { assertTrue(false); return null;}
        }
    }
    /* <<< Class specific using reflection <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */

}
//http://www.biostat.wisc.edu/bcg/sup/macsup/unixtips.html
