package charite.christo;
import charite.christo.interfaces.*;
import java.io.*;
import java.awt.*;
import java.util.*;
import java.awt.dnd.*;
import javax.swing.*;

import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;

/**HELP
   <h2>STRAP-Utilities</h2>
   The directory ~/Desktop/strap_utils/ contains useful utilities in form of shell-scripts.

   The program logic resides in the file strap.jar which contains also
   the Strap workbench for proteins and the entire source code.

   <h3>Utilities for everyone</h3>
   <ul>
   <li><u>textViewer</u>  starts a fast text viewer for

   <a href="http://en.wikipedia.org/wiki/Plain_text">plain text</a> files.
   It has extended search and  highlighting features and is suitable for large texts.

   <br><br>
   <u>Start:</u>

   The script file textViewer can be started from the command line
   and takes the text file to be viewed as commmand line parameter.

   It can also be started by
   double click. In this case
   a green bar appears which serves as a drag target for the files to be viewed:
   <span style=" background-color: #00FF00; ">Drag text file here</span>

   Some desktop systems support
   <a href="http://en.wikipedia.org/wiki/Drag-and-drop">Drag_and_drop</a>
   also for script files (Windows, some Linux desktops, but not yet Mac-OSX4).

   On Macintosh Terminal window will be visible which can be closed by the user.

   Try to drag text files  onto the  textViewer script file.
   <br>
   <br>

   <u>Maximum file size:</u> The text-viewer copes with very large
   files up to several hundreds of Megabytes. The file size cannot
   exceed the RAM size of the PC. The maximum is set in the script
   file to 500 Megabytes and can be changed by making a modified copy
   of the shell script.

   <br>
   <br>
   <u>Windows</u>: The text-file to be viewed can be dragged onto textViewer.js.
   Alternatively, the following command line can be started at the  DOS shell:
   <pre>cscript textViewer.js textFile.txt</pre>
   This would not work if the script-engine is deactivated. In this case
   an equivalent batch file textViewer.bat can be created
   with the following content:
   <pre>
   CD /D %~dp0
   java -Xmx500M -jar strap.jar -viewTextFiles %*
   </pre>
   <br><u>Key-commands in the viewer:</u>
   When the text is viewed, press F1 for help on hot-keys in the text-viewer.

   </li>
   </ul>
   <h3>Utilities for life scientists</h3>
   <ul>
   <li><u>strap.jar:</u> contains the program  STRAP which is a program to compare and analyze protein sequences and structures.
   It is written for Biologists, Biochemists and Bioinformaticians.

   This can be started by double click or with the command line
   <pre>
   java -jar strap.jar
   </pre>

   See http://3d-alignment.eu for details.

   The file can be clicked to start Strap.

   When accessing the internet via a proxy server, the
   <a href="http://en.wikipedia.org/wiki/Web_proxy">Web_proxy</a>
   may need to be set manually.
   This is  because in offline mode the Web_proxy is not obtained from the Web-browser.
   </li>

   <li><u>toClustal</u>, <u>toMultipleFasta</u>, <u>toFasta</u>:
   Converting alignment and protein files.
   Please provide the protein and alignment files to be converted as
   command line parameter.</li>
   </ul>

   <h3>Deinstallation</h3>

   Remove the folders <u>~/Desktop/strap_utils the and ~/.StrapAlign/</u>
   (On Windows C:\StrapAlign\).

   @author Christoph Gille
*/
public final class DesktopUtils implements DropTargetListener, java.awt.event.ActionListener {
    public final static int UTIL_TEXT_VIEWER=1;
    private static void install(boolean becauseOfCB) {
        final File fSrc= thisJarFile();
        final int fSrcL=sze(fSrc);
        if (fSrcL>0) {
            final File fJar=file(dirSettings()+"/strap.jar");
            final int fJarL=sze(fJar);
            if (fSrcL!=fJarL || fSrc.lastModified()>fJar.lastModified()) {
                delFile(fJar);
                cpy(fSrc,fJar);
                final File dir=dirDesktopUtils();
                hardLinkOrCopy(fSrc,file(dir+"/strap.jar"));
                final String ext=isMacintosh() ? ".command" : isWindows() ? ".js" : ".sh";

                wrte(file(dir+"/textViewer"+ext), cmd(fJar, "-Xmx500M","-viewTextFiles"));
                wrte(file(dir+"/toFasta"+ext), cmd(fJar, "","-toFasta -tellOutputDir"));
                wrte(file(dir+"/toMultipleFasta"+ext), cmd(fJar, "","-toMultipleFasta -tellOutputDir"));
                wrte(file(dir+"/toClustal"+ext), cmd(fJar, "","-toClustal -tellOutputDir"));

                final StringBuffer sb=new StringBuffer(2222).append("<html><body>\n");
                if (becauseOfCB) sb.append("This directory has been created on the Desktop by STRAP because the respective check-box in the is activated.\n\n");
                sb.append(getHlp(DesktopUtils.class));

                wrte(file(dir+"/README.html"),sb.append("\n</body></html>"));
                if (!becauseOfCB) {
                    final String success=sze(fJar)>0 ?
                        "The utilities were successfully installed in <br>"+dir :
                        "Failed to install the utilities";
                    ChMsg.msgDialog(0L, success);
                    viewFile(dir, 0);
                }
            }
        }
    }
    private static String cmd(File fJar, String javaOptions, String strapOptions) {
        return isWindows() ?
            "var objShell = WScript.CreateObject('WScript.Shell')\n"+
            "var argv='';\n"+
            "for(i=WScript.Arguments.length; --i>=0;)  argv+=WScript.Arguments.Item(i).replace(/ /g, '%20')+' ';\n"+
            "objShell.Run('cmd /C java "+javaOptions+" -jar  \""+fJar.toString().replace('\\','/')+"\" "+strapOptions+" '+argv, 0)\n"
            : "#!/bin/sh\njava "+javaOptions+" -jar ${0%/*}/strap.jar "+strapOptions+" $*";
    }
    public static String OPTION_INSTALL="-installDesktopUtils";
    public static boolean main(String...argv) {
        final int iInstall=idxOfOption(OPTION_INSTALL,argv);
        final int iView=idxOfOption("-viewTextFiles",argv);
        final int iPubmed=idxOfOption("-pubmed",argv);

        if (iInstall>=0 || iView>=0 || iPubmed>=0) {
            final File dir=file("~/@/tempDesktopUtils");
            mkdrsErr(dir);
            setWorkingDir(dir);
        }
        if (iInstall>=0) {
            if (ChMsg.yesNo("Installing STRAP utility folder on ~/Desktop?")) install(false);
        }
        if (iView>=0) {
            int count=0;
            for(int i=iView+1;i<argv.length; i++) {
                final File f=file(argv[i]);
                if (sze(f)>0) {
                    new AsciiView(f).showInFrame(argv[i],ChFrame.STAGGER|ChFrame.DISPOSE_CHILDREN|DISPOSE_ON_CtrlW);
                    count++;
                }
            }
            ChFrame.setGlobalOptions(ChFrame.SYSTEM_EXIT_WHEN_ALL_CLOSED);
            if (count==0)  new DesktopUtils(UTIL_TEXT_VIEWER);
            return true;
        }
        if (iPubmed>=0) {
            Insecure.setFileModificationControl(false);
            final String s=delPfx("-pubmed=",argv[iPubmed]);
            openPmidManager(nxt(DIGT,s)>=0?s:null, ChFrame.SHUT_DOWN_ON_CLOSE);
            return true;
        }
        return false;
    }
    private static ChFrame fPmid;
    public static void openPmidManager(String txt, long options) {

        if (fPmid==null) {
            final String msg="Type a list of Pubmed or Uniprot references. \n"+
                "Move the mouse over these references.\n"+
                "Abstracts appear below.\n"+
                "Also try context-menu\n",
                example="\nPMID:20004206 PMID:16469097 PMID:11301311 PMID:16322575\n UNIPROT:Q9H936\n";
            final ChTextArea taPMID=new ChTextArea(msg+(sze(txt)>0? rplcToStrg(" PMID","\nPMID",txt):example)+"\n"), taAbstract=new ChTextArea("");
            final ChTextComponents tPMID=taPMID.tools(), tAbstract=taAbstract.tools();
            taPMID.tools().setReplaceWhenTextChanges(new ChTextReplacement("PMID: ","PMID"));
            if (sze(txt)==0) tPMID.saveInFile("DesktopUtils_PMID_LIST");
            tAbstract.enableUndo(true).underlineRefs(UNDERLINE_REFS);
            ChTextComponents.setTextAreaForAbstract(tAbstract);
            tPMID.enableUndo(true).underlineRefs(UNDERLINE_REFS|UNDERLINE_REFS_PMID_DOWNLOAD);
            pcp(KEY_IF_EMPTY,msg,taPMID);
            pcp(KEY_IF_EMPTY,"The Abstract of the PMID\nunder the mouse\nwill be shown here",taAbstract);
            Dimension d=dim(1,150);
            final Object
                pNorth=hB("Manage Pubmed abstracts","#",smallHelpBut(ChPubmed.class)),
                pMain=newCNSEW(new JSplitPane(JSplitPane.VERTICAL_SPLIT,scrllpn(0L,taPMID,d),scrllpn(0L,taAbstract,null)),pNorth, null,null,null);
            fPmid=new ChFrame("Pubmed").ad(pMain).size(444,444);
        }
        fPmid.shw(options);
    }
    public DesktopUtils(int id) {
        final Window f=new Window(new ChFrame(""));
        if (id==UTIL_TEXT_VIEWER) {
            final JComponent
                bClose=new ChButton("EXIT", ChButton.NO_FILL).t(null).ps(dim(16,1)).li(this),
                bHelp=smallHelpBut(DesktopUtils.class).ps(dim(20,1)),
                pan=newCNSEW("Drag text file here   ",C(0xFF00),null,bClose,bHelp);
            f.add(pan);

            ChRenderer.setSmallButtonXc( 0, bClose);
            for(JComponent c : new JComponent[]{pan,bClose,bHelp}) {
                new DropTarget(c,this);
                setDragMovesWindow(false, c);
            }
        }
        packW(f);
        f.show();
        ChFrame.setAOT(true,f);
    }
    public void actionPerformed(java.awt.event.ActionEvent ev) {
        final String cmd=ev.getActionCommand();
        if (cmd=="EXIT" && ChMsg.yesNo("Really exit text viewers?")) System.exit(0);
    }

    private java.awt.datatransfer.Transferable transferable;
    private File[] files;
    public void dropActionChanged(java.awt.dnd.DropTargetDragEvent event) {  }
    public void dragEnter(java.awt.dnd.DropTargetDragEvent ev) { ev.acceptDrag(java.awt.dnd.DnDConstants.ACTION_COPY);}

    public void dragOver(java.awt.dnd.DropTargetDragEvent event) {
        final java.awt.datatransfer.Transferable t=event.getTransferable();
        if (t!=transferable) {
            transferable=t;
            files=ChTransferable.getFiles(t,0);
        }
    }
    public void drop(java.awt.dnd.DropTargetDropEvent event) {
        putln("\nDrop");
        event.acceptDrop(java.awt.dnd.DnDConstants.ACTION_COPY);
        final java.awt.datatransfer.Transferable t=event.getTransferable();
        final File ff[]=sze(files)>0 ? files : ChTransferable.getFiles(t,0);
        for(int i=sze(ff); --i>=0;) {
            final File f=ff[i];
            if (sze(f)>0) {
                new AsciiView(f).showInFrame(f.getName(),ChFrame.STAGGER|ChFrame.DISPOSE_CHILDREN|DISPOSE_ON_CtrlW);
            }
        }
    }
    public void dragExit(java.awt.dnd.DropTargetEvent ev) { }

}