package charite.christo;
import charite.christo.interfaces.*;
import java.util.regex.Pattern;
import javax.swing.*;
import java.awt.*;
import java.io.File;
import java.util.*;
import java.util.List;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP
   Items in graphical list components can be selected using  string pattern.
   See WIKI:Regular_expression.
   @author Christoph Gille
*/
public class DialogStringMatch implements java.awt.event.ActionListener {
    public final static String
        PROVIDE_HAYSTACK="DSM$$PH", PROVIDE_STRINGS="DSM$$PS",
        ACTION_FINISHED="DSM$$F", ACTION_BEFORE="DSM$$B",
        METHODS[]={"Contains string","Starts with","Ends with","Regul expr"};
    public final static int NEVER_FILE_CONTENT=1, WHITE_BG=1<<1;
    private JComponent butSelectAnd, butSelectOr;
    private Object haystacks;
    private ChFrame frame;
    private UniqueList<DialogStringMatch> vRows=new UniqueList(DialogStringMatch.class);
    private JLabel labCount;
    private JComponent box, panFile;
    private ChTextField tfFileSize;
    private int options;
    private Component listComponent;
    private Collection vMatching;
    private String fileName;
    private boolean isAnd;
    public DialogStringMatch(long options, Component c, String fileName) {
        this.listComponent=c;
        this.options=(int)options;
        this.fileName=fileName;
    }
    public DialogStringMatch(long options, Object haystacks, Collection vMatching, String fileName) {
        this.haystacks=haystacks;
        this.vMatching=vMatching;
        this.options=(int)options;
        this.fileName=fileName;
    }
    public ChButton button(String txt0) {
        final String txt=txt0!=null ? txt0 : "Select ...";
        return new ChButton("NEW").li(this).t(txt).tt("Select items that match a string pattern.")
        .cp(DialogStringMatch.class,this);
    }
    private JComponent panel;
    private JComponent getParentPanel() {
        if (panel==null) {
            labCount=new JLabel();
            final Color BG=C((options&WHITE_BG)!=0 ? 0XffFFff : DEFAULT_BACKGROUND);
            final JComponent msg=new ScrollLabel("With this dialog you can select items  using string patterns");
            msg.setBackground(BG);
            JComponent pNorth=newCNSEW(msg,null,BG,smallSourceBut(DialogStringMatch.class),smallHelpBut(DialogStringMatch.class));
            this.butSelectAnd=new ChButton("Logical \"AND\"").li(this);
            this.butSelectOr=new ChButton("Logical \"OR\"").li(this).tt("combine all searches by boolean \"OR\"");

            if (0==(options&NEVER_FILE_CONTENT)) {
                this.tfFileSize=new ChTextField("10").ct(Integer.class).cols(9,true,false);
                this.panFile=hB("   Limit search within files to ",tfFileSize,"kByte");
                pcp(KOPT_HIDE_IF_DISABLED,"",this.panFile);
            }
            this.box=vB(pNorth,new DialogStringMatch(this).panel,BG);
            panel=newCNSEW(box,gcp(KEY_NORTH_PANEL,this),pnl(butSelectAnd,butSelectOr,labCount, panFile),BG,null);
            enableDisable(true);
        }
        return panel;
    }
    /* <<< Parent Instance <<< */
    /* ---------------------------------------- */
    private void enableDisable(boolean mayAddRow) {
        int countTxt=0;
        boolean isFile=false, emptyText=false;
        for(DialogStringMatch  m : vRows.asArray()) {
            final boolean b= sze(m.tf.getText())>0;
            if (b) countTxt++;
            setEnabld(b, m.butSelect);
            isFile|=m.cbFile.s();
            emptyText=emptyText || 0==sze(m.tf.getText());
        }
        setEnabld(countTxt>1, butSelectOr);
        setEnabld(countTxt>1, butSelectAnd);
        setEnabld(isFile, tfFileSize);
        setEnabld(isFile, panFile);
        if (mayAddRow && !emptyText) {
            final DialogStringMatch row=new DialogStringMatch(this);
            box.add(row.panel);
            if (frame!=null) {
                Dimension d=panel.getPreferredSize();
                frame.size(d.width+32,d.height+32);
            }
        }
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> selectElements >>> */
    private static int selectJListElements(Component list_or_tee, DialogStringMatch isEnabled) {
        int count=0;
        if (list_or_tee instanceof ChJList) {
            final ChJList l=(ChJList)list_or_tee;
            final ListModel m=l.getModel();
            final int N=m.getSize();
            final int iiOld[]= l.getSelectedIndices();
            final int ii[]=chSze(iiOld, N+iiOld.length);
            for(int j=0;j<N;j++) {
                ii[j+iiOld.length]=-1;
                if (isEnabled.isEnabled(m.getElementAt(j))) {
                    ii[j+iiOld.length]=j;
                    count++;
                }
            }
            l.setSelII(ii);
        } else if (list_or_tee instanceof JTree) {
            final JTree t=(JTree)list_or_tee;
            final Set set=new HashSet();
            final int N=t.getRowCount();
            for(int iRow=0;iRow<N;iRow++) {
                final javax.swing.tree.TreePath tp=t.getPathForRow(iRow);
                final Object o=tp.getLastPathComponent();
                if (o!=null && isEnabled.isEnabled(o)) set.add(tp);
            }
            count=sze(set);
            t.setSelectionPaths(toArry(set,javax.swing.tree.TreePath.class));
        } else if (list_or_tee instanceof JTable) {
            final JTable t=(JTable)list_or_tee;
            final int COLS=t.getColumnCount();
            for(int row=t.getRowCount(); --row>=0;) {
                for(int col=COLS; --col>=0;) {
                    final Object o=t.getValueAt(row, col);
                    if (o!=null && isEnabled.isEnabled(o)) {
                        t.addRowSelectionInterval(row, row);
                        count++;
                        break;
                    }
                }
            }
        }
        return count;
    }
    private static void recursiveSelect(Object haystacks, DialogStringMatch isEnabled, Collection matching) {
        if (haystacks==null) return;
        if (haystacks instanceof Object[]) {
            for(Object h : ((Object[])haystacks)) recursiveSelect(h,isEnabled, matching);
        } else {
            if (haystacks instanceof List) {
                final List v=(List)haystacks;
                for(int i=sze(v); --i>=0;) recursiveSelect(v.get(i),isEnabled,matching);
            }
            if (haystacks instanceof ChRunnable) {
                recursiveSelect(((ChRunnable)haystacks).run(PROVIDE_HAYSTACK,null),isEnabled,matching);
            }
            if(isEnabled.isEnabled(haystacks)) matching.add(haystacks);
        }
    }
    private void selectUnselect(DialogStringMatch isEnabled) {
        handleActEvt(this, ACTION_BEFORE,0);
        int count=0;
        if (haystacks!=null) {
            vMatching.clear();
            recursiveSelect(haystacks,isEnabled, vMatching);
            count=sze(vMatching);
        } else if (listComponent instanceof JComponent) {
            count=selectJListElements(listComponent, isEnabled);
        }
        ChDelay.setText(labCount,count+" items matched",1);
        ChDelay.revalidate(labCount,333);
        handleActEvt(this, ACTION_FINISHED,0);
    }
    /* <<< selectElements <<< */
    /* ---------------------------------------- */
    /* >>> Here starts the former inner class >>> */
    private ChTextField tf;
    private ChButton cbNot,cbLC, cbWr, cbWl, cbReg, cbFile, butSelect;
    private DialogStringMatch parent;
    //    private Set setEnabled, setNotEnabled;

    private DialogStringMatch(DialogStringMatch parent) {
        this.parent=parent;

        //setEnabled=new HashSet(999);    setNotEnabled=new HashSet(999);
        parent.vRows.add(this);
        tf=new ChTextField().li(this).tt("Enter a search String");
        tf.requestFocus();
        if (sze(parent.fileName)>0) tf.saveInFile("DialogStringMatch_"+parent.fileName+"_"+sze(parent.vRows));
        tf.tools().enableWordCompletion(parent.haystacks);

        final ChTextComponents tools=tf.tools();
        pcp(KEY_IF_EMPTY,"Enter search string",tf);
        tools.highlightOccurrence(" ", null,null,HIGHLIGHT_UPDATE_IF_TEXT_CHANGES,C(DEFAULT_BACKGROUND));
        tools.highlightOccurrence("\t",null,null,HIGHLIGHT_UPDATE_IF_TEXT_CHANGES,C(0x444444));        
        final String ttW="The character before/after is not a word character";
        cbWl=toggl("Word Left").tt(ttW);
        cbWr=toggl("Word Right").tt(ttW);
        cbReg=toggl("Regular expression");
        cbNot=toggl("not").tt("Selects entries <br>that do not match");
        cbLC=toggl("Ignore case").tt("Upper and lower case does not matter").li(this).s(true);
        cbFile=toggl("File content").li(this).tt("Tests the file content for string match.");
        butSelect=new ChButton(parent.haystacks!=null ? "Find<br>matching" : "Select<br>matching").li(this);
        Container pSouth=pnl(cbNot.cb()," ",cbReg.cb(),cbLC.cb(), cbWl.cb(), cbWr.cb(),
                             (parent.options&NEVER_FILE_CONTENT)==0 ? cbFile.cb():null);
        Container pCenter=pnl(new GridLayout(2,1),tf,pSouth);
        this.panel=newCNSEW(pCenter,null,null,null,pnl(butSelect));
        closeOnCtrlW(true,panel,frame);
        parent.enableDisable(false);
    }
    public boolean isEnabled(Object object0) {
        if (this.parent==null) { /* Is the parent */
            boolean success=false;
            for(DialogStringMatch m : this.vRows.asArray()) {
                if (m==null || sze(m.tf.getText())==0) continue;
                final boolean b=!m.isEnabled(object0);
                if (!b && !isAnd)  return true;
                if (b && isAnd) return false;
                success|=!b;
            }
            return success;
        } else  { /* Single fields */
            //if (setEnabled.contains(object0) || setNotEnabled.contains(object0)) return true;
            final String txt=tf.getText();
            Pattern pattern=null;
                if (cbReg.s()) {
                    try {
                        pattern=Pattern.compile(txt, (cbLC.s()?Pattern.CASE_INSENSITIVE : 0)|Pattern.MULTILINE|Pattern.DOTALL);
                    } catch(Exception e) { errorAC(toStrg(e)); }
                }
            final long strstrOptions=
                (cbLC.s() ? STRSTR_IC : 0) |
                (cbWl.s() ? STRSTR_w_L : 0) |
                (cbWr.s() ? STRSTR_w_R : 0);
            final int fileKB=!isSelctd(cbFile) ? 0 : maxi(0,atoi(parent.tfFileSize));
            final boolean result= recursiveMatches(txt, object0,  pattern, strstrOptions, fileKB);
            final boolean enabled=cbNot.s() ? !result : result;
            //(enabled ? setEnabled : setNotEnabled).add(object0);
            return enabled;
        }
    }

    private static boolean recursiveMatches(String needle, Object o, Pattern pattern, long strOpt, int fileKB) {
        if (o==null) return false;
        if (o instanceof Object[] || o instanceof List) {
            for(int i=sze(o); --i>=0;) {
                final Object o2=get(i,o);
                if (o2!=null && recursiveMatches(needle, o2, pattern, strOpt, fileKB)) return true;
            }
        }
        if (o instanceof ChRunnable && recursiveMatches(needle, ((ChRunnable)o).run(PROVIDE_STRINGS,null),pattern, strOpt, fileKB)) return true;
        CharSequence txt=null;
        if (o instanceof File) {
            txt= fileKB>0 ?
                readBytes((File)o, 1024*fileKB,  baFromPool("Protein",1024*10)) :
                o.toString();
        }
        if (txt==null && o instanceof CharSequence) txt=(CharSequence)o;
        if (txt==null && o instanceof HasName) txt=((HasName)o).getName();
        if (txt!=null) {
            if (pattern!=null) {
                final java.util.regex.Matcher m=pattern.matcher(txt);
                while(m!=null && m.find()) {
                    final char prev=chrAt(m.start()-1, txt), nxt=chrAt(m.end(),txt);
                    if (
                        (0==(strOpt&STRSTR_w_R) || !is(LETTR_DIGT_US,nxt)) &&
                        (0==(strOpt&STRSTR_w_L) || !is(LETTR_DIGT_US,prev))
                        ) return true;
                }
                return false;
            }
            return strstr(strOpt,needle,txt)>=0;
        }
        return false;
    }
    /* <<< Match <<< */
    /* ---------------------------------------- */
    /* >>> AWTEvent >>> */
    public void actionPerformed(java.awt.event.ActionEvent ev) {
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        final JComponent box=(parent!=null ? parent : this).box;
        if (cmd=="NEW") {
            if (frame==null) {
                String title=gcps(KEY_TITLE,this);
                frame=new ChFrame(title==null ? "Select by text pattern" : title).ad(getParentPanel());
                final ChTextField tf=child(box,ChTextField.class);
                if (tf!=null) tf.requestFocus();
            }
            frame.shw(ChFrame.TO_FRONT|ChFrame.AT_CLICK|ChFrame.PACK|ChFrame.SIZE_REGARD_RESOLUTION);
        }
        if (parent!=null) {
            if (q==butSelect || q==tf && cmd==ACTION_ENTER) { // Matcher
                parent.labCount.setText("Computing ...");
                parent.selectUnselect(DialogStringMatch.this);
            }
            if (q==tf || q==cbFile) {
                parent.enableDisable(q==tf);
            }
        } else {
            if (q==butSelectAnd || q==butSelectOr){
                this.isAnd= q==butSelectAnd;
                selectUnselect(DialogStringMatch.this);
                //for(DialogStringMatch m : vRows.asArray()) { m.setEnabled.clear();m.setNotEnabled.clear(); }
            }
        }
    }
}
