package charite.christo;
import charite.christo.interfaces.*;
import charite.christo.hotswap.*;
import javax.swing.*;
import static charite.christo.ChUtils.*;
/**
   allows class hotswapping
   @author Christoph Gille
*/
public class HotswapUtil {
    private HotswapUtil(){}
    public static Object instance(final Object o0) { return instance(o0, (Class)null,false); }
    public static <T extends Object>T instance(final Object o0, final Class<T> clazz) { return instance(o0,clazz,false); }
    public static <T extends Object>T instance(final Object o0, final Class<T> clazz, boolean error) {
        if (o0==null) return null;
        final ChCombo combo=o0 instanceof ChCombo ? (ChCombo)o0 : null;
        Object o=o0, inst=null;
        if (combo!=null) o=combo.getSelectedItem();
        else if (o instanceof JList) o=((JList)o).getSelectedValue();
        if (o instanceof String) o=getClass(o);
        try {
            if (o instanceof ProxyClass) inst=((ProxyClass)o).newInstance();
        } catch(Throwable e){ stckTrc(e);}
        if (inst==null && o instanceof Class) {
            try {
                inst=  ((Class)o).newInstance();
            }catch(Exception e){ if (error) putln(" caughtIn HotswapUtil ",e); }
        }
        if (inst!=null) {
            if (combo!=null) setSharedParas(combo.instanceOfSelectedClass(null), inst);
            if (clazz!=null && !isAssignableFrom(clazz,inst)) inst=null;
        }
        if (inst==null && error) error("Could not instantiate "+o0+" "+clazz);
        return (T)inst;
    }

    public static boolean isAssignableFrom(Class interf,Object o) {
        Class c=HotswapUtil.getClass(o);
        return interf!=null && c!=null && interf.isAssignableFrom(c);
    }
    static boolean isInstanceOf(Class interf,Object o) {
        if (o==null || interf==null) return false;
        if (o instanceof HotswapProxy) {
            ProxyClass pc=((HotswapProxy)o).hotswap_getProxyClass();
            return pc!=null && idxOf(interf,pc.getProxyInterfaces())>=0;
        }
        return  interf.isAssignableFrom(o.getClass());
    }

    public static String classNameWithSlash(Object o) { String s=className(o); return s!=null ? s.replace('.','/') :null;}
    public static String shortClassName(Object o) {
        final String cn=className(o);
        return cn!=null ? cn.substring(cn.lastIndexOf('.')+1) : "null";
    }
    public static String className(Object o) {
        if (o instanceof String) return (String)o;
        final Class c=getClass(o);
        if (c==null) return null;
        final String name=c.getName();
        final int iDollar=name.indexOf('$');
        return iDollar>0 ? name.substring(0,iDollar) : name;
    }
    public static Class getClass(Object o) {
        if (o==null) return null;
        if (o instanceof JComboBox) o=((JComboBox)o).getSelectedItem();
        if (o instanceof JList) o=((JList)o).getSelectedValue();
        if (o==null) return null;
        final ProxyClass pc=o instanceof ProxyClass ? (ProxyClass)o : o instanceof HotswapProxy ?  ((HotswapProxy)o).hotswap_getProxyClass() : null;
        if (pc!=null) return pc.getClassInstance();

        if (o instanceof String ) {
            try {
                return Class.forName(o.toString());
            } catch(Throwable ex){}
        }
        else if (o instanceof Class) return (Class)o;
        return o.getClass();
    }
}
