package charite.christo;
import charite.christo.interfaces.*;
import java.awt.*;
import java.io.File;
import java.net.URL;
import java.util.*;
import javax.swing.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
/**HELP
<h2>Download Dialog</h2>

The download dialog appears when files are loaded from the Internet.

<table style="caption-side: bottom">
<CAPTION><b>Figure:</b> Download dialog. It is save to press "Download without asking".
</CAPTION>
<tr><td><i>JCOMPONENT:InteractiveDownload#docuAskDownload()</i></td></tr>
</table>

<h2>Proxy Server</h2>
<i>INCLUDE_DOC:Web</i>

<h2>Privacy</h2>

If you work with sensitive data you might want to activate the toggle
<i>CB:ChUtils#buttn(TOG_NASK_BEFORE_UPLOADING_DATA)</i> to be asked for each data transfer from your computer to the web.

@author Christoph Gille

*/
public class InteractiveDownload implements ChRunnable {
    private final static String
        SUFFICES[]={"","TMP",".gzTMP",".ZTMP",".zipTMP",".bz2TMP"},
        DL_ADL_M_C[]={"Download","More info","Cancel"},
        NEEDS_DL="The program needs to download some file[s] in order to proceed";
    private final static Object SYNC_IaW=new Object(), SYNC=new Object();
    public Object run(String id,Object arg) {
        final Object argv[]=arg instanceof Object[] ? (Object[])arg : null;
        if (id=="downloadFilesEDT") {
            final URL[] uu=(URL[])argv[0];
            final File[] ff=(File[])argv[1];
            final File dir=(File)argv[2];
            final boolean[] ret=(boolean[])argv[3];
            ret[0]=downloadFilesEDT(uu, ff, dir);
        } else if (id=="WATCH") {
            final StringBuffer sbFileSize=new StringBuffer();
            final File[] FILES=(File[])((Object[])arg)[0];
            final ChTextArea taDownloaded=(ChTextArea)((Object[])arg)[1];
            final boolean[] isEnabled=(boolean[])((Object[])arg)[2];
            while(isEnabled[0]) {
                sleep(222);
                sbFileSize.setLength(0);
                for(File f0:FILES) {
                    if (f0==null) continue;
                    boolean exists=false;
                    for(String suffix : SUFFICES) {
                        final File f=file(f0+suffix);
                        if (f.exists()) {
                            sbFileSize.append("  ").append(f.getName());
                            format10(sze(f),11,true,sbFileSize).append("  \n");
                            exists=true;
                        }
                    }

                    if (!exists) sbFileSize.append(' ').append(f0.getName()).append(": not yet\n");
                }
                taDownloaded.t(sbFileSize);
            }
            sbFileSize.setLength(0);
        } else assertTrue(false);
        return null;
    }

    public static boolean downloadFiles(URL urls[],File files[]) {return dlFiles(urls,files,null);}
    public static boolean downloadFilesAndUnzip(URL urls[], File directory) { return dlFiles(urls,null,directory); }
    private static boolean dlFiles(final URL urls[],final File files[], final File directory) {
        if (isGraphicsAvailable() && !isEDT()) {
            final boolean[] retValue={false};
             synchronized(SYNC_IaW) {
                 ChDelay.invokeAndWait(new InteractiveDownload(),"downloadFilesEDT",new Object[]{urls,files,directory, retValue});
             }
            return retValue[0];
        } else return downloadFilesEDT(urls,files,directory);
    }
    private static boolean downloadFilesEDT(final URL urls[], final File files[], final File directory) {
        final boolean isGui=isGraphicsAvailable();
        final File[] vFILE=new File[urls.length];
        final URL[]
            vURL=new URL[urls.length],
            srcParents=new URL[urls.length];
        final String[] destParents=new String[urls.length];
        final ChTextArea taSrc=isGui?new ChTextArea("").ro():null, taDest=isGui?new ChTextArea("").ro():null;
        boolean isPlugin=false;
        int count=0;
        boolean useEdtFTP=false;

        for(int i=0; i<urls.length; i++){
            final URL url=urls[i];
            if (url==null) continue;
            final String
                sUrl=url.toString(),
                fName=withoutLastComponent(sUrl.substring(sUrl.lastIndexOf('/')+1),'?'),
                fNameNoCompress;
            isPlugin|= sUrl.endsWith("?plugins");
            useEdtFTP=useEdtFTP || useEdtFTP(sUrl);
            if (!isPlugin &&  fName.endsWith(".pack.gz")) {
                if (sze(file(directory, delSfx(".pack.gz",fName)))>0) continue;
                fNameNoCompress=fName;
            } else fNameNoCompress=delSfx(".Z",delSfx(".bz2",delSfx(".gz",fName)));

            final File file=
                files!=null ? files[i] :
                directory!=null ? file(directory, fNameNoCompress) : null;
            if (file==null || (!isPlugin && sze(file)>0)) continue;
            destParents[count]=toStrg(mkParentDrs(file));
            vFILE[count]=file;
            vURL[count]=url;
            if (isGui) {
                taSrc.appnd(url,"\n");
                taDest.appnd(filepathUnixStyle(file),"\n");
            }
            srcParents[count]=url(withoutLastComponent(delSfx("/",url),'/'));
            count++;
        }
        if (count==0) return true;
        if (useEdtFTP) new AnonymousFTPPROXY().proxyObject();
        if (directory==dirPlugins()) {
            final ChTextArea sb=new ChTextArea("Going to download the following plugin data into ").appnd(directory,"\n\n");
            for(URL u:vURL) { sb.appnd("  ",u,"\n"); currVers(toStrg(u)); }
            sb.appnd("\n",C(0xFFaaAA) ,"Warning:",C(0xFFffFF)," Plugins can potentially contain WIKI:Malware.\n\n"+
                      "Do not download, unless you fully trust the plugin.\n",
                      "You may want to protect yourself in a sandbox environment (See "+URL_STRAP+"security.html).\n\n"+
                      "Trusted URLs are highlighted in green\n");
            for(String needle:TRUSTED_PLUGINS)  sb.tools().highlightOccurrence(needle, null,null, 0, C(0xaaffaa));
            if (ChMsg.option(SYNC, 0L, sb,new String[]{"I trust the plugin","I do not want to load the plugin"})!=0) return false;
        }

        final StringBuffer sb=new StringBuffer(333).append("WIKI:Web_proxy\n");
        for(int i=2; --i>=0;) {
            Object last=null;
            sb.append("Download ").append(i==0 ? "to:\n" : "from:\n");
            for(Object p: i==0 ? destParents : srcParents)
                if (!eqNz(p,last)) sb.append(' ').append(last=p).append("  \n");
        }
        final ChButton togAsk=buttn(TOG_NASK_BEFORE_DOWNLOAD);
        final JComponent
            pSouth=isGui?monospc(sb) : null,
            pCenter=isGui?vBhB( "Loading the following files:", hBL(taSrc," ==> ",taDest),null) : null,
            msg1=isGui ? newCNSEW(scrllpn(0L,pCenter, dim(666,222)), togAsk.cb(),pSouth,null,null) : null;
        final String DC[]={"Download","Cancel"};
        if (isPlugin && ChMsg.option(SYNC, 0L, msg1,DC)!=0) return false;

        if (isGui && !togAsk.s()) {
            addTab(null,false,ChFrame.CENTER);
            final int option=ChMsg.option(SYNC, 0L, vBhB(NEEDS_DL,togAsk.cb()) ,DL_ADL_M_C);
            if (option==2 || option==1 && ChMsg.option(SYNC, 0L, msg1,DC)!=0) return false;
        }
        final ThreadDownload thread=new ThreadDownload(vFILE,vURL);
        if (isGui) {
            thread.start();
            final ChTextArea taDownloaded=new ChTextArea("");
            final String wget=isMacintosh() ? "curl -O  " : "wget -N ";
            thread.butLog.a("# If download fails you can try ").a(wget).a( " \n");
            for(int i=0; i<vURL.length; i++) if (vURL[i]!=null) thread.butLog.a(wget).a(vURL[i]).a(" -O '").a(vFILE[i]).a("' \n");
            thread.MSG_DOWNLOAD=vBhB(
                                     thread.pNorth,
                                     " ",
                                     pSouth,
                                     " ",
                                     monospc("Downloaded bytes:"),
                                     hB(taDownloaded,"#",  thread.butLog.t("Log"))
                                     );

            taDownloaded.setPreferredSize(dim(1,count*(EX+4)));
            final boolean isRunning[]={true};
            ChDelay.newThread(new InteractiveDownload(),"WATCH",new Object[]{vFILE,taDownloaded.appnd("\n"),isRunning}).start();
            addTab(null,false, ChFrame.CENTER);
            ChMsg.option(SYNC, 0L, thread.MSG_DOWNLOAD,new String[]{"Abort download"});
            isRunning[0]=false;
        } else thread.run();
        boolean success=thread.finished && thread.success;
        thread.interrupted=true;
        return success;
    }

    private static class ThreadDownload extends Thread implements ChRunnable,IsEnabled {
        @Override public void interrupt() { success=false;  super.interrupt();}
        boolean success=true;
        final LogFileButton butLog=isGraphicsAvailable()?new LogFileButton("Log-messages"):null;
        final URL[] urls;
        final File files[];
        boolean interrupted,finished;
        JComponent MSG_DOWNLOAD;
        final JComponent pNorth=isGraphicsAvailable()?pnl():null;
        ThreadDownload(File[] vF, URL[] vU) {
            urls=vU; files=vF;
            if (pNorth!=null) pNorth.add(new JLabel("loading ... "));
        }
        public boolean isEnabled(Object o){return !interrupted;}
        @Override public void run() {
            for(int i=0; i<urls.length;i++) {
                final File f=files[i];
                final URL url=urls[i];
                if (f==null || url==null || sze(f)>0) continue;
                delFile(f);
                final String sUrl=url.toString();
                if (sze(f)==0) downloadAndDecompress(sUrl,f,butLog,((IsEnabled)this));
                if (sze(f)==0) {
                    if (butLog!=null) butLog.a("Error:  could not load ").a(f).a("\n");
                    success=false;
                } else {
                    if (butLog!=null) butLog.a("Download: success\n");
                    if (f.getName().toLowerCase().endsWith(".exe")) {
                        if (butLog!=null) butLog.a("Make executable\n");
                        makeExecutable(f);
                    }
                }
                if (butLog!=null) butLog.a("\n");
            }
            finished=true;
            ChDelay.invokeAndWait(this,"AFTER_DOWNLOAD",null);
            if (butLog!=null) butLog.closeStream();
        }
        public final Object run(String id, Object arg) {
            if (id=="AFTER_DOWNLOAD") {
                if (pNorth!=null) {pNorth.removeAll();
                    if (success) pNorth.add(pnl(C(0x00b200),"download succeeded"));
                    else pNorth.add((pnl(C(0xFF0000),"download failed",smallHelpBut(InteractiveDownload.class))));
                    final Window w=parentWndw(MSG_DOWNLOAD);
                    if (w instanceof JDialog)  w.dispose();
                    addTab(MSG_DOWNLOAD,success, ChFrame.CENTER);
                }
            }
            return null;
        }
    }
    public static void addTab(Container c,boolean success, long options) {
        if (frameDownloads==null)  frameDownloads=new ChFrame("Downloads: "+ChMsg.titleReopenCtrl_I()).ad(tabbedPane=new ChTabPane(ChTabPane.LEFT)).size(555,222);
        if (c==null)  { frameDownloads.setVisible(false); return; }
        final JScrollPane sp=scrllpn(0L,c);
        TabItemTipIcon.set(String.valueOf(++iDownload),null,null,null,sp);
        TabItemTipIcon.set(null,null,null,success ? IC_HAPPY:IC_BUG,sp);
        tabbedPane.addTab(sp);
        if (UserProfile.isWebMode()) frameDownloads.setExtendedState(JFrame.ICONIFIED);
        else frameDownloads.shw(options);
    }
    private static ChTabPane tabbedPane;
    static ChFrame frameDownloads;
    private static int iDownload;

    public static File downloadFileIfNewer(Object urls) {
        final File ff[]=urls==null ? null : downloadFilesIfNewer(new Object[]{urls});
        return get(0,ff,File.class);
    }

    public static File[] downloadFilesIfNewer(Object urls[]) {
        final URL uu[]=new URL[urls.length];
        final File ff[]=new File[urls.length];
        int count=0;
        for(int i=urls.length; --i>=0;) {

            if (urls[i]==null  || urls[i] instanceof File || urls[i] instanceof String && !looks(LIKE_EXTURL,urls[i])) continue;
            final URL u=url(urls[i]);
            if (u==null) continue;
            ff[i]=file(u);
            if (ff[i]==null) continue;
            if (isUpToDate(ff[i],u)==0) {
                if (sze(ff[i])>0) {
                    delFile(ff[i]);
                    putln("Going to deleted ",ff[i], sze(ff[i])>0?ANSI_FG_GREEN+" success"+ANSI_RESET : " failed");
                }
                uu[i]=u;
                count++;
            }

        }
        if (count>0) {
            putln(ANSI_FG_GREEN+">>>>>>>>>>>>>>>>>>> "+ANSI_RESET);
            puts("InteractiveDownload: Going to Download ");  puts(uu);
            downloadFiles(uu,ff);
            putln(" "+ANSI_FG_GREEN+"DONE"+ANSI_RESET);
            putln(ANSI_FG_GREEN+"<<<<<<<<<<<<<<<<<<<<<< "+ANSI_RESET);
        }
        return ff;
    }
    public final static JComponent docuAskDownload() {
        return
            vBpnl(
                  NEEDS_DL,
                  pnl(new JButton(DL_ADL_M_C[0]),new JButton(DL_ADL_M_C[1]),new JButton(DL_ADL_M_C[2]))
                  );
    }
}

