package charite.christo;
import charite.christo.interfaces.*;
import java.awt.event.*;
import java.util.*;
import java.net.URL;
import java.io.*;
import javax.swing.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP
WIKI:Windows32 is the world's most widely applied operation system.

<br>
Windows differs to standard OS with respect to
<ul>
<li>Command line syntax</li>
<li>File separator character "\". On normal systems backslash is an WIKI:Escape_character and not a file separator character.</li>
<li>Path separator character ";"</li>
<li>File system:
<ul>
<li>File and folder names: Case insensitive, some names not possible because reserved for MS_DOS</li>
<li>File system: Language dependent directory tree</li>
<li>Missing hard links</li>
<li>Limitation of file path length to 260 characters. See WIKI:Path_(computing).</li>
<li>Unusual characters in file and directory names like space and Umlaute</li>
</ul>
</li>
</ul>


<br><br>

<b>Some important Windows tips for STRAP users</b><br>

<ul>

<li>If STRAP cannot download files because of  Web-proxy problems it might help to webstart STRAP with IE.</li>

<li>The editor WIKI:Notepad does not cope with file ends of some protein files. Use WIKI:Notepad_plus_plus or WIKI:Wordpad instead.</li>

<li>
Certain file-names are reserved by the system such as "aux.pdb", "con.swissprot", "prn.fasta", "icon" and "icons".
Their use leads to loss of data.
</li>

<li>
Some bioinformatics programs expect slash as a file separator
rather than the back-slash which is the native separator on MS-Windows.
</li>

</ul>

<br>

<br><br><b>Cygwin:</b>

To install Bioinformatics programs  STRAP uses the Gnu-compiler of WIKI:Cygwin if no other specified.

<br><br><b>The directory for program settings:</b>

WIKI:Dot_files are used for files that store settings of application programs.

Since there had been problems on some Windows versions  with file names starting with a dot, the dot is omitted.

Consequently the  settings directory is "%HOMEPATH%\StrapAlign\" rather than ~/.StrapAlign/".

But if %HOMEPATH%  contains white space then "C:\StrapAlign\" is used instead.

An alternative directory can be provided in the file %HOMEPATH%\.location_of_StrapAlign.ini.

@author Christoph Gille

But file:///C:/ is fine but failes file:/C:/...
Do not use quotes. rundll32.exe shell32.dll,Control_RunDLL "file:///C:/....html" does not work. Without quotes it works.

*/
public class Microsoft implements ActionListener,ChRunnable {
    public final static String CYGWIN_CORE="gcc-g++ gcc-g77 make python grep coreutils wget findutils ", CYGWIN_LATEX="tetex tetex-extra gv ";
    private Microsoft(){}
    private static Microsoft inst;
    private static Microsoft instance() {if (inst==null) inst=new Microsoft(); return inst;}
    public static String cmd() { return isSystProprty(IS_WINDOWS_95_FAMILY)  ? "command.com" : "cmd.exe";}

    public void actionPerformed(ActionEvent ev) {
        final Object q=ev.getSource(), roots=gcp("R",q);
        final String cmd=ev.getActionCommand(), packages=gcps("P", q);
        if (cmd=="CW") {
            final boolean log=isCtrl(ev) || setupFailed || cbCygwinUnattended!=null&&!cbCygwinUnattended.s();
            if (ChMsg.yesNo(ChMsg.AT_CLICK, "Going to install/update the cygwin packages<pre>"+packages+"</pre>This will take some time.")) {
                final Object  para=new Object[]{packages, longObjct(log ? ChExec.SHOW_STREAMS : 0L), roots};
                ChDelay.newThread(instance(),"INST_CW", para).start();
            }
        }
        if (cmd=="CWD") {
            new ChFrame("Install Cygwin").ad(panInstallCygwin(packages)).shw(ChFrame.AT_CLICK|ChFrame.PACK|ChFrame.ALWAYS_ON_TOP);
        }
    }

    public static ChButton newCygwinButton(String packages) {
        return new ChButton("CWD").t("Install cygwin packages ...").i(IC_CYGWIN).li(instance()).cp("P",packages);
    }
    private static ChButton cbCygwinUnattended;
    private static JComponent panInstallCygwin(String packages0) {
        final String packages=sze(packages0)>0 ? packages0 : CYGWIN_CORE;
        if (cbCygwinUnattended==null) cbCygwinUnattended=toggl("Run Cygwin setup.exe unattended and automatically ").s(true);
        Object roots=null;
        {
            final File dirCygwin=dirCygwin();
            roots=dirCygwin;
            putln("dirCygwin="+dirCygwin+"  "+dirCygwin.isDirectory());
            if (!dirCygwin.isDirectory()) {
                final Object ff[]=listExistingFileRoots();
                if (sze(ff)>0) {
                    final String rr[]=new String[ff.length];
                    for(int i=sze(ff); --i>=0;) rr[i]=delSfx("\\",ff[i])+"\\cygwin";
                    final ChCombo c=new ChCombo(rr);
                    roots=c;
                    for(String r:rr) {
                        if ((32|chrAt(0,toStrg(r)))=='c') c.s(r);
                    }
                }
            }
        }
        final JComponent
            panProblem=vB("#TB In case of problems with Cygwin installation",
                          " Unselect the following check-box and try installation again. ",
                          cbCygwinUnattended.cb()
                          ),
            b=new ChButton("CW").t(ChButton.GO).li(instance()).actionClose(CLOSE_PARENT_WINDOW)
            .cp("P",packages).cp("R",roots),
            pan=vBhB(
                     hBL("WIKI:Cygwin"," is required for Bioinformatics programs embedded in STRAP."),
                     " ",
                     "Installation of Cygwin requires 200 to 500 MBytes.",
                     hBL("Root directory: ", roots),
                     " ",
                     hBL("Install the Cygwin packages  ",packages, " ",b),
                     " ",
                     pnl(" ",panProblem)
                     );

        return pan;
    }

    private static boolean setupFailed;
    public Object run(String id,Object arg) {
        final Object argv[]=arg instanceof Object[] ? (Object[]) arg : null;
        if (id=="INST_CW") {
            final long time0=System.currentTimeMillis();
            final String packages=(sze(argv[0])>0 ? argv[0].toString() : CYGWIN_CORE).replaceAll("  *",",");
            final long options=atol(argv[1]);
            final File dirCygwin=file(argv[2]);
            final File f=file(dirBin()+"/cygwinSetup/setup.exe");
            if ((System.currentTimeMillis()-f.lastModified())/1000 > 60*60*24* 2) delFile(f);
            InteractiveDownload.downloadFiles(new URL[]{url("http://www.cygwin.com/setup.exe")},new File[]{f});
            final File fBash=file(dirCygwin()+"/bin/bash.exe");
            final java.net.InetSocketAddress proxy=proxyForUrl(url("http://www.cygwin.org/"));
            new ChExec(options)
                .setCommandLineV(cmd(),"/c",f,
                                 cbCygwinUnattended==null || cbCygwinUnattended.s()?"-q":null,
                                 "-R",dirCygwin,"-l",dirCygwin,
                                 proxy==null?null:"-p", proxy==null?null:proxy,
                                 "-d",
                                 "-P",packages
                                 ).run();
            if (sze(fBash)==0)  { setupFailed=true; error("Can not find <br>"+fBash); }
            inEDT(ChDelay.newThread(this,"DEL_CW_TMP",null));
            final File ff[]=isSelctd(buttn(TOG_CYGWIN_DEL_TMP_FILES)) ? dirCygwin().listFiles() : null;
            for(int i=sze(ff); --i>=0;) {
                if (ff[i].toString().indexOf("%3a%2f%2f")>0 && ff[i].lastModified()>time0)  deleteTree(ff[i]);
            }
        }
        return null;
    }
    public static String toCygwinPath(Object f) {
        final String p=toStrg(f);
        if (p==null) return null;
        return isWindows() ? "/cygdrive/"+p.charAt(0)+"/"+p.substring(2).replace('\\','/') : p;
    }

    public static String regQuery(String query) {
        File fTmp=createTempFile(".reg");
        final ChExec exec=new ChExec(0).setCommandLineV("c:\\winxp\\i386\\regedit.exe","/e",fTmp,"\"HKEY_CLASSES_ROOT\\Software\\DeLano Scientific\\PyMOL\\PYMOL_PATH\"");
        exec.run();
        new ChFrame("").ad(getCtrlPanel(exec)).shw(0);
        final ChExec exec2=new ChExec(ChExec.STDOUT).setCommandLineV("type",fTmp);
        new ChFrame("").ad(getCtrlPanel(exec2)).shw(0);
        final String stdout[]=splitStrg(getCtrlPanel(exec2),'\n');
        if (stdout==null) return null;
        new ChTextViewer(joinLines(stdout)).shw(0);
        for(String s:stdout) {
            if (s.startsWith("@=")) return s.substring(2);
        }
        return null;
    }

}
/*

which curl

lt-8 -> {
alt-5 -> [
alt-shift-7 -> \
alt-n -> ~

http://cygwin.com/setup-1.7.exe
http://alumnit.ca/~apenwarr/netselect/
http://sourceware.org/cygwin-apps/package-server.html
Or you could modify the attached Windows batch file (I've made it
a TXT file so it passes through most filters) and change the
mirror name to one that's closer.

You might have to make one batch file per country :-)

Ralph

*/
