package charite.christo;
import charite.christo.interfaces.*;
import java.io.*;

import static charite.christo.ChUtils.*;
/**HELP
   Proteins read from a PDB-file are selected.
   When a protein has more than one chain a new protein file is written for each chain.
   An underscore and the chain identifier letter is appended to the file name.
   @author Christoph Gille
*/
public  class PDB_separateChains extends AbstractInteractiveFileProcessor {
     {
        setTextNorth("The following list shows all PDB  files with more than one chain.<br>"+
                     "Select the files to be split into one-chain- PDB-files.");
        setErrorNonSelected("none of these files are PDB files with more than one chain");
    }
    private static boolean alsoIfOnlyOneChain=true;
    private final static int CHAIN_ID=21;
    @Override public String fileEntry(ByteArray byteArray,File f) {
        if (f==null || byteArray==null) return null;
        boolean chains[]=chainsInPDBfile(byteArray);
        if (!alsoIfOnlyOneChain && countTrue(chains)<2) return null;
        final StringBuffer sb=new StringBuffer().append(f).append(" ");
        final int last=lastTrue(chains);
        for(int i=0;i<sze(chains);i++)
            if (chains[i])
                sb.append((char)i).append(i==last-1 ? " and " : " ");
        return sb.toString();
    }

    @Override public File[] processFile(ByteArray byteArray, final File f, boolean overwrite) {
        final File[] vCreated=new File[128];
        int count=0;
        if (byteArray==null) {
            final int l=sze(f);
            if (l>0) byteArray=readAndDecompress(f,baFromPool("PDB_separateChains",l));
        }
        if (byteArray!=null) {
            final boolean chains[]=chainsInPDBfile(byteArray);
            for(char iChain=' '; chains!=null && iChain<='z';iChain++) {
                if (chains[iChain]) {
                    final File fNew=file(newName(f,(char)iChain));
                    if (sze(fNew)==0 || overwrite) {
                        delFile(fNew);
                        writePdbOnlyChain(fNew, iChain, byteArray);
                    }
                    if (sze(fNew)>0) vCreated[count++]=fNew;
                }
            }
        }
        return chSze(vCreated,count,File.class);
    }
public static boolean[] chainsInPDBfile(ByteArray byteArray) {
        final int ends[]=byteArray.getLineEnds();
        final byte[] T=byteArray.getBytes();
        final int length=byteArray.getEnd();
        boolean rr[]=null;
        for(int iL=0; iL<ends.length; iL++) {
            final int b= iL==0 ? 0 : ends[iL-1]+1;
            final int l=ends[iL]-b;
            if (b+CHAIN_ID>=length) break;
            final int c0=T[b], c1=T[b+1], c2=T[b+2], c3=T[b+3], c4= T[b+4], c5=T[b+5];
            final boolean isA=c0=='A' && c1=='T' && c2=='O' && c3=='M' && c4==' '&& c5==' ';
            final boolean isH=c0=='H' && c1=='E' && c2=='T' && c3=='A' && c4=='T'&& c5=='M';
            if ( l>CHAIN_ID && ( isA||isH) ) {
                if (rr==null) rr=new boolean[128];
                rr[T[b+CHAIN_ID] & 127]=true;
            }
        }
        return rr;
    }
    private static boolean writePdbOnlyChain(final OutputStream fos, final char chain, ByteArray ba) throws IOException {
        final int
            ATOM=('A')+('T'<<8)+('O'<<16)+('M'<<24),
            HEAD=('H')+('E'<<8)+('A'<<16)+('D'<<24),
            CONE=('C')+('O'<<8)+('N'<<16)+('E'<<24),

            TER =('T')+('E'<<8)+('R'<<16)+(' '<<24),
            HETA=('H')+('E'<<8)+('T'<<16)+('A'<<24),
            HELI=('H')+('E'<<8)+('L'<<16)+('I'<<24),
            SHEE=('S')+('H'<<8)+('E'<<16)+('E'<<24),
            SEQR=('S'<<0)+('E'<<8)+('Q'<<16)+('R'<<24),
            ANIS=('A')+('N'<<8)+('I'<<16)+('S'<<24);  // ANISOU  OU

        if (ba==null) return false;
        final int ends[]=ba.getLineEnds();
        final byte[] T=ba.getBytes();
        if (T==null) return false;
        boolean success=false;
        //putln(ba.newString());

        nextLine:
        for(int iL=0; iL<ends.length; iL++) {
            final int b= iL==0 ? 0 : ends[iL-1]+1;
            final int l=ends[iL]-b;
            if (l>6) {
                final int first32bits= charsTo32bit(T,b);
                final int c4= T[b+4], c5=T[b+5];
                int idxChain=-1;
                switch(first32bits){
                case ANIS: if (c4=='O'&& c5=='U')
                    continue nextLine;
                    break;
                case CONE:
                    if (c4=='C'&& c5=='T') continue nextLine;
                    break;
                case TER:
                    if (c4==' '&& c5==' ') continue nextLine;
                    break;
                case ATOM:
                    if (c4==' '&& c5==' ') idxChain=CHAIN_ID;
                    break;
                case HETA:
                    if (c4=='T'&& c5=='M') {
                        idxChain=CHAIN_ID;
                        /*
                          if (l>19) {
                          final int resName=charsTo32bit(T,17);
                          if (resName==HETERO_HOH||resName==HETERO_HHO||resName==HETERO_OHH||resName==HETERO_H2O||
                          resName==HETERO_DOD||resName==HETERO_DDO||resName==HETERO_ODD||resName==HETERO_D2O) skip=true;
                          }
                        */
                    }
                    break;
                case HELI:
                    if (c4=='X'&& c5==' ') idxChain=19;
                    break;
                case SHEE:
                    if (c4=='T'&& c5==' ') idxChain=21;
                    break;
                case SEQR:
                    if (c4=='E'&& c5=='S') idxChain=11;
                    break;
                }
                if (idxChain>0 && idxChain<l) {
                    final char c=(char)T[b+idxChain];
                    if (c==chain) success=true;
                    else continue nextLine;
                }
            }
            fos.write(T,b,l);
            fos.write('\n');
        }
        fos.flush();
        return success;
    }
    public static boolean writePdbOnlyChain(final File f, final char chain, final ByteArray ba) {
        if (ba==null) return false;
        OutputStream fos=null;
        boolean success=false;
        try {
            fos=fileOutStream(f);
            success=writePdbOnlyChain(fos,chain,ba);
            return success;
        } catch(IOException e) { putln("Caught PDB_separateChains#writePdbOnlyChain ",e); return false;}
        finally {
            closeStrm(fos);
            if (success)  {
                for(int i=0;i<3; i++) if (sze(f)==0) sleep(99);
            } else delFile(f);
        }

    }
    private static String newName(File  f, char chain) {
        if (f==null) return null;
        final String fn=delSfx(".zip",delSfx(".Z",delSfx(".bz2",delSfx(".gz", f))));
        if (chain==' ') chain='_';
        final int i=fn.lastIndexOf('.');
        if (i<0 || i<fn.lastIndexOf('/') || i<fn.lastIndexOf('\\')) return fn+'_'+chain+".pdb";
        return fn.substring(0,i)+'_'+chain+fn.substring(i);
    }
}
