package charite.christo;
import charite.christo.interfaces.*;
import java.util.*;
import javax.swing.Icon;
import java.net.URL;
import java.io.File;
import java.util.zip.ZipFile;

import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
public class UniqueList<T> implements List<T>,  HasRendererText,TabItemTipIcon.Interface, ChRunnable, HasPropertyMap, HasModiCount {
    public final static int UNIQUE=1<<1, USE_HASH_SET=1<<2, ASSERT_EDT=1<<3, TO_STRING_SUPER=1<<4;
    private int mc, options=UNIQUE;
    private Set hashSet;
    private List _v;
    public List<T> v() {
        if (_v==null) {
            _v=new MyVect();
        }
        return _v;
    }
    private final Class clazz;
    private Map _clientObjects;
    public UniqueList(Class c){
        assertInstAllowed();
        clazz=c;
    }
    public Class getClazz() { return clazz;}
    public UniqueList rmOptions(long opt) { options&=~opt; return this; }
    public UniqueList addOptions(long opt) {
        options|=opt;
        return this;
    }
    public int getModificationCount() { return mc;}
    @Override public boolean equals(Object o) { return o==this;}
    public Map getPropertyMap(boolean create) {return _clientObjects==null&&create?_clientObjects=new ChMap() : _clientObjects;};
    /* <<< Instance <<< */
    /* ---------------------------------------- */
    /* >>> indexOf >>> */
    public int indexOf(Object o, int from0, int to0) {
        final List v=_v;
        final Object[] elDat=v instanceof MyVect ? ((MyVect)v).getData() : null;
        final int size=_v==null ? 0 : elDat!=null ? mini(elDat.length,size()) : size();
        if (size==0) return -1;
        final boolean fwd=from0<to0;

        final int to=fwd?mini(to0,size):to0;
        final boolean equals2=o instanceof File || o instanceof URL || o instanceof ZipFile || o instanceof CharSequence;
        for(int i=mini(size-1,from0); fwd?i<to:to<i; i+=fwd ? 1 : -1) {
            final Object el=elDat!=null ? elDat[i]:v.get(i);
            if (equals2 ? equals2(el,o) : el==o) return i;
        }
        return -1;
    }
    private boolean equals2(Object o1, Object o2) {
        if (o1 instanceof CharSequence) return o1.equals(o2);
        else if (o1 instanceof ChFile && o2 instanceof ChFile) return o1.hashCode()==o2.hashCode() && o1.toString().equals(o2.toString());
        else if (o1 instanceof URL    && o2 instanceof URL) return o1.toString().equals(o2.toString());
        else if (o1 instanceof File && o2 instanceof File) return o1.equals(o2);
        else if (o1 instanceof ZipFile && o2 instanceof ZipFile) return ((ZipFile)o1).getName().equals(((ZipFile)o2).getName());
        return o1==o2;
    }
    public UniqueList ensureCapacity(int n) {
        final List v=v();
        if (v instanceof Vector) ((Vector)v).ensureCapacity(n);
        return this;
    }
    /* <<< indexOf <<< */
    /* ---------------------------------------- */
    /* >>> Override  >>> */
    @Override public List<T> subList(int from, int to) { return v().subList(from,to);}
    @Override public int size() {
        runHooks(HOOK_ACCESS);
        return _v==null ? 0 : _v.size();
    }
    @Override public boolean isEmpty() { return size()>0; }
    @Override public ListIterator<T> listIterator() {return listIterator(0); }
    @Override public ListIterator<T> listIterator(int i) {return v().listIterator(i); }
    @Override public Iterator<T> iterator() { return v().iterator(); }
    @Override public Object[] toArray() { return size()>0?v().toArray() : NO_OBJECT;}
    @Override public <T> T[] toArray(T[] a) { return v().toArray(a); }
    @Override public boolean containsAll(Collection<?> c) { return _v==null ? false : sze(c)==0 ? true :  v().containsAll(c); }
    @Override public synchronized boolean contains(final Object o) {
            final List v=_v;
            final int size=v==null ? 0 : size();
            if (size==0) return false;

            if (0!=(this.options&USE_HASH_SET)) {
                if (hashSet==null) (hashSet=new HashSet()).addAll(this);
                if (hashSet.size()!=size) {
                    assertTrue(false);
                    hashSet.retainAll(this);
                    hashSet.addAll(this);
                }
                return hashSet.contains(o);
            }

            return indexOf(o)>=0;
        }
    @Override public int indexOf(Object o) { return indexOf(o, 0,MAX_INT); }
    @Override public int lastIndexOf(Object o) { return indexOf(o, MAX_INT, -1); }

    @Override public synchronized T get(int i) {
            if (i<0 || i>=size()) return null;
            try{ return v().get(i);} catch(Throwable e){ return null;}
        }

    /* *** Primary Methods that change *** */

    /* --- Remove List --- */
    @Override synchronized public T remove(int i)  {
            if (0!=(options&ASSERT_EDT)) assertEDT();
            final T removed=i>=0 && i<size() ? v().remove(i) : null;
            if (removed!=null) {
                _changed();
                if (hashSet!=null) hashSet.remove(removed);
            }
            return removed;
        }
   @Override synchronized public boolean retainAll(Collection c)  {
            if (0!=(options&ASSERT_EDT)) assertEDT();
            if (v().retainAll(c)) {
                _changed();
                if (hashSet!=null) hashSet.retainAll(c);
                return true;
            } else return false;
        }
    @Override public boolean removeAll(Collection<?> c) {
        if (0!=(options&ASSERT_EDT)) assertEDT();
        if (hashSet!=null) hashSet.removeAll(c);
        boolean changed= _v==null ? false : v().removeAll(c);
        if (changed) _changed();
        return changed;
    }

     @Override synchronized public final void clear() {
            if (0!=(options&ASSERT_EDT)) assertEDT();
            if (size()>0) {
                v().clear();
                _changed();
                if (hashSet!=null) hashSet.clear();
            }
        }
    /* --- Add --- */

    @Override synchronized public boolean add(T s) {
            if (0!=(options&ASSERT_EDT)) assertEDT();
            assertTrue(!(s instanceof java.lang.ref.SoftReference));
            if (s!=null && !(0!=(options&UNIQUE) && contains(s)) && v().add(s)) {
                _changed();
                if (hashSet!=null) hashSet.add(s);
                return true;
            } else return false;
        }

    @Override public boolean addAll(Collection<? extends T> c) { return addAll(0,c); }

    @Override public boolean addAll(int pos, Collection<? extends T> c) {
        final int L=sze(c), size=size();
            if (L==0) return false;
            if (0!=(options&ASSERT_EDT)) assertEDT();
            Collection cAdd=c;
            if (0!=(options&UNIQUE)) (cAdd=c instanceof Set ? c : new HashSet(c)).removeAll(v());
            v().addAll(pos, cAdd);
            final boolean changed=size!=size();
            if (changed) {
                _changed();
                if (hashSet!=null) hashSet.addAll(c);
            }
            return changed;
        }

    @Override synchronized public final void add(int idx, T s) {
            if (0!=(options&ASSERT_EDT)) assertEDT();
            if (s!=null && !(0!=(options&UNIQUE)&&contains(s))) {
                final int i=idx<0 ? 0 : idx;
                if (i<size()) try { v().add(i,s); } catch(Throwable e){ v().add(s);}
                else v().add(s);
                _changed();
                if (hashSet!=null) hashSet.add(s);
            }
        }
    /* --- Mixed add and remove --- */

    @Override synchronized public boolean remove(Object o) {
            if (0!=(options&ASSERT_EDT)) assertEDT();
            if (v().remove(o)) {
                 _changed();
                 if (hashSet!=null) hashSet.remove(o);
                return true;
            } else return false;
        }

    @Override synchronized public T set(int i,T o) {
            if (0!=(options&ASSERT_EDT)) assertEDT();
            if (i>=size() || i<0) return null;
            if (o==null || get(i)==o) return o; /* no check, otherwise Collections.sort does not work */
            _changed();
            if (hashSet!=null) hashSet.remove(get(i));
            return v().set(i,o);
        }

    /* <<< implements List  <<< */
    /* ---------------------------------------- */
    /* >>> asArray  >>> */
    public void debug_reset() { refArray=null;}
    private int asArray_mc;
    private Object refArray;
    public synchronized T[] asArray() {
            final List v=v();
            final int size=size();
            final Class c=getClazz();
            if (size==0) return (T[])emptyArray(c);
            Object[] a=(Object[])deref(this.refArray);
            if (a==null || a.length!=size || asArray_mc!=mc) {
                asArray_mc=mc;
                if (clazz==java.lang.ref.Reference.class) {
                    for(int i=size; --i>=0;) {
                        try {
                            final Object ref=v.get(i);
                            if (deref(ref)==null) remove(ref);
                        } catch(Throwable e){}
                    }
                }
                try {
                v().toArray(a=(Object[])java.lang.reflect.Array.newInstance(c,size));
                } catch(Throwable ex){ putln("Caught in UniqueList "+c+" "+this+"  "+ex); };
                final Object[] noNull=rmNullA(a,c);
                if (a!=noNull) while(v.remove(null));
                this.refArray=weakRef(a=noNull);
            }
            if (isMyComputer() && cntains(null,a)) {
                putln("UniqueList null ",clazz);
            }
            return (T[]) a;
        }
    public Object[] elementData() {
        final List v=_v;
        final Object[] oo=v==null ?  NO_OBJECT : v instanceof MyVect ? ((MyVect)v).getData() : null;
        return oo!=null ? oo : asArray();
    }
    /* <<< asArray <<< */
    /* ---------------------------------------- */
    /* >>> Convenience >>> */

    public boolean addAllA(T[] oo) {
        boolean changed=false;
        if (oo!=null ) for(T o : oo) changed|=add(o);
        return changed;

    }
    /* <<< Convenience <<< */
    /* ---------------------------------------- */
    /* >>>  Appearance >>> */
    private String labelText="";
    private Object ic;
    public String getText() { return labelText;}
    public UniqueList<T> t(String text) { labelText=text;return this;}
    public UniqueList<T> i(Icon icon) { ic=icon; return this;}
    public UniqueList<T> i(String icon) { ic=icon; return this;}
    public TabItemTipIcon getTabItemTipIcon(java.awt.AWTEvent ev) {
        final Icon icon=ic instanceof Icon ? (Icon)ic : ic instanceof String ? iicon((String)ic) : null;
        return TabItemTipIcon.shared(null,labelText,null, icon);
    }
    public UniqueList like(Object l) {  return  t(getTxt(l)).i(getIcn(l)); }

    @Override public String toString() {
        return
            0!=(options&TO_STRING_SUPER) ? v().toString() :
            labelText!=null ? labelText+"#"+sze(this) : v().toString();
    }

    /* <<< Appearance <<< */
   //========================================
    /* --- Convenience  */
    public boolean removeAll(Object remove[]) {
        if (0!=(options&ASSERT_EDT)) assertEDT();
        boolean modified=false;
        for(int i=size();--i>=0;) {
            final Object o=get(i);
            if (ChUtils.idxOf(o,remove)>=0) {
                modified=true;
                remove(o);
            }
        }
        return modified;
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Hooks >>> */
   public final static int HOOK_ACCESS=1, HOOK_CHANGE=0;
    private ChRunnable[][] hooks;
    private String[][] hooks_ID;
    public UniqueList addHook(int type, ChRunnable r, String s) {
        if (r!=null && s!=null) {
            if (hooks==null) { hooks=new ChRunnable[2][]; hooks_ID=new String[2][]; }
            hooks_ID[type]=ad(s,hooks_ID[type],0);
            hooks[type]=adToArray(r, hooks[type],0, ChRunnable.class);
        }
        return this;
    }
    private void runHooks(int type) {
        if (hooks==null || hooks[type]==null) return;
        for(int i=hooks[type].length; --i>=0;) hooks[type][i].run(hooks_ID[type][i],this);
    }
    private void _changed() {
        mc++;
        if (hooks!=null && hooks[HOOK_CHANGE]!=null) runHooks(HOOK_CHANGE);
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */
    private Object nodeId, rendererComponent;
    public Object run(String id, Object arg) {
        if (id==PROVIDE_TEXT_FOR_SORTING) return labelText;
        if (id==PROVIDE_JTREE_NODE_ID) return labelText!=null ? labelText :  nodeId==null ? nodeId=intObjct(new Object().hashCode()) : nodeId;
        if (id==SET_TREE_VALUE) rendererComponent=arg;
        if (id==MAY_PROVIDE_TREE_VALUE) return rendererComponent;
        return null;
    }
    public CharSequence getRendererText(int options){
        return labelText+" #"+size();
    }

    private static class MyVect extends Vector {
        static boolean error;
        public Object[] getData() {
            try {
                return error ? null : super.elementData;
            } catch(Throwable ex){ error=true; assertTrue(false); }
            return null;
        }
    }
}
//http://java.sun.com/j2se/1.5.0/docs/guide/collections/reference.html  Collections
