package charite.christo.blast;
import static charite.christo.ChUtils.*;
import charite.christo.*;
import charite.christo.interfaces.*;
import java.util.*;
import java.io.*;
import static charite.christo.ChConstants.*;
/**
   @author Christoph Gille InputStream
*/
public abstract class AbstractBlaster implements SequenceBlaster, HasControlPanel,Disposable,java.awt.event.ActionListener {
    /** This can be overridden */
    private boolean isAAQuery=true,  hasResultXml;
    private String urlResultHtml, query, fromCache="found in cache\n";
    private int numberOfAlignments=2000, sensitivity=5;
    private String matrix="blosum62";
    public void setSensitivity(int n) { sensitivity=n; }
    public int getSensitivity() { return sensitivity; }
    public void setAAQuerySequence(String s) {query=onlyLetters(s.toUpperCase()); isAAQuery=true;}
    public void setNTQuerySequence(String s) {query=onlyLetters(s.toUpperCase()); isAAQuery=false;}
    public String getQuerySequence() { return query;}
    public int getNumberOfAlignments() { return numberOfAlignments;}
    public void setNumberOfAlignments(int n) {numberOfAlignments=n;}
    public boolean isAAQuery() {return isAAQuery;}
    public String getBlastProgram(){ return isAAQuery() ? "blastp" : "blastn";}
    public String getQuery() { return query;}
    public ByteArray getResultXml(){
        final String q=this.query;
        if (q==null) return null;
        final String cacheS=cacheSK('S'), cacheK=cacheSK('K');
         {
             CacheResult.getLastModified(getClass(),cacheS,cacheK);
             ByteArray ba= CacheResult.getValue(getClass(),cacheS,cacheK,(ByteArray)null);
            if (ba!=null && strstr("FATAL: ",ba)>0) {
                putln("AbstractBlaster Error: found \"FATAL: \"");
                return null;
            }
            if (ba!=null && !this.hasResultXml) {
                this.hasResultXml=true;
                log(fromCache);
            }
            return ba;
        }
    }
    public void setUrlResultHtml(String url) {
        this.urlResultHtml=url;
        inEDT(threadM("enableDisable", this, NO_OBJECT));
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Matrix  >>> */
    public void setMatrix(String mx) throws IllegalArgumentException {
        for(String m:getAvailableMatrices())
            if (mx.equals(m)) {matrix=m; return; }
        throw(new IllegalArgumentException(mx+": no suXbch blast similarity matrix"));
    }
    /**FIXME*/
    public String[] getAvailableMatrices() { return new String[]{"blosum64"}; }
    public String getMatrix() {return matrix;}
    /* <<< Matrix <<< */
    /* ---------------------------------------- */
    /* >>> Databases >>> */
    public void setDatabase(String s) {
      this.database=delToLastChar(s,' ');
    }
    public String getDatabase(){return database;}
    /* <<< Databases <<< */
    /* ---------------------------------------- */
    /* >>> log >>> */
    private AsciiView _log=new AsciiView(999);
    public AsciiView logStdout(CharSequence... oo) { putln(oo); return log(oo); }
    public AsciiView log(CharSequence... oo) {
        if (_log==null) _log=new AsciiView(999);
        return _log.appendSS(oo);

    }
    /* <<< log <<< */
    /* ---------------------------------------- */
    /* >>> GUI >>> */
    private java.awt.Container _ctrlPnl;
    @Override public java.awt.Container getControlPanel() {
        if (_ctrlPnl==null) {
            _ctrlPnl=newCNSEW(
                              scrllpn(0L,log(),dim(99,99)),
                              TabItemTipIcon.homePage(this),
                              ChUtils.pnl(button(BUT_XML), button(BUT_XMLb)," ", button(BUT_TXT),button(BUT_TXTe), " ", button(BUT_IN_BROWSER)),
                              null,null
                              );

        }
        return _ctrlPnl;
    }

    public final static int BUT_XML=1, BUT_XMLb=2, BUT_TXT=3, BUT_TXTe=4, BUT_IN_BROWSER=5;
    public final ChButton[] button=new ChButton[6];
    public ChButton button(int type) {
        ChButton b=button[type];
        if (b==null) {
            if (type==BUT_XMLb) b=new ChButton().li(this).t("XML in file browser");
            if (type==BUT_XML)  b=new ChButton().li(this).t("Result XML");
            if (type==BUT_TXT)  b=new ChButton().li(this).t("Result text");
            if (type==BUT_TXTe) b=new ChButton().li(this).t("Text in editor");
            if (type==BUT_IN_BROWSER) b=new ChButton().li(this).t("View in Browser");
            button[type]=b;
            enableDisable();
        }
        return b;
    }
    private String database=null;
    /* <<< GUI <<< */
    /* ---------------------------------------- */
    /* >>> Cache >>> */
    private final PrgParas prgPara=new PrgParas();
    public final PrgParas getPrgParas() { return prgPara;}

    private String cacheS, cacheK;
    private final static StringBuffer sb=new StringBuffer();
    public String cacheSK(char type) {
        final String q=this.query;
        if (q==null) return null;
        synchronized(sb) {
            sb.setLength(0);
            if (type=='K') {
                if (cacheK==null) {
                    sb.append(q);
                    join(getPrgParas().asStringArray(HasPrgParas.EQUALS_SIGN), "_", sb);
                    cacheK=sb.toString();
                }
                return cacheK;
            } else {
                if (cacheS==null) cacheS="/"+getDatabase()+"/"+
                                      (sensitivity==5?"":sensitivity+"/")+
                                      CacheResult.sectionForSequence(q);

                return cacheS;
            }
        }
    }
    /* <<< Cache <<< */
    /* ---------------------------------------- */
    /* >>> Disposable >>> */
    private boolean isDisposed;
    public void dispose(){isDisposed=true;}
    public boolean isDisposed(){return isDisposed;}
    /* <<<  Disposable <<< */
    /* ---------------------------------------- */
    /* >>> Compute >>> */
    public abstract ByteArray computeIt();
    public final void compute() {
        final String q=this.query;
        if (sze(q)<5) {
            log(RED_ERROR," in AbstractBlaster query sequence too short",q,"\n");
            return;
        }
        if (database==null) {
            log(RED_ERROR," in AbstractBlaster: database=null");
            return;
        }
        final String sync=("AB$$SYNCc"+getClass()+" "+hashCd(cacheS)+" "+hashCd(cacheK)).intern();
        synchronized(sync) {

            final String cacheS=cacheSK('S'), cacheK=cacheSK('K');
            final long lastModified=maxAge==MAX_INT ? 0:CacheResult.getLastModified(getClass(), cacheS, cacheK);

            if (CacheResult.isEnabled() && maxAge*1000>System.currentTimeMillis()-lastModified) {
                final ByteArray ba=getResultXml();
                if (ba!=null) {
                    log(ANSI_GREEN+"Found in cache\n"+ANSI_RESET);
                    if (lastModified>0) log("Computation was performed on ",day(lastModified),"\n");
                }
            }
            if (!this.hasResultXml) {
                final ByteArray result=computeIt();
                fromCache=null;
                if (result!=null){
                    CacheResult.putValue(getClass(),cacheS,cacheK,result,true);
                    ChMsg.sound(ChMsg.SOUND_SUCCESS);
                    this.hasResultXml=result!=null;

                } else ChMsg.sound(ChMsg.SOUND_FAILURE);
            }
        }
        inEDT(threadM("enableDisable", this, NO_OBJECT));

    }
    public void enableDisable() {
        for(int i=button.length;--i>=0;) {
            setEnabld(i==BUT_IN_BROWSER ? 0<sze(urlResultHtml) : hasResultXml, button[i]);
        }
    }
    private File fTxt;
    public void actionPerformed(java.awt.event.ActionEvent ev) {
        final int modi=modifiers(ev), iBut=idxOf(ev.getSource(), button);
        if (iBut==BUT_XML) new AsciiView(getResultXml()).showInFrame("Blast XML");
        if (iBut==BUT_IN_BROWSER) visitURL(urlResultHtml, modi);
        if (iBut==BUT_TXTe || iBut==BUT_TXT) {
            final ByteArray xml=getResultXml();
            if (xml!=null) {
                if (sze(fTxt)==0) {
                    fTxt=createTempFile(".txt");
                    try{
                        final OutputStream fOut=fileOutStream(fTxt);
                        new BlastParser().parseXML(getResultXml(), 99999, fOut);
                        closeStrm(fOut);
                    } catch(IOException e){putln("Caught in AbstractBlaster ",e);}
                }
                if (sze(fTxt)>0) {
                    if (iBut==BUT_TXTe) viewFile(fTxt, modi);
                    if (iBut==BUT_TXT ) new AsciiView(fTxt).showInFrame("Blast result");
                }
            }
        }
        if (iBut==BUT_XMLb) {
            final File fTmp=createTempFile(".xml");
            final ByteArray xml=getResultXml();
            if (xml!=null) {wrte(fTmp,xml); viewFile(fTmp, modi);}
        }
    }

    protected boolean isAaDb() {
        return 0<=idxOfStrg(STRSTR_END|STRSTR_IC|STRSTR_w_L, getDatabase(), getAvailableDatabases(true, false));
    }

    private long maxAge=MAX_INT;
    @Override public void cachedResultsNotOlderThanSec(int time) { maxAge=time; }

}
