package charite.christo.blast;
import charite.christo.*;
import charite.christo.interfaces.*;
import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static charite.christo.protein.ChSoapEBI.*;
/**
   @author Christoph Gille
   // See SocketOutputStream  SocketInputStream
*/
public class Blaster_SOAP_ebi extends AbstractBlaster implements NeedsInternet {
    private final static String[][]
        DB={
        { "EMBL Release em_rel", "EMBL Updates emnew", "EMBL Coding Sequence emcds"},
        { "*aminos_from_3D_structures pdb", "aminoacid uniprotkb",
          "*aminoacids uniref100", "aminoacids uniref90","aminoacids uniref50"}
    }, availDB={null,null};
    public String[] getAvailableDatabases(boolean isAmino, boolean withoutDelay) {
        final int isAA=isAmino?1:0;
        if (withoutDelay) return availDB[isAA]!=null ? availDB[isAA] : DB[isAA];
        synchronized("BSE%%SYNCgad") {
            if (availDB[isAA]==null) {
                final ByteArray paramDetailDB=getParamDetail("database");
                for(int type=2; --type>=0;) {
                    String ss[]=parseDatabases(type!=0, paramDetailDB);
                    if (sze(ss)==0) ss=DB[type];
                    availDB[type]=ss;
                    charite.christo.blast.BlastParser.addToProteinDatabases(charite.christo.blast.BlastParser.URL_EBI,ss);
                }
            }
            return availDB[isAA];
        }
    }

    @Override public String[] getAvailableMatrices() { return new String[]{ isAaDb() ?  "blosum65" : "identity.4.2"}; }
    @Override public ByteArray computeIt() {
        try {
            final String
                db0=getDatabase(),
                db=db0.substring(1+db0.lastIndexOf(' ')),
                num=String.valueOf(mini(500,getNumberOfAlignments()));
            final int sens=getSensitivity();
            final boolean isPdb=equls(STRSTR_IC,"pdb",db);

            String resultId=null;

            ByteArray answer=null,query=null;
            query=getRsc("soapWublastEbi");
            if (query!=null) {
                query
                    .replace("SENSITIVITY", sens>=9?"high":sens>=7?"high":sens>=5?"normal":sens>=3?"low":"vlow")
                    .replace("ALIGNMENTS", num)
                    .replace("SCORES", num)
                    .replace("EXP", sens<=3 ? "1e-1" : "100")         /*10*/
                    .replace("PROGRAM", isAAQuery() ? (isAaDb() ? "blastp" : "tblastn") : (isAaDb() ? "blastx" : "blastn"))
                    .replace("DATABASE", db)
                    .replace("SEQUENCE", getQuery());
                log(ANSI_GREEN+"\nSocket Write"+ANSI_RESET+"\n",query,"\n");
                synchronized(toStrgIntrn("BSEBI$$SYNC"+db+sens)) {
                    putln("Start Blast at EBI !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! ");
                    sleep(3000);
                    answer=getResponse(ASK_FOR_PERMISSION, query);
                }
            }
            resultId=answer==null?null : answer.newStringToGt(strstr(STRSTR_AFTER,"<jobId>",answer));
            setUrlResultHtml("http://www.ebi.ac.uk/Tools/services/web/blastresult.ebi?tool=wublast&jobId="+resultId);
            boolean error=sze(resultId)==0;
            if (!error) {
                int status=0;
                for(int iTry=1000; --iTry>=0 && (status==0 || status==STATUS_RUNNING);) {
                    sleep( (isPdb?1000:3000)*sens);
                    status=getStatus(resultId);
                    putln("#"+iTry+" Loop Blaster_SOAP_ebi status="+status);
                }
                error|= (status!=STATUS_FINISHED);
            }
            //new ChTextViewer(urlResultHtml+"&format=xml\n\n"+resultXml).shw(0);
            final byte[] rXml=error ? null : getResult(resultId, "xml");
            error |= sze(rXml)==0;
            if (!error) return new ByteArray(rXml);
            error(new ByteArray(sze(query)+sze(answer)+999).append(query).append('\n',2).append('-',60).append('\n').append(answer));
        } catch(Throwable ex){stckTrc(ex);}
        return null;
    }

    public static ByteArray getParamDetail(String paramDetail) {
        final ByteArray query=getRsc("soapWublastParamDetailEBI");
        if (query==null)  return null;
        query.replace("PARAMETERID",paramDetail);
        return getResponse(0L, query);
    }

    //public static ByteArray getResoib
    public static String[] parseDatabases(boolean protein, ByteArray xml) {
        if (xml==null) return null;
        final byte T[]=xml.getBytes();
        final int E=xml.getEnd();
        final List v=new ArrayList(99);
        for(int value=0; 0<(value=strstr(STRSTR_AFTER,"<value>",T, value+1,E));) {
            final int end=strstr(STRSTR_AFTER, "</properties>",T, value,E);
            if (end<0) break;
            // putln(strstr(">protein<",T, value,end)+  " ----------   "+((0<strstr(">protein<",T, value,end)) == protein));
            if ((0<strstr(">protein<",T, value,end)) == protein) {
                final String item, id=xml.newStringToGt(strstr(STRSTR_AFTER, "<value>",xml, value,end));
                if (id.indexOf("ensemblgenomes")>=0) item=id;
                else {
                final String label=xml.newStringToGt(strstr(STRSTR_AFTER, "<label>",xml, value,end));
                item=label+"   "+id;
                }

                v.add(item);
            }
            value=end;
        }
        return strgArry(v);
    }
    public String getUrlMask(String db) {
        return "http://www.ebi.ac.uk/cgi-bin/dbfetch?db="+db+"&style=raw&id=";
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Static utils >>> */
    /* <<< Static utils  <<< */
    /* ---------------------------------------- */
    /* >>> Main >>> */
    // java charite.christo.protein.Blaster_SOAP_ebi EYAFKATNQTNINSLAVRGKDCTVVIS
    /*
    public static void main(String argv[]) {
        setPrgParameters("",argv,true);
        CacheResult.setEnabled(false);

        final SequenceBlaster b=new Blaster_SOAP_ebi();
        b.getAvailableDatabases(true,false);

        b.setAAQuerySequence(argv[0]);
        b.setDatabase("pdb");
        b.setSensitivity(1);
        putln("xxxxxxxxxxxxxxxxxxxxxxxxxxxxx compute");
        b.compute();
        putln("Result=\n", b.getResultXml());

    }
    */
}

/*
 <hit number="485" database="SP" id="PSA_THEAC" ac="P25156" length="233" description="Proteasome subunit alpha OS=Thermoplasma acidophilum (strain ATCC 25905 / DSM 1728 / JCM 9062 / NBRC 15155 / AMRC-C165) GN=psmA PE=1 SV=2">

http://www.ebi.ac.uk/cgi-bin/dbfetch?db=SP&style=raw&id=PSA_THEAC
http://www.ebi.ac.uk/cgi-bin/dbfetch?db=TR&style=raw&id=A2BKB7
"A2BKB7

 */