package charite.christo.hotswap;
import charite.christo.*;
import java.util.*;
import static charite.christo.ChUtils.*;

class ChInvocationHandler implements HotswapProxy, java.lang.reflect.InvocationHandler {
    private final ProxyClass cls;
    private Object obj, tmp;

    // ================================
    // Constructors
    // ================================
    private static String lastErrorMessage;
    public Object invoke(Object src, java.lang.reflect.Method method, Object[] args) throws Throwable {
        if (obj==null || !(obj.getClass().equals(cls.getClassInstance()))) obj=hotswap_newInstance();
        hotswap_enqueue();
        cls.mayHotswap();
        try {
            final Object result=method.getName().startsWith("hotswap")
                ? method.invoke(this,args)
                : method.invoke(hotswap_getInstance(), args);
            lastErrorMessage=null;
           return result;
        }catch(Throwable e) {
          putln("caught in ChInvocationHandler: ",e);
            Throwable cause=e;
            while( (cause=cause.getCause())!=null) {
                String msg=cause.getMessage();
                if (msg==null || msg.equals(lastErrorMessage)) {
                    putln("ChInvocationHandler cause=",cause);
                    lastErrorMessage=msg;
                    cause.printStackTrace(System.out);
                }
            }
            return null;
        }
    }
    ChInvocationHandler(ProxyClass cls){
        this.cls=cls;
    }
    @Override public String toString(){
        return "(Proxy of class "+cls.getClassInstance().getName()+": "+obj+")";
    }
    @Override public boolean equals(Object other){
        if (other==this) return true;
        if (!(other instanceof ChInvocationHandler))  return false;
        final ChInvocationHandler that=(ChInvocationHandler)other;
        return (this.cls.getClassInstance().equals(that.cls.getClassInstance()));
    }

    public ProxyClass hotswap_getProxyClass(){
        return cls;
    }

    public synchronized Object hotswap_getInstance() {
        // If the object is null, this is the first time the object
        // has been constructed or the arguments have been updated.
        if (obj==null)  obj=hotswap_newInstance();
        return obj;
    }
    public synchronized void hotswap_enqueue(){
        if (!cls.isEnqueued(this)) cls.enqueue(this);
    }
    public synchronized void hotswap_dequeue(){
        if (cls.isEnqueued(this))  cls.dequeue(this);
    }
    public synchronized boolean hotswap_isEnqueued(){
        return cls.isEnqueued(this);
    }
    public synchronized void hotswap_release(){
        // Release from the proxy class
        cls.dequeue(this);
        tmp=null;
    }
    synchronized boolean hotswap_prepare(Class newClass){
        try {
            tmp=newClass.getDeclaredConstructor().newInstance();
        } catch (Throwable ex) {
            ex.printStackTrace(System.out);
            return false;
        }
        return true;
    }
    synchronized void hotswap_commit(){
        // Swap the old instance with the new.
        Object old=obj;
        obj=tmp;
        tmp=old;
    }
    synchronized void hotswap_rollback(){
        tmp=null;
    }
    synchronized Object hotswap_newInstance(){
        try {
            return cls.getClassInstance().getDeclaredConstructor(NO_CLASS).newInstance();
        } catch (Exception ex) {
            putln("caught in ChInvocationHandler ",cls);
            putln(ex);
            return null;
        }
    }
}
