package charite.christo.hotswap;
import java.util.*;
import java.io.File;
import charite.christo.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**
   These hotswap classes are adapted from the inxar.hotswap library by Paul Cody Johnston.
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
*/
public class ProxyClass implements charite.christo.interfaces.HasRendererText{
    private final File classfile,sourcefile;
    private ProxyClass(File f,String className){
        this.className=className;
        final String fn=f.getAbsolutePath();
        final String noExt=ChUtils.withoutDotSuffix(fn);
        classfile=new File(noExt+".class");
        sourcefile=new File(noExt+".java");
        this.proxies=new Stack();
    }
    public Class[] getProxyInterfaces() {
        if (interfaces_==null) {
            final List set=new ArrayList();
            Class cls=getClassInstance();
            adUniq(HotswapProxy.class, set);
            while (cls!=null) {
                adAllUniq(cls.getInterfaces(), set);
                cls=cls.getSuperclass();
            }
            interfaces_= toArry(set,Class.class);
        }
        return interfaces_;
    }
    public HotswapProxy newInstance(){
        HotswapProxy hp= newProxy(new ChInvocationHandler(this));
        return hp;
    }
    private HotswapProxy newProxy(ChInvocationHandler proxy){
        try {
            final Class ii[]=getProxyInterfaces();
            final ClassLoader cl=Insecure.CLASSLOADING_ALLOWED ? classLoader() : CLASSLOADER;
            return (HotswapProxy) java.lang.reflect.Proxy.newProxyInstance(cl,ii, (java.lang.reflect.InvocationHandler)proxy );
        } catch (Exception ex) { putln(RED_CAUGHT_IN,ex); return null; }
    }

    private Class[] interfaces_;
    public File getSourcefile(){ return sourcefile; }
    public File getClassfile(){ return classfile;  }
    public String getName(){ return className; }
    public Class getClassInstance() {
        if (cls==null) cls=newClassInstance();
        return cls;
    }
    private boolean needsHotswap;
    public void hotswap(){
        interfaces_=null;
        cls=null;
        needsHotswap=true;
        mayHotswap();
    }
    public boolean mayHotswap(){
        synchronized(this) {
            if (sze(proxies)==0) return false;
            // Disallow recursive hotswaps.
            if (isHotswapping) return false;
            isHotswapping=true;
        }
        boolean result=false;
        try {
            result=hotswapInternal();
        } finally { isHotswapping=false; }
        return result;
    }
    private synchronized boolean hotswapInternal(){
        //if (whenModified==classfile.lastModified()) return false;
        if (!needsHotswap) return false;
        needsHotswap=false;
        Class newClass=null;
        try {
            newClass=newClassInstance();
            //whenModified=classfile.lastModified();
        } catch (Exception ex) {
            putln(RED_CAUGHT_IN+"ProxyClass#hotswapInternal() ",className," ",ex);
            /* Remove all proxies, the hotswap died. */
            proxies.clear();
            return false;
        }
        final UniqueList<ChInvocationHandler> prepared=new UniqueList(ChInvocationHandler.class);
        boolean commit=true;
        try {
            // First prepare all proxies.
            while (!proxies.isEmpty()) {
                ChInvocationHandler p=proxies.pop();
                if (p.hotswap_prepare(newClass)) prepared.add(p);
                else commit=false;
            }
        } catch (Throwable ex) {
            /* might be nice to know the error */
            putln(RED_CAUGHT_IN+"ProxyClass#hotswapInternal() ");
            stckTrc(ex);
            commit=false;
        }
        if (commit) {
            // Do the commit
            for(ChInvocationHandler p : prepared.asArray()) p.hotswap_commit();
            // The transaction has succeeded. Drop the old class and
            // replace it with the new. Also update classload
            // timestamp.
            this.cls=newClass;
        } else {
            for(ChInvocationHandler p : prepared.asArray()) p.hotswap_rollback();
        }
        return commit;
    }
    synchronized void enqueue(ChInvocationHandler proxy){
        if (!proxies.contains(proxy)) proxies.add(proxy);
    }
    synchronized void dequeue(ChInvocationHandler proxy){
        proxies.remove(proxy);
    }
     synchronized boolean isEnqueued(ChInvocationHandler proxy){
        return proxies.contains(proxy);
    }
    static ChClassLoader1 classLoader() {
        final ChZip chZip=ChZip.getInstance(dirHotswap());
        return ChClassLoader1.getInstance(true, chZip.jarFiles(),dirHotswap());
    }
     synchronized Class newClassInstance(){
        try {
            interfaces_=null;
            return classLoader().loadClass(className);
        } catch(Exception e){ putln(RED_CAUGHT_IN+"ProxyClass#newInstance ",className+" ",e); stckTrc(e); return null;}
    }
    private final String className; // the name of the class to proxy
    private Class cls; // the instance of the class to proxy
    private final Stack<ChInvocationHandler> proxies; // a list of enqueued proxies
    private boolean isHotswapping; // flag used to detect hotswap recursion
    public static synchronized ProxyClass getProxyClass(String className){
        if (className==null) return null;
        for(ProxyClass pc : vProxyClass.asArray()) {
            if (className.equals(pc.getName())) return pc;
        }
        return null;
    }
    public static synchronized ProxyClass getProxyClass(File f, String className){
        ProxyClass pc=getProxyClass(className);
        if (pc==null) {
            pc=new ProxyClass(f, className);
            vProxyClass.add(pc);
        }
        return pc;
    }
    public static synchronized void remove(ProxyClass cls){
        vProxyClass.remove(cls);
    }
    @Override public boolean equals(Object o) {
        return o instanceof ProxyClass && ((ProxyClass)o).getSourcefile()==getSourcefile();
    }
    private final static UniqueList<ProxyClass> vProxyClass=new UniqueList(ProxyClass.class);
    public static ProxyClass[] getProxyClasses() { return vProxyClass.asArray();}
    @Override public String toString() { return getRendererText(0);}

    private final int id=count++;
    private static int count;
    public String getRendererText(int options) {
        final StringBuffer SB=strBuf(SB_TIP);
        SB.append(getName()).append(" #").append(id);
        final File fS=getSourcefile(), fC=getClassfile();
        SB.append(
                  sze(fS)==0 ? " (no sourcefile)" :
                  sze(fC)==0 ? " (no classfile)" :
                  fC.lastModified()<fS.lastModified() ? " (no classfile)" : " (ok)" );
        return SB.toString();
    }
}
