package charite.christo.protein;
import charite.christo.interfaces.*;
import java.util.*;
import java.awt.Image;

import charite.christo.*;
import static charite.christo.ChUtils.*;
/**
  A skelleton of an Selection of Residues {@link ResidueSelection ResidueSelection}.
  @author Christoph Gille
*/
public class AbstractResidueSelection extends AbstractVisibleIn123 implements HasWeakRef, ResidueSelection,HasImage, HasPropertyMap, HasName, HasRendererOptions {
    private Protein protein;
    private ResidueSelection sel0;

    public AbstractResidueSelection(ResidueSelection selection) { sel0=selection;}
    public AbstractResidueSelection() {}
    public final Object WEAK_REF=newWeakRef(this);
    public Object getWeakRef() { return WEAK_REF;}

    /* ---------------------------------------- */
    /* >>> Protein >>> */
    public boolean setProtein(Protein newP) {
        protein=newP;
        return true;
    }
    public Protein getProtein() { return protein;};
    /* <<< Protein <<< */
    /* ---------------------------------------- */
    /* >>> Selected Residues >>> */
    private int offset=0;
    private boolean selected[];
    public boolean[] getSelectedAminoacids() {
        final ResidueSelection s=this.sel0;
        if (s!=null) {
            s.setProtein(protein);
            return s.getSelectedAminoacids();
        }
        final boolean bb[]=selected;
        return bb!=null ? bb:NO_BOOLEAN;
    }
    public int getSelectedAminoacidsOffset() {
        final ResidueSelection s=this.sel0;
        return s!=null ? s.getSelectedAminoacidsOffset() : offset;
    }
    public void setSelectedAminoacids(boolean bb[],int offset) {
        Protein.incrementModificationCount(ProteinModiCount.MC_RESIDUE_SELECTIONS,getProtein());
        this.selected=bb;
        this.offset=offset;
    }
    public void setSelectedRange(int from, int to) {
        if (from<0) from=0;
        if (to<from) to=from;
        boolean bb[];
        if(to>from) {
            bb=this.selected;
            if (sze(bb)<to-from) bb=new boolean[to-from+9];
            else Arrays.fill(bb,to-from, bb.length,false);
            Arrays.fill(bb,0,to-from,true);
        } else bb=NO_BOOLEAN;
        setSelectedAminoacids(bb,from);
    }
    /* <<< Selected Residues  <<< */
    /* ---------------------------------------- */
    /* >>> Renderer >>> */
    private Image image;
    private String name;
    @Override public String toString(){
        final ResidueSelection s=this.sel0;
        return s!=null && name==null ? s.toString() : name!=null ? name : super.toString();
    }
    public void setName(String s){ name=s; }
    public String getName(){
        final ResidueSelection s=this.sel0;
        return name!=null ? name : getNam(s);
    }
    @Override public void setImage(Image i) { image=i;}
    public void setImage(String i) { javax.swing.ImageIcon ic=iicon(i); if (ic!=null) image=ic.getImage();}

    @Override public Image getImage(java.awt.Component imageObserver) {
        final  ResidueSelection s=this.sel0;
        final Image im=super.getImage(imageObserver);
        return im!=null ? im : image!=null ? image : s instanceof HasImage ? ((HasImage)s).getImage(imageObserver) : null;
    }
    public void getRGB(int rgb[]) {
        final boolean bb[]=getSelectedAminoacids();
        if (bb==null) return;
        final Protein p=getProtein();
        final int selOffset=p!=null ? getSelectedAminoacidsOffset() : 0;
        final int color=getColor()!=null ? getColor().getRGB(): 0xFF0000;
        final int to=mini(bb.length,rgb.length-selOffset);
        Arrays.fill(rgb,0);
        for(int i=maxi(0,-selOffset); i<to; i++)
            if (bb[i]) rgb[i+selOffset]=color;
    }

    public long getRendererOptions() { return getProtein()==null ? HasRendererOptions.STRIKE_THROUGH : 0L; }
    /* <<< Renderer <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */
    private Map _clientObjects;
    public Map getPropertyMap(boolean create) {return _clientObjects==null&&create?_clientObjects=new HashMap() : _clientObjects;}

}
