package charite.christo.protein;
import java.util.*;
import charite.christo.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
public class ChAlignUtils  {
    /**
       The gapped sequences may be 0-terminated!
    */
    private static void appendName(String prefix, int i, StringBuffer sb) {
        sb.append(prefix);
        for(int j=8-stringSizeOfInt(i); --j>=0;) sb.append('0');
        sb.append(i).append(' ');
    }
    public static StringBuffer toGappedFasta(byte gapped[][]) { return toGappedMSF(gapped,"s", 0, true, new StringBuffer()); }
    public static StringBuffer toGappedMSF(byte gapped[][], String headerPrefix, int offset, boolean fastaFormat, StringBuffer sb0) {
        final StringBuffer sb=sb0!=null ? sb0 : new StringBuffer();
        int size=sze(gapped)*(1+12);
        if (size==0) return sb;
        for(byte bb[]:gapped) size+=bb.length;
        sb.ensureCapacity(size+sb.length());
        for(int i=0; i<gapped.length; i++) {
            if (fastaFormat) sb.append('>');
            if (fastaFormat) sb.append('>').append(headerPrefix).append(i+offset).append('\n');
            else appendName(headerPrefix,i+offset,sb);
            for(byte c: gapped[i]) {
                if (c==0) break;
                sb.append('a'<=c && c<='z' || 'A'<=c && c<='Z' ? (char)c : '-');
            }
            sb.append('\n');
        }
        return sb;
    }

    public static void msfHeader(byte gapped[][],  char type, StringBuffer sb) {
        sb.append("!!AA_MULTIPLE_ALIGNMENT 1.0\n\n myAlignment.msf  MSF: 224  Type: ")
            .append(type).append("  September  9, 2009 09:09  Check: 9999 ..\n\n");
        for(int i=0; i<sze(gapped); i++) {
            sb.append(" Name: ");
            appendName("s",i,sb);
            sb.append(" Len:").append(gapped[i].length).append(" Check: 1 Weight:1\n");
        }
        sb.append("\n\n//\n\n");
    }

    public static StringBuffer toUngappedMultipleFasta(byte sequences[][],String prefix, int offset) {
        int size=0;
        for(byte[] s: sequences) size+=s.length+20;

        final StringBuffer sb=new StringBuffer(size);
        for(int i=0; i<sequences.length; i++) {
            sb.append('>').append(prefix).append(offset+i).append("\r\n");
            for(byte c :sequences[i]) {
                if ('a'<=c && c<='z' || 'A'<=c && c<='Z') sb.append((char)c);
            }
            sb.append("\r\n");
        }
        return sb;
    }

    public static byte[][] blockFormatAlignment(byte[][] gapped, char gap) {
        int maxCol=-1;
        for(byte gg[] : gapped) maxCol=maxi(maxCol,prev(LETTR,gg, sze(gg)-1,-1));
        final int w=maxCol+1;
        byte[][] t=gapped;
        for(int i=gapped.length; --i>=0;) {
            if (gapped[i]==null) gapped[i]=t[i]=new byte[w];
            if (sze(gapped[i])!=w) {
                if (t==gapped) t=gapped.clone();
                t[i]=chSze(t[i],w);
            }
            for(int k=sze(t[i]); --k>=0;) {
                if (!is(LETTR,t[i],k)) {
                    if (t==gapped) t=gapped.clone();
                    if (t[i]==gapped[i]) t[i]=gapped[i].clone();
                    t[i][k]=(byte)gap;
                }
            }

        }
        return t;
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> C-Alpha atoms and Superimpose3D >>> */
    public static byte[][] insertWhereNoCoordinates(int startIdxR,byte[] gapped3dR, int startIdxM, byte[] gapped3dM,   byte[] seqR, byte[] seqM) {
        if (gapped3dR==null || gapped3dM==null || seqR==null || seqM==null) return null;
        byte[] outR=null,outM=null;
        while(true) {
            int oR=0,oM=0,  colR=0, colM=0, iR=0, iM=0;
            for(int i=0; i<startIdxR && iR<seqR.length;) { final byte c=seqR[iR++]; if ('A'<=c && c<='Z') i++; }
            for(int i=0; i<startIdxM && iM<seqM.length;) { final byte c=seqM[iM++]; if ('A'<=c && c<='Z') i++; }
            boolean alreadyFoundLetterR=false, alreadyFoundLetterM=false;
            while(iR<seqR.length && iM<seqM.length && (colR<gapped3dR.length || colM<gapped3dM.length) ) {
                final boolean isLetterR=is(LETTR,gapped3dR,colR++);
                final boolean isLetterM=is(LETTR,gapped3dM,colM++);
                int countXR=0, countXM=0;
                if (isLetterR) while(iR<seqR.length && !('A'<=seqR[iR] && seqR[iR]<='Z')) {
                        if (alreadyFoundLetterR) {
                            if (outR!=null) outR[oR]=seqR[iR];
                            oR++;
                            countXR++;
                        }
                        iR++;
                    }
                if (isLetterM) while(iM<seqM.length && !('A'<=seqM[iM] && seqM[iM]<='Z')) {
                        if (alreadyFoundLetterM) {
                            if (outM!=null) outM[oM]=seqM[iM];
                            oM++;
                            countXM++;
                        }
                        iM++;
                    }
                for(int i=countXM-countXR; --i>=0;) {
                    if (outR!=null) outR[oR]=' ';
                    oR++;
                }
                for(int i=countXR-countXM; --i>=0;) {
                    if (outM!=null) outM[oM]=' ';
                    oM++;
                }

                if (iR<seqR.length) {
                    if (outR!=null) outR[oR]=isLetterR ? seqR[iR] : (byte)' ';
                    oR++;
                }
                if (iM<seqM.length) {
                    if (outM!=null) outM[oM]=isLetterM ? seqM[iM] : (byte)' ';
                    oM++;
                }
                if (isLetterR) { alreadyFoundLetterR=true; iR++; }
                if (isLetterM) { alreadyFoundLetterM=true; iM++; }
            }
            if (outR==null) {
                outR=new byte[oR];
                outM=new byte[oM];
            }
            else break;
        }

        return new byte[][]{outR,outM};
    }

    public static int[] gaps2Columns(int gg[], int toIndex) {
        final int cc[]=new int[gg.length], to=mini(gg.length,toIndex);
        int c=0;
        for(int i=0; i<to; i++) {
            c+=gg[i];
            cc[i]=c++;
        }
        return cc;
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Convert idx to columns  >>> */
    public static int[] letterIdx2column(byte text[]) {
        int indices[]=new int[countLetters(text,0,MAX_INT)];
        for(int i=0,iLetter=0;i<text.length;i++) {
            final int  b=text[i];
            if ('a'<=b && b<='z' || 'A'<=b && b<='Z' ) indices[iLetter++]=i;
        }
        return indices;
    }

    public static int[] column2letterIdx(byte text[]) {
        final int n=text.length;
        int indices[]=new int[n];
        Arrays.fill(indices,-1);
        final boolean charClass[]=charClass(LETTR);
        for(int i=0,iLetter=0;i<n;i++) {
            if ( text[i]>0 && charClass[text[i]]) indices[i]=iLetter++;
        }
        return indices;
    }
    public static int column2letterIdx(byte seq[],int from) {
        int count=0;
        for(int i=mini(from,seq.length);--i>=0;) {
            final int c=seq[i]|32;
            if ('a'<=c && c<='z') count++;
        }
        return count;
    }

 /* <<< Convert idx to columns <<< */
    /* ---------------------------------------- */
    /* >>> Convert idx to columns  >>> */

    /*
      Prefix "t" means templeate
      Prefix "m" means the gapped sequence to be modified
    */
    public static int[] useGapsFromExternalAlignment(byte[] tGapped, byte[] mGapped, byte[] tResidues, byte[] mResidues,  int tGaps[], int mGaps[]) {
        final int[] tPos=gaps2Columns(tGaps,MAX_INT);
        final int[] mPos=gaps2Columns(mGaps,MAX_INT);

        int tIdx=maxi(0,idxOfLetters(tGapped,tResidues,0, MAX_INT));
        int mIdx=maxi(0,idxOfLetters(mGapped,mResidues,0, MAX_INT));

        for(int col=0; col<tGapped.length; col++) {
            final boolean tLetter=is(LETTR,tGapped, col);
            final boolean mLetter=is(LETTR,mGapped, col);
            if (tLetter && mLetter) {
                //putln(" tIdx="+tIdx+" mIdx="+mIdx+" Correctur="+(sumOf(mGaps,mIdx)+mIdx));
                mGaps[mIdx]+=  tPos[tIdx] - (sumOf(mGaps,mIdx+1)+mIdx);
                if (mGaps[mIdx]<0) mGaps[mIdx]=0;
            }
            if (tLetter) tIdx++;
            if (mLetter) mIdx++;
        }
        for(int iA=mIdx; iA<mResidues.length; iA++) {
            mGaps[iA]+=  mPos[iA] - (sumOf(mGaps,iA+1)+iA);
            if (mGaps[iA]<0) mGaps[iA]=0;
        }
        return mGaps;
    }

    public static byte[][] gappedForPerfectMatch(byte[]...ungapped) {
        if (ungapped.length!=2) return null;
        final int iNeedle, iHaystack;
        if (ungapped[0].length<ungapped[1].length) { iNeedle=0; iHaystack=1; } else {iNeedle=1; iHaystack=0;}
        final int idx=strstr(STRSTR_IC,ungapped[iNeedle],ungapped[iHaystack]);
        byte gapped[][]=null;
        if (idx>0) {
            final byte[] seq=new byte[ungapped[iHaystack].length];
            System.arraycopy(ungapped[iNeedle],0, seq, idx, ungapped[iNeedle].length);
            Arrays.fill(seq,0,idx,(byte)' ');
            gapped=new byte[2][];
            gapped[iHaystack]=ungapped[iHaystack];
            gapped[iNeedle]=seq;
            return gapped;
        }
        return gapped;

    }

  /*
    public static boolean isLocal(byte[][] aliLocal, byte[][] aliGlobal) {
        if (aliLocal.length!=aliGlobal.length) {
            debug("aliLocal.length="+aliLocal.length+" aliGlobal.length="+aliGlobal.length+"\n");
            stckTrc();
        }
        for(int i=mini(aliLocal.length,aliGlobal.length); --i>=0;) {
            if (countLetters(aliLocal[i],0,MAX_INT)!=countLetters(aliGlobal[i],0,MAX_INT)) return true;
        }
        return false;
    }
  */
    /* <<<   <<< */
    /* ---------------------------------------- */
    /* >>>Trim >>> */
    public static boolean atLeastNumLettersInColumn(int num, int col, byte[][] ali) {
        int count=0;
        for(byte row[] : ali) {
            if (is(LETTR,row,col) && ++count>=num) return true;
        }
        return false;
    }

    public static byte[][] trimAlignment(int numLetters, byte[][] ali) {
        if (ali==null) return null;
        final int L=maxSze(ali);
        int from=0, to=L;
        while(from<L && !atLeastNumLettersInColumn(numLetters, from,ali)) from++;
        while(to>=0 && !atLeastNumLettersInColumn(numLetters,  to-1,ali)) to--;
        final byte[][] neu=new byte[ali.length][];
        for(int i=neu.length; --i>=0;) {
            neu[i]=substrgB(ali[i],from,to);
        }
        return neu;

    }
    public final static long PRINT_ALI_IDENT_LETTER=1<<1, PRINT_ALI_TRUNCATE=1<<2;
    public static void aliWithMidline(long options, byte[] gg1, int start1, String header1, byte[] gg2, int start2, String header2, int width, ByteArray ba) {

        final int toCol1=prev(LETTR,gg1,MAX_INT,-1), toCol2=prev(LETTR,gg2,MAX_INT,-1);
        final int toCol=0!=(options&PRINT_ALI_TRUNCATE)?mini(toCol1,toCol2) : maxi(toCol1,toCol2);
        final int fromCol=0!=(options&PRINT_ALI_TRUNCATE)?maxi(nxt(LETTR,gg1,0,MAX_INT), nxt(LETTR,gg2)) : 0;

        final int hL1=sze(header1), hL2=sze(header2), hL=maxi(hL1,hL2)+1;
        final int numL=start1==MIN_INT?0:maxi(stringSizeOfInt(start1+toCol),stringSizeOfInt(start2+toCol));
        int num1=countLetters(gg1,0,fromCol)+start1, num2=countLetters(gg2,0,fromCol)+start2;

        for(int fromC=fromCol; fromC<toCol; fromC+=width) {
            for(int j=0;j<3;j++) {

                if (ba!=null) {
                    if (hL1+hL2>0) {
                        if (j==0) ba.append(header1).append(' ', hL-hL1);
                        if (j==1) ba.append(' ', hL);
                        if (j==2) ba.append(header2).append(' ', hL-hL2);
                    }
                    if (numL>0) {
                        if (j==0) ba.append(num1, numL).append(' ');
                        if (j==1) ba.append(' ', numL+1);
                        if (j==2) ba.append(num2, numL).append(' ');
                    }
                }
                puts(" > "+fromC+" "+toCol);
                for(int i=fromC; i<fromC+width && i<toCol; i++)  {

                    final int c=
                        j==0?gg1[i] :
                        j==1 && 0!=(options&PRINT_ALI_IDENT_LETTER) && (gg1[i]|32)==(gg2[i]|32) ? (gg1[i]&~32) :
                        j==1 ? Blosum.BL2SEQ_MIDLINE[gg1[i]&127][gg2[i]&127] :
                        gg2[i];

                    ba.append((char)c);
                    if (j==0 && is(LETTR, gg1, i)) num1++;
                    if (j==2 && is(LETTR, gg2, i)) num2++;
                }
                puts("< ");
                ba.append('\n');
            }
            ba.append('\n');
        }

    }
}
