package charite.christo.protein;
import charite.christo.*;
import charite.christo.interfaces.*;
import java.awt.Component;
import javax.swing.*;
import java.io.File;
import java.util.*;
import static charite.christo.protein.Protein3dUtils.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static java.awt.event.MouseEvent.*;
public class ChAstex implements ProteinViewer, ChProcessEvent, astex.MoleculeRendererListener, ChRunnable {
    /*
      http://www.mattryall.net/blog/2007/03/composition-stream-concatenation#comment935
      InputStream FileInputStream

      GUI:

      Select Ligands-menu: Tell the name of the Ligand
      Save View: Tell that run script. And define default file extension like .astexScript.

     */
    /* ---------------------------------------- */
    /* >>> Conversions >>> */
    public astex.MoleculeRenderer getMoleculeRenderer() {
        final astex.MoleculeViewer mv=getAstex();
        return mv!=null ? mv.getMoleculeRenderer() : null;
    }
    private astex.Molecule getMolecule() {
        final astex.MoleculeRenderer mr=getMoleculeRenderer();
        return mr!=null ? mr.getMolecule(this.proteinName) : null;
    }
    private astex.Residue getResidue(int rN, char rI, char rC) {
        final astex.Molecule mol=getMolecule();
        final Protein p=getProtein();
        if (mol==null || p==null) return null;
        for(int iC=mol.getChainCount(); --iC>=0;) {
            final astex.Chain chain=mol.getChain(iC);
            if (rC==0||chrAt(0,chain.getName())==rC) {
                for(int iR=chain.getResidueCount(); --iR>=0;) {
                    final astex.Residue res=chain.getResidue(iR);
                    if (res.getNumber()==rN && (rI==0||res.getInsertionCode()==rI)) return res;
                }
            }
        }
        return null;
    }
    /* <<< Conversions <<< */
    /* ---------------------------------------- */
    /* >>> Protein >>> */
    private Protein protein;
    private String proteinName;
    public Protein getProtein() { return protein;}
    public boolean setProtein(long options, Protein p, ProteinViewer inSameView) {
        if (this.protein!=null) return false;
        final File file=p.getFileMayBeWithSideChains();
        if (file==null) return false;
        this.protein=p;
        this.proteinName=p.getName();
        final ChAstex inSameViewPO= (ChAstex) ( inSameView instanceof ChAstex ? (ChAstex)inSameView : inSameView instanceof ChAstexPROXY ? ((ChAstexPROXY)inSameView).pv() : null);
        astex.MoleculeViewer moleculeViewer=inSameViewPO!=null ? inSameViewPO.getAstex() : null;
        if (moleculeViewer==null) {
            moleculeViewer=new astex.MoleculeViewer();
            moleculeViewer.setApplication(true);
            final String rplc[][]={
                {"Exit"},
                {"Save Molecule"},
                {"Save All"},
                {"Run Script...","Load view / Run Script..."},
                {"Write BMP...", "Write image in BMP format"}
            };
            final JMenuBar mb=menuBar2j(moleculeViewer.createMenuBar(), rplc);
            this.shared_menubar=weakRef(mb);
            mb.add(pnl());
            final ChButton b=new ChButton("SHW_OBJECT",ChButton.NO_BORDER).t(" Surfaces and ribbons ...").li(LI);
            mb.add(pnl(b));
            moleculeViewer.setUsePopupMenu(true);
            this.shared_refPanel=weakRef(moleculeViewer);
            pcp("ChAstex$$MB_NO_GC", mb,moleculeViewer);
            this.shared_vView=new UniqueList(ChAstex.class);
            this.shared_ref_astexMV=weakRef(moleculeViewer);
        } else {
            this.shared_menubar=inSameViewPO.shared_menubar;
            this.shared_ref_astexMV=inSameViewPO.shared_ref_astexMV;
            this.shared_refPanel=inSameViewPO.shared_refPanel;
            this.shared_vView=inSameViewPO.shared_vView;
        }
        getMoleculeRenderer().addMoleculeRendererListener(this);
        adUniq(this, shared_vView);
        // getMoleculeRenderer().addMolecule(file.toString(), p.getName());
        astex.Molecule molecule=null;

        final File ff[];
        {
            final List<File>vH=new ArrayList();
            vH.add(file);
            for(HeteroCompound h : p.getDnaAndRnaAndHeteros()) adUniq(h.getFile(), vH);
            ff=toArry(vH, File.class);
        }
        final File fileLoad;
        if (ff.length>1) {
            int size=0;
            for(File f : ff) size+=sze(f)+1;
            final ByteArray ba=new ByteArray(size);
            for(File f : ff) ba.append(readBytes(f)).append('\n');
            wrte(fileLoad=createTempFile(".pdb"), ba);
        } else fileLoad=file;
        try {
            molecule=astex.MoleculeIO.readPDB(new astex.FILE(ChUtils.inStream(fileLoad.toString(), STREAM_UNZIP)));
        } catch(Exception ex) {stckTrc(ex); return false; }
        if (molecule==null) return false;
        molecule.setSymmetry(null);
        molecule.setFilename(file.toString());
        molecule.setName(p.getName());
        getMoleculeRenderer().addMolecule(molecule);
        final Component canvas=deref(shared_refPanel, Component.class);
        LI.addTo("mw",canvas);
        return true;
    }
    /* <<< Protein <<< */
    /* ---------------------------------------- */
    /* >>> Disposable >>> */
    private boolean _disposed;
    public void dispose() {
        if (!_disposed) {
            _disposed=true;
            final UniqueList v=this.shared_vView;
            interpret(0L, "molecule remove '"+this.proteinName+"';");
            v.remove(this);
            AbstractProxy.disposeProxy(this);
        }
    }
    /* <<< Disposable <<< */
    /* ---------------------------------------- */
    /* >>> CommandInterpreter >>> */
    private final static String supported[]={
        COMMANDselect, COMMANDselection_name,
        COMMANDset_rotation_translation,
        COMMANDzoom,COMMANDrotate,COMMANDcenter, COMMANDbackground,
        COMMANDspheres,COMMANDlines,COMMANDsticks,COMMANDribbons,COMMANDcolor,
        COMMANDsurface, COMMANDsurface_color,
        COMMANDchange_object_color,
        COMMANDcartoon, COMMANDlabel,
        COMMANDobject_delete,
        COMMANDcenter_amino
    };
    public void interpret(long options, String command0) {
        final Selection3D atoms[]=this.selectedResidues;
        final boolean hasSelection=sze(atoms)>0;
        final String id=makeSelId(atoms, this.selection_id);
        final String command=command0.trim();
        log3D(true,this,command);
        final astex.MoleculeRenderer mr=getMoleculeRenderer();
        final astex.Renderer r=mr!=null ? mr.renderer:null;
        final String word0=toStrgIntrn(frstTkn(command));
        final int L0=sze(word0);
        final int para=nxtE(NO_SPC,command,L0,999);

        String objectName=_objectName=null;

        if (r==null || L0==0) return;
        String send=idxOfStrg(word0, allCommands())>=0 ? null : rplcToStrg(RESIDUES,id, rplcToStrg(PROTEIN,this.proteinName,command));

        if (supports(word0, this)) {
            final boolean off=command.endsWith(" off");
            final String
                group=hasSelection ? " ( group '"+id+"' ) " : " ( molecule \""+this.proteinName+"\" and aminoacid ) ",
                objectPfx= word0==COMMANDribbons ? "ribbon_" : word0==COMMANDcartoon ? "schematic_" :  word0==COMMANDsurface ? PFX_SURFACE_OBJECT : null;
            if (objectPfx!=null) {
                objectName=objectPfx+this.proteinName+"_"+(hasSelection ? selectionToText(SURFACE_ID|SURFACE_ID_COLON|SURFACE_ID_KOMMA|SURFACE_ID_DOT|SURFACE_ID_DASH,atoms):id);
                if (!off) addRemoveObject(true,objectName,false);
                putln("objectName="+objectName +"   off="+off);
            }

            final long color=rgbInScript(command);
            if (objectName!=null && off) send=addRemoveObject(false,objectName,true);
            else if (word0==COMMANDobject_delete) send=addRemoveObject(false, objectIdInScript(command),true);
            else if (word0==COMMANDset_rotation_translation) { transform(matrixInScript(command)); send="";}
            else if (word0==COMMANDbackground) send="background 0x"+toHex(color,6,null)+";\n";
            else if (word0==COMMANDzoom) r.applyZoom((100.0-atoi(command,L0))/100);
            else if (word0==COMMANDsurface_color) { surface_color=color;   send=null;}
            else if (word0==COMMANDribbons) send="schematic -name '"+objectName+"' -alltube  true  -ribbonwidth 20 "+group+";\n";
            else if (word0==COMMANDchange_object_color) {
                final String oid=objectIdInScript(command);
                send="object \""+oid+"\" colour 0x"+rgbHex((int)(color&0xFFffFF))+";\n";
                if ( (color>>24)!=0xFF) send+="object \""+oid+"\" transparency "+((int) (color>>24))+";\n";
            }
            else if (word0==COMMANDsurface) {
                long c=surface_color;
                if (c==-1) c=0xFFff00ff;
                send="surface -solid true '"+objectName+"' 0x"+rgbHex((int)(c&0xFFffFF))+" "+group+";\n";
                if ( (c>>24)!=0xFF) send+="object \""+objectName+"\" transparency "+((int) (c>>24))+";\n";
            }
            else if (word0==COMMANDcenter_amino) send="center (aminoacid );\n";
            else if (word0==COMMANDcartoon) send="schematic -name '"+objectName+"' "+group+";\n";
            else if (word0==COMMANDcenter) {
                send="center  ( sphere 20 around "+group+" ); ";
                ChDelay.repaint(getMoleculeRenderer(), 333);
            }
            else if (word0==COMMANDlabel) {
                send="label "+(off?"clear": "\""+delSfx('"', delPfx('"', command.substring(para)))+"\"")+" "+group+";\n";
            } else if (word0==COMMANDcolor)  {
                /* was ist mit alpha anteil? */
                send="color 0x"+toHex(color, 6, null)+group+";\n";
            } else if (word0==COMMANDrotate) {
                final char axis=chrAt(para,command);
                final double f=atof(command,para+1);
                if (axis=='x') r.rotateX(f);
                if (axis=='y') r.rotateY(f);
                if (axis=='z') r.rotateZ(f);
            } else if (word0==COMMANDselect) {
                send=selectCmd(options, command.substring(para).getBytes());

            } else if (word0==COMMANDselection_name) {
                next_selection_id=command.substring(para);
                send="";
            }
            final String s=
                word0==COMMANDspheres?"spheres":
                word0==COMMANDlines?"lines":
                word0==COMMANDsticks?"sticks":
                null;
            if (s!=null) send="display "+s+(off?" off":" on")+group+";\n";
        }
        if (sze(send)>0) {
            log3D(false,this,send);
            mr.executeInternal(send);
            if (word0==COMMANDcolor) mr.executeInternal("select ( none  );");

        }
        _objectName=objectName;
        ChDelay.runAfterMS(EDT,threadRepaint,555);
    }
    /* <<< CommandInterpreter <<< */
    /* ---------------------------------------- */
    /* >>> ResidueSelection >>> */
    private Selection3D selectedResidues[];
  private String selection_id, next_selection_id;

    private String selectCmd(long options, byte[] expr) {
        final Selection3D atoms[]=Protein3dUtils.toSelectionT(expr, getProtein());
        this.selectedResidues=atoms;
        final String sel=selectionToText(ASTEX,atoms);
        final StringBuffer sb=new StringBuffer(99).append("select ( molecule \"").append(this.proteinName).append('"');
        if (nxt(NO_SPC,sel)>0) sb.append(" and ").append(sel);
        sb.append(" ) ");
        if (0!=(options&SET_SELECTED_RESIDUES_HIGHLIGHT)) {
            highlightCalpha(atoms,"D");
            ChDelay.runAfterMS(EDT, thread_highlightCalpha(atoms,"B"),222);
        }
        final String id=makeSelId(atoms, this.selection_id=this.next_selection_id);
        this.next_selection_id=null;
        sb.append(";\ndefine ").append(id).append(" current;");
        return sb.append(';').toString();
    }
    private Thread thread_highlightCalpha(Selection3D atoms[],String type) {
        return threadM("highlightCalpha",this, new Object[]{atoms,type});
    }
    public void highlightCalpha(Selection3D[] atoms, String type) {
        for(int iA=sze(atoms); --iA>=0;) {
            final Selection3D a=atoms[iA];
            if (a==null) continue;
            for(int rn=a.getFirst(); rn<=a.getLast(); rn++) {
                final astex.Residue res=getResidue(rn, a.getInsertion(), a.getChain());
                if (res!=null) {
                    for(int i=res.getAtomCount(); --i>=0;){
                        final astex.Atom atom=res.getAtom(i);
                        if ("CA".equals(atom.getAtomLabel())) sphere(atom,type);
                    }
                }
            }
        }
    }
    private void sphere(astex.Atom atom, String type) {
        final Component canvas=deref(shared_refPanel, Component.class);
        final java.awt.Graphics g=canvas!=null ? canvas.getGraphics() : null;
        if (atom!=null && g!=null) {
            final astex.MoleculeRenderer mr=getMoleculeRenderer();
            final astex.Renderer r=mr!=null ? mr.renderer :null;
            if (r!=null) {
                g.setColor(C(0xFF8080));
                final int x=atom.xs>>astex.Renderer.FixedBits;
                final int y=atom.ys>>astex.Renderer.FixedBits;
                final int R=EX;
                if(type=="D") g.fillOval(x-R,y-R,2*R,2*R);
                if(type=="B") ChIcon.drawCursor(g, x-R,y-R,2*R,2*R, canvas);

            }
        }
    }

    /* <<< ResidueSelection <<< */
    /* ---------------------------------------- */
    /* >>> Ribbons Cartoons Schematic >>> */

    private long surface_color=0xFFff00ff;
    private String _objectName;
    private final List<String> vObjects=new ArrayList();
    private String addRemoveObject(boolean add, String ribbon, boolean rmFromList) {

        if (add) adUniq(ribbon, vObjects);
        else {
            if (rmFromList)vObjects.remove(ribbon);
            return "object remove '"+ribbon+"';";
        }
        new StrapEvent(this, StrapEvent.PROTEIN_VIEWER_SURFACES_CHANGED).run();
        return "";
    }

    /* <<< Objects <<< */
    /* ---------------------------------------- */
    /* >>> Viewer Panel >>> */
  private Object shared_refPanel,  shared_ref_astexMV, shared_menubar;
    private UniqueList<ChAstex> shared_vView;
    private astex.MoleculeViewer getAstex() { return deref(shared_ref_astexMV, astex.MoleculeViewer.class); }
    public ChAstex[] getViewersSharingView() {
        final UniqueList<ChAstex> v=this.shared_vView;
        return v!=null ?  v.asArray() : null;
    }

    private final static Object
    FLAGS=new Long(PROPERTY_NAMED_SELECTIONS|PROPERTY_MULTI_MOLECULE | PROPERTY_MULTI_VIEW | PROPERTY_HEAVY_WEIGHT |
                   PROPERTY_RIBBON_COLOR_NO_CHANGE | PROPERTY_NEEDS_UPDATE_SURFACES_WHEN_ROTATED),
        MASK_SEL[]={null, new Long(SHIFT_MASK|ALT_MASK)};
    public void setProperty(String id, Object value) {}
    public Object getProperty(String getID) {
        if (getID==GET_AWTMASK_DISC_SELECTION) {
            if (MASK_SEL[0]==null) MASK_SEL[0]=new Long(isMacintosh() ? shortCutMask() : CTRL_MASK);
            return MASK_SEL[0];
        }
        if (getID==GET_AWTMASK_CONT_SELECTION) return MASK_SEL[1];
        if (getID==GET_ATOM_SELECTION_EXAMPLE) return "atom CA CB";
        if (getID==GET_SUPPORTED_COMMANDS) return supported;
        if (getID==GET_FLAGS) return FLAGS;
        if (getID==GET_CANVAS) return deref(shared_refPanel);
        if (getID==GET_JMENUBAR) return deref(shared_menubar);
        if (getID==GET_SURFACEOBJECTS) return vObjects;
        if (getID==GET_LAST_CREATED_OBJECT) return _objectName;
        return null;
    }

    /* <<< Viewer Panel <<< */
    /* ---------------------------------------- */
    /* >>> AWTEvent >>> */
    private final ChEventAdapter LI=new ChEventAdapter(this);
    public java.awt.AWTEvent processEvent(JComponent component, java.awt.AWTEvent ev) {
        final int id=ev.getID();

        final astex.MoleculeRenderer mr=getMoleculeRenderer();
        final Protein p=getProtein();
        if (mr==null || p==null) return null;
        final int x=x(ev), y=y(ev);
         if (id==MOUSE_CLICKED) {
            final astex.Atom atom=mr.getNearestAtom(x, y);
            final astex.Molecule molecule=getMolecule();
            final astex.Residue residue=atom==null?null:atom.getResidue();
            if (residue!=null && residue.isStandardAminoAcid() && atom.getMolecule()==molecule) {
                final astex.Chain chain = residue.getParent();
                final char chainChar=chrAt(0,chain!=null ? chain.getName() : null);
                final char insCode= residue.getInsertionCode();

                Protein3dUtils.dispatchEvtPicked(this, residue.getNumber(), chainChar, is(LETTR_DIGT,insCode)?insCode:(char)0);
            }
        }

        mr.repaint();
        return ev;
    }

    /* <<< AWTEvent <<< */
    /* ---------------------------------------- */
    /* >>> Matrix3D >>> */
    //private final static WeakHashMap<astex.Molecule, float[]> mapMolecule2xyz=new WeakHashMap();
    private float[] xyzORIG;
    private void transform(Matrix3D m3d) {
        final astex.MoleculeRenderer mr=getMoleculeRenderer();
        final astex.Molecule molecule=mr!=null ? mr.getMolecule(this.proteinName) : null;
        if (molecule==null) return;
        final int N=molecule.getAtomCount();
        if (xyzORIG==null) {
            if (m3d==null) return;
            this.xyzORIG=new float[N*3];
            for(int i=mini(N,xyzORIG.length/3); --i>=0;) {
                final astex.Atom atom=molecule.getAtom(i);
                for(int j=3;--j>=0;) xyzORIG[i*3+j]=(float)atom.get(j);
            }
        }
        assertTrue(xyzORIG.length==3*N);
        final float[] xyz;
        if (m3d==null || m3d.isUnit()) xyz=xyzORIG;
        else {
            xyz=new float[3*N];
            m3d.transformPoints(xyzORIG, xyz, N);
        }
        for(int i=mini(N,xyz.length/3); --i>=0; ) {
            final astex.Atom atom=molecule.getAtom(i);
            final int i3=3*i;
            atom.set(xyz[i3],xyz[i3+1],xyz[i3+2]);
        }
        interpret(0L, "center  ( sphere 20 around ( molecule '"+this.proteinName+"' ) ); ");
        mr.repaint();
    }
    public void moleculeAdded(astex.MoleculeRenderer renderer, astex.Molecule molecule){}
    public void genericAdded(astex.MoleculeRenderer renderer, astex.generic.Generic generic){}
    public void genericRemoved(astex.MoleculeRenderer renderer, astex.generic.Generic generic){}
    public void mapAdded(astex.MoleculeRenderer renderer, astex.Map map){}
    public void mapRemoved(astex.MoleculeRenderer renderer, astex.Map map){}
    public void atomSelected(astex.MoleculeRenderer renderer, astex.Atom atom){}
    public void moleculeRemoved(astex.MoleculeRenderer renderer, astex.Molecule molecule){
        if (molecule!=null && ChAstex.this.proteinName.equals(molecule.getName())) ChAstex.this.dispose();
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */
    private final Thread threadRepaint=ChDelay.newThread(this,"REPAINT",null);
    public Object run(String id, Object arg) {
        if (id=="REPAINT") {
            final astex.MoleculeRenderer mr=getMoleculeRenderer();
            if (mr!=null)  mr.repaint();
            final Component canvas=deref(shared_refPanel, Component.class);
            if (canvas!=null) canvas.repaint();
        }
        return null;
    }
}
/*
MoleculeRendererListener#atomSelected should be atomPicked
Atom#x  should be private
schematic should move with protein
remove group
how to list groups
how to draw cursor

java.lang.NegativeArraySizeException
	at astex.Molecule.connect2(Molecule.java:836)
	at astex.MoleculeIO.readPDB(MoleculeIO.java:1019)

*/
