package charite.christo.protein;
import charite.christo.*;
import charite.christo.interfaces.*;
import java.awt.*;
import java.awt.event.MouseEvent;
import java.io.File;
import javax.swing.*;
import java.util.List;
import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static charite.christo.protein.Protein3dUtils.*;
import static java.awt.event.MouseEvent.*;
import org.jmol.api.JmolAdapter;
import org.jmol.api.JmolCallbackListener;
import org.jmol.api.JmolViewer;
import org.jmol.viewer.JmolConstants;

import org.openscience.jmol.app.jmolpanel.AppConsole;
import org.jmol.viewer.Viewer;
/**
   <b>JMol Home:</b> http://jmol.sourceforge.net/<br>
   <b>License:</b> WIKI:LGPL
   @author Christoph Gille
   svn co https://jmol.svn.sourceforge.net/svnroot/jmol/trunk/Jmol
   ant
   ant jar

*/
public class ChJmol implements ProteinViewer, ChRunnable, PaintHook, JmolCallbackListener {
    /* <<< Objects <<< */
    /* ---------------------------------------- */
    /* >>> Viewer Panel >>> */
    private ChFrame shared_appConsole[];
    private float shared_zoom[];
    private Object shared_panel, shared_viewer, shared_adapter;
    private UniqueList<ChJmol> shared_vView;
    private final static Object
    FLAGS=new Long(PROPERTY_MULTI_MOLECULE | PROPERTY_MULTI_VIEW | PROPERTY_HAS_SCRIPT_PANEL | PROPERTY_HAS_CONTEXTMENU),
        MASK_SEL[]={new Long(CTRL_MASK), new Long(SHIFT_MASK)};

    public void setProperty(String id, Object value) {}
    public Object getProperty(String getID) {
        if (getID==GET_AWTMASK_DISC_SELECTION) return MASK_SEL[0];
        if (getID==GET_AWTMASK_CONT_SELECTION) return MASK_SEL[1];
        if (getID==GET_SUPPORTED_COMMANDS) return supported;
        if (getID==GET_ATOM_SELECTION_EXAMPLE) return " *.CA or *.CB";
        if (getID==GET_FLAGS) return FLAGS;
        if (getID==GET_CANVAS) return deref(shared_panel);
        if (getID==GET_SURFACEOBJECTS) return vObjects;
        if (getID==GET_LAST_CREATED_OBJECT) return _objectName;
        return null;
    }
    /* <<< Viewer Panel <<< */
    /* ---------------------------------------- */
    /* >>> Protein >>> */
    private Protein  protein;
    private String filePath, proteinName, rgbHex;
    public Protein getProtein() { return protein;}

    public boolean setProtein(long options, Protein p, ProteinViewer inSameView) {
        if (this.protein!=null) return false;
        this.proteinName=p.getName();
        final File file=p.getFileMayBeWithSideChains();
        if (0==sze(file)) return false;
        this.protein=p;
        final ChJmol inSameViewPO= (ChJmol) ( inSameView instanceof ChJmol ? (ChJmol)inSameView : inSameView instanceof ChJmolPROXY ? ((ChJmolPROXY)inSameView).pv() : null);
        org.jmol.viewer.Viewer viewer=inSameViewPO!=null ? inSameViewPO.getJmolViewer() : null;
        if (viewer==null) {
            this.shared_vView=new UniqueList(ChJmol.class);
            final JmolAdapter adapter=new org.jmol.adapter.smarter.SmarterJmolAdapter();
            this.shared_adapter=weakRef(adapter);
            final JComponent pnl=new ChPanel() {
                     public void processEvent(final AWTEvent ev0) {
                        AWTEvent ev=ev0;
                        final int m=modifiers(ev), m1=m&MASK_KEYMODIFIERS, id=ev.getID();
                        if (ev0 instanceof MouseEvent && (m1==SHIFT_MASK || m1==CTRL_MASK) &&
                            (id==MOUSE_PRESSED || id==MOUSE_MOVED || id==MOUSE_CLICKED)) {
                            final MouseEvent mev=(MouseEvent)ev0;
                            final int mNeu=m & ~(CTRL_MASK|SHIFT_MASK);
                            ev=new MouseEvent((Component)ev.getSource(), id, mev.getWhen(), mNeu, x(ev), y(ev), mev.getClickCount(),false,BUTTON2);
                            pcp(KEY_EVT_MODI,intObjct(m),this);
                        } else pcp(KEY_EVT_MODI,null, this);
                        super.processEvent(ev);
                    }
                };
            addPaintHook(this, pnl);
            this.shared_panel=weakRef(pnl);
            viewer=(Viewer)JmolViewer.allocateViewer(pnl, adapter);
            this.shared_viewer=weakRef(viewer);
            viewer.setJmolCallbackListener(this);
            this.shared_appConsole=new ChFrame[1];
            this.shared_zoom=new float[]{100};
            //org.jmol.popup.JmolPopup jmp=org.jmol.popup.JmolPopup.newJmolPopup(viewer, false, "menu", true);
            for(java.awt.event.MouseWheelListener wl : pnl.getListeners(java.awt.event.MouseWheelListener.class)) {
                pnl.removeMouseWheelListener(wl);
            }
            final Container c=(Container)viewer.getProperty("DATA_API","getPopupMenu",null);
            setUndockble(true,c);
        } else {
            this.shared_adapter=inSameViewPO.shared_adapter;
            this.shared_panel=  inSameViewPO.shared_panel;
            this.shared_viewer= inSameViewPO.shared_viewer;
            this.shared_vView=  inSameViewPO.shared_vView;
            this.shared_zoom=inSameViewPO.shared_zoom;
            this.shared_appConsole=inSameViewPO.shared_appConsole;
        }
        viewer.loadModelFromFile(null, this.filePath=toStrg(file) , null, null, true, null, null, 0);
        interpret(0L, "model all; center protein");
        this.shared_vView.add(this);
        /*
        org.jmol.util.Logger.setLogger(new ChJmol_logger());
        org.jmol.util.Logger.setActiveLevel(org.jmol.util.Logger.LEVEL_MAX, true);
        */
        return true;
    }
    /* <<< Protein <<< */
    /* ---------------------------------------- */
    /* >>> ResidueSelection >>> */
    public void highlightCalpha(Selection3D[] atoms, String type) {
        final Viewer viewer=getJmolViewer();
        final String iModel=getModelNumber();
        final StringBuffer sb=new StringBuffer();
        for(int iA=sze(atoms); --iA>=0;) {
            final Selection3D a=atoms[iA];
            if (a==null) continue;
            for(int rn=a.getFirst(); rn<=a.getLast(); rn++) {
                //final BitSet bsAtoms = viewer.getAtomBitSet("atomno="+iAtom+" and "+iModel);
                sb.setLength(0);
                sb.append(rn);
                if (a.getInsertion()!=0) sb.append('^').append(a.getInsertion());
                sb.append(".CA and ").append(iModel);
                final BitSet bsAtoms = viewer.getAtomBitSet(sb.toString());
                final int atomIndex=bsAtoms.nextSetBit(0);
                if (atomIndex<0) return;
                final javax.vecmath.Point3f p3f=viewer.getAtomPoint3f(atomIndex);
                final javax.vecmath.Point3i p3i=p3f!=null ? viewer.transformPoint(p3f) : null;
                if (p3i==null) return;
                int d=16;
                final JComponent canvas=deref(shared_panel, JComponent.class);
                ChIcon.drawCursor(canvas.getGraphics(), p3i.x-d/2,p3i.y-d/2,d,d, canvas);
            }
        }
    }
    private Selection3D selectedResidues[];
    private String selection_id, selection_Expr;
    private String selectCmd(long options, String expr) {
        final Selection3D atoms[]=Protein3dUtils.toSelectionT(expr, getProtein());
        this.selectedResidues=atoms;
        final String sel=selectionToText(JMOL,atoms);
        final StringBuffer sb=new StringBuffer(99)
            .append(sel).append(" and ").append(getModelNumber()).append('\n');
        if (0!=(options&SET_SELECTED_RESIDUES_HIGHLIGHT)) {
            //ChDelay.runAfterMS(EDT, thread_highlightCalpha(atoms,"B"),222);
            //highlightCalpha(atoms,"D");
        }
        return sb.toString();
    }
    /* <<< ResidueSelection <<< */
    /* ---------------------------------------- */
    /* >>> ug workaround  >>> */
    public Object run(String id,Object arg) {
        return null;
    }

    /* <<< bug workaround <<< */
    /* ---------------------------------------- */
    /* >>> CommandInterpreter  >>> */
    private final static String
    supported[]={
        COMMANDselect, COMMANDselection_name,
        COMMANDset_rotation_translation,
        COMMANDchange_object_color,
        COMMANDzoom,COMMANDcenter, COMMANDbackground,
        COMMANDspheres,COMMANDsticks,COMMANDribbons,COMMANDcolor,
        COMMANDsurface, COMMANDsa_surface, COMMANDsurface_color,
        COMMANDcartoon, COMMANDlabel, COMMANDdots,
        COMMANDlines,COMMANDrotate, COMMANDobject_delete, COMMANDchange_object_color,
        COMMANDshowScriptPanel,
        COMMANDcenter_amino
    };

    public void interpret(long options, String command0) {
        if (command0==null) return;
        final String iModel=getModelNumber();
        final String command=rplcToStrg(PROTEIN, iModel, command0.trim(), charClass(NO_LETTR_DIGT_US), charClass(NO_LETTR_DIGT_US), 0);
        final Selection3D atoms[]=this.selectedResidues;
        final boolean hasSelection=sze(atoms)>0;
        final String id=makeSelId(atoms, this.selection_id);
        log3D(true,this,command);
        //final String command=command0.trim();
        final String word0=toStrgIntrn(frstTkn(command));
        final int L0=sze(word0);
        if (L0==0) return;
        final int para=nxtE(NO_SPC,command, L0,MAX_INT);
        final boolean off=command.endsWith(" off");
        String send=null;
        final String objectPfx=word0==COMMANDsurface || word0==COMMANDsa_surface ? PFX_SURFACE_OBJECT : null, objectName;

        if (objectPfx!=null) {
            final String pn=rplcChars(charClass(NO_LETTR_DIGT_US), '_', this.proteinName);
            objectName=objectPfx+"_"+pn+"_"+(hasSelection ? selectionToText(SURFACE_ID,atoms):id);
            if (!off) addRemoveObject(true,objectName);
        } else objectName=null;
        _objectName=objectName;
        if (supports(word0, this)) {
            final long color=rgbInScript(command);
            final int alpha=(int)(color>>24);
            final String sOff=off?"off":"on";
            if (word0==COMMANDchange_object_color) {
                send="";
                final String oid=objectIdInScript(command);
                if (oid!=null && oid.startsWith(PFX_SURFACE_OBJECT)) {
                    send="isosurface "+oid+"\n color isosurface \"#"+rgbHex(color)+"\"\n";
                    if (color!=-1 && alpha!=0xFF) send+="isosurface translucent "+mini(10,10-alpha*10/256)+"\n";
                }
            }
            else if (word0==COMMANDsurface || word0==COMMANDsa_surface) {
                send=
                    "frame "+iModel+"\n"+
                    "select "+selectionToText(JMOL,atoms)+"\n"+
                    "isosurface "+objectName+ (word0==COMMANDsa_surface ? " sasurface " : " solvent ")+(off ? sOff : "")+"\n"+
                    "color isosurface \"#"+rgbHex(surface_color)+"\"\n";
                final int sAlpha=(int)((surface_color>>>24)&0xff);
                if (sAlpha!=0xff) send+="isosurface translucent "+mini(10,10-sAlpha*10/256)+"\n";
                send+="frame all;\n";
            }
            else if (objectName!=null && off)  send="";
            else if (word0==COMMANDcenter_amino) send="center amino";
            else if (word0==COMMANDset_rotation_translation) { transform(matrixInScript(command)); send=""; }
            else if (word0==COMMANDobject_delete) send=addRemoveObject(false, objectIdInScript(command));
            else if (word0==COMMANDbackground || word0==COMMANDrotate) send=delPfx("3D_",command);
            else if (word0==COMMANDzoom) {
                float z=this.shared_zoom[0] * atoi(command,para) * .01f;
                if (z<5) z=5; else if (z>200000) z=200000;
                this.shared_zoom[0]=z;
                send="zoom "+ z;
            }
            else if (word0==COMMANDsurface_color) { this.surface_color=color; send=""; }

            else if (word0==COMMANDribbons) {
                send="ribbon "+(off?sOff:"40");
                if (!off && rgbHex!=null) send+="; color ribbon \"#"+rgbHex+"\"\n";
            }
            else if (word0==COMMANDshowScriptPanel) {
                send="";
                ChFrame f=shared_appConsole[0];
                if (f==null) {
                    new AppConsole(getJmolViewer(), f=new ChFrame("Jmol"), AppConsole. ALL_BUTTONS);
                    closeOnCtrlW(true, f, CLOSE_PARENT_WINDOW);
                }
                f.shw(ChFrame.AT_CLICK);
            } else if (word0==COMMANDselect) {
                send="selectionhalo on\nselect "+(this.selection_Expr=selectCmd(options, command.substring(para)));
                selection_id=null;
            } else if (word0==COMMANDselection_name) {
                selection_id=command.substring(para);
                send="";
            }
            else if (word0==COMMANDcolor)   send="color \"#"+(this.rgbHex=rgbHex(color))+"\"";
            else if (word0==COMMANDspheres) send="cpk "+sOff;
            else if (word0==COMMANDcartoon) send="cartoon "+(off ? sOff : "100");
            else if (word0==COMMANDdots)    send="set solvent off; dots "+(off?sOff:"+1.2");
            else if (word0==COMMANDcenter)  send="center "+selection_Expr;
            else if (word0==COMMANDlabel)   send="label "+(off?sOff: "\""+delSfx('"', delPfx('"', command.substring(para)))+"\"");
            else if (word0==COMMANDlines)   send="wireframe "+sOff;
            else if (word0==COMMANDsticks)  send=off? "wireframe off; spacefill off" :"wireframe 0.2; spacefill 25%";
        }
        toJmol(send!=null?send:command);
    }
    private void toJmol(String s) {
        if (sze(s)>0) {
            final Viewer viewer=getJmolViewer();
            if (viewer==null) return;
            assertEDT();
            log3D(false,this,s);
            viewer.scriptWait(s);
            //viewer.evalStringQuiet(send);
        }

    }

    /* <<< CommandInterpreter <<< */
    /* ---------------------------------------- */
    /* >>> Disposable >>>  select 1.1; cartoon off; zap 1.1 */
    private boolean _disposed;
    public void dispose() {
        if (!_disposed) {
            _disposed=true;
            final UniqueList v=this.shared_vView;
            final String iModel=getModelNumber();
            interpret(0L, "select "+iModel+"; cartoon off; zap "+iModel);
            if (v!=null) v.remove(this);
            AbstractProxy.disposeProxy(this);
            if (sze(v)==0) {
                final Viewer vi=getJmolViewer();
                if (vi!=null) {
                    vi.setJmolCallbackListener(null);
                    vi.setModeMouse(JmolConstants.MOUSE_NONE);
                    dispos(DISPOSE_RM_CHILDS, this.shared_appConsole[0]);
                }
            }
        }
    }
    public void mayDeleteJmol() {
        final Viewer viewer=getJmolViewer();
        if (viewer==null) return;
        for(ChJmol jm : getViewersSharingView()) {
            if (jm.getModelNumber()==null) jm.dispose();
        }
    }
    /* <<< Disposable <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */
    public final static ChJmol[] NONE={};
    public ChJmol[] getViewersSharingView() {
        final UniqueList<ChJmol> v=this.shared_vView;
        return v!=null ?  v.asArray() : NONE;
    }

    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Paint >>> */
    private final Dimension currentSize=new Dimension();
    private final Rectangle rectClip=new Rectangle();
    @Override public boolean paintHook(JComponent c, Graphics g, boolean after) {
        if (c==deref(this.shared_panel)) {
            if (!after) {
                final org.jmol.viewer.Viewer viewer=getJmolViewer();
                if (viewer!=null) {
                    c.getSize(currentSize);
                    g.getClipBounds(rectClip);
                    viewer.renderScreenImage(g, currentSize, rectClip);
                    //g.setColor(C(0xFF0000));
                    //g.drawString("This viewer is not yet fully stable in Strap ",EM,charA(g));

                }
                else {
                    g.setColor(C(0xFF));
                    g.fillRect(0,0,9999,9999);
                }
                return false;
            }
            return false;
        } else putln("ChJmol paintHook shared_panel="+deref(this.shared_panel));
        return true;
    }

    /* <<< Paint <<< */
    /* ---------------------------------------- */
    /* >>> Helper methods >>> */
    public Viewer getJmolViewer() {
        return deref(shared_viewer, Viewer.class);
    }
    private String modelNumberDotted;
    private String getModelNumber() {
        final Viewer viewer=getJmolViewer();
        final String path=this.filePath;
        if (viewer!=null && path!=null) {
            for(int i=viewer.getModelCount(); --i>=0;) {
                //     putln("getModelCount "+i+" "+viewer.getModelFileName(i)+" "+viewer.getModelNumberDotted(i));
                String fn=null;
                try { fn=viewer.getModelFileName(i); } catch(Exception ex){}
                if (eqNz(path, fn)) {
                        final String iModel=viewer.getModelNumberDotted(i);
                    if (iModel!=null){
                        if (modelNumberDotted==null) modelNumberDotted=iModel;
                        assertTrue(eq(iModel,modelNumberDotted));
                    }
                    return iModel;
                }
            }
        }
        return modelNumberDotted;
    }
    /* <<< Helper methods <<< */
    /* ---------------------------------------- */
    /* >>> JmolCallbackListener >>> */
    public void setCallbackFunction(String callbackType, String callbackFunction) {
        // putln(ANSI_YELLOW+"setCallbackFunction "+ANSI_RESET+callbackType+"    "+callbackFunction);
    }
    public void notifyCallback(int type, Object[] data) {
        final Viewer viewer=getJmolViewer();
        if (viewer==null) return;
        /*
        if (shared_appConsole[0]==null) {
           final Object console=viewer.getProperty("DATA_API","getAppConsole", null); //AppConsole
            if (console!=null) {
                final JComponent tp=deref(invokeMthd("getTextPane",console), JComponent.class);
                if (tp!=null) {
                    final Window w=parentWndw(tp);
                    shared_appConsole[0]=weakRef(w);
                    putln("w="+w+ "  "+shrtClasNam(tp));
                    closeOnKeystroke(tp, new Object[]{w}, CLOSE_ON_CtrlW);
                    addKeyLi(KLI_ALWAYS_ON_TOP,tp);
                }
            }
        }
        */
        if (type==JmolConstants.CALLBACK_LOADSTRUCT) {
            putln(get(5,data)+ANSI_YELLOW+"notifyCallback "+ANSI_RESET+type+" ",data);
        }

        if (type==JmolConstants.CALLBACK_LOADSTRUCT && JmolConstants.FILE_STATUS_MODELS_DELETED==atoi(get(5,data))) {
            ChDelay.runAfterMS(EDT, threadM("mayDeleteJmol",this, NO_OBJECT),11);
        }

        if (type==JmolConstants.CALLBACK_LOADSTRUCT) {
            final File path=file(data[1]);
            final StrapEvent ev=new StrapEvent(AbstractProxy.getProxy(this,Object.class),StrapEvent.VIEWER3D_MOLECULE_LOADED);
            ChDelay.runAfterMS(EDT,ev.setParameter(path),99);
        }
        if (type==JmolConstants.CALLBACK_PICK) {
            final String strInfo=ChUtils.get(1,data,String.class);
            // putln("strInfo="+strInfo);
            //[GLU]143^A:2.N
            final int
                slash=strInfo.lastIndexOf('/'),
                colon=strInfo.indexOf(':');
            final boolean caret=chrAt(colon-2,strInfo)=='^';
            final int
                numStart=prev(NO_DIGT_DASH,strInfo,  colon-(caret?3:1), -1)+1,
                num=atoi(strInfo,numStart);
            final String iModel=slash<0?null : atoi(strInfo, slash+1)+".1";
            final char chain=chrAt(colon+1, strInfo), ins=caret?chrAt(colon-1,strInfo):0;
            putln("strInfo="+strInfo+"   model="+iModel+" chain="+chain+" num="+num+" ins="+ins);
            for(ChJmol chJm : getViewersSharingView()) {
                final Protein p=chJm.getProtein();
                if (p!=null && (iModel==null || eq(iModel,chJm.getModelNumber()))) {

                    Protein3dUtils.dispatchEvtPicked(chJm, num, chain, ins);
                }
            }
        }
    }
    public boolean notifyEnabled(int callback_pick) { return true; }
    /* <<< JmolStatusListener/JmolCallbackListener <<< */
    /* ---------------------------------------- */
    /* >>> Surfaces >>> */
    private final List<String> vObjects=new ArrayList();
    private String _objectName;
    private long surface_color=0xFFff00ff;

    private String addRemoveObject(boolean add, String ribbon) {
        if (add) adUniq(ribbon, vObjects);
        else {
            vObjects.remove(ribbon);
            return "isosurface "+ribbon+" off";
        }
        new StrapEvent(this, StrapEvent.PROTEIN_VIEWER_SURFACES_CHANGED).run();
        return "";
    }
    /* <<< Surfaces <<< */
    /* ---------------------------------------- */
    /* >>> Matrix3D >>> */
    private boolean alreadyTransformed, showInfoSurf=true;
    private void transform(Matrix3D m3d) {
        final boolean isUnit=m3d==null;
        if (isUnit && !alreadyTransformed) return;
        final Viewer viewer=getJmolViewer();
        if (viewer==null) return;
        final String iModel=getModelNumber();
        final StringBuffer sb=new StringBuffer(333)
            .append("select ").append(iModel).append('\n')
            .append(alreadyTransformed ? "restore":"save").append(" coordinates X").append(hashCode()).append('m')
            .append(atoi(iModel)).append('\n');
        if (!isUnit) {
            sb.append(m3d.toText(Matrix3D.FORMAT_JMOL,null)).append("rotate SELECTED @r\n translate SELECTED @t\n");
        }
        toJmol(sb.toString());
        alreadyTransformed=true;
        if (showInfoSurf && sze(vObjects)>0) {
            showInfoSurf=false;
            error("Warning: There are surface objects which have not been moved.<br>"+
                  "Please use the menu item \"Style&gt;>Surface objects\" to updated the surfaces.");
        }

    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Testing >>> */

}

