package charite.christo.protein;
import charite.christo.*;
import java.net.*;
import java.io.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**
   @author Christoph Gille
   // SocketOutputStream  BufferedInputStream
*/
public class ChSoapEBI {
    private ChSoapEBI() {}
    public final static String TYPE_OUT="out";
    public final static int STATUS_RUNNING=1, STATUS_FINISHED=2;

    public final static long ASK_FOR_PERMISSION=1<<1;
    public static ByteArray getResponse(long options, ByteArray query) { return getResponse(options, query, "www.ebi.ac.uk"); }
    public static ByteArray getResponse(long options, ByteArray query, String url) {
        if (sze(query)<10 || 0!=(options&ASK_FOR_PERMISSION) && !Web.uploadPermission(ChFrame.CENTER, query,url)) return null;
        OutputStream os=null;
        InputStream is=null;
        ByteArray answer=null;
        Socket socket=null;
        try {
            socket=new Socket(url,80);
            os=socket.getOutputStream();
            setContentLength(query);
            os.write(query.getBytes(),0,query.getEnd());

            os.flush();
            is=socket.getInputStream();
            answer=readBytes(is);
        } catch(IOException ex){ stckTrc(ex);}
        closeStrm(os);
        closeStrm(is);
        closeStrm(socket);
        return answer;
    }
    private static ByteArray rsc;
    public static ByteArray getRsc(String s) {
        final String fn="soap.rsc";
        synchronized(fn) {
            if (rsc==null) rsc=readBytes(rscAsStream(0L,_CCP,fn));
            assertTrue(rsc!=null);
        }
        int from=strstr(STRSTR_w|STRSTR_AFTER, "#"+s,rsc);

        if (from<0) {
            putln(RED_ERROR,"ChSoapEBI.getRsc ",s);
            assertTrue(false);
            return null;
        }
        from=nxtE(NO_SPC,rsc,from,MAX_INT);
        final int to=strstr(STRSTR_E|STRSTR_AFTER, "</soapenv:Envelope>",rsc,from, MAX_INT);
        //putln(" from="+from+" "+to);
        final byte[] bb=new byte[to-from];
        System.arraycopy(rsc.getBytes(), from, bb,0,to-from);
        final ByteArray ba=new ByteArray(bb);
        return ba;
    }
    public static int getStatus(String jobId) {
        final ByteArray query=getRsc("soapGetStatusEBI");
        if (query==null) return 0;
        final ByteArray answer=getResponse(0L, query.replace("JOBID",jobId));
        final int i=strstr(STRSTR_AFTER,"<status>",answer);
        return i<0 ? 0 :
            equls("FINISHED", answer, i) ? STATUS_FINISHED :
            equls("RUNNING", answer, i) ? STATUS_RUNNING : 0;
    }

    public static boolean setContentLength(ByteArray ba) {
        final int iXml=strstr("<?xml ",ba), iCont=strstr(STRSTR_AFTER, "Content-Length: ",ba);
        if (iXml<0 || iCont<0) return false;
        final byte T[]=ba.getBytes();
        final int e= strstr(STRSTR_AFTER|STRSTR_E,"Envelope>",ba);
        int len=e-iXml, pos=iCont;
        while(chrAt(pos+1,ba)==' ') pos++;
        while(len>0) {
            if (pos<=iCont) {assertTrue(false); return false;}
            T[pos--]=(byte)('0'+len%10);
            len/=10;
        }
        return true;
    }

    public static byte[] getResult(String jobId, String type) {
        final ByteArray query=getRsc("soapGetResultEBI");
        if (query==null) return null;
        final ByteArray answer=getResponse(0L, query.replace("JOBID",jobId).replace("TYPE",type));
        //putln("answer="+query);
        final int i=strstr(STRSTR_AFTER,"<output>",answer);
        final int gt=i<0?-1:strchr('<',answer,i,MAX_INT);
        return gt<0 ? null :  Base64.decode(answer.getBytes(), i, gt-i);
    }

    // java  charite.christo.strap.extensions.ChSoapEBI
    /*
    public static void main(String argv[]) throws Exception {
        putln(getRsc("soapGetResultEBI"));
    }
    */
}
