package charite.christo.protein;
import charite.christo.*;
import java.io.File;
import java.util.*;
import java.net.URL;
import charite.christo.interfaces.ChRunnable;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/*
FYI, I have put up a static mirror of the /das/sources and /das/coordinatesystem documents:

http://www.ebi.ac.uk/das-srv/registry/das/sources
http://www.ebi.ac.uk/das-srv/registry/das/coordinatesystem

*/
public class DAS extends ChRenderer implements Comparator  {

    public final static int DAYS_VALID=2;
    public final static String
    URL_REGISTRY="http://www.dasregistry.org/das1/sources/",
        CAPABILITY_TYPE_feature="das1:features",
        UniProt="UniProt",
        AUTHORITIES[]={UniProt, "UniParc","PDBresnum","IPI","EMBL","Entrez", "NCBI", "Ensembl", "Other"},
        POSSIBLE_CAPABILITY_TYPES[]={"das1:alignment", "das1:entry_points", "das1:error_segment", CAPABILITY_TYPE_feature, "das1:interaction", "das1:maxbins", "das1:sequence", "das1:sources", "das1:structure", "das1:stylesheet", "das1:types", "das1:unknown_feature", "das1:unknown_segment"};
    /* ---------------------------------------- */
    /* >>> Classes >>> */
    public static class Capability {
        private String type, query_uri;
        public String getType() { return type;}
        public String getQueryUri() { return query_uri;}
        public String toString() { return "DAS.Capability("+type+"   "+query_uri+")";}
    }
    public static class Coordinates {
        private String  authority, test_range, source;
        public String getAuthority() { return authority;}
        public String getTestRange() { return test_range;}
        public String getSource() { return source;}
        public String toString() { return "DAS.Coordinates( source="+source+"  authority="+authority+"  test_range="+test_range+")";}
    }
    public static class Entry {
        private String title, description, doc_href, leaseTime;
        public String getTitle() { return title;}
        public String getDescription() { return description;}
        public String getDocHref() { return doc_href;}
        public String getLeaseTime() { return leaseTime;}
        private Coordinates[] coordinates;
        private Capability[] capabilities;
        public Capability[] getCapabilities() { return capabilities;}
        public Coordinates[] getCoordinates() { return coordinates;}
        public String toString() {
            return "DAS.Entry("+title+"   "+description+"  "+doc_href+"  \n"+joinAsStrg(capabilities,"\n")+"\n"+joinAsStrg(coordinates,"\n");
        }
        public String getQueryUri(String type) {
            for(Capability c : capabilities)
                if (type==null || c.getType()==type)
                    return c.getQueryUri();
            return null;
        }
    }
    /* <<< Classes <<< */
    /* ---------------------------------------- */
    /* >>> Download Entry >>> */
    public static URL getURL(String ref, char chain,  Entry entry, String type) {
        final ByteArray baTmp=new ByteArray(99);
        baTmp.setEnd(0)
            .append(entry.getQueryUri(type))
            .append(type==CAPABILITY_TYPE_feature?"?segment=":"?query=")
            .urlEncode(ref,ref.indexOf(':')+1,MAX_INT);
        if (chain!=0) baTmp.append('.').append(chain);
        return url(baTmp.toString());
    }
    public static File downloadURL(String ref, char chain,  Entry entry, String type, ChRunnable log) {
        final URL url=DAS.getURL(ref,chain,  entry, type);
        if (log!=null) log.run(APPEND,"DAS.URL= "+url+"\n");
        final String leaseTime=entry.getLeaseTime();
        final File fCAched=file(url), fLeaseTime=leaseTime!=null ? file(url+".leaseTime") : null;
        if (sze(fCAched)>0 &&  CacheResult.isEnabled() && equls(leaseTime, readBytes(fLeaseTime))) {
            if (log!=null)  log.run(APPEND, "DAS use "+ANSI_FG_MAGENTA+" cached "+ANSI_RESET+"value leaseTime="+leaseTime+"\n");
            return fCAched;
        }
        final int timeValid=CacheResult.isEnabled() && leaseTime==null ? DAYS_VALID : 0;
        final File f=ChUtils.download_TS(url, timeValid, (File[])null, log);
        if (sze(f)>0) wrte(fLeaseTime,leaseTime);
        if (log!=null) log.run(APPEND,"DAS f="+f+" "+sze(f)+" Bytes\n");
        return f;
    }
    /* <<< Download <<< */
    /* ---------------------------------------- */
    /* >>> readRegistry >>> */
    public final static List<String> vRegistry=new ArrayList();
    static {vRegistry.add(URL_REGISTRY); }
    private static Entry entries[];
    final static List vEntries=new ArrayList(333);
    public synchronized static Entry[] readRegistry(ChRunnable log) {
        if (entries==null) {
            addAlternativeUrl("http://www.ebi.ac.uk/das-srv/registry/das/sources/",    URL_REGISTRY);
            vEntries.clear();
            for(int i=sze(vRegistry); --i>=0;) {
                final String reg=vRegistry.get(i);
                readReg(url(reg), vEntries, log);
            }
            entries=toArry(vEntries,Entry.class);
        }
        return DAS.entries;
    }
    private synchronized static void readReg(URL urlReg, final List vEntries, ChRunnable log) {
        if (urlReg==null) return;
        final File fReg=download_TS(urlReg, CacheResult.isEnabled() ? DAYS_VALID : 0, (File[])null, log);
        final ByteArray ba=readBytes(fReg);
        if (ba==null) return;
        final byte[][] AUTHORITIESbb=strings2bytes(AUTHORITIES);
        final byte[][] TYPESbb=strings2bytes(POSSIBLE_CAPABILITY_TYPES);
        final byte[] T=ba.getBytes();
        final int E=sze(ba);
        final int ltSOURCESgt=strstr("<SOURCES>",T,ba.getBegin(),E);
        final List vTmp=new ArrayList();
        if (ltSOURCESgt>0) {
            for(int ltSOURCE=ltSOURCESgt; (ltSOURCE=strstr(STRSTR_AFTER,"<SOURCE ",T,ltSOURCE,E))>0; ) {
                final Entry entry=new Entry();
                final int end=strstr("</SOURCE>",T,ltSOURCE,E);
                final int sourceGT=strchr('>',T,ltSOURCE,end);
                if (null==(entry.title=xmlAttribute("title",T,ltSOURCE,sourceGT))) continue;
                entry.description=xmlAttribute("description",T,ltSOURCE,sourceGT);
                entry.doc_href=xmlAttribute("doc_href",T,ltSOURCE,sourceGT);
                vTmp.clear();
                for(int lt=ltSOURCE; (lt=strstr(STRSTR_AFTER,"<CAPABILITY ",T,lt,end))>0; ) {
                    final int gt=strchr('>',T,lt,end);
                    final Capability capability=new Capability();
                    if (null==(capability.query_uri=xmlAttribute("query_uri", T, lt,gt))) continue;
                    if (null==(capability.type=xmlAttribute("type", T, lt,gt, POSSIBLE_CAPABILITY_TYPES, TYPESbb))) continue;
                    vTmp.add(capability);
                }
                if (sze(vTmp)==0) continue;
                entry.capabilities=toArry(vTmp,Capability.class);
                vTmp.clear();
                for(int lt=ltSOURCE; (lt=strstr(STRSTR_AFTER,"<COORDINATES ",T,lt,end))>0;) {
                    final int gt=strchr('>',T,lt,end);
                    if (gt<0) continue;
                    final Coordinates coordinates=new Coordinates();
                    coordinates.authority=xmlAttribute("authority",T,lt,gt, AUTHORITIES, AUTHORITIESbb);
                    coordinates.test_range=xmlAttribute("test_range",T,lt,gt);
                    coordinates.source=xmlAttribute("source",T,lt,gt);
                    vTmp.add(coordinates);
                }
                if (sze(vTmp)==0) continue;
                entry.coordinates=toArry(vTmp, Coordinates.class);

                for(int lt=ltSOURCE; (lt=strstr(STRSTR_AFTER,"<PROP ",T,lt,end))>0;) {
                    final int gt=strchr('>',T,lt,end);
                    if (entry.leaseTime==null) {
                        entry.leaseTime=strstr("name=\"leaseTime\"", T,lt,gt)>=0 ?  xmlAttribute("value", T,lt,gt) : null;
                    }
                }
                vEntries.add(entry);
            }
        }
    }
    /* <<< Get <<< */
    /* ---------------------------------------- */
    /* >>> Comparator  >>> */
    public final static String[] IMPORTANT_FEATURES={"CSA - extended", "cbs_total", "netphos", "netoglyc",
                                                     "Prosite Features (matches)","FireDB", "cbs_total"};

    public int compare(Object o1, Object o2) {
        if (!(o1 instanceof Entry && o2 instanceof Entry)) return 0;
        final Entry e1=(Entry)o1, e2=(Entry)o2;
        final String s1=e1.title, s2=e2.title;
        final int i1=idxOfStrg(s1,IMPORTANT_FEATURES), i2=idxOfStrg(s2,IMPORTANT_FEATURES);
        if (i1!=i2) return i1>i2 ? -1 : 1;
        return s1.compareTo(s2);
    }
    /* <<< Comparator <<< */
    /* ---------------------------------------- */
    /* >>> pdb2uniprot  >>> */
    public final static String DB_SOURCE[]={UniProt};
    private static byte[] DB_SOURCEbb[];

    private final static Map<String,String> pdb2uniprot=new HashMap();
    public static String pdb2uniprot(String pdb, char chain, ChRunnable log)  {
        if (pdb==null) return null;
        final String pdb_chain=(pdb+"_"+chain).intern();
        String unip;
        synchronized(pdb_chain) {
            unip=pdb2uniprot.get(pdb_chain);
            if (unip==null) {
                if (DB_SOURCEbb==null) DB_SOURCEbb=strings2bytes(DB_SOURCE);
                for(Entry e : readRegistry(log)) {
                    if (!"pdb_uniprot_sifts_mapping".equals(e.getTitle())) continue;
                    final File f=downloadURL(pdb, chain,  e, "das1:alignment" , log);
                    final ByteArray ba=readBytes(f);
                    if (ba==null) continue;
                    final byte[] T=ba.getBytes();
                    final int E=ba.getEnd();
                    for(int lt=ba.getBegin(); unip==null && (lt=strstr(STRSTR_AFTER,"<alignObject ",T,lt,E))>0; ) {
                        final int gt=strchr('>',T,lt,E);
                        final String dbSource=xmlAttribute("dbSource", T, lt,gt, DB_SOURCE, DB_SOURCEbb);
                        if (UniProt==dbSource)  unip=xmlAttribute("dbAccessionId",T,lt,gt);
                    }
                    if (unip!=null) break;

                }
            }
        }
        return unip;
    }

}

