package charite.christo.protein;
import charite.christo.*;
import java.io.*;
import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static charite.christo.protein.ChSoapEBI.*;
/*
http://www.ebi.ac.uk/ws/services/WSDbfetchDoclit?wsdl

http://www.ebi.ac.uk/ws/services/WSDbfetchDoclit?wsdl
*/
public final class FetchEBI_SOAP {
    private static boolean DEBUG=false;
    public final static long SKIP_EXISTING=1;
    public final static String
        CMD_WSDBFETCH_SEND_QUERY="CMD_WSDBFETCH_SEND_QUERY",
        CMD_WSDBFETCH_ERROR="CMD_WSDBFETCH_ERROR",
        CMD_WSDBFETCH_SUCCESS="CMD_WSDBFETCH_SUCCESS",
        DB_EMBL="EMBL",DB_UNIPARC="UNIPARC",
        DB_UNIPROT="UNIPROT",DB_UNIPROTKB="UNIPROTKB",
        DATABASES[]={DB_UNIPROTKB,DB_UNIPROT, DB_EMBL,DB_UNIPARC};
    private FetchEBI_SOAP(){}
    private static String preferredFormat(String db) {
        if (DB_UNIPROTKB.equalsIgnoreCase(db) || DB_UNIPROT.equalsIgnoreCase(db) ) return "uniprot";
        if (DB_EMBL.equals(db)) return "embl";
        return "fasta";
    }
    private static Object mapEntryFile;
    public static File dbColonID2file(String entry) {
        Map<String,File> map=deref(mapEntryFile, Map.class);
        if (map==null) mapEntryFile=newSoftRef(map=new HashMap());
        if (entry==null || sze(entry)>66) return null;
        File f=map.get(entry);
        if (f==null) {
            final int colon=entry.indexOf(':');
            if (colon<0) return null;
            final String database=( entry.startsWith("EXPASY:") || entry.startsWith("SWISS:") ) ? DB_UNIPROT : entry.substring(0,colon);
            map.put(entry,f=id2file(database,entry.substring(colon+1), null));
        }
        return f;
    }
    private static File dir_;
    private static File dir() {
        if (dir_==null) {
            dir_=file(dirSettings() +"/wsdbfetch/");
            final String README=
                "You may remove the files in the directory to save memory\n"+
                "This directory contains protein files which are downloaded\n"+
                "via SOAP WSDbfetch from EBI\n";
            mkdrsErr(dir_,README);
        }
        return dir_;
    }
    private static File id2file(String database,String id,String format0) {
        final String format=sze(format0)>0 ? format0 : preferredFormat(database);
        if (id==null||sze(id)>33 || database==null||sze(database)>33) return null;
        return file(dir()+"/"+(database+"_"+rplcChars(charClass(NO_FILENM),'_',id)).toUpperCase()+"."+format);
    }

    public static Thread thread_download(long options, String tokens[]) {
        return threadM("download", FetchEBI_SOAP.class, new Object[]{longObjct(options), tokens});
    }

    public static void download(long options, String tokens[]) {
        final Set<String> vQuery=new HashSet();
        for(String db :DATABASES) {
            vQuery.clear();
            final String dbColon=db+":";
            if (tokens!=null) {
                for(String dbColonId:tokens) {
                    if (dbColonId!=null && dbColonId.startsWith(dbColon) ||
                        dbColonId!=null && DB_UNIPROT==db && (dbColonId.startsWith("EXPASY:")||dbColonId.startsWith("SWISS:"))) {
                        final int colon=dbColonId.indexOf(':'), pipe=dbColonId.indexOf('|');
                        final String s=pipe>0 ? dbColonId.substring(0,pipe) : dbColonId;
                        final String id=s.substring(colon+1);
                        if ( (options&SKIP_EXISTING)!=0 &&
                             (sze(id2file(db,id,preferredFormat(db)))>=PROTEIN_MIN_FILE_SZE || sze(file(Hyperrefs.toUrlString(s,Hyperrefs.PROTEIN_FILE)))>=PROTEIN_MIN_FILE_SZE)) {
                            //if (isMyComputer()) putln("FetchEBI_SOAP: skipping ",s);
                            continue;
                        }
                        vQuery.add(id);
                    }
                }
            }
            FetchEBI_SOAP.download(db,toArry(vQuery,String.class),FetchEBI_SOAP.preferredFormat(db));
        }
    }

    private final static int MAXNUM=100;
    public static void download(final String database,final String[] IDs,final String format0)  {
        int fromIdx=0;
        while(fromIdx<IDs.length) {
            final String format=format0!=null && format0!="" ? format0 : preferredFormat(database);
            final StringBuffer sb=new StringBuffer(IDs.length*10);
            for(int i=fromIdx, count=0; i<IDs.length && count<MAXNUM; i++) {
                final File f=id2file(database,IDs[i],format);
                if (sze(f)==0) {
                    if (sze(sb)>0) sb.append(' ');
                    if (IDs[i]!=null) { sb.append(IDs[i]); count++; }
                }
                fromIdx=i+1;
            }
            final String queries=sb.toString();
            if (sze(queries)>0) {
                ByteArray answer=null,query=null;
                query=getRsc("soapWSDBFetch");
                if (query!=null) {
                    query
                        .replace("DATABASE", database)
                        .replace("FORMAT", format)
                        .replace("QUERIES", queries);
                    if (DEBUG) putln(ANSI_GREEN+"\nSocket Write"+ANSI_RESET+"\n");
                    if (DEBUG) putln(query,"\n");
                    answer=getResponse(0L, query);
                }
                if (DEBUG) putln("\n\nFetchEBI_SOAP answer ",answer,"\n\n");
                if (answer!=null) {
                    final byte T[]=answer.getBytes();
                    final int
                        end0=answer.getEnd(),
                        beginO=strstr("<fetchBatchReturn" ,T,0,end0),
                        end=strstr("</fetchBatchReturn",T,beginO,end0),
                        begin=1+strchr('>',T,beginO,end0),
                        ends[]=answer.getLineEnds();
                    if (begin>0) {
                        answer.setBegin(begin);
                        answer.setEnd(end);
                        int from=0;
                        for(int iL=0; iL<ends.length; iL++) {
                            final int b=iL==0 ? begin : ends[iL-1]+1, e=ends[iL];
                            if (e-b<8) continue;
                            final boolean entryStarts=T[b]=='I' && T[b+1]=='D' && T[b+2]==' ';
                            if (entryStarts || T[b]=='A' && T[b+1]=='C' && T[b+2]==' ') {
                                from=b;
                                for(int iID=IDs.length; --iID>=0;) {
                                    final String id=IDs[iID];
                                    final int idxID=strstr(STRSTR_IC,id,T,b,e);
                                    if (idxID>0 && !is(LETTR_DIGT,T[idxID-1]) && !is(LETTR_DIGT,T,idxID+sze(id))) {
                                        final int slashSlash=strstr("\n//",T,b,end);
                                        if (slashSlash>0) {
                                            final File f=id2file(database,id,format);
                                            wrte(f, T, from,slashSlash+1);
                                        } else stckTrc();
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    /*
    // java charite.christo.protein.FetchEBI_SOAP
    public static void main(String argv[]) {
        //setPrgParameters("", argv,true);
        final String ids[]={"Q27488", "Q476Y4"};
        for(String id : ids) {
            final File f=id2file(DB_UNIPROT, id, "");
            delFile(f);
            assertTrue(sze(f)==0);
        }

        download(DB_UNIPROT, ids,"");
        for(String id : ids) {
            final File f=id2file(DB_UNIPROT, id, "");
            putln("\n f="+f+"  "+sze(f));
        }
    }
    */
}
/*
  echo rm ~/.StrapAlign/wsdbfetch/*
  echo rm -r ~/.StrapAlign/wsdbfetch/* ~/.StrapAlign/web/*
*/

