package charite.christo.protein;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import java.awt.Container;
import java.util.*;

import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**
This class converts SequenceAligner  for two proteins into one for many proteins.
The  utility  is used to produce multiple sequence alignments with classes that
can align only two proteins.
@author Christoph Gille
MultiFromPairAligner
*/
public class MultiFromPairAligner implements CanBeStopped,NeedsProteins,SequenceAlignerSorting, HasControlPanel {
    /* >>> Constructor and Settings >>> */
    public MultiFromPairAligner(Object c) {
        assertTrue(isClassOrPC(c));
        classes[0]=c;
    }
    private final Object classes[]={null,null};
    public MultiFromPairAligner setSuperpositionMethod(Class c) {classes[1]=c;

return this;}
    public Object[] getClasses() { return this.classes; }
    //public Class getClazzes() { return HotswapUtil.getClass(clazz); }
    /* <<< Constructor and Settings <<< */
    /* ---------------------------------------- */
    /* >>> Interface Methods >>> */
    private int[] sortedIndices;
    private Protein proteins[];
    private byte seqs[][], alignedSequences[][];
    public byte[][] getSequences() { return seqs;}
    public byte[][] getAlignedSequences() { return alignedSequences;}
    public void setSequences(byte[]...ss) {    seqs=ss;   }

    public void setProteins(Protein... pp) {
        proteins=pp;
        log("\n"+ANSI_YELLOW+"MultiFromPairAligner setProteins"+ANSI_RESET+"\n");
        for(int i=0; i<sze(pp);i++) log(i+" "+pp[i]+"\n");
    }
    public void setOptions(long flags) {}
    public long getPropertyFlags() { return 0L;}
    public long getOptions() { return 0L;}
    public int[] getIndicesOfSequences() { return sortedIndices;}
    /* <<<  Interface Methods <<< */
    /* ---------------------------------------- */
    /* >>> ControlPanel >>> */
    private Container ctrlPan;
    private final static boolean LOG=true;
    private AsciiView taResult;
    public AsciiView log(String... oo) {
        AsciiView ta=this.taResult;
        if (ta==null && isGraphicsAvailable())
            ta=taResult=new AsciiView(999)
                .appendS("This is the log Panel of MultiFromPairAligner("+shrtClasNam(classes[0])+")\n3d-Superposition: "+shrtClasNam(classes[1])+"\n");
        if (ta!=null) ta.appendSS(oo);
        return ta;
    }
    public Container getControlPanel() {
        AsciiView ta=log();
        if (ctrlPan==null && ta!=null) ctrlPan=scrllpn(0L,ta);
        return ctrlPan;
    }
    /* <<< ControlPanel <<< */
    /* ---------------------------------------- */
    /* >>> Compute >>> */
    private final static int FLAG_setProteins=1, FLAG_3D=1<<1;
    private int[] flags;
    private int flags(int i01) {
        if (flags==null) {
            flags=new int[2];
            for(int i=2;--i>=0;) {
                flags[i]=
                    ( isAssignblFrm(NeedsProteins.class, this.classes[i]) ? FLAG_setProteins : 0)|
                    ( isAssignblFrm(Superimpose3D.class, this.classes[i]) ? FLAG_3D : 0);
            }
        }
        return flags[i01];
    }

    private AProtein[] _aPP;
    private AProtein[] aProts() {
        if (_aPP==null) {
            final Protein pp[]=this.proteins;
            _aPP=new AProtein[pp.length];
            for(int i=pp.length;--i>=0;)
                _aPP[i]=new AProtein(seqs[i],pp!=null && pp.length>i ? pp[i]: null,i);
        }
        return _aPP;
    }

    private AlignmentResult alignmentResults[][];
    private AlignmentResult[][] getAlignments() {
        final String MSG_STOP="MultiFromPairAligner stopped";
        long timePG=0;
        final StringBuffer sbPG=new StringBuffer();
        if (alignmentResults==null) {
            alignmentResults=new AlignmentResult[seqs.length][seqs.length];
            final AProtein[] pp=aProts();
            final int N=pp.length*pp.length;

            final String constProgress= "/"+(pp.length*pp.length)+" "+niceShrtClassNam(classes[0])+" "+(niceShrtClassNam(classes[1]==null?"":classes[1]))+" ";
            for(int iR=pp.length, count=0; --iR>=0;) {
                for(int iM=pp.length; --iM>=0; count++) {
                    final Protein pR=pp[iR].protein, pM=pp[iM].protein;
                    if (alignmentResults[iR][iM]==null) {
                        AlignmentResult result=alignmentResults[iR][iM]=new AlignmentResult(pp[iR], pp[iM]);
                        SequenceAligner aligner=null;
                        byte[][] rrLocal=null;
                        double score=Double.NaN;
                        boolean is3D=false;
                        for(int try3d=2; --try3d>=0 && rrLocal==null;) {
                            is3D=try3d>0;
                            final Object claz=classes[try3d];
                            if (claz==null || is3D && !(pM!=null && pR!=null && pM.getResidueCalpha()!=null && pR.getResidueCalpha()!=null)) continue;
                            aligner=HotswapUtil.instance(claz, SequenceAligner.class,true);
                            if (aligner==null) continue;
                            if (pM!=null && (flags(try3d)&FLAG_setProteins)!=0) ((NeedsProteins)aligner).setProteins(pR,pM);
                            final boolean isSuper=isAssignblFrm(Superimpose3D.class, aligner) &&  pM!=null && (flags(try3d)&FLAG_3D)!=0;
                            if (isSuper) {
                                ((Superimpose3D)aligner).setReferenceProtein(pR);
                                ((Superimpose3D)aligner).setMobileProtein(pM);
                            } else aligner.setSequences(pp[iR].ungapped, pp[iM].ungapped);
                            if (stopped || get(0,isInterrupted)) { putln(MSG_STOP); return null;}
                            refLastAligner=weakRef(aligner);

                            aligner.compute();
                            if (System.currentTimeMillis()-timePG>200) {
                                sbPG.setLength(0);
                                sbPG.append(KEY_PROGRESS).append(count).append(constProgress);
                                if (pR!=null && pM!=null) sbPG.append(pR).append(" ./. ").append(pM);
                                timePG=System.currentTimeMillis();
                                handleActEvt(this, sbPG.toString(),0);
                            }
                            if (stopped) { putln(MSG_STOP); return null;}
                            rrLocal=aligner.getAlignedSequences();
                            final String iRiM=ANSI_RESET+" ("+iR+","+iM+") ";
                            if (get(0,rrLocal)==null || get(1,rrLocal)==null) {
                                log(ANSI_RED+"Error"+iRiM+"getAlignedSequences() returns null "+claz+"\n");
                                return null;
                            } else {
                                if (isSuper) {
                                    if (isInstncOf(HasScore.class, aligner)) score=((HasScore)aligner).getScore();
                                    else score=Blosum.pairAignScore(-20f,-4f,rrLocal[0],rrLocal[1],   -Blosum.BLOSUM62_MIN);
                                } else score=Blosum.pairAignScore(-20f,-5f,rrLocal[0],rrLocal[1],  0);

                                //putln(isSuper+" score="+score);
                                if (LOG) log(ANSI_GREEN,"Success",iRiM+"score="+score+" ",(try3d>0?" 3D":""),"\n");
                            }
                        }
                        if (get(0,rrLocal)==null || get(1,rrLocal)==null) {
                            return null;
                        }
                        final byte oldAli[][]={pp[iR].ungapped, pp[iM].ungapped};
                        if (isMyComputer() && sze(rrLocal)!=2) putln(RED_ERROR+"sze(rrLocal)!=2");
                        result.aligned=rrLocal;
                        result.score=score;
                        result.superimposed=is3D;
                        if (isInstncOf(Disposable.class,aligner)) ((Disposable)aligner).dispose();
                        if (LOG) log("result("+iR+","+iM+")="+result.score+"\n");

                        count++;
                        if (rrLocal!=null) alignmentResults[iM][iR]=result.invert();
                    }
                }
            }
            handleActEvt(this, KEY_PROGRESS+N+" Alignments done" ,0);
        }
        return alignmentResults;
    }
    public void compute(){
        for(int i=2;--i>=0;) {
            if (this.classes[i]!=null && HotswapUtil.instance(classes[i],SequenceAligner.class,true)==null) {
                error("Could not instanciate <pre>"+this.classes[i]+"</pre>");
                debugExit("classes[i]="+classes[i]);
                return;
            }
        }
        final AlignmentResult results[][]=getAlignments();
        if (results==null) {
            putln(RED_ERROR+" MultiFromPairAligner results=null");
            return;
        }

        final int nP=seqs.length;
        if (nP<2) return;
        final ProteinStack stack=new ProteinStack();
        final AProtein aProtein[]=aProts();
        int topIdx=0;
        for(int try3d=2; --try3d>=0;) {
            double maxScoreSum=Float.MIN_VALUE;
            for(int i=nP; --i>=0;) {
                if (try3d>0 && aProtein[i].protein.getResidueCalpha()==null) continue;
                float scoreSum=0;
                for(int j=nP;--j>=0;) {
                    if (i==j) continue;
                    final AlignmentResult result=results[i][j];
                    scoreSum+=result.score*result.score;
                }
                if (maxScoreSum<scoreSum) {
                    maxScoreSum=scoreSum;
                    topIdx=i;
                }
            }
            if (topIdx>=0) break;
        }
        log("\n A(i,j) means sequence alignment of protein i,j and S(i,j) means 3D superposition\n");
        stack.top=aProtein[topIdx];
        adUniq(stack.top,    stack.all);
        final List<AProtein> notAlignedV=new ArrayList();
        for(int i=nP;--i>=0;) {
            if (i!=stack.top.number && aProtein[i]!=null) notAlignedV.add(aProtein[i]);
        }
        while(notAlignedV.size()>0 && isRunning)  compute(stack,notAlignedV);
        alignedSequences=new byte[nP][];
        sortedIndices=new int[nP];
        for(int i=0;i<nP && isRunning;i++) {
            final AProtein p=stack.all.get(i);
            alignedSequences[p.number]=p.gapped;
            sortedIndices[i]=p.number;
        }
        alignmentResults=null;
        proteins=null;
    }
    private void compute(ProteinStack stack, List<AProtein> notAlignedV) {
        final AlignmentResult bestAli=bestAlignmentResult(stack.all, notAlignedV);
        final AProtein protStack=bestAli.proteinR, protNotAl=bestAli.proteinM;
        final byte aligned[][]=bestAli.aligned;
        if (protStack==null||aligned==null||protNotAl==null||aligned==null) {
            if (isMyComputer()) log(ANSI_RED,"Error",ANSI_RESET," in protStack="+(protStack!=null)+" aligned="+aligned+" protNotAl="+protNotAl+" aligned="+(aligned!=null ?  aligned.length : -1)+"\n");
            isRunning=false;
            return;
        }
        log( (bestAli.superimposed? " S(" : " A(")+protStack.number+","+protNotAl.number+") ");
        final byte[][]  ist={protStack.gapped , protNotAl.gapped};
        final byte[][] soll={aligned[0],aligned[1] };
        final int[][] inserts=new MergeAlignments(MergeAlignments.START_IDX_AUTO,soll,ist).getInsertsAtCol();
        protNotAl.gapped=insertGapsAtColumns(inserts[1],protNotAl.gapped);
        for(int i=stack.all.size(); --i>=0; ) {
            final AProtein p=stack.all.get(i);
            p.gapped=insertGapsAtColumns(inserts[0],p.gapped);
        }
        notAlignedV.remove(bestAli.proteinM);
        adUniq(bestAli.proteinM, stack.all);
    }
    /* <<< Compute <<< */
    /* ---------------------------------------- */
    /* >>> Best Comparison  >>> */
    private AlignmentResult bestAlignmentResult(List<AProtein> pp1,List<AProtein> pp2) {
        AlignmentResult best=null;
        final AlignmentResult aa[][]=getAlignments();
        for(int i1=pp1.size(); --i1>=0;) {
            for(int i2=pp2.size(); --i2>=0;) {
                final AlignmentResult result=aa[pp1.get(i1).number][pp2.get(i2).number];
                if (best==null || result!=null && result.isBetterThan(best)) best=result;
            }
        }
        return best;
    }
    /* <<< Best Comparison <<< */
    /* ---------------------------------------- */
    /* >>>  Classes  >>> */
    private static class ProteinStack { AProtein top;  List<AProtein> all=new ArrayList();}
    private static class AlignmentResult {
        final AProtein proteinM,proteinR;
        byte[][] aligned;
        double score;
        boolean superimposed;
        public AlignmentResult(AProtein pR, AProtein pM) { proteinR=pR; proteinM=pM; }
        public AlignmentResult invert() {
            AlignmentResult r=new AlignmentResult(proteinM, proteinR);
            r.score=score;
            r.superimposed=superimposed;
            r.aligned=new byte[][]{aligned[1], aligned[0]};
            return r;
        }
        private boolean isBetterThan(AlignmentResult r2) {
            if (r2==null) return true;
            if (superimposed && !r2.superimposed) return true;
            if (!superimposed && r2.superimposed) return false;
            return score>r2.score;
        }
    }
    private static class AProtein  {
        final byte[] ungapped;
        final int number;
        final Protein protein;// may be null
        byte[]  gapped;
        public AProtein(byte[] seq,Protein p,int number) {
            gapped=seq;
            protein=p;
            this.number=number;
            this.ungapped=seq;
        }
        @Override public String toString() {return "AProtein#"+number+" ";}
    }
    /* <<< Classes  <<< */
    /* ---------------------------------------- */
    /* >>> User Interface >>> */
    //    public final int getProgressMaximum() {int n=getSequences().length; return (n*n-n)/2;}
    //    public final int getProgress() {return count;}

    /* <<<  User Interface <<< */
    /* ---------------------------------------- */
    /* >>> Threading >>> */
    private boolean stopped;
    private boolean isRunning=true;
    private Object refLastAligner;
    public void stop() {
        stopped=true;
        final CanBeStopped cbs=deref(refLastAligner, CanBeStopped.class);
        if (cbs!=null) cbs.stop();
    }
    {vALIGNMENTS.add(weakRef(this));}
    private boolean isInterrupted[];
    public void setIsInterrupted(boolean[] boolArraySize1) {
        this.isInterrupted=boolArraySize1;
    }
    /* <<< Threading <<< */
    /* ---------------------------------------- */
    /* >>> Utils >>> */
    public static byte[] insertGapsAtColumns(int insert[], byte orig[]) {
        int size=orig.length;
        for(int i=orig.length<insert.length ? orig.length:insert.length; --i>=0;) size+=insert[i];
        final byte txt[]=new byte[size];
        int iT=0;
        for(int i=0;i<orig.length; i++) {
            if (i<insert.length) {
                for(int j=insert[i]; --j>=0;) txt[iT++]=32;
            }
            txt[iT++]=orig[i];
        }
        return txt;
    }

}
