package charite.christo.protein;
import charite.christo.*;
import charite.christo.interfaces.*;
import java.util.*;
import java.io.*;
import java.awt.Component;
import javax.swing.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
import static charite.christo.protein.ProteinViewer.*;
import static java.awt.event.MouseEvent.*;

public class Protein3dUtils implements ChRunnable {
    private final static Object KEY_SHARED_MAP=new Object();
    private final static String KEY_NATIV_MENUS="P3U$$NM";
  public final static int GENERIC=0,  SURFACE_ID=1, JMOL=2, PYMOL=3, ASTEX=4, RASMOL=5,
      SURFACE_ID_COLON=1<<17, SURFACE_ID_DASH=1<<18, SURFACE_ID_KOMMA=1<<19, SURFACE_ID_DOT=1<<20;
    public final static String
    KEY_ADD_TO_NATIVE_MENU="P3U$$ATNM",
        SCRIPT_select_3D="select_3D ",
        VARIABLES="$FIRST $LAST $NTERM $CTERM $ALL $AMINOACID ";

    private static Protein3dUtils _inst;
    private static Protein3dUtils instance() {if (_inst==null) _inst=new Protein3dUtils(); return _inst; }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Cache >>> */
    private final static int CACHE_splitAtoms=1;
    private static Object rCached;
    private static synchronized Object[] cached() {
        Object[] r=(Object[])deref(rCached);
        if (r==null) rCached=newSoftRef(r=new Object[99]);
        return r;
    }

    private static String[] _atoms;
    public static String[] atomsOfAminos() {
        if (_atoms==null) {
            String s="C C5 C5W CA CB CD CD1 CD2 CE CE1 CE2 CE3 CF CG CG1 CG2 CH0 CH1E CH1S CH2 CH2E CH2G CH2P CH3E CR1E CR1H CR1W CRHH CW CY CY2 CZ CZ2 CZ3 "+
                "N NC2 ND1 ND2 NE NE1 NE2 NH1 NH1S NH2 NH3 NZ "+
                "O OC OD1 OD2 OE1 OE2 OG OG1 OH OH1 OS OXT "+
                "SD SG SH1E SM ";
                _atoms=splitTokns(0L, s);
        }
        return _atoms;
    }
    private static List<Protein> vPIV=new ArrayList();
    public static Protein[] getProteinsInView(ProteinViewer...vv) {
        synchronized(vPIV) {
            vPIV.clear();
            if (vv!=null) for(ProteinViewer v : vv) getProteinsInViewV(v,vPIV);
            return sze(vPIV)==0?Protein.NONE:toArry(vPIV,Protein.class);
        }
    }

    public static List<Protein> getProteinsInViewV(ProteinViewer viewer, List<Protein> vProts) {
        if (viewer==null) return vProts;
        for(ProteinViewer v : getViewersSharingView(viewer)) {
            adUniq(p(v), vProts);
        }
        return vProts;
    }
    public static ProteinViewer[] getViewersSharingView(ProteinViewer pv) {
        if (pv==null) return ProteinViewer.NONE;
        final ProteinViewer vv[]=rmNullA(pv.getViewersSharingView(),ProteinViewer.class);
        return sze(vv)>0 ? vv : new ProteinViewer[]{pv};
    }

    public static String titleFor3dViewers(Protein pp[]) {
        final StringBuffer sb=strBuf(SB_TIP).append(" #").append(pp.length).append(' ');
        for(Protein p : pp) {
            sb.append(delPfx("pdb",delSfx(".pdb",delSfx(".ent",delSfx(".Z",delSfx(".gz",p.getName())))))).append(' ');
        }
        return sb.toString();
    }
    public static Protein p(ProteinViewer pv) { return pv==null?null:pv.getProtein();}
    public static boolean isActive(ProteinViewer v) {
        final Protein p=p(v);
        return p==null ? false : cntains(v, p.getProteinViewers());
    }

    public static ProteinViewer[] getViewers(Protein[] pp, Class clazz) {
        final List v=new ArrayList();
        for(Protein p : pp) {
            for(ProteinViewer pv : p.getProteinViewers()) {
                if (clazz==null || isAssignblFrm(clazz,pv)) adUniq(pv,v);
            }
        }
        return toArry(v,ProteinViewer.class);
    }

    public static Component viewerCanvas(ProteinViewer pv) { return pv==null?null: (Component)pv.getProperty(ProteinViewer.GET_CANVAS);}
    public static JMenuBar jMenuBar(ProteinViewer pv) { return pv==null?null: (JMenuBar)pv.getProperty(ProteinViewer.GET_JMENUBAR);}

    public static JComponent[] getNativeMenus(ProteinViewer pv) {
        final Component panel=viewerCanvas(pv);
        if (panel==null) return null;
        JComponent[] mm=gcp(KEY_NATIV_MENUS, panel, JComponent[].class);
        if (mm==null) {
            final JMenuBar mb=jMenuBar(pv);
            if (mb!=null) {
                //final JMenu mOther=jMenu(0L, pvp().m('b',null),"");
                removeMenuItems(new String[]{"file"}, new String[]{"close","exit","Close","Exit"}, mb);
                List v=new ArrayList();
                for(Component c : mb.getComponents()) if (c instanceof JMenu) v.add(c);
                final JComponent mOther=gcp(KEY_ADD_TO_NATIVE_MENU, panel, JComponent.class);
                if (mOther!=null) {
                    v.add(Box.createHorizontalGlue());
                    v.add(mOther);
                }
                mm=toArry(v,JComponent.class);

            } else mm=new JComponent[0];
            pcp(KEY_NATIV_MENUS, mm,panel);
        }
        return mm;
    }

    public static ProteinViewer[] askViewersSharingView(ProteinViewer pv0, Protein pSuggestion, String msg) {
        final ProteinViewer[] vv=getViewersSharingView(pv0), vv2=vv.clone();
        if (vv.length<2) return vv;
        final Protein pp[]=new Protein[vv.length];
        int sel=-1;

        for(int i=vv.length; --i>=0;) {
            pp[i]=p(vv[i]);
            if (pSuggestion!=null && pSuggestion==pp[i]) sel=i;
        }

        final ChJList jl=new ChJList(pp,0);

        if (sel>=0) jl.setSelI(sel); else selectAllItems(jl);

        final Object pan=newCNSEW(scrllpn(0L,jl),"The 3D-View has more than one Proteins. Please select at least one.",msg,null,null);
        if (!ChMsg.yesNo(pan)) return ProteinViewer.NONE;
        final Object selected[]=jl.getSelectedValues();
        for(int i=sze(vv2); --i>=0;) if (!cntains(p(vv2[i]), selected)) vv2[i]=null;
        return rmNullA(vv2,ProteinViewer.class);
    }
    public static ProteinViewer findProteinViewerWithProteins(Class clazz, Protein pp[], boolean exactly) {
        for(int iP=sze(pp); --iP>=0; ) {
            if (pp[iP]==null) continue;
            for(ProteinViewer pv : pp[iP].getProteinViewers()) {
                if (clazz!=null && !isAssignblFrm(clazz,pv)) continue;
                final Protein ppV[]=getProteinsInView(pv);
                if (!cntainsAll(pp, ppV)) continue;
                if (exactly && pp.length!=ppV.length) continue;
                return pv;
            }
        }
        return null;
    }
    public static ProteinViewer getViewerForProtein(Protein p, ProteinViewer pv) {
        if (p!=null && pv!=null)
            for(ProteinViewer v : getViewersSharingView(pv))
                if (p(v)==p) return v;
        return null;
    }
    public static ProteinViewer derefViewer(Object o) {
        return (ProteinViewer) (o instanceof ProteinViewer ? o : deref(ProteinViewer.mapViewer.get(o)));
    }
    //selectionString(
    //select ( molecule "a1_SaccharomycesCerevisiae.pdb" and (  (  residue 10 to 160  and chain '2'  and atom CA CB )  ) );
    //select Unnamed = (protein and ( ///2/183-187/CA,CB  or  ///A/183-187 ) )
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Scripting >>> */
    private final static String COMMANDS_WITH_COLOR[]={COMMANDcolor,COMMANDsurface_color, COMMANDbackground, COMMANDchange_object_color};

    private static String[] _allCmds;
    public static String[] allCommands() {
        if (_allCmds==null) _allCmds=getFinalStaticStrings("COMMAND", ProteinViewer.class);
        return _allCmds;
    }

    public static long rgbInScript(String command) {
        for(String word0 : COMMANDS_WITH_COLOR) {
            if (strstr(STRSTR_w,word0, command)<0) continue;
            final int begin=command.indexOf('#')+1;
            final int end=begin>0?nxtE(NO_HEX_DIGT,command,begin,MAX_INT) : -1;
            return end-begin<2?-1: Long.parseLong(command.substring(begin,end),16) | (end-begin>6 ? 0 : 255L<<24);
        }
        return -1;
    }
    public static String objectIdInScript(String c) {
        int spc=nxt(SPC,c), b=spc<0?-1 : nxt(NO_SPC, c, spc,MAX_INT), e=b<0?-1 : nxtE(SPC, c, b,MAX_INT);
        while(chrAt(e-1,c)==';') e--;
        if(chrAt(e-1,c)=='"') e--;
        if(chrAt(b,c)=='"') b++;

        return b<e && b>0 ? c.substring(b,e) : null;
    }

    public static Matrix3D matrixInScript(String c) {
        if (c.startsWith(COMMANDset_rotation_translation)) {
            if (c.indexOf("null")>0) return null;
            Matrix3D m3d=new Matrix3D();
            if (m3d.parsePlain(c.getBytes(), COMMANDset_rotation_translation.length(), MAX_INT)) {
                return  m3d.isUnit() ? null : m3d;
            }
            else putln("Error in Matrix: ",c);
        }
        return null;
    }
    public static String makeSelId(Selection3D[] atoms, String selId) {
        return sze(selId)>0 ? selId : sze(atoms)==0 ? "entire" : "unnamed";
    }
    /* <<< Scripting <<< */
    /* ---------------------------------------- */
    /* >>> Log >>> */
    public final static String LOG_PROTEIN_PFX=SCRIPT_select_3D+"*, ", LOG_FILE=STRAPOUT+"/commands.3d.txt";
    private final static Object KEY_LOG[]={new Object(), new Object()};

    public static ChLogger log3D(boolean isGeneric, ProteinViewer pv0, String command) {
        final String KEY_p="P3U$$logP";
        ProteinViewer pv=AbstractProxy.getProxy(pv0,ProteinViewer.class);
        if (pv==null) pv=pv0;
        final Protein p=p(pv);
        if (p==null) return null;
        final Object key=KEY_LOG[isGeneric?0:1];
        final Map map=getSharedHashMap(pv);
        ChLogger log=(ChLogger)map.get(key);
        if (log==null) {
            map.put(key,log=new ChLogger(0L, 10*1000, isGeneric?file(LOG_FILE):null));
            if (isGeneric) log.setPreprocessor(instance());
        }
        if (equls("3D_",command)==isGeneric && sze(command)>0) {
            final Object lastP=map.get(KEY_p);

            if (p!=lastP) {
                log.appendLine(LOG_PROTEIN_PFX+p);
                map.put(KEY_p,p);
            }
            log.appendLine(command);
            if (isMyComputer() && !isGeneric) putln(niceShrtClassNam(pv),ANSI_GREEN+"> ", ANSI_RESET, command);
        }
        return log;
    }

    public static void showLogPanel(boolean isGeneric, ProteinViewer pv) {
        if (pv==null) return;
        final ChLogger log=(ChLogger) log3D(isGeneric, pv, null);
        ChFrame.frame(niceShrtClassNam(pv)+(isGeneric?" Generic":" Native"), log.getPanel(), CLOSE_ON_CtrlW|ChFrame.PACK)
            .shw(ChFrame.CENTER);
    }

    private final static String LOG_SKIP[]={
        COMMANDzoom, COMMANDrotate,  COMMANDcenter_amino,
        COMMANDselection_name+" "+SELECTION_CURSOR,
        COMMANDselection_name+" "+ResidueSelection.STANDARD,
        COMMANDselection_name+" "+SELECTION_PICKED
    };
    private final int BB[]=new int[2];
    private void lineStarts(ByteArray ba) {
        final int E=ba.getEnd();
        final byte[] T=ba.getBytes();
        for(int i=BB.length; --i>=0;) BB[i]=-1;
        for(int i=E, count=0; --i>=0;) {
            if (i==0 || T[i-1]=='\n' ) {
                BB[count++]=i;
                if (BB.length==count) break;
            }
        }
    }
    public Object run(String id, Object arg) {
        final Object argv[]=arg instanceof Object[] ? (Object[]) arg : null;
        if (id==ChLogger.RUN_PREPROCESS) {
            final String line=(String)argv[0];
            final ByteArray ba=(ByteArray)argv[1];
            putln("line="+line);
            lineStarts(ba);
            final boolean isSel=equls(STRSTR_w_R, COMMANDselect, line,0);
            for(String cmd : LOG_SKIP) {
                if (equls(STRSTR_w_R, cmd, ba, BB[0])) {
                    ba.setEnd(BB[0]);
                    lineStarts(ba);
                    break;
                }
            }
            if (0<=BB[0] && equls(STRSTR_w_R, COMMANDselect, ba, BB[0])) {
                if (isSel || equls(STRSTR_w_R, COMMANDselect, ba, BB[1])) {
                    if (!equls(STRSTR_w_R, COMMANDselection_name, ba, BB[1])) {
                        ba.setEnd(BB[0]);
                        lineStarts(ba);
                    }
                }
            }
            for(String cmd : COMMANDS_WITH_COLOR) {
                if (equls(STRSTR_w_R, cmd,line,0) && equls(STRSTR_w_R, cmd, ba, BB[1])) {
                    final int colorStart=line.indexOf('#');
                    if (colorStart>0 && equls(0L, line,0,colorStart, ba, BB[1])) {
                        ba.setEnd(BB[0]);
                        lineStarts(ba);
                    }
                }
            }
            return "";
        }
        return null;
    }

    /* <<< Log <<< */
    /* ---------------------------------------- */
    /* >>> Properties >>> */
    public static long flags(ProteinViewer v) {
        final Long f=v==null?null: (Long)v.getProperty(ProteinViewer.GET_FLAGS);
        return f==null ? 0 : f.longValue();
    }
    public static boolean supports(String command, ProteinViewer v) {
        final String[] ss=v==null?null:(String[])v.getProperty(GET_SUPPORTED_COMMANDS);
            final int b=nxt(NO_SPC,command);
            if (ss!=null && b>=0) {
                int e=command.indexOf(' ', b);
                if (e<0) e=command.length();
                for(String s : ss) {
                    if (e==b+s.length() && equls(s, command, b)) return true;
                }
            }
            return false;
    }

    public static Map getSharedHashMap(ProteinViewer pv) {
        if (pv==null) return null;
        Map m=gcp(KEY_SHARED_MAP, pv, Map.class);
        if (m==null) {
            for(ProteinViewer v : getViewersSharingView(pv)) {
                if (m==null && v!=pv) m=gcp(KEY_SHARED_MAP, v, Map.class);
            }
            if (m==null) pcp(KEY_SHARED_MAP, m=new HashMap(),pv);
        }
        return m;
    }
    /* <<< Properties <<< */
    /* ---------------------------------------- */
    /* >>> Add structure  >>> */
    public static Thread thread_addPdbId(String id_chain, Protein3d p3d, Class proteinClass) {
        return threadM("addPdbId",Protein3dUtils.class, new Object[]{id_chain,p3d,proteinClass});
    }

    public static void addPdbId(String id_chain, Protein3d p3d, Class proteinClass) {
      if (p3d==null || sze(id_chain)<0) return;
      pcp(Protein3d.KEY_MSG_NO_PROTEIN, "Downloading "+id_chain+" ...",p3d);
      final File f=downloadPdbChain(id_chain,  pdbChain(id_chain));
      inEDT(thread_addStructure(f,p3d, proteinClass));
      pcp(Protein3d.KEY_MSG_NO_PROTEIN, null,p3d);
    }

    public static Thread thread_addStructure(File f, Protein3d p3d, Class proteinClass) {
        return threadM("addStructure",Protein3dUtils.class,new Object[]{f,p3d, proteinClass});
    }
    public static void addStructure(File f, Protein3d p3d, Class proteinClass) {
        Protein p=null;
        try { p=(Protein)proteinClass.newInstance(); } catch(Exception ex){}
        if (p==null) p=new Protein();
        p.setFile(f);
        final ByteArray ba=readBytes(f);
        if (sze(ba)>0) {
            new PDB_Parser().parse(p,0L, ba);
            p3d.addProteins(p);
            setWndwState('F', p3d);
        }
    }

    /* <<< Add structure <<< */
    /* ---------------------------------------- */
    /* >>> Event >>> */
    public static void dispatchEvtPicked(ProteinViewer pv, int resNum, char chain, char insertion) {
        if (p(pv)!=null) {
            final Selection3D atoms=new Selection3D(0L, resNum, resNum, insertion, chain, null);
            //putln("dispatchEvtPicked ",selectionToText(GENERIC, new Selection3D[]{atoms}));
            new StrapEvent(pv,StrapEvent.PROTEIN_VIEWER_PICKED).setParameter(atoms).run();
        }
    }
    /* <<< Event <<< */
    /* ---------------------------------------- */
    /* >>> from Selection3D >>> */
    /* ( /b1_SaccharomycesCerevisiae and (////10-20    ////99 ) );  */
    public static String selectionToText(int viewerId, Selection3D[] ss) {
        if (countNotNull(ss)==0) return "$NONE";
        final boolean
            isP=viewerId==PYMOL,
            isJ=viewerId==JMOL,
            isA=viewerId==ASTEX,
            isR=viewerId==RASMOL,
            isG=viewerId==GENERIC,
            isJR=isJ||isR,
            isLDU= (viewerId&0xFFFF)==SURFACE_ID;

        final StringBuffer sb=new StringBuffer(99);

        if (isA || isP) sb.append(" ( ");
        for(int i=0; i<ss.length; i++) {
            final Selection3D s=ss[i];
            final String name=s.getResidueName();
            if (0<i && (isA || isP)) sb.append(" or ");
            if (isA) sb.append(" ( "); /* Bugfix */
            if ("$ALL".equals(name)) {
                sb.append(isA||isJR?" all "   : isG?name : isP?"/" : "");
            } else if ("$AMINOACID".equals(name)) {
                sb.append(
                          isA?" aminoacid " :
                          isJR? " amino " :
                          isG?name :
                          isP?" all and not het " : /* Todo no DNA no RNA */
                          "");
            } else if ("$NONE".equals(name)) {
                sb.append(isA||isJR||isP ? " none " :isG?name : "");
            } else if (name!=null) {
                if (0<i) sb.append(!isLDU ? " " : 0!=(viewerId&SURFACE_ID_KOMMA)?",":"__");
                sb.append(isJR?" [":isA?" name ":isP?" resn ":" ");
                if (isG || isA || isP || isJR) sb.append(name);
                if (isJR) sb.append("] ");
            } else {

                final char ins=s.getInsertion(), chain=spc2us(s.getChain());
                final int fst=s.getFirst(), lst=s.getLast();
                if (0<i) sb.append(isJR?" or ":!isLDU?" " : 0!=(viewerId&SURFACE_ID_KOMMA)?",":"__");

                if (isA) sb.append(" residue ");
                if (isP) {
                    sb.append("///");
                    if (is(LETTR_DIGT,chain)) sb.append(chain); /* Bug: chain is space not treated properly */
                    sb.append('/');
                }
                sb.append(fst);
                if (ins!=0) {
                    if (isJR) sb.append('^').append(ins);
                    if (isG || isLDU)  sb.append(ins);
                }
                if (fst<lst) {
                    sb.append(isA ? " to " : !isLDU || 0!=(viewerId&SURFACE_ID_DASH) ? "-" : "_").append(lst);
                    if (ins!=0) {
                        if (isJ) sb.append('^');
                        if (isJ || isG || isLDU) sb.append(ins);
                    }
                }
                if (ins!=0 || fst==lst) {
                    if (isA) sb.append(" and insertion '").append(ins==0?' ':ins).append('\'');
                    if (isP && ins!=0) sb.append(ins);
                }
                if (chain!=0) {
                    if (isJR||isG) { sb.append(':'); if (isG||chain!='_') sb.append(chain); }
                    if (isLDU)     sb.append( 0!=(viewerId&SURFACE_ID_COLON)?':':'c').append(chain);
                    if (isA)       sb.append(" and chain '").append(chain).append('\'');
                }
            }
            final String nativeA=s.getNativeAtoms();
            final String[] atoms=s.getAtoms();
            if (sze(atoms)+sze(nativeA)>0) {
                sb.append(isP?"/" : isA?" and " : isJR?" and ( " : "");
                if (sze(nativeA)>0) {
                    if (isLDU) sb.append(nativeA.hashCode());
                    else if (!isG) sb.append(nativeA);
                } else {
                    if (isA) sb.append(" atom ");
                    for(int iAt=0; iAt<atoms.length; iAt++) {
                        final String a=atoms[iAt];
                        if (0<iAt) {
                            if (isJR) sb.append(" or ");
                            if (isP)  sb.append(',');
                            if (isA)  sb.append(' ');
                        }
                        if (isLDU) sb.append(0!=(viewerId&SURFACE_ID_DOT)?'.':'a');
                        if (isJR)  sb.append("*.");
                        if (isG)   sb.append('.');
                        if (isA || isP || isG || isJR || isLDU) sb.append(a);
                    }
                }
                if (isJR) sb.append(" ) ");
                if (isG && sze(nativeA)>0) {
                    sb.append(".NATIVE").append(toStrg(Hex2bin.bin2hex(nativeA.getBytes())));
                }
            }
            if (isA) sb.append(" ) "); /* Bugfix */
        }
        if (isA || isP) sb.append(" ) ");
        return sb.toString();
    }

    public static boolean[] selectionToBoolean(Selection3D atoms[], Protein p) {
        if (p==null || sze(atoms)==0) return NO_BOOLEAN;
        final boolean bb[]=new boolean[p.countResidues()];
        for(Selection3D a : atoms) {
            final char chain=a.getChain(), ins=a.getInsertion();
            final int first=p.resNumAndChain2resIdx(false, a.getFirst(), chain, ins);
            if (first<0) continue;
            final int last=p.resNumAndChain2resIdx(false, a.getLast(), chain, ins);
            if (last<first) continue;
            setTrue(first,last+1, bb);
        }
        return bb;
    }

    public static int selectionToI(boolean fullLength, Selection3D at, Protein p) {
        return at==null||p==null ? -1 :
            p.resNumAndChain2resIdx(fullLength, at.getFirst(), at.getChain(), at.getInsertion());
    }

    /* <<< from Selection3D <<< */
    /* ---------------------------------------- */
    /* >>> to Selection3D >>> */

    public static Selection3D[] toSelectionI(int first, int last, Protein p) {
        if (p==null) return Selection3D.NONE;
        final float[] xyz=p.getResidueCalpha();
        final int rn[]=p.getResidueNumber();
        final int nR=mini(sze(xyz)/3, sze(rn),  p.countResidues());
        int i1=mini(nR-1,maxi(0,first));
        int i2=mini(nR-1,maxi(0,last));
        while(i1+1<nR && Float.isNaN(get(i1*3, xyz))) i1++;
        while(i2>0 && Float.isNaN(get(i2*3, xyz))) i2--;

        if (i1>i2 || i2>=nR || Float.isNaN(get(i1*3, xyz)) || Float.isNaN(get(i2*3, xyz))) return Selection3D.NONE;
        Selection3D sel=new Selection3D(0L,  rn[i1], rn[i2],
                                        i1==i2?chrAt(i1,p.getResidueInsertionCode()):(char)0,
                                        chrAt(i1,p.getResidueChain()),
                                        null);
        return new Selection3D[]{sel};
    }

    public static Selection3D[] toSelectionBB(boolean bb[], Protein p) {
        return toSelectionBB(false, bb,0,p,"");
    }
    public static Selection3D[] toSelectionS(ResidueSelection s) {
        return
            s==null ? Selection3D.NONE :
            toSelectionBB(false, s.getSelectedAminoacids(), s.getSelectedAminoacidsOffset(),s.getProtein(),null);
    }
    public static Selection3D[] toSelectionBB(boolean fullLen, boolean bb[], int offset, Protein p, String atoms) {
        final int last=lastTrue(bb);
        if (p==null || last<0) return Selection3D.NONE;
        final List<Selection3D> v=new ArrayList();
        final int rn[]=   fullLen?  p.getResidueNumberFullLength()       : p.getResidueNumber();
        final byte rc[]=  fullLen? p.getResidueChainFullLength()         : p.getResidueChain();
        final byte ri[]=  fullLen? p.getResidueInsertionCodeFullLength() : p.getResidueInsertionCode();
        final float xyz[]=fullLen? p.getResidueCalphaOriginalFullLength(): p.getResidueCalpha();
        final int nR=mini(sze(xyz)/3, sze(rn), mini(sze(rc), p.countResidues(), last+1+offset));
        for(int i=offset; i<nR; ) {
            if (!bb[i-offset] || Float.isNaN(xyz[i*3])) {
                i++;
                continue;
            }

            final int rn1=rn[i];
            final byte rc1=rc[i], ri1=get(i,ri);
            int i2=i+1;

            while (i2<nR && bb[i2-offset]  && !Float.isNaN(xyz[i2*3]) &&  rc1==rc[i2] && ri1==get(i2,ri)) i2++;
            if (i2<=nR) {
                v.add(new Selection3D(0L,  rn1, rn[i2-1], (char)ri1, (char)rc1,  splitAtoms(atoms)));
            }
            i=i2;
        }
        return toArry(v, Selection3D.class);
    }

    public static byte[] rplcVars(byte txt[], Protein p) {
        final int[] resNum=p==null?null:p.getResidueNumber();
        final int nR=resNum==null?0:p.countResidues();
        if (nR==0 || strchr('$',txt)<0) return txt;
        final float xyz[]=p.getResidueCalpha();
        byte[] T=txt;
        for(int iN=4; --iN>=0;) {
            final String n=iN==0?"$FIRST":iN==1?"$LAST":iN==2?"$NTERM":"$CTERM";
            if (strstr(STRSTR_w, n,T)<0) continue;
            final boolean first=iN==0||iN==2;
            int i=first?0: nR-1;
            while(Float.isNaN(get(3*i, xyz)) && i<nR && i>=0) i+=first?1:-1;
            final int rn=get(i, resNum);
            final int add=iN==2?-1:iN==3?1:0;
            final char chain=(char)get(0,p.getResidueChain());
            final String replacement=is(LETTR_DIGT,chain)?(rn+add)+":"+chain : toStrg(rn+add);
            T=rplcBB(STRSTR_w, n, replacement, T);
        }
        return T;
    }

    private static ChTokenizer _toSel_tok;

    public static Selection3D[] toSelectionT(Object text, Protein p) { return toSelectionT(false,text,p); }

    public static Selection3D[] toSelectionT(boolean convert_i2rn, Object text, Protein p) {
        final byte[] T1=text==null?null : text instanceof byte[] ? (byte[])text : ((String)text).getBytes();
        final byte[] T=rplcVars(T1,p);
        final int[] resNum=p==null?null:p.getResidueNumber();
        if (sze(T)==0) return Selection3D.NONE;
        ChTokenizer TOK=_toSel_tok;
        if (TOK==null) TOK=_toSel_tok=new ChTokenizer().setDelimiters(charClass(", "));
        TOK.setText(T, 0, MAX_INT);
        final List<Selection3D> v=new ArrayList();
        while(TOK.nextToken()) {
            String error=null, name=null;
            final int B=TOK.getFrom(), E=TOK.getTo(), endGrp=nxtE(SPC, T, B,MAX_INT);
            final int colon=strchr(':', T,B,endGrp);
            int resFrom=-1, resTo=-1;
            char insF=0, insT=0, chain=0;

            final int noWord1=nxtE(NO_LETTR_DIGT_US, T,B+1,MAX_INT);
            if (!looks(LIKE_NUM,T,B,E)) {
                name= chrAt(B,T)!='"' ? toStrg(T,B,noWord1) : toStrg(T,B+1, strchr(STRSTR_E,'"', T, B+1, noWord1));
            }
            else {
                resFrom=resTo=atoi(T,B);
                final int noDigt=nxtE(NO_DIGT, T,B+1,MAX_INT);

                if (is(LETTR, T, noDigt)) insF=insT=(char)T[noDigt];
                final int dashPos=
                    chrAt(noWord1, T)==':' && chrAt(noWord1+1, T)=='-' ? noWord1+1 :
                    chrAt(noWord1, T)==':' && chrAt(noWord1+2, T)=='-' ? noWord1+2 :
                    chrAt(noWord1, T)=='-' ? noWord1 :
                    -1;
                if (dashPos>0) {
                    if (!looks(LIKE_NUM, T, dashPos+1, E)) error="Number expected after \"-\" ";
                    else {
                        resTo=atoi(T,dashPos+1);
                        final int posIns=nxtE(NO_DIGT, T,dashPos+2, MAX_INT);
                        insT=is(LETTR, T, posIns) ? (char)T[posIns] : 0;
                    }
                }

                chain=colon>0 && is(LETTR_DIGT_US, T,colon+1) ? (char)T[colon+1] : 0;
            }
            final int atomsB=strchr('.', T,B,endGrp);
            final int atomsE=atomsB<0?-1:nxtE(SPC,T,atomsB,MAX_INT);
            final String nativeA=equls(".NATIVE",T,atomsB) ? toStrg(Hex2bin.hex2bin(T,atomsB+7,atomsE)) : null;
            final String atoms[]=nativeA!=null||atomsB<0?null : splitAtoms(T, atomsB, atomsE);

            if (convert_i2rn && colon<0 && resNum!=null) {
                resFrom=get(resFrom, resNum);
                resTo=get(resTo, resNum);
                chain=(char)get(resFrom,p.getResidueChain());
                if (resFrom==INT_NAN || resFrom>resTo) continue;
            }
            Selection3D s1=null, s2=null;
            if (name!=null) {
                s1=new Selection3D(0L, name, chain,atoms);
            } else if (insF==insT || resFrom==resTo) {
                s1=new Selection3D(0L, resFrom,resTo, insF,chain,atoms);
            } else {
                s1=new Selection3D(0L, resFrom,resFrom, insF,chain,atoms);
                s2=new Selection3D(0L, resFrom+1,resTo, insT,chain,atoms);
            }
            for(int i=2; --i>=0;) {
                Selection3D s=i==0?s1:s2;
                if (s==null) continue;
                v.add(nativeA!=null ? s.newInstanceNativeAtoms(nativeA) :s);
            }
            if (error!=null) putln("Warning in toSelectionT:", toStrg(T,B,E)," ", error);
        }
        final Selection3D ss[]=toArry(v, Selection3D.class);
        Arrays.sort(ss);
        return ss;
    }
    /* <<< to Selection3D <<< */
    /* ---------------------------------------- */
    /* >>> Atoms >>> */
    public static String[] splitAtoms(byte[] T, int B, int E) { return B>=E?NO_STRING : splitAtoms(toStrg(T,B,E)); }
    public static String[] splitAtoms(String s) {
        if (sze(s)==0) return NO_STRING;
        final Object cached[]=cached();
        Map<String,String[]> m=(Map)cached[CACHE_splitAtoms];
        if (m==null) cached[CACHE_splitAtoms]=m=new HashMap();
        String atoms[]=m.get(s);
        if (atoms==null) {
            atoms=splitTokns(0L, s, ".");
            List v=null;
            for(int iA=atoms.length; --iA>=0;) {
                final String a=atoms[iA];
                final int L=sze(a);
                final boolean asteriskB=a.charAt(0)=='*', asteriskE=L>0 && a.charAt(L-1)=='*';
                if (asteriskB || asteriskE) {
                    for(String aType : atomsOfAminos()) {
                        final int longer=sze(aType)-L+1;
                        if (longer<0) continue;
                        if (
                            asteriskB && equls(a, 1, L,   aType, longer) ||
                            asteriskE && equls(a, 0, L-1, aType, 0)
                            ) {
                            if (v==null) v=new ArrayList();
                            adUniq(aType, v);
                            atoms[iA]=null;
                        }
                    }
                }
                if (v!=null) {
                    adAllUniq(atoms, v);
                    atoms=toStrgArryO(v);
                }
            }
            m.put(s,atoms);
        }
        return atoms;
    }

    public static void changeAtoms(boolean isGeneric, Selection3D ss[], String expr) {
        final String aa[]=isGeneric ? splitAtoms(expr) : null;
        for(int i=sze(ss); --i>=0;) {
            if (ss[i]==null) continue;
            if (isGeneric && aa.length==0 && sze(ss[i].getAtoms())==0) continue;
            ss[i]=
                isGeneric ? ss[i].newInstanceChangeAtoms(aa) :
                ss[i].newInstanceNativeAtoms(expr);
        }
    }

    private static boolean sameChain(Selection3D s, char ins,char chain) {
        return s!=null&&s.getChain()==chain&&s.getInsertion()==ins && s.getResidueName()==null;
    }
    public static Selection3D[] addOrRemove(char operation, Selection3D[] ss0, int frst,int last, char ins, char chain) {
        Selection3D[] ss=ss0;
        final boolean add;
        if (operation=='!') {
            boolean a=true;
            for(int i=sze(ss); --i>=0;) {
                final Selection3D s=ss[i];
                if (!sameChain(s,ins,chain)) continue;
                final int F=s.getFirst(), L=s.getLast();
                if (F<=frst-1 && frst<=L || F<=last && last<=L) {
                    a=false;
                    break;
                }
            }
            add=a;
        } else add=operation=='+';
        boolean needsAdd=add;
        for(int i=sze(ss); --i>=0;) {
            final Selection3D s=ss[i];
            if (!sameChain(s,ins,chain)) continue;
            final int F=s.getFirst(), L=s.getLast();
            if (add) {
                if (F<=frst && last<=L) { needsAdd=false; break; }
                if (F<=frst && frst<=L+1  && L<last)  { ss[i]=new Selection3D(0L, F, last, ins, chain, null); needsAdd=false; }
                if (frst<F  && F<=last+1  && last<=L) { ss[i]=new Selection3D(0L, frst, L, ins, chain, null); needsAdd=false; }
            } else {
                if (frst<=F && L<=last) ss[i]=null;
                else {
                    final boolean wF=F<=frst-1 && frst<=L, wE=F<=last && last<=L;
                    final Selection3D
                        a1=wF && F<=frst-1 ? new Selection3D(0L, F, frst-1, ins, chain, null) : null,
                        a2=wE && last+1<=L ? new Selection3D(0L, last+1, L, ins, chain, null) : null;
                    if (wF||wE) {
                        ss[i]=a1;
                        if (a2!=null) ss=adToArray(a2,ss,0, Selection3D.class);
                    }
                }
            }
        }
        if (needsAdd) ss=adToArray(new Selection3D(0L, frst,last,ins,chain,null), ss,0, Selection3D.class);
        Arrays.sort(ss= rmNullA(ss, Selection3D.class));
        return ss;
    }

    /* <<< Atoms <<< */
    /* ---------------------------------------- */
    /* >>> main >>> */

    // java charite.christo.protein.Protein3dUtils  '10A-30:G.CA.CB'
    // java charite.christo.protein.Protein3dUtils  'ALA.CA.CB'
    /*
    public static void main(String argv[]) throws Exception {
        final Selection3D sel[]=toSelectionT(argv[0], null);
  //       putln("ONLY_LETTR_DIGT_US "+selectionToText(ONLY_LETTR_DIGT_US, sel));
         putln("GENERIC "+selectionToText(GENERIC, sel));
         putln("ASTEX "+selectionToText(ASTEX, sel));
//         putln("JMOL "+selectionToText(JMOL, sel));
//         putln("PYMOL "+selectionToText(PYMOL, sel));
    }
    */

}

