package charite.christo.protein;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import charite.christo.*;
import java.util.*;
import static charite.christo.protein.ProteinModiCount.*;
/**
   This protein model extends <i>JAVADOC:Protein</i>.
   It has the additional feature that it can be defined from a nucleotide sequence
   which is translated into amino acid sequence.
   The class ProteinNT does not depend on classes of STRAP which reside in the folder charite/christo/strap/.
   STRAP extends ProteinNT by the class <i>JAVADOC:charite.christo.strap.StrapProtein</i>.
   @author Christoph Gille
*/
// getNucleotidesCurrentStrand()
public class ProteinNT extends Protein implements charite.christo.interfaces.ChRunnable {
    private final static int CACHE_Ni_2_Ti=40+0, CACHE_Ti_2_Ni=40+1, CACHE_NCS=40+2, CACHE_RES=40+3,  CACHE_CURR_STRG=40+4, CACHE_N_STRG=40+5, CACHE_N_STRG_UC=40+6, CACHE_NR_STRG_UC=40+7;
    /* >>> Modification Count >>> */
    private final static boolean LOG=false;

    public final static boolean AMINO_ACIDS=true, NUCLEOTIDES=false;
    public final static String FORWARD="FORWARD",REVERSE="REVERSE",
        FORWARD_COMPLEMENT="FORWARD_COMPLEMENT",REVERSE_COMPLEMENT="REVERSE_COMPLEMENT",
        ORIENTATIONS[]={FORWARD,REVERSE,FORWARD_COMPLEMENT,REVERSE_COMPLEMENT};
    public boolean parseFrameShift(final byte z[], final int l) {
        if (z==null || l<3) return false;
        final int rc= z[0]+z[1];
        this.orient=
            rc=='C'+'R' ? REVERSE_COMPLEMENT :
            rc=='C'+' ' ? FORWARD_COMPLEMENT :
            rc=='R'+' ' ? REVERSE :
            FORWARD;
        Protein.incrementModificationCount(MC_CDS,this);
        final int n=getCharacters().length;
        final boolean bb[]=new boolean[n];
        Arrays.fill(bb,true);
        for(int i=mini(n,l-2);--i>=0;) {
            final byte c=z[i+2];
            bb[i]= 'A'<=c && c<='Z';
        }
        setNucleotidesCurrentStrandTranslated(bb);
        return true;
    }
    // ----------------------------------------
    private String orient;
    /** Direction translation of nucleotides.*/
    public final boolean isReverse(){ return orient==REVERSE_COMPLEMENT || orient==REVERSE;}
    /** Complement for the translation of nucleotides. */
    public final boolean isComplement(){ return orient==REVERSE_COMPLEMENT || orient==FORWARD_COMPLEMENT;}
    public final String getOrientationOfReadingFrame() { return orient;}
    /**
       Set the way nucleotides are translated
    */
    public final void setTranslatedStrand(String orientation) {
        Protein.incrementModificationCount(MC_CDS,this);
        rCached=null;
        if (orientation==null) {
            isTranslated_=null;
            delFile(strapFile(null,getName(),mDNA));
        }
        this.orient=orientation;
    }
    // ---
    private boolean[] isTranslated_;
    /**
       All nucleotides (with introns) including the non-translated (introns) ones
       @return a byte array of nucleotides or null
    */
    public final byte[] getNucleotides() {
        final byte cc[]=getCharacters();
        final String o=this.orient;
        final int mc=MODI_COUNTS[MC_CHARACTER_SEQUENCE]+MODI_COUNTS[MC_CDS];
        setModificationCount(MC_NUCLEOTIDE_TYPE, mc);
        return o!=null ? cc : null;
    }
    // ----
    /**
       The nucleotides after applying the necessary reverse and complement operations.
       The string is 0-terminated.
    */
    public final byte[] getNucleotidesCurrentStrand() {
        final Object[] cached=cached();
        final byte nucl[]=getNucleotides();
        final int n=countNucleotides();
        final int mc=MODI_COUNTS[MC_NUCLEOTIDE_TYPE]+MODI_COUNTS[MC_CDS];
        if (cached[CACHE_NCS]==null || setModificationCount(MC_NUCLEOTIDE_STRAND,mc))  {
            if (nucl==null) return null;
            final byte nu[]=isComplement() ? DNA_Util.complement(nucl,n) : nucl;
            cached[CACHE_NCS]=isReverse() ? DNA_Util.reverse(nu,n) : nu;
            if (LOG) puts("cs ");
        }
        return (byte[])cached[CACHE_NCS];
    }

    public final byte[] getNucleotidesCurrentStrandExactLength() { return getNucleotidesCurrentStrand();}
    /** This method is slow the first time after a change. Then the cached result is returned. */
    public String getNucleotidesAsString() {
        final Object[] cached=cached();
        final byte bb[]=getNucleotides();
        final int n=countNucleotides();
        String s=(String)fromCache(cached,CACHE_N_STRG, MODI_COUNTS[MC_NUCLEOTIDE_TYPE]);
        if (s==null) cached[CACHE_N_STRG]=s=new String(bb,0,n);
        return s;
    }

    /** This method is slow the first time after a change. Then the cached result is returned. */
    public String getNucleotidesAsStringUC() {
        final Object[] cached=cached();
        final String nt=getNucleotidesAsString();
        if (cached[CACHE_N_STRG_UC]==null) cached[CACHE_N_STRG_UC]=nt.toUpperCase();
        return (String)cached[CACHE_N_STRG_UC];
    }
    public String getNucleotidesReverseComplementAsStringUC() {
        final Object[] cached=cached();
        final byte bb[]=getNucleotides();
        final int n=countNucleotides();
        String s=(String)fromCache(cached,CACHE_NR_STRG_UC, MODI_COUNTS[MC_NUCLEOTIDE_TYPE]);
            if (s==null) {
                DNA_Util.reverseComplementDestructive(bb,n); // !!!!
            cached[CACHE_NR_STRG_UC]=s=toStrg(bb,0,n).toUpperCase();
            DNA_Util.reverseComplementDestructive(bb,n);
        }
        return s;
    }

    public String getNucleotidesCurrentStrandAsString() {
        final Object[] cached=cached();        
        final byte bb[]=getNucleotidesCurrentStrand();
        if (bb==null) return null;        
        final int n=countNucleotides();
        String s=(String)fromCache(cached,CACHE_CURR_STRG, MODI_COUNTS[MC_NUCLEOTIDE_STRAND]);
        if (s==null) cached[CACHE_CURR_STRG]=s=new String(bb,0,n);
        return s;
    }
    public final byte getNucleotideCurrentStrand(int i) { return i<countNucleotides() ? get(i,getNucleotidesCurrentStrand()) : -1;}
    /** Get the nucleotide of index i or -1; */
    public final byte getNucleotide(int i) { return i<countNucleotides() ? get(i,getNucleotides()) : -1;}
    /** Number of nucleotides or 0; */
    public final int countNucleotides() { return sze(getNucleotides());}
    /** Sets the sequence of nucleotides. e.g. "CAGTTGTACTG".getBytes()  */
    public final void setNucleotides(byte rawNtSequence[],String orientation) {
        setCharSequence(rawNtSequence);
        setTranslatedStrand(orientation);
    }
    /** See <i>JAVADOC:ProteinNT#setNucleotides(byte[],ProteinNT$Orientation)</i> */
    public final void setNucleotides(String nts,String orientation) { setNucleotides(nts.getBytes(),orientation);}
    /**
       Determines which nucleotide needs to be translated (coding sequence).
       True for exon and false for intron or UTR.
    */
    public final void setNucleotidesCurrentStrandTranslated(boolean bb[]) {
        rCached=null;
        isTranslated_=bb;
        Protein.incrementModificationCount(MC_ARE_NUCLEOTIDES_TRANSLATED,this);
        cdsExpressionApplied=false;
    }
    // ---
    /**
       the transformed nucleotide of index i is translated or not depending on b
    */
    public final void setNucleotideCurrentStrandTranslated(int iNt,boolean isTranslated) {
        final int n=countNucleotides();
        if (iNt>=n) return;
        boolean isTrans[]=areNucleotidesCurrentStrandTranslated();
        if (isTrans==null) { isTrans=new boolean[n];Arrays.fill(isTrans,true);}
        isTrans=chSze(isTrans,n);
        if (isTrans[iNt]!=isTranslated) {
            isTrans[iNt]=isTranslated;
            Protein.incrementModificationCount(MC_ARE_NUCLEOTIDES_TRANSLATED,this);
            isTranslated_=isTrans;
        }
    }
    /**
       True means a nucleotide is part of an triplet (exons). False means the nucleotide is not translated (intron,UTR)
    */
    public final boolean[] areNucleotidesCurrentStrandTranslated() { return isTranslated_;}
    /**
       True means a nucleotide is part of an triplet (exons). False means the nucleotide is not translated (intron,UTR)
    */
    public final boolean isNucleotideCurrentStrandTranslated(int i) {
        final boolean[] isTrans=areNucleotidesCurrentStrandTranslated();
        final int nNt=countNucleotides();
        return i>=0 && i<nNt && (isTrans==null  || (i<isTrans.length && isTrans[i]));
    }
    // ---
    private int countCodingNucleotides, _mc_cnt;
    public final int countCodingNucleotides() {
        final boolean[] isTrans=areNucleotidesCurrentStrandTranslated();
        final byte nt[]=getNucleotidesCurrentStrand();
        if (nt==null) return 0;
        if (isTrans==null) return countNucleotides();
        final int mc=MODI_COUNTS[MC_ARE_NUCLEOTIDES_TRANSLATED]+MODI_COUNTS[MC_NUCLEOTIDE_STRAND];
        if (_mc_cnt!=mc) {
            _mc_cnt=mc;
            int count=0;
            for(int i=mini(nt.length,isTrans.length);--i>=0;)
                if (isTrans[i])
                    count++;
            this.countCodingNucleotides=count;
           if (LOG) puts("count ");
        }
        return this.countCodingNucleotides;
    }
    // ---
    @Override public final byte[] getResidueTypeFullLength() {
        final Object[] cached=cached();
        final boolean isTransl[]=areNucleotidesCurrentStrandTranslated();
        final byte transformed[]=getNucleotidesCurrentStrand();
        final int mc=MODI_COUNTS[MC_NUCLEOTIDE_STRAND]+MODI_COUNTS[MC_ARE_NUCLEOTIDES_TRANSLATED]+MODI_COUNTS[MC_CDS]+
            MODI_COUNTS[MC_CHARACTER_SEQUENCE];
        if (cached[CACHE_RES]==null || setModificationCount(MC_RESIDUE_TYPE_FULL,mc)) {
            if (transformed==null) return getCharacters();
            cached[CACHE_RES]=DNA_Util.translateTriplet2aaCompact(transformed,isTransl);
            if (LOG) puts("rtf ");
        }
        return (byte[])cached[CACHE_RES];
    }
    /**
       An array mapping nucleotide indices. E.g. the 100th translated
       nucleotide may be found at position 200 of the original
       nucleotide array if there is an untranslated region of 100
       nucleotides.
       <i>JAVADOC:ProteinNT#translatedNucleotideIndices2nucleotideIndices()</i>
    */
    public final int[] nucleotideIndices2translatedNucleotideIndices() {
        return (int[])nucleotideIndices2translatedNucleotideIndices_()[CACHE_Ni_2_Ti];
    }
    private final Object[] nucleotideIndices2translatedNucleotideIndices_() {
        final Object[] cached=cached();
        final boolean isTransl[]=areNucleotidesCurrentStrandTranslated();
        final int mc=MODI_COUNTS[MC_ARE_NUCLEOTIDES_TRANSLATED];
        if (cached[CACHE_Ni_2_Ti]==null || setModificationCount(MC_NUC_IDX2TRANSLATED_IDX,mc)) {
            final int n=countNucleotides();
            final int ii[]=redim( (int[])cached[CACHE_Ni_2_Ti],n,33); Arrays.fill(ii,0,n,-1);
            final int kk[]=redim( (int[])cached[CACHE_Ti_2_Ni],n,33); Arrays.fill(kk,0,n,-1);
            for(int i=0,idx=0;i<n;i++) {
                kk[idx]=i;
                if (isTransl==null || (isTransl.length>i && isTransl[i]))
                    ii[i]=idx++;
            }
            cached[CACHE_Ti_2_Ni]=kk;
            cached[CACHE_Ni_2_Ti]=ii;
           if (LOG) puts("n2i ");
        }
        return cached;
    }
    /**
       Given is the index of a nucleotide.
       @return the index of translated nucleotides, where introns and UTRs are not counted.
    */
    public final int nucleotideIdx2translatedNucleotideIdx(int idx) {
        return idx<countNucleotides() ? get(idx, nucleotideIndices2translatedNucleotideIndices()) : MIN_INT;
    }
    /**
       An array mapping nucleotide indices. E.g. the 100th translated
       nucleotide may be found at position 200 of the original
       nucleotide array if there is an untranslated region of 100
       nucleotides.
       See <i>JAVADOC:ProteinNT#nucleotideIndices2translatedNucleotideIndices()</i>
    */
    public final int[] translatedNucleotideIndices2nucleotideIndices() {
        return (int[])nucleotideIndices2translatedNucleotideIndices_()[CACHE_Ti_2_Ni];
    }
    /**
       computes the nucleotide index from the translated idx.
       Translated are exons but not introns and UTRs
    */
    public final int translatedNucleotideIndex2nucleotideIndex(int idx) {
        return get(idx, translatedNucleotideIndices2nucleotideIndices());
    }
    /**
       Maps amino acid indices upon nucleotide indices
       See <i>JAVADOC:ProteinNT#nucleotideIndex2aminoAcidIndex(int)</i>
    */
    public final int aminoAcidIndex2nucleotideIndex(int i){
        return i<countCodingNucleotides() ? get(3*i, translatedNucleotideIndices2nucleotideIndices()) : MIN_INT;
    }
    /**
       Maps nucleotide indices upon amino acid indices
       See <i>JAVADOC:ProteinNT#aminoAcidIndex2nucleotideIndex(int)</i>
    */
    public final int nucleotideIndex2aminoAcidIndex(int i){
        return i<countNucleotides() ?  get(i, nucleotideIndices2translatedNucleotideIndices())/3 : MIN_INT;
    }
    /**
       The triplet at amino acid position iAs
       @param iAa the index of the amino acid
       @param buffer a buffer of 3 bytes
    */
    public final byte[] getResidueTriplet(final int iAa,final byte buffer[]) {
        final int i0=aminoAcidIndex2nucleotideIndex(iAa);
        final byte bb[]=getNucleotidesCurrentStrand();
        final boolean isTransl[]=areNucleotidesCurrentStrandTranslated();
        if (i0<0) return null;
        int i1=i0+1; while(isTransl!=null && i1<bb.length && !(isTransl.length>i1 && isTransl[i1])) i1++;
        int i2=i1+1; while(isTransl!=null && i2<bb.length && !(isTransl.length>i2 && isTransl[i2])) i2++;
        if (i1<0 || i2<0) return null;
        if (i2<bb.length) {
            buffer[0]=bb[i0];
            buffer[1]=bb[i1];
            buffer[2]=bb[i2];
        }
        return buffer;
    }
    public final boolean[] nucleotideSelection2aminoAcidSelection(final boolean selNts[],final int offset) {
        if (selNts==null) return NO_BOOLEAN;
        final int ii[]=nucleotideIndices2translatedNucleotideIndices(), cR=countResidues(), cR3=cR*3, cN=countNucleotides();
        if (ii==null) return NO_BOOLEAN;
        final boolean
            selAA[]=new boolean[cR],
            isRev=isReverse();
        for(int i=mini(mini(selNts.length+offset,ii.length),cN); --i>=offset;)
            if (selNts[i-offset]) {
                final int aa3=ii[isRev ? cN-i-1 : i];
                if (aa3>=0 && aa3<cR3) selAA[aa3/3]=true;
            }
        return trimArray(selAA,0);
    }
    /**
       join(7940..7996,8095..8174,8287..8368,8506..8686,
       complement(join(1787209..1787388,1787427....
       join(complement(2261..2639),complement(881..1620))
     */
    public void applyCDS_expression(String s) {
        if (s==null ) return;
        if (!isEDT()) ChDelay.invokeLater(this,"applyCDS_expression",s);
        else {
            final boolean isComp= s.indexOf("compl")>=0;
            setTranslatedStrand(isComp ? REVERSE_COMPLEMENT : FORWARD);
            final byte txt[]=s.getBytes();
            final ChTokenizer T=new ChTokenizer();
            T.setText(txt, nxt(DIGT,txt), MAX_INT);
            T.setDelimiters(charClass(",()"));
            final boolean bb[]=new boolean[sze(getCharacters())];
            while(T.nextToken()) {
                final int idx1=T.getFrom();
                final int idx2=maxi(strstr(STRSTR_AFTER,"..",txt,idx1,T.getTo()), strstr(STRSTR_AFTER,">",txt,idx1,T.getTo()));
                if (idx2<0) continue;
                final int from=atoi(txt,idx1)-1,  to0=atoi(txt,idx2);
                if (from<0 || from>=to0) {
                    //if (isMyComputer()) putln("Error in ProteinNT#applyCDS_expression "+this+" from="+from+" to="+to+" bb.length="+bb.length);
                    continue;
                }
                final int to=mini(to0,bb.length);
                if (isComp) Arrays.fill(bb, maxi(0, bb.length-to), mini(bb.length, bb.length-from),true);
                else if (from<to && from>=0 && to<=bb.length) Arrays.fill(bb, from,to,true);
            }
            setNucleotidesCurrentStrandTranslated(bb);
            cdsExpressionApplied=true;
            final String refs=get(CDS_XREFS, geneProductProteinNoteForCDS(s));
            if (refs!=null) {
                removeAllSequenceRefs();
                for(String xRef:splitTokns(0L, refs,"\t")) addSequenceRef(xRef);
            }
        }
    }
    private boolean cdsExpressionApplied;
    public boolean cdsExpressionApplied() { return cdsExpressionApplied;}
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Threads >>> */

    public Object run(String id, Object arg) {
        if (id=="applyCDS_expression") applyCDS_expression((String)arg);
        return null;
    }

}
