package charite.christo.protein;
import charite.christo.interfaces.*;
import charite.christo.*;
/**HELP

Proteins are stored in plain text files. A typical error is to save  protein files as  <b>MS-Word</b>-Document
because the proprietary WORD-format  is not recognized  by STRAP.


Recognition of file formats in STRAP works by  try-and-error disregarding of the file suffix.

<br><br>At first the <b>DSSP-format</b> is assumed which in addition to the
amino acid sequence also contains the C-alpha coordinates and the
secondary structure definition.

<i>INCLUDE_DOC:DSSP_Parser</i>

<br><br> In case the protein file does not comply with the DSSP format the    <b>PDB-format</b> (WIKI:Protein_Data_Bank_(file_format))  is tried.

<i>INCLUDE_DOC:PDB_Parser</i>

<br><br>If the file is not in PDB-format then the fasta-format is tested.

<i>INCLUDE_DOC:SingleFastaParser</i>

The first  series of non blank characters  should consist exclusively of  digits  followed  by white space of any length and an amino acid sequence.
EMBL-, WIKI:Genbank- and WIKI:Swissprot -files follow this scheme and should be parsed correctly.
The header is almost ignored. We only look for the name of the compound and the organism to create some information texts.

<br><br>

WIKI:Genbank files usually contain nucleotide sequence rather than amino
acids and nucleotides will be seen in the protein alignment.

Genbank files can be interpreted with the dialog <i>ITEM:charite.christo.strap.DialogGenbank</i>.

For other nucleotide sequences the reading frame and the translated regions can be set manually
(<i>ITEM:charite.christo.strap.EditDnaParentPane</i>).
Three nucleotide bases yield one amino acid.

<br><br>
Finally, when no specific format was recognized all letters in the file are used as one letter codes of amino acids.
<br><br><b>File compression: </b>
Files ending with <b>.gz,</b>   <b>.bz2,</b>    <b>.Z</b> or <b>.zip </b> will be decompressed automatically.

<br><br><b>Problems: </b>
<ul>
<li>pdb1a6y.ent.Z: last residue in SEQRES is a MET and is not in ATOMS</li>
<li>pdb1acb: SEQRES   1 I   70  THR GLU PHE GLY  where is it ?</li>
</ul>
<i>SEE_CLASS:PDB_Parser</i>
<i>SEE_CLASS:DSSP_Parser</i>
<i>SEE_CLASS:NumberedSequence_Parser</i>
<i>SEE_CLASS:ProteinParser</i>
<i>SEE_CLASS:SingleFastaParser</i>
<i>SEE_CLASS:StupidParser</i>
<i>SEE_CLASS:XML_SequenceParser</i>
<i>SEE_CLASS:SwissHeaderParser</i>

@author Christoph Gille
At least the amino acid sequence should be set with the Method {@link Protein#setResidueType(byte[]) setResidueType}
*/
public interface ProteinParser {
    long IGNORE_SEQRES=1, SIDE_CHAIN_ATOMS=1<<1, SEQUENCE_FEATURES=1<<2;
/**
       @return true: success, false: inappropriate file format
       @param text the entire file contents
       It is a byte array and not a String Object for performance reasons.
*/
    boolean parse(Protein p, long options, ByteArray text);
}
