package charite.christo.protein;
import charite.christo.ByteArray;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**
   Implemented by plugins that  perform a 3D-superposition of the C-alpha atoms of two protein chains.
   @author Christoph Gille
*/
public interface Superimpose3D  {
    public final static long PRINT_HUMAN_READABLE=1;
    public final static long PARSE_ALIGNMENT=1, PARSE_MATRIX=1<<1, PARSE_SCORE=1<<2, PARSE_MSA_START_S=1<<3;
    public final static int COLUMN_SEQUENCE=10;

    void setOptions(long optionFlags);
    long getOptions();
    /**
       STRAP calls this method to set the mobile protein.
    */
    void setMobileProtein(Protein p);
    /**
       STRAP calls this method to set the reference protein.
    */
    void setReferenceProtein(Protein p);
    /**
       @return The plugin should return  the  reference protein.
    */
    Protein getReferenceProtein();
    /**
       @return The plugin should return  the  mobile protein.
    */
    Protein getMobileProtein();
    /**
       STRAP calls this method to invoke computation.
       It is run in a different thread so that STRAP is still responding to user interactions.
    */
    void compute();
    /**
       returns a 3X3 rotation matrix and a translation vector.
       returns a  {dx,dy,dz} translation vector
       This method is invoked by STRAP after the compute() method returned.
    */
    Result getResult();
    /**
        The Result class bundles the result returned after computation
    */
    byte[] NO_BYTE={};
    public static class Result {
        private final int numSeq;
        public Result(int numberOfSequences) { numSeq=numberOfSequences; }

    /* ---------------------------------------- */
    /* >>> Matrix3D >>> */
        private Matrix3D m3d;
        public void setMatrix(Matrix3D m) { m3d=m;}
        public Matrix3D getMatrix() { return this.m3d; }
        /* <<< Matrix3D <<< */
        /* ---------------------------------------- */
        /*
        private boolean isInverse;
        public void setInverse(boolean b) {
            if (isInverse!=b) {
                if (m3d!=null) this.m3d=m3d.newInverse();
                isInverse=b;
            }
        }

        /* <<< Swap <<< */
        /* ---------------------------------------- */
        /* >>> Gapped Sequences >>> */
        private byte[][] gapped;
        public void setGappedSequences(byte[] ss[]) {
            assertTrue(sze(ss)>=this.numSeq);
            gapped=ss;
        }
        public byte[][] getGappedSequences() { return gapped; }
        public int[] getIndicesOfSequences() { return indicesOfProteins;}
        public void setIndicesOfSequences(int ii[]) {
            final int N=this.numSeq;
            assertTrue(ii.length>=N);
            for(int i=N; --i>=0;) {
                if (idxOf(i,ii,0,N)<0) {
                    putln(RED_ERROR+"Superimpose3D setIndicesOfSequences numSeq="+N+"  ii="+intArrayToTxt(ii,4));
                    for(byte[] bb : getGappedSequences()) putln(bb);
                    this.indicesOfProteins=count01234(ii.length);
                    stckTrc();
                    return;
                }
            }
            this.indicesOfProteins=ii;
        }
        private int[] indicesOfProteins;
        /* <<< Gapped Sequences <<< */
        /* ---------------------------------------- */
        /* >>> Score >>> */
        private float score=Float.NaN, rmsd=Float.NaN;
        public float getScore() { return score;}
        public float getRmsd() { return rmsd;}
        public void setScore(double s) { score=(float)s;}
        public void setRmsd(double s) { rmsd=(float)s;}
        /* <<< Score <<< */
        /* ---------------------------------------- */
        /* >>> Write Out >>> */
        public ByteArray toText(ByteArray sb0, long options) {
            final ByteArray sb=sb0!=null ? sb0 : new ByteArray(999);
            //sb.append(isInverse ? " Inverse:\n" : "");
            final boolean hr=(options&PRINT_HUMAN_READABLE)!=0;
            if (m3d!=null) sb.append( hr ? m3d.toText(Matrix3D.FORMAT_EXPLAIN, "mobile ") : this.m3d);
            final byte[] gg[]=getGappedSequences();
            if (sze(gg)>0) {
                final int[] ii=this.getIndicesOfSequences();
                if (!hr) sb.append("<alignment>\n");
                for(int i=0; i<gg.length; i++) {
                    final int idx=sze(ii)>=gg.length ? idxOf(i,ii,0, Integer.MAX_VALUE) : i;
                    if (idx<0 && ii!=null) {
                        System.err.println("Superimpose3D getIndicesOfSequences");
                        for(int j : ii)  System.err.println(j);
                    }
                    sb.append('s').append(idx).append(' ', COLUMN_SEQUENCE-1-stringSizeOfInt(idx));
                    sb.append(gg[idx]).append('\n');
                }
                if (!hr) sb.append("</alignment>\n");
            }
            if (!Float.isNaN(score)) sb.append(" Score=").append(score);
            if (!Float.isNaN(rmsd)) sb.append(" Rmsd=").append(rmsd,0,4);
            //if (isInverse) sb.append(" swapped");
            return sb;
        }

    }
}
