package charite.christo.strap;
import charite.christo.*;
import charite.christo.strap.interfaces.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import java.io.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**
   @author Christoph Gille
*/
public abstract class AbstractSuperposition implements IsEnabled, Disposable, CanBeStopped, HasNativeExec {
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Superimpose3D >>> */
    private Superimpose3D.Result superpositionResult;
    public Superimpose3D.Result getResult() { return superpositionResult; }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Proteins >>> */
    private transient Object refMobileProtein,refReferenceProtein;
    public void setMobileProtein(Protein p) { refMobileProtein=weakRef(p); }
    public void setReferenceProtein(Protein p) { refReferenceProtein=weakRef(p); }
    public Protein getMobileProtein() { return (Protein)deref(refMobileProtein);}
    public Protein getReferenceProtein() { return (Protein)deref(refReferenceProtein);}
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> NeedsProteins >>> */
    private Protein[] proteins;
    public void setProteins(Protein... pp) { proteins=pp; }
    /* <<< NeedsProteins <<< */
    /* ---------------------------------------- */
    /** >>> compute >>> */
    final static StringBuffer sbKey=new StringBuffer(999), sbSection=new StringBuffer(999);

    public abstract Superimpose3D.Result computeResult(byte[][] sequences, Protein[] pp);
    public void compute() {
        final long options=getOptions();
        final boolean
            isSuper=isAssignblFrm(Superimpose3D.class, this),
            isAlign=isAssignblFrm(SequenceAligner.class, this),
            isCache=CacheResult.isEnabled(),
            doSort=(cbCommutative==null || cbCommutative.s());
        final Protein
            proteins0[]=isSuper ? new Protein[]{getReferenceProtein(),getMobileProtein()} : 
        this.proteins!=null ? this.proteins.clone() : null,
            pp[]=proteins0!=null ? proteins0.clone() : null;

        final byte[][] seq, seq0=getSequences();
        final String errPfx=RED_ERROR+"AbstractSuperposition ";
        if (isSuper && (get(0,pp)==null || get(1,pp)==null) ) {
            putln(errPfx,"proteins are null ");
            putln(errPfx,"getReferenceProtein()=",getReferenceProtein(), "\n this=",this);
            stckTrc();
            return;
        }
        if (pp!=null) {
            if (doSort) sortArry(pp, ProteinUtils.comparatorSequenceAlphaIC());
            seq=Protein.getResidueType(pp);
        } else {
            seq=seq0!=null ? seq0.clone() : null;
            if (doSort) sortArry(seq, comparator(COMPARE_ALPHABET));
        }
        if (sze(seq)<2) {
            putln(errPfx,"#seq="+sze(seq));
            return;
        }

        final byte debugSeq[][]=seq.clone();
        Superimpose3D.Result result=null;
        if (isAlign && !isSuper) {
            byte[][] aligned=null;
            if ( (getOptions()&SequenceAligner.OPTION_CHECK_PERFECT_MATCH)!=0)
            {
                aligned=ChAlignUtils.gappedForPerfectMatch(seq);
                if (aligned!=null) {
                    putln(shrtClasNam(this),":"+ANSI_GREEN+"perfect match"+ANSI_RESET+"\n");
                    result=new Superimpose3D.Result(seq.length);
                    result.setGappedSequences(aligned);
                    result.setScore(SequenceAligner.SCORE_FOR_PERFECT_MATCH);
                }
            }
        }
        if (result==null) {
            synchronized(sbSection) {
                sbKey.setLength(0);
                sbSection.setLength(0);
                CacheResult.sectionForSequence(seq[0],sbSection).append(implementationID);
                if (isSuper) {
                    final int hc0=hashCd(seq[0], 0, MAX_INT, true), hc1=hashCd(seq[1],0 ,MAX_INT, true);
                    for(int i=0;i<2;i++) CacheResult.sectionForSequence(seq[i],sbKey);
                    sbKey.append(hc0).append('_').append(hc1).append('_');
                    for(int i=0;i<2;i++) sbKey.append(CacheResult.keyForCalphaAtoms(pp[i].getResidueCalpha(null))).append('-');

                } else if(isAlign) {
                    if (sze(seq)>1 && (options&SequenceAligner.OPTION_USE_SECONDARY_STRUCTURE)==0)  {
                        sbKey.append(CacheResult.keyForUngappedSeqences(seq));
                        final byte[][][] profiles=getProfiles();
                        for(int i=0; i<sze(profiles); i++) {
                            if (profiles[i]==null) continue;
                            for(byte[] gapped : profiles[i]) {
                                if (gapped==null) sbKey.append("null");
                                else sbKey.append('P').append(hashCd(gapped,0,MAX_INT,true)).append('l').append(hashCd(gapped,0,MAX_INT,true));
                            }
                        }
                    }
                }
                sbKey.append('_');
                join(getPrgParas().asStringArray(HasPrgParas.EQUALS_SIGN), "_", sbKey);
                if (isCache) {

                    final ByteArray output=CacheResult.getValue(getClass(), sbSection.toString(), sbKey.toString(), (ByteArray)null);
                    if (output!=null && checkIntegrity(output)) {
                        final long parseOptions=
                            (isSuper ? Superimpose3D.PARSE_MATRIX : 0) |
                            (isAlign ? Superimpose3D.PARSE_ALIGNMENT : 0) |
                            (isAssignblFrm(HasScore.class, this)  ? Superimpose3D.PARSE_SCORE : 0);
                        result=parse(seq.length, output, null, parseOptions);

                        if (AbstractSuperpositionDebug.errorInPairalignment(result,pp," from cache")) result=null;
                    }
                }
            }
        }
        final boolean isSwapped=isSuper &&  proteins0[0]!=pp[0];
        if (result==null) {
            result=computeResult(seq, pp);
            if (result!=null)  {
                CacheResult.putValue(getClass(), sbSection.toString(), sbKey.toString(), result.toText(null,0L),true);
                if (isMyComputer()) AbstractSuperpositionDebug.checkCache_ClassSectionKey(getClass(),sbSection.toString(), sbKey.toString(),seq);
            }
        }
        if (result!=null) {
            final Matrix3D m3d=result.getMatrix();
            if (isSwapped && m3d!=null) result.setMatrix(m3d.newInverse());
            final byte gg[][]=result.getGappedSequences();
            final int ii[]=result.getIndicesOfSequences();
            if (gg!=null) {
                final byte[][]gg0=new byte[gg.length][];
                final int[] ii0=sze(ii)==gg.length ? new int[ii.length] : null;

                final Object[] set0=pp!=null?proteins0:seq0, set=(pp!=null?pp:seq).clone();
                for(int i=seq0.length; --i>=0;) {
                    final int idx=idxOf(set0[i], set);
                    if (idx>=0) set[idx]=null;
                    gg0[i]=gg[idx];
                    if (ii0!=null) ii0[i]=ii[idx];
                }
                result.setGappedSequences(gg0);
                if (ii0!=null) result.setIndicesOfSequences(ii0);
            }
            this.superpositionResult=result;
        } else putln(RED_ERROR+" result=null ",shrtClasNam(this));

        if (isMyComputer()) AbstractSuperpositionDebug.errorInPairalignment(result,proteins0, "Am Ende ");
        if (isSuper) assert seq[0]==debugSeq[0] && seq[1]==debugSeq[1];
    }

    /* <<< Compute <<< */
    /* ---------------------------------------- */
    /* >>> Cache >>> */
    private String  implementationID="_2";
    protected void setImplementationID(String s) { implementationID="_"+s; }
    /* <<< Cache <<< */
    /* ---------------------------------------- */
    /* >>> Score >>> */
    public final float getScore() {
        final Superimpose3D.Result r=getResult();
        return r!=null ? r.getScore(): Float.NaN;
    }
    public final float getRmsd()  {Superimpose3D.Result r=getResult(); return r!=null ? r.getRmsd() : Float.NaN;}
    /* <<< Score <<< */
    /* ---------------------------------------- */
    /* >>> SequenceAligner >>> */
    private long options;
    public final void setOptions(long o) { options=o; }
    public final long getOptions() { return options;}
    public long getPropertyFlags() { return 0L;}

    /* <<< SequenceAligner  <<< */
    /* ---------------------------------------- */
    /* >>>Sequences  >>> */
    private byte[][] _sequences;
    public void setSequences(byte[]... ss) { _sequences=ss; }
    public byte[][] getSequences() {
        if (_sequences!=null) return _sequences;
        final Protein pR=getMobileProtein(), pM=getMobileProtein();
        return pM!=null && pR!=null ? new byte[][]{pR.getResidueTypeExactLength(), pM.getResidueTypeExactLength()} : null;
    }
    public byte[][] getAlignedSequences(){
        final Superimpose3D.Result r=getResult();
        return r!=null ? r.getGappedSequences() : null;
    }
    /* <<< Sequences <<< */
    /* ---------------------------------------- */
    /* >>> Sorting >>> */
    public int[] getIndicesOfSequences() {
        final Superimpose3D.Result r=getResult();
        return r!=null ? r.getIndicesOfSequences() : null;
    }
    /* <<< Sorting <<< */
    /* ---------------------------------------- */
    /* >>> Profiles >>> */
    private java.util.ArrayList<byte[][]> vProfiles;
    public void addProfile(byte ss[][]) {
        if (ss!=null) {
        if (vProfiles==null) vProfiles=new java.util.ArrayList();
        if (vProfiles.size()<getMaxNumberOfProfiles()) vProfiles.add(ss);
        }
    }
    public byte[][][] getProfiles() {  return vProfiles!=null ? toArry(vProfiles, byte[][].class) : null; }
    public int getMaxNumberOfProfiles() { return 3;}
    /* <<< Profiles <<< */
    /* ---------------------------------------- */
    /* >>> IsEnabled >>> */
    public boolean isEnabled(Object p) {
        if (this instanceof Superimpose3D) {
            return p instanceof Protein && ((Protein)p).getResidueCalpha()!=null;
        }
        return true;
    }
    /* <<< IsEnabled <<< */
    /* ---------------------------------------- */
    /* >>> Disposable >>> */
    public void dispose() { if (executable!=null) executable.dispose(); }
    /* <<< Disposable <<< */
    /* ---------------------------------------- */
    /* >>> File >>> */
    private static int countInstances;
    private File dirTmp;
    public final File dirTemp() {
        if (dirTmp==null)  {
            dirTmp=executable!=null ? executable.dirTemp() : file(dirTmp()+"/"+shrtClasNam(this)+"/"+ ++countInstances);
        }
        return dirTmp;
    }
    public final static String MFA_FILE="input.fa";
    public final File mfaInFile(byte[][] ss) {
        if (sze(ss)==0) return null;
        final File f=file(dirTemp()+"/"+MFA_FILE);
        wrte(f,ChAlignUtils.toUngappedMultipleFasta(ss,"s",0));
        return f;
    }
    public final File pdbFile(Protein p, int i, boolean res[],  long protWriterOptions) {
        final ByteArray sb=new ByteArray(p.countResidues()*120);
        final ProteinWriter1 pw=new ProteinWriter1();
        pw.selectResidues(res);
        pw.setChainForAllResidues('A');
        if ( (protWriterOptions&ProteinWriter.SIDE_CHAIN_ATOMS)!=0) p.loadSideChainAtoms(null);
        pw.toText(p,null, protWriterOptions,sb);
        final File  f=file(dirTemp(),  i==1 ?  "mobile.pdb" : "reference.pdb");
        wrte(f,sb);
        return f;
    }

    /* <<< File <<< */
    /* ---------------------------------------- */
    /* >>> Exec >>> */
    private AbstractExecutable executable;
    public boolean setExecutable(AbstractExecutable e, String srcUrls[], String instScript, String...binUrls) {
        if (!Insecure.EXEC_ALLOWED) {
            Insecure.tellNoPermission('X',"");
            return false;
        }
        e.setSourcePackageURLs(srcUrls);
        e.setSourceInstallationScript(instScript);
        e.setBinaryPackageURLs(binUrls);
        executable=e;
        e.initExecutable();
        return true;
    }
    public AbstractExecutable getNativeExec() { return executable;}
    private static ChButton cbCommutative;
    public static javax.swing.AbstractButton newCbCommutative() {
        if (cbCommutative==null) cbCommutative=toggl("Superposition is commutative").s(true).tt("A with B gives the inverse result of B with A.\nSaves computation time.");
        return cbCommutative.cb();
    }
    public void stop() {
        final AbstractExecutable e=this.executable;
        if (e!=null) e.stop();
    }
    /* <<< Exec <<< */
    /* ---------------------------------------- */
    /* >>> Command line options >>> */
    private final PrgParas prgPara=new PrgParas();
    public final PrgParas getPrgParas() { return prgPara;}
    /* <<<  Command line options <<< */
    /* ---------------------------------------- */
    /* >>>  Static utils >>> */

    public static byte[][] insertWhereNoCoordinates(int startR, byte ggR[], int startM, byte ggM[], Protein pR, Protein pM) {
        return ChAlignUtils.insertWhereNoCoordinates(startR,ggR, startM, ggM, pR.getResidueTypeLowerCaseMeansNo3dCoordinates(), pM.getResidueTypeLowerCaseMeansNo3dCoordinates());
    }
    public static Superimpose3D.Result parse(int numSeq, ByteArray txt, Superimpose3D.Result result0, long options) {
        if (txt==null) {
            putln(ANSI_RED+"Error in AbstractAligner"+ANSI_RESET +"#parse: txt is null");
            return null;
        }

        final byte[] T=txt.getBytes();
        final int E=txt.getEnd(), ends[]=txt.getLineEnds();
        final Superimpose3D.Result result= result0!=null ? result0 : new Superimpose3D.Result(numSeq);
        boolean error=false;
        if ( (options&Superimpose3D.PARSE_MATRIX)!=0) {
            final Matrix3D m3d=new Matrix3D();
            result.setMatrix(m3d);
            if (!m3d.parse(txt)) error=true;
        }
        final boolean parseScore=(options& (Superimpose3D.PARSE_SCORE|Superimpose3D.PARSE_MATRIX))!=0;
        if (parseScore) {
            float score=Float.NaN, rmsd=Float.NaN;
            String t;
            int pos;
            for(int i=E-6; --i>=1;) {
                final int c0=T[i]|32;
                if (c0=='s' && (T[i+1]|32)=='c' && !is(LETTR_DIGT_US,T,i-1)) {
                    if ( (equls(STRSTR_IC,t="score=",T,i)||equls(STRSTR_IC,t="score =",T,i)) && is(DIGT,T,pos=nxt(NO_SPC,T,i+t.length(),E))) score=(float)atof(T,pos);
                }
                if (c0=='r' && (T[i+1]|32)=='m' && !is(LETTR_DIGT_US,T,i-1)) {
                    if ( (equls(STRSTR_IC,t="rmsd=",T,i)||equls(STRSTR_IC,t="rmsd =",T,i)) && is(DIGT,T,pos=nxt(NO_SPC,T,i+t.length(),E))) rmsd=(float)atof(T,pos);
                }
                if (c0=='a' && T[i+1]=='l' && T[i+2]=='i' && chrAt(i-1,T)=='\n') { /* ClustalW */
                    if ( equls(STRSTR_IC,t="Alignment Score ",T,i) && is(DIGT,T,pos=nxt(NO_SPC,T,i+t.length(),E))) score=(float)atof(T,pos);
                }
            }
            if (!Float.isNaN(score)) result.setScore(score);
            if (!Float.isNaN(rmsd)) result.setRmsd(rmsd);
        }
        if ( (options&Superimpose3D.PARSE_MSA_START_S)!=0) {
            final MultipleSequenceParser parser=new MultipleSequenceParser().setNamesStartWith('s').setText(txt);
            final byte gg0[][]=parser.getGappedSequences();
            if (sze(gg0)>=numSeq) {
                final String names[]=parser.getSequenceNames();
                final byte gg[][]=chSze(gg0,numSeq, byte[].class);
                final byte sorted[][]=new byte[gg.length][];
                final int[] ii=new int [gg.length];
                for(int i=ii.length; --i>=0;) ii[i]=-1;
                boolean errorI=false;
                for(int i=0;i<gg.length;i++) {
                    //putln("names["+i+"]="+names[i]);
                    final int j=atoi(names[i],nxt(DIGT,names[i]));
                    if (j<sorted.length && j>=0) {
                        sorted[j]=gg[i];
                        if (ii[j]==-1) ii[j]=i;
                    } else errorI=true;
                }
                errorI|=idxOf(-1,ii,0,MAX_INT)>=0;
                if (errorI) {
                    putln(RED_ERROR+"AbstractAligner names=",names);
                    return null;
                }
                result.setIndicesOfSequences(ii);
                result.setGappedSequences(sorted);
            }
            else error=true;
        }

        if ( (options&Superimpose3D.PARSE_ALIGNMENT)!=0) {
            int aliLineStart=-1, aliLineEnd=-1;
            for(int iL=0; iL<ends.length; iL++) {
                final int b0= iL==0 ? 0 : ends[iL-1]+1,  e=prev(NO_SPC,T,ends[iL]-1,b0-1)+1, b=nxtE(NO_SPC,T,b0, e);
                if (e-b<10 || T[b]!='<') continue;
                else if (equls("<alignment>",T,b)) aliLineStart=iL+1;
                else if (equls("</alignment>",T,b)) aliLineEnd=iL;
            }
            if (aliLineStart>0 && aliLineStart<aliLineEnd) {
                byte gg[][]=null;
                int ii[]=null;
                while(true) {
                    int count=0;
                    for(int iL=aliLineStart; iL<aliLineEnd; iL++) {
                        final int b=iL==0 ? 0 : ends[iL-1]+1,  e=prev(NO_SPC,T,ends[iL]-1,b-1)+1;
                        if (e-b>1 && T[b]=='s' && is(DIGT, T, b+1)) {
                            if (gg!=null) {
                                final int spc=strchr(' ',T,b+2,e), idx=atoi(T,b+1);
                                ii[idx]=count;
                                gg[idx]=txt.subSequenceAsBytes(b+Superimpose3D.COLUMN_SEQUENCE,e);
                            }
                            count++;
                        }
                    }
                    if (gg!=null) break;
                    else {
                        gg=new byte[count][];
                        ii=new int[count];
                    }
                }
                result.setGappedSequences(gg);
                result.setIndicesOfSequences(ii);
            } else error=true;
        }
        return !error ? result : null;
    }

    private final static byte[] TAGS[]={"rotation".getBytes(),"translation".getBytes(), "alignment".getBytes()};

    public final static boolean checkIntegrity(ByteArray txt) {
        boolean ok=true;
        int end=0;
        final byte[] T=txt.getBytes();
        final int ends[]=txt.getLineEnds();
        for(byte[] tag : TAGS) {
            final int L=tag.length;
            int open=0, close=0;
            for(int i=0; i<ends.length; i++) {
                final int b=i==0?txt.getBegin():ends[i-1]+1;
                final int e=ends[i];
                if (end<e) end=e;
                if (e-b<L+2 || T[b]!='<') continue;
                if (T[b+1]=='/' && e-b>L+2 && T[b+L+2]=='>' && equls(tag,T,b+2)) close++;
                if (T[b+L+1]=='>' && equls(tag,T,b+1)) open++;
            }
            if (open!=close) {
                putln(RED_ERROR+" checkIntegrity ", tag, " "+open+" "+close);
                ok=false;
            }
        }
        return ok;
    }
}
