package charite.christo.strap;
import charite.christo.protein.HeteroCompound;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import java.util.*;
import java.awt.Component;
import javax.swing.JPopupMenu;
import static charite.christo.ChUtils.*;

public final class ContextObjects {
    private Object  lastV;
    final Object vObjects;

    private ProteinViewerPopup pvMenu;
    private ProteinPopup pMenu;
    private final HeteroPopup hMenu[]=new HeteroPopup[2];
    private final ResidueSelectionPopup sMenu[]=new ResidueSelectionPopup[3];

    private int modificationCount=-1;
    private final UniqueList vResSel[]=new UniqueList['z'+1];
    private final boolean isImmutable;
    /* <<< Fields <<< */
    /* ---------------------------------------- */
    /* >>> Constructor  >>> */
    public ContextObjects(boolean isImmutable, Object v) {
        this.isImmutable=isImmutable;
        this.vObjects=v;
        for(char c=' '; c<'z'; c++) {
            if (id2class(c)!=null)  this.vResSel[c]=new UniqueList(id2class(c));
        }
    }

    /* <<< Constructor <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */
   public ResidueSelection[] residueSelections(char A_or_S) {
        update();
        return vResSel[A_or_S]==null ? ResidueSelection.NONE : (ResidueSelection[])vResSel[A_or_S].asArray();
    }
    public StrapProtein[]          proteins() { update(); return   (StrapProtein[])vResSel['P'].asArray(); }
    public HeteroCompound[] heteroCompounds() { update(); return (HeteroCompound[])vResSel['H'].asArray();}
    public ProteinViewer[]   proteinViewers() { update(); return  (ProteinViewer[])vResSel['3'].asArray(); }

    public ProteinViewerPopup proteinViewerMenu() { if (pvMenu==null) pvMenu=new ProteinViewerPopup(this.isImmutable,this); return pvMenu;}
    public ProteinPopup proteinMenu() {if (pMenu==null) pMenu=new ProteinPopup(this); return pMenu;}
    public HeteroPopup heteroCompoundMenu() {
        if (hMenu[1]==null) hMenu[1]=new HeteroPopup(this);
        return hMenu[1];
    }
    public ResidueSelectionPopup selectionMenu(char A_or_S) {
        final int i=A_or_S=='A' ? 0 :  A_or_S=='F' ? 1 : 2;
        if (sMenu[i]==null) sMenu[i]=new ResidueSelectionPopup(this,A_or_S);
        return sMenu[i];
    }

    final javax.swing.JPopupMenu popupMenuForType(char id) {
        return
            id=='F' || id=='A' || id=='S' ? selectionMenu(id).menu() :
            id=='P' ? proteinMenu().menu() :
            id=='3' ? proteinViewerMenu().menu() :
            id=='H' ? heteroCompoundMenu().menu() :
            null;
    }

    private void update() {
        Object vSO=this.vObjects;
        if (vSO==KEY_CONTEXT_OBJECTS) {
            vSO=KEY_CONTEXT_OBJECTS[0];
            if (vSO==null)  vSO=KEY_CONTEXT_OBJECTS[2];
        }
        int mc=modiCount(vSO);
        if (vSO instanceof Object[]) {
            for(Object o0 : ((Object[])vSO))  {
                final Object o=deref(o0);
                if (o!=null) mc+=o.hashCode();
            }
        }
        if (this.modificationCount!=mc  || lastV!=vSO || mc<0) {
            final Object oo[]=oo(getSelValues(vSO));
            this.modificationCount=mc;
            lastV=vSO;
            for(UniqueList vS : vResSel) if (vS!=null) vS.clear();
            for(int i=0; i<oo.length; i++) {
                final Object o=get(i,oo);
                if (o instanceof UniqueList) for(Object el : ((UniqueList)o).asArray()) myAdd(el);
                myAdd(o);
            }
        }
    }
    private final void myAdd(Object o) {
        final char id=menuID(o);
        if (id!=0) {
            if (id=='F' || id=='A' || id=='S') vResSel[' '].add(o);
            vResSel[id].add(o);
        }
    }
    /* <<< Update  <<< */
    /* ---------------------------------------- */
    /* >>> Static utils >>> */

    public static char menuID(Object o) {
        return
            o instanceof HeteroCompound ? 'H' :
            o instanceof ResidueAnnotation ? ( ResidueAnnotation.isAnno(o) ? 'A' : 'F') :
            o instanceof StrapProtein ?  'P' :
            o instanceof UniqueList &&  ((UniqueList)o).getClazz()==ResidueAnnotation.class ? 'G' :
            isInstncOf(ResidueSelection.class, o) ?  'S' :
            isInstncOf(ProteinViewer.class,o) ?  '3' :
            (char)0;
    }

    public static Class id2class(char c) {
        return
            c=='A' || c=='F' ? ResidueAnnotation.class :
            c=='S' || c==' ' ? ResidueSelection.class :
            c=='G' ? UniqueList.class :
            c=='P' ? StrapProtein.class :
            c=='3' ? ProteinViewer.class :
            c=='H' ? HeteroCompound.class :
            null;

    }
    // --------------------------------------------------

    public static void showContextMenu(char menuType, Object[] ooMouse, Object[] ooSelected) {
        if (sze(ooMouse)==0) return;
        assertNotNull(id2class(menuType));
        final java.awt.AWTEvent mev=getLastMouseEvt();
        final Object q=mev==null?null:mev.getSource();
        if (q!=null) undockSetTarget(gcp(KEY_UNDOCK_TARGET,q, ChJList.class));

        final Set v=new HashSet();
        for(Object o : ooMouse) if (menuID(o)==menuType) v.add(o);
        if (cntainsAtLeastOne(ooMouse,ooSelected)) {
            for(Object o : ooSelected) if (menuID(o)==menuType) v.add(o);
        }
        KEY_CONTEXT_OBJECTS[0]=v.toArray();
        final JPopupMenu jpm=StrapAlign.getContextObjects().popupMenuForType(menuType);
        shwPopupMenu(jpm);
        //if (jpm!=null) jpm.show(component,x,y);
    }
}
