package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import java.awt.*;
import java.util.ArrayList;
import java.awt.event.*;
import javax.swing.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP
   The Dialog is used to align a
   number of selected proteins (see WIKI:Sequence_alignment).

   The user selects an alignment method <i>COMBO:charite.christo.strap.extensions.MultipleAlignerClustalW</i>.

   By pressing the button <i>LABEL:ChButton#BUTTON_GO</i> the alignment
   computation is started. It takes some seconds to minutes, before the
   computation the result is shown in a preview panel.

   To accept the alignment in the preview and to infer the alignment into
   the working alignment panel
   <i>BUTTON:DialogAlignResult#BUT_LAB_Apply</i> needs to be pressed to.

   <br><br><b>Types of alignment procedures:</b>

   So-called <b>local</b> alignment procedures identify local matches
   whereas <b>global</b> alignment algorithms align the entire sequences.

   Some alignment programs align <b>two</b> sequences only. The derived multiple
   alignments are not as good as those from methods that
   natively support multiple alignments.

   <b>3D-superposition</b> programs consider only the 3D-structure and do not
   regard the residue type.

   <br><br><b>Profiles:</b> ClustalW and T-coffee can use pre-aligned
   sequences as a so-called sequence profile. For more information
   see the manual of T-coffee.

   <i>SEE_DIALOG:DialogAlignOneToAll</i>

   @author Christoph Gille
*/
public final class DialogAlign extends AbstractDialogJTabbedPane implements StrapListener, ActionListener, ChRunnable {
    private final AbstractButton

        cbSecStru=cbox("Use secondary structure"),
        radioAllProteins[]=radioGrp(new String[]{"Selecting proteins", RADIO_REGION , RADIO_SELECTION},0,this);
    private final ChButton togProfile[]=new ChButton[3];
    public final static String
        RADIO_REGION="Rubber band selection",
        RADIO_SELECTION="Text list of protein names";
    final JComponent pCenter[]=new JComponent[3], pMain;
    private final ChTextArea
    taResidueSelection=new ChTextArea("").saveInFile("DialogAlign_namesResSel"),
        taProfile[]=new ChTextArea[3];
    private final ProteinList proteinList=new ProteinList(align,0).selectAll(0);
    private final ChCombo comboClass=StrapAlign.classChoice(SequenceAligner.class).save(DialogAlign.class,"SequenceAligner").li(this);
    private int resultNumber=1;
    /* <<< Fields <<< */
    /* ---------------------------------------- */
    /* >>> Constructor  >>> */

    public DialogAlign() {
        pcp(KEY_CHOICE_CLASS,comboClass,this);
        pcp(KEY_ENABLED,comboClass,proteinList);

        pCenter[0]=proteinList.scrollPane();
        pCenter[1]=vBhB("The residues within the rubber band selection are taken.",
                        "It is shown as red-white  \"Marching ants\" in the alignment pane.",
                        "It looks like:"
                        );
        addMarchingAnt(ANTS_SEARCH, weakRef(pCenter[1]), new Rectangle(10*EM,2*EX, 5*EM, EX), C(0xffFFff),C(0xFF00));
        final Object msgList=
            "Enter a list of proteins. Optionally, specify residue ranges by appending slash residue selection like \"myProtein/100-200\".<br>"+
            "You may also give a named residue selection like \"myProtein/myResidueSelection\".";
        addActLi(this,taResidueSelection);
        pCenter[2]=newCNSEW(scrllpn(0L,taResidueSelection),msgList, null,null,null);
        taResidueSelection.tools()
            .enableWordCompletion(new Object[]{align.vProteins, ResidueAnnotation.vAll})
            .highlightOccurrence(align, charClass(SPC), charClass(NO_LETTR_DIGT_US), HIGHLIGHT_UPDATE_IF_TEXT_CHANGES, C(0xFFff14,90));
        final JComponent
            panProfile=pnl(new GridLayout(1,togProfile.length)),
            panDetails=vBhB("#ETBem",
                                hBL("Method: ",comboClass.panel()),
                                buttn(TOG_CACHE).cb(),
                                hBL("Use some pre-aligned sequences as a profile.", panProfile),
                                hBL("Currently only for T-coffee:", cbSecStru),
                                (isSystProprty(IS_LINUX386) || isSystProprty(IS_WINDOWS)) ? null :  "Note: some alignment methods require certain packages to be installed."
                              ),
            pSouth=vB(
                      hBL("Options: ",toggl().actionCollapse(false,panDetails), pnl(new ChButton("GO").t(ChButton.GO).li(this))),
                      panDetails
                      ),
            pNorth=vBpnl(
                         StrapAlign.headLine(this),
                         hBL(" ",hB("#TB You can specify the proteins to be aligned by ...",radioAllProteins)," ")
                         );
        pMain=newCNSEW(null, pNorth,pSouth, null,null);

        setCenter();
        for(int i=0; i<togProfile.length; i++) panProfile.add((togProfile[i]=toggl("profile "+(i+1)).li(this)).cb());

        adMainTab(pMain,this, null);
    }
    /* <<< Constructor <<< */
    /* ---------------------------------------- */
    /* >>> Profiles >>> */
    private static boolean takesProfile(Object a, boolean msg) {
        if (!isAssignblFrm(SequenceAlignerTakesProfile.class,a)) {
            if (msg) {
                StringBuffer s=new StringBuffer(shrtClasNam(a)+" does not take profiles.<br>Only the following methods can work with alignment profiles:<ul>");
                for(Object o : derefArray(StrapPlugins.allClassesV(SequenceAligner.class).asArray(), SequenceAlignerTakesProfile.class))
                  s.append("<li>").append(shrtClasNam(o)).append("</li>");
                error(s.append("</ul>"));
            }
            return false;
        }
        return true;
    }
    public void addProfiles(SequenceAligner aligner) {
        StringBuffer sbIgnore=null, sbError=null;
        final Color COLOR_IGNORE=C(0xFF0000,66),  COLOR_OK=C(0x00ff00,66);
        int count=0;
        for(int i=0; i<togProfile.length; i++) {
            if (!isSelctd(togProfile[i]) || !takesProfile(aligner,true)) continue;
            final SequenceAlignerTakesProfile alignerP=(SequenceAlignerTakesProfile)aligner;
            final int maxProfile=alignerP.getMaxNumberOfProfiles();
            final ChTextArea ta=taProfile[i];
            final ByteArray ba=ta.tools().byteArray();
            final int ends[]=ba.getLineEnds();
            final byte[] T=ba.getBytes();
            wrte(file(DIR_ANNOTATIONS+"/profile"+i),T,0,ba.getEnd());
            final ArrayList<byte[]> vGapped=new ArrayList();
            int blockBegin=0;
            for(int iL=0; iL<ends.length;  iL++) {
                final int b= iL==0 ? 0 : ends[iL-1]+1;
                final int e=prev(NO_SPC,T,ends[iL]-1,b-1)+1;
                int countLetters=0;
                Color color=null;
                for(int iT=b; iT<e; iT++) {
                    final byte c=T[iT];
                    if (is(LETTR,c)) countLetters++;
                    else if (c!=' ' && c!='.' && c!='-') {
                        countLetters=-1;
                        if (sbIgnore==null) sbIgnore=new StringBuffer().append("The following lines are ignored: "); else sbIgnore.append(", ");
                        color=COLOR_IGNORE;
                        sbIgnore.append(iL);
                        break;
                    }
                }
                if (countLetters>0) {
                    final byte seq[]=new byte[e-b];
                    for(int iT=b; iT<e;iT++) seq[iT-b]=is(LETTR,T,iT) ? T[iT] : (byte)'-';
                    if (sze(vGapped)==0) blockBegin=b;
                    vGapped.add(seq);
                }
                if (countLetters==0 || iL==ends.length-1) {
                    final byte[][] alignment=toArry(vGapped,byte[].class);
                    if (alignment.length==1) {
                        (sbError==null ? sbError=new StringBuffer() : sbError).append("Error line "+iL+" profile has only one sequence.\n");
                    }
                    if (alignment.length>1) {
                        if (++count>maxProfile) StrapAlign.errorMsg(shrtClasNam(aligner)+" takes not more than "+maxProfile+" profiles.");
                        else {
                            alignerP.addProfile(alignment);
                            color=COLOR_OK;
                        }
                    }
                    vGapped.clear();
                }
                if (color!=null) {
                    try {
                         ta.getHighlighter().addHighlight(color==COLOR_OK ? blockBegin : b,e,ChTextArea.highlightPainter(color));
                    } catch(Exception ex){}
                }
            }
        }
        if (sbIgnore!=null) (sbError==null ? sbError=new StringBuffer() : sbError).append("\n").append(sbIgnore);
        if (sbError!=null) error(sbError);
    }

    /* <<< Profiles <<< */
    /* ---------------------------------------- */
    /* >>> ActionEvent >>> */
    @Override public void actionPerformed(ActionEvent ev) {
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        final boolean isSelctd=isSelctd(q);
        if (q==comboClass) proteinList.repaint();
        for(int i=togProfile.length; --i>=0;) {
            if (q==togProfile[i] && isSelctd) {
                if (taProfile[i]==null) {
                    final ChTextArea ta=taProfile[i]=new ChTextArea(readStrg(DIR_ANNOTATIONS+"/profile"+i));
                    final Object msg=
                        "Enter a sequence alignment without protein name. This alignments will be used as a profile.<br>"+
                        "More than one profile can be given by separating the alignments by a blank line.<br>"+
                        "A rectangular region (\"walking ants\") in the alignment pane can be inserted as a profile: <br>"+
                        "Open a rectangular region by dragging the mouse. right-click the box and select \"Copy\"  in the context menu. Type CTRL-V to insert.",
                        c=newCNSEW(scrllpn(0L,ta),msg, null,null,null);
                    addActLi(this,ta);
                    ta.saveInFile("DialogAlign_profile"+i);
                    pcp("MSG",c,ta);
                }
                ChFrame.frame(getTxt(q), gcp("MSG",taProfile[i]),CLOSE_ON_CtrlW).shw(ChFrame.AT_CLICK);
            }
        }

        if (cmd=="GO") tryAlignment();
        setCenter();
    }

    /* <<< AWTEvent <<< */
    /* ---------------------------------------- */
    /* >>> Compute >>> */
    private void tryAlignment() {
        final Protein[] pp,ppLocal;
        switch(radioGrpIdx(radioAllProteins)) {
        case 1:
            final Rectangle rect=StrapView.getRectangle(null);
            pp=rect==null ? null : StrapView.getProteinsInRectangle(null);
            ppLocal=new Protein[sze(pp)];
            for(int iP=ppLocal.length; --iP>=0;) {
                ppLocal[iP]=StrapProteinUtils.newProteinColumnFromTo((StrapProtein)pp[iP],rect.x, rect.x+rect.width);
            }
            break;
        case 2:
            final Object[] pp_ss=StrapAlign.getProteinsAndAnnotationsWithRegex(toStrg(taResidueSelection),align.getProteins(),'A');
            final ResidueSelection ss[]=derefArray(pp_ss,ResidueSelection.class);
           
            final ArrayList vP=new ArrayList(pp_ss.length);
            for(ResidueSelection s : ss) adUniq(s.getProtein(), vP);
            adAllUniq( derefArray(pp_ss,Protein.class),vP);
            pp=toArry(vP, Protein.class);
            ppLocal=new Protein[pp.length];
            for(int iP=pp.length; --iP>=0;) {
                final Protein pL=StrapProteinUtils.newProteinNarrowToResidueSelection(pp[iP],ss);
                ppLocal[iP]=pL!=null ? pL : pp[iP];
            }
            break;
        default:
            pp=ppLocal=proteinList.getSelectedProteinsAtLeast(2);
            break;
        }

        if (sze(pp)<2) { error("Required at least 2 proteins."); return; }
        final SequenceAligner aligner=StrapPlugins.instanceSA(comboClass, sze(pp),true);
        if (aligner==null)  { error("Could not instantiate the alignment class.");return;}
        final Object o=aligner instanceof MultiFromPairAligner ? ((MultiFromPairAligner)aligner).getClasses()[0] : aligner;
        if (isAssignblFrm(Superimpose3D.class,o)) {
            for(Protein p : pp) {
                if (p.hasCalpha()) continue;
                error("The method<br>"+delSfx("PROXY",shrtClasNam(o))+"<br>requires 3D-structures for all proteins");
                break;
            }
        }
        StrapProteinUtils.setSeqs(aligner,ppLocal);
        addProfiles(aligner);
        aligner.setOptions(isSelctd(cbSecStru)?SequenceAligner.OPTION_USE_SECONDARY_STRUCTURE:0L);
        final DialogAlignResult r=new DialogAlignResult(aligner,pp, DialogAlignResult.USE_THREADS);
        TabItemTipIcon.set("preview "+resultNumber++,null,null,null,r);
        ChDelay.addTab(this,r,333);
    }
    /* <<< Compute <<< */
    /* ---------------------------------------- */
    /* >>> StrapEvent >>> */
    public void handleEvent(StrapEvent e) {
        final int t=e.getType();
        if (t==StrapEvent.RESIDUE_SELECTION_CHANGED || t==StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR || t==StrapEvent.AA_SHADING_CHANGED) {
            for(DialogAlignResult c : childs(getSelectedComponent(),DialogAlignResult.class)) c.repaint();
        }
    }

    /* <<< StrapEvent <<< */
    /* ---------------------------------------- */
    /* >>> Thread >>> */
    public Object run(String id, Object arg) {
        final Object argv[]=arg instanceof Object[] ? (Object[])arg : null;
        final int iRadio=id==RADIO_REGION ? 1 : id==RADIO_SELECTION ? 2 : -1;
        if (iRadio>=0) {
            radioGrpSet(iRadio,radioAllProteins);
            if (id==RADIO_SELECTION) taResidueSelection.t(toStrg(arg));
            setCenter();
        }
        return null;
    }
    private JComponent pCenterCurrent;
    private void setCenter() {
        final JComponent c=pCenter[maxi(0,radioGrpIdx(radioAllProteins))];
        if (pCenterCurrent!=c) {
            rmFromParent(pCenterCurrent);
            pMain.add(pCenterCurrent=c, BorderLayout.CENTER);
        }
    }

}
