package charite.christo.strap;
import charite.christo.interfaces.*;
import charite.christo.*;
import charite.christo.protein.*;

import charite.christo.strap.interfaces.*;
import java.awt.*;

import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP

With this dialog a reference protein can be aligned to many other
proteins producing a pair-alignments for each protein.

On pressing <i>LABEL:ChButton#BUTTON_GO</i>  the  sequence alignments appear in the preview.

<br><br><b>Inferred into the working alignment</b>
Sequence alignments can be  applied to the working sequence alignment.
Only the gaps of the reference protein are changed in the working alignment.
The other proteins are not changed.

<br><br><b>Highlighting of the match:</b>
Table rows can be selected using the <i>ICON:IC_SHOW</i> toggles.
The matches of selected rows are highlighted in the alignment.

<br><br><b>Limitations:</b>
The induction of gaps from the calculated sequence alignment into the working
sequence  alignment may fail because only the residue positions of the
reference protein but not of the other sequences are changed by
this procedure.

Therefore, it might not find sufficient white space in the multiple sequence
alignment.

<i>SEE_DIALOG:DialogAlign</i>
<i>SEE_DIALOG:DialogSuperimpose3D</i>
@author Christoph Gille
*/
public class DialogAlignOneToAll extends AbstractDialogJTabbedPane implements StrapListener, java.awt.event.ActionListener {

    private final ProteinCombo comboProtein;
    private final ProteinList proteinList=new ProteinList(StrapAlign.getInstance(),0).selectAll(0);
    private final ChCombo comboClass=StrapAlign.classChoice(SequenceAligner.class).save(DialogAlign.class,"SequenceAligner");

    public void actionPerformed(java.awt.event.ActionEvent ev) {
        StrapAlign.errorMsg("");
        final String cmd=ev.getActionCommand();
        if (cmd=="GO") {
            final StrapProtein pp[]=proteinList.getSelectedOrAllProteins();
            if (pp.length==0) { StrapAlign.errorMsg("Select at least one protein");return;}
            if (pp.length==0) return;
            final Container c=new DialogAlignOneToAllResult(pp,comboProtein.getProtein(),comboClass);
            adTab("<sup>"+comboProtein+"<br>"+shrtClasNam(comboClass)+"</sup>",c, this);
        }
    }
    public DialogAlignOneToAll() {
        comboProtein=new ProteinCombo(align,0);
        final Container
            pWestNorth=vBhB(
                            "Reference protein:", comboProtein,
                            "Alignment Method:",comboClass.panel(),
                            pnl(new GridLayout(1,2),new ChButton("GO").t(ChButton.GO).li(this),buttn(TOG_CACHE).cb())
                            ),
            pWest=newCNSEW(null,pWestNorth,null,null,KOPT_TRACKS_VIEWPORT_WIDTH),
            pCenter=pnl(new GridLayout(1,2),scrllpn(0L,pWest),proteinList.scrollPane()),
            pMain=newCNSEW(pCenter,StrapAlign.headLine(this),null,null,null);
        adMainTab(pMain,this,null);
        pcp(StrapAlign.KEY_CHOICE_PROTEIN,comboProtein,this);
    }
    //    private DialogAlign dialogAutoAlign=new DialogAlign(null);

    public void handleEvent(StrapEvent e) {
        final int t=e.getType();
        if (t==StrapEvent.RESIDUE_SELECTION_CHANGED || t==StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR) {
            ChDelay.repaint(this,333);
        }
    }
}
