package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.strap.interfaces.*;
import charite.christo.protein.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP

The bar chart visualizes values assigned to single proteins.

In the bar chart dialog the Java class must be selected that performs the calculation.

As a simple test  "number of residues in proteins" can be chosen.

The bar chart is shown within the row header of the alignment window.

<br>
There are two possibilities:
<ol>
  <li>
    The user chooses a class implementing  <i>JAVADOC:ValueOfProtein</i>.
  </li>
  <li>
    The user chooses a class implementing  <i>JAVADOC:ProteinProteinDistance</i> and a
    reference protein.
  </li>
</ol>
<i>SEE_DIALOG:DialogCompareProteins</i>
   @author Christoph Gille
*/
public class DialogBarChart extends AbstractDialogJPanel  implements VisibleIn123,StrapListener,ActionListener   {
    private final ChCombo comboVOP,comboPPD;
    private final ProteinCombo comboOtherProtein;
    private final ChTextField tfFrom,tfTo;
    private final ChButton butOnVOP,butOnPPD;
    // ----------------------------------------
    /* --- Interface VisibleIn123 --- */
    private static Color color=Color.GREEN;
    public Color getColor() { return color;}
    public void setColor(Color c) { color=c;}
    public int getStyle() { return VisibleIn123.FULL_BACKGROUND;}
    public int getVisibleWhere() { return VisibleIn123.SEQUENCE;}
    public void setVisibleWhere(int w){}
    // ----------------------------------------
    private void myUpdate() {
        //align.removeListener(instance);
        final StrapView view=StrapAlign.getInstance().getAlignmentPanelWithFocus();
        if (view==null) return;
        butTable.enabled(false);
        final boolean isPPD=butOnPPD.s(), isVOP=butOnVOP.s();
        final Object c=!isVOP && !isPPD?null : (isPPD?comboPPD:comboVOP).getSelectedItem();
        final Object sharedInstance= !isVOP && !isPPD?null : (isPPD ? comboPPD : comboVOP).instanceOfSelectedClass(null);
        view.setBarchart(c, sharedInstance, atoi(tfFrom,0), atoi(tfTo), isPPD ? comboOtherProtein.getProtein():null);
        if (sharedInstance!=null) {
            butTable.enabled(true);
            if (sharedInstance instanceof StrapListener) align.addListener((StrapListener)sharedInstance);
        }
        view.getRowHeader().repaint();
        StrapAlign.dispatchLater(StrapEvent.VALUE_OF_PROTEIN_CHANGED,1);
    }
    // ----------------------------------------
    public void actionPerformed(ActionEvent ev) {
        StrapAlign.errorMsg("");
        final Object q=ev.getSource();
        if (q==butTable)  table();
        if (q==butOnVOP && butOnVOP.s()) butOnPPD.s(false);
        if (q==butOnPPD && butOnPPD.s()) butOnVOP.s(false);
        myUpdate();
    }
    private final ChButton butTable=new ChButton(null,ChButton.DISABLED).t("As table").li(this);
    private void table() {
        final StringBuffer sb=new StringBuffer();
        String title="";
        for(StrapProtein p:align.getVisibleProteins()) {
            final ValueOfProtein  vop=p.getValueOfProtein();
            if (vop==null) continue;
            vop.setProtein(p);
            title=vop.getClass().getName();
            sb.append(p).append('\t').append(((float)vop.getValue())).append('\n');
        }
        if (sb.length()>0) new ChTextViewer(sb).title(title).shw(ChFrame.TO_FRONT);
        else StrapAlign.errorMsg("No values to show");
    }
    // ----------------------------------------
    public DialogBarChart() {
        tfFrom=new ChTextField("  1").cols(6,true,true).li(this).ct(Integer.class);
        tfTo=new ChTextField("9999").cols(6,true,true).li(this).ct(Integer.class);
        butOnVOP=toggl("Show chart",ChButton.SHOW_HIDE).li(this);
        butOnPPD=toggl("Show chart",ChButton.SHOW_HIDE).li(this);
        setUndockble(butOnVOP);
        setUndockble(butOnPPD);
        comboVOP=StrapAlign.classChoice(ValueOfProtein.class).li(this);
        pcp(KEY_CHOICE_CLASS,comboVOP,this);
        comboPPD=StrapAlign.classChoice(ProteinProteinValue.class).li(this);
        (comboOtherProtein=new ProteinCombo(align,0)).li(this);
        JComponent
            pVoP=vBpnl(
                       Color.WHITE,
                       getBackground().darker(),
                       pnl("Computation method",comboVOP.panel()),
                       pnl(enlargeFont(butOnVOP,2f),butTable),
                       "#ETBem"
                       ),
            pVoAP=vBpnl(
                        Color.WHITE,
                        pnl("Computation method",comboPPD.panel()),
                        pnl("Reference protein " ,comboOtherProtein),
                        pnl("Alignment position from ",tfFrom," to ",tfTo,enlargeFont(butOnPPD,2f),butTable.cln()),
                        "#ETBem"
                        );
        pVoAP=pnl(pVoAP);
        pVoP=pnl(pVoP);
        Container pCenter=vB(KOPT_TRACKS_VIEWPORT_WIDTH,pVoP,pVoAP);
        cnsew(this,scrllpn(0L,pCenter),pTop(),null,null,null);
        myUpdate();
    }
    public void handleEvent(StrapEvent ev){
        final int t=ev.getType();
        final StrapView view=StrapAlign.getInstance().getAlignmentPanelWithFocus();
        if(t==StrapEvent.VALUE_OF_PROTEIN_CHANGED && butOnVOP.s() || t==StrapEvent.PROTEIN_PROTEIN_DISTANCE_CHANGED && butOnPPD.s() ) {
            if (view!=null)  view.getRowHeader().repaint();
        }
    }
    @Override public void dispose() {
        final StrapView view=StrapAlign.getInstance().getAlignmentPanelWithFocus();
        align.removeListener(this);
        if (view!=null) view.setBarchart(null,null,0,0,null);
    }
}
