package charite.christo.strap;
import charite.christo.interfaces.*;
import charite.christo.*;
import charite.christo.blast.*;
import charite.christo.protein.*;
import java.util.*;
import java.io.*;
import java.awt.event.*;
import javax.swing.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP

STRAP allows to search public and private sequence databases to
identify sequences that are similar to a query sequences.

Depending on the type of sequences, there are four search methods:
<ol>
<li>The query is an amino acid (aa) sequence and the database contains nucleotide (nt) sequences.</li>
<li>  ...   nt   ...  nt. ...  </li>
<li>  ...   aa   ...  aa  ...  </li>
<li>  ...   nt   ...  aa  ...  </li>
</ol>

Most users are already familiar with WIKI:BLAST Web services which are
accessible with a Web browser.

STRAP communicates with the Web services and presents the results.
It provides functionality that is not available in Web-browsers:

<ul>

<li>Many jobs can be started simultaneously.</li>

<li>The Blast result is saved in the cache. If  the same query is used a
second time the result is loaded from the cache and appears immediately.
</li>

<li>
Blast alert:

The sequence databases are permanently
growing.

With the cache disabled   STRAP conducts a new Blast
and compares the new result with the previous result stored in the cache.

It reports if there are new hits since last run and  highlights all new hits.
</li>

<li>In STRAP several patterns can be highlighted in different colors.
STRAP identifies patterns even if they are disrupted  by a gap.
</li>

<li>Sequences from the hit-list can be loaded  with one click.</li>
</ul>
   See UBIC:blastall UBIC:blat UBIC:wu_blast
   @author Christoph Gille
*/
public class DialogBlast extends AbstractDialogJTabbedPane implements ActionListener,ChRunnable  {
    private final static String SELECTION_NAME="blast_match";
    private final ProteinList proteinList;
    private final List[] vDatabases={new Vector(),new Vector()};

    private final ChButton
    togAminoDB=toggl("Database of Amino acid sequences").s(true).li(this),
        butGo=new ChButton("GO").t(ChButton.GO).li(this).tt("Invokes BLAST computation.");

    private final ChCombo
        comboClass=StrapAlign.classChoice(SequenceBlaster.class).li(this).save(DialogBlast.class,"SequenceBlaster"),
        comboNT=new ChCombo("Aminos or nucleotides: Strap's guess","Query is nucl sequ with introns","Query is the coding nucleotide sequence").li(this),
        comboDB[]={ new ChCombo(0L,vDatabases[0]), new ChCombo(0L,vDatabases[1])};
    private final ChLabel
    labFirst=new ChLabel("first residue "),
        labLast=new ChLabel(" last residue "),
        labStatus=new ChLabel();
    private transient SequenceBlaster lastRunning;
    private final Object[] radioDatabase;
    public DialogBlast() {
        pcp(KEY_CHOICE_CLASS,comboClass,this);
        (proteinList=new ProteinList(align,0)).li(this);
        pcp(StrapAlign.KEY_CHOICE_PROTEIN, weakRef(proteinList), this);
        proteinList.li(this);
        comboClass.setPreferredSize(dim(prefWidth(comboClass), prefHeight(butGo)));
        updateDatabases();
        radioDatabase=radioGrp(new String[]{"Proteins with 3D-structure", "All proteins", "Non-redundant set of proteins"},0,this);
        final JComponent
            pSimple= vBhB(
                          "Searching for similar sequences in:",
                          hBL(" ",vB(radioDatabase)),
                          " "
                          ),
            pSettings=vBhB(
                           pnl(new java.awt.GridLayout(2,2),hB("#",labFirst),tfFrom,hB("#",labLast),tfTo),
                           hBL("Database:  ", togAminoDB.cb()),
                           comboDB[1],
                           comboDB[0],
                           "Blast service:",
                           hBL(" ",comboClass.panel()),

                           hBL(" ",comboNT),
                           buttn(TOG_CACHE).cb()
                           ),
            pSide=vBpnl(
                        KOPT_TRACKS_VIEWPORT_WIDTH,
                        pSettings,
                        pSimple,
                        hBL("Details ",toggl().actionCollapse(false,pSettings).actionCollapse(true, pSimple),"#",pnl(butGo)),
                        labStatus
                        ),
            pLeftRight=pnl(new java.awt.GridLayout(1,2),proteinList.scrollPane(),pSide),
            pMain=newCNSEW(pLeftRight,StrapAlign.headLine(this),null,null,null);
        togAminoDB.actionCollapse(false, comboDB[1] ).actionCollapse(true, comboDB[0]);
        adMainTab(pMain,this,null);
    }

    private final ChTextField
        tfOutputWidth=new ChTextField("999999").cols(7,true,true),
        tfFrom=new ChTextField("1").cols(6,true,true),
        tfTo=new ChTextField("9999").cols(6,true,true)/*, tfMaxLen=new ChTextField("  3000")*/;
    private void updateDatabases(){
        final SequenceBlaster b=HotswapUtil.instance(comboClass,SequenceBlaster.class,true);
        if (b==null) return;

        for(int isAmino=2; --isAmino>=0;) {
            vDatabases[isAmino].clear();

            final String dd[]=b.getAvailableDatabases(isAmino!=0, true);
            if (sze(dd)>0) {
                adAll(dd,vDatabases[isAmino]);
                comboDB[1].s(dd[0]);
            }
        }
        final String s=comboNT.i()==0 ? " residue" : " nucleotide";
        labFirst.t("First "+s);
        labLast.t("Last "+s);
        proteinList.repaint();
        final Thread t2=threadM("fromRadioDatabase", this, NO_OBJECT);
        newThrd(thread_setAvailableDB(b,vDatabases,comboDB), t2).start();
    }
    private Runnable todo[];
    private void compute() {
        StrapProtein pp[]=proteinList.getSelectedProteins();
        if (pp.length==0 && align.getVisibleProteins().length<4) pp=align.getVisibleProteins();
        if (pp==null||pp.length==0) { StrapAlign.errorMsg("Select a protein!"); return;}
        int count=0;
        this.todo=new Runnable[pp.length];
        final int mode=comboNT.i();
        for(StrapProtein p : pp) {
            if (!proteinList.isEnabled(p)) continue;
            final SequenceBlaster blaster=lastRunning=HotswapUtil.instance(comboClass,SequenceBlaster.class,true);
            if (blaster==null)  { StrapAlign.errorMsg("Could not instantiate"+comboClass);return;}
            final String seq;
            if (mode==0)  seq=p.getResidueTypeAsStringUC();
            else {
                final byte nts[]=p.getNucleotidesCurrentStrand();
                if (nts==null  && p.containsOnlyACTGNX()) seq=p.getResidueTypeAsStringUC();
                else if(nts!=null) {
                    seq=toStrg(mode==1 ? nts:DNA_Util.onlyTranslated(nts,p.areNucleotidesCurrentStrandTranslated()));
                } else seq=null;
            }
            if (seq==null) return;
            final int from=mini(maxi(1,pp.length>1 ? 1 : atoi(tfFrom)),seq.length());
            final int to=mini(maxi(from,pp.length>1 ? MAX_INT : atoi(tfTo)),seq.length());
            final String query=seq.substring(from-1,to);
            if (mode==0 && !(p.countResidues()>50 && p.containsOnlyACTGNX())) blaster.setAAQuerySequence(query);
            else blaster.setNTQuerySequence(query);

            blaster.setDatabase(comboDB[togAminoDB.s()?1:0].toString());
            final Result r=new Result(blaster,p,mode==0,query);
            todo[count++]=ChDelay.newThread(r,"RUN",null);
            adTab(r.tabTitle(" ..."),r, (JTabbedPane)this);
        }
        ChDelay.newThread(this,"RUN",null).start();
    }
    public Object run(final String id,final  Object arg) {
        if (id=="RUN") {
            final Runnable rr[]=todo;
            if (rr!=null)
                for(Runnable r : rr)
                    if (r!=null && !isDisposed())
                        r.run();

        }
        return null;
    }
    public void actionPerformed(ActionEvent ev) {
        final Object q=ev.getSource();
        final String cmd=actionCommand(ev);
        final int nP=proteinList.getSelectedProteins().length;
        if (nP>0) StrapAlign.errorMsg("");
        if (q==togAminoDB) {
            if (!togAminoDB.s()) radioGrpSet(-1, radioDatabase);
        }
        if (q==comboNT || q==comboClass) {
            final int mode=comboNT.i();
            proteinList.setFilter(mode==1 ? ProteinList.NT_or_ACTGN : mode==2 ? ProteinList.NT : 0);
            proteinList.repaint();
            updateDatabases();
        }
        if (q==butGo) compute();
        if (q==proteinList) {
            tfFrom.setEnabled(nP==1);
            tfTo.setEnabled(nP==1);
        }
        if (0<=radioGrpIdx(q,radioDatabase)) fromRadioDatabase();
    }

    public void fromRadioDatabase() {
        final int iDB=radioGrpIdx(radioDatabase);
        int idx=-1;
        if (iDB>=0) {
            final java.util.List[] vv=vDatabases;
            for(int i=sze(vv[1]); --i>=0;) {
                String s=(String)vv[1].get(i);
                s=s.substring(s.lastIndexOf(' ')+1);
                if (
                    iDB==0 && "pdb".equalsIgnoreCase(s) ||
                    iDB==1 && ( "uniprotkb".equalsIgnoreCase(s) || "uniprot".equalsIgnoreCase(s)) ||
                    iDB==2 && "uniref90".equalsIgnoreCase(s)) {
                    idx=i; break;
                }
            }
            if (idx>=0) {
                comboDB[1].s(idx);
                togAminoDB.s(true);
                setFG(0, labStatus.t(""));
            } else setFG(0xFF0000, labStatus.t("Not available: "+getTxt(radioDatabase[iDB])));
        }
    }
    private class Result extends ChPanel implements Disposable,ChRunnable, ActionListener {
        private final SequenceBlaster blaster;
        private final AsciiView tpResult=new AsciiView((File)null);
        private final ChButton BUT_applyFormat=new ChButton("Apply",ChButton.DISABLED);
        private final StrapProtein protein;
        private final ChLabel labInfo=new ChLabel();
        private final String query;
        private String database;
        private boolean isComputed;

        public void dispose() {
            dispos(0L, blaster);
            dispos(0L, lastRunning);
        }

        Result(SequenceBlaster bl,StrapProtein p, boolean isAA,String query) {
            this.query=query;
            protein=p;
            blaster=bl;
            final String txt=
                "query="+query+
                "\nprogram="+blaster.getClass().getName()+
                "\ndatabase="+blaster.getDatabase()+
                "\nis calculating please wait ...\n";

            tpResult.setText(new ByteArray(txt));
            disposeOnCtrlW(0L, tpResult,Result.this);
            pcp(KOPT_CONTAINS_AA,"",tpResult);
            pcp(ACTION_DATABASE_CLICKED,"load", tpResult);
            pcp(KEY_TITLE, "Blast "+p, tpResult);
            final String msgFetch=
                "<h2>Downloading proteins from the Blast result</h2>"+
                "<i>PACKAGE:charite.christo.strap.</i>"+
                "Entries in the  blast-result start with a greater than sign "+
                "followed by a datatbase and a sequence identifier. <br>"+
                "Consider the the following example and observe the mouse cursor when the mouse is moved over the identifiers:"+
                "<pre>\n  71&gt; UNIPROTKB:P56817\n</pre>"+
                "Here \"UNIPROTKB\" acts as a hyperlink and opens the entry in the Web-browser. "+
                "Clicking  \"P56817\", loads the sequence into the alignment pane."+
                "<br><br>"+
                "<i>SEE_DIALOG:DialogFetchSRS</i>"+
                "<i>SEE_DIALOG:DialogFetchPdb</i>",
                infoCreateSel=
                "The selected proteins (I.e. the proteins marked in alignment row header) are analyzed.<br>"+
                "The sequences of the blast hits are considered (I. e. the lines starting with the letter \"H\").<br>"+
                "If a hit sequence is identical or part of a protein sequence,<br>"+
                "then a residue selection for this protein is created.";
            final Runnable run=ChDelay.newThread(this,"RUN",null);
            final JComponent changeFormat=vBhB(
                                               "The format of the BLAST output width can be changed here.",
                                               "The advantage of unfolded lines is that the output can be searched for sequence patterns (Ctrl F).",
                                               hB("Residues per text line ", tfOutputWidth),
                                               BUT_applyFormat.r(run)
                                               ),

                pTools=vBhB(
                            hBL(new ChButton("RES_SEL").t("Create residue selection of matched regions").li(this),smallHelpBut(infoCreateSel)),
                            hBL(ChButton.actionViewMessage("Change Output format",changeFormat)),
                            ChButton.actionCtrl(blaster),
                            " ",
                            hBL("Options for result-text ", new ScrollLabel(ChTextViewer.KEYBINDINGS))
                          ),
                pNorth=vB(
                          hBL(
                              toggl().actionCollapse(false,pTools),"Tools   ",
                              ChButton.actionViewMessage("Download proteins...",msgFetch).cp(ACTION_DATABASE_CLICKED,"load"),
                              ChButton.actionCtrl(blaster),
                              labInfo
                              ),
                          pTools
                          );
            cnsew(
                  this,
                  tpResult.tools().scrollPane(),
                  newCNSEW(pNorth,null,null,new ChButton().actionClose(this, DISPOSE_RM_CHILDS),null),
                  null,	null,null
                  );
        }
        public Object run(String id, Object arg) {
            if (id=="RUN") {
                this.database=delToLastChar(comboDB[togAminoDB.s()?1:0].toString(),' ');
                ByteArray xml=CacheResult.isEnabled() ? blaster.getResultXml() : null;
                if (xml==null) { blaster.compute();  this.isComputed=true; xml=blaster.getResultXml();}
                if (xml!=null) ChDelay.invokeAndWait(this,"RUN_EDT",xml);
            }
            if (id=="RUN_EDT") {
                ByteArray xml=(ByteArray)arg;
                final File fResult=createTempFile(".blast");
                final OutputStream fOut;
                try {  fOut=fileOutStream(fResult); } catch(IOException e){ putln(" caught in DialogBlast ",e); return null;}
                final BlastParser parser=new BlastParser();
                ByteArray when=null;
                String whenSection=null;
                final AbstractBlaster abstrB=deref(blaster, AbstractBlaster.class);
                if (abstrB!=null) when=CacheResult.getValue(abstrB.getClass(), whenSection="/day/"+abstrB.cacheSK('S'), abstrB.cacheSK('K'), null);
                if (when==null) when=new ByteArray(0);
                parser.setDay(when, this.isComputed);
                try {
                    fOut.write((fResult+"\n"+shrtClasNam(blaster)+" "+database+"\nQuery="+query+"\n").getBytes());
                } catch(IOException ex){}
                parser.parseXML(xml,maxi(10,atoi(tfOutputWidth)),fOut);
                closeStrm(fOut);
                final int count=parser.countHits(), countN=parser.countNovelHits();
                setTtleAt(this, tabTitle(" #"+count+(countN>0?"*":"")), DialogBlast.this);
                //if (isComputed) CacheResult.put(blaster,"/day/",when,true); !!!!!!!!!!!!!!!!!
                if (abstrB!=null) CacheResult.putValue(abstrB.getClass(), whenSection, abstrB.cacheSK('K'), when,true);

                xml=null;
                tpResult.tools().rmAllHighlights();
                tpResult.setFile(fResult);
                pcp(AsciiView.KOPT_SEARCH_SKIP_BLANKS,"",tpResult);

                final ChTextComponents tools=tpResult.tools();
                tools.underlineRefs(0);
                labInfo.t(countN>0 ? countN+" Novel Blast hits." : "#hits = "+count);
                if (count>countN && countN>0) tools.highlightOccurrence("(NEW)",null,null,HIGHLIGHT_NOT_IN_SCROLLBAR, C(0xffafaf));
                BUT_applyFormat.setEnabled(true);
                revalAndRepaintC(tpResult);

            }
            return null;
        }
        private String tabTitle(String suffix) {
            final String n=protein.getName();
            return n+ suffix;
        }

        private final List<ResidueAnnotation> vSelections=new Vector();
        public void actionPerformed(ActionEvent ev) {
            final String cmd=ev.getActionCommand();
            if (cmd=="RES_SEL") {
                StrapProtein pp[]=StrapAlign.getInstance().getSelectedProteins();
                if (sze(pp)==0) pp=StrapAlign.getInstance().getVisibleProteins();

                final ByteArray xml=blaster.getResultXml();
                if (sze(xml)==0) { error("Blast not finished yet"); }
                else {
                    BlastResult bResult=new BlastResult(xml);
                    final List<ResidueAnnotation> v=this.vSelections;
                    v.clear();
                    for(StrapProtein p : pp) {
                        final boolean[] selected=bResult.getMatchingAminoacids(p.getResidueTypeExactLength());
                        if (countTrue(selected)>0) {
                            ResidueAnnotation a=null;
                            for(ResidueAnnotation ra : p.getResidueAnnotations()) if (SELECTION_NAME.equals(ra.getName())) a=ra;
                            if (a==null) a=new ResidueAnnotation(p);
                            a.setValue(ResidueAnnotation.POS, boolToTxt(selected,1,",","-",(ByteArray)null));
                            a.setValue(ResidueAnnotation.NAME,SELECTION_NAME);
                            a.setValue(ResidueAnnotation.GROUP, "blast");
                            p.addResidueSelection(a);
                            v.add(a);
                        }
                    }
                    if (sze(v)>0)  {
                        final String T="Residue selections created by Blast";
                        final Object
                            pApply=vBpnl(
                                         "#TB Apply created residue selections",
                                         hBL("Compute multiple sequence alignment using only these matching parts ",new ChButton("REALIGN").t("Realign ...").li(this))
                                         ),
                            pNorth=vBpnl(T,hB(" ",pApply," "));
                        StrapAlign.showInJList(0L,null, v.toArray(), pNorth, null).showInFrame(ChFrame.AT_CLICK|ChFrame.SIZE_REGARD_RESOLUTION,T);
                        StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,9);
                    }
                    else ChMsg.msgDialog(0L, "I analyzed "+pp.length+" proteins but could not create residue selections from blast result");
                }
            }
            if (cmd=="REALIGN") ResSelUtils.openDialogRealign(toArry(vSelections,ResidueAnnotation.class));
        }
    }

    public static Thread thread_setAvailableDB(SequenceBlaster b, java.util.List v[], JComponent jc[]) {
        return threadM("setAvailableDB", DialogBlast.class, new Object[]{b,v,jc});
    }
    public static void setAvailableDB(SequenceBlaster b, java.util.List v[], JComponent jc[]) {
        for(int isAmino=2; --isAmino>=0;) {
            final String ss[]=b.getAvailableDatabases(isAmino>0, false);

            if (sze(ss)>0 && get(isAmino,v)!=null) {
                v[isAmino].clear();
                adAll(ss,v[isAmino]);
            }
            if (sze(jc)>isAmino) revalAndRepaintC(jc[isAmino]);
        }
    }
}
