package charite.christo.strap;
import charite.christo.interfaces.*;
import charite.christo.*;
import charite.christo.strap.interfaces.*;
import java.awt.*;

import javax.swing.*;
import java.io.File;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static charite.christo.strap.interfaces.AlignmentWriter.*;
/**HELP
The sequence alignment can be exported in various formats:
<ol>
<li>HTML</li>
<li>WIKI:Fasta_format</li>
<li>HSSP-format is not frequently used: http://www.predictprotein.org/Dexa/out_hssp.html</li>
<li>MSF-format:  http://bmerc-www.bu.edu/examples/output/seqlist.msf.html</li>
<li>ClustalW-format: http://www.molecularevolution.org/mbl/resources/fileformats/</li>
</ol>
<i>SEE_DIALOG:Texshade</i>
<i>SEE_DIALOG:DialogImportMFA</i>
<i>SEE_DIALOG:DialogExportProteins</i>

@author Christoph Gille
*/
public class DialogExportAlignment  extends AbstractDialogJTabbedPane implements Runnable {
    private final ProteinList proteinList=new ProteinList(align,0);
    private final ChCombo
        comboClass=StrapAlign.classChoice(AlignmentWriter.class),
        comboFormat=new ChCombo("HTML","ClustalW","MSF","Fasta"),
        comboDNA=new ChCombo("amino acid sequence","nucleotide sequence");
    private final ChTextField
        tfColumns=new ChTextField(" 50").cols(6,true,true),
        tfFrom=new ChTextField("1").cols(7,true,true),
        tfTo=new ChTextField("99999").cols(7,true,true);
    private final ChButton
        cbExtensions=toggl("Suppress file suffices such as  \".pdb\"").s(true),
        cbSpace=toggl("Extra space in MSF and ClustalW format");
    private int count;
    public void run() {
        final StrapProtein pp[]=proteinList.getSelectedOrAllProteins();
        if (pp.length==0) return;
        AlignmentWriter alWr=HotswapUtil.instance(comboClass,AlignmentWriter.class,true);
        if (alWr==null) return;
        Container c=export(pp,alWr);
        Class clas=HotswapUtil.getClass(comboClass.getSelectedItem());
        adTab(shrtClasNam(clas),c, (JTabbedPane)this);
    }

    public DialogExportAlignment() {
        JComponent
            pTop=StrapAlign.headLine(this),
            pLeft=vBhB(
                       KOPT_TRACKS_VIEWPORT_WIDTH,
                       comboClass.panel(),
                       hBL("Format ",comboFormat," ",comboDNA),
                       hBL("Residues per line ",tfColumns),
                       hBL("Alignment column  from ",tfFrom," to ",tfTo),
                       cbExtensions.cb(),
                       cbSpace.cb(),
                       pnl(new ChButton("GO").t(ChButton.GO).r(this)),
"##"
                       ),
            pCenter=pnl(new GridLayout(1,2),pLeft, proteinList.scrollPane()),
            pMain=newCNSEW(pCenter,pTop,null,null,null);
        adMainTab(pMain, this,null);
        pcp(KEY_CHOICE_CLASS,comboClass,this);
    }
    private Container export(StrapProtein pp[],AlignmentWriter alWr) {
        final String names[]=new String[pp.length];
        for(int iP=0; iP<pp.length; iP++) {
            final String s=pp[iP].getName();
            names[iP]=cbExtensions.s() ? withoutDotSuffix(s) : s;
        }
        alWr.setProteins(pp);
        alWr.setNames(names);
        alWr.setColumnRange(maxi(atoi(tfFrom)-1,0), maxi(atoi(tfTo)-1,0) );
        alWr.setResiduesPerLine(maxi(atoi(tfColumns),1));

        final int iF=comboFormat.i();
        final long options=
            (comboDNA.i()>0 ? NUCLEOTIDE_TRIPLET: 0) |
            (cbSpace.s() ? EXTRA_SPACE :0) |
            (iF==0?HTML:iF==1?CLUSTALW:iF==2?MSF:iF==3?FASTA:0);
        final StringBuffer sb=new StringBuffer(99999);
        alWr.getText(options,sb);
        final String ext="."+alWr.getFileExtension();
        final File f=file(STRAPOUT+"/alignments/"+ ++count+ext), f2=file(STRAPOUT+"/alignments/current"+ext);
        final String txt=sb.toString();
        wrte(f,txt);
        wrte(f2,txt);
        final ChTextComponents tools= ChTextComponents.tools( (options&HTML)!=0 ? new ChTextViewer(txt) : new AsciiView(f));
        final JComponent pnl= newCNSEW(tools.scrollPane(),null, pnl(f,f2),null,null);
        disposeOnCtrlW(0L, tools.jComponent(), pnl);
        return pnl;
    }
}
