package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import java.awt.*;
import javax.swing.*;
import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP

<pre class="data" style="float: right;">
   1    AAACATGGCG CTGGCTAGCG TGTTGCAGCG ...
   51   ACGGGTTTTT TGGGCTCGGA GGTGGTGCAG ...
   101  GGGAGTCCTG GTGATGGGCT GAGCCTAGCC ...
...
</pre>

The Genbank file format and Embl file format are widely used file
format for annotated nucleotide sequences (format specifications:
http://www.ncbi.nlm.nih.gov/Sitemap/samplerecord.html and
ftp://ftp.ncbi.nih.gov/genbank/gbrel.txt and
http://www.ebi.ac.uk/embl/Documentation/User_manual/usrman.html).
When a Genbank file is loaded into STRAP the entire nucleotide
sequence will be extracted and appear in the multiple sequence
alignment pane.

These nucleotides can be translated into
amino acids using the CDS-sequence-features within the protein file.
The Genbank file contains fields with the keyword "CDS" (coding sequence).
The text following "CDS" tells the translation direction and the translated nucleotides.
The sequence may be continuous or interrupted by introns.
The following example of a CDS field is from the file NCBI_NT:5757659.

<pre class="data" style="float: right;">
   CDS     5. .799
   /gene="Psmb5"
   /codon_start=1
   /product="proteasome subunit X"
   /protein_id="AAD50536.1"
   /db_xref="GI:5757659"
   /translation="MALA ...VSVP"
</pre>

If the Genbank file contains several genes the user can chose one of them.

Sometimes Genbank files already contain the translation in a field
<I>/translation.</I> The  sequence computed by STRAP should be identical to
this sequence.

<br>

When the project is saved to disk an additional file will be created with
the ending ".dna" coding the translation direction (forward/reverse
complement) and the translated and un-translated positions.
<i>SEE_DIALOG:DialogFetchSRS</i>
<i>SEE_DIALOG:EditDnaParentPane</i>
<span class="clearall"></span>
@author Christoph Gille
*/
public class DialogGenbank extends AbstractDialogJTabbedPane implements java.awt.event.ActionListener,NeedsProteins,Runnable {
    private int count;
    public DialogGenbank() {
        final Container
            pSouth=pnl(new ChButton("GO").t(ChButton.GO).li(this)),
            box=newCNSEW("Select proteins in Genbank or EMBL format", StrapAlign.headLine(this),pSouth,null,null);
        adMainTab(box, this,null);
    }
    private Object[] vProteins;
    public void setProteins(Protein... pp) {
        final int n=sze(pp);
        vProteins=new Object[n];
        for(int i=n; --i>=0;) vProteins[i]=weakRef(pp[i]);
    }
    public void actionPerformed(java.awt.event.ActionEvent ev) {
        final String cmd=ev.getActionCommand();
        final Object q=ev.getSource();
        final StrapProtein p= gcp("P",q,StrapProtein.class);
        final ChJTable table= gcp("T",q,ChJTable.class);
        final Object[][] data=gcp("D",q,Object[][].class);
        final Object[] row=gcp("R",q,Object[].class);
        if (q instanceof ChCombo && row!=null) {
            final int idx=((ChCombo)q).i();
            for(Object o : row) {
                if (o instanceof ChCombo)  ((ChCombo)o).s(idx);
            }
            repaint();
        }
        if (cmd=="GO") {
            StrapAlign.errorMsg("");
            setProteins(align.getSelectedProteins());
            run();
        }
        if (cmd=="K" && p!=null) { p.setTranslatedStrand(null); new StrapEvent(this,StrapEvent.NUCL_TRANSLATION_CHANGED).setParameter(p).run() ; }
        if (cmd=="A" && table!=null && data!=null) {
            final int ii[]=ChJTable.selectedRowsConverted(table);
            StrapProtein pp[]=null;
            for(int iRow=data.length; --iRow>=0;) {
                if (ii.length>0 && idxOf(iRow,ii,0,ii.length)<0) continue;
                final StrapProtein prot=(StrapProtein)deref(data[iRow][1]);
                final Object cds=data[iRow][2];
                if (prot!=null) {
                    prot.applyCDS_expression(toStrg(cds));
                    if (pp==null) pp=new StrapProtein[data.length];
                    pp[iRow]=prot;
                }
            }
            if (pp!=null) new StrapEvent(this,StrapEvent.NUCL_TRANSLATION_CHANGED).setParameters(pp).run();
        }
    }

    public void run() {
        final StrapProtein pp[]=derefArray(vProteins,StrapProtein.class);
        final Object[][] data=new Object[pp.length][];
        ChButton butReset=null;
        for(int iP=0; iP<pp.length; iP++) {
            final StrapProtein  p=pp[iP];
            final String CDS[]=p.getCDS();
            if (sze(CDS)==0) {
                data[iP]=new Object[]{String.valueOf(iP+1), weakRef(p), "Error: No CDS", "", "","", "",""};
                continue;
            }
            butReset=new ChButton("K",ChButton.ICON_SIZE).li(this).t(null).i(IC_KILL).tt("Discard all reading frame and intron/exon information")
                .cp("P",weakRef(p)).cp(ChRenderer.KEY_EDITOR_COMPONENT,"");

            final Object cc[]=new Object[5];
            for(int col=5; --col>=0;) {
                String ss[]=null, oneVal="";
                for(int i=CDS.length; --i>=0;) {
                    final String cds=CDS[i], s0;
                    if (col==0) s0=cds;
                    else {
                        final String gpps[]=p.geneProductProteinNoteForCDS(cds);
                        s0=gpps!=null && gpps.length>col-1?gpps[col-1]:"";
                    }
                    if (sze(s0)>0) {
                        if (ss==null) ss=new String[CDS.length];
                        ss[i]=s0;
                        if (oneVal=="") oneVal=s0;
                        else oneVal=null;
                    }
                }
                if (sze(oneVal)>0 || ss==null) cc[col]=oneVal;
                else {
                    final ChCombo c=new ChCombo(ss).li(this);
                    cc[col]=c;
                    pcp("R",cc,c);
                    pcp(ChRenderer.KEY_EDITOR_COMPONENT,c,c);
                    pcp(ChRenderer.KEY_RENDERER_COMPONENT,c,c);
                }
            }
            pcp(ChRenderer.KEY_RENDERER_COMPONENT,butReset,iicon(IC_KILL));
            data[iP]=new Object[]{String.valueOf(iP+1), weakRef(p), cc[0], cc[1], cc[2], cc[3], cc[4], butReset};
        }
        final ChJTable jTable=new ChJTable(new ChTableModel("#", "protein", "CDS", "gene", "product", "protein", "note",   "reset")
                                           .setData(data),ChJTable.DEFAULT_RENDERER|ChJTable.ICON_ROW_HEIGHT);
        jTable.setColumnWidth(false, 0,ChJTable.COLUMN_WIDTH_4_DIGITS);
        jTable.setColumnWidthC(false,7, butReset);
        final JComponent
            butApply=new ChButton("A").t("Apply all or selected rows").li(this).cp("D",data).cp("T",jTable),
            pSouth=pnl(butApply),
            pNorth=newCNSEW(smallHelpBut(this),null,null,null,null),
            pnl=newCNSEW(scrllpn(0L,jTable),pNorth,pSouth,null,null);
        adTabTITI(pnl,String.valueOf(++count), this);
        pNorth.add(new ChButton().actionClose(pnl),BorderLayout.EAST);
    }

}
/*
Problem:
AL935121
FT   CDS             complement(join(104984..105040,105385..105531,
FT                   AL772299.10:2661..2732,AL772299.10:3575..3646,
FT                   AL772299.10:4101..4220,AL772299.10:5400..5462,
FT                   AL772299.10:5872..6063,AL772299.10:9023..9139,
FT                   AL772299.10:10369..10503,AL772299.10:12643..12788,

*/