package charite.christo.strap;
import charite.christo.interfaces.*;
import charite.christo.*;
import charite.christo.protein.MultipleSequenceParser;
import java.awt.*;
import javax.swing.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;

/**HELP
Multiple sequence files (<i>WIKI:FASTA_format</i>, MSF, BIOWIKI:NexusFormat, WIKI:Stockholm_format ) can be directly dragged into the
alignment panel.

But the import dialog provides more control for the import of these files.

The alignment text is pasted into the text-area
and <i>LABEL:ChButton#BUTTON_GO</i> is pressed.

The result is a table of proteins.

The  column <i>LABEL:DialogImportMFA#HEADER_PROTEIN_NAME</i> must be
filled-in. this column containse the names for the new
protein files.

<br><br><b>Setting the protein names</b>:

Often the header consists of several fields separated by a vertical bar and the first or
secondary field can be used as protein name.
More general but also more complicated are regular expressions to extract parts of the header text (WIKI:Regular_expression).

@author Christoph Gille
*/
public class DialogImportMFA  extends AbstractDialogJTabbedPane implements java.awt.event.ActionListener {
    public void setText(CharSequence s) { ta.setText(toStrg(s));}
    public final static String
        HEADER_PROTEIN_NAME="Protein name",
        BUT_LAB_clearNames="Clear column \""+HEADER_PROTEIN_NAME+"\"",
        BUT_LAB_FillNamesRE="Fill the column \""+HEADER_PROTEIN_NAME+"\" using regular expressions";
    private final ChTextArea ta=new ChTextArea("");
    private Object[] togOverwrite=radioGrp(new String[]{"Skip","Overwrite protein"},0,null);

    private ChFileChooser fs;
    private ChFrame frame;
    private final ChTextField tfSeparator=new ChTextField("|").cols(3,true,true).li(this);
    public void actionPerformed(java.awt.event.ActionEvent ev) {
        StrapAlign.errorMsg("");
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        if (cmd=="EX") ta.t("\n>homo_sapiens \nMY-MISS--ISAH-IPPIE\n>mus_musculus \nI--MISS-MISSISSIPPI\n");
        if (cmd=="FF")  {
            if (fs==null) {
                fs=new ChFileChooser(DialogImportMFA.class).setVisibleFalseParentFrame();
                addActLi(this,fs);
                fs.filter().acceptSuffix(true,".fasta",".fa",".mfa");
                fs.addButton(fs.butProject());
                frame=new ChFrame("DialogImportMFA").size(400,400).ad(fs.getPanel());
            }
            frame.show();
        }
        if (cmd==JFileChooser.APPROVE_SELECTION) ta.t(readStrg(fs.getSelectedFile()));
        if (cmd=="GO") {
            Parse c=new Parse(new MultipleSequenceParser());
            if (c.isSuccess()) adTab("Parsed",c, (JTabbedPane)this);
        }
        if (q==tfSeparator && (cmd==ACTION_ENTER || cmd==ACTION_FOCUS_LOST)) {
            String s=tfSeparator.toString();
            setText(s.length()>0 ? s.substring(0,1) : "|");
        }
    }
    public DialogImportMFA() {
        JComponent
            pButtons=pnl(
                         new ChButton("FF").t("Load from file").li(this),
                         new ChButton("EX").t("Show example").li(this), "  ",
                         new ChButton("GO").t(ChButton.GO).li(this)
                         ),
            pNote=pnl("If the multiple sequence  file contains protein<br>IDs you should also consider the Dialog",align.b(DialogFetchSRS.class)),
            pSouth=vB(pButtons,pNote);
        pcp(KEY_IF_EMPTY,"Enter an alignment in multiple fasta format",ta);
        adMainTab(newCNSEW(ta.tools().enableUndo(true).enableWordCompletion(dirWorking()).scrollPane(),StrapAlign.headLine(this),pSouth,null,null), this,null);

    }
    //========================================
    private final static int HEADER=0,NAME=1,SEQUENCE=2,OVERWRITE=3;
    private String[] header;
    private final class Parse extends JPanel implements java.awt.event.ActionListener,javax.swing.table.TableModel,javax.swing.table.TableCellRenderer {
        /* --- TableModel --- */
        public String getColumnName(int c) {return header[c];}
        public int getRowCount() {return data.length;}
        public int getColumnCount() {return header.length;}
        public Object getValueAt(int row,int c) {return data[row][c];}
        public boolean isCellEditable(int row,int column) { return column==NAME;}
        public void setValueAt(Object o,int row,int col) {
            data[row][col]=o;
            if (col==NAME) { table.revalidate(); table.repaint(); }
        }
        public Class getColumnClass(int col) {return classes[col];}
        public void  addTableModelListener(javax.swing.event.TableModelListener l) {}
        public void removeTableModelListener(javax.swing.event.TableModelListener l) {}
        /* --- EndTableModel --- */
        /* --- Table Renderer --- */
        private final ChLabel LABEL=new ChLabel();
        { monospc(LABEL);}
        public Component getTableCellRendererComponent(JTable t,Object v,boolean isSel,boolean hasFocus,int row,int col) {
            if (v==null) v="Error";
            LABEL.cp(KOPT_IS_SELECTED,isSel?"":null).t(v.toString()).setEnabled(true);
            switch(col) {
            case NAME:
                if (v!=null && v.equals(ChRegex.NO_MATCH)) {
                    LABEL.setEnabled(false);
                    return LABEL.t(v.toString());
                }
                if (v==null || v.toString().trim().equals("")) {
                    LABEL.setEnabled(false);
                    return LABEL.t("Enter a name for the protein");
                }
                break;
            case OVERWRITE:
                final String name;
                {
                    Object o=data!=null && data.length>row ? data[row][NAME] : null;
                    String n=o!=null ? o.toString() : null;
                    name=n!=null ? n : null;
                }
                final boolean exists=name!=null && !"".equals(name) && file(name).exists();
                final String tt= exists ? "a protein with this name <br>does already exist<br>in the project directory<br>check if you want to overwrite"
                    : "a protein with this name <br>does not yet exist";
                return LABEL.tt(tt).t(exists ? "exists" : "").fg(0xFF0000);

            }
            return LABEL;
        }
        /* --- EndTable Renderer --- */
        public void actionPerformed(java.awt.event.ActionEvent ev) {
            StrapAlign.errorMsg("");
            final Object q=ev.getSource();
            final String cmd=ev.getActionCommand();
            if (cmd==BUT_LAB_FillNamesRE) {
                final String replacement=tfReplacement.toString();
                final java.util.regex.Pattern pattern=ChRegex.pattern(tfExpression.toString());
                for(Object[] d:data) {
                    final String header=d[HEADER].toString().trim();
                    d[NAME]=ChRegex.replace(pattern,replacement,header).replaceAll("\\W","_");
                }
                table.repaint();
            }
            if (cmd=="FIELD") {
                String sep=tfSeparator.toString();
                sep=sep.length()>0 ? sep.substring(0,1) : "|";
                final int col=atoi(((ChButton)q).getText());
                for(Object dd[]:data) {
                    final java.util.StringTokenizer st=new java.util.StringTokenizer(dd[HEADER].toString().trim(),sep);
                    if (st.countTokens()<col) return;
                    for(int iT=0;iT<col;iT++)
                        dd[NAME]=st.nextToken();
                }
                table.repaint();
            }
            if (cmd==BUT_LAB_clearNames) {
                for(Object[] dd:data) dd[NAME]="";
                table.repaint();
            }
            if (cmd=="GO") {
                final int nR=table.getRowCount();
                vProteins=new UniqueList(StrapProtein.class);
                vAlreadyExists=new UniqueList(StrapProtein.class);
                if (nR==0) StrapAlign.errorMsg("Cannot create protein objects: no table rows");
                veryLongName=false; countNoName=countCreated=0;
                final int selected[]=ChJTable.selectedRowsConverted(table);
                if (selected.length==0) {
                    if (ChMsg.yesNo("No table rows selected! <br>Do you want to make protein objects from all rows?"))
                        for(int i=0;i<nR;i++) processRow(i);
                    else return;
                } else {
                    for(int s:selected) processRow(s);
                }
                String msgLong="At least one of the protein names is unusal long.\nDo you want to review the protein names before creating the protein objects ?";
                if (veryLongName && ChMsg.yesNo(msgLong)) return;
                final StringBuffer sb=new StringBuffer();
                if (vProteins.size()>0) {
                    sb.append("The following proteins were created:\n");
                    for(Object o: vProteins.asArray()) sb.append(o).append("\n");
                }
                if (countNoName>0) sb.append("\n\n"+countNoName+" table rows were skipped because the field in the column 'Protein name' was empty");
                if (vAlreadyExists.size()>0) {
                    sb.append("\n\nThe following proteins were not created because proteins with the given name  already exist: \n");
                    for(Object o: vAlreadyExists.asArray()) sb.append(o).append("\n");
                }
                if (sb.length()>0) new ChTextViewer(sb).title("Created").shw(ChFrame.AT_CLICK);
                align.addProteins(-1,vProteins.asArray());
                table.repaint();
            }
        }
        boolean veryLongName;
        private int countCreated,countNoName;
        private UniqueList<StrapProtein> vProteins,vAlreadyExists;
        private void processRow(int i) {
            String name=getNameAt(i);
            //putln("name=",name);
            if (name==null || name.equals("")) { countNoName++; return;}
            String seq=table.getValueAt(i,SEQUENCE).toString().trim();
            String header=table.getValueAt(i,HEADER).toString().trim();
            // already known ?
            StrapProtein p=StrapProtein.getProteinWithName(name,null);
            if (p!=null) {
                if (radioGrpIdx(togOverwrite)==1) align.removeProteins(false,p);
                else {
                    vAlreadyExists.add(p);
                    return;
                }
            }
            p=new StrapProtein(align);
            p.setName(name);
            p.setProteinParserClass(MultipleSequenceParser.class);
            if (name.length()>30) veryLongName=true;
            final byte bb[]=seq.getBytes();
            p.setResidueType(onlyLetters(bb,0,MAX_INT));
            p.inferGapsFromGappedSequence(bb);
            vProteins.add(p);
            countCreated++;
            final java.io.File f=file(name);
            p.setFile(f);
            wrte(f,">"+header+"\n"+seq);
        }
        private ChJTable table;
        private ChTextField
            tfExpression=new ChTextField("([^|]*)\\|([^|]*)\\|([^|]*)\\|([^|]*).*").saveInFile("DialogImportMFA_expression"),
            tfReplacement=new ChTextField("HELLO $1 WORLD $2 ").tt("the replacement expression").saveInFile("DialogImportMFA tfReplacement");
        private Class classes[]={String.class,String.class,String.class,String.class};
        private Object[][] data;
        private boolean success=true;
        boolean isSuccess() { return success;}
        private String getNameAt(int i) {
            String name=table.getValueAt(i,NAME).toString().trim();
            name=rplcToStrg(" ","_",name);
            return name;
        }

        public Parse(MultipleSequenceParser parser) {
            if (nxt(NO_SPC, ta.tools().byteArray())<0) {StrapAlign.errorMsg("Enter multiple sequence text"); success=false;return;}
            final ByteArray ba=ta.tools().byteArray();
            ba.delBlanksL();
            parser.setText(ba);
            final byte seq[][]=parser.getGappedSequences();
            String headers[]=toStringArray(parser.getFastaHeader());
            if (headers==null) headers=parser.getSequenceNames();
            if (headers==null || headers.length==0) {StrapAlign.errorMsg("Error parsing the text: No protein names found"); success=false;return;}
            data=new Object[headers.length][];
            for(int i=0;i<headers.length;i++) {
                final String header=headers[i];
                final String name=nxt(NO_FILENM,header)<0 ? header : "";
                final String s=seq[i]!=null ? new String(seq[i]) : "";
                data[i]=new Object[] {header,name,s,""};
            }
            header= new String[]{"Header",HEADER_PROTEIN_NAME,"Sequence",""};
            table=new ChJTable(/*TableModel*/this,ChJTable.NO_REORDER);
            table.setDefaultRenderer(String.class,this);
            table.setDefaultEditor(String.class,new ChRenderer());
            table.setColumnWidthC(false, OVERWRITE,LABEL.t("exists"));
            final Container pFill=pnl("Take column in protein header ");
            for(int i=7;--i>=0;) {
                final ChButton b=new ChButton("FIELD").t(String.valueOf(7-i)).li(this);
                pFill.add(b);
                if (i==1) pFill.add(new JLabel(" or "));
            }
            pFill.add(new JLabel("  Seperator:"));
            pFill.add(tfSeparator);
            final Object explanation=
                "Short explanation: \n"+
                "The headers in fasta files often consist of fields that are separated by vertical bar.\n"+
                "Assume that you want only the 4rd field to be the protein name.\n"+
                "Press the button \"column 4\"\n";                 ;
            pFill.add(smallHelpBut(explanation));
            Object explanationRG=
                 "Short explanation: \n"+
                 "The headers in fasta files often consist of fields that are separated by vertical bar.\n"+
                 "Assume that you want only the 3rd field to be the protein name.\n"+
                 "[^|] means any character but vertical bar.\n"+
                 "[^|]* means a sequence of characters without vertical bar. Can have length zero.\n"+
                 "([^|]*)\\| means a sequence of characters without vertical bar followed by a vertical bar.\n"+
                 "([^|]*)\\|([^|]*)\\|([^|]*)\\| Is the same 3 times.\n"+
                 "Each parenthetical group can be referenced using the dollar sign followed by a number.\n"+
                 "$3 means the third field in the fasta header.\n"+
                "For more information consult the literature on regular expressions.\n",
                panFill=pnl(
                            new GridLayout(3,1),
                            newCNSEW(tfExpression,null,null,null,"Expression for<br> fasta header:"),
                            newCNSEW(tfReplacement,null,null,null,"Replacement for<br> protein name:"),
                            pnl(
                                new ChButton(BUT_LAB_FillNamesRE).li(this).tt("Find protein names  applying the regular <br>expressions on fasta header"),
                                smallHelpBut(explanationRG)
                                )
                            ),
                togPanFill=UserProfile.set('e',ChButton.actionViewMessage("Advanced users may process protein names with regular expressions",panFill)),
                pSouth=vBpnl(
                             pnl(new ChButton("GO").li(this).t(ChButton.GO),"If a protein  with given name already exists:",pnl(togOverwrite)),
                             vBpnl("#TBem Protein header infers  protein name",pFill,new ChButton(BUT_LAB_clearNames ).li(this)),
                             togPanFill);
            cnsew(this,new JScrollPane(table),hB("Preview: A protein file could be created for each selected table row  ","#",
                                                 new ChButton().actionClose(this, DISPOSE_RM_CHILDS)),pSouth,null,null);
        }
    }
}

