package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import charite.christo.strap.interfaces.*;
import java.awt.*;
import javax.swing.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
import static charite.christo.protein.StrapEvent.*;
import static charite.christo.strap.StrapUtils.*;
/**HELP

With this dialog numeric data is plotted along the sequence/alignment.
<ol>
<li>Plotting values assigned to amino acids such as sequence hydrophobicity profile</li>
<li>Values belonging to columns of the multiple sequence alignment such as sequence variable</li>
</ol>

Values are assigned to each column of the sequence alignment and a
diagram is plotted in the alignment pane and the horizontal scrollbar.

<i>PACKAGE:charite.christo.strap.extensions.</i>

The values are calculated by a class which implements
<I>ValueOfAlignPosition
</I> or <I>ValueOfResidue</I>.

The user chooses a Java class <i>COMBO:CountResidues</i> and some
proteins and selects the toggle
<i>ICON:IC_SHOW</i>.

<i>SEE_DIALOG:DialogPlot</i>
<i>SEE_DIALOG:DialogBarChart</i>

@author Christoph Gille
*/
public class DialogPlot extends AbstractDialogJTabbedPane implements ValueOfResidue, ValueOfAlignPosition, ResidueSelection, java.awt.event.ActionListener,IsEnabled, VisibleIn123, StrapListener, HasModiCount  {
    /* --- Fields for Plot ---  */
    private final Object hasValues;
    private final ValueOfResidue vor;
    private final ValueOfAlignPosition vap;
    private ProteinList listProteins;

    /* --- Fields for ResidueSelection ---  */
    private ChButton CB[];
    private ChCombo comboStyle;
    private boolean[] selectedAA;

    private ChSlider sliderL, sliderU;
    /* --- Fields for both, ResidueSelection and Plot ---  */
    private final JComponent panel;
    private final ChButton cbShow=toggl("show",ChButton.SHOW_HIDE|ChButton.HEIGHT_I).li(this);
    private final ProteinCombo comboProt;
  private StrapProtein p() { ProteinCombo c=this.comboProt; return c==null?null:c.getProtein();}
    private Color color=Color.GREEN;
    private int modiCount=-1;
    private JTabbedPane tabPane;
    private JComponent panSelection;

    /* --- Fields  and Methods for parent ---  */
    private final ChCombo comboClassVoR,comboClassVoAP;

    public void setVoR(Object o) { comboClassVoR.s(o);}
    public void setVoAP(Object o) { comboClassVoAP.s(o);}
    public DialogPlot() {
        comboClassVoR=StrapAlign.classChoice(ValueOfResidue.class).classRenderer(ValueOfResidue.class);
        comboClassVoAP=StrapAlign.classChoice(ValueOfAlignPosition.class).classRenderer(ValueOfAlignPosition.class);
        final JComponent
            butVoR=new ChButton("VOR").t("plot...").li(this),
            pVoR=hB("#TBem Values can be assigned to each amino acid."," ",enlargeFont(butVoR,2f),comboClassVoR.panel()),
            butVoAP=new ChButton("VOAP").t("plot...").li(this),
            pVoAP=vB("#TBem Values can be assigned to each alignment position", hB(enlargeFont(butVoAP,2f)," ",comboClassVoAP.panel())),
            pCenter=vB(a().headLine(this),pVoR,pVoAP,KOPT_TRACKS_VIEWPORT_WIDTH);
        adMainTab(scrllpn(0L,pCenter), this,null);
        hasValues=null;
        vap=null;
        vor=null;
        comboProt=null;
        panel=null;
    }
    /* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
    private DialogPlot(final Object v, Container sharedCtrlPanel, JTabbedPane tabbed) {
        this.hasValues=v;
        this.tabPane=tabbed;
        TabItemTipIcon.set(TabItemTipIcon.tab(v),TabItemTipIcon.item(v),TabItemTipIcon.tip(v),TabItemTipIcon.icn(v),this);
        if (v instanceof ValueOfAlignPosition) {vor=null; vap=(ValueOfAlignPosition)v;} else { vor=(ValueOfResidue)v; vap=null; }
        final boolean isVor=vor!=null;
        if (v instanceof StrapListener) a().addListener((StrapListener)v);
        comboProt=new ProteinCombo(align,0);
        listProteins=!isVor ? new ProteinList(align,0) : null;

        if (comboProt!=null) pcp(KEY_ENABLED, weakRef(this),comboProt.li(this));
        if (listProteins!=null) pcp(KEY_ENABLED, weakRef(this),listProteins.li(this));

        cbShow.s(isVor);
        final JComponent pNorth=newCNSEW(a().headLine(v),null,null, new ChButton().actionClose(this),null);
        //pNorth.add(new ChButton().actionClose(this),BorderLayout.EAST);
        //pcp(KOPT_NOT_INDICATE_TOO_SMALL,"",pNorth);
        panSelection=vB("#TB Select Residues ");
        final Container
            ctrlPanel=getCtrlPanel(v),
            pCenter=listProteins!=null ? newCNSEW(listProteins.scrollPane(),sharedCtrlPanel,ctrlPanel,null,null) : newCNSEW(sharedCtrlPanel,comboProt,ctrlPanel,null,null),
            pSouth=pnl(cbShow,
                       new ChButton("SEL").t("Select residues").li(this),
                       new ChButton("TABLE").t("As table").li(this),
                       new ButColor(0L,Color.GREEN,this).options(ChButton.NO_FILL|ChButton.NO_BORDER|ChButton.ICON_SIZE),
                       isVor ? StrapView.button(StrapView.TOG_PLOT_ENTIRE_PANEL).cln() : null
                       );
        this.panel=newCNSEW(pCenter,pNorth,pSouth,panSelection,null);
        StrapView.getPlottersV().add(this);
        a().addListener(this);

        comboClassVoR=comboClassVoAP=null;
    }

    public Color getColor() { return color; }
    public void setColor(Color c) {
        final int ev=CB!=null ? RESIDUE_SELECTION_CHANGED_COLOR : vor!=null ? VALUE_OF_RESIDUE_CHANGED_COLOR : VALUE_OF_ALIGN_POSITION_CHANGED_COLOR;
        if (c!=color) StrapAlign.dispatchLater(ev, 111);
        color=c;
        cbShow.s(true);
    }
    public void actionPerformed(java.awt.event.ActionEvent ev) {
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        final Protein p=getProtein();
        StrapProtein pp[]=getProteins();
        final boolean isVor=cmd=="VOR";
        if (isVor || cmd=="VOAP"){
            final ChCombo choice=isVor ? comboClassVoR : comboClassVoAP;
            final Object o=HotswapUtil.instance(choice, isVor ? ValueOfResidue.class : ValueOfAlignPosition.class,true);
            adTabTITI(new DialogPlot(o, getSharedCtrlPan(choice,true), this).panel,o,this);
        }
        if (cmd=="SEL") {
            rmFromParent(q);
            addActLi(this,sliderL=new ChSlider("sliderLower",0,100,0).endLabels("Values that are higher than ...",null));
            addActLi(this,sliderU=new ChSlider("sliderUpper",0,100,100).endLabels("Values that are lower than ...",null));
            comboStyle=AbstractVisibleIn123.newComboStyle().li(this);
            CB=new ChButton[]{
                toggl("1D",ChButton.SHOW_HIDE).li(this).tt("Show in alignment pane").s(true),
                toggl("SB",ChButton.SHOW_HIDE).li(this).tt("Show in scroll-bar").s(true),
                toggl("3D",ChButton.SHOW_HIDE).li(this).tt("Show in 3D-backbones")
            };
            panSelection.add(sliderL);
            panSelection.add(sliderU);
            panSelection.add(pnl(CB[0],CB[2],CB[1]));
            panSelection.add(comboStyle);
            panSelection.add(pnl(dim(100,9999)));
            if (vap!=null) panSelection.add(comboProt);
            updateSelection();
            panSelection.revalidate();
        }
        if (cmd=="TABLE") {
            final StringBuffer sb=new StringBuffer(99);
            JTable table=null;
            final ChRenderer rend=new ChRenderer();
            monospc(rend.LABEL);
            final double dd[]=vap!=null ? vap.getValues() : vor!=null ? vor.getValues() : null;
            if (p!=null && vor!=null && vap==null && dd!=null) {
                final Object data[][]=new Object[mini(dd.length,p.countResidues())][2];
                for(int i=0; i<data.length; i++) {
                    sb.setLength(0);
                    final int resNum[]=p.getResidueNumber();
                    if (resNum!=null) {
                      sb.append(get(i,resNum));
                        final int ins=p.getResidueInsertionCode(i);
                        if (ins>0) sb.append((char)ins);
                        final byte chain=p.getResidueChain(i);
                        if (chain>0) sb.append(':').append((char)chain);
                    }
                    data[i]=new Object[]{intObjct(i+1), String.valueOf((char)p.getResidueType(i)) ,sb.toString() ,new Float(dd[i])};
                }
                table=new JTable(data,new Object[]{"Idx","Amino acid","residue Number","value"});
            }
            if (vap!=null && dd!=null && sze(pp)>0) {
                {
                    final int maxCol=StrapProtein.getMaxColumn(pp);
                    final Object data[][]=new Object[mini(maxCol,dd.length)][2];
                    for(int col=0; col< data.length; col++) {
                        sb.setLength(0);
                        for(StrapProtein prot:pp){
                            final int idx=prot.column2index(col);
                            sb.append( idx<0 ? '.' : (char)prot.getResidueType(idx));
                        }
                        data[col]=new Object[]{intObjct(col+1),sb.toString(),new Float(dd[col])};
                    }
                    table=new JTable(data,new Object[]{"alignment column","residues","value"});
                }
                final StrapProtein pCurs=(StrapProtein)a().getProteinAtCursor();
                if (pCurs!=null) {
                    final Object data[][]=new Object[pCurs.countResidues()][3];
                    final Object FLOAT_NaN=new Float(Float.NaN);
                    for(int i=0; i<data.length; i++) {
                        final int col=pCurs.getResidueColumn(i);
                        final Object val= (col>=0 && col<dd.length) ?  new Float(dd[col]) : FLOAT_NaN;
                        sb.setLength(0);
                        final int resNum[]=pCurs.getResidueNumber();
                        if (resNum!=null) {
                          sb.append(get(i,resNum));
                            final int ins=p.getResidueInsertionCode(i);
                            if (ins>0) sb.append((char)ins);
                            final byte chain=pCurs.getResidueChain(i);
                            if (chain>0) sb.append(':').append((char)chain);
                        }
                        data[i]=new Object[]{intObjct(i+1),String.valueOf(((char)pCurs.getResidueType(i))),sb.toString(), intObjct(col+1),val};
                    }
                    final JTable tab=new JTable(data,new Object[]{"idx","aa","pdb-no","alignment column","value"});
                    tab.setDefaultRenderer(Object.class,rend);
                    final Container pMain=newCNSEW(scrllpn(0L,tab),pCurs+"   (cursor)","use copy and paste CTRL-A  CTRL-C CTRL-V",null,null);
                    new ChFrame("values from alignment positions mapped to protein at cursor").ad(pMain).shw(ChFrame.STAGGER);
                }
            }
            if (table!=null) {
                table.setDefaultRenderer(Object.class,rend);
                new ChFrame(hasValues.getClass().getName())
                    .ad(newCNSEW(scrllpn(0L,table),null,"use copy and paste CTRL-A  CTRL-C CTRL-V",null,null))
                    .shw(ChFrame.AT_CLICK);
            }
        }
        if (q instanceof ChSlider) {
            if (sliderU.getValue()<sliderL.getValue())
            if (q==sliderU) sliderL.setValue(sliderU.getValue()); else sliderU.setValue(sliderL.getValue());
            CB[0].s(true);
            final double vv[]=getValues();
            if (vv!=null)  {
                double maxVal=Double.MIN_VALUE, minVal=Double.MAX_VALUE;
                for(double val:vv) { if (maxVal<val) maxVal=val; if (minVal>val) minVal=val; }
                final double
                    f=(maxVal-minVal)/100,
                    lower= minVal+sliderL.getValue()*f,
                    upper= minVal+sliderU.getValue()*f;
                if (q==sliderU) sliderU.endLabels("Values that are >"+(float)upper,null);
                else sliderL.endLabels("Values that are <"+(float)lower,null);
            }
        }
        if (CB!=null) updateSelection();
        for(int isSelect=CB!=null ? 2 : 1; --isSelect>=0;) {
            int evType=0;
            if (q==comboStyle) evType=RESIDUE_SELECTION_CHANGED_COLOR;
            if (q==listProteins || q==comboProt || q instanceof ChButton || q instanceof ChSlider) evType=isSelect>0 ? RESIDUE_SELECTION_CHANGED : VALUE_OF_RESIDUE_CHANGED;
            if (evType!=0)  StrapAlign.dispatchLater(evType, 111);
        }
    }
    public final boolean isEnabled(Object o) {
        final Object v=this.hasValues;
        return o==null ? false : v instanceof IsEnabled ? ((IsEnabled)v).isEnabled(o) : true;
    }
    public Protein getProtein() { return  vor!=null ? sp(vor) : sp(comboProt);}
    public StrapProtein[] getProteins() { return vap!=null ? vap.getProteins() : null; }
    public void setProteins(StrapProtein... proteins) {}
    public boolean setProtein(Protein p) { if (comboProt!=null) comboProt.s(p); return true; }
    @Override public String toString() { return hasValues!=null ? hasValues.toString() : super.toString();}

    public double[] getValues() {
        if (CB==null && !cbShow.s()) return null;
        if (vor!=null) {
            if (comboProt!=null) {

              final Protein pCombo=p();
              if (sp(vor)!=pCombo) vor.setProtein(pCombo);
            }
            return vor.getValues();
        } else if (vap!=null) {
            StrapProtein pp[]=listProteins.getSelectedProteins();
            if (pp.length==0) pp=a().getVisibleProteins();
            final int modi=modiCount(this.listProteins);
            if ( (this.modiCount!=modi) || sze(vap.getProteins())!=pp.length) { this.modiCount=modi; vap.setProteins(pp); }
            return vap.getValues();
        } else return null;
    }
    @Override public void dispose() {
        for(Protein p: a().getProteins()) p.removeResidueSelection(this);
        a().removeListener(hasValues);
        if (isAssignblFrm(Disposable.class, hasValues)) ((Disposable)hasValues).dispose();
        StrapAlign.dispatchLater(CB!=null ? RESIDUE_SELECTION_CHANGED : VALUE_OF_RESIDUE_CHANGED,111);
        StrapView.getPlottersV().remove(this);
        if (tabPane!=null) tabPane.remove(panel);
        super.dispose();
    }

    /* --- Methods for ResidueSelection and Plotting  --- */
    public int getModificationCount() {
        getValues();
        return modiCount(sliderL)+modiCount(sliderU)+modiCount(comboProt)+modiCount(vor)+modiCount(vap);
    }
    /* --- Methods for ResidueSelection --- */
    public void setSelectedAminoacids(boolean bb[], int offset) {}
    public int getSelectedAminoacidsOffset() { return 0;}
    private void updateSelection() {
        final Protein prot=p();
        for(Protein p : a().getProteins()) {
            if (p!=prot) p.removeResidueSelection(this);
        }
        if (prot!=null) {
            prot.addResidueSelection(this);
            StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
        }
    }
    private int selectedModi;
    public boolean[] getSelectedAminoacids() {
        final StrapProtein p=p();
        final double[] vv=  getValues();
        //putln("getSelectedAminoacids "+p+"  "+vv);
        if (vv==null || p==null) return null;
        final int modi=getModificationCount();
        if (selectedAA==null || selectedModi!=modi) {
            selectedModi=modi;
            Protein.incrementModificationCount(ProteinModiCount.MC_RESIDUE_SELECTIONS, p);
            final int nR=p.countResidues();
            final boolean bb[]=redim(selectedAA, nR, 99);
            double maxVal=Double.MIN_VALUE, minVal=Double.MAX_VALUE;
            for(double v:vv) { if (maxVal<v) maxVal=v; if (minVal>v) minVal=v; }
            final double
                f=(maxVal-minVal)/100,
                lower= minVal+sliderL.getValue()*f,
                upper= minVal+sliderU.getValue()*f;

            for(int iA=nR; --iA>=0; ) {
                final int i=vor!=null ? iA : p.getResidueColumn(iA);
                final double v=i>=0 && i<vv.length ? vv[i] : Double.NaN;
                bb[iA]=!Double.isNaN(v) && v<=upper && v>=lower;
            }
            selectedAA=bb;
        }
        return selectedAA;
    }

    public int getStyle() {return comboStyle.getSelectedIndex();}
    public int getVisibleWhere() {
        if (CB==null) return 0xFFffFFff;
        return
            (CB[0].s() ? VisibleIn123.SEQUENCE : 0) |
            (CB[1].s() ? VisibleIn123.SB : 0) |
            (CB[2].s() ? VisibleIn123.STRUCTURE : 0);
    }
    public void setVisibleWhere(int w){
        if (CB!=null) {
            CB[0].s( (w&VisibleIn123.SEQUENCE)!=0);
            CB[2].s( (w&VisibleIn123.STRUCTURE)!=0);
            CB[1].s( (w&VisibleIn123.SB)!=0);
        }
    }

    public final void handleEvent(StrapEvent ev) {
        if (CB!=null && ev.getType()==VALUE_OF_ALIGN_POSITION_CHANGED) {
            final Object q=ev.getSource(), oo[]=ev.getParameters();
            if (q==hasValues || oo.length==0 || idxOf(hasValues,oo)>=0) {
                getValues();
                if (getModificationCount()!=modiCount) StrapAlign.dispatchLater(RESIDUE_SELECTION_CHANGED,111);
            }
        }
    }
}
