package charite.christo.strap;
import charite.christo.interfaces.*;
import charite.christo.*;
import javax.swing.JComponent;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static charite.christo.strap.StrapScriptInterpreter.*;
/**HELP

<div class="figure">

    <table style="caption-side: bottom">
        <CAPTION><b>Figure: </b>
        Generation of a Web link. The included information  is specified with check-boxes and radio buttons.
        The link text appears in the text field from where it can be copied into the html source. A test button allows to quickly test the
        generated link.
        </CAPTION>
        <tr><td> <i>JCOMPONENT:new DialogPublishAlignment("")</i></td></tr>

      </table>

</div>
With this dialog a Web-Link can be formed which loads the proteins from the public databases into STRAP and displays an alignment.
      When the link is clicked on a computer with Java, the alignment will be displayed in STRAP.

      Depending on the purpose of the link and on how much information should be transfered
      two different types are available:
    <ol>
      <li>A compact URL</li>
      <li>A Web form</li>
    </ol>

    In both types, proteins that are stored in protein databases are
    included by database reference. This has the advantage that always the most recent
    version of the protein file is loaded and  the current state of
    sequence features and cross links are available.
    <h3>Single URL</h3>

    Since the URL contains the information in a compact form in a single line, not all information is stored.

    The advantage is that the WIKI:URL can not only be included in web-pages, but also in  e-mails, Office-documents.

    The generation of the URL is conducted in two steps:
    <ol>
      <li> <i>BUTTON:DialogPublishAlignment#BUT_LAB_P</i>
        The parameter String is written into the first text-field and can be modified by the user.
        The following table summarizes the "|"-separated fields.
        <div class="floatR">
          <table border="1" >
            <caption>  WIKI:Vertical_bar separated fields of  protein entries</caption>
            <tr><th>No</th><th>Description</th><th>Example</th></tr>
            <tr><th>1</th><td>URL of protein file or database colon ID</td> <td>EMBL:M57965</td> </tr>
            <tr><th>2</th><td>Protein name. Optionally with exclamation mark and residue subset. </td> <td>b_myosin_heavy_chain</td> </tr>
            <tr><th>3</th><td>Icon </td> <td>http://www.ebi.ac.uk/thornton.../duc_temp.gif</td> </tr>
            <tr><th>4</th><td>Residue selections. Supported  3D-styles: sticks, spheres and dots. </td> <td>#00FFFF,sticks,16-20,#FF00FF,spheres,40-50</td> </tr>
            <tr><th>5</th><td>Coding sequence of a nucleotide sequence </td> <td>reverse,40-100</td> </tr>
          </table>
        </div>
        The 5th field is required only for nucleotide sequence files. It contains either the index of the CDS  such as "#1" for the first or "#2" for the second CDS or the
        CDS expression directly.
      </li>
      <li> <i>BUTTON:DialogPublishAlignment#BUT_LAB_E</i>
        From the text in the first text-field the web-link will be generated using WIKI:Url_encoding and written into the 2nd text-field.
        The generated URL acts as a hyperlink and can be tested by clicking. A new STRAP session will be opened in web-mode and an alignment will be loaded with the specified information.
      </li>
    </ol>

   <h3>Web form</h3>

    Since the web form has no  size limitation,  the entire information for the alignment can be included.
    The draw-back is that it can only be included in web-pages, but not in office documents or e-mails.
    A minimal html-page including the web link for the selected proteins is generated on pressing <i>BUTTON:DialogPublishAlignment#BUT_LAB_W</i>.

    For testing it is loaded into the web browser upon pressing <i>BUTTON:DialogPublishAlignment#BUT_LAB_B</i>.

    From this html code the text between the opening and closing and &lt;body&gt; tags can be used in any html-page.

    An overview of the STRAP scripting commands is given in http://3d-alignment.eu/web/strap_script.html.
    <br><br>
    The following commands are available:
    <p class="data">
    <i>STRING_ARRAY:StrapScriptInterpreter#allScriptCommands()</i>
    </p>

    @author Christoph Gille
*/
public class DialogPublishAlignment extends AbstractDialogJPanel implements java.awt.event.ActionListener {
    public final static String
        BUT_LAB_B="Test web page in Web Browser",
        BUT_LAB_W="\u2193\u2193\u2193 Generate Web Form \u2193\u2193\u2193",
        BUT_LAB_P="\u2193\u2193\u2193 Generate parameter String  \u2193\u2193\u2193",
        BUT_LAB_E="\u2193\u2193\u2193 Form URL \u2193\u2193\u2193";

    private final ChTextArea taPARA=new ChTextArea(""), taENCODED=new ChTextArea(""), taForm=new ChTextArea("");
    private final ChButton toggles[]=new ChButton['z'];
    private Object[] radioAnnotations;
    public void actionPerformed(java.awt.event.ActionEvent ev) {
        String php="http://www.bioinformatics.org/strap/strap.php";

        final String cmd=ev.getActionCommand();
        StrapProtein[] pp=align.getSelectedProteinsInVisibleOrder();
        if (sze(pp)==0) pp=align.getVisibleProteins();
        final ByteArray sb=new ByteArray(333);
        final int iRadioAnnotations=radioGrpIdx(radioAnnotations);
        final ResidueAnnotation selectedAnnotations[]=(ResidueAnnotation[]) align.SELECTED_OBJECTS.residueSelections('A');
        if (cmd=="F") sb.append("<form action=\"").append(php).append("?separateInstance=t&no3D=all\"  method=\"POST\" >\n <input  type=\"hidden\"  name=\"")
            .append(toggles['g'].s()?"load":"align").append("\" value=\"\n");
        if (cmd=="P" || cmd=="F") {
            for(StrapProtein p : pp) {
                final String acc=p.getAccessionID(), nam=p.getName(), pdb=p.getPdbID();
                String pName=null;
                if (pdb!=null) {
                    final String ch=p.getChainsAsString();
                    pName= (sze(ch)>0 ? pdb+ "_"+ch : pdb)+".pdb";
                    sb.append(pdb);
                    if (sze(p.getChainsAsString())>0) sb.append('_').append(p.getChainsAsString());
                } else if (acc!=null)  {
                    pName=acc; /* KEGG */
                    sb.append(acc);
                } else {
                    if (p.getURL()==null) continue;
                    sb.append(p.getURL());
                }
                sb.append('|');
                if (pName!=null && !pName.substring(pName.indexOf(':')+1).replace(':','_').equals(nam)) sb.append(nam);
                if (cmd!="F") {
                    sb.append('|');  /* ---  Icon --- */
                    if (toggles['i'].s()) {
                        final String iconUrl=readStrg(p.iconLink());
                        if (sze(iconUrl)>0 && ! iconUrl.startsWith(URL_RCSB_IMAGES))  sb.append(iconUrl.trim());
                    }
                    sb.append('|'); /* ---  ResidueAnnotation --- */
                    if (iRadioAnnotations!=2) {
                        for(ResidueAnnotation a:p.getResidueAnnotations()) {
                            if ((iRadioAnnotations==0 || cntains(a,selectedAnnotations)) && a.getFeatureName()==null) {
                                final java.awt.Color color=a.getColor();
                                if (color!=null && a.isEnabled() && a.getFeatureName()==null) {
                                    toHexB(color.getRGB(),6, sb.append('#')).append(',').append(a.getValue(ResidueAnnotation.POS));
                                }
                            }
                        }
                    }
                    sb.append('|'); /* --- CDS --- */
                    final boolean translated[]=p.areNucleotidesCurrentStrandTranslated();
                    if (sze(translated)>0 && ( !p.cdsExpressionApplied() || sze(p.getCDS())>1) ) {
                        if (p.isReverse() && p.isComplement()) sb.append("complement,");
                        boolToTxt(translated,0, ",", "..", sb);
                    }
                    sb.append('|');
                }
                taPARA.t(sb.append('\n').replace("|\n","\n").newString()).tools().underlineRefs(UNDERLINE_WEB_COLORS);
            }
            if (cmd=="F") {
                sb.append("\"/>\n <input  type=\"hidden\" name=\"script\" value=\"\n");
                for(StrapProtein p : pp) {
                    if (iRadioAnnotations!=2) {
                        for(ResidueAnnotation a:p.getResidueAnnotations()) {
                            final String an=a.getName();
                            if ((iRadioAnnotations==0 || cntains(a,selectedAnnotations))  && a.getFeatureName()==null) {
                                sb.append(SCRIPT_new_aminoacid_selection+p+"/"+an+", "+a.getValue(ResidueAnnotation.GROUP)+", "+a.getValue(ResidueAnnotation.POS)+"\n");
                                sb.append(SCRIPT_set_annotation+p+"/"+an+", Color ,#");
                                toHexB(a.getColor().getRGB(),6,sb).append('\n');
                                for(ResidueAnnotation.Entry e : a.getEntries()) {
                                    String k=e.getKey(), v=e.getValue();
                                    if (k==ResidueAnnotation.POS || k==ResidueAnnotation.GROUP || k==ResidueAnnotation.NAME || !e.isEnabled()) continue;
                                    if (v.indexOf(Hyperrefs.PAIR_ALIGNMENT)>=0 || v.indexOf(SequenceFeatures.ALI_SCORE)>=0) continue;

                                    final String kShort=StrapPlugins.mapL2S(k);
                                    if (kShort!=null) k=kShort;

                                    sb.append(SCRIPT_add_annotation+p+"/"+an+", "+k+" ,"+v+"\n");
                                }
                            }
                        }
                    }
                    final String iconUrl=readStrg(p.iconLink());
                    if (toggles['i'].s() && sze(iconUrl)>0 && ! iconUrl.startsWith(URL_RCSB_IMAGES)) sb.append(SCRIPT_icon +p+", "+iconUrl+"\n");
                    if (toggles['g'].s()) sb.append(SCRIPT_gaps+p+", "+p.getGappedSequenceAsString().replace(' ','-')+"\n");
                    final charite.christo.protein.Matrix3D m3d=toggles['3'].s() ? p.getRotationAndTranslation() : null;
                    if (m3d!=null && !m3d.isUnit()) sb.append(SCRIPT_rotate_translate+p+", "+m3d.toText(0,null)+"\n");
                }
                sb.append("\n\"/> <input type=\"SUBMIT\" name=\"SUBMIT\" value=\"This is the Web-Button to start STRAP.\"/></form>\n");
                taForm.t(sb.newString()).tools().underlineRefs(UNDERLINE_WEB_COLORS|UNDERLINE_REFS_NOT_CLICKABLE|UNDERLINE_REFS_NO_ICON);
            }
        }
        if (cmd=="E") {
            ByteArray ba1=new ByteArray(333).append(php).append("?align=")
                .urlEncode(taPARA.toString().replaceAll("\\s+"," ").replaceAll("\\|+ "," ").getBytes(),0,MAX_INT);
            taENCODED.setText(ba1.newString());
        }
        if (cmd=="B"){
            final java.io.File fHtml=createTempFile(".html");
            wrte(fHtml,"<html><body>\n"+taForm.toString()+"\n</body></html>\n");
            visitURL(fHtml, modifiers(ev));
        }
    }

    public DialogPublishAlignment() {  this((String)null);}
    public DialogPublishAlignment(String isHtmlDocumentation) {
        taENCODED.setLineWrap(true);
        taPARA.setLineWrap(true);
        taPARA.tools().underlineRefs(UNDERLINE_WEB_COLORS|UNDERLINE_REFS_NOT_CLICKABLE|UNDERLINE_REFS_NO_ICON);
        for(ChTextArea ta : new ChTextArea[]{taPARA, taENCODED, taForm}) {
            ta.tools().enableWordCompletion(new Object[]{dirWorking(),allScriptCommands()});
            ta.tools().enableUndo(true);
        }
        toggles['i']=toggl("Include protein icons").s(true).tt("Protein images dragged from Web pages or file browsers onto proteins");
        radioAnnotations=radioGrp(new String[]{"All","Only selected", "None"}, 0, this);
         final javax.swing.JTabbedPane tabbed=new javax.swing.JTabbedPane();
         final JComponent pNorth=newCNSEW( null, pTop(),null,toggles['i'].cb(),pnl("#ETB","  Included residue annotations: ",radioAnnotations));

         {
             final JComponent
                 butP=new ChButton("P").li(this).t(BUT_LAB_P),
                 butE=new ChButton("E").li(this).t(BUT_LAB_E);
            tabbed.add("Simple URL", pnl(new java.awt.GridLayout(2,1),newCNSEW(taPARA.tools().scrollPane(),pnl(butP),null,null,null), newCNSEW(taENCODED,pnl(butE),null,null,null)));
        }
        {
            toggles['g']=toggl("Include the alignment gaps").tt("The alignment gaps are either transfered or computed on the client useing ClustalW");
            toggles['3']=toggl("Include the 3D-rotation/translation").tt("Only for proteins with 3D-structure. 3x3 Rotation Matrix and translation Vector.");
            final JComponent
                but=new ChButton("F").li(this).t(BUT_LAB_W),
                pNorth2=newCNSEW(pnl(but),null,null,null, pnl(toggles['g'].cb(),toggles['3'].cb()));
            tabbed.add("Comprehensive Web form", newCNSEW(taForm.tools().scrollPane(),pNorth2,hB(new ChButton("B").li(this).t(BUT_LAB_B)),null,null));
        }
        if (isHtmlDocumentation!=null) tabbed.setSelectedIndex(1);
        cnsew(this,  tabbed, pNorth, null,null,null);
    }

}
