package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import java.awt.*;
import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.*;
import java.awt.event.*;
import java.util.List;
import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP

The dialog  allows to perform string replacements in many annotations, simultaneously.

First the entry type like "Remark" or "Note" needs to be chosen in the choice menu.

In the alignment panel the annotations  need to be selected.
Selected annotations have WIKI:marching_ants.

All residue annotations that have the specified  entry type like "Remark"  appear
in the table.

In the  first and second  row the search strings and replacements are typed.

The preview of the string replacements is immediately shown in the table.

<br><br>
There are three options:
<ol>

<li>Pressing
<i>BUTTON:DialogResidueAnnotationChanges#BUT_LAB_Replace</i>
changes the selected entries.</li>

<li>Pressing <i>BUTTON:DialogResidueAnnotationChanges#BUT_LAB_Add</i>
keeps the old entry and generates a further one.
</li>

<li>Pressing
<i>BUTTON:DialogResidueAnnotationChanges#BUT_LAB_Kill</i> deletes
the entries.</li>

</ol>

Frequently it is necessary to add <b>identical</b> entries to a number of selections.
For this task the dialog  of the context menu :
<i>ITEM:ResidueSelectionPopup#ACTION_addLine</i> can be used.

<br><br>

Note WIKI:Regular_expressions of Java are  unusal concerning capturing of groups: Groups are captured with parenthesis and not backslash parenthesis.
The captured groups can be references by dollar number.

<i>SEE_CLASS:AddAnnotation</i>
<i>SEE_CLASS:ResidueAnnotation</i>
<i>SEE_CLASS:ResidueAnnotationView</i>
<i>SEE_CLASS:ActivateDeactivate</i>

@author Christoph Gille

*/
public class DialogResidueAnnotationChanges extends AbstractDialogJPanel implements TableModel,StrapListener,TableCellRenderer,TableCellEditor,ActionListener,Runnable {
    private final ChJTable jTable;
    private String key="Note";
    public final static String
        BUT_LAB_Add="Add",
        BUT_LAB_Kill="Kill",
        BUT_LAB_Replace="Replace";
    private final static int NUM=4;
    private final String[] replacement=new String[NUM], searchPattern=new String[NUM];
    private final List<ResidueAnnotation.Entry> vEntries=new UniqueList(ResidueAnnotation.class);
    private final List<CellEditorListener> listeners=new ArrayList();
    private final JComboBox comboKeys;
    private final ChLabel labInfo=new ChLabel().fg(0xFF0000), LABEL=new ChLabel();
    private final ChButton cbRegex=toggl("Use regular expression").li(this);
    public void run() {
        final Object oKey=comboKeys.getSelectedItem();
        if (oKey!=null) {
            key=oKey instanceof Class ? ((Class)oKey).getName() : toStrg(oKey);
            final ResidueAnnotation aa[]=(ResidueAnnotation[])align.SELECTED_OBJECTS.residueSelections('A');
            vEntries.clear();
            for(ResidueAnnotation s:aa) {
                for(ResidueAnnotation.Entry e : s.getEntries()) {
                    if (e.getKey().equals(key)) vEntries.add(e);
                }
            }
            labInfo.t(
                      aa.length==0 ? "Error: no residue annotations is selected  !" :
                      sze(vEntries)==0 ? "Error: non of the "+aa.length+" selected ResidueAnnotation's has an entry  "+key+" " :
                      "");
            if (isAlreadyPainted()) {
                jTable.repaint();
                jTable.revalidate();
            }
        }
    }
    @Override public void paintComponent(Graphics g){}
    public void actionPerformed(ActionEvent ev) {
        final String cmd=ev.getActionCommand();
        final Object q=ev.getSource();
        final StrapAlign align=StrapAlign.getInstance();
        if (q==comboKeys) ChDelay.runAfterMS(EDT,this,444);
        if (cmd==ACTION_ENTER)  fireEditingStopped();
        if (cmd==BUT_LAB_Replace || cmd==BUT_LAB_Add || cmd==BUT_LAB_Kill) {
            final Replacement rr[]=getSelected();
            if (sze(rr)>0) {
                if (cmd==BUT_LAB_Replace) change(rr,'R');
                if (cmd==BUT_LAB_Add) change(rr,'A');
                if (cmd==BUT_LAB_Kill) change(rr, 'K');
                ChDelay.runAfterMS(EDT,this,11);
                StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
            }
        }
        if (cmd=="EDITOR") StrapAlign.editAnnotation(editedResidueAnnotation);
        if (q==cbRegex) jTable.revalidate();
    }
    private static class Replacement{ ResidueAnnotation a; ResidueAnnotation.Entry entry; String newValue;}
    private Replacement[] getSelected() { // 0:ResidueSelection  1:new Value 2:old Value
        final int ii[]=ChJTable.selectedRowsConverted(jTable);
        final Replacement[] repl=new Replacement[ii.length];
        boolean isSelected=false;
        for(int i=0; i<ii.length; i++)
            if (ii[i]>1) {
                final Replacement r=repl[i]=new Replacement();
                final Object valAt=getValueAt(ii[i],0);
                if (!(valAt instanceof ResidueSelection)) continue;
                r.a=(ResidueAnnotation)valAt;
                r.newValue=getValueAt(ii[i],getColumnCount()-1).toString();
                r.entry=vEntries.get(ii[i]-2);
                isSelected=true;
            }
        if (!isSelected) {
            labInfo.t("No rows in table selected").fg(0xFF0000);
            return null;
        }
        return repl;
    }
    private void change(Replacement rr[], char mode) {
        for(Replacement r:rr) {
            final ResidueAnnotation a=r!=null ? r.a : null;
            if (a!=null) {
                final ResidueAnnotation.Entry e=r.entry;
                if (mode=='K') a.removeEntry(e);
                else if (mode=='A') a.addE(e.getKey(),r.newValue);
                else if (mode=='R') e.setValue(r.newValue);
            }
        }
    }
    public DialogResidueAnnotationChanges() {
        comboKeys=new ChCombo(ChJTable.CLASS_RENDERER, new ComboKeys(true)).li(this).s("Remark");
        Arrays.fill(searchPattern,"");
        Arrays.fill(replacement,"");
        jTable=new ChJTable(this,0);
        jTable.setDefaultRenderer(ResidueAnnotation.class,this);
        jTable.setDefaultRenderer(String.class,this);
        jTable.setDefaultEditor(ResidueAnnotation.class,this);
        final ChRenderer renderer=new ChRenderer();
        final ChTextField tf=renderer.textField();
        tf.li(this).tools().highlightOccurrence(" ",null,null,HIGHLIGHT_UPDATE_IF_TEXT_CHANGES,C(DEFAULT_BACKGROUND));
        jTable.setDefaultEditor(String.class,renderer);
        final JComponent
            pKey=hB("Key:  ",comboKeys),
            pNorth=vBhB(pTop(),pKey,labInfo),
            pButtons=pnl(
                         cbRegex.cb(),
                         new ChButton(BUT_LAB_Replace).li(this)
                         .tt("<b>Caution</b><br>replace the entries in the  residue selections<br> by the table entries in the right column"),
                        " ", new ChButton(BUT_LAB_Add).li(this)
                        .tt("add new key value pairs <br>to the sets of residues"),
                        " ", new ChButton(BUT_LAB_Kill).li(this)
                        .tt("<b>Caution</b><br>delete the selected key-value pairs <br>from the residue selections")
                         );
        cnsew(this,scrllpn(0L,jTable),pNorth,pButtons,null,null);
        ChDelay.runAfterMS(EDT,this,999);
    }
    public final void handleEvent(StrapEvent ev) {
        int type=ev.getType();
        if (type==StrapEvent.OBJECTS_SELECTED || type==StrapEvent.RESIDUE_SELECTION_CHANGED) {
            ChDelay.runAfterMS(EDT,this,444);
        }
    }
    /* >>>   TableModel  >>> */
    public int getRowCount(){ return 2+sze(vEntries);};
    public int getColumnCount(){return 2+2;};

    public String getColumnName(int col){
        return
            col==0 ? "ResidueAnnotation" :
            col==1 ? "original value" :
            "replacement "+(col-1);
    }
    public Class getColumnClass(int col){ return col==0 ? ResidueAnnotation.class:String.class;}
    public boolean isCellEditable(int row, int col){return col>1&& row<2 || col==0 && row>1;}
    public Object getValueAt(int row, int col){
        Object ret=null;
        if (row<2) {
            if (col==0) ret=row==0 ? "search string" : "replacement";
            else if (col==1) ret="";
            else if (row==0) ret=searchPattern[col-2];
            else if (row==1) ret=replacement[col-2];
        } else if  (row-2>=sze(vEntries)) ret="error";
        else {
            final ResidueAnnotation.Entry entry=vEntries.get(row-2);
            final ResidueAnnotation a=entry.getResidueAnnotation();
            if (col==0) ret=a;
            else if (col==1) ret=entry.getValue();
            else if (col>1) {
                final String haystack=toStrg(getValueAt(row,col-1)) ;
                final String needle=searchPattern[col-2];
                final String replacement=this.replacement[col-2];
                if (sze(needle)==0 || sze(replacement)==0 || sze(haystack)==0) ret=haystack;
                else {
                    if (cbRegex.s()) {
                        try {
                            ret=ChRegex.replace(ChRegex.pattern(needle),replacement,haystack);
                        } catch(Exception ex) {labInfo.t("Error on processing regexpr. "+needle+"  "+ex);}
                    } else ret=rplcToStrg(needle,replacement,haystack);
                }
            }
        }
        return ret!=null ? ret:"";
    }
    public void setValueAt(Object o,int row,int col) {
        if (o==null) return;
        if (row==0 && col>1)
            try {
                searchPattern[col-2]=toStrg(o);
            } catch(Exception e) {labInfo.t("Error on processing regexpr. "+o+"  "+e);}
        if (row==1 && col>1) replacement[col-2]=toStrg(o);
    }
    public void addTableModelListener(TableModelListener li){};
    public void removeTableModelListener(TableModelListener li){};
    /* >>> Renderers  >>> */
    public Component getTableCellRendererComponent(JTable t,Object v,boolean isSel,boolean hasFocus,int row,int col) {
        return component(v,isSel,row,col,false);
    }
    private Component component(Object v,boolean isSel,int row,int col,boolean isEditor) {
        final JComponent c;
        if (v instanceof ResidueAnnotation) {
            final ResidueAnnotation s=(ResidueAnnotation)v;
            final StrapProtein p=(StrapProtein)s.getProtein();
            if (isEditor) editedResidueAnnotation=s;
            return (isEditor ? BUT_SORE_EDITOR : BUT_SORE).t(p+"/"+s.getName()).tt("The residue<br>double click: edit it");
        } else c=LABEL.t("error").fg(isSel && row>1 ? 0xFF : 0);
        String tt=null;
        c.setEnabled(true);
        if (v==null) return LABEL;
        if (v instanceof String) {
            LABEL.t(v.toString());
        }
        if (v instanceof ResidueAnnotation.Entry) {
            LABEL.setText( ((ResidueAnnotation.Entry)v).getValue());
        }
        if (row<2 && col<2) {
            LABEL.t(" ");
        }
        final int nCol=getColumnCount();
        if (row==0 && col>1) {
            final String s=searchPattern[col-2];
            if (sze(s)==0) { LABEL.t("Enter a search string here").fg(0xFF0000); }
            else setFG(0x00b200, c);
            tt="<b>Search pattern</b> <sub>(regular expression)</sub><br>all occurrence in this column<br> will be replaced<br>by the replacement below";
        }
        if (row==1 && col>1) {
            if (sze(replacement[col-2])==0) LABEL.t("enter a replacement here").fg(0xFF0000);
            else setFG(0x00b200, c);
            tt="<b>Replacement</b> for the <br>search string above";
        }
        if (row>1 && col==1) tt="This is the current text";
        if (row>1 && col==nCol-1) tt="This is the  text<br>that would be replaced.";
        if (row>1 && col>1 && col<nCol-1) tt="This is an intermediate text.<br>some replacements may require to steps.";
        LABEL.tt(tt);

        return c;
    }
    private ResidueAnnotation editedResidueAnnotation;
    private final ChButton BUT_SORE_EDITOR=new ChButton("EDITOR").li(this), BUT_SORE=new ChButton();
    /**
       TableCellEditor
    */
    void fireEditingCanceled() {
        ChangeEvent ev=new ChangeEvent(this);
        for(int i=sze(listeners)-1; i>=0;i--) {
            listeners.get(i).editingCanceled(ev);
        }
    }
 void fireEditingStopped() {
        ChangeEvent ev=new ChangeEvent(this);
        for(int i=sze(listeners)-1; i>=0;i--) {
            listeners.get(i).editingStopped(ev);
        }
    }
    public boolean stopCellEditing() {fireEditingStopped();return true; }
    public void cancelCellEditing() {  fireEditingCanceled();}
    public void addCellEditorListener(CellEditorListener li) {listeners.add(li);}
    public void removeCellEditorListener(CellEditorListener li) { listeners.remove(li);}
    public boolean isCellEditable(EventObject o) { return true;}
    public boolean shouldSelectCell(EventObject ev) { return true;}
    public Object getCellEditorValue() {return null;}
    public Component getTableCellEditorComponent(JTable t,Object v,boolean hasFocus,int row,int column) {
        return component(v,false,row,column,true);
    }
}
