package charite.christo.strap;
import charite.christo.protein.*;
import charite.christo.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP

   Only protein  that have 3D-coordinates, can be visualized in STRAP three-dimensionally.

   But proteins loaded from sequence files like Swissprot files lack 3D-information.

   This dialog allows to add 3D-coordinates from a similar structure entry such that
   3D-views can be opened also in these cases.

   <br><br>
   Please note that in many cases no suitable 3D-structure exists yet in the PDB database.
   <br><br>

   For a given amino acid sequence the pdb database is searched for a 3D-structure with the same or a highly similar
   amino acid sequence.

   Only those proteins are considered here that are loaded from sequence files lacking 3D-coordinates.
   Conversely,  proteins that are loaded from a 3D-structure file will be skipped.

   <br><br>

   The search result is presented in a table.
   Each table row represents one protein. It shows candidates of matching PDB-IDs.

   The location of the hit relative to the query sequence is shown graphically as a gray bar.

   <br>

   To infer the 3D-coordinates of the PDB entry upon the amino acid sequence,
   the user needs to press the button <i>BUTTON:"Apply"</i> in the last table column.

   Those amino acids in the amino acid sequence that inherited xyz-3D-coordinates will be written as an upper case
   letter. Those, that are not assigned 3D-coordinates are shown as lower case letter.

   Amino acid positions where the residue type is different to the one of the PDB-file are marked.

   <br><br>

   The associated 3D-structure can be dissociated using the context menu of the protein.

   <br><br>
   You may also search for suitable 3D-structures using <i>BUTTON:DialogBlast!</i> and assign the
   pdb-ID to the proteins manually.

   @author Christoph Gille
*/
public class DialogSimilarStructure extends AbstractDialogJTabbedPane implements java.awt.event.ActionListener, StrapListener {
    private final ProteinList proteinList=new ProteinList(align,0);//.selectAll(ProteinList.NO_CALPHA);
    private final ChCombo comboClass=StrapAlign.classChoice(SimilarPdbFinder.class);
    private final ChButton togSkip=toggl("Skip proteins that are loaded from a 3D-file").s(true);
    private int count;
    public DialogSimilarStructure() {
        pcp(KEY_CHOICE_CLASS,comboClass,this);
        pcp(StrapAlign.KEY_CHOICE_PROTEIN, weakRef(proteinList), this);
        final javax.swing.JComponent
            pNorth=StrapAlign.headLine(this),
            pDetails=vBhB(
                          " ",
                          "Method: ",
                          comboClass.panel(),
                          ChButton.actionShowResults().t("Log blast"),
                          buttn(TOG_CACHE).cb(),
                          togSkip.cb(),
                          " ",
                          new ChButton("M").t("Mannually assign structures").li(this)
                          ),
            pLeft=vBhB(
                       new ChButton("GO").t("Automatically search suitable structures").li(this),
                       hBL("Options ",toggl().actionCollapse(false,pDetails)),
                       pDetails
                       ),
            pRight=proteinList.scrollPane(),
            pCenter=pnl(new java.awt.GridLayout(1,2), pLeft,pRight),
            pMain=newCNSEW(pCenter, StrapAlign.headLine(this),null,null,null);
        adMainTab(pMain,this,null);
    }

    public void actionPerformed(java.awt.event.ActionEvent ev) {
        final String cmd=ev.getActionCommand();
        if (cmd=="M" || cmd=="GO") {
            final Protein ppSel[]=proteinList.getSelectedProteins();
            Protein pp[]=ppSel.clone();
            final boolean skip=isSelctd(togSkip);
            for(int iP=pp.length; --iP>=0;) if (skip&&pp[iP].isLoadedFromStructureFile()) pp[iP]=null;
            pp=rmNullA(pp, Protein.class);
            if (pp.length==0) {
                error("You need to select at least one protein "+(skip?"which has not been loaded from a 3D-file":""));
                return;
            }
            Object clazz="M";
            if (cmd=="GO") {
                if (pp.length>30 && !ChMsg.yesNo("Are you sure to start the search for similar structures for that many proteins?")) return;
                clazz=comboClass.getSelectedItem();
            }
            if (clazz!=null) {
                final DialogSimilarStructureResult r=new DialogSimilarStructureResult(clazz, pp);
                adTab("#"+ ++count+"   ",r,this);
            }
        }
    }
    public void handleEvent(StrapEvent e) {
        final int id=e.getType();
        if (0!=(id|StrapEvent.FLAG_ROW_HEADER_CHANGED)) ChDelay.repaint(this,99);
    }

}