package charite.christo.strap;
import charite.christo.interfaces.*;
import charite.christo.*;
import charite.christo.strap.extensions.*;
import charite.christo.protein.*;
import java.util.*;
import javax.swing.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**
   @author Christoph Gille
*/
public class DialogSimilarStructureResult extends ChPanel implements java.awt.event.ActionListener, Disposable {
  private final int COL_P=0, COL_R=1, COL_ALI=2, COL_WWW=3, COL_3D=4, COL_DETAILS=5, COL_APPLY=6, COLS=7;
    private ChJTable jTable;
    private Object tableData[][];
    private final Object LAB_COMPUTING="Computing ...", LAB_NOT_FOUND="No 3D-structure found";
    private final ChLog chLog=new ChLog(new ByteArray(999));
    private final ChButton buttons[]=new ChButton[COLS];

    private final JProgressBar progress=new JProgressBar(); { progress.setPreferredSize(dim(0,0)); }
    private boolean _disposed;

    private ChTextArea taManual;
    public DialogSimilarStructureResult(Object classes, Protein pp[]) {

        final Object bLog=ChButton.actionViewMessage("Log", chLog.getText());
        if (classes=="M") {
            final ByteArray sb=new ByteArray(999);
            for(Protein p : pp) {
                sb.append(p).append(' ',4);
                if (p.getInferredPdbID()!=null) sb.append(p.getInferredPdbID());
                sb.append('\n');
            }
            final Object wc=StrapAlign.getInstance().vProteins;
            final ChTextComponents tools=(taManual=new ChTextArea(sb)).tools();
            tools.enableUndo(false)
                .enableWordCompletion(wc)
                .highlightOccurrence(wc, charClass(SPC), charClass(SPC), HIGHLIGHT_UPDATE_IF_TEXT_CHANGES, C(0xFFff14,90));
            final Object
                pNorth="Type a pdb-ID after each protein name. Optionally append underscore chain like \"1IRU_D\".",
                pCenter=tools.scrollPane(),
                pButtons=hBL(pnl(new ChButton("M").t("Apply").li(this)," ",bLog));
            cnsew(this,pCenter,pNorth,pButtons,null,null);
            return;
        }

        final Object data[][]=this.tableData=new Object[pp.length][COLS];
        for(int iP=pp.length; --iP>=0;) {
            final List v=new ArrayList();
            v.add(LAB_COMPUTING);
            data[iP][COL_P]=pp[iP];
            data[iP][COL_R]=new ChCombo(ChJTable.DEFAULT_RENDERER, v);
        }
        jTable=new ChJTable(null, ChJTable.DEFAULT_RENDERER|ChJTable.ICON_ROW_HEIGHT|ChJTable.NO_REORDER);
        jTable.setModel(new ChTableModel("Protein","Search Hit from PDB-database", "Alignment", "Web", "3D", "Details", "Apply").setData(tableData));

        for(int i=COLS; --i>=0;) {
            final String t=
                i==COL_ALI ? "Preview alignment" :
                i==COL_WWW ? "WWW" :
                i==COL_3D ? "3D-Preview" :
                i==COL_DETAILS ? "Details" :
                i==COL_APPLY ? "Apply" :
                null;
            if (t==null) continue;
            buttons[i]=new ChButton(t).li(this);
            buttons[i].cp(ChRenderer.KEY_EDITOR_COMPONENT, buttons[i].cln());
            jTable.setColumnWidthC(false, i,buttons[i]);
        }
        final Object
            pButtons=hB(
                        "Act on selected table rows: ",
                        new ChButton("GO").t("Associate Structure").li(this),
                        new ChButton("RM").li(this).t("Remove"),
                        "#",
                        bLog
                        ),
            pCenter=scrllpn(0L, jTable),
            info="After computation has been finished, You need to press the \"Apply\" button in the result rows.",
            pDetails=hBL("Details ", toggl().actionCollapse(false,pButtons));
        cnsew(this,pCenter,info,vB(pDetails, pButtons,progress),null,null);
        progress.setPreferredSize(dim(EX, ICON_HEIGHT));
        progress.revalidate();
        setJPrgrss(progress, 0, 0, pp.length, "");
        for(Object o : oo(classes))  threadM("compute",this,new Object[]{o,pp}).start();
    }

    public void compute(Object clazz, Protein pp[]) {
        for(int iP=0; iP<pp.length; iP++) {
            if (_disposed) return;
            if (pp[iP]==null) continue;
            final String msg="Searching similar structure for "+pp[iP]+". Please wait!";
            setJPrgrss(progress, iP+1, 0, pp.length, msg);
            final Object row[]=this.tableData[iP];
            final SimilarPdbFinder sim=HotswapUtil.instance(clazz,SimilarPdbFinder.class,true);
            sim.setProtein(pp[iP]);
            final SimilarPdbFinder.Result[] rr=sim.getSimilarStructures();
            final ChCombo combo=(ChCombo)row[COL_R];
            final List v=(List)combo.getData();
            v.remove(LAB_COMPUTING);
            v.remove(LAB_NOT_FOUND);
            adAll(rr, v);
            if (sze(v)==0) v.add(LAB_NOT_FOUND);
            else {
                for(int col=COLS; --col>=0;) {
                    if (buttons[col]!=null) row[col]=buttons[col];
                }
            }
            combo.revalidate();
            ChDelay.repaint(jTable, 33);
        }
        setJPrgrss(progress, 0, 0, pp.length,"Finished "+(pp.length<2?"" : "all "+pp.length+" proteins") );

    }
    public void dispose() {
        _disposed=true;
    }

    public void actionPerformed(java.awt.event.ActionEvent ev) {
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        final int modi=ev.getModifiers(), iButton=idxOf(q,buttons);
        final StrapAlign align=StrapAlign.getInstance();
        if (iButton>=0) {
            final int  iRow=jTable.getEditingRow();
            final SimilarPdbFinder.Result r=result(iRow);
            if (r==null) return;
            final Protein p=r.getProtein();
            final String id_chain=delPfx("PDB:",r.getPdbID()), id=withoutLastComponent(id_chain,'_');
            if (iButton==COL_ALI) {
                final int iChain=id_chain.lastIndexOf('_');
                final char chain=iChain<0?0:chrAt(iChain+1,id_chain);
                final ByteArray ba=new ByteArray(3333)
                    .append(
                            "This alignment will show how well the associated structure file fits the sequence under consideration.\n"+
                            "Downloading "+id+" chain="+chain+", please wait ... ");
                final AsciiView view=new AsciiView(ba).showInFrame(p+" ./. "+id_chain);
                thread_showAli(p, id,  chain,  ba, view).start();
            }
            if (iButton==COL_WWW) {
                final CharSequence txt=joinLines(ChTextComponents.webLinks(id, getCustomSettings(Customize.pdbLinks)));
                ChTextComponents.showFrameWeblinks(Customize.pdbLinks, txt, "B");
            }
            if (iButton==COL_3D) {
                final Protein3d p3d=new Protein3d();
                new ChFrame(id_chain).ad(p3d).shw(ChFrame.AT_CLICK|ChFrame.DISPOSE_CHILDREN|ChFrame.DISPOSE_ON_CLOSE|CLOSE_ON_CtrlW);
                Protein3dUtils.thread_addPdbId(id_chain,p3d, StrapProtein.class).start();
            }
            if (iButton==COL_DETAILS) {
                showCtrlPanel(ChFrame.AT_CLICK|ChFrame.PACK, "Control-panel "+id_chain, r.simPdbFinder);
            }
            if (iButton==COL_APPLY && p!=null) {
                StrapProteinUtils.threadInferCoordinates(new Protein[]{p}, new String[]{id_chain}, chLog, 0L).start();
            }
        }
        if (cmd=="GO" || cmd=="RM") {
            final int N=sze(tableData), ii[]=ChJTable.selectedRowsConverted(jTable);
            final boolean all=sze(ii)==0 && ChMsg.yesNo(ChFrame.AT_CLICK, "Set 3D-coordinates for "+N+" proteins?");
            final Protein ppInf[]=new Protein[tableData.length];
            final String ids[]=new String[tableData.length];
            for(int i=all?N:sze(ii); --i>=0;) {
                final int iR=i<sze(ii) ? ii[i] : i;
                final SimilarPdbFinder.Result r=result(iR);
                final Protein p=r==null?null:r.getProtein();
                if (p==null) continue;
                if (cmd=="GO") { ppInf[iR]=p; ids[iR]=r.getPdbID();}
                if (cmd=="RM" && p.getInferredPdbID()!=null) ppInf[iR]=p;
            }
            if (countNotNull(ppInf)!=0) {
                if (cmd=="RM") {
                    if (ChMsg.yesNo("Remove associated 3D-structures from "+countNotNull(ppInf)+" proteins?")) {
                        for(Protein p : ppInf) if (p!=null) p.detach3DStructure();
                        StrapAlign.dispatchLater(StrapEvent.ATOM_COORDINATES_CHANGED,1);
                    }
                }
                if (cmd=="GO") StrapProteinUtils.threadInferCoordinates(ppInf, ids, chLog, 0L).start();
            }
        }
        if (cmd=="M") {

            final ByteArray ba=taManual.tools().byteArray();
            final byte T[]=ba.getBytes();
            final int[] ends=ba.getLineEnds();
            final ChTokenizer TOK=new ChTokenizer();
            final Protein all[]=align.getProteins(), pp[]=new Protein[ends.length];
            final String ids[]=new String[ends.length];
            for(int iL=0; iL<ends.length; iL++) {
                final int b=iL==0?0:ends[iL-1]+1,  e=ends[iL];
                TOK.setText(T,b,e);
                final String
                    prot=TOK.nextToken() ? TOK.asString() : null,
                    pdb=TOK.nextToken() ? TOK.asString() : null;

                final Protein p=StrapProtein.getProteinWithName(prot, all);
                if (p!=null) {
                    pp[iL]=p;
                    ids[iL]=pdb;

                }
                //putln(" prot="+prot+"  pdb="+pdb+"  p="+p);
            }
            StrapProteinUtils.threadInferCoordinates(pp, ids, chLog, 0L).start();
        }
    }
    private SimilarPdbFinder.Result result(int row) {
        final ChCombo combo=row<0?null:(ChCombo)this.tableData[row][COL_R];
        return combo==null?null: deref(combo.getSelectedItem(), SimilarPdbFinder.Result.class);
    }
    private static Thread thread_showAli(Protein p, String pdbId, char chain, ByteArray ba, JComponent ta) {
        return threadM("showAli", DialogSimilarStructureResult.class, new Object[]{p,pdbId,charObjct(chain),ba,ta});
    }
    public static void showAli(Protein p, String pdbId, char chain, ByteArray ba, JComponent ta) {
        final byte seq[]=SequenceForPdbID.getSequence(pdbId, chain);
        final byte[] rt=p==null?null:p.getResidueTypeExactLength();
        final SequenceAligner aligner=new PairAlignerNeoBioPROXY();
        aligner.setSequences(new byte[][]{seq,rt});
        aligner.compute();
        final byte[][] aligned=aligner.getAlignedSequences();
        if (sze(aligned)==2) {
            ba.append('\n');
            ChAlignUtils.aliWithMidline(ChAlignUtils.PRINT_ALI_TRUNCATE, aligned[0], 1, pdbId+(chain!=0?"_"+chain:""), aligned[1], 1, toStrg(p), 60, ba);
            ChDelay.repaint(ta,22);
            packW(ta);

        }

    }
}