package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.strap.interfaces.*;
import charite.christo.protein.*;
import java.awt.*;
import java.util.*;

import javax.swing.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static java.awt.event.KeyEvent.*;
import static charite.christo.interfaces.PredictSubcellularLocation.*;
/**HELP

The cellular compartment is predicted from the amino acid sequence.
@author Christoph Gille
*/
public class DialogSubcellularLocalization extends AbstractDialogJTabbedPane implements java.awt.event.ActionListener,ChRunnable, StrapListener {
    private final ChJList classList;
    private final ProteinList proteinList;
    private final JComboBox comboOrgansims;
    private final ChTextArea LOG=new ChTextArea("");
    public DialogSubcellularLocalization() {
        final UniqueList l=StrapPlugins.allClassesV(PredictSubcellularLocation.class);
        classList=new ChJList(l,ChJTable.CLASS_RENDERER);
        classList.setSelII(new int[]{0,1,2,3,4,5,6,7,8,9});
        comboOrgansims=new JComboBox(Prediction.getOrganisms());
        OverlayScrollPane sp=new OverlayScrollPane(classList).addButton(smallHelpBut(l.toArray())).addButton(vB("Right-click for context-menu",OPAQUE_FALSE," "));
        final JComponent
            pCenter=pnl(new GridLayout(1,2),sp, scrllpn(0L,proteinList=new ProteinList(align,0))),
            pButtons=pnl(comboOrgansims," ",new ChButton("GO").t(ChButton.GO).li(this)),
            pMain=newCNSEW(pCenter,StrapAlign.headLine(this),pButtons,null,null);
        adMainTab(pMain, this,null);
        LOG.tools().underlineRefs(0);
        pcp(StrapAlign.KEY_CHOICE_PROTEIN,proteinList,this);
    }

    public void process(StrapProtein[] pp) {
        final Object cc[]=classList.getSelectedValues();
        final Object[][] tableData=new Object[pp.length][];
        for(int iP=pp.length;--iP>=0;) {
            tableData[iP]=new Object[cc.length+1];
            tableData[iP][0]=pp[iP];
            for(int iC=cc.length; --iC>=0;) {
                tableData[iP][iC+1]=iicon(IC_HOURGLASS);
            }
        }
        final String colNames[]=new String[cc.length+1];
        for(int i=cc.length; --i>=0;) colNames[i+1]=delPfx("SubcellularLocation",shrtClasNam(cc[i]));
        colNames[0]="protein";
        final ChTableModel tm=new ChTableModel(colNames).setData(tableData);
        final JComponent
            jTable=new ChJTable(tm,ChJTable.ICON_ROW_HEIGHT|ChJTable.DEFAULT_RENDERER),
            butLog=ChButton.actionViewMessage("Log",LOG),
            pnl=newCNSEW(scrllpn(0L,jTable),null,pnl(butLog, new ChButton("debug")) ,null,null);
        closeOnCtrlW(false,jTable, pnl);
        adTab(String.valueOf(++countTab),pnl,  (JTabbedPane)this);
        for(int iC=0; iC<cc.length; iC++) {
            ChDelay.newThread(this,"GO",new Object[]{tableData,pp,cc[iC],intObjct(iC),jTable,pnl}).start();
        }
    }

    private int countTab;
    private PredictSubcellularLocation lastPredictor;
    public void actionPerformed(java.awt.event.ActionEvent ev) {
        final String cmd=ev.getActionCommand();
        if (cmd=="GO") {
            final StrapProtein pp[]=proteinList.getSelectedOrAllProteins();
            process(pp);
        }
        if (cmd=="debug"){
            final String scn=shrtClasNam(lastPredictor);
            if (scn==null) error("Error: No prediction performed!\nThis button shows the control-panel of the last prediction process.\n");
            else if (!(isAssignblFrm(HasControlPanel.class, lastPredictor))) error("Error: "+scn+" does not have a control panel.\n");
            else new ChFrame(scn).ad(getCtrlPanel(lastPredictor)).shw(0);
        }
    }
    public Object run(String id, Object arg) {
        final Object[] argv= arg instanceof Object[] ? (Object[])arg:null;
        if (id=="GO") {
            final Object[][] td=(Object[][])argv[0];
            final StrapProtein pp[]=(StrapProtein[])argv[1];
            final int iC=atoi(argv[3]);
            final JTable jTable=(JTable)argv[4];
            final JComponent tab=(JComponent)argv[4];
            final StringBuffer sbF=new StringBuffer(), sbP=new StringBuffer(33);
            for(int iP=0; iP<pp.length && !isDisposed() && SwingUtilities.getRootPane(tab)!=null;  iP++) {
                sbF.setLength(0);
                sbP.setLength(0);
                final byte[] rt=pp[iP].getResidueTypeExactLength();
                final PredictSubcellularLocation predictor=HotswapUtil.instance(argv[2],PredictSubcellularLocation.class,false);
                if (predictor==null) continue;
                lastPredictor=predictor;
                final String cn=shrtClasNam(predictor);
                predictor.setResidueType(rt);
                predictor.setOrganism(comboOrgansims.getSelectedIndex());
                sbF.append(STRAPTMP).append("/where/").append(delPfx("SubcellularLocation",cn)).append('_');
                CacheResult.sectionForSequence(rt,sbF);
                final PredictSubcellularLocation.Prediction[] predictions=predictor.getCompartments();
                java.io.File f=null;
                if (predictions==PredictSubcellularLocation.ERROR || predictions==null) td[iP][iC+1]=iicon(IC_BUG);
                else {
                    for(PredictSubcellularLocation.Prediction pred : predictions) {

                        sbP.append( get(pred.getCompartment(),  getFinalStaticInts(PredictSubcellularLocation.class, PredictSubcellularLocation.MAX_VARIABLE)));
                        if (pred.getScore()!=Integer.MIN_VALUE) sbP.append('(').append(pred.getScore()).append(')');
                        sbP.append(' ');
                    }
                    final String appnd=pp[iP]+" "+filepathUnixStyle(f)+" "+sbP+"\n";
                    final ByteArray baLOG=LOG.tools().byteArray();
                    if ( strstr(appnd,baLOG)<0) LOG.appnd(appnd);
                    td[iP][iC+1]=sbP.length()==0 ? "unknown" : sbP.toString();
                }
                ChDelay.repaint(jTable,333);
            }
        }
        return null;
    }
    public void handleEvent(StrapEvent ev) {
        final int t=ev.getType();
        if ( (t&StrapEvent.FLAG_ROW_HEADER_CHANGED)!=0) ChDelay.repaint(this,333);
    }
}
