package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP

Two protein structures can be superimposed three-dimensionally using
C-alpha coordinates so that they are oriented in 3D-space in the same
way.

The mobile protein and the reference protein[s] are selected by the
user.

The user can specify a superposition method
<i>COMBO:charite.christo.strap.extensions.Superimpose_CEPROXY</i>.

   Pressing <i>LABEL:ChButton#BUTTON_GO</i>  finds an optimal  rotation/translation that minimizes
   the square root mean distance (RMSD) between C-alpha  atoms of both proteins.
   The result is displayed in a preview panel.

   Only after pressing <i>BUTTON:DialogSuperimpose3D#BUT_LAB_Apply</i>
   this transformation is applied to the mobile protein whereas the
   target protein remains unchanged.

   When the alignment is saved the transformation of the superimposed
   protein will be written to HD.

   These files are found int "./gaps/" and have the suffix ".<I>3dTrans</I>".

   <br><br><b>Residue subset</b> Usually the entire protein structures
   are used, but the calculation can also be confined to a subset of
   residues such as "10-100,120-150".

   <br><br><b>Highlighting structurally equivalent residues:</b>
   Superposition can facilitate the alignment of proteins when the
   3D-structures are similar while their sequence are not due to
   divergent evolution (see PUBMED:7373651).  With the
   <i>DIALOG:DialogSelectionOfResiduesMain</i> all residues adjacent
   to the cursor position can be highlighted.

   <br><br><b>Writing PDB-files in new coordinates:</b> See
   <i>DIALOG:DialogExportProteins</i>.

   <br><br>
   <b>Related sites:</b> http://www.cgl.ucsf.edu/home/meng/grpmt/structalign-content.html

   @author Christoph Gille
*/
public class DialogSuperimpose3D  extends AbstractDialogJTabbedPane implements ActionListener,ChRunnable {
    public final static String BUT_LAB_Apply="Approve";
    private final ChTextField[] tfSelection=new ChTextField['R'+1];
    private final ProteinCombo comboProteinR=new ProteinCombo(align,ProteinList.CALPHA);
    private final ProteinList comboProteinM=new ProteinList(align,ProteinList.CALPHA);
    private final ChCombo comboClass;
    private final JComponent togOptions;
    private StrapProtein[] setProteins;
    public DialogSuperimpose3D() {
        comboProteinM.li(this);
        {
            final StrapProtein pp[]=align.getProteins();
            int iP=0;
            while(iP<pp.length) if (pp[iP++].hasCalpha()) { comboProteinR.s(pp[iP-1]); break;}
            while(iP<pp.length) if (pp[iP++].hasCalpha()) { comboProteinM.setSelI(iP-1); break;}
        }

        final JComponent[] panSel=new JComponent['R'+1];
        for(char i='C'; i<'R'+1; i++) {
            final String
                ttE=".<br>If kept empty all c-Alpha-atoms are taken.<br>Example: \"10-100,120-199\" "+
                "<br><br>Residue selections can be copied via Drag-and-Drop.",
                ttSel="Insert alignment cursor position";
            final ChTextField tf=i=='C' || i=='M' || i=='R' ? new ChTextField() : null;
            if (tf==null) continue;
            tfSelection[i]=tf;
            tf.tools().enableUndo(false).saveInFile("DialogSuperimpose3D_tf"+i);
            StrapAlign.newDropTarget(tf,false);
            panSel[i]=newCNSEW(tf,null,null,new ChButton("C"+i,ChButton.ICON_SIZE).li(this).i(IC_CURSOR16).t(null).tt(ttSel),null);
            if (i=='M' || i=='R') {
                pcp(KEY_IF_EMPTY,"Optional: residue range like \"10-20\"",tf);
                tf.tt("Selected amino acids"+ttE);
                pcp(ResidueSelection.class, weakRef(i=='M' ? comboProteinM : comboProteinR),tf);
            } else {
                pcp(StrapAlign.KEY_ALIGNMENT_COLUMNS, "",tf);
                tf.tt("Selected alignment positions"+ttE);
                pcp(KEY_IF_EMPTY,"Optional: alignment columns like \"10-100,120-200\"",tf);
            }
        }

        comboClass=StrapAlign.classChoice(Superimpose3D.class).save(DialogSuperimpose3D.class,"Superimpose3D");
        pcp(KEY_CHOICE_CLASS,comboClass,this);
        final JComponent
            cbCache=buttn(TOG_CACHE).cb(),
            panProteinM=newCNSEW(comboProteinM.scrollPane(),"<b>Mobile protein[s]</b>", vB(panSel['M'],panSel['C']),null,"#ETB"),
            pSettings=vB(comboClass.panel(),
                        cbCache,
                        AbstractSuperposition.newCbCommutative()),
            pWestNorth=
            vBpnl(
                  newCNSEW(comboProteinR,"<b>Reference protein</b>", panSel['R'],null,"#ETB"),
                  " ",
                  new ChButton("GO").t(ChButton.GO).li(this)
                  .tt("<ul><li>SHIFT+CTRL: worse protein as reference</li><li>CTRL: optimal protein as reference</li><li>SHIFT: removes results with a  score below a threshhold</li></ul>"),
                  hBL("Options", this.togOptions=toggl().actionCollapse(false,new Object[]{pSettings, panSel, tfSelection})),
                  pSettings
                  ),
            pWest=newCNSEW(null,pWestNorth,null,null,KOPT_TRACKS_VIEWPORT_WIDTH),
            pCenter=pnl(new GridLayout(1,2),scrllpn(0L,pWest),panProteinM),
            pMain=newCNSEW(pCenter, StrapAlign.headLine(this),null,null,null);
        adMainTab(pMain,this, null);
    }

    /* <<< Constructor <<< */
    /* ---------------------------------------- */
    /* >>> Thread >>> */
    public Object run(String id, Object arg) {
        if (id=="COMPUTING") {
            try {
                pcp("COMPUTING","",this);
                final StrapProtein pR=get(0,arg,StrapProtein.class);
                final StrapProtein ppM[]=get(1,arg,StrapProtein[].class);
                StrapProtein lastMobile=null;
                if (ppM!=null)
                    for(StrapProtein p : ppM)
                        if (p!=null && p.hasCalpha()) lastMobile=p;
                if (lastMobile==null) { StrapAlign.errorMsg("Select at least one mobile protein with a C-alpha atoms");return null;}
                final boolean selectResR[];
                final String sTf[]=new String['R'+1];
                for(int i=sTf.length; --i>=0;) {
                    JComponent jc=tfSelection[i];
                    if (jc!=null && jc.isEnabled() && isSelctd(togOptions)) sTf[i]=jc.toString();
                }
                final String err3="<br>No or too few residues";
                if (sze(sTf['R'])>0) {
                    selectResR=pR.residueSubsetAsBoolean1(sTf['R']);
                    if (countTrue(selectResR)<3) { StrapAlign.errorMsg(pR+" "+sTf['R']+err3); return null;}
                } else { selectResR=null; }
                final UniqueList<DialogSuperimpose3DResult> vSuperimpose=new UniqueList(DialogSuperimpose3DResult.class);
                final Protein protR=pSubset(pR,selectResR);
                for(StrapProtein pM:ppM) {
                    if(isDisposed()) break;
                    if (pR==null || pM==null || !pM.hasCalpha() || ppM.length>1 && pM==pR) continue;
                    sleep(11);
                    final boolean selectResM[];
                    if (sze(sTf['M'])>0 && ppM.length==1) selectResM=pM.residueSubsetAsBoolean1(sTf['M']);
                    else if (sze(sTf['C'])>0) {
                        selectResM=ResSelUtils.columnsToResiduePosition(parseSet(sTf['C'], pM.countResidues(),-1),pM);
                    } else selectResM=null;
                    if (selectResM!=null && countTrue(selectResM)<3) {error(pM+" "+selectResM+err3); continue;}
                    final Superimpose3D s3d=HotswapUtil.instance(comboClass,Superimpose3D.class,true);
                    if (s3d==null) StrapAlign.errorMsg("Could not instantiate the Superimpose ");
                    else {
                        s3d.setMobileProtein(pSubset(pM,selectResM));
                        s3d.setReferenceProtein(protR);
                        final DialogSuperimpose3DResult s=new DialogSuperimpose3DResult(s3d,this);
                        TabItemTipIcon.set(addHtmlTagsAsStrg("<font size=2>"+noExt(pM)+"<br>"+noExt(pR)+"</font>"),null,null,null,s);
                        ChDelay.addTab(this,s,333);
                        vSuperimpose.add(s);
                        s.run();
                    }
                }
                sleep(999);
            } finally{
                pcp("COMPUTING",null,this);
                revalidate();
        }
        }
        return null;
    }
    /* <<< Thread <<< */
    /* ---------------------------------------- */
    /* >>> AWTEvent >>> */
    public void actionPerformed(ActionEvent ev) {
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        StrapAlign.errorMsg("");
        if (q==comboProteinM) setEnabledIncludingChildren(tfSelection['M'].getParent(),comboProteinM.getSelectedProteins().length==1);
        final StrapProtein pR=comboProteinR.getProtein(), ppM[]=comboProteinM.getSelectedProteins();
        final char c1=chrAt(1,cmd);
        if (sze(cmd)==2 && chrAt(0,cmd)=='C' && (c1=='R'||c1=='M'||c1=='C')) {
            final ChTextField tf=tfSelection[c1];
            if (!tf.isEnabled()) return;
            final StrapProtein p=align.getProteinAtCursor();
            final int iA=align.getIndexOfAminoAcidAtCursor();
            if (iA<0 || p==null) StrapAlign.errorMsg("Inserting Cursor-Position<: first place the cursor in the alignment pane");
            else {
                if (c1=='M' && ppM.length!=1 || c1=='R' && pR==null) return;

                final StrapProtein pDest=c1=='R' ? pR : c1=='M' ? ppM[0] : null;
                final int col=p.getResidueColumn(iA);
                if (col>=0) {
                    int ins=pDest!=null ? pDest.column2nextIndex(col) : col;
                    ChTextComponents.insertAtCaret(0L, " "+(ins+1)+" ",tf);
                }
            }
        }
        if (cmd=="GO") {
            if (isCtrl(ev)) setRepresentingProtein(isShift(ev));
            else if (isShift(ev)) removeAllBelowThreshold();
            else {
                if (pR==null) StrapAlign.errorMsg("Select a reference protein");
                else if(!pR.hasCalpha()) StrapAlign.errorMsg("No c-alpha coordinates in the reference protein");
                else {
                    final StrapProtein pp[]=align.getProteins();
                    final Thread t=ChDelay.newThread(this,"COMPUTING", new Object[]{pR,sze(ppM)==0 && sze(pp)<5 ? pp : ppM});
                    t.setPriority(Thread.MIN_PRIORITY);
                    t.start();
                }
            }
        }
    }
    /* <<< AWTEvent <<< */
    /* ---------------------------------------- */
    /* >>> Score >>> */
    private void removeAllBelowThreshold() {
        float f=(float)atof(ChMsg.input(ChMsg.AT_CLICK, "This removes all superposition resutls<br>with a score below a threshhold.<br>Enter threshhold",8,"3.1"),0);
        removeAllBelowThreshold(f);
    }

    private void removeAllBelowThreshold(double threshhold) {
        for(int i=getTabCount();--i>=0;) {
            final DialogSuperimpose3DResult s3d=deref(getComponentAt(i),DialogSuperimpose3DResult.class);
            final Superimpose3D.Result result=s3d==null?null:s3d.getResult();
            if (result!=null && result.getScore()<threshhold) {
                DialogSuperimpose3D.this.remove(s3d);
                s3d.dispose();
            }
        }
    }
    /* <<< Score <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */
    @Override public void paintChildren(Graphics g){
        super.paintChildren(g);
        if (gcp("COMPUTING",this)==null || getSelectedIndex()>0) return;
        g.setFont(getFnt(ICON_HEIGHT,true,Font.BOLD));
        g.setColor(Color.PINK);
        g.drawString("Computing ...",1,getHeight()-3*ICON_HEIGHT);
        g.drawString("please wait",1,getHeight()-2*ICON_HEIGHT);

    }

    /* <<< Graphics  <<< */
    /* ---------------------------------------- */
    /* >>> Rep  >>> */
    private void  setRepresentingProtein(boolean isBest) {
        StrapProtein pp[]=align.getSelectedProteins();
        if (pp.length==0) pp=align.getProteins();
        final StrapProtein pBestWorse[]=findRepresentingProtein3D(pp,comboClass);
        if (pBestWorse!=null) { comboProteinR.s(pBestWorse[isBest ? 0 : 1]); comboProteinR.revalidate();}
    }

    private static StrapProtein[] findRepresentingProtein3D(StrapProtein pp[], Object comboClass) {
        float bestScoreSum=0,worseScoreSum=99999;
        StrapProtein best=null,worse=null;
        for(StrapProtein pI:pp) {
            float scoreSum=0;
            for(StrapProtein pJ:pp) {
                if (pJ==null || !pJ.hasCalpha() || pI==null || !pI.hasCalpha()) continue;
                final Superimpose3D suim=HotswapUtil.instance(comboClass,Superimpose3D.class,true);
                if (pI==pJ) continue;
                suim.setMobileProtein(pI);
                suim.setReferenceProtein(pJ);
                suim.compute();
                try {
                    final float f=((HasScore)suim).getScore();
                    scoreSum+=f*f;
                } catch(Exception e){putln("caught in findRepresentingProtein3D ",e); return null;}
            }
            if (bestScoreSum<scoreSum) {bestScoreSum=scoreSum;best=pI;}
            if (worseScoreSum>scoreSum) {worseScoreSum=scoreSum;worse=pI;}
        }
        if (Strap.listFileName!=null) {
            final String base=withoutDotSuffix(Strap.listFileName);
            mkdrs(file("info"));
            if (best!=null) wrte(file("info/"+base+".REP"),best+"\n");
            if (worse!=null) wrte(file("info/"+base+".EXTREME"),worse+"\n");
        }
        return new StrapProtein[]{best , worse};
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> static utils >>> */
    private String noExt(Object s) {  return delSfx(".PDB",delSfx(".ENT",delSfx(".pdb",delSfx(".ent",s.toString())))); }
    public static Protein pSubset(Protein p, boolean selRes[]) {
        if (countTrue(selRes)<3) return p;
        final Protein pNew=p.cloneProtein(new StrapProtein());
        pNew.setResidueSubset(selRes);
        return pNew;
    }

}
// prosup www.came.sbg.ac.at
//PUBMED: 15609341 FAST: a novel protein structure alignment algorithm.
// dalilight FAST K2
//Aysam guerler@chemie.fu-berlin.de Prof Knapp

// http://bioinformatics.oxfordjournals.org/cgi/reprint/22/17/2166
//MATRAS: a program for protein 3D structure comparison
// qCOPS  TopMatch

// http://sourceforge.net/apps/mediawiki/lajolla/index.php?title=Main_Page http://lajolla.sf.net