package charite.christo.strap;
import charite.christo.*;
import charite.christo.blast.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import charite.christo.strap.interfaces.*;
import java.awt.*;
import java.awt.event.*;
import java.io.File;
import java.util.*;
import javax.swing.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
/**HELP

   BL2SEQ produces all local alignments of two given sequences  (UBIC:bl2seq). The
   high scoring alignments are displayed in a two-dimensional diagram.
   If bl2seq is
   not installed the Web server http://www.ncbi.nlm.nih.gov/blast/bl2seq/wblast2.cgi is used.

   @author Christoph Gille

*/
public class Dialog_bl2seq extends AbstractDialogJTabbedPane implements Runnable {
    public void run() {
        final StrapProtein pX=comboProtX.getProtein(),pY=comboProtY.getProtein();
        if (pX==null || pY==null) return;
        Bl2seq bl2seq=new Bl2seq(pX,pY);
        adTab("<sub>"+pX+"<br>"+pY+"</sub>",bl2seq, (JTabbedPane)this);
        new Thread(bl2seq).start();
    }

    @Override public void dispose() {
        comboProtX=comboProtY=null;
        disposeChildren(this);
    }
    private ProteinCombo comboProtX,comboProtY;
    public Dialog_bl2seq() {
        comboProtX=new ProteinCombo(align,0);
        comboProtY=new ProteinCombo(align,0);
        comboProtY.s(1);
        Container pMain=vBpnl(
                              pnl("Horizontal",comboProtX),
                              pnl("Vertical",comboProtY),
                              " ",
                              pnl(buttn(TOG_CACHE).cb()," ", new ChButton().t(ChButton.GO).r(this)),
                              " "
                              );
        adMainTab(newCNSEW(pMain,StrapAlign.headLine(this),null,null,null), this, null);
    }
    private static class Bl2seq extends DialogDotPlot implements Runnable {

        private ChExec myExec;
        public Bl2seq(StrapProtein pX, StrapProtein pY) {
            super("ft2",pX,pY);
            pcp(KEY_IMAGE_FILE_NAME,pX+"+"+pY+"-bl2seq.png",this);
        }

        public synchronized void run() {
            final byte[] sequenceX=getSeq('X'), sequenceY=getSeq('Y');

            final String cacheKey,cacheSection;
            ByteArray result=null;
            if (CacheResult.isEnabled()) {
                final StringBuffer sb=new StringBuffer(30);
                CacheResult.sectionForSequence(sequenceX,sb);
                cacheSection=sb.toString();
                sb.setLength(0);
                sb.append(sequenceX).append('_').append(sequenceY);
                cacheKey=sb.toString();
                result=CacheResult.getValue(getClass(), cacheSection, cacheKey,(ByteArray)null);
            } else cacheSection=cacheKey=null;

            log(ByteArray.CLEAR,"x: ",sequenceX,"\n\ny: ",sequenceY,"\n\n");
            if (sequenceX==null || sequenceY==null) return;
            final boolean ntX=nxt(NO_ACTGUN,sequenceX)<0;
            final boolean ntY=nxt(NO_ACTGUN,sequenceY)<0;
            final String prg=ntX && ntY ? "blastn" : !ntX && ntY ? "tblastn" :  ntX && !ntY ? "blastx" : "blastp";
            log((ntX?"x Nucl  ":"x Amino "),sequenceX,"\n\ny: ",(ntY?"y Nucl  ":"x Amino "),sequenceY,"\n\n","Programm: ",prg,"\n");
            final int[][] vHits=new int[sequenceY.length][];

            if (result!=null) log("Found result in cache: \n",result);
            if (result==null) {
                log("Trying locally installed bl2seq. \nIt can be installed from  ftp://ftp.ncbi.nlm.nih.gov/blast/executables/LATEST/ \n");
                myExec=new ChExec(ChExec.STDOUT | ChExec.STDERR | ChExec.IGNORE_ERROR | ChExec.LOG);
                myExec.setCommandLineV( "bl2seq","-p",prg,"-i",writeFasta(sequenceX,"x"),"-j",writeFasta(sequenceY,"y"));
                myExec.run();
                result=myExec.getStdoutAsByteArray();
                if (result==null) log("Local bl2seq failed\n\n");
            }
            if (result==null) {
                final String SERVER="http://www.ncbi.nlm.nih.gov/blast/bl2seq/wblast2.cgi?1";
                log("Trying ",SERVER,"\n");
                log("\nError: could not start the program bl2seq \ngoing to try server ",SERVER,"\n");
                final boolean isN=prg=="blastn";
                result=Web.getServerResponse(0L, SERVER,new String[][]{
                    {"program",prg},
                    {"matrix", isN ? "" : "BLOSUM62"},
                    {"gopen",isN ? "5" : "11"},
                    {"gext",isN ? "2" : "1"},
                    {"dropoff","50"},
                    {"expect","10"},
                    {"word",isN ? "10" : "3"},
                    {"Filer","0"},
                    {"seq",">sx\n"+toStrg(sequenceX)},
                    {"sseq",">sy\n"+toStrg(sequenceY)},
                    {"submit","align"}
                });

                if (result==null) {
                    log("\nNo response from server!\n");
                    return;
                } else log("From server:\n",result);
                CacheResult.putValue(getClass(),cacheSection, cacheKey,result,false);
            }
            if (result !=null) {
                final int ends[]=result.getLineEnds();
                final int end=result.getEnd();
                final byte txt[]=result.getBytes();
                int fromX=-1, fromY=-1,toX=-1,toY=-1, seqX=-1, seqY=-1;
                double score=0;

                for(int iL=0; iL<ends.length; iL++) {
                    final int b=iL==0?0: ends[iL-1]+1;
                    final int e=ends[iL];
                    if (e-b<12) continue;
                    final byte c0=txt[b+0], c1=txt[b+1], c2=txt[b+2], c3=txt[b+3], c4=txt[b+4], c5=txt[b+5];
                    if (c1=='S' && c2=='c' && c3=='o'  && c4=='r' && c5=='e') {
                        score=atof(txt,b+8);
                    }
                    final boolean
                        isX=c0=='Q' && c1=='u' && c2=='e' && c3=='r' && c4=='y',
                        isY=c0=='S' && c1=='b' && c2=='j' && c3=='c' && c4=='t';
                    if (isX || isY) {
                        int to=-1;
                        for(int i=b+12; i<e; i++) {
                            final char c=(char)txt[i];
                            if (to==-1  && '0'<=c && c<='9') to=atoi(txt,i);
                        }
                        final int from=atoi(txt,b+6);
                        int seq=b+6;
                        while(seq<e) {
                            final char c=(char)txt[seq];
                            if ('a'<=c && c<='z' || 'A'<=c && c<='Z') break;
                            seq++;
                        }
                        if (isX) { fromX=from-1; toX=to; seqX=seq;}
                        if (isY) { fromY=from-1; toY=to; seqY=seq;}
                    }
                    if (isY && toX>=0 && toY>=0 && fromX>=0 && toX>=0) {
                        super.addPoints(txt,txt,seqX,seqY,fromX>toX,fromY>toY, end,end, fromX,fromY,score,vHits);
                    }
                }
            }
            setBl2seq(vHits,myExec);
        }
        private File writeFasta(byte[] seq,String name) {
            final File fOut=file(dirTmp()+"/bl2seq/"+name+".fa");
            delFile(fOut);
            wrte(fOut,">"+name+"\n"+toStrg(seq)+"\n");
            return fOut;
        }
    }

}
