package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.strap.interfaces.*;
import charite.christo.protein.*;

import java.awt.*;
import java.util.*;
/**HELP
This operation creates residue selections at all beginnings of  continuous blocks of translated nucleotides (WIKI:Exon, WIKI:Intron).

   @author Christoph Gille
*/
public class DnaExonStart {
    public final static void mark(StrapProtein pp[]) {
        StrapAlign.errorMsg("");
        if (pp==null || pp.length==0) { StrapAlign.errorMsg("Select proteins"); return;}
        int nP=pp.length,count=0;
        for(int i=0;i<nP;i++) {
            StrapProtein p=pp[i];
            byte bb[]=p.getNucleotides();
            boolean isTr[]=p.areNucleotidesCurrentStrandTranslated();
            if (bb==null || isTr==null) continue;
            mark(p,isTr);
            count++;
        }
        if (count==0)  { StrapAlign.errorMsg("No exons and introns in proteins"); return;}
    }

    private final static void killResidueAnnotations(StrapProtein p) {
        for(ResidueAnnotation s: p.getResidueAnnotations()) {
            if ("start_of_exons".equals(s.getValue(ResidueAnnotation.GROUP)) &&
                s.getValue(ResidueAnnotation.NAME).startsWith("exon")) p.removeResidueSelection(s);
        }
    }
    private final static void mark(StrapProtein p, boolean isTr[]) {
        killResidueAnnotations(p);
        int n=isTr.length;
        boolean lastIsTr=false;
        UniqueList<ResidueAnnotation> v=new UniqueList(ResidueAnnotation.class);
        final boolean isReverse=p.isReverse();
        for(int i=0;i<n;i++) {
            if (!lastIsTr && isTr[i]) {
                ResidueAnnotation s=new ResidueAnnotation(p);
                s.setValue(ResidueAnnotation.NAME, "exon"+ (1+v.size()));
                int idx=isReverse ? n-i : i+1;
                s.setValue(ResidueAnnotation.POS,""+idx);
                s.setEnabled(false,ResidueAnnotation.POS);
                s.setValue(ResidueAnnotation.GROUP,"start_of_exons");
                s.setColor(Color.PINK);
                p.addResidueSelection(s);
                v.add(s);
            }
            lastIsTr=isTr[i];
        }
        StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_ADDED,111);
    }
}
