package charite.christo.strap;
import charite.christo.*;
import charite.christo.protein.*;
import charite.christo.interfaces.*;
import charite.christo.strap.interfaces.SelectorOfNucleotides;
import java.awt.*;
import javax.swing.*;
import java.awt.image.BufferedImage;
import java.util.*;
import java.util.List;
import static charite.christo.ChUtils.*;
import static java.awt.event.MouseEvent.*;
import static java.awt.event.KeyEvent.*;
import static charite.christo.ChConstants.*;
/**
   This is the upper part in the NT view showing the translated AA.
   @author Christoph Gille
*/
class EditDnaView extends ChPanel implements PaintHook ,StrapListener{
    private final static Rectangle RECT=new Rectangle(), RECT_CURSOR[]={new Rectangle(),new Rectangle()};
    private final static byte[] BUF=new byte[12];
				private final EditDnaParentPane parent;
    private JComponent panel;
    private Rectangle cb,cbSmall;
    private transient Object refImage;
    private final StrapProtein protein;
    private boolean  isSelecting, isDragged;
    private final ChScrollBar hSB=new ChScrollBar(JScrollBar.HORIZONTAL); {addPaintHook(this,hSB); }
    private int fontSize=18, mouseAA=-1, ySelOfNucl;
    private long mayBeUpdateImage_modiCount=-1;

    private int caret, selectionStart=-1, selectionEnd=-1, strapCursor;
    public StrapProtein  getProtein() { return protein;}
    private Rectangle rectCursorStrap(JComponent component, boolean neu) {
        final Rectangle R=RECT_CURSOR[component==hSB?0:1];
        final StrapProtein p=getProtein();
        final int
            iCursor=StrapAlign.getInstance().getIndexOfAminoAcidAtCursor(protein),
            nNt=p!=null ? p.countNucleotides() : 0,
            iAa=neu ? iCursor :  strapCursor;
        if (nNt==0) return null;
        if (component==hSB) {
            final Rectangle track=hSB.getTrack();
            final int
                h=track.height,
                iNt=p.aminoAcidIndex2nucleotideIndex(iAa);
            if (iNt<0) return null;
            R.y=h/2;
            R.x= (int) (((long)track.width)*iNt/nNt + track.x-h/2);
            R.width=R.height=h;
        }
        if (component==this) {
            int tNt=3*iAa;
            tNt+=1-tNt%3;
            final int iNt=p.translatedNucleotideIndex2nucleotideIndex(tNt);
            R.y=0;
            R.x=(iNt-1)*cb.width;
            R.width=3*cb.width;
            R.height=cb.height;
        }
        return R;
    }

    @Override public void processEvent(AWTEvent ev) {
        if (cb==null) changeFontSize(true);
        final int id=ev.getID();
        final StrapProtein p=getProtein();
        final int nNt=p!=null ? p.countNucleotides() :0;
        if (nNt==0) return;
        final StrapAlign align=StrapAlign.getInstance();
        final byte nucleotides[]=p.getNucleotidesCurrentStrand();
        final int charW=cb.width, charH=cb.height, h=getHeight(),w=getWidth();
        final boolean shift=isShift(ev),  ctrl=isCtrl(ev);
        final int lastSelection=selectionEnd;
        final int oldCaret=this.caret;
        final int kcode=keyCode(ev);
        final int scaleUpDown=scaleUpDown(ev);
        if (scaleUpDown!=0) changeFontSize(scaleUpDown<0);
         {
            final int x=x(ev),y=y(ev);
            final int rotation=wheelRotation(ev);
            if (rotation!=0 && !ctrl) {
                RECT.x=hSB.getValue()+ (rotation<0 ? -fontSize : fontSize);
                RECT.y=RECT.height=1;
                RECT.width=hSB.getVisibleAmount();
                scrollRectToVisible(RECT);
            }
            if (id==MOUSE_PRESSED) {
                requestFocus();
                caret=x/cb.width;
                selectionStart=selectionEnd=-1;
            }
            if (id==MOUSE_DRAGGED) {
                if (isDragged) selectionEnd=x/charW;
                else  { isDragged=true; selectionStart=selectionEnd=x/charW; }
            }
            if (id==MOUSE_MOVED) isDragged=false;

            if (id==MOUSE_CLICKED) {
                if (isPopupTrigger(ev)) StrapAlign.showContextMenu('S',getResidueSelections(ev));
                else if (mouseAA>=0 && y<=2*charH) {
                    final StrapView v=align.getAlignmentPanelWithFocus();
                    if (v!=null) {v.setCursorToResidue(p,mouseAA-p.getFirstResidueIndex()); v.scrollCursorToVisible();}
                }
            }
            if (id==MOUSE_MOVED) {
                if (y>0-3 && y<charH+3)  {
                    final int
                        nt=x/charW,
                        tNt=p.nucleotideIdx2translatedNucleotideIdx(nt),
                        iAa=mouseAA=(tNt-tNt%3+1)/3;
                    if (iAa>=0) Protein.mouseOver(p,iAa);
                    setCursor(getCursr(iAa>=0 ? 'H':'D'));
                }
            }
            if (id==MOUSE_EXITED) {
                int start=maxi(0,mini(selectionStart,selectionEnd));
                int end=mini(nNt-1, maxi(selectionEnd,selectionStart));
                if (start!=end) toClipbd(new String(nucleotides,start,end-start+1));
            }
        }
        if (id==KEY_PRESSED) {
            final byte[] selNuc=p.getSelectedNucleotides();
            final byte[] selAmi=p.getSelectedAminoacids();
            int go=0;
            if (isAlt(ev)) {
                if (kcode==VK_RIGHT || kcode==VK_LEFT) {
                    final boolean areTranslated[]=p.areNucleotidesCurrentStrandTranslated();
                    int pos=caret;
                    final boolean translated=p.isNucleotideCurrentStrandTranslated(pos);
                    final int iA0=p.nucleotideIdx2translatedNucleotideIdx(pos)/3;
                    final boolean isSeled=pos<selNuc.length && selNuc[pos]!=0 || iA0>=0 && iA0<selAmi.length && selAmi[iA0]>0;
                    while(pos>=0 && pos<nNt && p.isNucleotideCurrentStrandTranslated(pos)==translated) {
                        final int iA=p.nucleotideIdx2translatedNucleotideIdx(pos)/3;
                        if ( (pos<selNuc.length && selNuc[pos]!=0  || iA>=0 && iA<selAmi.length && selAmi[iA]>0)!=isSeled) break;
                        pos+= ((kcode==VK_LEFT) ? -1 : 1);
                    }
                    go=maxi(0,mini(nNt-1,pos))-caret;
                }
            } else {
                switch(kcode) {
                case VK_A: if(ctrl) { selectionStart=0; selectionEnd=MAX_INT; } ;break;
                case VK_LEFT: go= ctrl ? -MAX_INT/2 : -1; break;
                case VK_RIGHT: go=ctrl ? MAX_INT/2 :  1; break;
                case VK_PAGE_DOWN:  go=100; break;
                case VK_PAGE_UP:  go=-100; break;
                case VK_E:  go=MAX_INT/2; break;
                case VK_DOWN:  go=10; break;
                case VK_UP: go=-10; break;
                }
            }
            if (go!=0) {
                if (shift && !isSelecting)  selectionStart=caret;
                this.caret=maxi(0,mini( nucleotides.length-1,caret+go));
                if (shift) {selectionEnd=caret;
                }
                isSelecting=shift;
            }
            else if (!shift && (kcode==VK_INSERT||kcode==VK_SPACE||kcode==VK_DELETE||kcode==VK_BACK_SPACE)) {
                final int  offset;
                final boolean translate;
                switch(kcode) {
                case VK_INSERT:     offset= 0; go= 1; translate=true; break;
                case VK_SPACE:      offset=go=-1; translate=true; break;
                case VK_DELETE:     offset= 0; go= 1; translate=false; break;
                default: offset=go=-1; translate=false; break;
                }
                final boolean noRegion=selectionEnd==selectionStart;
                final int end=noRegion ? caret : maxi(this.selectionStart,this.selectionEnd);
                final int start=noRegion ? caret :  mini(this.selectionStart,this.selectionEnd);
                boolean isChanged=false;
                for(int idx=start;  idx<=end; idx++) {
                    final int changeIdx=start==end ? idx+offset : idx;
                    if (changeIdx<0  || changeIdx>nNt) continue;
                    if (p.isNucleotideCurrentStrandTranslated(changeIdx) != translate) { isChanged=true;  p.setNucleotideCurrentStrandTranslated(changeIdx,translate);}
                    if (noRegion) caret+=go;
                }
                if (isChanged) {
                    StrapAlign.dispatchLater(StrapEvent.NUCL_TRANSLATION_CHANGED,111);
                    repaint(charW*start,0,99999,charW*(end-start));
                }
            }
        }
        if (selectionEnd>nNt) selectionEnd=nNt;
        if (lastSelection!=selectionEnd) repaint(0,h-charH,w,99);

        if (oldCaret!=caret) {
            repaint(oldCaret*charW,h-charH, charW,charH);
            repaint(caret*charW,h-charH, charW,charH);
            RECT.y=RECT.height=1;
            RECT.width=charW*4;
            RECT.x=maxi(0,mini(w-1,(caret-1)*charW));
            scrollRectToVisible(RECT);
        }
        if (id==MOUSE_MOVED || kcode!=0) parent.setFocused(this,true);
        if (kcode==0 && scaleUpDown==0) super.processEvent(ev);
    }
    public EditDnaView(StrapProtein p,EditDnaParentPane pan) {
        protein=p;
        parent=pan;
        setToolTipText("");
        generateMiddleForLeftButton(hSB);
        this.enableEvents(MOUSE_EVENT_MASK|MOUSE_WHEEL_EVENT_MASK|FOCUS_EVENT_MASK|FOCUS_EVENT_MASK|KEY_EVENT_MASK);
        StrapAlign.getInstance().addListener(this);
        setAutoscr(this);
        addPaintHook(this,this);
    }
    public boolean paintHook(JComponent component,Graphics g_, boolean after) {
        if (!after) return true;
        if (component==null) return false;
        final int W=component.getWidth(), H=component.getHeight();
        final Graphics2D g=(Graphics2D)g_;
        final int charW=cb.width, charW2=charW/2,  charH=cb.height,  charWs=cbSmall.width, charHs=cbSmall.height;
        final int clipX,clipY, clipX2, h=getHeight();
        {
            final Rectangle c=g.getClipBounds();
            if (c!=null) {
                clipX=c.x;
                clipY=c.y;
                clipX2=clipX+c.width;
            } else {
                clipX=clipY=0;
                clipX2=MAX_INT/2;
            }
        }
        final Font
            font=getFnt(fontSize,true,0),
            fontSmall=getFnt(this.fontSize/2+1,true,0);

        final StrapProtein p=this.protein;
        final int nNt=p!=null ? p.countNucleotides() : 0;
        if (nNt==0) return true;
        final boolean isTranslated[]=p.areNucleotidesCurrentStrandTranslated();
        final boolean isReverse=p.isReverse();
        final ResidueSelection[] selections=p.getAllResidueSelections();
        final byte[] selectedNucleotides=p.getSelectedNucleotides();
        AbstractResidueSelection selCursor=null;
        for(ResidueSelection s : selections) {
            if (s instanceof AbstractResidueSelection) {
                AbstractResidueSelection as=(AbstractResidueSelection)s;
                if (as.getName()=="cursor") selCursor=as;
            }
        }
        final Color colorFg=C(isWhiteBG()?0:0xFFffFF), colorBg=C(isWhiteBG()?0xFFffFF:0);
        if (component==hSB) {
            if (selCursor!=null) selCursor.setStyle(VisibleIn123.IMAGE_NO_FILL);
            final Rectangle track=hSB.getTrack();
            final int width=track.width, height=track.height;
            g.setColor(colorBg);
            fillBigRect(g, 0,0,W,H);
            if (width<0) return true;
            BufferedImage image=(BufferedImage)deref(refImage);
            {
                final int rgbExon=isWhiteBG() ? 0xFFffFFff : 0xFF000000;
                if (image!=null && image.getWidth()<width) image=null;
                final long mc=
                  p.getModificationCount(ProteinModiCount.MC_SELECTED_NT)+
                  p.getModificationCount(ProteinModiCount.MC_ARE_NUCLEOTIDES_TRANSLATED)+ (((long)width)<<32);
                if (image==null || mayBeUpdateImage_modiCount!=mc) {
                    mayBeUpdateImage_modiCount=mc;
                    if (image==null) refImage=newSoftRef(image=new BufferedImage(width,1,BufferedImage.TYPE_INT_ARGB));
                    final int rgbs[]=sharedRGB(width);
                    if (isTranslated!=null) {
                        for(int i=width; --i>=0;) {
                            final int iNt=i*nNt/width;
                            rgbs[i]= (iNt>=isTranslated.length || isTranslated[isReverse ? nNt-iNt-1 : iNt]  ? rgbExon : 0xffc8c8c8);
                        }
                    } else Arrays.fill(rgbs,0xffc8c8c8);
                    for(ResidueSelection s:selections) {
                        if (s==this) continue;
                        Color col=C(0xFF00);
                        if (s instanceof VisibleIn123) {
                            VisibleIn123 v123=(VisibleIn123)s;
                            if ((v123.getVisibleWhere()&VisibleIn123.SB)==0) continue;
                            col=v123.getColor();
                        }
                        final int rgb=col==null ? 0 : col.getRGB();
                        final SelectorOfNucleotides selNt=deref(s, SelectorOfNucleotides.class);
                        final boolean nt[]= selNt!=null ? selNt.getSelectedNucleotides() : null;
                        if (nt!=null && nt.length>0) {
                            final int ntOffset=selNt.getSelectedNucleotidesOffset();
                            for(int iNt=mini(nt.length+ntOffset,nNt);--iNt>=ntOffset;) {
                                if (nt[iNt-ntOffset]) {
                                    final long iNtR=isReverse ? nNt-iNt-1 : iNt;
                                    final int idx1= (int) (iNtR*width/nNt);
                                    final int idx2= mini(width, (int) ((iNtR+1)*width/nNt)+1);
                                    for(int i=idx1; i<idx2; i++) if (i>0) rgbs[i]=rgb;
                                }
                            }
                        } else {
                            final boolean selAA[]=s.getSelectedAminoacids();
                            if (selAA!=null) {
                                final int selOffset=s.getSelectedAminoacidsOffset();
                                final int iMax=mini(selAA.length+selOffset, p.countResidues());
                                for(int iAa=maxi(0,selOffset); iAa<iMax; iAa++) {
                                    if (selAA[iAa-selOffset]) {
                                        for(int i3=0, iAa3=3*iAa; i3<3; i3++) {
                                            final int iNt=p.translatedNucleotideIndex2nucleotideIndex(iAa3+i3);
                                            if (iNt<0) continue;
                                            final int iRgb=isReverse ? nNt-iNt-1 : iNt;
                                            final int idx1= (int) ((long)(iRgb)*width/nNt);
                                            final int idx2= mini(width,(int) ((long)(iRgb+1)*width/nNt)+1);
                                            for(int i=idx1; i<idx2; i++) rgbs[i]=rgb;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    image.setRGB(0,0,width,1,rgbs,0,width);
                }
            }
            {
                final int h2=height>>1,  hSel=maxi(height/3,2);
                int xDest1=track.x,xDest2=track.x+width;
                if (isReverse)  { xDest2=track.x; xDest1=track.x+width;}
                if (image!=null) g.drawImage(image,xDest1,h2-hSel,xDest2,h2+hSel,  0,0,width,1,hSB);
            }
            {
                g.setColor(C(0x8080ff,100));
                final Rectangle t=hSB.getThumb();
                fillBigRect(g, t.x,t.y,t.width,t.height);
                g.drawRect(t.x,t.y,t.width,t.height);
                final Rectangle r=rectCursorStrap(hSB,true);
                if (r!=null) ChIcon.drawCursor(g, r.x, r.y, r.width, r.height,this);
            }
        } else if (component==this) {
            parent.setFocused(this,false);
            g.setColor(colorBg);
            fillBigRect(g, 0,0,999999,999);
            final byte nucleotides[]=p.getNucleotidesCurrentStrand();
            if (nucleotides==null) return true;
            final int
                ntTo=mini( nNt-1, clipX2/charW, nucleotides.length-1),
                ii[]=p.translatedNucleotideIndices2nucleotideIndices(),
                kk[]=p.nucleotideIndices2translatedNucleotideIndices();
            g.setColor(colorBg);
            final int ntFrom, tNtFrom;
            {
                final int nt=clipX/charW;
                if (nt>=kk.length) return true;
                final int tNt=kk[nt];
                ntFrom=nt-nt%10;
                tNtFrom=tNt-tNt%3;
            }
            if (clipY<charHs+2*charH) { /* --- aminos --- */
                final int y0=charHs, x0=0;
                g.setColor(colorFg);
                for(int i=tNtFrom-3;i<=ntTo;i+=3) {
                    if (i<0) continue;
                    final int i3=i/3;
                    if (ii.length<=i+2 || nNt<=i+2) break;
                    final int i0=ii[i],i1=ii[i+1],i2=ii[i+2];
                    if (nucleotides.length<=i2 || i0<0 || i1<0 || i2<0 ) break;
                    g.setColor(C(0x808080));
                    final int x=x0+i1*charW;
                    for(int j=3;--j>=0;) g.drawLine(x+charW2,y0+charH,x0+charW2+ii[j+i]*charW,h-charH);
                    if (x+charW<clipX || x>clipX2) continue;
                    BUF[0]=charite.christo.protein.DNA_Util.triplet2aa(nucleotides[i0],nucleotides[i1],nucleotides[i2]);
                    g.setFont(font);
                    RECT.width=charW*2; RECT.height=charH*2;
                    RECT.x=x-1;
                    RECT.y=0;
                    ResSelUtils.drawResidueSelection(StrapProtein.AMINO_ACIDS,BUF[0],p,selections,i3,g,RECT,cb,this);
                    g.setColor(colorFg);
                    g.drawBytes(BUF,0,1,x,cb.y);
                    g.setFont(fontSmall);
                    if (i3%5==0) {
                        itoa(i3+1,BUF);
                        int len=0;
                        while(len<BUF.length && BUF[len]!=0) len++;
                        g.drawBytes(BUF,0,len,x+charW,cb.y);
                    }
                }
            }
            if (clipY<h-charH) { /* --- ruler for all nucleotides --- */
                final int y0=2*charHs;
                ySelOfNucl=h-charHs;
                g.setFont(fontSmall);
                RECT.height=charHs; RECT.width=charWs;
                for(int i=ntFrom-4;i<=ntTo+4;i++) {
                    if (i<0 || i>ntTo) continue;
                    final int x=i*charW+charW2;
                    if (i%10==0) {
                        g.setColor(colorFg);
                        itoa(i+1,BUF);
                        int len=0;
                        while(len<BUF.length && BUF[len]!=0) len++;
                        g.drawBytes(BUF,0,len,x-len*charWs/2,cbSmall.y+y0);
                        itoa(nNt-i,BUF);
                        len=0;
                        while(len<BUF.length && BUF[len]!=0) len++;
                        g.drawBytes(BUF,0,len,  x-len*charWs/2,1+y0+cbSmall.y+charHs-1);
                        g.drawLine(x, h-charH-charHs, x,h-charH-1);
                    }
                }
            }
            final Font bold=getFnt(fontSize,true,Font.BOLD);
            for(int i=maxi(0,ntFrom);i<=ntTo; i++) {
                BUF[0]=nucleotides[i];
                final int x=i*charW+charW2;
                final int y=h-charH;
                boolean isBG=false;
                if (selectionStart!=selectionEnd && (selectionStart<=i && i<=selectionEnd || selectionEnd<=i && i<=selectionStart)) {
                    g.setColor(C(0xFF));
                    fillBigRect(g, x-charW2, y,charW, charH);
                    isBG=true;
                }
                if(i==caret) {
                    g.setColor(C(0xFFffFF));
                    fillBigRect(g, x-charW2, y+2,charW, charH);
                    isBG=true;
                }
                if (selectedNucleotides!=null && selectedNucleotides.length>i && selectedNucleotides[i]!=0) {
                    g.setColor(C(0xFFAFAF));
                    RECT.x=x-charW2;
                    RECT.y=y-2;
                    RECT.width=charW+1;
                    RECT.height=charH+6;
                    ResSelUtils.drawResidueSelection(StrapProtein.NUCLEOTIDES,(byte)0,p,selections,i,g,RECT,cb,this);
                    isBG=true;
                }
                if (isBG) {
                    g.setColor(colorBg);
                    g.setFont(bold);
                    g.drawBytes(BUF,0,1,x-charW2-1,y+cb.y);
                    g.drawBytes(BUF,0,1,x-charW2+1,y+cb.y+2);
                }
                g.setColor(isTranslated==null || i<isTranslated.length && isTranslated[i] ? colorFg : C(0x808080));
                g.setFont(font);
                g.drawBytes(BUF,0,1,x-charW2,y+cb.y+1);
            }
            if (selCursor!=null) selCursor.setStyle(VisibleIn123.FULL_BACKGROUND);
        } else {
            final Font f=getFnt(14,true,Font.BOLD);
            g.setFont(f);
            g.setColor(colorFg);
            g.drawString(toStrg(p.getRendTxt()),EM, charBounds(f).y+1);
        }
        return true;
    }
    @Override public String getToolTipText(java.awt.event.MouseEvent ev) {
        final ResidueSelection[] v=getResidueSelections(ev);
        if (v.length==0) return null;
        final StringBuffer sb=strBuf(SB_TIP);
        int i=0;
        for(ResidueSelection s:v) {
            if (s==null || s==Protein.getMouseOverSelection()) continue;
            final String tip=TabItemTipIcon.tip(s);
            if (i++>0) sb.append("<br>");
            sb.append(tip!=null ? tip : toStrg(s));
        }
        return addHtmlTagsAsStrg(sb);
    }
    private final static List<ResidueSelection> vRS=new ArrayList();
    private ResidueSelection[] getResidueSelections(AWTEvent ev){
        if (cb==null) paintComponent(null);
        vRS.clear();
        final StrapProtein protein=getProtein();
        if (protein==null) return new ResidueSelection[0];
        final int
            x=x(ev),
            y=y(ev),
            iN=(protein.isReverse() ? getWidth()-x : x)/cb.width;
        final Object[] oo;
        if (y>ySelOfNucl) oo=ResSelUtils.selOfNuclAt(protein,iN);
        else if (y<cb.height) {
            final int iA=mouseAA-protein.getFirstResidueIndex();
            oo=protein.getResidueSelectionsAt(0L,iA,iA+1,VisibleIn123.SEQUENCE);
        } else oo=null;
        if (oo!=null)
            for(Object s: oo)
                if (s instanceof ResidueSelection)
                    adUniq((ResidueSelection)s,vRS);
        vRS.remove(Protein.getMouseOverSelection());
        return toArry(vRS,ResidueSelection.class);
    }

    public JComponent getPanel() {
        if (panel==null) {
            final Color bg=C(isWhiteBG()?0xFFffFF:0);
            final JComponent pName=hB(bg);
            addPaintHook(this,pName);
            pName.setPreferredSize(dim(100,24));
            final JScrollPane sp=new JScrollPane(this,JScrollPane.VERTICAL_SCROLLBAR_NEVER,JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
            panel=newCNSEW(sp,null,"#JS",bg,null);
            final JComponent pNorth=newCNSEW(pName,bg,null,new ChButton().actionClose(panel),null);
            panel.add(pNorth,BorderLayout.NORTH);
            changeFontSize(false);
            addMoliRefreshTip(hSB);
            sp.setHorizontalScrollBar(hSB);
            noInsets(sp);

            pNorth.setPreferredSize(dim(1,24));
        }
        return panel;
    }
    private void changeFontSize(boolean enlarge) {
        fontSize=mini(33,maxi(3,fontSize + (enlarge ? -1 : 1)));
        hSB.setUnitIncrement(3*fontSize);
        hSB.setBlockIncrement(30*fontSize);
        cbSmall=charBounds(getFnt(this.fontSize/2+1,true,0));
        cb=charBounds(getFnt(fontSize,true,Font.BOLD));
        final int
            w=cb.width*getProtein().countNucleotides(),
            hSmall=cbSmall.height,
            h=hSmall+2*cb.height+hSmall+hSmall+hSmall/4;
        setPreferredSize(dim(w, h));
        panel.setMaximumSize(dim(w, h+EX));
        revalidate();
        parent.revalidate();
        panel.repaint();
    }
    private void paintCursor() {
        for(int i=0;i<2;i++) {
            final Rectangle rSB=rectCursorStrap(hSB, i==1),  r=rectCursorStrap(this, i==1);
            if (rSB!=null) hSB.paintImmediately(rSB);
            if (r!=null) paintImmediately(r);
        }
        this.strapCursor=StrapAlign.getInstance().getIndexOfAminoAcidAtCursor(protein);
    }

    public final void handleEvent(StrapEvent ev) {
        final int t=ev.getType();
        if (t==StrapEvent.CURSOR_CHANGED_PROTEIN || t==StrapEvent.CURSOR_MOVED_WITHIN_PROTEIN)  paintCursor();
        else if (t==StrapEvent.PROTEIN_RENAMED || t==StrapEvent.AA_SHADING_CHANGED || t==StrapEvent.RESIDUE_SELECTION_CHANGED ||
                 t==StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR || t==StrapEvent.RESIDUE_SELECTION_ADDED || t==StrapEvent.CHARACTER_SEQUENCE_CHANGED)
            getPanel().repaint();
        else if ( (t&StrapEvent.FLAG_RESIDUE_TYPES_CHANGED)!=0) {
            final StrapProtein p=getProtein();
            if (p==null || p.countNucleotides()==0) rmFromParent(getPanel());
            ChDelay.repaint(getPanel(),333);

        }
    }

}
