package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static charite.christo.strap.SequenceFeatures.*;
public class ExpasyGff  {

    public static void load(long options, Protein pp[], ChRunnable log) {
        boolean hasFeatures=false;
        List<StrapProtein> vNoGFF=null;
        final Set<String> vID_already=new HashSet();
        ByteArray baGFF=null;
        try {
            final Set<String> vID=new HashSet();
            for(int step : new int[]{0, F_NO_SEC_STRU, F_SEC_STRU}) {
                nextProtein:
                for(int iP=0; iP<pp.length; iP++) {
                    final StrapProtein p=(StrapProtein)pp[iP];
                    if (p==null) continue;
                    synchronized(mkIdObjct("EGFF$$ks",p)) {
                        if (p.containsOnlyACTGNX()) continue;
                        if (log!=null) log.run(ChRunnable.PROGRESS, progressReport(ExpasyGff.class,iP+1,pp.length+1,"Expasy-GFF: "+p+" ..."));
                        final int countAnnotationsOld=p.getResidueAnnotations().length;
                        int countFeatures=0;
                        final String[] ids=(options&OPT_ID_BY_HOMOLOGY)!=0 ? FindUniprotID.ids(false,p).clone() : new String[]{p.getUniprotID(),p.getAccessionID()};
                        if (countNotNull(ids)==0) continue;

                        for(int j=ids.length; --j>=0;) {
                            if (!equls("UNIPROT:",ids[j],0) && !equls("SWISS:",ids[j],0) &&
                                !(equls("TREMBL:",ids[j],0) && !equls("PDB:",ids[j],0))) ids[j]=null;
                        }
                        vID_already.clear();
                        boolean foundGff=false;

                        for(String id0 :  ids) {
                            if (id0==null || !id0.startsWith("UNIPROT:") && !id0.startsWith("SWISS:") && !id0.startsWith("TREMBL:") && !id0.startsWith("PDB:")) continue;
                            String id=id0.startsWith("PDB:") ? pdbID(id0) : id0.substring(id0.indexOf(':')+1);
                            if (!vID_already.contains(id)) vID_already.add(id); else continue;
                            final String gff_expasy_colon_Id="GFF_EXPASY:"+id;
                            final java.net.URL url=url(Hyperrefs.toUrlString(gff_expasy_colon_Id,Hyperrefs.PLAIN_TEXT));
                            final java.io.File f=download_TS(url(url),  CacheResult.isEnabled() ? MAX_DAYS:0);
                            if (f!=null && sze(f)==0 &&  sze(file(f+TMP_DOWNLOAD_SFX))==0) wrte(f,"0");
                            if (step==0 && log!=null) log.run(ChRunnable.APPEND, p+" "+gff_expasy_colon_Id+" "+(f!=null ? sze(f) : -1)+"bytes\n");
                            baGFF=readBytes(f,baGFF);
                            if (baGFF!=null) {
                                hasFeatures=foundGff=true;
                                for(ResidueAnnotation aNew : parseFeatures(step, baGFF,  url, p, "", step==0 ? vID : null)) {
                                    aNew.addEnS(CLASS_Hyperrefs,gff_expasy_colon_Id);
                                    if (aNew.isEnabled()) countFeatures++;
                                }
                                break;
                            }
                        }

                        if (step==0) continue nextProtein;
                        final ResidueAnnotation[] aa=p.getResidueAnnotations();
                        if (step>0 && countAnnotationsOld!=aa.length)  {
                            SequenceFeatures.keepFeaturesWithBestScore(p);
                            SequenceFeatures.disableSecondaryStructure(aa);
                        }
                        if (countFeatures>0) {
                            Protein.incrementModificationCount(ProteinModiCount.MCA_SEQUENCE_FEATURES_V,null);
                            StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_ADDED, 999);
                        }
                    }
                }
                if (step==0) {
                    final String msg="GFF SOAP uniprot "+sze(vID)+" files ... ";
                    StrapAlign.drawMessage("@1"+msg);
                    FetchEBI_SOAP.download(FetchEBI_SOAP.DB_UNIPROTKB,toStrgArryO(vID),"");
                    StrapAlign.drawMessage("@1"+ANSI_GREEN+msg);
                }
            }
        } finally {
            if (log!=null) log.run(ChRunnable.PROGRESS, progressReport(ExpasyGff.class,1,1,"Expasy-GFF: "+pp.length+" proteins done"));
        }
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Also for StrapDAS >>> */

    static ResidueAnnotation[] parseFeatures(int mode, ByteArray ba, java.net.URL url, StrapProtein p, String sequenceDatabase,  Set vOnlyCollectIDs) {
        final ByteArray log=SequenceFeatures.log();
        if (0==sze(ba) || p==null) return ResidueAnnotation.NONE;
        final boolean isResnum="PDBresnum".equals(sequenceDatabase);
        final String xmlSegment, sequenceDB= isResnum ? "PDB" : sequenceDatabase;
        final byte[] T=ba.getBytes();
        final int E=ba.getEnd(), TABS[]=new int[9], ends[]=ba.getLineEnds();
        final boolean isXML=strstr("</FEATURE>",T,0,E)>0;
        if (isXML) {
            final int segment=strstr(STRSTR_w|STRSTR_AFTER,"<SEGMENT",T,0,E);
            xmlSegment=xmlAttribute("id",T,segment, E);
        } else xmlSegment=null;
        List<ResidueAnnotation> vRA=null;
        final ByteArray baTmp=new ByteArray(99);
        int b=0;
        nextFeature:
        for(int iL=0; ;iL++) {
            final int e;
            if (isXML) {
                b=strstr(STRSTR_w|STRSTR_AFTER,"<FEATURE",T,b,E);
                e=strstr("</FEATURE>",T,b,E);
            } else {
                if (iL>=ends.length) break;
                b=iL==0 ? 0 : ends[iL-1]+1;
                e=get(iL,ends);
                if (get(b,T)=='#') continue;
            }
            if (e<0 || b<0) break;
            final CharSequence line=isXML ? new ByteArray(T,b,e) : ba.newString(b,e);
            if (p.containsInvalidFeatureLine(line)) continue;
            for(ResidueSelection a:p.getAllResidueSelections()) {
                final CharSequence srcTxt= a instanceof ResidueAnnotation ? ((ResidueAnnotation)a).getSrcText() : null;
                if (sze(srcTxt)==sze(line) && equls(srcTxt, line,0))  {  continue nextFeature; }
            }
            final String featureName, seqname, source, attribute, viewSrcText;
            final int origStart, origEnd;
            String disabledWhy=null;
            if (isXML) {
                final int featureGT=strchr('>',T,b,e);
                final String sLabel=xmlAttribute("label",T,b,featureGT);
                final int type=strstr(STRSTR_w|STRSTR_AFTER,"<TYPE",T,b,e), typeGT=strchr('>',T,type,e);
                final String sType=htmlDecodAsStrg( T[typeGT-1]!='/' ? ba.toString(typeGT+1,  strstr("</TYPE>",T,type,e), ByteArray.STRING_POOL) : xmlAttribute("id",T, type,typeGT));
                final String category=xmlAttribute("category",T, type,typeGT);
                if (category!=null && category.indexOf(" similarity")>0) disabledWhy="Occures 'similarity' in category";
                final boolean[] found={false};
                final String sTypeTrans=SequenceFeatures.translateFeatureName(sType,found);
                final String
                    sTypeLC=lCase(sType),
                    sTypeLC1=delSfx("_one",delSfx("_two",delSfx("_three",delSfx("_four",delSfx("_five",delSfx("_six",delSfx("_seven",delSfx("_eight",sTypeLC))))))));
                if (found[0] || setSequenceOntologyTermsLC().contains(sTypeLC) || setSequenceOntologyTermsLC().contains(sTypeLC1)) {
                    featureName=sTypeTrans;
                    attribute=htmlDecodAsStrg(sLabel);
                } else {
                    featureName=htmlDecodAsStrg(sType);//translateFeatureName(htmlDecodAsStrg(sLabel), (boolean[])null );
                    attribute=htmlDecodAsStrg(capitalize(lCase(sType)));
                    //putln("\n",ba+"\n sType="+sType+"\n sLabel="+sLabel);
                }

                origStart=atoi(T,strstr(STRSTR_AFTER,"<START>",T,b,e))-1;
                origEnd=atoi(T,strstr(STRSTR_AFTER,"<END>",T,b,e));
                source=uCase(sequenceDB);
                seqname=xmlSegment;
                if (SequenceFeatures.isDeactivated(sType)) disabledWhy="isDeactivated(sType)";
                viewSrcText= url!=null ? file(url)+ChTextComponents.RANGE_SUFX_XML+(b-8)+"-"+(e+9) : null;
            } else { /* P25787	UniProtKB	Modified residue	24	24	.	.	.	Note=Phosphotyrosine	*/
                viewSrcText=null;
                if (tabulatrs('\t', T, b, e, TABS)<8) continue;
                seqname=ba.newString(b,TABS[0]);
                source=uCase(ba.newString(1+TABS[0],TABS[1]));

                if (vOnlyCollectIDs!=null) { if (source.startsWith("UNIPROT")) vOnlyCollectIDs.add(seqname); continue; }
                final String feature0=ba.newString(1+TABS[1], TABS[2]).replace(' ','_').intern();
                if (cntainsOnly(DIGT, T, 1+TABS[2],TABS[3])) origStart=atoi(T,1+TABS[2])-1; else continue;
                if (cntainsOnly(DIGT, T, 1+TABS[3],TABS[4])) origEnd=atoi(T,1+TABS[3]); else continue;
                attribute=ba.newString(1+TABS[7],TABS[8]>0 ? TABS[8]:e);
                if ( (feature0=="Modified_residue" || feature0=="Topological_domain" || feature0=="Region") && attribute.startsWith("Note=")) {
                    final int semicolon=attribute.indexOf(';');
                    featureName=attribute.substring(5,semicolon>0 ? semicolon : sze(attribute));
                } else featureName=feature0;
            }
            if (featureName==null) continue;
            final String fName=capitalize(lCase(featureName)).intern();
            if ("Chain"==fName) continue;
            final boolean isSecStru= fName=="Turn" || fName=="Helix" || fName=="Beta_strand";
            if ( (mode==F_SEC_STRU || mode==F_NO_SEC_STRU) &&  isSecStru != (F_SEC_STRU==mode)) continue;
            final ResidueAnnotation a=new ResidueAnnotation(p);
            if (isResnum) {
                final String sChain=is(LETTR_DIGT,p.getChainsAsString(),0) ? p.getChainsAsString().substring(0,1) : "";
                a.addEnS(ResidueAnnotation.POS, (origStart+1)+":"+sChain+"-"+origEnd+":"+sChain);
                disabledWhy=null;
            } else {
                final ReferenceSequence uniprot=p.getReferenceSequence(source+":"+seqname, null,false,(Runnable)null);
                if (uniprot==null) continue;
                final float score=uniprot.getAlignmentScore();
                final int start=get(origStart,uniprot.getUniprotIdx2Idx());
                final int end=  get(origEnd-1, uniprot.getUniprotIdx2Idx())+1;
                if (start<0 || end<0) {
                    final String msg="Protein "+p+" "+start+".."+end+": failed to map positions   \n\n";
                    if (log!=null) log.append(msg);
                    continue;
                }
                final boolean poorAlignment;
                if (uniprot!=null && (score<50 ||  uniprot.countMismatch(start,end,null)/(end-start+4f)>.3f)) {
                    final String s="Protein "+p+" poor alignment "+uniprot.getLinkForAlignment()+ALI_SCORE+score+"\n\n";
                    if (log!=null && strstr(s,log.getBytes())<0) log.run(ChRunnable.APPEND,s);
                    poorAlignment=true;
                } else poorAlignment=false;
                final int idx0=p.getFirstResidueIndex();
                final String rule=featureName2color(fName).rule;
                final String interval=rule=="-" ? "-" : ",";
                baTmp.setEnd(0).append(idx0+start+1,4);
                if (start+1<end) baTmp.append(interval).append(idx0+end);
                final String positions=baTmp.newString();
                a.addEnS(ResidueAnnotation.POS, positions);
                a.setRendererText(fName+" "+positions+" "+ALI_SCORE+score);
                a.addEnS("Note",uniprot.getLinkForAlignment());
                a.addEnS("Remark",ALI_SCORE+score);
                a.setReferenceSequence(uniprot, start,end );
                final int selOffset=a.getSelectedAminoacidsOffset();
                if (selOffset>=p.countResidues()) disabledWhy="selOffset>=p.countResidues()";
                else if (poorAlignment) disabledWhy="Poor alignment ";
                else if (strstr("Status=Potential",T,b,e)>=0) disabledWhy="Occurs 'Status=Potential'";
                if (rule!=null && rule!="-") {
                    final char amino=(char) (p.getResidueType(selOffset)&~32);
                    if (rule.indexOf(amino)<0) disabledWhy="Only one amino and "+amino+" not in "+rule;
                }
                if (isSecStru && p.getResidueSecStrType()!=null) disabledWhy="Secondary structure is already infered";
            }
            if (viewSrcText!=null) a.addEnS("SourceXML",viewSrcText);
            a.setSrcText(line);
            a.setStyle(VisibleIn123.LINES | (1<<VisibleIn123.BIT_SHIFT_LINE));
            a.addEnS(ResidueAnnotation.NAME,fName);
            if (!isXML) a.addEnS("SourceLine",line.toString().replace('\t',' '));
            a.addEnS("Remark",attribute);
            setFN(fName, isXML ? SRC_DAS : SRC_GFF_EXPASY,a);
            a.setFeatureSrc(source, seqname);
            addFeature(a,p);
            a.setValue(ResidueAnnotation.GROUP, (disabledWhy==null ? "" : "deactivated ")+(isXML ? "DAS-":"GFF-")+"features");
            if (countTrue(a.getSelectedAminoacids())==0) disabledWhy="No aminos selected";
            if (disabledWhy!=null) {
                a.setEnabled(false);
                a.addEnS("Disabled","Why disabled: "+disabledWhy);
            }
            a.readFeatureAnnotations();
            (vRA==null ? vRA=new ArrayList() : vRA).add(a);
        }
        return toArry(vRA, ResidueAnnotation.class);
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Ontology >>> */
    private static Set<String> setSeqOnt;
    private static Set<String> setSequenceOntologyTermsLC() {
        if (setSeqOnt==null) {
            setSeqOnt=new HashSet();
            final ByteArray ba=readBytes(rscAsStream(0L,_CC,"sequence_ontology_terms.rsc"));
            if (ba!=null) {
                final byte T[]=ba.getBytes();
                for(int i=ba.getEnd(); --i>=0;)  if ('A'<=T[i] && T[i]<'Z') T[i]=(byte)(T[i]|32);
                final ChTokenizer TOK=new ChTokenizer().setText(ba);
                while(TOK.nextToken()) setSeqOnt.add(TOK.asString());
            } else stckTrc();
        }
        return setSeqOnt;
    }
    /* <<< Ontology <<< */
    /* ---------------------------------------- */
    /* >>> InProteinFile >>> */
    /*
      pdb1vyp.ent:
      REMARK 800 SITE
      REMARK 800 SITE_IDENTIFIER: AC1
      REMARK 800 SITE_DESCRIPTION: TNF BINDING SITE FOR CHAIN X
      SITE     1 AC1 17 HOH C 699  HOH C 700  HOH C 701  HOH C 702
      SITE     2 AC1 17 ALA X  23  PRO X  24  LEU X  25  THR X  26
      SITE     3 AC1 17 ALA X  58  GLN X 100  HIS X 181  HIS X 184
      SITE     4 AC1 17 ARG X 233  ALA X 302  GLY X 323  ARG X 324
      SITE     5 AC1 17 TYR X 351
      SITE     1 AC2 10 HOH C 703  HOH C 704  THR X  26  ALA X  58
      SITE     2 AC2 10 TYR X  68  PHE X 102  HIS X 181  HIS X 184
      SITE     3 AC2 10 TYR X 186  TYR X 351
    */
    public static ResidueAnnotation[] readFeaturesInProteinFile(ByteArray text, int maxLen, StrapProtein... pp) {
        final List<ResidueAnnotation> vAnnotations=new ArrayList();
        ByteArray ba=null;
        boolean found=false;

        for(StrapProtein p : pp) {
            boolean thisFound=false;
            ba=pp.length==1 && text!=null ? text : readBytes(p.getFile(),ba);
            if (ba==null) continue;
            final byte[] T=ba.getBytes();
            final int B=ba.getBegin(), E=ba.getEnd(),  ends[]=ba.getLineEnds();
            final ChTokenizer TOK=new ChTokenizer();
            Map<String,String> mapRemark=null;
            for(int iL=0; iL<ends.length; iL++) {
                final int b=iL==0 ? B :ends[iL-1]+1, e=ends[iL];
                if (e-b>28 && T[b+7]=='8' && T[b+8]=='0' && T[b+9]=='0' && T[b]=='R' && equls("REMARK 800 SITE_IDENTIFIER: ",T,b)) {
                    final String id=ba.toString(b+28,e, ByteArray.STRING_TRIM);
                    for(int jL=iL+1; jL<ends.length; jL++) {
                        final int lineBegin=ends[jL-1]+1;
                        if (is(LETTR_DIGT, T, lineBegin)) {
                            if (equls("REMARK 800 SITE_DESCRIPTION: ",T,lineBegin)) {
                                (mapRemark==null ? mapRemark=new HashMap() : mapRemark).put(id,ba.toString(lineBegin+29, ends[jL], ByteArray.STRING_TRIM));
                            }
                            break;
                        }
                    }
                }
            }
            int b=0, e=0;
            ResidueAnnotation lastAnno=null;
            nextLine:
            for(int iL=0; iL<ends.length; iL++, b=e+1) {
                e=prev(NO_SPC,T,ends[iL]-1,b-1)+1;
                final ResidueAnnotation a;
                final String featureName;
                if (e-b>10 && T[b]=='S' && equls("SITE: ",T,b)) {
                    final String siteID=tokn(T,b+11,e);
                    featureName=capitalize(lCase(siteID));
                    final StringBuffer sb=strBuf(SB_TIP);
                    for(int i=0; i<3; i++) {
                        final int resnum=b+26+i*11;
                        if (resnum>=e || T[resnum]==' ') break;
                        final char iCode=(char)T[resnum+1];
                        sb.append(atoi(T,resnum-3));
                        if (iCode!=' ') sb.append(iCode);
                        sb.append(':').append((char)T[resnum-4]).append(' ');
                    }
                    final String remark=mapRemark!=null ? mapRemark.get(siteID) : null;
                    a=new ResidueAnnotation(p);
                    a.addEnS(ResidueAnnotation.POS, sb.toString());
                    if (remark!=null) a.addEnS("Remark",remark);
                    thisFound=found=true;
                } else if (T[b]=='M' && equls("MODRES ",T,b) && e-b>25) {
                    /*
                      MODRES 1IVO ASN A   32  ASN  GLYCOSYLATION SITE
                    */
                    a=new ResidueAnnotation(p);
                    featureName=capitalize(lCase(translateFeatureName(ba.toString(b+29, e, ByteArray.STRING_TRIM|ByteArray.STRING_POOL), (boolean[])null)));
                    a.addEnS("Remark",featureName);
                    a.addEnS("Note",ba.newString(b,e));
                    final char iCode=(char)T[b+22];
                    //final String pos= atoi(T,b+18)+(iCode!=' '?iCode+":":":")+(char)T[b+16];
                    a.addEnS(ResidueAnnotation.POS, atoi(T,b+18)+(iCode!=' '?iCode+":":":")+(char)T[b+16]);
                    thisFound=found=true;
                } else if (T[b]=='F' && T[b+1]=='T' &&  T[b+2]==' ') {
                    TOK.setText(T,b+3, e);
                    if (!TOK.nextToken()) continue nextLine;
                    if (lastAnno!=null && equls(" /FTId=",T,b+33)) {
                        lastAnno.addEnS("Remark", ba.newString(b+35, T[e-1]=='.' ? e-1 : e));
                        continue nextLine;
                    }
                    {
                        final int f=TOK.getFrom();
                        final byte c0=T[f];
                        if ( (c0=='C'&&TOK.eq("CDS")||TOK.eq("CHAIN")) || c0=='e'&&TOK.eq("exon")  || !is(LETTR,c0) || !TOK.containsOnly(LETTR_DIGT_US)) continue nextLine;
                    }
                    final String feature0=TOK.asStringIP();
                    if (!TOK.nextToken() || !TOK.containsOnly(DIGT)) continue;
                    final int start=TOK.asInt()-1;
                    if (!TOK.nextToken() || !TOK.containsOnly(DIGT)) continue;
                    final int end=TOK.asInt();
                    if (end-start<1 || end-start>maxLen) continue;
                    featureName=capitalize(lCase(translateFeatureName(feature0, (boolean[])null)));
                    final String interval=featureName2color(featureName).rule=="-" ? "-" : ",";
                    a=new ResidueAnnotation(p);
                    a.addEnS("Remark",ba.newString(b,e));
                    a.setStyle(VisibleIn123.LINES | (1<<VisibleIn123.BIT_SHIFT_LINE));
                    a.addEnS(ResidueAnnotation.POS, start+1==end ? String.valueOf(end) : (1+start)+interval+(end));
                    thisFound=found=true;
                } else { a=null; featureName=null;}
                if (a!=null && featureName!=null) {
                    a.setFeatureSrc("ProteinFile","null");
                    a.readFeatureAnnotations();
                    addFeature(a,p);
                    setFN(featureName, SRC_PROTEIN_FILE,a);
                    a.addEnS(ResidueAnnotation.NAME,featureName);
                    if (isDeactivated(featureName)) {
                        a.setEnabled(false);
                        a.addEnS("Disabled","Why disabled: isDeactivated(featureName)");
                    }
                    lastAnno=a;
                    vAnnotations.add(a);
                }
            }
            if (thisFound) SequenceFeatures.keepFeaturesWithBestScore(p);
        }
        final ResidueAnnotation aa[]=toArry(vAnnotations,ResidueAnnotation.class);
        if (found) {
            disableSecondaryStructure(aa);
            Protein.incrementModificationCount(ProteinModiCount.MCA_SEQUENCE_FEATURES_V,null);
            StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_ADDED, 999);
        }
        return aa;
    }
    /* <<<  In Protein File <<< */
    /* ---------------------------------------- */

}