package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import java.io.*;
import java.awt.*;
import javax.swing.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP

Jalview is an excellent well-known alignment viewer.<br>
To name a few features:
<ul>
<li>Evolutionary dendrograms</li>
<li>Export of Encapsulated Postscript (eps) Images which can be included in OpenOffice documents.</li>
<li>Jalview can retrieve and visualize features from many BioDAS* sources at once.</li>
</ul>
When Jalview is started the first time ~9 Megabyte are downloaded.<br><br>
*/

public class Export2Jalview {
    public final static int SECSTRU=1;
    private static CharSequence gff(StrapProtein pp[], int mode) {
        final StringBuffer sb=new StringBuffer(9999).append("##gff-version 3\n");
        for(StrapProtein p : pp) {
            final String pName=withoutDotSuffix(p.getName());
            for(ResidueAnnotation a : p.getResidueAnnotations()) {
                final String feature= a.getFeatureName();
                if (feature==null || (a.getVisibleWhere()&VisibleIn123.JALVIEW)==0 || !a.isEnabled()) continue;
                final int selOffset=a.getSelectedAminoacidsOffset();
                final boolean bb[]=a.getSelectedAminoacids();
                sb.append(pName).append('\t').append(a.getFeatureSrcDb()).append('\t').append(feature).append('\t')
                    .append(firstTrue(bb)+selOffset+1).append('\t')
                    .append(lastTrue(bb)+selOffset+1).append("\t.\t.\t.\t")
                    .append(a.getValue("Remark")).append("\t\n");

            }
            final byte[] ss=p.getResidueSecStrType();
            if ( (mode&SECSTRU)!=0 && ss!=null) {
                for(int  iType=2; --iType>=0;) {
                    final char type=iType==0 ? 'h' : 'e';
                    for(int i=0; i<ss.length; i++) {
                        if ((ss[i]|32)==type) {
                            sb.append(pName).append("\tPDB\t").append(type=='h' ? "Helix" : "Beta strand").append('\t')
                                .append(i+1).append('\t');
                            for(; i<ss.length && (ss[i]|32)==type; i++);
                            sb.append(i).append("\t.\t.\t.\tfrom structure\t\n");
                        }
                    }
                }
            }
        }
        return sb;
    }

    public void launchJalview(StrapProtein pp[], int mode) {
        if (pp==null) return;
        final File fFeature=file(STRAPOUT+"/jalview/features.gff"), fAlign=file(STRAPOUT+"/jalview/align.msf");
        delFile(fFeature);
        delFile(fAlign);
        wrte(fFeature,gff(pp,mode));
        StringBuffer sb=new StringBuffer();
        ExportAlignment.export(pp,ExportAlignment.MSF,null,60,null,0f,sb);
        wrte(fAlign,sb);
        sb.setLength(0);
        sb.append("http://www.jalview.org/services/launchApp?open=").append(java.net.URLEncoder.encode(fAlign.toString()))
            .append("&features=").append(java.net.URLEncoder.encode(fFeature.toString()));
        visitURL(sb.append(VISIT_URL_JAVAWS),0);
    }
}
// http://www.compbio.dundee.ac.uk/~ws-dev1/jalview/latest/webstart/jalview.jnlp
// http://www.compbio.dundee.ac.uk/~ws-dev1/jalview/release/jalview-source.tar.gz
