package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.VisibleIn123;
import charite.christo.interfaces.Colored;
import charite.christo.protein.*;
import charite.christo.strap.interfaces.AlignmentWriter;
import java.awt.Color;
import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static java.awt.Color.*;
import static charite.christo.HtmlUtil.*;
/**HELP

<div class="NOPRINT">

<br><br><u>View</u>
<ul>
<li>Sequence features like Glycosylation are underlined. They have a balloon message.</li>
<li>Helices and beta sheets have a red and yellow background.</li>
<li>Residue annotations are underlined.</li>
</ul>

<br><u>User Interaction</u>
<ul>
<li>Proteins with an accession ID have a link to the protein file database.</li>
</ul>

<br><u>Known Problems:</u>
<ul>
<li>Line-wrap: In text-processors the alignment lines may be broken.
Work around: select entire text and set a small character size. Use font "Courier".
</li>
<li>Openoffice and TextEdit: Feature lines too high. Please vote in bugzilla: http://www.openoffice.org/issues/show_bug.cgi?id=99259</li>
<li>Abiword: Not yet working.</li>
</ul>
<br>

<br><u>Tips and Tricks</u>
<ul>
<li>Drag the mouse with the alt-key to select columns in MS-Word (so-called rectangular selection).</li>
<li>Images can be searched in Google and dragged with the mouseonto proteins (See <a href="http://3d-alignment.eu/dragProteinLink.html">Dragging Web links</a>). They appear as protein icon in the alignment.</li>

</ul>
</div>

   @author Christoph Gille
*/
public class ExportAlignment implements AlignmentWriter {

    private String[] names;
    private StrapProtein proteins[];
    private int firstCol=0, toColumn=MAX_INT, resPerLine=50;
    public StrapProtein[] getProteins() { return proteins;}
    public String[] getNames(){return names;}
    public void setProteins(StrapProtein... pp) { proteins=pp;}
    public void setNames(String[] names) { this.names=names;}
    public void setColumnRange(int first, int last){ firstCol=first; toColumn=last; }

    public void setResiduesPerLine(int n) { resPerLine=n;}
    public void setCharForGapAndTerminalBlanks(char gap,char blankN, char blankC) { }
    public String getFileExtension() { return suffix;}
    private String suffix;
    private static int imgSize;
    static void setIconSize(int pxl) {  imgSize=pxl; }

    public void getText(long options,StringBuffer sb) {
        final float sim=(options&HTML)!=0 ?  StrapView.sliderSimilarity().getValue()*.01f : 0;
        if ( (options&FASTA)!=0) exportFasta(options,sb);
        else export(proteins,options, new Range(firstCol,toColumn),resPerLine,getNames(),sim,sb);
        this.suffix= (options&CLUSTALW)!=0 ? "clustalw" : (options&MSF)!=0 ? "msf" : (options&FASTA)!=0 ? "fasta" : "html";
    }

    public static StringBuffer export(StrapProtein pp[],long mode,Range fromTo,int width, String names[],float shadingThreshold, StringBuffer sb0) {
        final boolean
            isHtml= (mode&HTML)!=0,
            isMsf= (mode&MSF)!=0,
            isClustalW=(mode&CLUSTALW)!=0,
            isDna=(mode&NUCLEOTIDE_TRIPLET)!=0,
            isBrowser= (mode&CSS_BROWSER)!=0,
            isFeatures= (mode&SEQUENCE_FEATURES)!=0,
            isSpace=!isHtml && (mode&EXTRA_SPACE)!=0;

        final int wide2narrow[]=new Gaps2Columns().computeWide2narrow(pp), cols[][]=new int[pp.length][], cols2idx[][]=new int[pp.length][];
        final Range columnRange=StrapAlign.translateColumnsToLessProteins(pp, fromTo!=null ? fromTo.getFrom():0, fromTo!=null ? fromTo.getTo():MAX_INT);
        if (columnRange==null) return sb0;
        final String[] pNames=names!=null ? names:new String[pp.length], pIcons=isHtml && (mode&PROTEIN_ICONS)!=0 ? new String[pp.length] : null;
        int maxCol=0;
        for(int iP=pp.length; --iP>=0;) {
            final StrapProtein p=pp[iP];
            final int nRes=p.countResidues();
            cols[iP]=Gaps2Columns.computeColumns(wide2narrow,p.getResidueGap(),nRes, null);
            cols2idx[iP]=col2nextIdx(cols[iP],nRes);
            maxCol=maxi(maxCol,maxCol(cols[iP],nRes));
            if (pNames[iP]==null) {
                final String na=p.getName();
                pNames[iP]=(mode&FILE_SUFFIX)==0 && na.lastIndexOf('.')>0 ? na.substring(0,na.lastIndexOf('.')) : na;
                if (pIcons!=null) pIcons[iP]=p.getIconUrl();
            }
        }
        int nLenMax=0,maxLen=0;
        ArrayList<String> vFeatureNames=null;
        for(int i=pNames.length;--i>=0;) maxLen=maxi(maxLen,pNames[i].length());

        if (maxCol>columnRange.getTo()) maxCol=columnRange.getTo();
        final int maxFeaturesPos=isFeatures ? SequenceFeatures.layoutFeatures(pp) : 0;
        final int[] current=isHtml ? createBuffer() : null;

        long linesWithFeature=0;
        final StringBuffer sb;
        if (sb0==null) {
            sb=new StringBuffer(999999);
        } else sb=sb0;

        if (isHtml) {
            for(StrapProtein p : pp) {
                for(ResidueAnnotation a:p.getResidueAnnotations()) {
                    if (a.getFeatureName()==null || (a.getVisibleWhere()&VisibleIn123.HTML)==0 || !a.isEnabled()) continue;
                    final int line=(a.getStyle()>>VisibleIn123.BIT_SHIFT_LINE)&255;
                    linesWithFeature |= (1<<line);
                }
            }
            if ((mode&HTML_HEAD_BODY)!=0) {
                sb.append("<html><head>\n<style type=\"text/css\">\n");
                final Customize custCss= HtmlUtil.customize('A');
                if (custCss!=null) for(String s : custCss.getSettings()) sb.append(s).append('\n');
                if (isBrowser && isFeatures) SequenceFeatures.cssColors(pp, sb);
                sb.append("</style>\n");
                if (isBrowser && isFeatures) {
                    sb.append("<!--[if IE]><style>\n");
                    for(String s : SequenceFeatures.customizeCSS_IE().getSettings()) sb.append(s).append('\n');
                    sb.append("</style><![endif]-->\n");
                }
                sb.append("<title>alignment</title></head>\n<body>\n");
                if (isBrowser) {
                    sb.append("\n<script language=\"JavaScript\" type=\"text/javascript\">\n");
                    appendRsc("alignBrowser.js",sb);
                    sb.append("\n</script>\n");
                    sb.append(HtmlDoc.getCssOrJS("js",DOCTYPE_HTML, true));
                    appendRsc("alignBrowserConfig.html",sb);

                }
            }
        } else if (isMsf) {
            sb.append("!!AA_MULTIPLE_ALIGNMENT 1.0\n\n test.msf  MSF: 224  Type: P  September  9, 2009 09:09  Check: 9999 ..\n\n");
        } else if (isClustalW) {
            sb.append("CLUSTAL W 2.1 multiple sequence alignment\n\n");
        }
        // Are there selected residues ?
        for(int i=0;i<pp.length;i++) {
            final int j=pNames[i].length();
            if (j>nLenMax) nLenMax=j;
            if (isMsf) sb.append(" Name: ").append(pNames[i]).append(" Len:").append(pp[i].countResidues()).append(" Check: 1 Weight:1\n");
        }
        if (isMsf) sb.append("\n//\n\n");
        final byte[] TRIPLET=new byte[3];
        final char c13[]=new char[isDna ? 3:1];
        final Color shadingColors[]=isHtml ? StrapAlign.getShadingColors() : null;
        final short[][] frequencies=isHtml ? computeFrequenciesAtColumn(pp,cols,columnRange.getFrom(), columnRange.getTo()+1) : null;
        for(int h=columnRange.getFrom(); h<=maxCol;h+=width) {
            for(int iP=0;iP<pp.length;iP++) {
                if (isHtml) {
                    sb.append("<pre class=\"").append( pIcons!=null && pIcons[iP]==null && imgSize==0 ? "iseq\">" : "seq\">");
                    if (pIcons!=null) {
                        if (pIcons[iP]!=null) {
                            sb.append("<img class=\"proteinIcon\" alt=\" \" src=\"").append(pIcons[iP]).append('"');
                            if (imgSize!=0) sb.append(" width=\"").append(imgSize).append("\" height=\"").append(imgSize).append('"');
                            sb.append(" />");
                        }
                        else if(imgSize!=0) HtmlUtil.blankImage(imgSize,imgSize,sb);
                    }
                }
                final StrapProtein p=pp[iP];
                final byte[] selected=isHtml ? p.getSelectedAminoacids() : null;
                final int nRes=p.countResidues();
                final byte resType[]=p.getResidueType();
                final boolean hasNt=p.getNucleotides()!=null;
                final int column2nextIdx[]=cols2idx[iP];
                final byte resSecStr[]=(mode&SECONDARY_STRUCTURE)!=0 ? p.getResidueSecStrType() : null;
                final int resCol[]=cols[iP];
                final int firstCol=resCol.length>0 ? resCol[0] : 0;
                final int toColumn=nRes>0 ? resCol[nRes-1] : 0;
                sb.append(pNames[iP]);
                final String accessionId=isBrowser ? p.getAccessionID() : null;
                if (accessionId!=null)  sb.append("<a class=\"proteinLink\" href=\"javascript:shw('").append(accessionId).append("');\">*</a>");
                for(int j=pNames[iP].length()-(accessionId!=null? 0 :1); j<maxLen; j++) sb.append(' ');
                if (isHtml) {
                    int num=MIN_INT;
                    if (h+width>=firstCol && h<=toColumn && h>=0 && h<=column2nextIdx.length) {
                        final int idx=column2nextIdx[h], nn[]=p.getResidueNumber();
                        num= (mode&PDB_RESIDUE_NUMBER)!=0 && nn!=null ? get(idx,nn) : idx+1;
                    }
                    final String s=num==MIN_INT ? "" : String.valueOf(num);
                    for(int j=5-s.length(); --j>=0;) sb.append(' ');
                    sb.append(s).append(' ');
                }
                for(int hh=h;hh<h+width && hh<=maxCol; hh++) {
                    final int iAA= hh<column2nextIdx.length  ? column2nextIdx[hh] : -1;
                    if (iAA>=0 && hh==resCol[iAA]) { // is no gap
                        if (isDna && hasNt) {
                            final byte trip[]=p.getResidueTriplet(iAA,TRIPLET);
                            if (trip!=null) {
                                c13[0]=(char)(trip[0]&~32);
                                c13[1]=(char)(trip[1]&~32);
                                c13[2]=(char)(trip[2]|32);
                            } else c13[0]=c13[1]=c13[2]='?';
                        }
                        else {
                            c13[0]=(char)resType[iAA];
                            if (!is(LETTR,c13[0])) c13[0]='X';
                            if (isDna) {c13[1]=c13[2]='~';}
                        }
                        if (isHtml && iAA>=0 && nRes>iAA) {
                            final int iCol=resCol[iAA];
                            final boolean isCons;
                            if (frequencies!=null && frequencies.length>iCol) {
                                final int c=(resType[iAA]|32)-'a';
                                isCons=frequencies[iCol]!=null ?  c>=0 && c<1+'z'-'a' && frequencies[iCol][c]>shadingThreshold*pp.length : true;
                            } else isCons=true;
                            final int bg;
                            if (resSecStr!=null&&iAA<resSecStr.length) {
                                final byte ss=resSecStr[iAA];
                                bg=ss=='H' ? 0xFFCCBB : ss=='E' ? 0xFFFF99 : -1;
                            } else bg=-1;
                            Color color=(isCons ? shadingColors[resType[iAA]] : BLACK);
                            if (color==WHITE) color=BLACK;
                            Color selection=null;
                            String tooltip=null;
                            if (selected!=null && selected.length>iAA && selected[iAA]!=0) {
                              for(ResidueSelection s : p.getResidueSelectionsAt(0L,iAA,iAA+1,VisibleIn123.HTML)) {
                                    if (s instanceof ResidueAnnotation) {
                                        if (!ResidueAnnotation.isEnabledAnno(s)) continue;
                                        String tip=TabItemTipIcon.tip(s);
                                        if (tip!=null && !"null".equals(tip))  tooltip= tooltip==null ? tip : tooltip+"<br>"+tip;
                                    }
                                    if (s instanceof Colored) selection=((Colored)s).getColor();
                                    if (selection!=null) break;
                                }
                            }
                            final boolean underline=selection!=null && (mode&UNDERLINE_ANNOTATIONS)!=0;
                            setStyles(  (underline?U:0) | (isCons?B:0) , rgb(color),bg,current,"","",sb);
                        }
                        sb.append(c13);
                        if(isSpace) sb.append(' ');
                    } else {
                        for(int j=isDna ? 3 : 1; --j>=0;) {
                            sb.append(isHtml&&(hh<firstCol||hh>toColumn)?' ' :  isClustalW  ? '-' : '.');
                            if(isSpace) sb.append(' ');
                        }
                    }

                }// hh
                if (isHtml) {
                    setStyles(0,0,-1,current,"","",sb);
                }
                if (vFeatureNames==null) vFeatureNames=new ArrayList();
                if (isHtml) {
                    if (isFeatures) {
                        tellFeatures(p,h,h+width,current,resCol,column2nextIdx, sb,vFeatureNames);
                        sb.append("&nbsp;</pre>\n");
                        showFeatures(p,h,mini(h+width,maxCol+1),maxLen+7,current, isBrowser ,maxFeaturesPos,resCol,column2nextIdx, pIcons!=null,  sb);
                    } else sb.append("&nbsp;</pre>");
                }
                sb.append('\n');
            }//iP

            if (isHtml) sb.append("<hr>");
            sb.append("\n\n");
        }//h
        if (isHtml) {
            if (sze(vFeatureNames)>0) {
                sb.append("<table summary=\"sequence feature\">\n");
                final Set vNames=new HashSet();
                for(StrapProtein p: pp) {
                    for(ResidueAnnotation a:p.getResidueAnnotations()) {
                        final String fn=a.getFeatureName();
                        if (!vNames.contains(fn)) {
                            vNames.add(fn);
                            final Color color=a.getColorForWhiteBG();
                            sb.append("<tr><td>");
                            setStyles(0,0,rgb(color), current,"","",sb);
                            sb.append(" &nbsp; ");
                            setStyles(0,0,-1,current,"","",sb);
                            sb.append("</td><th>").append(SequenceFeatures.toShortFeatureName(fn)).append("</th><td>").append(fn).append("</td></tr>\n");
                        }
                    }
                }
                sb.append("</table>\n");
            }
            if (isHtml && isFeatures) {
                final String hlp=getHlp(ExportAlignment.class);
                if (hlp!=null) sb.append(hlp);
            }
            if ((mode&HTML_HEAD_BODY)!=0) sb.append("</body></html>\n");
        }
        return sb;
    }

    private static HashSet set_TF;
    private static void tellFeatures(StrapProtein p, int hFrom, int hTo, int[] current, int cols[], int c2i[], StringBuffer sb, Collection<String> vA) {
        setStyles(0,0,-1,current,"","",sb);
        final int nRes=p.countResidues();
        final int iFrom=hFrom>=0 && hFrom<c2i.length ? c2i[hFrom] : -1;
        if (iFrom<0) return;
        int iTo=hTo>=0 && hTo<c2i.length ? c2i[hTo] : maxCol(cols,nRes);
        if (iTo>=0 && iTo<cols.length && cols[iTo]>hTo) iTo--;
        /* sb.append(" ("+iFrom+" "+iTo+") "); */
        boolean isFirst=true;
        if (set_TF==null) set_TF=new HashSet();
        set_TF.clear();
        for(ResidueAnnotation a:p.getResidueAnnotations()) {
            final String fn=a.getFeatureName();
            if (fn==null || !a.isEnabled()) continue;
            //  if ((a.getVisibleWhere()&VisibleIn123.HTML)==0
            if (set_TF.contains(fn)) continue;
            final int selOffset=a.getSelectedAminoacidsOffset();
            final boolean bb[]=a.getSelectedAminoacids();
            boolean contains=false;
            for(int iA=maxi(iFrom,selOffset); iA<iTo; iA++) {
                final int iB=iA-selOffset;
                if (iB>=bb.length) break;
                if (iB>=0 &&  bb[iB]) { contains=true; break;}
            }
            if (!contains) continue;
            sb.append(isFirst?" (":",");
            isFirst=false;
            final Color color=a.getColorForWhiteBG();
            setStyles(0, rgb(color), -1,current,"","",sb);
            sb.append(SequenceFeatures.toShortFeatureName(fn));
            if (!vA.contains(fn)) vA.add(fn);
            set_TF.add(fn);
        }
        setStyles(0,0,-1,current,"","",sb);
        if (!isFirst) sb.append(")");
    }
    private static void showFeatures(StrapProtein p, int hFrom, int hTo, int leadingSpace, int current[],  boolean isBrowser, int maxFeaturesPos, int resCol[], int c2i[], boolean hasIcon, StringBuffer sb) {
        final int nRes=p.countResidues();
        final int maxCol=mini(maxCol(resCol,nRes), c2i.length-1, hTo-1);
        final ResidueAnnotation aa[]=p.getResidueAnnotations();
        final byte bb[]=p.getSelectedAminoacids();
        final StringBuffer sbTmp=new StringBuffer();
        int count=0;
        for(int iF=0; iF<10; iF++) {
            boolean hasFeature=false;
            for(ResidueAnnotation a:aa) {
                if (a.getFeatureName()==null || !a.isEnabled()  || (a.getVisibleWhere()&VisibleIn123.HTML)==0 ) continue;
                if (iF==((a.getStyle()>>VisibleIn123.BIT_SHIFT_LINE) & 0xFF)) { hasFeature=true; break;}
            }
            if (!hasFeature) continue;
            sb.append("<pre class=\"").append(hasIcon &&  imgSize==0 ? "iftr\">" : "ftr\">");
            if (imgSize!=0) HtmlUtil.blankImage(imgSize,1,sb);
            count++;
            int space=leadingSpace;
            String lastAttr="";
            nextCol:
            for(int h=hFrom; h<=maxCol; h++) {
                final int iAa=c2i[h];
                if (iAa>=bb.length) break;
                if (iAa<0 || iAa>nRes || h!=resCol[iAa] || bb[iAa]==0) {   space++; continue;}
                for(ResidueAnnotation a:aa) {
                    if (a.getFeatureName()==null || !a.isEnabled() ||  (a.getVisibleWhere()&VisibleIn123.HTML)==0 ) continue;
                    final boolean sel[]=a.getSelectedAminoacids();
                    final int selOffset=a.getSelectedAminoacidsOffset();
                    final int iB=iAa-selOffset;
                    if (sel==null || iB<0 || iB>=sel.length || !sel[iB]) continue;
                    if (iF!=((a.getStyle()>>VisibleIn123.BIT_SHIFT_LINE) & 0xFF)) continue;
                    if (space>0) {
                        setStyles(0,0xFFffFF,0xFFffFF,current,"","",sb);
                        while(space>0) { space--; sb.append(' ');}
                    }
                    String url=a.getLastValue("DAS");
                    if (url==null) url=a.getLastValue(CLASS_Hyperrefs);
                    //putln("url="+url);
                    final int color0=0xFFffFF & rgb(a.getColorForWhiteBG());
                    final int color=color0==0xffFFff ? 0xddDDdd : color0;
                    sbTmp.setLength(0);
                    if (isBrowser) sbTmp.append(" class=\"").append(a.getCssClass()).append("\" ");
                    else toHex(color, 6, sbTmp.append(" style=\"background-color: #")).append(";\" ");
                    HtmlUtil.attributeTooltip(a.getFeatureName(), color, sbTmp);
                    final String attr=sbTmp.toString();
                    if (!attr.equals(lastAttr)) {setStyles(0,0xFFffFF,0xFFffFF,current,"","",sb); lastAttr=attr;}
                    setStyles( (url!=null ? A : 0) | F, color, -1, current, "javascript:shw('"+url+"');",attr,sb);
                    sb.append('-');
                    continue nextCol;
                }
                space++;
            }
            setStyles(0,0,-1,current,"","",sb);
            sb.append(" </pre>\n");
        }
        for(;count<maxFeaturesPos; count++) sb.append("<pre class=\"ftr\"> &nbsp; </pre>\n");
    }
    private static int[] col2nextIdx(int cc[], int nRes) {
        if (nRes==0) return new int[0];
        final int nI=cc[nRes-1];
        final int ii[]= new int[nI+1];
        Arrays.fill(ii,0,nI+1,-1);
        for(int iAa=0;iAa<nRes;iAa++) {
            final int c=cc[iAa];
            if (c>=0 && c<ii.length) ii[c]=iAa;
        }
        int last=-1;
        for(int i=nI;i>=0;i--)
            if (ii[i]<0) ii[i]=last;
            else last=ii[i];
        return ii;
    }
    private static int maxCol(int col[], int nRes) {
        return col.length>=nRes && nRes>0 ? col[nRes-1] : -1;
    }
    // &#x2500;

    private static short[][] computeFrequenciesAtColumn(StrapProtein[] pp, int[][] columns, int fromCol, int toCol) {
        final short[][] frequencies=new short[toCol]['Z'-'A'+1];
        for(int iP=pp.length; --iP>=0;) {
            final StrapProtein p=pp[iP];
            final int cols[]=columns[iP];
            final byte rt[]=p.getResidueType();
            for(int iA=p.countResidues(); --iA>=0; ) {
                final int c=cols[iA];
                if (c<fromCol || c>=toCol) continue;
                final byte a=rt[iA];
                if ('a'<=a && a<='z' || 'A'<=a && a<='Z') frequencies[c][ (a|32) -'a']++;
            }
        }
        return frequencies;
    }

    private void exportFasta(long options, StringBuffer sb){
        final StrapProtein pp[]=this.proteins;
        for(int iP=0;iP<pp.length;iP++) {
            final StrapProtein  p=pp[iP];
            final String s=names!=null && names[iP]!=null ? names[iP] : p.getName();
            if (sb!=null) sb.append('>').append(s).append('\n');
            final byte[] res=p.getGappedSequence();
            if (res==null) return;
            final int from=firstCol, to=mini(p.getMaxColumn()+1, res.length, toColumn+1);
            int count=0;
            for(int i=from; i<to;i++) {
                final byte c=res[i];
                final boolean let=is(LETTR,c);
                if ( (options&NO_GAPS)!=0 && let) continue;
                sb.append(let?(char)c :'-');
                if (++count % resPerLine ==0 && i!=to-1) sb.append('\n');
            }
            sb.append('\n');
        }
    }
}
