package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import charite.christo.strap.interfaces.SelectorOfNucleotides;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.util.List;
import java.util.regex.Pattern;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
class HighlightPattern extends AbstractVisibleIn123 implements ActionListener {
    private byte[][] patterns;
    private boolean isNucleotide;
    private String patternText="", sPatterns[];
    public String getText() { return patternText;}
    private ButColor butColor;
  final NextAndPrevSelection nextAndPrevSelection=new NextAndPrevSelection();
    public ButColor butColor(){
        if (butColor==null)  butColor=new ButColor(0L,null,this);
        return butColor;
    }

    public void actionPerformed(ActionEvent ev) {
        if (ev.getActionCommand()==ACTION_COLOR_CHANGED)  {
            StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR,111);
        }
    }

    public boolean  setText(String text0) {
        final String s=text0!=null ? text0.trim().toUpperCase() : "";
        try {
            return !s.equals(patternText);
        } finally {
            sPatterns=splitTokns(0L, patternText=s);
            patterns=new byte[sPatterns.length][];
            for(int i=patterns.length;--i>=0; ) patterns[i]=sPatterns[i].getBytes();
        }
    }
    final boolean isNucleotide() {return isNucleotide;}
    final boolean setNucleotide(boolean b) {  try { return isNucleotide==b; } finally { isNucleotide=b; } }
    void save(StringBuffer sb) {
        String pat=patternText;
        if (pat.length()>0){
            final Color c=getColor();
            final int rgb=c!=null?c.getRGB() : 0xff00, a=c!=null?c.getAlpha():255;
            sb.append(pat+"\t"+((rgb>>16)&255)+","+ ((rgb>>8)&255)+","+(rgb&255)+","+a+"\t"+getVisibleWhere()+"\t"+isNucleotide+"\n");
        }
    }
    private void parse(String line) {
        final String ss[]=splitTokns(0L, line,"\t");
        if (sze(ss)>3) {
            setText(ss[0]);
            setColor(str2color(ss[1]));
            setVisibleWhere(atoi(ss[2]));
            setNucleotide("true".equals(ss[3]));
        }
    }
    // ----------------------------------------
    private final DialogHighlightPattern dialog;
    public HighlightPattern(String line,DialogHighlightPattern d) {
        dialog=d;
        parse(line);
        setStyle(FULL_BACKGROUND);
        setVisibleWhere(0xffffffff);
        setColor(Color.GREEN);
    }
    // ----------------------------------------
    public Selector getResidueSelection(StrapProtein p) {
        for(ResidueSelection s : p.getResidueSelections()) {
            if (s instanceof Selector) {
                if (((Selector)s).highlightPattern==this) {
                    return (Selector)s;
                }
            }
        }
        return null;
    }
   public ResidueSelection[] getResidueSelections(StrapProtein pp[]) {
     final List v=new ArrayList();
     for(StrapProtein p : pp) {
       final ResidueSelection s=p==null ? null : getResidueSelection(p);
       if (s!=null) v.add(s);
     }
     return toArry(v,ResidueSelection.class);
   }
    void addToProteins(StrapProtein[] pp) {
        final boolean remove=sze(getText())==0 || getVisibleWhere()==0;
        for(StrapProtein p : pp) {
            final Selector s0=getResidueSelection(p);
            if (remove) {
                if (s0!=null) {
                    p.removeResidueSelection(s0);
                    s0.setProtein(null);
                }
            } else {
                final Selector s=s0!=null ? s0: new Selector();
                s.highlightPattern=this;
                if (s!=s0) p.addResidueSelection(s);
                s.setProtein(p);
            }
        }
    }
    // ----------------------------------------
    final class Selector extends AbstractResidueSelection implements SelectorOfNucleotides, HasRendererText {
        private HighlightPattern highlightPattern;
        @Override public void setColor(Color c) { highlightPattern.setColor(c);}
        @Override public Color getColor() { return highlightPattern.getColor();}
        @Override public int getVisibleWhere() {return highlightPattern.getVisibleWhere();}
        @Override public int getStyle() {return highlightPattern.getStyle();}
        @Override public String getRendererText(int options) { return highlightPattern.patternText;}
        private boolean[] selRes,selNT, selAAfromNT;
        private byte[] selRes_L, selNT_L;
        private String nt_lastText;
        private int getSelectedNucleotides_modificationCount;
        private byte[] resultLength;
        public int getSelectedNucleotidesOffset() { return 0;}
        public byte[] getMatchLenghts() { return selRes_L;}
        private boolean[] getMatches(String haystack, String needles[], boolean buffer[], final byte[] bufferMatchLength, boolean clearBuffer) {
            final boolean bb[]=redim(buffer, haystack.length(),9);
            if (clearBuffer && bb==buffer) Arrays.fill(bb,false);
            final byte[] resL=this.resultLength=redim(bufferMatchLength, haystack.length(), 9);
            for(String needle : needles) {
                try {
                    final Pattern pattern=Pattern.compile(needle);
                    final java.util.regex.Matcher matcher=pattern.matcher(haystack);
                    while(matcher.find()) {
                        final int start=matcher.start();
                        bb[start]=true;
                        resL[start]=(byte)mini(127,matcher.end()-start);
                    }
                }catch(Exception e){ dialog.setError("Incomplete regular expression:   \""+needle+"\"");}
            }
            return bb;
        }
        private boolean[] getMatches(final byte haystack[], final int length, final byte needles[][], final boolean[] buffer, final byte[] bufferMatchLength, boolean clearBuffer) {
            if (haystack==null || needles==null) return NO_BOOLEAN;
            final boolean bb[]=redim(buffer, haystack.length ,9);
            final byte[] resL=this.resultLength=redim(bufferMatchLength, haystack.length, 9);
            if (clearBuffer && bb==buffer) Arrays.fill(bb,false);

            for(byte[] needle : needles) {
                final int firstLetter=nxt(LETTR,needle);
                if (firstLetter<0) continue;
                final int needle0=needle[firstLetter];
                nextPos:
                for(int i=mini(haystack.length,length)-needle.length+1; --i>=0; ) {
                    if (needle0!= (haystack[i+firstLetter]&(~32))) continue nextPos;
                    for(int j=needle.length; --j>=0;) if (needle[j]!='.' && needle[j] != (haystack[i+j]&(~32)))  continue nextPos;
                    bb[i]=true;
                    resL[i]=needle.length<127 ? (byte) needle.length : 127;
                }
            }
            return bb;
        }

        public boolean[] getSelectedNucleotides() {
            final ProteinNT protein=(ProteinNT)getProtein();
            final int orientations=dialog.getOrientations();
            final boolean
                isForward=(orientations&1)!=0,
                isReverseComplement=(orientations&2)!=0;
            if (patternText.length()==0 || !isNucleotide() || !isReverseComplement&&!isForward) return NO_BOOLEAN;
            final byte[] nts=protein.getNucleotides();
            protein.getResidueType();
            final int countNT=protein.countNucleotides();
            if (nts==null) return null;
            final int mc=
                protein.getModificationCount(ProteinModiCount.MC_RESIDUE_TYPE)+
                protein.getModificationCount(ProteinModiCount.MC_NUCLEOTIDE_TYPE)+(orientations<<8);

            if (selNT==null || nt_lastText!=patternText || getSelectedNucleotides_modificationCount!=mc) {
                getSelectedNucleotides_modificationCount=mc;
                nt_lastText=patternText;
                final boolean isRegex=isRegex(patternText);
                if (isReverseComplement) {
                    if (isRegex) {
                        selNT=getMatches(protein.getNucleotidesReverseComplementAsStringUC(),sPatterns,selNT, selNT_L, true);
                    } else {
                        DNA_Util.reverseComplementDestructive(nts,countNT);
                        selNT=getMatches(nts,countNT,patterns, selNT, selNT_L, true);
                        DNA_Util.reverseComplementDestructive(nts,countNT);
                    }
                    selNT_L=resultLength;
                    reverseArray(selNT,0,countNT);
                }
                if (isForward) {
                    selNT=  isRegex ?
                        getMatches(protein.getNucleotidesAsStringUC(),sPatterns,selNT, selNT_L, !isReverseComplement) :
                        getMatches(nts,countNT,patterns, selNT, selNT_L, !isReverseComplement);
                }
                selNT_L=resultLength;
                selAAfromNT=protein.nucleotideSelection2aminoAcidSelection(selNT,0);
            }
            return selNT;
        }
        private int getResidueType_modificationCount;
        private String aa_lastText;
        @Override public boolean[] getSelectedAminoacids() {
            if (patternText.length()==0) return NO_BOOLEAN;
            final StrapProtein protein=(StrapProtein)getProtein();
            if (protein==null) return NO_BOOLEAN;
            if (isNucleotide) {
                if (protein.countNucleotides()>0) { getSelectedNucleotides(); return selAAfromNT;}
                else return NO_BOOLEAN;
            }
            final byte[] aa=protein.getResidueType();
            final int modiCount=protein.getModificationCount(ProteinModiCount.MC_RESIDUE_TYPE);
            if (selRes==null || aa_lastText!=patternText || getResidueType_modificationCount!=modiCount) {
                getResidueType_modificationCount=modiCount;
                Protein.incrementModificationCount(ProteinModiCount.MC_RESIDUE_SELECTIONS, protein);

                dialog.setError("");
                final boolean selected[]=isRegex(patternText) ?
                    getMatches(protein.getResidueTypeAsStringUC(),sPatterns,selRes, selRes_L, true) :
                    getMatches(aa,protein.countResidues(), patterns, selRes, selRes_L, true);
                selRes_L=resultLength;
                aa_lastText=patternText;
                selRes=selected;

            }
            return selRes;
        }
        private boolean isRegex(String ex) {
            for(int i=ex.length(); --i>=0;) {
                final char c=ex.charAt(i);
                if (c!='.' && c!=' ' && (c<'A' || c>'Z'))  return true;
            }
            return false;
        }

    }
}
