package charite.christo.strap;
import charite.christo.interfaces.*;
import charite.christo.*;
import charite.christo.protein.*;
import java.awt.*;
import javax.swing.*;
import java.io.File;
import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP
   The file selector shows the protein files of the current project
   directory. They are ordered in different lists according the file endings.

   Protein files can be selected and loaded into STRAP by double
   clicking or with the button
   <i>BUTTON:DialogProjectFiles#BUT_LAB_Load</i>.  Multiple selections
   are possible  with the keys <b>CTRL</b> (same as STRG) and
   <b>SHIFT</b>.

   <br><br>

   <b>Drag-And-Drop:</b> The file selector is capable of WIKI:Drag_and_drop:

   Files from the native file browser or  desktop can be dragged into the
   file selector or into  to the STRAP alignment panel (see
   See http://www.bioinformatics.org/strap/dragProteinLink.html).

   <br><br><b>Alternative directories:</b> In the choice menu at the top
   a number of other folders can be chosen.  For example a
   local mirror of the PDB-file collection may given.

  <br><br><b>Flags:</b> The flags <b>g</b>,<b>s</b>,<b>3</b> and <b>n</b> which precede the file names indicate
   the existence of gaps, residue-annotations, 3D coordinate transformations and nucleotide reading
   frames, respectively.

   <br><br><b>Multiple sequence files</b> with the ending .msf or .clustalw contain the aligned sequences of several proteins: WIKI:Clustalw: WIKI:NexusFormat.

   <br><br><b>Customization:</b> The list of file endings and the list of alternative directories can be altered.

   <br><br><b>File formats:</b>

   Many <i>CLASS_REF:ProteinParser</i>file formats are recognized: WIKI:Swissprot, WIKI:Protein_Data_Bank, WIKI:Fasta, WIKI:Genbank, EMBL.

   See http://www.molecularevolution.org/mbl/resources/fileformats/.

   The file text is analyzed and the file type identified without  considered the name of the file.
   If no appropriate protein file type is detected then the sequences is formed from the
   letters in the file.

   <i>SEE_CLASS:ProteinParser</i>

   @author Christoph Gille
*/
public class MainDialogProjectFiles extends AbstractDialogJPanel implements  HasControlPanel,java.awt.event.ActionListener{

    @Override public void dispose() {
        for(Object d:hashMap.values().toArray()) ((DialogProjectFiles)d).dispose();
        hashMap.clear();
    }
    private JComponent
        visibleDialog;

    public MainDialogProjectFiles() {
        setSettings(this, Customize.proteinFileExtensions, Customize.fileBrowsers, Customize.file_directories);
        final String msg=
            "The project direktory is denoted by dot slash \"./\".<br><br>"+
            "With Drag-and-Drop, files can be dragged from the Desktop or from any other location.<br><br>"+
            "<b>Advanced users:</b> You can implement other directories such as a local copy of the PDB using the customize button <i>ICON:charite.christo.ChConstants#IC_CUSTOM</i>.";
        final JComponent
            butHelpOtherDir=smallHelpBut(msg),
            pNorth=newCNSEW(comboDir(), StrapAlign.headLine(this),null,butHelpOtherDir,null),
            b2=StrapAlign.b(CLASS_DialogProteinList).t("Load proteins in list"),
            butFileBrowser=new ChButton("FB").t(isWindows() ? "Open Explorer" : isMacintosh() ? "Open Finder" : "Open native file browser").li(this)
            .tt("Use Drag and Drop to load proteins from other file directories into STRAP."),
            butMFA=StrapAlign.b(CLASS_DialogImportMFA)
            .tt("Also try to load multiple sequence files directly with Drag-and-Drop"),
            pSouth=pnl("#TB Further options",new GridLayout(1,4),butFileBrowser,butMFA,b2, WATCH_MOVIE+MOVIE_Load_Proteins);
        cnsew(this,null,pNorth,pSouth,null,null);
        putDialog();
    }
    private ChCombo comboDir_;
    private ChCombo comboDir() {
        if (comboDir_==null) comboDir_=Customize.customize(Customize.file_directories).newComboBox().li(this);
        return comboDir_;
    }
    private void putDialog() {
        try { rmFromParent(visibleDialog);} catch(Exception e){}
        String dir=comboDir().toString();
        if (dir.indexOf(' ')>0) dir=dir.substring(0,dir.indexOf(' '));
        visibleDialog=dialogProjectFiles(file(dir));
        add(visibleDialog,BorderLayout.CENTER);
        revalidate();
    }
    private final Map<File,DialogProjectFiles> hashMap=new HashMap();
    private Container controlPanel;
    public Container getControlPanel() {
        if (controlPanel==null) controlPanel=vBpnl(StrapAlign.button(StrapAlign.TOG_SKIP_IDENTICAL_SEQUENCE).cb());
        return controlPanel;
    }
    private DialogProjectFiles dialogProjectFiles(File directory) {
        DialogProjectFiles o=hashMap.get(directory);
        if (o==null) {
            o= new DialogProjectFiles(new StrapFileCellRenderer(null),directory);
            hashMap.put(directory,o);
        }
        return o;
    }
    public void actionPerformed(java.awt.event.ActionEvent ev) {
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        final int modi=ev.getModifiers();
        if (q==comboDir_) putDialog();
        if (cmd=="FB") {
            final String s=comboDir().toString();
            final File f=file(s.indexOf(' ')>0 ? s.substring(0,s.indexOf(' ')) : s);
            final Customize cust=Customize.customize(Customize.fileBrowsers);
            final ChExec ex=new ChExec(0).setCommandsAndArgument(cust.getSettings(),f);
            if ( (modi&java.awt.event.ActionEvent.CTRL_MASK)!=0) ex.log(cust);
            new Thread(ex).start();
        }
    }
}
