package charite.christo.strap;
import charite.christo.protein.*;
import charite.christo.interfaces.*;
import charite.christo.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP
   A set of proteins is generated from the selected proteins<br>
   where the sequence identity ratio is not higher than the given threshold.<br><br>

   Proteins are excluded from the set for which another very similar protein exists.<br>
   This procedure is rather primitive and more sophisticated programs for this task exist (E.g. nrdb90).<br><br>
   Because it is aligning each sequence against all others by ClustalW the time rises with the square of the number of sequences.<br><br>
   Enter the threshold ( 0 ... 1 ).<br>
   The larger the number, the more sequences  will be returned.

   @author Christoph Gille
   see also http://www.ebi.ac.uk/~holm/nrdb90/nrdb90.pl.txt

*/
public class MakeNonRedundantSequenceSet implements ChRunnable  {
    private static StrapProtein[] nonredundantSequenceSet(StrapProtein pp0[],float threshold) {
        final StrapProtein pp[]=pp0.clone();
        final boolean done[][]=new boolean[pp.length][pp.length];
        final byte[] bothResTypes[]=new byte[2][];
        for(int i=0;i<pp.length;i++) {
            for(int j=0;j<pp.length;j++) {
                if (i==j || pp[i]==null||pp[j]==null || done[i][j] || done[j][i] ) continue;
                done[i][j]=true;
                final byte[] s0=bothResTypes[0]=pp[i].getResidueTypeExactLength();
                final byte[] s1=bothResTypes[1]=pp[j].getResidueTypeExactLength();
                final int maxLen=maxi(s0.length,s1.length);
                if (maxLen==0) continue;
                final SequenceAligner aligner= StrapPlugins.instanceSA(charite.christo.strap.extensions.MultipleAlignerClustalW.class,0,true);
                aligner.setSequences(bothResTypes);
                aligner.compute();
                final byte aligned[][]=aligner.getAlignedSequences();
                if (sze(aligned)<2) continue;
                final Ratio r=CrudeSequenceSimilarity.getIdentity(aligned[0],aligned[1],0,Integer.MAX_VALUE);
                final float f=r.getNumerator()/(float)maxLen;
                if (f>threshold) {
                    if (countLetters(s0,0,MAX_INT) > countLetters(s1,0,MAX_INT)) pp[j]=null; else pp[i]=null;
                }
            }
        }
        return rmNullA(pp,StrapProtein.class);
    }
    private float threshold;
    public Object run(String id, Object para) {
        final StrapProtein pp[]=(StrapProtein[])para;
        if (id=="GO") {
            StrapProtein ppNR[]=nonredundantSequenceSet(pp,threshold);
            ChDelay.invokeLater(this,"EDT",ppNR);
        }
        final String title="Non-redundant List";
        if (id=="EDT") StrapAlign.showInJList(0L, null, pp, title, null).showInFrame(ChFrame.AT_CLICK, title);

        if (id=="START") {
            StrapAlign.errorMsg("");
            if (ChMsg.noProteinsSelected(pp)) return null;
            final String msg=getHlp(MakeNonRedundantSequenceSet.class);
            final float th=(float)atof(ChMsg.input(0L, msg,7,"0.8"),0);
            if (th<=0 || th>=1)  StrapAlign.errorMsg("The value must be in the range of 0 to 1");
            else {
                final MakeNonRedundantSequenceSet instance=new MakeNonRedundantSequenceSet();
                instance.threshold=th;
                ChDelay.newThread(instance,"GO",pp).start();
            }
        }
        return null;
    }
}
