package charite.christo.strap;
import charite.christo.protein.*;
import charite.christo.interfaces.*;
import charite.christo.*;
import java.awt.*;
import javax.swing.*;
import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**
Underlining residues with the mouse.
   @author Christoph Gille
*/
public class MouseSelection extends AbstractResidueSelection implements HasRendererComponent,Runnable {

    {
        setColor(Color.CYAN);
        setVisibleWhere(0xffffffff);
        setStyle(VisibleIn123.DOTTED);
        setName(STANDARD);
    }
    private int lastFrom,lastTo;

    public void set(final Protein protein, final int col1, final int col2, boolean[] selRes) {
        final StrapProtein p=(StrapProtein)protein;
        final Protein oldP=getProtein();
        setVisibleWhere(getVisibleWhere()|VisibleIn123.SEQUENCE);
        if (oldP!=p) {
            if (oldP!=null) oldP.removeResidueSelection(this);
             StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,22);
        }
        super.setProtein(p);
        if (p==null) return;
        p.addResidueSelection(this);
        final int nR=p.countResidues(), from, to;
        if (selRes==null) {
            final int colMin=col1<col2 ? col1 : col2;
            final int colMax=col1>col2 ? col1 : col2;

            from=p.column2nextIndex(colMin);
            final int to1=p.column2nextIndex(colMax);
            to=to1<0 ? nR : to1;
            super.setSelectedRange(from,to);
        } else {
            super.setSelectedAminoacids(selRes, 0);
            from=firstTrue(selRes);
            to=lastTrue(selRes)+1;
        }
        if (lastFrom!=from || lastTo!=to){
            lastFrom=from;
            lastTo=to;
            StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,22);
            if ((getVisibleWhere()&VisibleIn123.STRUCTURE)!=0) {
                if (from>=0 && from<to && to<=nR) toClipbd(toStrg(p.getResidueTypeExactLength(),from,to));
                ChDelay.runAfterMS(EDT,this,222);
            }
        }
    }
    public void run() {
      final Protein p=getProtein();
      if (p!=null) V3dUtils.residueSelectionsTo3D(V3dUtils.RSto3D_NOT_SAME_WIRE|V3dUtils.RSto3D_SET_FOCUSED_PV, new ResidueSelection[]{MouseSelection.this}, p.getProteinViewers());
    }
    private final ChLabel RENDERER_COMPONENT=new ChLabel();
    public JComponent getRendererComponent() {
        final Protein p=getProtein();
        if (p==null) RENDERER_COMPONENT.t("");
        final int from=getSelectedAminoacidsOffset();
        final int to=from+lastTrue(getSelectedAminoacids());
        final int num[]=p.getResidueNumber();
        final byte chain[]=p.getResidueChain();
        String s="Mouse selection "+((p==null || num==null) ? null : " ("+get(from,num)+":"+chrAt(from,chain)+" to "+get(to-1,num)+":"+chrAt(to-1,chain)+")");
        return RENDERER_COMPONENT.t(s);
    }
}
