package charite.christo.strap;
import charite.christo.*;
import charite.christo.protein.*;
import charite.christo.interfaces.*;
import javax.swing.*;
import java.awt.Component;
import java.util.*;
import javax.swing.event.*;
import java.awt.event.*;
import java.io.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static charite.christo.protein.Protein3dUtils.*;
import static charite.christo.strap.V3dUtils.*;
import static charite.christo.protein.ProteinViewer.*;
/**
   The menu for protein viewers (Pymol, JMol ... ).
   @author Christoph Gille
*/
public class ProteinViewerPopup implements ActionListener, IsEnabled, ChRunnable {
    private final static String
        BUT_GEN_MENUS="PVU$$KGM", MENU_OTHER_BAR="   \u25baOther-menubar",  KEY_WORD0="PVP$$w0";
    final static String TOG_SEND[]={"Send init script","Send 3D-commands attached to residue selections"};
    private final ContextObjects objects;
    private final ChLabel
    labTitle=new ChLabel().cp(KOPT_NOT_FOR_MENUBAR,""),
        labContext=new ChLabel("Note: This 3D-View has a context menu (Right click the 3D-pane)")
        .cp(KOPT_HIDE_IF_DISABLED,"").fg(0xAA)
        .cp(KEY_ENABLED,this);
    private final boolean isImmutable;
    public ProteinViewerPopup(boolean isImmutable, ContextObjects o) {
        this.isImmutable=isImmutable;
        objects=o;
    }
    /* ---------------------------------------- */
    /* >>> ActionListener  >>> */
    public void actionPerformed(ActionEvent ev) {
        final String cmd=ev.getActionCommand();
        final Object q=ev.getSource();
        if (q==null) return;
        final int modi=modifiers(ev);
        final boolean isSelctd=isSelctd(q), shift=0!=(modi&MouseEvent.SHIFT_MASK);
        Protein3d p3d=null;
        final List<Protein> vAllPP=new ArrayList();
        for(ProteinViewer pv : objects.proteinViewers()) {
            final Protein p=p(pv);
            if (p==null) continue;
            final Class clazz=pv.getClass();
            final Component panel=viewerCanvas(pv);
            if (panel instanceof Protein3d) p3d=(Protein3d)panel;
            final Protein pp[]=getProteinsInView(pv);
            adAllUniq(pp,vAllPP);
            if (cmd=="SUPER_GO") {
                StrapAlign.dispatchLater(StrapEvent.COMPUTATION_STARTED,333);
                final String msg="It seems that a superposition job is already running.<br>Start new job anyway?";
                if (sze(pp)<2) error("Superposition requires at least \n"+
                                     "two proteins in the same 3D-view.\n\n"+
                                     "But only "+sze(pp)+" protein  is in this view\n\n"+
                                     (viewerCanvas(pv)==null?"":
                                      "Use WIKI:Drag_and_drop to load\n"+
                                      "at least one more protein.\n\n"+
                                      "Watch movie "+MOVIE_Drag_protein_to_3D));
                else if (!StrapProteinUtils.IS_SUPERIMPOSING[0] || ChMsg.yesNo(msg)) {
                    final ArrayList v=new ArrayList();
                    for(int i=pp.length; --i>=0;) {
                        final Matrix3D m3d=p.getRotationAndTranslation();
                        pcp("SUPER_UNDO",m3d!=null ? m3d : new Matrix3D(),pp[i]);
                        if (pp[i].getResidueCalpha()!=null) v.add(pp[i]);
                    }
                    StrapProteinUtils.threadSuperimposeProteins(v,CAN_BE_STOPPED, StrapProteinUtils.SUPERIMP_EVENT).start();
                }
            }
            for(int i=pp.length; --i>=0;) {
                if (cmd=="SUPER_UNDO" || cmd=="SUPER_HOME") {
                    pp[i].setRotationAndTranslation(cmd=="SUPER_HOME" ? null : gcp("SUPER_UNDO",pp[i],Matrix3D.class));
                    new StrapEvent(Protein3d.class,StrapEvent.PROTEIN_3D_MOVED).setParameters(pp).run();
                }
            }
            if (cmd=="SUPER_STOP") stopAllAlignments();
            if (cmd==CLASS_DialogNeighbors) StrapProteinUtils.structureNeighbors(pp);
            if (cmd==IC_CAMERA) {
                final File f=file(STRAPOUT+"/images/"+niceShrtClassNam(pv)+".png");
                writePng(panel,f);
                visitURL(f, modi);
            }
            if (cmd=="INFO") showHelp(clazz==Protein3d.PView.class ? Protein3d.class : clazz);
            if (cmd=="H") showHelp(V3dUtils.class);
            if (cmd==IC_JAVA) showJavaSource(clazzName(pv),(String[])null);
            if (cmd=="CTRL") ChFrame.frame("Ctrl "+pv.getProtein()+" @ "+niceShrtClassNam(pv), getCtrlPanel(pv), 0L).shw(ChFrame.AT_CLICK);
            if (cmd=="CUST") for(Customize c : Customize.cc(pv)) Customize.addDialog(c);
            if (cmd=="LOG" || cmd=="LOGg" || cmd=="SAVE") showLogPanel(cmd!="LOG", pv);
            if (cmd=="LOAD") loadSessionI();
            if (cmd==JFileChooser.APPROVE_SELECTION) loadSession(SESSION_LOAD_PROTS, pv,fs.getSelectedFile());
            if (cmd=="ADD") ChMsg.msgDialog(0L, vBpnl("To add proteins to the view please use Drag-and-Drop",
                                                      "Simply drag a protein label or protein file into the view.",
                                                      WATCH_MOVIE+MOVIE_Drag_protein_to_3D));
            if (cmd=="EXPORT") {
                Object msg=gcp(cmd,this);
                if (msg==null) {
                    msg=vB(
                           "<br><b><u>Export using Drag-and-Drop</u></b><br>"+
                           "To export proteins use Drag-and-Drop:<br>"+
                           "Start dragging the mouse on a protein label and drop the protein on the desktop.<br>"+
                           "Protein labels contain the protein name and maybe a protein icon.<br>"+
                           "They are found in the alignment panel or with above menu item \"List proteins\".",
                           WATCH_MOVIE+MOVIE_Export_Proteins,
                           "<br><b><u>Export using the context menu</u></b><br>"+
                           "Alternatively, you can use the export-function in the context menu.",
                           StrapAlign.b(CLASS_DialogExportProteins),
                           WATCH_MOVIE+MOVIE_Context_Menu);
                    pcp(cmd,msg,this);
                }
                ChFrame.frame(cmd, msg, CLOSE_ON_CtrlW).shw(ChFrame.AT_CLICK|ChFrame.PACK);
            }
            if (cmd=="ADD_DIALOG") StrapAlign.addDialogSetClass(CLASS_Dialog3DViewer,pv.getClass());
            final int iList=idxOf(cmd,V3dListModel.TITLES);
            if (iList>=0) {
                final ChJList jl=V3dListModel.jList(iList,true,1);

                if (iList==V3dListModel.HET || iList==V3dListModel.NUCL || idxOf(iList, V3dListModel.TYPES_SEL)>=0) {
                    final Object info=new ChLabel(INFO_HETERO_NOT_SELECTABLE).cp(KOPT_HIDE_IF_DISABLED,"");
                    pcp(KEY_SOUTH_PANEL,info,jl);
                    vEnableDisable.add(info);
                }
                jl.showInFrame(ChFrame.AT_CLICK|ChFrame.SIZE_REGARD_RESOLUTION,cmd);
                enableDisable();
            }
            if (cmd=="STRAP_MB") { StrapAlign.setMenuBar(null, 0,0); StrapAlign.setToolpane(null); }
            if (cmd==BUT_GEN_MENUS  || cmd==BUT_NATIV_MENUS) {
                pcp(KEY_WHICH_MENU,cmd, viewerCanvas(pv));
                setPV(PV_FOCUSED, pv);
            }
            if (cmd=="RM") {
                final ProteinViewer vv[]=getViewersSharingView(pv), vvDel[]=new ProteinViewer[vv.length];
                int iDel=0;
                if (vv.length==1) {
                    if (ChMsg.yesNo("Close 3D-viewer "+p(vv[0]))) vvDel[iDel++]=vv[0];
                } else if (vv.length>0) {
                    final List<Protein> vP=new ArrayList();
                    for(ProteinViewer v : vv) vP.add(p(v));
                    final ChJList list=new ChJList(vP,ChJTable.DEFAULT_RENDERER);
                    if (ChMsg.yesNo(newCNSEW(list,"Please select the proteins to be removed.","Remove proteins from 3D-view?",null,null))) {
                        final Object ppDel[]=list.getSelectedValues();
                        for(ProteinViewer v : vv) if (cntains(p(v),ppDel)) vvDel[iDel++]=v;
                    }
                }
                if (iDel>0) disposeV(vvDel);
            }
            if (cmd=="PV" || cmd=="WIRE_DEFAULT_V" || cmd==BUT_DEFAULT_VIEWER) {
                final Class c3d=HotswapUtil.getClass(cmd==BUT_DEFAULT_VIEWER ? StrapAlign.getDefault(ProteinViewer.class) :  gcp(ProteinViewer.class,q));
                if (sze(pp)>5 && !ChMsg.yesNo("Are you sure that you want to see "+sze(pp)+" proteins in "+niceShrtClassNam(c3d)+
                                              ".<br>Showing too many protein 3D-structures may overload the Java session.")) continue;
                final long options=(b(TOG_SEND[1]).s() ? OPEN_SEND_ALL_ANNOTATIONS:0) | (b(TOG_SEND[0]).s() ? OPEN_INIT_SCRIPT:0);
                final ProteinViewer pvNew=c3d!=null ? open3dViewer(options, pp, c3d) : null;
                if (pv instanceof Protein3d.PView) {
                    final ProteinViewer vvNew[]=getViewersSharingView(pvNew);
                    boolean isPreview=false;
                    for(int pass=0;pass<2; pass++) {
                        for(ProteinViewer v0 : getViewersSharingView(pv)) {
                            final Protein p0=p(v0);
                            final Matrix3D m3d=((Protein3d.PView)v0).getTransformationPreview();
                            for(ProteinViewer v : vvNew) {
                                if (v==null || p(v)!=p0) continue;
                                if (pass==0) isPreview=isPreview||(m3d!=null && !m3d.isUnit());
                                else if (isPreview) {
                                    V3dUtils.transform3D(m3d,v);
                                    pcp(KOPT_IS_PREVIEW,"",v);
                                }
                            }
                        }
                    }
                } else {
                    final Runnable run=ChDelay.newThread(this,"CPY_SESSION", new Object[]{log3D(true, pv,""),weakRef(pvNew)});
                    final Object
                        b1=new ChButton("Apply?").r(run).actionClose(CLOSE_PARENT_WINDOW),
                        b2=new ChButton("Cancel").actionClose(CLOSE_PARENT_WINDOW),
                        pnl=vBhB(
                                 "You opened a new 3D-view<br>"+
                                 "Also apply the rendering styles<br>"+
                                 "from the previous 3D-view?",
                                 pnl(b1,b2));
                    new ChFrame("Yes-No").ad(pnl).shw(ChFrame.PACK|ChFrame.AT_CLICK|ChFrame.ALWAYS_ON_TOP);
                }
            }
            if (cmd=="SCRIPT") showScriptPanel();
            if (cmd=="SEL_ATOM")  openDialogSelect(SELECT_ATOMS);
            if (cmd=="SEL_ATOMg") openDialogSelect(SELECT_ATOMS_G);
            if (cmd=="SEL_AMINO") openDialogSelect(SELECT_AMINO);
            { /* --- Selections --- */
                Object select=chrAt(0,cmd)=='$' || chrAt(0,cmd)=='-'&&chrAt(1,cmd)=='9' ? cmd : null;
                if (cmd=="SEL_I") {
                    final boolean bb[]=selectionToBoolean(getSelA(pv), p);
                    for(int i=sze(bb); --i>=0;) bb[i]=!bb[i];
                    select=bb;
                }
                if (cmd=="SEL_H" ||  cmd=="SEL_E" ||  cmd=="SEL_HE") {
                    final byte ss[]=p.getResidueSecStrType();
                    if (sze(ss)>0) {
                        final boolean bb[]=new boolean[ss.length];
                        for(int i=ss.length; --i>=0;) bb[i]=strchr(0L, (char)ss[i], cmd, 4, MAX_INT)>=0;
                        select=bb;
                    }
                }
                final Selection3D ss3D[]=
                    select instanceof boolean[] ? toSelectionBB((boolean[])select,p) :
                    select instanceof String ? toSelectionT((String)select, p):
                    null;
                if (ss3D!=null) {
                    setPV(PV_FOCUSED,pv);
                    setSelAminos(0L, ss3D, "",  pv);
                }
            }
            final String word0=gcps(KEY_WORD0, q);
            if (word0!=null) sendCommand(0L,  shift&&equls(word0,cmd,0) ? rplcToStrg(" on", " off", cmd) : word0, pv);
        }/* for(ProteinViewer pv */
        if (cmd=="CUST_INIT") {
            for(Object o : StrapPlugins.allClassesV(ProteinViewer.class).asArray()) {
                Customize.addDialog(Customize.getCustomizableForClass(HotswapUtil.getClass(o), Customize.CLASS_InitCommands));
            }
        }
        if (cmd==ACTION_BECOME_INVISIBLE) { /* Otherwise the menus are not reacting */
            ChDelay.runAfterMS(EDT,ChDelay.newThread(this, "SET_PV",null), 99);
            //if (popup!=null) popup.setVisible(false);
        }
        if (cmd=="SHOW_HETERO") {
            final String s="This can be done in the tool-panel.<br>"+
                "It is opened with the \"List proteins\"-toggle in the lower right corner.";
            ChMsg.msgDialog(0L, s);
        }
        if (cmd=="MBT") {
            final List vID=new ArrayList();
            for(Protein p : toArry(vAllPP,Protein.class)) adUniq(withoutLastComponent(delPfx("PDB:",p.getPdbID()),'_'),vID);
            if (sze(vAllPP)==0) error("No PDB ID");
            else {
                final ChCombo comboPdbID=new ChCombo(0L,vID);
                final Object msg=vBhB("Starting MBT viewer. see",
                                      "http://mbt.sdsc.edu/",
                                      "This requires a modern computer and fast internet.",
                                      "PCs with older video cards may crash.",
                                      "Very large proteins may cause problems.",
                                      hB("PDB ID ", comboPdbID));
                final int choice=ChMsg.option(0L, msg,new String[]{"Simple Viewer", "Protein Workshop","Cancel"});
                final String pdbID=toStrg(comboPdbID), view=choice==0?"SV":choice==1?"PW":null;
                if (view!=null) visitURL("http://www.pdb.org/pdb/Viewers/RCSBViewers/view.jsp?viewerType="+view+"&structureId="+pdbID+"&structIdFromStrutsAction="+pdbID,0);
            }
        }
        if (p3d!=null) {
            if (cmd=="WIRE_DIFF_COLORc" || cmd=="WIRE_DIFF_COLOR") {
                p3d.eachProteinInDifferentColor(cmd=="WIRE_DIFF_COLORc",false);
                p3d.enableDisable();
            }
            if (cmd=="WIRE_RESET")  { p3d.getRotationAndTranslationOfView().unit().scale(5f); p3d.repaint(); }
            if (cmd=="WIRE_HIDE_DIST_HET") p3d.setHideDistantDNS(isSelctd);
        }

    }
    /* <<< ActionEvent <<< */
    /* ---------------------------------------- */
    /* >>> Menu >>> */
    Object[] m(char id) {
        if((id|32)=='b') {
            final boolean extra=LAFChooser.isUseScreenMenuBar();
            return new Object[]{
                MENU_OTHER_BAR,
                extra ? null : b("STRAP_MB"),
                b(id=='b'?BUT_GEN_MENUS:BUT_NATIV_MENUS),
                labContext
            };
        }
        if (id=='S') {
            return
                new Object[]{
                "Superimpose",
                b("SUPER_GO"), b("SUPER_STOP"), b("SUPER_UNDO"), b("SUPER_HOME"),
                b(CLASS_DialogNeighbors).i(IC_SEARCH),
                KOPT_DISABLE_IF_CHILDS_DISABLED
            };
        }
        if (id=='T') {
            return new Object[]{
                "Tools",
                b(COMMANDcenter),
                b(IC_CAMERA),
                "-",
                b("CTRL"),
                b("SCRIPT"),
                b(COMMANDshowScriptPanel),
                b("CUST"),
                "-",
                b(CLASS_DialogNeighbors),
                menuForAccelerators(),
            };
        }
        if (id=='P') return new Object[]{"Protein", b("ADD"), b("ADD_DIALOG"), b(V3dListModel.TITLES[V3dListModel.PROTEINS]), b("EXPORT"),b("RM")};
        if (id=='H') return new Object[]{"Help",b("INFO"),b("H")};
        if (id=='v') {
            final Object vv[]=StrapPlugins.allClassesV(ProteinViewer.class).asArray(),
                ii[]=new Object[vv.length+9];
            int j=0;
            if (vv.length>0) {
                ii[j++]="Alternative protein Viewers";
                ii[j++]="The proteins are opened in an external viewer";
                ii[j++]="i";
                ii[j++]=IC_JMOL;
                for(int i=vv.length; --i>=0;) {
                    final Object c=vv[i];
                    if (c==null) continue;
                    ii[j++]=new ChButton("PV").li(this)
                        .t("Open in "+niceShrtClassNam(c)).i(TabItemTipIcon.imageIcon(c))
                        .cp(ProteinViewer.class,c);
                };
                if (!Insecure.EXEC_ALLOWED) ii[j++]=b("MBT").t("   \u25baMBT ...");
                ii[j++]=new Object[]{
                    "Options for start",
                    b("CUST_INIT"),
                    b(TOG_SEND[0]),
                    b(TOG_SEND[1])
                };
            }
            return ii;
        }
        return null;
    }
    /* <<< Menu <<< */
    /* ---------------------------------------- */
    /* >>> Button >>> */
    private final Map<String,ChButton> mapButtons=new HashMap();
    ChButton b(String id) {
        final ProteinViewer vv[]=isImmutable ? objects.proteinViewers():null;
        if (sze(vv)==1 && isEnabledPV(id, vv[0])==Boolean.FALSE) return null;
        ChButton b=mapButtons.get(id);
        if (b==null) {
            String t=null, tt=null, clazz3D=null, word0=null;
            Object i=null;
            boolean toggle=false, select=false;
            final String niceName=sze(vv)==1?niceShrtClassNam(vv[0]):null;
            final int iList=idxOf(id, V3dListModel.TITLES);
            if (id=="SEL_ATOM")    t="Specific for this 3D-program ...";
            else if (id=="SEL_ATOMg")   t="Generic ...";
            else if (id=="SEL_AMINO") { t="Type selections in text-box ..."; i=IC_EDIT; }
            else if (id=="SEL_H")     { t="Select alpha-helices"; i=IC_HELIX; }
            else if (id=="SEL_E")     { t="Select beta-sheets";   i=IC_SHEET ; }
            else if (id=="SEL_HE")      t="Select alpha-helices and beta-sheets";
            else if (id=="SEL_I")       t="Invert selection";
            else if (id=="-99-$NTERM")  t="Select hidden N-terminus";
            else if (id=="$CTERM-9999") t="Select hidden C-terminus";
            else if (id=="$ALL")        t="Select everything";
            else if (id=="$FIRST-$LAST")t="Select all amino acids";
            else if (id==BUT_NATIV_MENUS) t="Native menu-bar of this particular 3D-viewer";
            else if (id==BUT_GEN_MENUS) t="Generic menu-bar for 3D-viewers in STRAP";
            else if (id=="STRAP_MB") t="Menu-bar of the Strap program (Or simply click into alignment)";
            else if (id=="CTRL")   { t="Control panel"; i=IC_CONTROLPANEL; }
            else if (id=="CUST") { t="Customize"; i=IC_CUSTOM; }
            else if (id=="LOG")    t="Native commands ";
            else if (id=="LOGg")   t="Generic commands ";
            else if (id=="SAVE") { t="Save session"; i=IC_DIRECTORY; }
            else if (id=="LOAD") { t="Load session"; i=IC_DIRECTORY; }
            else if (id=="INFO") t="Information on "+(niceName!=null?niceName:"this Protein viewer");
            else if (id=="H") t="Using protein 3D viewers in STRAP";
            else if (id==IC_JAVA) { t="Java source"; i=IC_JAVA; }
            else if (id==IC_CAMERA) { t="Capture image (print)"; i=IC_CAMERA; }
            else if (id=="RM") { t="Remove proteins / close view ...";  i=IC_CLOSE; }
            else if (iList>=0)  {
                t=id;
                i=iList==V3dListModel.HET ? IC_FLAVIN : iList==V3dListModel.NUCL? IC_DNA : null;
            }
            else if (id=="ADD" || id=="ADD_DIALOG") { t="Add proteins to the view"; if (id=="ADD") i=IC_CURSOR_DRAG; }
            else if (id=="EXPORT") { t="Export proteins"; i=IC_CURSOR_DRAG; }
            else if (id=="SUPER_GO") {
                t="Superimpose all proteins in this view";
                tt="All structures will be superimposed upon the optimal structure (OS)<br>"+
                    "The OS is the one where the sum of scores to the others is maximal.<br><br>"+
                    "To superimposed only fewer proteins,  open another 3D-view with only the proteins under consideration.<br><br>"+
                    "The Computation is interupted when the view is closed.<br>"+
                    "An 'Undo' button will appear";
                i=IC_SUPERIMPOSE;
            }
            else if (id=="SUPER_STOP") { t="Stop all superposition processes"; i=IC_STOP; }
            else if (id=="SUPER_UNDO") { t="Reset coordinates to what it was before superposition"; }
            else if (id=="SUPER_HOME") {
                t="Reset coordinates";
                tt="Reset coordinates to home the original coordinates as recorded in the protein file";
            }
            else if (id=="SHOW_HETERO") t="Show/Hide single hetero compounds";
            else if (id=="WIRE_DIFF_COLOR") {
                t="Each protein in a different color";
                tt="Each protein color according to the chain ID using JMol colors.<br>SHIFT: Each in a different color indipendent of the chain ID.";
            }
            else if (id=="WIRE_DIFF_COLORc") t="Each protein in a different color according to chain";
            else if (id=="WIRE_RESET") {
                t="Revert original orientation";
                tt="Orientation as at the beginning";
            }
            else if (id=="CUST_INIT") { t="Customize init scripts"; i=IC_CUSTOM; }
            else if (id=="SCRIPT") { t="Generic script panel";  i=IC_BATCH; }
            else if (id==BUT_DEFAULT_VIEWER) clazz3D=CLASS_ChAstexPROXY;
            else if (id=="WIRE_DEFAULT_V" && getPrgParameters()!=null) {
                for(String s0 : getPrgParameters()) {
                    final String s=delPfx("-prefere",s0);
                    if (s==s0) continue;
                    for(Object c : StrapPlugins.allClassesV(ProteinViewer.class).asArray()) {
                        final String cn=shrtClasNam(c);
                        if (cn!=null && (s.equalsIgnoreCase(cn)) || s.equalsIgnoreCase(niceShrtClassNam(c))) clazz3D=cn;
                    }
                }
                if (t==null) return null;
            }
            else if (id==TOG_SEND[0] || id==TOG_SEND[1] || id==TOG_CURSOR_SELECT) { t=id; select=toggle=true; }
            else if (id=="WIRE_HIDE_DIST_HET") {
                t="Hide nucleotide chains and heteros that are distant from current peptide chain";
                toggle=true;
            } else {
                for(String cmd : allCommands()) {
                    if (equls(STRSTR_w_R, cmd, id, 0)) {
                        t=id;
                        word0=cmd;
                        break;
                    }
                }
            }
            if (id==COMMANDcenter) { t="Center"; tt="Center view to current selection";}
            if (id==COMMANDshowScriptPanel) { t="Native script-panel"; tt="The script panel originally provided by this viewer."; i=IC_BATCH; }
            if (clazz3D!=null) {
                final String nice=niceShrtClassNam(clazz3D);
                t=id==BUT_DEFAULT_VIEWER ? id : "View in "+nice;
                i=TabItemTipIcon.imageIcon(clazz3D);
                tt=nice+" is an advanced 3D-viewer";
            }
            b=(toggle ? toggl(id).s(select) : new ChButton(id)).t(t).tt(tt).i(i).li(this).cp(KEY_ENABLED,this);
            if (word0!=null
                || id=="SAVE" || id=="LOG" || id=="LOAD"
                || id==BUT_DEFAULT_VIEWER || id==BUT_NATIV_MENUS || id=="SCRIPT"
                || id=="CTRL" || id=="CUST" || id=="STRAP_MB" || id=="ADD" || id=="ADD_DIALOG"
                || id=="SEL_ATOM" || id=="SEL_ATOMg"
                || id=="SEL_H" || id=="SEL_E" || id=="SEL_HE"
                || id=="SUPER_STOP" || id=="SUPER_GO" || id=="SUPER_UNDO" || id=="SUPER_HOME" || id==CLASS_DialogNeighbors
                ) {
                b.cp(KOPT_HIDE_IF_DISABLED,"");
                if (word0!=null) b.cp(KEY_WORD0, word0);
            }
            if (id==CLASS_DialogNeighbors) b.tabItemTipIcon(CLASS_DialogNeighbors);
            if (clazz3D!=null) pcp(ProteinViewer.class, clazz3D, b);
            if (id=="SEL_ATOM") pcp(KEY_FOREGROUND, C(0x8800),b);
            mapButtons.put(id, b);
        }
        return b;
    }
    public boolean isEnabled(Object o) {
        final boolean cpOnlyWire=null!=gcp("ONLY_WIRE",o);
        final String s=gcps(KEY_ACTION_COMMAND,o);
        boolean e=false, c=false;
        for(ProteinViewer v : objects.proteinViewers()) {
            final Boolean E=isEnabledPV(s,v);
            if (E!=null) return E==Boolean.TRUE;
            final long flags=flags(v);
            final Component panel=viewerCanvas(v);
            final Protein3d p3d=deref(panel, Protein3d.class);
            final ProteinViewer vv[]=v.getViewersSharingView();
            for(int iV=sze(vv); --iV>=-1;) {
                final ProteinViewer v2=iV<0?v:vv[iV];
                final Protein p=p(v2);
                if (p==null) continue;
                final boolean[] subs=p.getResidueSubsetB();
                final boolean isPreview=p3d!=null && p3d.isPreview() || gcp(KOPT_IS_PREVIEW,v2)!=null;
                final int iList=idxOf(s,V3dListModel.TITLES);
                if (iList>=0){ c=true; e=0<sze(V3dListModel.model(iList).items()); }
                else if (s=="RM") c=e=true;
                else if (s=="SAVE" ||  s=="LOG" || s=="LOAD" || s=="$ALL") { c=true; e=p3d==null;}
                else if (s=="SEL_E" || s=="SEL_H" || s=="SEL_HE") { c=true; e=p.getResidueSecStrType()!=null; }
                else if (s=="-99-$NTERM") { c=true; e=p.getResidueSubsetB1()>0 || p.getInferredPdbID()!=null; }
                else if (s=="$CTERM-9999") {
                    c=true;
                    e=subs!=null && lastTrue(subs)+1<sze(p.getResidueTypeFullLength()) || p.getInferredPdbID()!=null;
                }
                else if (s=="SUPER_UNDO") { c=true; e=!isPreview && gcp("SUPER_UNDO",p)!=null; }
                else if (s=="SUPER_HOME") {
                    c=true;
                    final Matrix3D m3d=p.getRotationAndTranslation();
                    e=!isPreview && m3d!=null && !m3d.isUnit();
                }
                else if (s=="WIRE_HIDE_DIST_HET" || s=="SHOW_HETERO") {c=true; e=p3d!=null && p.getDnaAndRnaStructures().length>0; }
                else if (s=="SUPER_STOP" ||s=="SUPER_GO"|| s==CLASS_DialogNeighbors) { c=true; e=!isPreview; }
                if (e) break;
            }
            if (o==labContext) { c=true; e=0!=(flags&PROPERTY_HAS_CONTEXTMENU); }
            //if (s=="STRAP_MB") { c=true; e=parentWndw(panel)==StrapAlign.getFrame() && !LAFChooser.isUseScreenMenuBar(); }
            else if (s==BUT_SURF_OBJ) { c=true; e=v.getProperty(GET_SURFACEOBJECTS)!=null;}
            else if (s=="WIRE_RESET") {c=true; e=p3d!=null && !p3d.getRotationAndTranslationOfView().isUnit(); }
            else if (s=="WIRE_DIFF_COLOR" || s=="WIRE_DIFF_COLORc"  || cpOnlyWire) {
                c=true;
                e=p3d!=null;
            }
            final String word0=gcps(KEY_WORD0,o);
            if (word0!=null && cntains(word0, allCommands())) { c=true; e=supports(word0,v); }
            if (e) break;
        }
        return !c || e;
    }
    public Boolean isEnabledPV(String s, ProteinViewer v) {
        if (v==null) return null;
        final long flags=flags(v);
        final boolean isWire=v instanceof Protein3d.PView, hasPanel=null!=viewerCanvas(v);
        boolean c=false, e=false;
        if (s==BUT_NATIV_MENUS) { c=true; e=jMenuBar(v)!=null; }
        else if (s==BUT_DEFAULT_VIEWER) {  c=true; e=isWire; }
        else if (s=="CTRL") { c=true; e=getCtrlPanel(v)!=null; }
        else if (s=="SEL_ATOM")  { c=true; e=v.getProperty(GET_ATOM_SELECTION_EXAMPLE)!=null; }
        else if (s=="SEL_ATOMg") { c=true; e=!isWire; }
        else if (s=="CUST") { c=true; e=sze(Customize.cc(v))>0; }
        else if (s==IC_CAMERA) { c=true; e=  hasPanel && 0==(flags&PROPERTY_OPENGL); }
        else if (s=="ADD" || s=="ADD_DIALOG") {
            c=true;
            e= hasPanel==(s=="ADD") && 0!=(flags&PROPERTY_MULTI_MOLECULE);
        }
        return !c?null:e?Boolean.TRUE:Boolean.FALSE;
    }
    /* <<< Button <<< */
    /* ---------------------------------------- */
    /* >>> MenuBar 3D >>> */
    private JPopupMenu popup;
    public JPopupMenu menu() {
        final ProteinViewer[] vv=objects.proteinViewers();
        ProteinViewer oneViewer=null;
        for(int i=sze(vv); --i>=0;) {
            if (vv[i]!=null) {
                if (oneViewer==null || !(vv[i] instanceof Protein3d.PView)) oneViewer=vv[i];
            }
        }
        if (oneViewer==null) return null;
        labTitle.setText("Context menu for "+vv.length+" protein viewer"+(vv.length>1 ? "s ":" ")+(vv.length==1?niceShrtClassNam(vv[0]):""));
        final Object mm[]=getGenericMenus(false).clone();
        for(int i=mm.length; --i>=0;) {
            if (mm[i]==b("MBT") || mm[i]==b(BUT_DEFAULT_VIEWER) ||
                getTxt(mm[i])==MENU_OTHER_BAR) mm[i]=null;
        }
        if (popup==null) {
            popup=jPopupMenu(0L, "Contextmenu for ProteinViewer", mm);
        }
        popup.setLightWeightPopupEnabled(false);
        return popup;
    }
    private JComponent[][] genMenus={null,null};
    JComponent[] getGenericMenus(boolean isBar) {
        if (genMenus[isBar?0:1]==null) {
            final ProteinViewer pv=!isImmutable?null:get(0,objects.proteinViewers(),ProteinViewer.class);
            final boolean notP3d=!(pv instanceof Protein3d.PView);
            final long flags=flags(pv);
            final Object
                acpMenus[]={ARRAY_CP, "ONLY_WIRE","",  KOPT_HIDE_IF_DISABLED,"", KEY_ENABLED,this},
                acpWire[]={ARRAY_CP, "ONLY_WIRE",""},
                    ooo[]={
                        labTitle,
                        new Object[]{
                            "File",
                            m('P'),
                            new Object[] {
                                "Session",
                                b("SAVE"),
                                b("LOAD"),
                                "-",
                                new Object[]{ "Log", b("LOGg"), b("LOG")}
                            },
                            b(IC_CAMERA),
                            b("RM")
                        },
                        new Object[]{
                            "View",
                            m('v'),
                            new Object[]{
                                "Animation",
                                new Object[]{
                                    "Delay when picking amino acid. Off=0  Fast=1  Medium=10  Slow=30 ",
                                    new ChTextField(((ChTextField)getJComp(GET_ANIM_SPEED)).getDocument(), 0L).cols(9,true,true).ct(Float.class)
                                }
                            },
                            !notP3d?null: new Object[] {
                                "i", IC_COLOR,
                                "Color",
                                b("WIRE_DIFF_COLOR"),
                                b("WIRE_DIFF_COLORc"),
                                acpMenus,
                                acpWire
                            },
                            !notP3d?null: new Object[] {
                                "i", IC_FLAVIN,
                                "Hetero compounds",
                                b("WIRE_HIDE_DIST_HET"),
                                b("SHOW_HETERO"),
                                acpMenus,
                                acpWire
                            },
                            b("WIRE_RESET")
                        },
                        m('S'),
                        new Object[] {
                            "Select",
                            !notP3d ? null : new Object[]{
                                "Specify atom types",
                                "i", IC_EDIT,
                                b("SEL_ATOMg"),
                                b("SEL_ATOM"),
                                KOPT_DISABLE_IF_CHILDS_DISABLED
                            },
                            b("SEL_AMINO"),
                            b(V3dListModel.TITLES[V3dListModel.HET]),
                            b(V3dListModel.TITLES[V3dListModel.NUCL]),
                            new Object[] {
                                "By secondary structure",
                                b("SEL_H"),
                                b("SEL_E"),
                                b("SEL_HE")
                            },
                            new Object[]{
                                "Termini not shown in the alignment",
                                "i", IC_SCISSOR,
                                b("-99-$NTERM"),
                                b("$CTERM-9999"),

                                KOPT_DISABLE_IF_CHILDS_DISABLED
                            },
                            "-",
                            new Object[] {
                                "Residue selection",
                                "Note: Residue selections are underlined residues in alignment pane.",
                                ResidueSelectionPopup.b('3').t("From residue selections"),
                                b(V3dListModel.TITLES[V3dListModel.SEL]),
                                b(V3dListModel.TITLES[V3dListModel.SFEATURE]),
                                b(V3dListModel.TITLES[V3dListModel.ANNO])
                            },
                            "-",
                            new Object[] {
                                "All",
                                b("$FIRST-$LAST"),
                                b("$ALL"),
                            },
                            b("SEL_I"),
                            "-",
                            b(TOG_CURSOR_SELECT)
                        },
                        styleMenu(pv),
                        m('T'),
                        m('H'),
                        !notP3d ? null : b(BUT_DEFAULT_VIEWER),
                        pv!=null&&jMenuBar(pv)==null && 0==(flags&PROPERTY_HAS_CONTEXTMENU)?null:m('B')
                    };
                    for(Object o : ooo) {
                        if (o instanceof ChButton) ((ChButton)o).options(ChButton.NO_FILL|ChButton.NO_BORDER).i(null);
                    }
                    final JComponent mm[]=genMenus[isBar?0:1]= derefArray(StrapAlign.makeMenus(ooo, null), JComponent.class);
                    for(JComponent menu : mm) {
                        final JMenu m=deref(menu,JMenu.class);
                        if (m!=null && getTxt(m)!=MENU_OTHER_BAR ) {
                            m.getPopupMenu().addPopupMenuListener(chEvLi());
                            pcp( ACTION_BECOME_INVISIBLE, oo(this), m.getPopupMenu());
                        }
                    }
                    if (isBar) undockItms(genMenus[0]);
        }
        return genMenus[isBar?0:1];
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> static utils  >>> */
    private static ChButton[] _withAccelerator;
    private static void withAccelerator(Object object, List<ChButton> v, Set<KeyStroke> setKS) {
        final Object oo[]=
            object instanceof JMenu ? childs((Component)object) :
            object instanceof Object[] ? (Object[])object:
            null;
        if (oo!=null) for(Object o : oo) withAccelerator(o,v,setKS);
        else {
            final JMenuItem mi=deref(object, JMenuItem.class);
            final KeyStroke ks=mi==null?null:mi.getAccelerator();
            if (ks!=null && setKS.add(ks)) addIfNotNull(gcp(KEY_CLONED_FROM,mi,ChButton.class), v);
        }
    }
    private static JMenu menuForAccelerators() {
        if (_withAccelerator==null) {
            final List<ChButton> v=new ArrayList(99);
            final Set<KeyStroke> setKS=new HashSet();
            withAccelerator(StrapAlign.getStrapMenus(), v, setKS);
            _withAccelerator=toArry(v,ChButton.class);
        }
        final ChJMenu m=new ChJMenu(KEY_ACCELERATOR);
        for(ChButton b : _withAccelerator) m.add(b.mi(null));
        return m;
    }
    private JMenuBar mbGeneric;
    public Object run(String id, Object arg) {
        final Object argv[]=arg instanceof Object[] ? (Object[]) arg : null;
        if (id=="SET_PV") setPV(PV_FOCUSED, getPV(PV_FOCUSED));
        if (id=="CPY_SESSION") {
            final ProteinViewer  pvNew=get(1,argv,ProteinViewer.class);
            final ChLogger log=get(0,argv, ChLogger.class);
            final File f=createTempFile(".txt");
            log.save(0L, f);
            loadSession(SESSION_NO_ASK, pvNew,f);
        }
        if (id==KEY_GET_MENUBAR) {
            final ProteinViewer vv[]=objects.proteinViewers(), pv=vv.length==1 ? vv[0] : null;
            if (!isActive(pv)) return null;
            JMenuBar mb=this.mbGeneric;
            if (mb==null) {
                final JComponent mm[]=getGenericMenus(true);
                mb=this.mbGeneric=toMenuBar(0L, mm, new ChJMenuBar());
            }
            return mb;
        }
        return null;
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Load Session >>> */
    private static ChFileChooser fs;
    public final static long SESSION_LOAD_PROTS=1<<1, SESSION_NO_ASK=1<<2;
    private void loadSessionI() {
        if (fs==null) {
            final File f=file(LOG_FILE);
            final String fn=f.getName();
            addActLi(this, fs=new ChFileChooser(ProteinViewerPopup.class).setVisibleFalseParentFrame());
            fs.setCurrentDirectory(mkParentDrs(f));
            fs.setMultiSelectionEnabled(false);
            fs.filter().acceptSuffix(isWindows(), fn.substring(fn.indexOf('.')+1));
        }
        ChFrame.frame("Choose a script input file",fs,ChFrame.ALWAYS_ON_TOP|ChFrame.PACK).shw(ChFrame.CENTER);
    }
    private static void loadSession(long options, ProteinViewer pv, File f) {
        if (pv==null || f==null) return;
        final Protein pp[]=getProteinsInView(pv);
        if (sze(f)==0) { error("File "+f+" has size 0"); return; }
        final ByteArray baLine=new ByteArray(99);
        final ByteArray baNotFound=new ByteArray(999);
        final List<Protein> vNeedsLoading=new ArrayList();
        InputStream is=null;
        if (0!=(options&SESSION_LOAD_PROTS)) {
            try {
                is=new FileInputStream(f);
                final ChInStream in=new ChInStream(is);
                final Set<String> set=new HashSet();
                while(in.readLine(baLine.setEnd(0))) {
                    if (equls(LOG_PROTEIN_PFX, baLine)) {
                        final String pn=baLine.newString(LOG_PROTEIN_PFX.length(), baLine.getEnd());
                        if (set.contains(pn)) continue;
                        set.add(pn);
                        final Protein p=StrapProtein.getProteinWithName(pn, StrapAlign.getInstance().getProteins());
                        if (p==null) baNotFound.append(pn).append('\n');
                        else if (!cntains(p,pp)) vNeedsLoading.add(p);
                    }
                }
            } catch(IOException iox) { error(0L, iox, "");}
            closeStrm(is);
        }
        if (0==(options&SESSION_NO_ASK)) {
            final JComponent
                pAlready=sze(vNeedsLoading)==0?null:
                vBhB(
                     " ",
                     "The following proteins will be loaded:",
                     scrllpn(0L, monospc(joinLines(vNeedsLoading.toArray())),dim(99,99))
                     ),
                pNotFound=sze(baNotFound)==0?null:
                vBhB(
                     " ",
                     "The following proteins could not be found and will be ignored:",
                     scrllpn(0L, monospc(baNotFound),dim(99,99))
                     ),
                msg=vBhB(
                         "<h2>Running script</h2>",
                         hB(new ChJList(new File[]{f},ChJList.OPTIONS_FILES), " ",formatSze(f,null)),
                         pAlready,
                         pNotFound,
                         " ",
                         "Continue?"
                         );
            if (!ChMsg.yesNo(msg)) return;
        }
        open3dViewer(0L, toArry(vNeedsLoading,Protein.class),pv);
        final Map<String,ProteinViewer> mapName2V=new HashMap();
        for(ProteinViewer v : getViewersSharingView(pv)) {
            final String n=getNam(p(v));
            if (n!=null) mapName2V.put(n,v);
        }
        ProteinViewer v=pv;
        try {
            is=new FileInputStream(f);
            final ChInStream in=new ChInStream(is);
            while(in.readLine(baLine.setEnd(0))) {
                if (equls(LOG_PROTEIN_PFX, baLine)) {
                    v=mapName2V.get(baLine.newString(LOG_PROTEIN_PFX.length(), baLine.getEnd()));
                } else if (v!=null) v.interpret(0L, baLine.newString());
            }
            sendCommand(0L, COMMANDcenter_amino,v);
        } catch(IOException iox) { error(0L, iox, "");}
        closeStrm(is);
    }
    /* -------------------------------------------------------------------------------- */
    private JMenu styleMenu(ProteinViewer pv) {
        final String INFO_SHIFT="<br>Holding shift-key while clicking,<br>means \"off\" instead of \"on\".\n<br>";
        final JMenu m=new ChJMenu("Style "),
            mOff=ChUtils.jMenu(0L, oo("Off"), "Off"),
            mOn=ChUtils.jMenu(0L, new Object[]{"On", INFO_UNDOCK,INFO_SHIFT}, "On");
        for(String cmd : STYLE_COMMANDS) {
            if (pv!=null&&!supports(cmd,pv)) continue;
            if (strstr("surface",cmd)>0) {
                final String script=COMMANDsurface_color+" #COLOR\n"+cmd+" off\n"+ cmd+" on\n";
                final ChButton miSurface=butColor(0L,C(0xFF00), script, null)
                    .cp(KOPT_HIDE_IF_DISABLED,"")
                    .cp(KEY_FRAME_TITLE, cmd)
                    .cp(KEY_WORD0, cmd)
                    .cp(KEY_ENABLED, this);
                m.add(miSurface.mi(cmd));
            } else mOn.add(b(cmd+" on").tt(INFO_SHIFT).mi(null));
            mOff.add(b(cmd+" off").cp(KEY_WORD0,cmd).mi(null));
        }
        if (pv==null||null!=pv.getProperty(GET_SURFACEOBJECTS)) {
            final ChButton b=new ChButton(BUT_SURF_OBJ).li(V3dUtils.li()).cp(KEY_ENABLED,this);
            m.add(b.mi("Surface (and ribbon) objects"));
            m.addSeparator();
        }
        if ((pv==null||supports(COMMANDcolor,pv))) {
            m.add(b(BUT_COLOR).li(V3dUtils.li()).i(IC_COLOR).cp(KEY_WORD0, COMMANDcolor).mi("Color"));
        }
        m.add(mOn);
        m.add(mOff);
        if (pv==null||supports(COMMANDlabel, pv)) {
            final Object oo[]={
                "Text labels", "i", IC_ANNO,
                new ChButton(BUT_LABEL).li(V3dUtils.li()).t("Type free label text ..."),
                " ",
                b(COMMANDlabel+" $FIRST_Aaa $FIRST_NUMBER"),
                b(COMMANDlabel+" $FIRST_Aaa"),
                b(COMMANDlabel+" $FIRST_A"),
                " ",
                b(COMMANDlabel+" off")
            };
            m.add(ChUtils.jMenu(0L, oo, "Label"));
        }
        vStyleButton.add(m);
        return m;
    }
}
