package charite.christo.strap;
import charite.christo.*;
import charite.christo.protein.*;
import charite.christo.interfaces.*;
import charite.christo.strap.interfaces.*;
import java.awt.*;
import java.util.*;
import javax.swing.AbstractButton;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static charite.christo.interfaces.VisibleIn123.*;
import static charite.christo.strap.StrapUtils.*;
/**
   @author Christoph Gille
 */
public class ResSelUtils implements Comparator, ChRunnable {
    private final static int ARROW_LEN=4;
    private static Polygon arrowL, arrowR;
    private static ResSelUtils _instance;
    private ResSelUtils() {}
    private static ResSelUtils instance() {
        if (_instance==null) _instance=new ResSelUtils();
        return _instance;
    }
    /* Cursor nach vorne */
    public static void sortResidueSelectionsByVisibility(ResidueSelection[] ss) {
        sortArry(ss,instance());
    }

    /* -1 if o1 should be drawn on top of o2 */
    public int compare(Object o1, Object o2)  {
        final ResidueSelection s1=o1 instanceof ResidueSelection ? (ResidueSelection)o1 : null;
        final ResidueSelection s2=o2 instanceof ResidueSelection ? (ResidueSelection)o2 : null;
        if (s1!=null && s2!=null) {
            final String n1=s1 instanceof HasName ? ((HasName)s1).getName() : null;
            final String n2=s2 instanceof HasName ? ((HasName)s2).getName() : null;
            if (n1==ResidueSelection.CURSOR) return -1;
            if (n2==ResidueSelection.CURSOR) return  1;
            final int style1=s1 instanceof VisibleIn123 ? ((VisibleIn123)s1).getStyle() : -1;
            final int style2=s2 instanceof VisibleIn123 ? ((VisibleIn123)s2).getStyle() : -1;
            if (style1<style2) return 1;
            if (style1>style2) return -1;
            final int count1=countTrue(s1.getSelectedAminoacids());
            final int count2=countTrue(s2.getSelectedAminoacids());
            if (count1<count2) return -1;
            if (count1>count2) return 1;

        }
        return 0;
    }

    static void killResidueAnnotation(long frameOpts, ResidueAnnotation ...ss) {
        if (ss.length==0) return;
        final StringBuffer msg=new StringBuffer("Really kill ").append(ss.length).append(" residue annotations: \n");
        if (ss.length<10)
            for(int i=0;i<ss.length;i++) {
                msg.append(ss[i].getName()).append("  ").append(sp(ss[i])).append('\n');
            }
        if (ChMsg.yesNo(frameOpts, monospc(msg))) {
            StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_DELETED, 111);
            for(int i=0;i<ss.length;i++) ss[i].dispose();
        }

    }
    public static ResidueAnnotation[] convertToResidueAnnotation(boolean edit, ResidueSelection... ss) {
        if (countNotNull(ss)==0) return ResidueAnnotation.NONE;
        final ArrayList<ResidueSelection> vOrig=new ArrayList(),vA=new ArrayList();
        for(ResidueSelection s:ss) {
            final StrapProtein p=sp(s);
            if (p==null || s instanceof ResidueAnnotation) continue;
            final ResidueAnnotation a=new ResidueAnnotation(p);
            a.setValue(ResidueAnnotation.GROUP, toStrg(s)=="cursor" ? "cursor" : "copied");
            final boolean bb[]=s.getSelectedAminoacids();
            final String positions=StrapProtein.selectedPositionsToText(bb,s.getSelectedAminoacidsOffset(),p);
            a.setValue(ResidueAnnotation.POS,positions);
            a.setValue(ResidueAnnotation.NAME, "copied_from_"+ s);
            p.addResidueSelection(a);
            final Color color=colr(s);
            if (color!=null) a.setColor(color);
            if (edit) StrapAlign.editAnnotation(a);
            if (s instanceof MouseSelection || getNam(s)==ResidueSelection.NAME_BACKBONE) p.removeResidueSelection(s);
            else vOrig.add(s);
            vA.add(a);
        }
        final int n=sze(vOrig);
        if (n>0 && ChMsg.yesNo(numAsTxt(true,n)+" persistant annotated residue selection object"+(n==1?"":"s")+" have been created.<br>Remove the original amino acid selections?")) {
            for(ResidueSelection s  : toArry(vOrig,ResidueSelection.class)) sp(s).removeResidueSelection(s);
        }
        if (sze(vA)>0) {
        StrapAlign.setNotSaved();
        StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_ADDED,111);
        }
        return toArry(vA,ResidueAnnotation.class);
    }

    public static SelectorOfNucleotides[] selOfNuclAt(StrapProtein p,int iN) {
        if (iN<0) return new SelectorOfNucleotides[0];
        UniqueList<SelectorOfNucleotides> v=new UniqueList(SelectorOfNucleotides.class);
        for(Object o:p.getAllResidueSelections()) {
            if (!SelectorOfNucleotides.class.isAssignableFrom(o.getClass())) continue;
            final SelectorOfNucleotides selNt=(SelectorOfNucleotides)o;
            final boolean bb[]=selNt.getSelectedNucleotides();
            final int offset=selNt.getSelectedNucleotidesOffset();
            if (bb!=null && offset<=iN && bb.length>iN-offset && bb[iN-offset]) v.add((SelectorOfNucleotides)o);
        }
        return v.asArray();
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>>  asString >>> */
    public static StringBuffer selectedAsString(ResidueSelection sel,StringBuffer sb) {
        final Protein p=sp(sel);
        final byte[] aa=p!=null ? p.getResidueType() : null;
        if (aa==null) return sb;
        boolean isFirst=true;
        int lastIndex=0;
        final boolean bb[]=sel.getSelectedAminoacids();
        final int selOffset=sel.getSelectedAminoacidsOffset();
        final int to=mini(sze(bb), p.countResidues()-selOffset);
        for(int i=maxi(0,-selOffset); i<to; i++) {
            if (bb[i]) {
                if (!isFirst && i-lastIndex>1) sb.append(',');
                isFirst=false;
                lastIndex=i;
                sb.append((char)aa[i+selOffset]);
            }
        }
        return sb;
    }
    /* <<< asString <<< */
    /* ---------------------------------------- */
    /* >>> Draw >>> */
    final static int yT(int style,int cbH) {
        if ((style & 255)==LINES) {
            final int line=(style & 0xFF0000)>>BIT_SHIFT_LINE;
            return cbH+3*line-cbH/8;
        }
        return style==LOWER_HALF_BACKGROUND ?  cbH/2-1 : style==UNDERLINE ?  cbH-4 :  0;
    }

    final static int yB(int style,int cbH) { return  style==UPPER_HALF_BACKGROUND ? cbH/2 : style==UNDERLINE ? cbH-1 : cbH-5;}
    private final static Stroke STROKE_DOTTED=new BasicStroke(1f,BasicStroke.CAP_BUTT,BasicStroke.CAP_BUTT,1f,new float[]{1f,1f},0);
    static void drawResidueSelection(boolean isAmino,byte chr,StrapProtein p,ResidueSelection[] shownSelections,final int iAa0,
                                               Graphics2D graphics,
                                               Rectangle r, Rectangle cb,
                                               Component obs) {
        graphics.translate(r.x,r.y);
        final String strgLen1[]=getStringsLenght1();
        for(int i=shownSelections.length; --i>=0;) { // backward 'cos cursor
            final ResidueSelection s=shownSelections[i];
            final VisibleIn123 visible=isShownIn(s,VisibleIn123.SEQUENCE);
            if (visible==null) continue;
            final int iAa;
            final boolean bb[];
            final boolean notShowingAll;
            final int booleanOffset;
            if (isAmino==ProteinNT.AMINO_ACIDS) {
                final boolean[] selected=s.getSelectedAminoacids();
                bb=ResidueAnnotation.getSelectedAminoacidsDisplayed(s);
                notShowingAll=selected!=bb;
                iAa=iAa0;
                booleanOffset=s.getSelectedAminoacidsOffset();
            } else {
                notShowingAll=false;
                iAa=iAa0;
                if (s instanceof SelectorOfNucleotides) {
                    SelectorOfNucleotides selNt=(SelectorOfNucleotides)s;
                    bb=selNt.getSelectedNucleotides();
                    booleanOffset=selNt.getSelectedNucleotidesOffset();
                } else { booleanOffset=0; bb=null;}
            }
            final int iBB=iAa-booleanOffset;
            if (bb==null || iBB<0 || bb.length<=iBB || !bb[iBB]) continue;
            final int
                style0=visible.getStyle(),
                style=style0==CURSOR ? FULL_BACKGROUND : style0,
                cbH=cb.height,
                yT= yT(style,cbH),
                yB= yB(style,cbH);

            final boolean terminus= bb.length==iBB+1 ||  !bb[iBB+1];
            final int rW= terminus ? maxi(3,r.width-3) : r.width;
            final Color color=style==IMAGE_NO_FILL  ? null : visible.getColor();
            final Image image=(style==IMAGE || style==IMAGE_NO_FILL || style==LINES)
                && obs!=null && isInstncOf(HasImage.class,s) ? ((HasImage)s).getImage(obs) : null;
            if (color!=null) {
                switch(style&255) {
                case HIDDEN:
                    break;
                case CIRCLE:
                    graphics.setColor(color);
                    for(int d=0,w=rW,h=cbH;d<4;d++) {
                        graphics.drawOval( (rW-w)/2, (cbH-h)/2,w,h);
                        w-=3;h-=4;
                    }
                    break;
                case DOTTED:
                    final Stroke stroke0=graphics.getStroke();
                    graphics.setStroke(STROKE_DOTTED);
                    graphics.setColor(color);
                    for(int y=rW; --y>=0;) graphics.drawLine(y%2,y,rW-1,y);
                    graphics.setStroke(stroke0);
                    break;
                case LINES:
                    graphics.setColor(color);
                    if (notShowingAll && (iBB==1 || iBB>1 && iBB+2==bb.length)) {
                        if (arrowL==null) {
                            final int D=4;
                            final Polygon ar=arrowR=new Polygon(), al=arrowL=new Polygon();
                            ar.addPoint(ARROW_LEN,0);
                            ar.addPoint(0,D);
                            ar.addPoint(0,-D);
                            ar.addPoint(ARROW_LEN,0);
                            al.addPoint(0,0);
                            al.addPoint(ARROW_LEN,D);
                            al.addPoint(ARROW_LEN,-D);
                            al.addPoint(0,0);
                        }
                        graphics.setColor(C(color.getRGB(),100));
                        final int transX=iBB==1 ? 0 : rW-ARROW_LEN;
                        graphics.translate(transX,yT);
                        graphics.fillPolygon(iBB==1?arrowR:arrowL);
                        graphics.translate(-transX,-yT);
                    } else fillBigRect(graphics, 0,  yT, rW,2);
                    break;
                case IMAGE:
                case LOWER_HALF_BACKGROUND:
                case UPPER_HALF_BACKGROUND:
                case FULL_BACKGROUND:
                case UNDERLINE:
                    graphics.setColor(color);
                    fillBigRect(graphics, 0,  yT,rW,yB-yT);
                break;
                default:
                }
            }
            if (image!=null && image.getHeight(obs)>0) {

                final int w=image.getWidth(obs),h=image.getHeight(obs);
                final int chars=w*3/h/2;
                if (chars>0) {
                    final int w2=w/chars;
                    if (chars<2) graphics.drawImage(image,0,  yT,rW,yB-yT,obs);
                    else {
                        final int x1=w2*(iAa % chars), x2=x1+w2;
                        graphics.drawImage(image , 0,  yT,rW,yB-yT,   x1 ,0,x2,h, obs);
                    }
                }
            }
        }
        if (chr!=0 && chr!=32){
            graphics.setColor(C(0));
            final String draw=strgLen1[chr&127];
            final int a=cb.y,x=cb.x;
            graphics.drawString(draw, x+0,a  );
            graphics.drawString(draw, x+1,a+1);
            graphics.drawString(draw, x-1,a-1);
            graphics.drawString(draw, x-1,a+1);
            graphics.drawString(draw, x+1,a-1);
        }
        graphics.translate(-r.x,-r.y);
    }
    public static int getResidueindexOfAnnotation(StrapProtein p,String annoName) {
        for(ResidueAnnotation s: p.getResidueAnnotations()) {
            if (annoName.equals(s.getName())) {
                return firstTrue(s.getSelectedAminoacids())+s.getSelectedAminoacidsOffset();
            }
        }
        return -1;
    }
    public static VisibleIn123 isShownIn(ResidueSelection s, int where) {
        if (!(s instanceof VisibleIn123)) return null;
        VisibleIn123 v=(VisibleIn123)s;
        if ((v.getVisibleWhere()&where)==0) return null;
        if (s instanceof ResidueAnnotation) {
            ResidueAnnotation a=(ResidueAnnotation)s;
            if (a.getFeatureName()!=null && !a.isEnabled()) return null;
        }
        if (v.getColor()==null) return null;
        return v;
    }
    /* <<< Draw <<< */
    /* ---------------------------------------- */
    /* >>> Map >>> */
    private static AbstractButton cbProceedSame;
    private final static String
        REPLACE="Replace",
        SKIP="Skip it",
        CANCEL="cancel";

   public static boolean[] mapPositions(ResidueSelection ss[],  Protein pDest) {
       if (pDest==null||sze(ss)==0) return null;
       boolean selected[]=null;
       for(ResidueSelection s:ss) {
           final Protein p_src=sp(s);
           if (p_src==null) continue;
           selected=ResSelUtils.mapPositions(s.getSelectedAminoacids(), s.getSelectedAminoacidsOffset(), p_src, (StrapProtein)pDest, selected);
       }
       return selected;
   }

    public static boolean[] mapPositions(boolean bb_src[],int offset_src, Protein src, Protein dest, boolean[] buffer) {
        final boolean selected_d[]=buffer!=null ? buffer : new boolean[dest.countResidues()];
        for(int i=sze(bb_src)+offset_src; --i>=offset_src;) {
            if (bb_src[i-offset_src]) {
                final int col=((StrapProtein)src).getResidueColumn(i);
                final int idx_d=((StrapProtein)dest).column2nextIndex(col);
                if (idx_d>=0 && idx_d<selected_d.length && ((StrapProtein)dest).getResidueColumn(idx_d)==col) selected_d[idx_d]=true;
            }
        }
        return selected_d;
    }
    public final static long COPY_ADD_TO_PROTEIN=1<<1, COPY_SINGLE=1<<2, COPY_ASK_NAME=1<<3, COPY_REPORT=1<<4;
    private static ResidueAnnotation copyResidueSelectionsSingle(ResidueSelection ss[], StrapProtein pDest, long options) {
        final boolean selected[]=mapPositions(ss,pDest);
        if (firstTrue(selected)<0) return null;
        final String name= toStrgTrim(0!=(options&COPY_ASK_NAME) ? ChMsg.input(ChMsg.AT_CLICK,"Please enter the name of the new residue selection", 20, "Unnamed") : "Unnamed");
        if (sze(name)==0) return null;
        final ResidueAnnotation a=new ResidueAnnotation(pDest);
        a.setValue(ResidueAnnotation.NAME,name);
        a.setValue(ResidueAnnotation.GROUP,"Copied");
        a.setValue(ResidueAnnotation.POS,Protein.selectedPositionsToText(selected,0,pDest));
        if (ss[0] instanceof ResidueAnnotation) {
            nextEntry:
            for(ResidueAnnotation.Entry e0 : ((ResidueAnnotation)ss[0]).getEntries()) {
                for(ResidueSelection s : ss) {
                    if (s==null || s==ss[0]) continue;
                    boolean found=false;
                    if (s instanceof ResidueAnnotation) {
                        for(ResidueAnnotation.Entry e : ((ResidueAnnotation)s).getEntries()) {
                            if (e.getKey()==ResidueAnnotation.POS) continue;
                            if (e.getKey().equals(e0.getKey()) && e.getValue().equals(e0.getValue())) {
                                found=true;
                                break;
                            }
                        }
                    }
                    if (!found) continue nextEntry;
                }
                a.addE(0L, e0.getKey(), e0.getValue(), e0.isEnabled(), MAX_INT);
            }
        }

        final Color color=colr(ss[0]);
        if (color!=null) a.setColor(color);
        pDest.addResidueSelection(a);
        return a;
    }
    static ResidueSelection[] copyResidueSelections(ResidueSelection ss[], StrapProtein[] ppDest, long options) {
        if (sze(ss)==0 || sze(ppDest)==0) return ResidueSelection.NONE;
        final StringBuffer sb=new StringBuffer();
        final ResidueAnnotation aaCopied[];
        if (0!=(options&COPY_SINGLE)) {
            final ArrayList<ResidueAnnotation>v=new ArrayList();
            for(StrapProtein p : ppDest) addIfNotNull(copyResidueSelectionsSingle(ss,p,options),v);
            aaCopied=toArry(v,ResidueAnnotation.class);
        } else aaCopied=copyResidueSelections(ss,ppDest, "copied", sb,  options);
        if (0!=(options&COPY_REPORT)) {
            StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_ADDED,111);
            final String t="Copied residue selections";
            StrapAlign.showInJList(0L, null, aaCopied,t,null).showInFrame(ChFrame.AT_CLICK|ChFrame.SIZE_REGARD_RESOLUTION, t);
        }
        return aaCopied;
    }
    static ResidueAnnotation[] copyResidueSelections(ResidueSelection[] ssSrc, Protein ppDest[],String group, StringBuffer sbError, long options) {
        final ArrayList<ResidueAnnotation>  v=new ArrayList();
        if (sze(ssSrc)==0) {
            StrapAlign.errorMsg("Error: No residue annotation selected");
            return ResidueAnnotation.NONE;
        }
        for(ResidueSelection s:ssSrc) {
            if (s==null) continue;
            final ResidueAnnotation aSrc=deref(s, ResidueAnnotation.class);
            for(Protein p0:ppDest) {
                final StrapProtein p=deref(p0, StrapProtein.class);
                if( p==null) continue;
                if (sp(s)==p && !cntains(s, p.getAllResidueSelections())) {
                    p.addResidueSelection(s);
                    continue;
                }
                final String name_src=s instanceof HasName ? ((HasName)s).getName() : toStrg(s);
                if (cntains(s,p.getResidueAnnotations())) {
                    if (sbError!=null) sbError.append("Skipping Selection "+name_src+" of protein "+p+": source and destination protein are the identical\n");
                } else {
                    final StrapProtein pSrc=sp(s);
                    if (pSrc==null) {
                        sbError.append("Skipping Selection "+name_src+" because getProtein() returns null.\n");
                        continue;
                    }
                    final int offset_src=s.getSelectedAminoacidsOffset();
                    final boolean[] sel_src=s.getSelectedAminoacids();
                    final boolean[] newPos=mapPositions(sel_src, offset_src, pSrc, p, (boolean[])null);
                    final String name=name_src==ResidueSelection.STANDARD ? name_src+"_"+Protein.selectedPositionsToText(sel_src,offset_src,pSrc) : name_src;
                    if (countTrue(newPos)>0) {
                        final ResidueAnnotation  a=new ResidueAnnotation(p);
                        final String feature;
                        if (aSrc!=null) {
                            for(ResidueAnnotation.Entry e : aSrc.getEntries()) a.addE(0L, e.getKey(),e.getValue(),e.isEnabled(), MAX_INT);
                            feature=a.getFeatureName();
                        } else feature=null;
                        a.setColor(s instanceof Colored ? ((Colored)s).getColor() : C(0xAAaa44));
                        a.setValue(ResidueAnnotation.GROUP,group.trim().replaceAll(ProteinViewer.PROTEIN,pSrc.toString()));
                        a.setValue("Copied_from",name +" @ " +pSrc+( aSrc!=null ? "!"+aSrc.getValue(ResidueAnnotation.POS):""));
                        a.setValue(ResidueAnnotation.POS, Protein.selectedPositionsToText(newPos,0,p));
                        a.setValue(ResidueAnnotation.NAME, feature!=null && !feature.equals(name) ? feature+"_"+name : name!=null ? name : shrtClasNam(s) );
                        v.add(a);
                    }
                }
            }
        }
        if (0!=(options&COPY_ADD_TO_PROTEIN)) mayBeAddResidueAnnotations(toArry(v, ResidueAnnotation.class),DialogCopyAnnotations.class);
        return toArry(v,ResidueAnnotation.class);
    }
    static ResidueAnnotation[] mayBeAddResidueAnnotations(ResidueAnnotation[] annotations, Object evtSrc) {
        if (cbProceedSame!=null) cbProceedSame.setSelected(false);
        Object allChoice=null;
        final ArrayList<ResidueAnnotation> vAdded=new ArrayList();
        boolean deleted=false;
        for(ResidueAnnotation sNew:annotations) {
            final String name=sNew.getName();
            final StrapProtein pDest=sp(sNew);
            if (pDest==null || name==null) continue;
            final ResidueAnnotation[] existingAnnotations=pDest.getResidueAnnotations();
            int nExisting=0;
            for(ResidueAnnotation s:existingAnnotations) {
                if (sNew!=s && name.equals(s.getName())) nExisting++;
            }
            final Object choice=nExisting==0 ? "" : allChoice!=null ? allChoice : howProceedWhenExist(name,nExisting);
            if (choice==null) continue;
            if (CANCEL.equals(choice)) break;
            if (isSelctd(cbProceedSame)) allChoice=choice;
            if (SKIP.equals(choice)) continue;
            if (REPLACE.equals(choice)) {
                for(ResidueAnnotation s:existingAnnotations)
                    if (name.equals(s.getName())) {
                        pDest.removeResidueSelection(s);
                        s.dispose();
                        deleted=true;
                    }
            }
            pDest.addResidueSelection(sNew);
            adUniq(sNew,vAdded);
        }
        if (deleted) StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_DELETED,111);
        if (sze(vAdded)>0) StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_ADDED,111);
        return toArry(vAdded, ResidueAnnotation.class);
    }
    private final static Object howProceedWhenExist(String name,int nExisting) {
        if (nExisting<1) return null;
        if (cbProceedSame==null) cbProceedSame=cbox("Treat following duplicate names the same way");
        final String s=(nExisting>1 ?  nExisting+" annotations" : "An annotation")
            +" with the name <br><b><center>"+name
            +"</center></b><br>already exists for this protein.<br>It is possible to have several annotations with the same name.<br>How do you want to proceed ?";
        return ChMsg.optionVal(vBhB(s, cbProceedSame),new String[]{
            REPLACE,
            "<html><body>keep it and<br>create another</body></html>",
            SKIP,
            CANCEL
        });
    }
    /* <<< Map/Copy <<< */
    /* ---------------------------------------- */
    /* >>> Residue Position to Column >>> */
    public static boolean[] toAlignmentColumns(boolean sel[], int offset, StrapProtein p) {
        boolean bb[]=null;
        for(int i=lastTrue(sel); --i>=0; ) {
            if (!sel[i]) continue;
            final int col=p.getResidueColumn(i+offset);
            (bb==null ?  bb=new boolean[col+1] : bb)[col]=true;
        }
        return bb!=null ? bb : NO_BOOLEAN;
    }
    public static boolean[] columnsToResiduePosition(boolean bb[], StrapProtein p) {
        boolean sel[]=null;
        for(int i=lastTrue(bb); --i>=0; ) {
            if (!bb[i]) continue;
            final int iA=p.column2index(i);
            if (iA<0) continue;
            (sel==null ? sel=new boolean[iA+1] : sel)[iA]=true;
        }
        return sel!=null ? sel : NO_BOOLEAN;
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Texshade >>> */
    final static String RUN_TEX_ONE_PROTEIN="SP$$RTOP";
    public static void texshade(String txt, StrapProtein p) {
        final ProteinAlignment a=p.getAlignment();
        NeedsProteins texshade=a instanceof StrapAlign ? (NeedsProteins) ((StrapAlign)a).addDialog(CLASS_Texshade,false) : null;
        if (texshade!=null) {
            texshade.setProteins(p);
            ((ChRunnable)texshade).run(RUN_TEX_ONE_PROTEIN,txt);
        } else Insecure.tellNoPermission('X', "Cannot run TeXshade<br>");
    }
    /* <<< Texshade <<< */
    /* ---------------------------------------- */
    /* >>> Filter >>> */
    public final static int FILTER_NOT_DISABLED=1;
    public static ResidueSelection[] filter(int type, ResidueSelection ss0[]) {
        if (ss0==null) return ResidueSelection.NONE;
        ResidueSelection[] ss=ss0;
        for(int i=ss.length; --i>=0;) {
            final ResidueSelection s=ss[i];
            boolean del=false;
            switch(type) {
            case FILTER_NOT_DISABLED:
                if (getNam(s)==ResidueSelection.CURSOR) del=true;
                if (s instanceof ResidueAnnotation && !((ResidueAnnotation)s).isEnabled()) del=true;
                break;
            }
            if (del) {
                if (ss==ss0) ss=ss0.clone();
                ss[i]=null;
            }
        }
        return rmNullA(ss,ResidueSelection.class);
    }

    public static int countType(char type,  ResidueSelection ss[]) {
        if (ss==null) return 0;
        int countA=0, countF=0, countS=0;
        for(ResidueSelection s : ss) {
            if (s==null) continue;
            if (s instanceof ResidueAnnotation) {
                if (ResidueAnnotation.isAnno(s)) countA++; else countF++;
            } else countS++;
        }
        return type=='A' ? countA : type=='F' ? countF : countS;
    }

    public static void openDialogRealign(ResidueSelection ss[]) {
        final ByteArray sb=new ByteArray(999);
        for(ResidueSelection s : ss) {
            final Protein p=sp(s);
            if (p==null) continue;
            final boolean[] sel=s.getSelectedAminoacids();
            final int offset=s.getSelectedAminoacidsOffset();
            final int first=firstTrue(sel)+offset, last=lastTrue(sel)+offset;
            if (first>=0 && last-first>3) sb.append(p).append('/').append(first+1).append('-').append(last+1).append('\n');

        }
        ((ChRunnable)StrapAlign.addDialog(CLASS_DialogAlign,false)).run(DialogAlign.RADIO_SELECTION, sb.toString());
    }

    public static ResidueSelection[] onlyThoseWithProteins(boolean rmNull, ResidueSelection ss[], Protein pp[]) {
        if (sze(ss)==0) return ResidueSelection.NONE;
        ResidueSelection ss2[]=ss;

        for(int i=0; i<ss.length; i++) {
            final Protein p=sp(ss[i]);
            if (ss[i]==null || !cntains(p,pp)) {
                if (ss==ss2) ss2=ss.clone();
                continue;
            }
        }
        return rmNull ? rmNullA(ss2,ResidueSelection.class) : ss2;

    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Choose for Context or Annotation >>> */

    public final static long CHOOSE_SINGLE=1<<1, CHOOSE_CONTEXT=1<<2, CHOOSE_EDIT=1<<3;
    public static void chooseResSels(long options, ResidueSelection ss[], AWTEvent ev, String msg, Runnable run) {
        if (countNotNull(ss)==0) return;
        final boolean singl=0!=(options&CHOOSE_SINGLE);
        final Component jl=new ChJList(weakRefArry(ss), ChJTable.DEFAULT_RENDERER | (singl?ChJTable.SINGLE_SELECTION:0L)).setSelI(0);
        final String
            title="Chose "+(singl?"a":"")+" residue selection"+(singl?"":"s"), m,
            id=0!=(options&CHOOSE_CONTEXT) ? "CTXT" : 0!=(options&CHOOSE_EDIT) ? "EDIT" : null;
        if (0!=(options&CHOOSE_EDIT)) {
            m="Edit annotations of residue selections"+
                (countType('S',ss)==0?"":"<br>The selected residue selection will be converted into a persistant annotation object.");
        } else m=0!=(options&CHOOSE_CONTEXT) ? "Context menu for residue selections" : msg;
        final Component
            b=new ChButton("CRS").like(ChButton.BUTTON_GO).actionClose(CLOSE_PARENT_WINDOW)
            .r(id==null?run:ChDelay.newThread(instance(), id, new Object[]{jl})),
            pSouth=vBhB(title,pnl(b)),
            pMain=newCNSEW(scrllpn(SCRLLPN_ETCHED_BORDER|SCRLLPN_INHERIT_PREF_SIZE,jl), m, pSouth, null,null);
        new ChFrame(title).ad(pMain).shw(ChFrame.AT_CLICK|ChFrame.PACK|ChFrame.ALWAYS_ON_TOP);
    }

    public Object run(String id, Object arg) {
        final Object argv[]=deref(arg,Object[].class);
        final ChJList jList=get(0,argv,ChJList.class);
        final ResidueSelection ss[]=derefArray(jList==null?null : jList.getSelectedValues(), ResidueSelection.class);
        if (countNotNull(ss)>0) {
            if (id=="EDIT") {
                ResSelUtils.convertToResidueAnnotation(true, ss);
                for(ResidueSelection s : ss) StrapAlign.editAnnotation(deref(s , ResidueAnnotation.class));
            }
            if (id=="CTXT") ResidueSelectionPopup.showContextMenu(ss);
        }
        return null;
    }

    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */
    public static boolean isSelectedAA(int iA, ResidueSelection s) {
        if (iA<0 || s==null) return false;
        return get(iA-s.getSelectedAminoacidsOffset(), s.getSelectedAminoacids());
    }
    public static void setSelectedAA(boolean onOff, int iA, ResidueSelection s, boolean withoutOffset) {
        if (iA<0 || s==null) return;
        final boolean bb0[]=s.getSelectedAminoacids();
        final int offs0=s.getSelectedAminoacidsOffset();
        boolean bb[]=bb0;
        int offs=offs0;
        if (withoutOffset && offs0!=0) {
            bb=chSze(withoutOffset(bb0,offs0),   maxi(iA, offs0+lastTrue(bb0)));
            offs=0;
        }
        int i=iA-offs;
        if (i>sze(bb)) bb=chSze(bb,i+1);
        else if (i<0)  { bb=chSze(withoutOffset(bb,offs), (i=iA)+1); offs=0; }
        bb[i]=onOff;
        if (bb!=bb0 || offs!=offs0) s.setSelectedAminoacids(bb, offs);
    }
}