package charite.christo.strap;
import charite.christo.*;
import charite.christo.protein.*;
import charite.christo.interfaces.*;
import java.awt.*;
import java.util.*;
import java.util.List;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
import static charite.christo.strap.ResidueAnnotation.Entry;
import static java.awt.event.MouseEvent.*;
/**HELP
    <i>PACKAGE:charite.christo.strap.extensions.</i>
    <i>PACKAGE:charite.christo.protein.</i>

				To edit a <i>CLASS_REF:ResidueAnnotation</i>residue annotation The
				item <i>BUTTON:ResidueSelectionPopup#ACTION_edit</i> of the
				context menu must be activated.

    The annotation view contains a table. Each row is one annotation
    entry.
    <div class="figure">
      <table style="caption-side: bottom">
        <CAPTION><b>Figure: </b>Residue annotation panel. The annotated text is displayed as a table. The left column contains the category and the right column the free text.
          Each row has a toggle button to activate and inactivate the entry.
          The user can type into the table and add new table rows.</CAPTION>
        <tr><td>  <i>JCOMPONENT:ResidueAnnotationView#docuView()</i> </td></tr>
      </table>
    </div>
    These entries can be modified and are saved when the alignment is saved.
    Four obligatory
				fields <i>STRING_ARRAY:ResidueAnnotation#MAIN_KEYS</i> are unique
				and mandatory whereas the  fields
    <i>Note</i>, <i>Remark</i>,
    <i>CLASSICON_CLASSNAME:Texshade</i> and
    <i>CLASSICON_CLASSNAME:Pymol</i> may occur zero to several times.
    <ul>
      <li>
        The most important field is
        "<i>STRING:ResidueAnnotation#POS</i>"
        because it defines the indices of the selected residues.
        Here are some examples of valid entries
        <ol>
          <li>
            The expression "1,3,4,6,101-103,110-112"
            selects the residue indices  1,3,4,6,101,102,103,110,111,112.
            Instead of the kommas  spaces can by used.
          </li>
          <li>
            The expression
            "+2 1 3 4"
            selects the residues 3,5, and 6 because the preceding +2 adds an offset of 2 to all positions. The "+"-sign is the first character in this expression.
            For the user interface  the first residue has the index "1" whereas internally STRAP starts counting at zero.
            A negative offset of "2" is achieved by a leading  "+-2".
          </li>
          <li>
            The expression
            "1:G-3:G 5:G"
            selects the residues with the pdb-numbers 1,2, 3 and 5 of the chain G.
            If the protein has only one chain the chain identifier can be ommited like
            "1:-3: 5:"

            </li>
            <li>
            Refering to nucleotide positions:
            When the toggle button is pressed (default state) the indices refer to <b>amino acid</b> positions.
            Otherwise the positions indicate <b>nucleotides</b> in case the protein is translated from a nucleotide sequence.
            </li>

            <li>
            To any group an atom specification can be appended.
            This allows to select certain atoms before changing their style in the 3D-viewers.
            Even though protein 3D viewers usually have a specific language for atom selection,
            the atom expression used here is defined by STRAP and works for all 3D-viewers.
            <br>
            for example "10:-20:.CA.CB" narrows the selection to the atoms CA and CB for the residues 10 to 20.
            A later command  such as "3D_spheres" would affect only c-alpha and c-beta atoms.
            There may be more than one "Atoms"-entries. Each "Atoms" specification take effect only on the following 3D-style
            commands but not on the previous.

            Asterisk can be used as a wild card like "*.CA" or "10:-20:.C".
            For the proper atom identifiers have a look at the PDB file.

          </li>
        </ol>
      </li>
      <li><b><i>STRING:ResidueAnnotation#ATOMS</i></b> 
      Certain Atoms can be specified by an expression like ".CB.CA" which means only C&alpha; and C&beta; atoms.
      The next 3D-command will act on these rather than on all atoms of the amino acid.
      When sending the commands to a 3D-program, the rows are processed sequentially.
      If there comes yet another <i>STRING:ResidueAnnotation#ATOMS</i>-row, 
      the previous atom-specification is replaced by the new one.
      </li>
      <li><b><i>STRING:ResidueAnnotation#NAME</i></b> Each selection has a name. Un-checking the check box deactivates the residue annotation.</li>
      <li><b><i>STRING:ResidueAnnotation#GROUP</i></b> Several selections may be bundled in one group e.g. "active site"</li>
      <li><i>CLASSICON_CLASSNAME:Texshade</i>: TeXshade is a LaTeX-extension for
        exporting an alignment as PDF or PostScript. This field is for TeXshade rendering commands.  Active when the check-box is
        selected.</li>
      <li><i>CLASSICON_CLASSNAME:Rasmol</i>, <i>CLASSICON_CLASSNAME:ChJmolPROXY</i> and <i>CLASSICON_CLASSNAME:Pymol</i>:
        Rasmol, VMD,  Pymol and Jmol are protein structure viewers. 3D-rendering commands can be typed.
        On  pressing the icon-buttons the commands  are send to the viewers. </li>
      <li><i>Note, Remark</i>: Free text can be typed here.
        Clicking on the hyperlink icon opens the specified URLs in the Web-browser.
        There are two ways to specify an URL:
        <ol>
          <li>By typing an URL directly</li>
          <li>By using a database prefix and a key as for example PDB:1ryp, PUBMED<span></span>:0815. The list of databases can be changed if the  CTRL key is hold while the mouse is clicked.</li>
        </ol>
      </li>
      <li><b><i>ICON:IC_EDIT</i><i>Note:</i></b> free text</li>
      <li><b><i>ICON:IC_EDIT</i><i>Remark:</i></b>free text</li>
    </ul>
    <i>SEE_DIALOG:DialogResidueAnnotationChanges</i>
    <i>SEE_DIALOG:DialogResidueAnnotationList</i>
    <i>SEE_CLASS:ResidueAnnotation</i>
    <i>SEE_CLASS:ResidueSelection</i>
   @author Christoph Gille
*/
public final class ResidueAnnotationView extends DefaultTableModel
    implements HasPanel, ChProcessEvent, TabItemTipIcon.Interface, Disposable, StrapListener,
    ListCellRenderer, TableCellEditor,TableCellRenderer {

    public final static String
        POS_AA=addHtmlTagsAsStrg("<font size=2>Indices of <br>amino acids</font>"),
        POS_NUCL=addHtmlTagsAsStrg("<font size=2>Indices of <br>nucleotides</font>"),
        BUT_LAB_ton="Insert<br>variable...",
        BUT_LAB_VARIABLES="View current replacement for variables such as \"TEX_PROTEIN\" or \"First_Aaa\"";
    private final static int COL_KEY=1,COL_VALUE=2,COL_CB=0,COL_DELETE=3;
    private final ChEventAdapter LI=new ChEventAdapter(this);
    private static ChLabel LABEL_RENDERER; { if (LABEL_RENDERER==null)  monospc(LABEL_RENDERER=new ChLabel()); }
    private final ChButton
        BUT_DELETE_EDITOR=new ChButton(null,ChButton.ICON_SIZE|ChButton.PAINT_IN_TABLE).li(LI).i(IC_KILL),
        BUT_OPTIMIZE_EDITOR=new ChButton(null,ChButton.MAC_TYPE_ICON|ChButton.PAINT_IN_TABLE).li(LI).t("*"),
                                               BUTTON_RENDERER=new ChButton(null,ChButton.PAINT_IN_TABLE),
                                               BUTTON_EDITOR=new ChButton(null,ChButton.PAINT_IN_TABLE).li(LI),
                                               TOGGLE_RENDERER=toggl().options(ChButton.ICON_SIZE|ChButton.PAINT_IN_TABLE),
                                               TOGGLE_EDITOR=toggl().li(LI).options(ChButton.ICON_SIZE|ChButton.PAINT_IN_TABLE),
                                               CB_VARIABLES=toggl(BUT_LAB_VARIABLES).li(LI);
    private final ChJList listVariables=new ChJList(SelectionVariables.names(0),ChJTable.SINGLE_SELECTION).li(LI);
    private int editedRow=-1, editedColumn=-1, rowCurrentTextfield=-1;
    private boolean isDisposed;
    private final JComponent panNorth, panel;
    private final ChJTable jTable;
    private final ResidueAnnotation _annotation;
    private Color color;
    /* ---------------------------------------- */
    /* >>> Instance >>> */
    public ResidueAnnotationView(ResidueAnnotation a) {
        this._annotation=a;
        pcp(KEY_MODEL, weakRef(a),this);
        assert a!=null;
        final StrapAlign align=StrapAlign.getInstance();
        setDataVector(new Object[0][4],new Object[]{"","key","value",""});
        final ChJTable jt=this.jTable=new ChJTable(this,ChJTable.ICON_ROW_HEIGHT|ChJTable.NO_REORDER);
        jt.renderer(this).setDefaultEditor(Object.class,this);
        jt.setRowSelectionAllowed(false);
        jt.setCellSelectionEnabled(false);
        jt.setColumnSelectionAllowed(false);
        rtt(jt);
        scrollByWheel(jt);
        jt.setIntercellSpacing(dim(0,0));
        final Object oo[]={this,ResidueAnnotation.class};
        panNorth=newCNSEW(
                          pnl(
                              ChButton.actionViewMessage("Add entry",new AddAnnotation(a)).tt("Add a new entry consisting of <br>a key and a value"),
                              new ChButton("M").t("Menu \u25bc").li(LI),
                              new ChButton("C",ChButton.NO_FILL|ChButton.ICON_SIZE).li(LI).i(IC_CURSOR16).t(null).tt("Add alignment cursor position."),
                              (a.getFeatureName()!=null ? SequenceFeatures.customizeColors().newButton().magnifyUnderMouse(IC_COLOR).t(null) : new ButColor(0L,C(0xFF00FF),a)).options(ChButton.NO_FILL|ChButton.ICON_SIZE),
                              " ",
                              new ProteinLabel(ProteinLabel.ALWAYS_CONTEXT_MENU, a.getProtein())
                              ),
                          null,null,
                          pnl(new ChButton().actionClose(this).tt("Close this view.<br>The residue annotation is not deleted.").options(ChButton.NO_FILL|ChButton.NO_BORDER|ChButton.ICON_SIZE)),
                          pnl(smallHelpBut(oo),smallSourceBut(oo))
                          );
        final JComponent panVariable=
            newCNSEW(listVariables,"choose a variable to be inserted <br>at the current caret position",
                     "<br>Variables are replaced by string expressions<br>depending on the protein and the selected indices.",null,null),
            butVariable=ChButton.actionViewMessage(BUT_LAB_ton,panVariable).tt("variables contain text that is set dynamically"),
            pSouth=newCNSEW(CB_VARIABLES.cb(),null,null,null,butVariable);
        this.panel=newCNSEW(scrllpn(0L,jTable) ,panNorth,pSouth,null,null);
        this.panel.setMinimumSize(dim(1,1));
        jt.setColumnWidth(false, COL_CB,ICON_HEIGHT);
        jt.setColumnWidthC(false, COL_DELETE, new Object[]{BUT_OPTIMIZE_EDITOR, BUT_DELETE_EDITOR});
        jt.setColumnWidth(true, COL_KEY, EM*10);
        jt.setColumnWidth(true, COL_VALUE, EM*60);
        pcp(KEY_DROP_TARGET_REDIRECT,weakRef(a),this);
        pcp(KEY_DROP_TARGET_REDIRECT,weakRef(a),jt);
        for(Component jc:panNorth.getComponents()) {
            jc.setBackground(this.color=a.getColor());
            if (!(jc instanceof AbstractButton)) jc.addMouseListener(LI);
        }
        listVariables.setCellRenderer(this);
        if (align!=null) {
            align.addListener(this);
            StrapAlign.newDropTarget(this.panel,false);
            StrapAlign.newDropTarget(jt,false);
            jt.enableDragRows(this, new int[]{1,2});
            pcp(KEY_DRAG_SRC_ID, intObjct(a.hashCode()), jt);
        }
        pcp(KEY_GET_VIEW, weakRef(this), a);
        pcp(KEY_MODEL,weakRef(a), panel);
    }
    public Component getPanel(int mode) { return this.panel;}
    public void dispose(){
        final ResidueAnnotation a=_annotation;
        if (!isDisposed) {
            isDisposed=true;
            StrapAlign.getInstance().removeListener(this);
            rmFromParent(this);
            rmAllChilds(this.panel);
            pcp(KEY_GET_VIEW, null, a);
        }
    }
    /* <<< Instance <<< */
    /* ---------------------------------------- */
    /* >>> TableModel >>> */
    public int getRowCount() {
        final ResidueAnnotation a=_annotation;
        return a==null ? 0 : a.getEntries().length;
    }
    public Object getValueAt(int row,int col) {
        final ResidueAnnotation a=_annotation;
        final Entry e=a==null?null:get(row, a.getEntries(), Entry.class);
        return e==null?null:col==1?e.getKey():col==2?e.getValue():col==3 && e.isEnabled()?"true" : "";
    }
    public boolean isCellEditable(int row,int col) {
        final ResidueAnnotation a=_annotation;
        final Entry e=a!=null ? a.getEntryAt(row) : null;
        final String k=e!=null ? e.getKey() : null;
        return col!=COL_KEY || isButton(k);
    }
    public void setValueAt(Object o,int row,int col) {
      changed(false);
    }
    static void changed(boolean color) {
        StrapAlign.dispatchLater(color?StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR : StrapEvent.RESIDUE_SELECTION_CHANGED,111);
        StrapAlign.setNotSaved();
    }
    /* <<< TableModel  <<< */
    /* ---------------------------------------- */
    /* >>> CellRenderer >>> */
    private int a_modiCount, lastRowCount;
    public Component getListCellRendererComponent(JList l,Object o, int idx, boolean isSelected, boolean hasFocus) {
        final String txt=o.toString();
        return LABEL_RENDERER.t(txt+" ===> "+SelectionVariables.replace(txt,_annotation)).fg(isSelected ? 0xff : 0);
    }
    private void stopEditingTextField(int row) {
        final ResidueAnnotation a=_annotation;
        final Entry e=a.getEntryAt(row);
        if (e!=null) {
            final String k=e.getKey();
            if (k==ResidueAnnotation.BG_IMAGE) a.findIconImage();
        }
    }
    public Object getCellEditorValue() { return null; }
    public boolean isCellEditable(EventObject ev) { return true;}
    public boolean shouldSelectCell(EventObject ev) { return true;}
    public boolean stopCellEditing() {
        fireEditing('S');
        rowCurrentTextfield=-1;
        return true;
    }
    @Override public void cancelCellEditing() { fireEditing('C');}
    public void addCellEditorListener(CellEditorListener li) {listeners.add(li);}
    public void removeCellEditorListener(CellEditorListener li) { listeners.remove(li);};
    private final List<CellEditorListener> listeners=new ArrayList();
    private void fireEditing( char cancelOrStop) {
        final ChangeEvent ev=new ChangeEvent(this);
        for(int i=listeners.size();--i>=0;) {
            final CellEditorListener l=(CellEditorListener)get(i,listeners);
            if (l==null) continue;
            if (cancelOrStop=='C') l.editingCanceled(ev);
            else l.editingStopped(ev);
        }
        if (cancelOrStop=='S') rowCurrentTextfield=-1;
    }
     private ChTextField textField(Entry e, boolean createInstance, boolean sync) {
        if (e==null) return null;
        final String v=e.getValue();
        ChTextField tf=deref(e.TEXTFIELD[0],ChTextField.class);
        if (tf==null) {
            if (!createInstance) return null;
            e.TEXTFIELD[0]=weakRef(tf=new ChTextField(v));
            final ChTextComponents t=tf.tools().enableUndo(true);
            t.underlineRefs(UNDERLINE_WEB_COLORS);
            pcp(Entry.class,e,tf);
            addActLi(LI,tf);
        }
        if (sync && !v.equals(toStrg(tf))) {
            tf.tools().setTextTS(v);
            tf.tools().underlineRefs(UNDERLINE_REFS|UNDERLINE_WEB_COLORS);
        }
        return tf;
    }
    public Component getTableCellRendererComponent(JTable table,Object value,boolean isSelected,boolean hasFocus,int row,int column) {
        final JComponent c= component(row,column, false);
        final String tt=toolTip(row,column,_annotation);
        if (tt!=null) c.setToolTipText(addHtmlTagsAsStrg(tt));
        return c;
    }
    public Component getTableCellEditorComponent(JTable table,Object value,boolean hasFocus, int row,int column) {
        this.editedRow=row;
        this.editedColumn=column;
        final JComponent c=component(row,column, true);
        final String tt=toolTip(row,column,_annotation);
        if (tt!=null) c.setToolTipText(addHtmlTagsAsStrg(tt));
        return c;
    }
    private JComponent component(int row,int column, boolean isEditor) {
        final ChLabel lab=LABEL_RENDERER;
        final ChButton TOGGLE=  isEditor ? TOGGLE_EDITOR : TOGGLE_RENDERER;
        final ChButton BUTTON=  isEditor ? BUTTON_EDITOR : BUTTON_RENDERER;
        lab.i("").t("").fg(0);
        lab.setEnabled(true);
        final ResidueAnnotation a=_annotation;
        if (a==null) return lab.t("error: null");
        final StrapProtein p=(StrapProtein)a.getProtein();
        final Entry entry=a.getEntryAt(row);
        if (entry==null) return lab.t("error entry==null");
        final String value= entry.getValue().trim();
        final String key=entry.getKey();
        final boolean isEnabled=entry.isEnabled();
        BUTTON.tt(null);
        TOGGLE.tt(null);
        lab.tt(null).i(null);
        switch(column) {
        case COL_DELETE:
            if (key==ResidueAnnotation.POS) return BUT_OPTIMIZE_EDITOR;
            if (idxOf(key,ResidueAnnotation.MAIN_KEYS)>=0) return lab;
            return isEditor ? BUT_DELETE_EDITOR : BUTTON.i(IC_KILL);
        case COL_KEY:
            if (key==ResidueAnnotation.POS) return  lab.t(entry.isEnabled() ? POS_AA:POS_NUCL);
            final String t=delSfx("2StrapPROXY",key.substring(key.lastIndexOf('.')+1));
            if (isButton(key)) return BUTTON.i(TabItemTipIcon.icn(key)).t(t).tt(TabItemTipIcon.tip(key));
            return lab.i("").t(t);
        case COL_VALUE:
            final boolean empty=sze(value)==0;
            if (isEditor) {
                if (rowCurrentTextfield>=0) { stopEditingTextField(rowCurrentTextfield);rowCurrentTextfield=row; }
                rowCurrentTextfield=row;
                return textField(entry,true,true);
            }
            if (key==ResidueAnnotation.POS && !isEditor && column==COL_VALUE) {
                if (empty)  return lab.i("").t("Enter sequ. posit. e.g.24-26,29,31. !   For DNA unselect checkbox !").fg(0xff0000);
                final StringBuffer sb=new StringBuffer().append(value);
                if (!isEnabled && p!=null && p.getNucleotides()==null) return lab.t("Error: the Protein does not have nucleotides. Set the toggle to aminos !").fg(0xff0000);
                if (value!=null && value.indexOf(':')>=0 &&  !is(LETTR,value,0) && p!=null && p.getResidueNumber()==null) {
                    return lab.t("Error:  Colon is only used to refer to the PDB residue chain and number!").fg(0xff0000);
                }
                ResSelUtils.selectedAsString(a,sb.append(" ==> "));
                return lab.t(sb);
            }
            if (key==ResidueAnnotation.NAME) {
                return empty ? lab.t("enter a name for the residue selection e.g. 'active site'").fg(0xff0000) : lab.t(a.getName());
            }
            if (key==ResidueAnnotation.BG_IMAGE && value!=null && value.equals(a.getLastImage())) {
                return lab.i(a.getIcon()).t(value);
            }
            if (key==ResidueAnnotation.GROUP  && empty)
                return lab.i("").t("enter a name for the group e.g. 'polymorphisms'").fg(0xff0000);
            return CB_VARIABLES.s() ? lab.t(SelectionVariables.replace(value,a)) : textField(entry,true,true);
        case COL_CB:
            TOGGLE.s(entry.isEnabled());
            if (key==ResidueAnnotation.GROUP ||  key==ResidueAnnotation.COLOR) return lab.i("").t("");
            if (key==ResidueAnnotation.POS) return  TOGGLE.i(new Object[]{IC_DNA,IC_STRUCTURE});
            return  TOGGLE.i(new Object[]{IC_HIDE,IC_SHOW});
        }
        return lab.t("ERROR");
    }
    /* <<< Rendering <<< */
    /* ---------------------------------------- */
    /* >>> Tip >>> */
    public TabItemTipIcon getTabItemTipIcon(AWTEvent ev) {
        final ResidueAnnotation a=_annotation;
        return a!=null ? a.getTabItemTipIcon(ev) : null;
    }
    private static String toolTip(int row, int col,ResidueAnnotation s) {
        try {
            final Entry e=s.getEntryAt(row);
            if (e==null) return null;
            final String k=e.getKey();
            if(k==ResidueAnnotation.POS)
                switch(col) {
                case COL_KEY:  return "Sequence positions<br>E.g. 42, 100-200,202<br>negtive or pos offset: <br> first char '+' or '-' followed by a number<br>Refering to pdb-nuber and chain like 42:A ";
                case COL_VALUE: return "Enter nucleotide or amino acid position";
                case COL_CB: return "Toggle nucleotide or amino acid position";
                case COL_DELETE: return "Simplify the expression<br>For example \"4-5,6\" to \"4-6\"";
                default:
                }
            if (k==CLASS_Texshade)
                switch(col) {
                case COL_KEY: return "Click this button to generate a printable PDF document using LaTeX<br>Subsequently the PDF-document is viewed";
                case COL_VALUE: return "TeXshade commands for the printable output.";
                case COL_CB: return "Activate/deactivate TeXshade commands.";
                case COL_DELETE: return "Delete row";
                default:
                }
            if (k!=null && k.startsWith(_CCP))
                switch(col) {
                case COL_KEY: return "Click this button to realize the rendering commands in a 3d-viewer<br>See also the 3d-viewer dialog from the protein popup";
                case COL_VALUE: return "Rendering commands for a protein viewer";
                case COL_CB: return "Activate/deactivate this command.";
                case COL_DELETE: return "Delete row";
                default:
                }
            if (k==ResidueAnnotation.NAME)
                switch(col) {
                case COL_KEY:
                case COL_VALUE: return "Each selection has a name";
                case COL_CB: return "Activate/deactivate the entire selection.";
                default:
                }
            if (k==ResidueAnnotation.GROUP)
                switch(col) {
                case COL_KEY:
                case COL_VALUE: return "Selections are grouped";
                case COL_DELETE:
                case COL_CB: return "No function";
                default:
                }
            if (k=="Note" ||k=="Remark")
                switch(col) {
                case COL_KEY:
                case COL_VALUE: return "Free text";
                case COL_CB: return "No function";
                case COL_DELETE: return "Delete row";
                default:
                }
        } catch( Exception e) {}
        return "";
    }
    /* <<< Tooltip <<< */
    /* ---------------------------------------- */
    /* >>> StrapListener >>> */
   public void handleEvent(StrapEvent ev) {
        final int t=ev.getType();
        final ResidueAnnotation a=_annotation;
        final int mc=a.getModificationCount(), nRows=getRowCount();
        if (a_modiCount!=mc) {
            a_modiCount=mc;
            jTable.repaint();
        }
        if (lastRowCount!=nRows) {
            lastRowCount=nRows;
            jTable.revalidate();
        }

        if ( (t&StrapEvent.FLAG_ROW_HEADER_CHANGED)!=0) {
            panNorth.revalidate();
            panNorth.repaint();
        }
        if (t==StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR) {
            final Color col=a.getColor();
            if (this.color!=col) {
                this.color=col;
                for(Component c:panNorth.getComponents()) c.setBackground(col);
                panNorth.repaint();
            }
        }
        if (t==StrapEvent.RESIDUE_SELECTION_DELETED) {
            final StrapProtein p=(StrapProtein)a.getProtein();
            if (p==null || !cntains(a,p.getResidueAnnotations())) dispose();
        }
   }
    /* <<< StrapListener <<< */
    /* ---------------------------------------- */
    /* >>> ActionListener >>> */

    public AWTEvent processEvent(JComponent c, AWTEvent ev) {
        final String cmd=actionCommand(ev);
        final Object q=ev.getSource();
        final int id=ev.getID();
        StrapAlign.errorMsg("");
        final ResidueAnnotation a=_annotation;
        if (a==null) return ev;
        final StrapProtein p=(StrapProtein)a.getProtein();
        if (isPopupTrigger(ev) || cmd=="M") ResidueSelectionPopup.showContextMenu(new ResidueAnnotation[]{a});
        if (cmd!=null) {
            if (cmd=="C") {
                final int iA=StrapAlign.getInstance().getIndexOfAminoAcidAtCursor(p);
                if (iA<0) StrapAlign.errorMsg("Inserting Cursor-Position: First place the cursor in the alignment pane");
                else {
                    final String s=" "+Protein.selectedPositionsToText(new boolean[]{true}, iA, p)+" ";
                    final Entry e=a.getEntryWithKey(ResidueAnnotation.POS);
                    final ChTextField tf=textField(e,false,false);
                    if (tf!=null && tf.hasFocus()) ChTextComponents.insertAtCaret(0L, s, tf);
                    else if (!get(iA-a.getSelectedAminoacidsOffset(), a.getSelectedAminoacids())) {
                        a.setValue(ResidueAnnotation.POS, a.getValue(ResidueAnnotation.POS)+s);
                        jTable.repaint();
                    }
                    changed(false);
                }
            }
            if (cmd==ACTION_CLICKED && q==listVariables) {
                final Entry e=a.getEntryAt(rowCurrentTextfield);
                final ChTextField tf= textField(e,true,true);
                if (tf==null) StrapAlign.errorMsg("Error: Click into table cell where you want the variable to be inserted");
                else {
                    ChTextComponents.insertAtCaret(0L," "+listVariables+" ",tf);
                    changed(false);
                }
            }
            final int row=editedRow, col=editedColumn;
            final Entry entry=a.getEntryAt(row);
            if (cmd==BUT_LAB_VARIABLES) { jTable.repaint();}
            if (entry!=null && p!=null) {
                final String key=entry.getKey(), value=entry.getValue();
                if (editedRow<0) return ev;
                if (q==BUT_DELETE_EDITOR) {
                    a.removeEntry(a.getEntryAt(this.editedRow));
                    changed(false);
                }
                else if (q==BUT_OPTIMIZE_EDITOR) { rowCurrentTextfield=-1; ResidueSelectionPopup.optimizeExpression(a);}
                else if (q==BUTTON_EDITOR) {
                    if (key==CLASS_Texshade) {
                        final String txt=rplcToStrg("$PROTEIN_NO","1",value);
                        ResSelUtils.texshade(toStrg(SelectionVariables.replace(txt,a)),p);
                    } else {
                        final String w0=frstTkn(value);
                        boolean success=false;
                        final StringBuffer sbErr=new StringBuffer();
                        ProteinViewer vv[]=p.getProteinViewers().clone();
                        for(int i=vv.length; --i>=0;) if (!Protein3dUtils.supports(w0, vv[i])) vv[i]=null;
                        vv=rmNullA(vv,ProteinViewer.class);
                        if (vv.length==0) vv=V3dUtils.getViewersOfProteinAsk(p, name2class(StrapAlign.getDefault(ProteinViewer.class)));
                        for(ProteinViewer v:vv) {
                            a.setOnlyEntry(entry);
                            if (idxOfStrg(w0,Protein3dUtils.allCommands())>=0 && !Protein3dUtils.supports(w0,v)) sbErr.append("Command \""+w0+"\" is not yet supported by "+shrtClasNam(v)+"\n");
                            //if (idxOfStrg(w0,Protein3dUtils.allCommands())>=0 && !v.getSupportsCommand(w0)) sbErr.append("Command \""+w0+"\" is not yet supported by "+shrtClasNam(v)+"\n");
                            else {
                                success=true;
                                V3dUtils.residueSelectionsTo3D(V3dUtils.RSto3D_COMMANDS|V3dUtils.RSto3D_SET_FOCUSED_PV, new ResidueAnnotation[]{a},  new ProteinViewer[]{v});
                            }
                            a.setOnlyEntry(null);
                            V3dUtils.viewerToFront(V3dUtils.OPEN_ADD_TAB, v);
                        }
                        if (!success && sze(sbErr)>0) error(sbErr);
                    }
                } else if (q==TOGGLE_EDITOR) {
                    entry.setEnabled(TOGGLE_EDITOR.s());
                    fireEditing('S');
                    jTable.revalidate();
                    jTable.editCellAt(row,col);
                }
                if (key==ResidueAnnotation.BG_IMAGE) {
                    a.findIconImage();
                    changed(true);
                }
            }
            if (cmd==ACTION_FOCUS_LOST || cmd==ACTION_ENTER) {
                final Entry e=gcp(Entry.class,q, Entry.class);
                if (e!=null) a.setValue(toStrg(q),e,false);
            }
        }
        return ev;
    }
    /* <<< ActionListener <<< */
    /* ---------------------------------------- */
    /* >>> Utils >>> */
    private static boolean isButton(String k) {
        return isAssignblFrm(CommandInterpreter.class,k) || k.equals(CLASS_Texshade) || k==ResidueAnnotation.COLOR || k==ResidueAnnotation.VIEW3D;
    }
    /* <<< Utils <<< */
    /* ---------------------------------------- */
    /* >>> Html >>> */
    public static JComponent docuView() {
        final StrapProtein p=new StrapProtein();
        p.setName("test_protein");
        final ResidueAnnotation s=new ResidueAnnotation(p);
        p.addResidueSelection(s);
        s.setValue(ResidueAnnotation.GROUP,"new");
        s.setValue(ResidueAnnotation.POS,"6-8");
        s.addE(CLASS_Texshade,"\\feature{bottom}{$TEX_PROTEIN}{$TEX_RESIDUES}{fill:$\\uparrow$}{active site}");
        s.addE(ResidueAnnotation.VIEW3D,ProteinViewer.COMMANDspheres);
        final ResidueAnnotationView v=new ResidueAnnotationView(s);
        v.panel.setPreferredSize(dim(370,300));
        return v.panel;
    }
}
