package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import charite.christo.strap.interfaces.*;
import charite.christo.strap.extensions.AbstractAlignAndCompare;
import java.awt.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.table.*;

import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP
In each line of the input file the program expects two (or more) protein files.
Both proteins are loaded and compared.
If more than two proteins are given, each is compared with each other.

In the result list both protein names and the alignment score are written.

For some comparison methods the result will be stored in a cache.

For time consuming computations like 3D-superpositions, this can be
used to fill the cache with results over-night.
*/
class ResultCompareProteinPairs extends AbstractDialogJPanel implements ChRunnable, ActionListener  {
    private final Vector<Vector<String>> vResult=new Vector();
    private final Vector<String> vColumnNames=new Vector();
    private final ChTextArea ta_Error=new ChTextArea("");
    { vColumnNames.add("protein1"); vColumnNames.add("protein2"); vColumnNames.add("score");}
    private final JTable jTable=new JTable(vResult,vColumnNames);
    private final Object clazz,otherClass;
    private final int fromColumn,toColumn;
    private final ChTextField tfFile;
    private final ProteinProteinValue sharedInstance;
    private final ChButton butStop=new ChButton("stop").li(this), butGo=new ChButton("GO").t(ChButton.GO).li(this);
    public ResultCompareProteinPairs(Object clazz, ProteinProteinValue sharedInstance, int fromColumn, int toColumn) {
        this.fromColumn=fromColumn;
        this.toColumn=toColumn;
        this.clazz=clazz;
        this.sharedInstance=sharedInstance;
        otherClass=isInstncOf(AbstractAlignAndCompare.class, sharedInstance) ? ((AbstractAlignAndCompare)sharedInstance).getOtherClass() : null;
        final JComponent pButtons=pnl(butGo);
        pButtons.add(buttn(TOG_CACHE).cb());
        tfFile=new ChTextField().li(this).ct(Float.class).saveInFile("ResultCompareProteinPairs_inputFile");
        tfFile.tools().enableWordCompletion(dirWorking());
        pcp(KEY_IF_EMPTY,"enter input file name",tfFile);
        JComponent pSouth=pnl( new GridLayout(2,1),newCNSEW(tfFile,null,null,new ChButton("HD").t("Browse").li(this),"Input file:"),pButtons);
        cnsew(this, scrllpn(0L,jTable), pTop(), pSouth, null,null);
        butGo.enabled(sze(file(tfFile))>0);
        tfFile.tools().underlineRefs(UNDERLINE_REFS_NOT_CLICKABLE);
    }

    private ChFrame frameFs;

    public Object run(String id, Object arg) {
        if (id=="GO") {
            final ByteArray ba=(ByteArray)arg;
            final int ends[]=ba.getLineEnds();
            final byte[] text=ba.getBytes();

            vResult.clear();
            for(int iL=0; iL<ends.length; iL++) {
                if (isDisposed()) break;
                final int b=iL==0 ? 0 : ends[iL-1]+1 , e=ends[iL];
                final StringTokenizer st=new StringTokenizer(new String(text,b,e-b), " \t\n,");
                StrapProtein[] pp=null;
                final int countTokens=st.countTokens();
                for(int i=countTokens; --i>=0;) {
                    if (isDisposed()) break;
                    final StrapProtein p=protein(st.nextToken());
                    if (p!=null) {
                        if (pp==null) pp=new StrapProtein[i+1];
                        pp[i]=p;
                    }
                }
                if (pp==null) continue;
                long time0=System.currentTimeMillis();
                for(StrapProtein p0 : pp) {
                    if (p0==null) continue;
                    final String n0=p0.getName();
                    Set setNames=mapProteinNames.get(n0);
                    if (setNames==null) mapProteinNames.put(n0,setNames=new HashSet());
                    for(StrapProtein p1 : pp) {
                        if (p1==null || isDisposed()) continue;
                        final String n1=p1.getName();
                        if (setNames.contains(n1) || countTokens==2 && n0.equals(n1)) continue;
                        setNames.add(n1);
                        final ProteinProteinValue dis=HotswapUtil.instance(clazz,ProteinProteinValue.class,true);
                        if (dis==null) continue;
                        dis.setProtein(p0);
                        dis.setOtherProtein(p1);
                        dis.setColumnRange(fromColumn,toColumn);
                        dis.compute();
                        final double v=dis.getValue();
                        final Vector vR=new Vector();
                        vR.add(n0);
                        vR.add(n1);
                        vR.add(String.valueOf((float)v));
                        vResult.add(vR);
                        final long t=System.currentTimeMillis();
                        putln("vResult="+vResult.size());
                        if (t-time0>333) {
                            putln("TableModelEvent "+vResult.size());
                            jTable.revalidate();
                            jTable.repaint();
                            time0=t;
                        }
                        if (isInstncOf(Disposable.class,dis)) ((Disposable)dis).dispose();
                        sleep(10);
                    }
                    if (countTokens==2) break;
                }
                            jTable.revalidate();
            }
            ChDelay.invokeLater(this,"FINISH",null);
        }
        if (id=="FINISH") {
            final Container parent=butStop.getParent();
            try {
            parent.remove(butStop);
            parent.add(new ChLabel("All done").fg(0xFF0000));
            ((JComponent)parent).revalidate();
            }catch(Exception e){}
        }
        return null;
    }
    private final Map<String,Set<String>> mapProteinNames=new HashMap();
    private Object refMapProteins;
    private StrapProtein protein(String s) {
        StrapProtein p=StrapProtein.getProteinWithName(s,null);
        if (p==null) {
            Map<String,StrapProtein> map=(Map)deref(refMapProteins);
            if (map==null) refMapProteins=newSoftRef(map=new HashMap());
            p=map.get(s);
            if (p==null) map.put(s,p=StrapProtein.newInstance(file(s)));
        }
        if (p==null && ta_Error!=null) ta_Error.appnd("Could not load  protein ",s,"\n");
        return p;
    }
    public void actionPerformed(ActionEvent ev) {
        StrapAlign.errorMsg("");
        final String cmd=ev.getActionCommand();
        final Object q=ev.getSource();
        if (q==tfFile) {
            butGo.enabled(sze(file(tfFile))>0);
        }
        if (cmd=="HD") {
            if (frameFs==null) frameFs=new ChFrame("").ad(addActLi(this,new ChFileChooser(ResultCompareProteinPairs.class)
));
            frameFs.shw(ChFrame.CENTER);
        }
        if (cmd==JFileChooser.APPROVE_SELECTION) {
            tfFile.setText(toStrg(((ChFileChooser)q).getSelectedFile()));
        }
        if (cmd=="GO" || cmd==ACTION_ENTER) {
            final ByteArray txt=readBytes(file(tfFile));
            if (txt==null) { StrapAlign.errorMsg("Could not open input file "+tfFile); return;}
            JComponent c=(JComponent)q, p=parnt(c,JComponent.class);
            p.remove(c);
            if (ta_Error!=null) p.add(ChButton.actionViewMessage("Log",ta_Error));
            if (isAssignblFrm(HasControlPanel.class,clazz)) p.add(new ChButton("CTRL").t("Details").li(this));
            putln(" \n otherClass ="+otherClass);
            for(Class hasView : new Class[]{Superimpose3D.class,SequenceAligner.class}) {
                if (isAssignblFrm(hasView,clazz) || isAssignblFrm(hasView,otherClass)  ) {
                    p.add(new ChButton("VIEW").t("View result").li(this));
                    break;
                }
            }
            p.add(butStop);
            p.revalidate();
            ChDelay.newThread(this,"GO",txt).start();
            tfFile.setEditable(false);
        }
        if (cmd=="stop") dispose();
        if (cmd=="VIEW" || cmd=="CTRL") {
            final int i=jTable.getSelectedRow();
            if (i<0) { StrapAlign.errorMsg("Select a table row first"); return;  }
            final Vector<String> v=vResult.get(i);
            StrapProtein p0=protein(v.get(0));
            StrapProtein p1=protein(v.get(1));
            if (cmd=="VIEW") StrapProteinUtils.viewComparison(p0,p1,otherClass);
            if (cmd=="CTRL") {
                final ProteinProteinValue dis=HotswapUtil.instance(clazz,ProteinProteinValue.class,true);
                if (dis==null) return;
                setSharedParas(this.sharedInstance,dis);
                dis.setProtein(p0);
                dis.setOtherProtein(p1);
                dis.setColumnRange(fromColumn,toColumn);
                dis.compute();
                Container cp=getCtrlPanel(dis);
                if (cp!=null) new ChFrame(v.toString()).ad(new JScrollPane(cp)).shw(0);
            }
        }
    }

}
