package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import java.io.File;
import java.util.*;
import javax.swing.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
/**HELP

   STRAP starts on clicking  <b>strap.jar</b> or <b>strap2.jnlp</b>.

   It may also be started from the command line using the command "<b>java</b>"  in the
   <b>bin/</b> directory of the Java installation:

   <pre class="terminal">
   java -jar STRAP.jar
   </pre>

   In case of the error message "java: command not found" the complete path should be provided like:

   <pre class="terminal">
   /local/jdk1.5.0/bin/java -jar STRAP.jar
   </pre>

   <br><b>Options:</b>
   <ul>

   <li><b>-stderr</b> Lets the standard error stream flow to the
   terminal shell in the conventional way rather than redirecting it to
   a special STRAP window available from the <I>debug</I> menu of
   STRAP.</li>

   <li><b>-stdout</b> Same for  standard output  stream </li>

   <li><b>-noSeqres</b> If
   residues of a pdb-file are recorded in SEQRES-lines but not in ATOM-lines they
   appear in lower case.  With this switch the SEQRES-lines are
   complete ignored and only the ATOM-lines are read.  </li>

   </ul>

   <br><b>Examples with protein files: </b>

   <pre class="terminal">
   java -jar strap.jar a1_Homo.seq a1_Saccharomyces.seq
   </pre>

   starts STRAP and loads two protein files.
   <br><br><b>Example with sequence grouping: </b>

   <pre class="terminal">
   java -jar strap.jar \{ a*.swiss \} \{ b*.swiss \}
   </pre>

   loads all files starting with an "a" into one line and all files
   starting with "b" into the other line.
   (See <i>DIALOG:DialogManyInOneRow</i>).

   <br><br><b>Example with <I>grep:</I> </b>

   <pre class="terminal">
   java -jar strap.jar $(grep -l -i '^OS.*human'  *.swiss)
   </pre>

   loads all WIKI:Swissprot files with the organism human. The option
   -i makes the search case insensitive.

   <br><br><b>Example with pdb chain identifier:</b>

   <pre class="terminal">
   java -jar strap.jar pdb1ryp.ent:A
   </pre>
   loads only chain A of the proteasome.
   <br>

   <br><br><b>Example with pdb chain identifier using underscore:</b>

   <pre class="terminal">
   java -jar strap.jar pdb1ryp_A.ent
   </pre>
   loads only chain A of the proteasome.
   <br>

   <br><br><b>Example with residue subset: </b>
   <pre class="terminal">
   java -jar strap.jar 'pdb1ryp.ent!20-30,50-66'
   </pre>
   loads only the given residue ranges of the proteasome.
   <br>
   <br><br><b> Example with the at-sign: </b>
   <pre class="terminal">
   java -jar strap.jar  @list
   </pre>

   loads all files found in file <I>list.</I>
   The text file "proteins.list"  contains the previously loaded files.

   <br><br><b>Example with PDB-id: </b>
   <pre class="terminal">
   java -jar strap.jar -pdb=1sbc
   </pre>
   loads the  PDB:1sbc from the PDB-server.
   <br>

   @author Christoph Gille

*/

public class Strap implements java.awt.event.ActionListener,ChRunnable {
    final static String  DO[]=new String[4];
    final static int NO_ALIGNMENT=0, NO_SUPERPOSE=1, NO_3D=2, DO_CHANGE_ORDER=3;
    static String listFileName;
    private static File fProjects;
    private static String projects[];
    private static Strap inst;
    static Strap instance() {if (inst==null) inst=new Strap(); return inst;}

    static {

        System.setProperty("apple.laf.useScreenMenuBar",LAFChooser.METAL ? "false" : "true");
        if(!LAFChooser.METAL)  System.setProperty("com.apple.mrj.application.apple.menu.about.name", "Strap");

    }
    /* ---------------------------------------- */
    /* >>> Event >>> */
    public void actionPerformed(java.awt.event.ActionEvent ev) {
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        {
            final ChFrame f=StrapAlign.getFrame();
            final boolean cancel=cmd==ACTION_WINDOW_CLOSING && q==fQuit || cmd=="QUITc";
            if (cmd==ACTION_WINDOW_CLOSING && q==_frameProjects) shutDown();
            if (cancel) { f.setVisible(true); fQuit.setVisible(false); }
            if (cmd=="QUITs" || cmd=="QUIT" || cancel) {

                final ChTextArea ta=gcp(ChTextArea.class,fQuit,ChTextArea.class);
                final String txt=toStrgTrim(ta);
                if (nxt(NO_SPC, txt)>=0) {
                    ta.t("");
                    final Thread t=ChStdout.thread_uploadBug("\n"+ANSI_GREEN+"Comment"+ANSI_RESET+" "+txt+"\n\n");
                    if (cancel) t.start(); else t.run();
                }
                if (cmd=="QUITs") StrapAlign.getInstance().save();
                if (cmd=="QUIT" || cmd=="QUITs") shutDown();
            }
        }
        if (cmd=="HLP") {
            setWndwState('t', _frameProjects);
            showHelp(CLASS_DialogProjectFiles);
        }
        if (cmd=="H") {
            final File f=file(getTxt(q));
            openProject(f,"", (Object[])null,null,0,(Runnable)null,false);
            saveProjectDir(f);
        }
        if (cmd=="SEARCH_HD") {
            final ChFileChooser fs=new ChFileChooser(Strap.class).setVisibleFalseParentFrame();
            addActLi(this,fs);
            fs.setMultiSelectionEnabled(false);
            fs.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
            final String TITLE="Choose project directory";
            new ChFrame(TITLE).ad(fs)
                .size(800,600)
                .shw(ChFrame.ALWAYS_ON_TOP|ChFrame.CENTER);
            setWndwStateLaterT('F', 99, TITLE);
        }
        if (cmd==ACTION_ENTER) {
            final String s=tf_project.toString();
            if (""==s) { error("The text field is empty.<br>Enter the complete path"); return;}
            final File f=file(s);
            if (s.indexOf(' ')>0 || s.indexOf("%20")>0){
                final String waround=!isWindows() ? "" :
                    "<h2>Workaround, if you really have to use this directory</h2>"+
                    "Assign an unused drive letter ( WIKI:Drive_letter) such as T: for the directory under consideration.<br>"+
                    "Type into the Windows command shell ( DOS shell ):"+
                    "<pre class=\"terminal\">\n\n"+
                    "  subst T: \""+s+"\" \n\n"+
                    "</pre>\n\n"+
                    "  Test the assignment:"+
                    "<pre class=\"terminal\">\n\n"+
                    "  dir T:\\  \n\n"+
                    "</pre>"+
                    "You can now refer  to this directory  with the path  T:\\.<br><br>"+
                    "Undo the  assignment:"+
                    "<pre class=\"terminal\">\n\n"+
                    "  subst /D T:  \n\n"+
                    "</pre>";
                error("The directory <pre class=\"data\">"+s+"</pre>Is invalid because it contains white space!<br>"+
                      "Bioinformatics programs do not cope with blanks in file paths.<br>Chose a different file folder!<br><br>"+waround);
                return;
            }
            else if (f.exists() && !f.isDirectory()) error("The path<pre>"+s+"</pre>is not a directory. It is a file.");
            else {
                Insecure.securitySetAllowedPath(f);
                if (mkdrsErr(f)) {
                    openProject(f,"",(Object[])null,null,0,null,false);
                    saveProjectDir(f);
                }
            }
            //ChDelay.newThread(this,"DesktopUtils",null).start();
        }
        if (cmd==JFileChooser.APPROVE_SELECTION) tf_project.t(((ChFileChooser)q).getSelectedFile().toString());
    }
    /* <<< AWTEvent <<< */
    /* ---------------------------------------- */
    /* >>> Quit About >>> */
    public static void aboutDialog() {
        final String s="<h2>Multiple sequence alignment program Strap</h2>Author Christoph Gille";
        ChMsg.msgDialog(0L, vBhB(s,hBL("Home-page: ",URL_STRAP), hBL("In Publications please cite:  ","PMID:11301311", " or ","PMID:16322575")));
        setWndwState('t',_frameProjects);
    }
  private static ChFrame fQuit, _frameProjects;

    static void quitStrap() {
        final StrapAlign align=StrapAlign.getInstance();
        if (align==null) shutDown();
        final ChFrame f=StrapAlign.getFrame();
        final String[] securityViolations=strgArry(Insecure.VIOLATIONS.values());
        final String[] para=getPrgParameters();
        if (securityViolations.length>0) {
            ChStdout.uploadErrorReport(joinAsStrg(securityViolations,"\n-------------\n"),true);
            Insecure.VIOLATIONS.clear();
        } else  {
            if (isPrgOptionTF("-notAskQuit",true,para) && !isPrgOptionTF("-noExit",true,para)) {
               if (StrapAlign.needsSaving()) StrapAlign.save();
               shutDown();
            } else {
                if (fQuit==null) {
                    final ChTextArea taComments=new ChTextArea("");
                    pcp(KEY_IF_EMPTY,"Or type your comments here", taComments);
                    final Strap li=instance();
                    final Object
                        sp=scrllpn(0L,taComments,dim(1,4*EX)),
                        contactN=hBL("Send problems, feed-back, questions to ",monospc(" christophgil@googlemail.com ")),
                        contact=newCNSEW(sp,contactN,null,null,null),
                        msg=vBhB(
                                 hBL("Contact the author ",toggl().actionPack().actionCollapse(false, contact)),
                                 contact,
                                 " ",
                                 hB("#"," ",
                                    !StrapAlign.needsSaving() ? null : new ChButton("QUITs").li(li).t("Save and Quit"),"#"," ",
                                    new ChButton("QUIT" ).li(li).t("Quit without saving"),"#"," ",
                                    new ChButton("QUITc").li(li).t("Do not quit"),"#"," "
                                    ),
                                 " "
                                 );
                    addActLi(li,fQuit=new ChFrame("Quit Strap").ad(msg));
                    pcp(ChTextArea.class,taComments, fQuit);
                }
                fQuit.shw(ChFrame.PACK|ChFrame.ALWAYS_ON_TOP|ChFrame.CENTER);
                //final Rectangle rF=f.getBounds(),rQ=fQuit.getBounds();
                //fQuit.setLocation(rF.x+(rF.width-rQ.width)/2 , rF.y+(rF.height-rQ.height)/2);
                fQuit.pack();
                f.setVisible(false);
            }
            //if (CB_INTERPRET_SCRIPTLINES.s()) delFile(file("~/.StrapAlignCommands"));
            if (isPrgOptionTF("-noExit",true,para)) {
                StrapAlign.getDialogPanel().removeAll();
                align.removeProteins(false, align.getProteins());
                f.setVisible(false);
                f.getContentPane().removeAll();
                f.dispose();
                //if (winProgress!=null) winProgress.setVisible(false);
                stopAllAlignments();
            }
        }
    }
    /* <<< Quit About <<< */
    /* ---------------------------------------- */
    /* >>> main >>> */
    private final static String DISCLAIMER=
        "Email christophgil@googlemail.com\n\n\n"+
        "No warranty  that  embedded programs for alignment computation, structure prediction, 3D-Visualization ... and STRAP itself\nwill be free from harmful code, so-called Malware.\n\n"+
        "If you worry about safety, then run STRAP in a Sandbox .\n\n"+
        "See http://www.bioinformatics.org/strap/security.html for more info.\n\n"+
        "Run STRAP?";

    public static void main(final String[] argv) {
        //try {
        //debugExit(charite.christo.strap.extensions.PairAlignerNeoBioPROXY.class.newInstance());
        //} catch(Exception ex){}
        boolean security=true;
        for(String a : argv) {
            if (a==null) continue;
            if (a.startsWith("-toFasta") || a.startsWith("-toMultipleFasta") || a.startsWith("-toClustal")) security=false;
        }
        Insecure.setFileModificationControl(security);
        {
            final String h=sysGetProp(USER_DOT_HOME);
            putln(" --- STRAP by Christoph Gille --- \n ", isMyComputer()?"isMyComputer":"");
            putln(USER_DOT_HOME,"=",h);
            String varHome=isWindows() ? "HOMEPATH":"HOME";
            if (javaVsn()!=14) putln("Environment var ",varHome,"=",ChEnvironment.get(varHome));

            if (sze(h)>0 && h.indexOf('%')<0) {
                ChMsg.disclaimerMsg(DISCLAIMER, h+(isWindows()?"/":"/.")+Insecure.STRAP_DISCLAIMER_TXT);
            }
        }
        final int iShowHelp=idxOfOption("-showDocumentation",argv), iGenerateHtml=idxOfOption("-generateHtmlDocumentation",argv);
        if (iGenerateHtml>=0) {
            final String aa[]=argv.clone();
            aa[iGenerateHtml]="";
            HtmlDoc.main(aa);
        } else if (iShowHelp>=0) {
            for(int i=iShowHelp+1;i<argv.length; i++) {
                if (chrAt(0,argv[i])!='-') showHelp(argv[i]);
            }
        } else {
            if (isMacintoshLaf()) invokeMthd(StrapMacApplicationListener.INIT, StrapMacApplicationListener.CLASS);
            ChDelay.invokeAndWait(instance(),"MAIN",argv);
        }

    }
    /* <<< Main <<< */
    /* ---------------------------------------- */
    /* >>> Web-Mode  >>> */
    private static int countNewActivation;
    private static void removeJNLP_files(boolean sayClearFolder) {
        boolean msg=false;
        for(String sDir : new String[]{"~/", "~/Desktop", "~/Desktop/downloads", "~/Desktop/Downloads",
                                       "~/Desktop/Download",  "~/Documents/Download",
                                       isWindows() ? ChEnvironment.get("TEMP") : "/tmp",
                                       isWindows() ? ChEnvironment.get("TMP" ) : null }) {
            if (sDir==null) continue;
            final File dir=file(sDir);
            if (!dir.isDirectory()) continue;
            for(String fn : new String[]{"strap", "strap_3D_alignment", "web_strap", "strapLite"}) {
                for (int vers=3; --vers>=0; ) {
                    final File f=new File(dir, vers==0?fn+".jnlp" : fn+"-"+vers+".jnlp");
                    if (sze(f)>0) {
                        if (vers==0) renamFile(f, dirLog()+"/"+f.getName()); else delFile(f);
                    }
                }
            }
            msg=msg ||  sayClearFolder && sze(dir.list())>33;
        }
        if (msg) ChMsg.savedMsg("To improve download time you could clean up the \"<b>downloaded-files</b>\" window of the Web-Browser","clearDownloadFolder");
    }

    public static boolean isWebMode(String argv[]) {
        for(String s: new String[]{"-pdb","-load","-align","-alignAndRearange","-script", "-notToFront"}) if (idxOfOption(s,argv)>=0) return true;
        return false;
    }
    /* <<< Web-Mode <<< */
    /* ---------------------------------------- */
    /* >>> run >>> */
    public Object run(String id,Object arg) {
        final StrapAlign align=StrapAlign.getInstance();
        final boolean newAct= id==Sil.ACTION;
        final String[] argv=
            newAct && arg instanceof String ? splitTokns(0L, (String)arg,"\t") :
            arg instanceof String[] ? (String[])arg :null;
        //if (id=="DesktopUtils") { ChUtils.sleep(5555); DesktopUtils.main(DesktopUtils.OPTION_INSTALL); }
        if (id=="SCRPT") StrapAlign.interpretScriptLater((ByteArray[])arg);
        int webOption=StrapAlign.WEB_OPTION_USE_UNIPROT_SOAP;
        final boolean sepInstance=(StrapAlign.button(StrapAlign.TOG_IN_NEW_FRAME).s() || isPrgOptionTF("-separateInstance", true, argv));

        if (id=="MAIN" || newAct) {
            final String userHome=systProprty(SYSP_USER_HOME); /* First method, because may ChMsg.askUserHome */
            ChMsg.disclaimerMsg(DISCLAIMER, userHome+"/"+(isWindows()?"":".")+"/strapDisclaimer.txt");

            if (DesktopUtils.main(argv)) return null;
            final ArrayList<Object>vTargets=new ArrayList();
            final ByteArray msgSettingsDir=new ByteArray(99);
            ChUtils.setBufferForMsg(msgSettingsDir,'S');
            setDialogs.clear();
            StringBuffer webAlignment=null;
            if (isSystProprty(IS_WEBSTARTED)) setWndwStateT('I',0,0,"D");
            removeJNLP_files(countNewActivation++ >10);
            boolean isAtSign=false;
            final StringBuffer sb=new StringBuffer();
            String[] vars=null, varValues=null;
            int nVars=0;
            for(int i=DO.length; --i>=0;) DO[i]=null;
            ByteArray scripts[]=null;
            setPrgParameters("strap",argv,false);
            long loadMask=ProteinParser.SEQUENCE_FEATURES;
            for(String a_:argv) {
                for(int i=0; i<nVars; i++) a_=rplcToStrg(vars[i],varValues[i],a_);
                final String a=a_.trim();
                final int eq=a.indexOf('=');
                final String afterEqu=jnlpDecode(a.substring(eq+1));
                final char afterEqu0=afterEqu.length()>0 ? afterEqu.charAt(0) : ' ';
                if (a.toUpperCase().endsWith(".LIST")) Strap.listFileName=a;
                if ("@".equals(a)) { isAtSign=true; continue; }
                final String standardOpts[]=interpretPrgParameter(a,afterEqu);

                if (!a.startsWith("-")) {
                    if (isAtSign) { sb.append('@'); isAtSign=false;}
                    sb.append(a).append(' ');
                } else {
                    if (a.startsWith("-SUBMIT=")) continue;
                    final String A=a.intern();
                    final String OO[]=new String[33];
                    int iO=0;
                    final boolean changeOrder=a.startsWith(OO[iO++]="-alignAndRearange=");
                    final boolean isAlign=a.startsWith(OO[iO++]="-align=") || changeOrder, isLoad=a.startsWith(OO[iO++]="-load=");
                    if (changeOrder) DO[DO_CHANGE_ORDER]="all";
                    if (isLoad) DO[NO_ALIGNMENT]="all";
                    final String ali=a.startsWith(OO[iO++]="-pdb=") ? "PDB:"+delPfx("PDB",afterEqu.toUpperCase()) : isAlign||isLoad? afterEqu : null;
                    Class defaultC=null;
                    if (a.startsWith("-default")) {
                        for(Class c : StrapPlugins.interfaces()) {
                            final String cn=getNam(c);
                            if (equls(0L, a,8,eq,  cn, cn.lastIndexOf('.')+1)) defaultC=c;
                        }
                    }
                    if (defaultC!=null) {}
                    else if (a.startsWith(OO[iO++]="-v3d="))      defaultC=ProteinViewer.class;
                    else if (a.startsWith(OO[iO++]="-s3d="))      defaultC=Superimpose3D.class;
                    else if (a.startsWith(OO[iO++]="-aligner="))  defaultC=SequenceAligner.class;
                    else if (ali!=null) {  (webAlignment==null ? webAlignment=new StringBuffer() : webAlignment).append(ali).append(' '); }
                    else if (a.startsWith("-variable=")) {
                        for(String s : splitTokns(0L, afterEqu,",")) {
                            int colon=s.indexOf('=');
                            if (colon<1) continue;
                            if (sze(vars)<nVars+1) { vars=chSze(vars,nVars+10); varValues=chSze(varValues,nVars+10); }
                            vars[nVars]=s.substring(0,colon);
                            varValues[nVars++]=s.substring(colon+1);
                        }
                    }
                    else if (isPrgOptionTF("-separatePdbChains",true,a)) webOption|=StrapAlign.WEB_OPTION_splitChains;

                    else if ("-rename=sp"==A) webOption|=StrapAlign.WEB_OPTION_RENAME_SWISS;
                    else if (a.startsWith(OO[iO++]="-geometry=")) parseRect(afterEqu, StrapAlign.GEOMETRY);

                    else if (a.startsWith(OO[iO++]="-downloadOriginalProteins")) webOption|=StrapAlign.WEB_OPTION_DOWNLOAD_ORIGINAL_PROTEINS;
                    else if (a.startsWith(OO[iO++]="-noA=")) DO[NO_ALIGNMENT]=afterEqu;

                    else if (a.startsWith("-noSP=")) DO[NO_SUPERPOSE]=afterEqu;
                    else if (a.startsWith("-no3D=")) DO[NO_3D]=afterEqu;
                    else if (a.startsWith("-mlock=")) buttn(TOG_LOCK_JMENU).s(true);
                    else if (isPrgOptionTF(OO[iO++]="-hideAll",true,a)) {
                        if (newAct) {
                            final StrapProtein pp[]=StrapAlign.getInstance().getProteins();
                            StrapAlign.getInstance().setIsInAlignment(false,pp);
                            inEDT(new StrapEvent(Strap.class, StrapEvent.PROTEINS_HIDDEN).setParameter(pp));
                        }
                    }
                    else if (isPrgOptionTF("-notToFront",true,a)) vTargets.add("f");
                    else if ((OO[iO++]="-classloader")==A)  AbstractProxy.setSystemClassLoader(true);
                    else if (a.startsWith(OO[iO++]="-dialog=")) setDialogs.add(afterEqu);

                    else if (isPrgOptionTF(OO[iO++]="-noCache",true,a)) CacheResult.setEnabled(false);
                    else if ((OO[iO++]="-noSeqres")==A) loadMask=ProteinParser.IGNORE_SEQRES;
                    else if (a.startsWith(OO[iO++]="-noSequenceFeatures")) loadMask&=(~ProteinParser.SEQUENCE_FEATURES);
                    else if (a.startsWith(OO[iO++]="-dasFeatures=")) vTargets.add(new Object[]{StrapAlign.KEY_DAS,afterEqu});
                    else if (a.startsWith(OO[iO++]="-dasRegistry=")) {
                      for(String reg : splitTokns(0L,afterEqu)) adUniq(reg, DAS.vRegistry);
                    }
                    else if ((OO[iO++]="-help")==A || "--help"==A || "-h"==A) {
                        putln(rplcSb(0L,"\n\n\n","\n\n",toStringBuffer(getHlp(Strap.class))));
                        System.exit(0);
                    } else if ((OO[iO++]="-toMultipleFasta")==A) { convertToMSA(ExportAlignment.FASTA, argv); System.exit(0);
                    } else if ((OO[iO++]="-toFasta")==A) { convertToMSA(0, argv); System.exit(0);
                    } else if (a.startsWith(OO[iO++]="-toClustal")) { convertToMSA(ExportAlignment.CLUSTALW, argv); System.exit(0);
                    } else if (a.startsWith("-benchmark=")) addBenchmark(afterEqu0);
                    else if (a.startsWith("-script") && (eq==7  || cntainsOnly(DIGT,a,7,eq))) {
                        final int num=atoi(a,7);
                        (scripts=chSze(scripts, num+1, ByteArray.class))[num]=new ByteArray(afterEqu);
                    }
                    else if (a.startsWith(OO[iO++]="-noIdentical")) StrapAlign.button(StrapAlign.TOG_SKIP_IDENTICAL_SEQUENCE).s(true);
                    else if (a.startsWith("-soap=")) webOption&=(~StrapAlign.WEB_OPTION_USE_UNIPROT_SOAP);
                    else if (a.startsWith("-port=")) {
                        final int port=atoi(afterEqu);
                        if (port<1024) putln("Error option ",a," Please use port numbers > 1024 ");
                        else if (new Sil(this,port).isSuccess())
                            putln("\n"+ANSI_GREEN+"Accepting"+ANSI_RESET+" commands at port "+port+" Try:\n echo -load=PDB:1sbc | telnet localhost "+port+"\n");
                    }
                    else if (standardOpts!=null &&
                             strstr("customizeAdd",a)<0 &&
                             !isPrgOption(OO[iO++]="-logDnD",a) &&
                             !isPrgOption(OO[iO++]="-separateInstance",a) &&
                             !isPrgOption(OO[iO++]="-stdout",a) &&
                             !isPrgOption(OO[iO++]="-stderr",a) &&
                             !isPrgOption(OO[iO++]=PARA_USE_JV,a) &&
                             !isPrgOption("-noDisclaimer",a) &&
                             !isPrgOption("-noExit",a) &&
                             !isPrgOption("-notAskQuit",a) &&
                             !isPrgOption("-prefere",a)
                             ) {
                        putln("\n Warning: Unknown option ",a,"\nAvailable options: ");
                        for(String ss[] : new String[][]{standardOpts,OO}) {
                            for(Object o:ss) if (o!=null) putln(o);
                        }
                    }
                    //debugExit("a="+a+" defaultC="+defaultC);
                    if (defaultC!=null && sze(afterEqu)>0) StrapAlign.setDefaultC(defaultC, afterEqu);
                }
                StrapProtein.setParserOptions(loadMask);

            }
            if (!newAct) {
                LAFChooser.setLAF(LAFChooser.IF_NOT_ALREADY,false);
                ChStdout.accordingToCommandLineOptions(getPrgParameters());
                for(int i=1;i<=2;i++) TabItemTipIcon.add(readLines(rscAsStream(0L,_CC,"tabItemTipIcon"+i+".rsc")));
            }
            if (sze(msgSettingsDir)>0) InteractiveDownload.addTab(new AsciiView(msgSettingsDir),true,ChFrame.CENTER);
            vTargets.add(StrapView.KEY_AP);
            vTargets.add(StrapAlign.ACTION_SORT);
            final Runnable script=scripts==null?null: ChDelay.newThread(this,"SCRPT", scripts);

            if (newAct) openProject(null,null, vTargets.toArray(), webAlignment, webOption, script, true);
            else {
                final boolean isWebArguments=isWebMode(argv);
                Web.checkForNewVersion((isMyComputer() ? "christo_":"")+(!Insecure.EXEC_ALLOWED?"strapLite" : isWebArguments ? "webStrap" : "strap"));
                if (isWebArguments) {
                    assertInstAllowed();
                    if (isSystProprty(IS_WEBSTARTED)) {
                        if (!sepInstance && !new Sil(instance(),0).isSuccess()) {
                            Sil.toSocketAndWaitForAnswerThenDie(16968,argv);
                            if (!new Sil(this,16968).isSuccess()) {
                                error("Unable to start communication via localhost:16968");
                                if (javaVsn()==14)
                                    error(
                                          "The old Java version 1.4. is running.<br>Please update Java! <br>"+
                                          "Make sure that the higher Java versions are not deactivated in the Java settings.<br>"+
                                          "See "+fileJcontrol()+"<br>"
                                          );
                            }
                        }
                        setWndwStateT('I',0,0,"D");
                    }
                    final File f=dirWeb();
                    if (f.isDirectory()) {
                        buttn(TOG_NASK_BEFORE_DOWNLOAD).s(true);
                        buttn(TOG_NASK_BEFORE_UPLOADING_DATA).s(true);
                        File dir=null;
                        for(int i=1;i<444 && (dir==null || dir.exists()) ;i++) dir=file(f+"/temp"+i);
                        delFileOnExit(dir);
                        openProject(dir,"",vTargets.toArray(),webAlignment,webOption,script,false);
                    } else ChMsg.msgDialog(0L, "Error: Could not create directory<br>Unable to start the program  properly.");
                }
                UserProfile.init();
                //if (isMacintosh() || isWindows()) { NOT_AVAILABLE_CLASSES.add(Rasmol.class);  NOT_AVAILABLE_CLASSES.add(Vmd.class); }
                ChMsg.sound(ChMsg.SOUND_ON);
                if (!isWebArguments && sze(sb)>0 && !dirHome().equals(dirWorking()) && !dirWorking().equals(dirDesktop()))
                    openProject(dirWorking(),sb,vTargets.toArray(),null,webOption,script,false);
                else if(!isWebArguments) {
                    fProjects=file("~/@/StrapProjects");
                    projects=readLines(fProjects);
                    tf_project=new ChTextField("",COMPLETION_DIRECTORY_PATH).cols(80,true,false).li(this);
                    String lastProject=null;
                    if (projects!=null) {
                        for(int i=projects.length; --i>=0;) {
                            final File d=file(projects[i]);
                            if (d==null || !d.isDirectory()) projects[i]=null;
                            else lastProject=projects[i];
                        }
                    }
                    JComponent panPrevProjects=null;
                    if (lastProject!=null) {
                        tf_project.t(lastProject);
                        final JComponent pan=vBhB("<h2>Open Project</h2>");
                        for(String p : projects) {
                            if (p!=null) pan.add(new ChButton("H",ChButton.MAC_TYPE_ICON).t(filepathUnixStyle(p)).li(instance()));
                        }
                        panPrevProjects=scrllpn(0L,pan);
                    } else {
                        final String
                            base=!Insecure.ARBITRARY_WORKING_DIRECTORY_ALLOWED ? toStrg(dirSettings()) :
                            strchr(' ',userHome)>0 ? "c:\\temp" : userHome;
                        tf_project.t(base+"/testProject");
                    }
                    final File dirS=dirSettings();
                    final JComponent
                        butSearch=new ChButton("SEARCH_HD").li(instance()).t("Browse"),
                        pTF=newCNSEW(tf_project,null,null,butSearch," "),
                        pSouth=vBhB("#ETB",
                                    " ",
                                    "<h2>Create project folder</h2>",
                                    pTF,
                                    pnl(new ChButton(ACTION_ENTER).li(instance()).t(ChButton.GO)),
                                    " "
                                    ),
                        pSouth2=vB(
                                   pSouth," ",
                                   sze(msgSettingsDir)>0 ? hBL("Note: A folder ",dirS," has been created ", smallHelpBut(msgSettingsDir)) : null,
                                   javaVsn()==14 ? pnl("You use Java version 1.4 <font color=\"#FF0000\"> Please update! </font>", !isMacintosh() ? "http://java.com/download/" : "http://www.apple.com/search/downloads/") : null
                                   ),
                        pMain=newCNSEW(panPrevProjects,null,pSouth2,null,"#EBem");
                    ChDelay.parentWindowToFront(pMain,5000);

                    final ChFrame f=_frameProjects=new ChFrame("STRAP compiled "+dateOfCompilation()+"   Java "+systProprty(SYSP_JAVA_VERSION)).ad(pMain);
                    final JMenuBar mb=new JMenuBar();
                    mb.add(Box.createHorizontalGlue());
                    mb.add(jMenu(0L, new Object[]{"Help", "h", StrapAlign.class, new ChButton("HLP").t("Projects in Strap").li(this)},""));
                    f.setJMenuBar(mb);
                    f.shw(ChFrame.PACK|ChFrame.ALWAYS_ON_TOP);
                    addActLi(instance(), f);
                    tf_project.requestFocus();
                }
            }
            return "";
        }
        return null;
    }
    /* <<< run <<< */
    /* ---------------------------------------- */
    /* >>> Open projects >>> */
    private static ChTextField tf_project;
    private final static Set<String> setDialogs=new HashSet();
    private static void saveProjectDir(File o) {
        if (o==null) return;
        if (projects!=null) {
            for(int iP=projects.length; --iP>=0;) {
                final File f=projects[iP]!=null ? file(projects[iP].trim()) : null;
                if (f!=null && !f.isDirectory())  projects[iP]=null;
            }
        }
        final ArrayList<String> l=new ArrayList();
        l.add(o.toString());
        adAllUniq(projects,l);
        final ByteArray ba=byteArr(SB_TIP);
        for(String s : strgArry(l)) ba.append(s).append('\n');
        if (!wrte(fProjects,ba)) putln("Caught in Strap.java: unable to  write file strapProjects");
    }
    private static boolean isInitialized;
    private static void openProject(File projectDir, CharSequence files, Object[] targets, CharSequence webAlignment, int webOption, Runnable runFinished,boolean newAct) {
        StrapAlign align=null;
        UserProfile.setWebMode(webAlignment!=null);
        if (!newAct) {
            if (!isInitialized) {
                isInitialized=true;
                ChMsg.ubuntuFontError();
                final String s=sysGetProp("java.vm.name");
                if (strstr(STRSTR_IC,"GNU",s,0,999)>=0 || strstr(STRSTR_IC,"CACAO",s,0,999)>=0 ) {
                    error("java.vm.name="+s+"\n\nSTRAP may not work properly with this Java version.\nRecommended Java system:\nhttp://java.com/download\n\n");
                }
            }
            if (_frameProjects!=null) _frameProjects.dispose();
            Insecure.securitySetAllowedPath(projectDir);
            setWorkingDir(projectDir);

            align=new StrapAlign(isWebMode(getPrgParameters()));
            ChFrame.setMainFrame(align.getFrame());
            mkdrsErr(file(DIR_ANNOTATIONS));
            mkdrsErr(file(DIR_GAPS));
            align.init();
            if (isMacintoshLaf()) invokeMthd(StrapMacApplicationListener.INIT2, StrapMacApplicationListener.CLASS);
        } else align=StrapAlign.getInstance();
        if (align!=null) {
            int opt=webOption|StrapAlign.WEB_OPTION_SKIP_PDB_IF_ALREADY_LOADED|(runFinished==null?StrapAlign.WEB_OPTION_SCROLL:0);
            if (webAlignment!=null) newThrd(StrapAlign.threadWebAlignment(webAlignment,null,targets,opt), runFinished).start();
            else if (runFinished!=null) runFinished.run();
            if (files!=null) {
                align.loadTheProteinsInList(files,StrapAlign.OPTION_PROCESS_LOADED_PROTS|StrapAlign.OPTION_EVENT_PROTEINS_ADDED|StrapAlign.OPTION_SHOW_ORPHAN_HETEROS_IN_3D);
            }
            if (sze(setDialogs)>0 && align!=null) {
                for(Object o : setDialogs.toArray()) {
                    final Class clazz=
                        "L".equals(o)?DialogSubcellularLocalization.class :
                        "B".equals(o)?DialogBlast.class :
                        name2class(fullClassName(o.toString(),STRAP_PACKAGES));
                    if (clazz!=null) {
                        if (StrapPlugins.getDialogForPlugin(clazz).length>0) StrapPlugins.applyPlugin(clazz);
                        else StrapAlign.addDialog(clazz,false);
                    }
                }
            }
        }
        setFolderIcon(dirSettings(),"superimpose.jpg", "This folder contains program files of STRAP.");
    }
    /* <<< Open project <<< */
    /* ---------------------------------------- */
    /* >>> Static Utils >>> */

    private final static StrapProtein[] readProteins(String ...argv) {
        final ArrayList<StrapProtein> vP=new ArrayList();
        for(String a : argv) {
            if (sze(a)==0 || a.startsWith("-")) continue;
            final File f=file(a);
            final ByteArray ba=readBytes(f);
            if (sze(ba)==0) continue;
            final StrapProtein[] ppMsf= StrapAlign.proteinsFromMsfText(ba, f, 2, (StrapAlign)null);
            if (sze(ppMsf)>0) for(StrapProtein p : ppMsf)  vP.add(p);
            else {
                final StrapProtein p=new StrapProtein();
                p.setFile(f);
                p.parse(f,ba,0);
                vP.add(p);
            }
        }
        return toArry(vP,StrapProtein.class);

    }
    private static void convertToMSA(long format, String ...argv) {
        final StrapProtein pp[]=readProteins(argv);
        if (pp.length>0) {
            if (format==0) {
                final ByteArray baOut=new ByteArray(999);
                for(StrapProtein p : pp) {
                    final File fOut=file(STRAPOUT+"/"+p+".fasta");
                    baOut.setEnd(0).append('>').append(p).append('\n').append(p.getResidueTypeExactLength()).append('\n');
                    wrte(fOut,baOut);
                    putln("Strap.java: Written "+fOut+"  "+baOut.getEnd()+" bytes");
                }
            } else {
                final ExportAlignment writer=new ExportAlignment();
                writer.setProteins(pp);
                final StringBuffer sb=new StringBuffer(99999);
                writer.getText(format,sb);
                wrte(file(STRAPOUT+"/alignment"+writer.getFileExtension()), sb);
            }
        }
        if (idxOfOption("-tellOutputDir",argv)>=0) {
            if (pp.length>0)  ChMsg.msgDialog(0L, vBhB("Files written to ",file(STRAPOUT)));
            else {
                ChFrame.setGlobalOptions(ChFrame.SYSTEM_EXIT_WHEN_ALL_CLOSED);
                showHelp(DesktopUtils.class);
            }
        }
    }
}

