package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import java.io.File;
import java.util.*;
import javax.swing.*;
import static java.awt.event.MouseEvent.MOUSE_CLICKED;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static charite.christo.protein.DAS.Entry;
/**HELP
   BioDAS* services provide residue specific annotations for proteins.
   BioDAS* services are listed in <i>STRING:DAS#URL_REGISTRY</i> (http://www.dasregistry.org/).
   STRAP reads this BioDAS* registry file and sorts the services according to the authority.
   The most important authority is "UniProt" which is displayed in the first tab.
   The user can select one or several DAS services in the table and one or several proteins.
   When <i>LABEL:ChButton#BUTTON_GO</i> is pressed, the DAS features are read from the selected Services.
   For this purpose, the database references stored in the protein files are used.
   The database references of proteins can be seen in the balloon text over the protein labels of the alignment.
   If a services from the "UniProt" tab is used, then only those
   proteins can be processed, that have a "UniProt" reference.
   <br><br><b>Cache:</b> The downloaded files are cached in
   <i>STRING:ChUtils#dirWeb()</i> and are valid for <i>STRING:DAS#DAYS_VALID</i> days.
   <br><br><b>Example XML output</b>: Press <i>BUTTON:StrapDAS#butExamble!</i> to get a list of example XML files for each DAS service.
   These XML files are basis for the table column "Example".
   @author Christoph Gille
   ARMA AgamP BDGP BGI BROADD BROADE BROADO BROADS BUSHBABY Btau_ CAT CEL CHIMP COMMON_SHREW CSAV EMBL EMDB Ensembl Entrez EquCab FUGU GENCODE GUINEAPIG HEDGEHOG HIV HUGO_ID Havana IPI JGI KEGG MEDAKA MGI MICROBAT MMUL_ NCBI NCBI m NCBIM OANA PDBresnum PPYG Pfam RABBIT RGSC SGD SQUIRREL TAIR TENREC TETRAODON TIGR TREESHREW UniParc UniProt WASHUC WB WS ZFISH Zv cavPor dipOrd gorGor micMur pika proCap pteVam tarSyr turTru vicPac
*/
public class StrapDAS extends ChRenderer implements ChRunnable, ChProcessEvent  {
public final static long  STOP_ON_CLOSED_FRAME=1<<1;
    private static ChEventAdapter li;
    private static StrapDAS _inst;
    private static StrapDAS instance() { if (_inst==null) li=new ChEventAdapter(_inst=new StrapDAS()); return _inst;}
    private static ChButton butGO, butRenderer;
    private static JComponent pButtons;
    private static JTabbedPane tabbed;
    private static ChTextArea taSouth;
    private static boolean downloadExamples;
    private final static StringBuffer sbExamples=new StringBuffer(9999);
    public static ChButton butExamble=ChButton.actionViewMessage("Example outputs",sbExamples);
    private JProgressBar progress;
    private final Set setDownloaded=new HashSet();
    private static ChJTable refTable;
    private JLabel labMsg=new JLabel(" ");
/* "No","Authority","Title","Coordinate source","Example entry",  "Log", "List Created" */
    private final static int
        COLUMN_NO=0,
        COLUMN_AUTHORITY=1,
        COLUMN_TITLE=2,
        COLUMN_SOURCE=3,
        COLUMN_AVAILABLE=4,
        COLUMN_LOG=5,
        COLUMN_LIST=6,
        FIELD_ENTRY=10,
        FIELD_coordinates=11,
        FIELD_V_AA=12,
        FIELD_JLIST_AA=13,
        FIELD_LOG=14,
        FIELD_EXAMPLE_URL=15;
    /* ---------------------------------------- */
    /* >>> readRegistry >>> */

    private static Object[][][] featureTableData;
    private static Object[][][] getFeatureTableData() {
        final Entry ee[]=DAS.readRegistry((ChRunnable)instance());
        synchronized("SDAS$$syncFTD") {
        Object rr[][][]=StrapDAS.featureTableData;
        if (rr==null) {
            final String AA[]=DAS.AUTHORITIES;
            rr=StrapDAS.featureTableData=new Object[AA.length][][];
            final List<Object[]> v=new ArrayList();
            final ByteArray baTmp=new ByteArray(99);
            for(int iA=AA.length;--iA>=0; ) {
                v.clear();
                for(int iE=0, no=0; iE<ee.length; iE++) {
                    final Entry e=ee[iE];
                    final String uri=e.getQueryUri(DAS.CAPABILITY_TYPE_feature);
                    if (uri==null) continue;
                    for(DAS.Coordinates coordinates : e.getCoordinates()) {
                        final String a=coordinates.getAuthority();
                        if (iA==AA.length-1 ? idxOfStrg(a,AA)>=0 : AA[iA]!=a) continue;
                        /* "No","Authority","Title","Coordinate source","Example entry",  "Log", "List Created" */
                        final Object[] row=new Object[16];
                        row[COLUMN_NO]=intObjct(++no);
                        row[COLUMN_AUTHORITY]=a;
                        row[COLUMN_TITLE]=e.getTitle();
                        row[COLUMN_SOURCE]=coordinates.getSource();
                        row[FIELD_coordinates]=coordinates;
                        if(coordinates.getTestRange()!=null) {
                            final String urlEx=baTmp.setEnd(0).append(uri).append("?segment=").urlEncode(coordinates.getTestRange()).newString();
                            row[FIELD_EXAMPLE_URL]=urlEx;
                            baTmp.setEnd(0).filter(FILENM, e.getTitle()+"_"+coordinates.getTestRange()+".xml", 0, MAX_INT, '_');
                            sbExamples.append("wget -N -T 10 -O ").append((Object)baTmp).append(' ').append(urlEx).append('\n');
                        }
                        row[FIELD_ENTRY]=e;
                        v.add(row);
                    }
                }
                rr[iA]=toArry(v,Object[].class);
            }
        }
        return rmNullA(rr, Object[][].class);
        }
    }
    /* <<< Get <<< */
    /* ---------------------------------------- */
    /* >>> For a given Protein  >>> */
    public static Thread thread_load(long option, String title[], Protein pp[], ChRunnable log) {
        final Object[] para={longObjct(option), title, pp, log};
        return threadM("load",StrapDAS.class, para);
    }
    public static void load(long option, String titles[], Protein pp[], ChRunnable log) {
        final Object registry[][][]=getFeatureTableData();
        if (registry==null) { putln(RED_ERROR+"StrapDAS: registry==null"); return;}
        final List<Object[]> vRows=new ArrayList();
        for(String title : titles) {
            if (sze(title)==0) continue;
            final String t=rplcToStrg("%20"," ",title);
            boolean success=false;
            for(Object[][] table:registry) {
                for(Object[] row : table) {
                    if (eqNz(t,row[COLUMN_TITLE])) {
                        vRows.add(row);
                        success=true;
                    }
                }
            }
            if (!success) putln(RED_ERROR+" No such DAS feature ",title);
        }
        loadFeaturesInRows(0L, toArry(vRows,Object[].class), pp, log);
    }
    private final static Object KEY_DAS_F=new Object();
    /* <<< For Protein <<< */
    /* ---------------------------------------- */
    /* >>> For Proteins >>> */

    public static Thread thread_loadFeaturesInRows(long options, Object[] rows[], Protein pp[],  ChRunnable log) {
        return threadM("loadFeaturesInRows",StrapDAS.class, new Object[]{longObjct(options), rows,pp, log});
    }

    public static void loadFeaturesInRows(long options, Object[] rows[], Protein pp[], ChRunnable log) {
        if (sze(rows)==0) return;
        pdb2uniprotP(pp,log);
        setEnabld(false, butGO);
        {
            final Set<String> v=new HashSet();
            for(int iP=sze(pp); --iP>=0;) {
                if (pp[iP]==null) continue;
                v.add(pp[iP].getAccessionID());
                adAll(pp[iP].getDatabaseRefs(), v);
                adAll(pp[iP].getSequenceRefs(), v);
            }
            FetchEBI_SOAP.thread_download(FetchEBI_SOAP.SKIP_EXISTING, strgArry(v)).start();
        }

        for(int iRow=0;iRow<rows.length; iRow++) {
            final Object row[]=rows[iRow];

            for(int iP=0; iP<pp.length; iP++) {
                final StrapProtein p=get(iP,pp, StrapProtein.class);
                if (p==null) continue;
                if (log!=null)  {
                    final String msg="StrapDAS "+row[COLUMN_AUTHORITY]+" "+row[COLUMN_TITLE]+ " "+p+" ...";
                    log.run(ChRunnable.PROGRESS, progressReport(StrapDAS.class,iRow*pp.length+iP +1,rows.length*pp.length+1,msg));
                }
                synchronized(mkIdObjct("SDAS$$syncp",p)) {
                    _loadFeatures(options, row,p, log);
                }
                SequenceFeatures.updateTable(333);
            }
            final String msg="Das "+row[COLUMN_AUTHORITY]+" done for "+pp.length+" proteins and "+rows.length+" types";
            if (log!=null) log.run(ChRunnable.PROGRESS, progressReport(StrapDAS.class,1,1, msg));

        }
        setEnabld(true, butGO);
    }
    private static void _loadFeatures(long option, Object[] row, StrapProtein p, ChRunnable globalLog) {
        if (p==null) return;
        final StrapAlign align=StrapAlign.getInstance();
        final JTable jTable=focusedTable();
        final Entry entry=(Entry)row[FIELD_ENTRY];
        final ChRunnable log=(ChRunnable)(row[FIELD_LOG]==null ? row[FIELD_LOG]=new AsciiView(99) : row[FIELD_LOG]);
        final List listRA=(List) (row[FIELD_V_AA]==null ? row[FIELD_V_AA]=new ArrayList() : row[FIELD_V_AA]);

        final String key="SDAS$$SYNC";
        Object sync=gcp(key,p);
        if (sync==null) pcp(key,sync=new Object(), p);

        nextProtein:
        synchronized(sync) {
            log.run(ChRunnable.APPEND,p+"\n\n");
            final ResidueAnnotation aa[]=p.getResidueAnnotations();
            for(ResidueAnnotation a : aa) if (gcp(KEY_DAS_F, a)!=null) return;
            final String authority=(String)row[COLUMN_AUTHORITY];
            int count=0;

            final List<String> vId=new ArrayList();
            if (0!=(option&SequenceFeatures.OPT_ID_BY_HOMOLOGY)) adAllUniq(FindUniprotID.ids(false,p), vId);
            else {
                adUniq(p.getAccessionID(),vId);
                adAllUniq(p.getSequenceRefs(),vId);
                adAllUniq(p.getDatabaseRefs(),vId);
            }

            for(boolean download : FALSE_TRUE) {
                for(String ref : strgArry(vId)) {
                    if (frame!=null && (option&STOP_ON_CLOSED_FRAME)>0 && !frame.isVisible()) break nextProtein;
                    if (ref==null) continue;
                    final boolean isResnum=eqNz("PDBresnum",authority);
                    final String aut=isResnum ? "PDB" : authority;
                    final boolean equls=equls(STRSTR_IC, aut, ref) && chrAt(aut.length(),ref)==':';
                    if (equls) {
                        final String urlSeq=Hyperrefs.toUrlString(ref, Hyperrefs.PROTEIN_FILE);
                        log.run(APPEND,"reference sequence  ref = "+ref+"\nreference sequence  url = "+urlSeq+"\n");
                        final File
                            ffEquivalent[]={FetchEBI_SOAP.dbColonID2file(ref)},
                            fSeq=isResnum ? null : download_TS(download ? url(urlSeq) : null, 999, ffEquivalent, globalLog);
                            if (!download && sze(fSeq)==0) continue;
                            log.run(APPEND,"reference sequence f = "+fSeq+" "+sze(fSeq)+" Bytes\n");
                            if (!isResnum && sze(fSeq)<20) continue;
                            final char chain=ref.startsWith("UNIPROT:") ? 0 : chrAt(0,p.getChainsAsString());
                            final ByteArray ba=ChUtils.readBytes(DAS.downloadURL(ref,chain,  entry, DAS.CAPABILITY_TYPE_feature, globalLog));
                            final java.net.URL url=DAS.getURL(ref,chain,  entry, DAS.CAPABILITY_TYPE_feature);
                            log.run(APPEND,"url="+url+"\n");
                            if (ba==null) continue;
                            final ResidueAnnotation[] aaNew=ExpasyGff.parseFeatures(0, ba, url,   p, authority, (Set)null);
                            for(ResidueAnnotation a : aaNew) {
                                listRA.add(weakRef(a));
                                ChDelay.revalidate((JComponent)row[FIELD_JLIST_AA],999, (long[])null);
                                if (a.isEnabled()) count++;
                                a.addEnS("RefSequence", ref+" "+filepathUnixStyle(fSeq));
                                a.addEnS(CLASS_Hyperrefs,toStrg(url));
                            }
                            if (jTable!=null) { jTable.revalidate();  jTable.repaint(); }
                            log.run(APPEND,ANSI_FG_GREEN+"Created "+ANSI_RESET+" ResidueAnnotation\n");
                    }
                    if (count>0) break;
                }
            }
            if (count>0) {
                Protein.incrementModificationCount(ProteinModiCount.MCA_SEQUENCE_FEATURES_V,null);
                StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_ADDED, 999);
            }
            log.run(APPEND, "/////\n");
        }
    }

    /* <<<  For a given Protein <<< */
    /* ---------------------------------------- */
    /* >>> Layout >>> */
    private static ChFrame frame;
    private static boolean initialized;
    private JComponent panel;
    public static void show(boolean downloadExamples) {
        if (!initialized) {
            initialized=true;
            ChDelay.newThread(instance(),"REGISTRY",  ChDelay.newThread(instance(),"AFTER_REGISTRY",null)).start();
        }
        StrapDAS.downloadExamples=downloadExamples;
        if (frame==null) frame=new ChFrame("DAS sources").size(666,444).ad(instance().getPanel());
        frame.shw(ChFrame.AT_CLICK);
    }
    private JComponent getPanel() {
        if (panel==null) {
            if (progress==null) progress=new JProgressBar();
            tabbed=new JTabbedPane();
            final JComponent pNorth=newCNSEW(pnl("<h2>BioDAS Entries</h2>", "(Under construction)"),null,null, panHS(StrapDAS.class), null);
            taSouth=new ChTextArea(DAS.URL_REGISTRY);
            pcp(ACTION_DATABASE_CLICKED,"Load",taSouth);
            taSouth.setLineWrap(true);
            butGO=new ChButton().li(li).like(ChButton.BUTTON_GO);
            pButtons=newCNSEW(pnl("Load DAS features for selected proteins",butGO, "  ", SequenceFeatures.butFindUniprot()), null, labMsg,null,null);
            repaintOnMouseEnterAndExit(false,pButtons);
            final JComponent pSouth=newCNSEW(taSouth,progress, pButtons,null,null);
            panel=newCNSEW(tabbed,pNorth, pSouth,null,null);
            repaintOnMouseEnterAndExit(false,labMsg);
        }
        return panel;
    }
    /* <<< Layout <<< */
    /* ---------------------------------------- */
    /* >>> Renderer >>> */
    @Override public java.awt.Component getTableCellRendererComponent(JTable t, final Object value0,boolean selected,boolean focus,int row0,int col){
        if (butRenderer==null) butRenderer=new ChButton("",ChButton.PAINT_IN_TABLE);
        Object value=value0;
        final int row=ChJTable.rowIndexToModel(row0,t);
        if (col==COLUMN_AVAILABLE) {
            final Object[] rowData=get(row, gcp("D",t,Object[][].class) , Object[].class);
            Object val=rowData[COLUMN_AVAILABLE];
            if (val==null) {
                final Object url=rowData[FIELD_EXAMPLE_URL];
                if (downloadExamples && !setDownloaded.contains(url)) {
                    setDownloaded.add(url);
                    ChDelay.newThread(this,"DL",url).start();
                }
                final File f=file(url);
                ByteArray ba=sze(f)>0 ? readBytes(f) : null;
                if (ba!=null) {
                    final byte[] T=ba.getBytes();
                    final int E=ba.getEnd();
                    if (strstr("</FEATURE>", T, 0, E)>0) {
                        value="No sequence range in feature";
                        for(int iStart=0; (iStart=strstr(STRSTR_AFTER,"<START>",  T, iStart,E))>0;) {
                            if (atoi(T,iStart)>0) value="OK";
                        }
                    } else { value="Error"; delFileOnExit(f); }
                }
                rowData[COLUMN_AVAILABLE]=value;
            }
        }
        if (col==COLUMN_AVAILABLE && value!=null) return butRenderer.t((String)value);
        if (col==COLUMN_LIST || col==COLUMN_LOG) {
            final int iTab=tabbed.getSelectedIndex();
            if (iTab>=0) {
                Object rowData[]=get(row, featureTableData[iTab], Object[].class);
                final List v=(List) get(FIELD_V_AA,rowData);
                if (col==COLUMN_LOG && v!=null ||  col==COLUMN_LIST && sze(v)>0) {
                    final String label= col==COLUMN_LOG ? "Log" : col==COLUMN_LIST ? "# "+sze(v) : "";
                    return butRenderer.t(label);
                }
            }
        }
        return super.getTableCellRendererComponent(t,value,selected,focus,row, col);
    }
    /* <<< Renderer  <<< */
    /* ---------------------------------------- */
    /* >>> Threading >>> */
    public Object run(String id, Object arg) {
        final Object argv[]=arg instanceof Object[] ? (Object[])arg : null;
        if (id==ChRunnable.APPEND) drawMsg(toStrg(arg), labMsg);
        if (id==ChRunnable.PROGRESS) {
            SequenceFeatures.instance().run(id,arg);
            setJPrgrss(progress,argv);
        }

        if (id=="DL") {
            download_TS(url(arg), CacheResult.isEnabled() ? DAS.DAYS_VALID :0, (File[])null, (ChRunnable)this);
            ChDelay.repaint(refTable,555);
        }
        if (id==ChRunnable.SAY_DOWNLOADING) drawMsg(toStrg(arg), pButtons);
        if (id=="REGISTRY") {
            getFeatureTableData();
            inEDT((Runnable)arg);
        }
        if (tabbed!=null) {
            if (id==ChRunnable.DOWNLOAD_FINISHED) tabbed.repaint();
            final Object[][][] reg=StrapDAS.featureTableData;
            if (id=="AFTER_REGISTRY" && reg!=null) {
                tabbed.removeAll();
                for(int iCoord=0; iCoord<DAS.AUTHORITIES.length; iCoord++) {
                    final ChTableModel m=new ChTableModel("No","Authority","Title","Coordinate source","Example entry",  "Log", "List Created");
                    m.setData(reg[iCoord]);
                    final ChJTable t=StrapDAS.refTable=new ChJTable(m,0).renderer(this);
                    ChJTable.setRowSorter(new Comparator[]{t},t);
                    t.setColumnWidth(false, 0,ChJTable.COLUMN_WIDTH_4_DIGITS);
                    t.setRowHeight(EX+8);
                    pcp("D",reg[iCoord],t);
                    li.addTo("m",t);
                    adTab(DAS.AUTHORITIES[iCoord],  scrllpn(SCRLLPN_SEARCH_BUTTON,t), tabbed);
                    if (iCoord==0 && sze(reg[iCoord])>1) t.setRowSelectionInterval(0, 4);
                }
                tabbed.setSelectedIndex(0);
                li.addTo("C",tabbed);
            }
        }
        return null;
    }
    /* <<< Threading <<< */
    /* ---------------------------------------- */
    /* >>> AWTEvent >>> */
    private static ChJTable focusedTable() {
        final JTabbedPane tp=_inst!=null ? _inst.tabbed : null;
        return tp!=null ? child(tabbed.getSelectedComponent(), ChJTable.class) : null;
    }
    public java.awt.AWTEvent processEvent(JComponent c, java.awt.AWTEvent ev) {
        final int id=ev.getID();
        final Object q=ev.getSource();
        final String cmd=actionCommand(ev);
        final ChJTable t=focusedTable();
        if (t==null) return ev;
        final int rr[]=ChJTable.selectedRowsConverted(t);
        final Object[][] rows=gcp("D",t,Object[][].class);
        if (q==t && (id==MOUSE_CLICKED || cmd==ACTION_CHANGE_EVENT)) butGO.enabled(sze(rr)>0);
        if (id==MOUSE_CLICKED) {
            final int row=ChJTable.rowIndexToModel(t.rowAtPoint(point(ev)),t);
            final int col=t.columnAtPoint(point(ev));
            final Object[] rowData=get(row, rows, Object[].class);
            if (rowData!=null) {
                final Entry entry=(Entry)rowData[FIELD_ENTRY];
                final DAS.Coordinates coordinates=(DAS.Coordinates)rowData[FIELD_coordinates];
                final List<ResidueAnnotation> vAA= get(FIELD_V_AA, rowData,List.class);
                final String title= (String)rowData[COLUMN_TITLE];
                final AsciiView log=(AsciiView)rowData[FIELD_LOG];
                if (col==COLUMN_LOG && log!=null) {
                    pcp(ACTION_DATABASE_CLICKED,"Load",log);
                    ChFrame.frame(title, log, ChFrame.SCROLLPANE|CLOSE_ON_CtrlW).shw(ChFrame.AT_CLICK);
                    return null;
                } else if (col==COLUMN_LIST && sze(vAA)>0) {
                    final Object pSouth=pnl("Note: the list supports ","WIKI:Drag_and_drop");
                    if (rowData[FIELD_JLIST_AA]==null) rowData[FIELD_JLIST_AA]=new Object();
                    StrapAlign.showInJList(0L, rowData[FIELD_JLIST_AA], vAA.toArray(),title, pSouth).showInFrame(ChFrame.AT_CLICK|ChFrame.SIZE_REGARD_RESOLUTION, title);
                    return null;
                } else if (col==COLUMN_AVAILABLE) {
                    final File f=file(rowData[FIELD_EXAMPLE_URL]);
                    final ByteArray ba=readBytes(f);
                    if (ba!=null) {
                        ba.append("\n\nLoaded From ").append(rowData[FIELD_EXAMPLE_URL]).append("\nFile: ").append(f);
                        new AsciiView(ba).showInFrame((String)rowData[COLUMN_TITLE]).highlightXML(ba);
                    }
                }
                if (sze(rr)==1) {
                    final StringBuffer sb=new StringBuffer(333);
                    final Object testRange=coordinates.getTestRange();
                    if (rowData[COLUMN_AVAILABLE]==null) {
                        final Object url=rowData[FIELD_EXAMPLE_URL];
                        sb.append("Testing: "+uCase(rplcToStrg("PDBresnum","PDB",(String)rowData[COLUMN_AUTHORITY]))+":"+testRange+"  "+url+"  "+filepathUnixStyle(url(url)));
                    } else sb.append(testRange);
                    taSouth.t(sb.append('\n').append(entry.getDocHref()).append("\n\n").append(entry.getDescription()));
                } else taSouth.t(DAS.URL_REGISTRY);
                taSouth.tools().underlineRefs(UNDERLINE_REFS);
            }
        }
        if (q==butGO && sze(rr)>0) {
            final StrapAlign a=StrapAlign.getInstance();
            if (a==null) return ev;
            StrapProtein pp[]=a.getSelectedProteins();
            if (pp.length==0) pp=a.getVisibleProteins();
            SequenceFeatures.instance().showPanel();
            final List vRows=new ArrayList();
            for(int i=sze(rr); --i>=0;) adUniq(get(rr[i],rows), vRows);
            thread_loadFeaturesInRows(STOP_ON_CLOSED_FRAME, toArry(vRows,Object[].class), pp, (ChRunnable)instance() ).start();
        }
        return ev;
    }

    /* <<< AWTEvent <<< */
    /* ---------------------------------------- */
    /* >>> pdb2uniprot >>> */
    public static void pdb2uniprotP(Protein pp[], ChRunnable log) {
        final String key="SDAS$$p2u";
        for(int i=sze(pp); --i>=0;) {
            final Protein p=pp[i];
            if (p==null) continue;
            synchronized(mkIdObjct(key,p)) {
                final String pdbID=p.getPdbID();
                if (p.getUniprotID()==null && pdbID!=null) {
                    p.setUniprotID(DAS.pdb2uniprot(pdbID,chrAt(0,p.getChainsAsString()), log));
                }
            }
        }
    }
}
