package charite.christo.strap;
import charite.christo.hotswap.ProxyClass;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import charite.christo.strap.interfaces.*;
import javax.swing.*;
import java.awt.*;
import java.io.File;
import java.util.List;
import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static java.awt.event.MouseEvent.*;

/**HELP
The hotswap-dialog provides a comfortable environment for writing plugins.
<i>PACKAGE:charite.christo.protein.</i>
<i>PACKAGE:charite.christo.strap.extensions.</i>
The concept of class-hotswap has been published in PUBMED:16469097.
<br><br><b>Quick start:</b> Follow this short guide:
Activate the node <b>"StrapExtension"</b> in the tree.
Subsequently, press the button
<i>BUTTON:StrapHotswap#BUT_LAB_New</i>
to create a new plugin of type "StrapExtension".
You will be asked for a name of your new plugin. You can accept the
suggested name.   The WIKI:Source_code of the
new plugin  is opened with a text editor.
The source is a minimal example for "StrapExtension".
To start  press <i>BUTTON:StrapHotswap#BUT_LAB_Apply</i>.
All, what the plugin is doing is that it tells in a message box  how many proteins are currently loaded.
The program code  can be modified and extended.
For example the text in the message box can be changed.
When the altered source code is saved to hard disk  the  WIKI:Compiler is automatically started. The result of successful compilation is a WIKI:Class_file.
However, if the compilation fails, the WIKI:Compilation_error is shown in the log panel. This panel is in another tab. you need to go back to the main tab.
The next time, the plugin is started, the new text will be seen in the message box.
<br><hr><b>Hotswapped classes modified by the user:</b>
The Java source- and class-files of the hotswap plugins reside in the  directory  "~/.StrapAlign/hotswap".
In contrast to standard plugins hotswapped classes must not have a package statement.
<br><br><b>JAR-files:</b>
All WIKI:Jar_files in the directories ~/.StrapAlign/plugins  and  ~/.StrapAlign/jars are included to the WIKI:Classpath_(Java).
<br><br><b> Types of plugins:</b>
Each plugin implements  a certain WIKI:Java_interface which determins the purpose of the plugin.
The interfaces and plugins are shown in the graphical tree.
The proximal branches of the tree are the interfaces such as <i>JAVADOC:StrapExtension</i> or <i>JAVADOC:ProteinParser</i>.
When these nodes are expanded the classes which implement the interface become visible.
For example <i>JAVADOC:ConsensusSequence</i> is a plugin implementing
<i>JAVADOC:StrapExtension</i> whereas <i>JAVADOC:ResidueSolventAccessibility</i> is a plugin
implementing <i>JAVADOC:ValueOfResidue</i>.
When you select <i>JAVADOC:ConsensusSequence</i> in the tree the button
<i>BUTTON:StrapHotswap#BUT_LAB_Apply</i> becomes active.
On pressing this button  an instance of this
selected plugin is created and an appropriate action performed.
The <i>JAVADOC:ConsensusSequence</i> implements <i>JAVADOC:StrapExtension</i> and has
therefore a <I>run()</I> method. The specific action for this type of plugins is to call the  <I>run()</I>-method.
On the other hand <i>JAVADOC:ResidueSolventAccessibility</i> implements
<i>JAVADOC:ValueOfResidue</i>.
Therefore it works in the context of a certain dialog. This dialog is
opened and <i>JAVADOC:ValueOfResidue</i> is selected within the
dialog.
<br><br>
<br><br><b>Developing/changing plugins:</b>
Innovative in STRAP is that plugins can be loaded, <b>created and changed at runtime</b>.
Plugins do not need to be re-instantiated after changing the Java-source.
This is possible because STRAP does not communicate directly with the object but with a proxy object.
The technique was inspired by the hotswap API by Paul Cody Johnston .
<br>
The buttons act on the class or interface that is selected in the tree by the user.
<ul>
<li><i>BUTTON:StrapHotswap#BUT_LAB_New</i>
This button is used to create a new plugin. The tree node of the interface to be implemented must be selected.
The useer can chose a name of the new plugin.
A skeleton of a class implementing the selected interface is written to the directory <i>JAVADOC:StrapPlugins</i> and can be modified in
a text editor.
Whenever the text is saved to hard disk it is automatically compiled and compiler messages are displayed.
</li>
<li><i>BUTTON:StrapHotswap#BUT_LAB_Edit</i> Source code of user
defined interfaces and classes is shown and and plugin code is loaded
into the text editor. As soon as the Java-file is saved STRAP tries to
compile the Java-file. When the compilation succeeds the changes take
effect immediately.  </li>
</ul>
<b>Debugging:</b> In the "debuggin" menu
which resides in the "option" menu are some tools to find programming
errors in plugins.  The standard outputs <I>stderr</I> and
<I>stdout</I> which are written with the Java-statements
<i>System.out.println("hello world")</i> or <I>System.err.println("hello
world")</I>, respectively, can be viewed. Classes implementing
<i>JAVADOC:StrapListener</i> are notified when the cursor moves, when
proteins are added or deleted or when the multiple sequence alignment
is altered.  This information is carried in instances of
<i>JAVADOC:StrapEvent</i>.  Events can be monitored:
<i>DIALOG:StrapEventWatcher!</i>.
<br><br><b>Customization:</b>
<i>ICON:IC_CUSTOM</i> The preferred  WIKI:Text_editor or  WIKI:Integrated_development_environment used for Java source texts can be specified with the customize button at the top bar.
<br><br><b>Accessing data of the multiple sequence alignment by plugins:</b>
For the communication with the STRAP-model four classes are usually sufficient:
<ul>
<li><b><i>JAVADOC:StrapAlign</i></b>
The root object. It can be obtained with the method <I>getProteinAlignment()</I> from event- or protein-objects.
or with the static method <i>JAVADOC:StrapAlign#getInstance()</i>.
</li>
<li><b><i>JAVADOC:StrapEvent</i></b>
Events are broadcasted to inform all listeners when changes occur. Try <i>DIALOG:StrapEventWatcher</i>.
</li>
<li><b><i>JAVADOC:StrapProtein</i></b></li>
</ul>
<br><br><b>Arrays are read only:</b>
Some methods return arrays.  Never change array elements in returned arrays. The following is wrong:
<pre class="data">
StrapProtein p= ....;
byte[] aminoAcids=p.getResidueType();
aminoAcids[4]=(byte)'Y';</pre>
In order to change the 5th residue to Tyrosine write instead:
<pre class="data">
StrapProtein p= ....;
p.setResidueType(4,(byte)'Y');</pre>
<br><br><b>Using public methods of STRAP classes :</b>
If possible, use documented  methods description because they will be kept in future releases whereas un-documented methods may change in the future.
<br>
See <i>DIALOG:StrapPlugins</i>
<br><br>Publication: PUBMED:16469097
*/
public class StrapHotswap implements HasPanel, ChRunnable, ChProcessEvent, PaintHook {
    /* --- Buttons --- */
    public final  static String
        BUT_LAB_Apply="Apply or <br>start plugin",
        BUT_LAB_Edit="Edit/view<br> Java source",
        BUT_LAB_New="New <br>plugin";
    private final JTabbedPane tabPane=new JTabbedPane();

    private final ChJTree jTree;
    public StrapHotswap() {
        setSettings(this, Customize.javaSourceEditor);
        mkdrsErr(dirHotswap(),"Here hotswap-plugins are stored");
        final Object[] root={
            vNO_INTERFACE,
            vNOT_COMPILED,
            TabItemTipIcon.set("","Interfaces","","",StrapPlugins.getInterfacesAndClasses())
        };
        jTree=new ChJTree(0L, new ChTreeModel(0L,root));
        jTree.setCellRenderer(new ChRenderer().options(ChRenderer.CLASS_NAMES));
        LI.addTo("m",jTree);
        jTree.setRootVisible(false);
        jTree.setShowsRootHandles(true);
        scrollByWheel(jTree);
        addPaintHook(this,jTree);
        final AsciiView log=ChCompilerJavacPROXY.log();
        final JComponent
            spTree=scrllpn(0L, jTree),
            belowTop=pnl(dirHotswap()),
            pNorth=vBpnl(StrapAlign.getInstance().headLine(this),belowTop),
            panButtons=pnl(
                           new GridLayout(7,1),
                           butNew,butEdit,"",
                           butApply,"",
                           ChButton.actionViewMessage("Log <br>messages",ChClassLoader1.log()),
                           SourcefilesInJarfile.newButton().t("List all<br>source files")
                           ),
            lComponent=new OverlayScrollPane(log).addButton(log.tools().newClearButton().li(LI)).addButton(togCompiler.cb()),
            rComponent=newCNSEW(spTree,pNorth,null,null,panButtons);
        tabPane.setMinimumSize(dim(9,9));
        adTab("Interfaces and plugins",rComponent, tabPane);
        adTab("Compiler messages",lComponent, tabPane);
        TabItemTipIcon.set("","","Hotswap plugins",IC_HOTPLUG,tabPane);
        enableDisable();
    }
    /* <<< Constructor <<< */
    /* ---------------------------------------- */
     /* >>> AWTEvent >>> */
    private final ChEventAdapter LI=new ChEventAdapter(this);
    public AWTEvent processEvent(JComponent component, AWTEvent ev) {
        final int id=ev.getID();
        final Object q=ev.getSource();
        final String cmd=actionCommand(ev);
        final boolean ctrl=isCtrl(ev), shift=isShift(ev);
        final File file=this.theFile;
        if (cmd!=null) {
            StrapAlign.errorMsg("");
            if (cmd==CMD_CLEAR) tabPane.setSelectedIndex(0);
            if (q==togCompiler) {
                pcp(KEY_IF_EMPTY,null,ChCompilerJavacPROXY.log());
                ChCompilerJavacPROXY.log(togCompiler.s() ?
                                     "the compiler will automatically compile\nwhen java files are modified.\n" :
                                     "Internal compiler is switched off.\n\n"+
                                     "STRAP expects that the compiler is invoked by the program\n"+
                                     "that you use to edit the Java file.\n\n\n");
            }
            if (q==butNew) {
                if (shift) jTree.revalidate();
                else if (ctrl) jTree.updateUI();
                else create(theInterface);
            }
            if (q==butApply) {
                if(proxyClass!=null) {
                    if (ctrl) assignNewProxy(proxyClass.getSourcefile());
                    StrapPlugins.applyPlugin(proxyClass);
                }
                else StrapPlugins.applyPlugin(theClass);
            }
            if (q==butEdit) {
                if (file!=null) {
                    Customize c=Customize.customize(Customize.javaSourceEditor);
                    ChExec ex=new ChExec(0).setCommandsAndArgument(c.getSettings(),file.getAbsolutePath());
                    if (ctrl) ex.log(c);
                    new Thread(ex).start();
                }
                if (!shift) {
                    final Class c=theClass!=null ? theClass : theInterface;
                    if (c!=null) showJavaSource(c.getName(),(String[])null);
                    if (file!=null && proxyClass!=null) new ChTextViewer(file).shw(0).title(proxyClass.getName());
                }
            }
        }
        if (q==jTree) {
            if (id==MOUSE_CLICKED) {
                Object o=objectAt(ev);
                if (o instanceof UniqueList) o=((UniqueList)o).run(MAY_PROVIDE_TREE_VALUE,null);
                if (o!=null && this.selection!=o) selectionChanged(this.selection=o);
            }
        }
        return ev;
    }
    /* <<< AWTEvent <<< */
    /* ---------------------------------------- */
    /* >>> File >>> */
  private  File className2SourceFile(String s){return file(dirHotswap(),s+".java");}
    private long lastModified;
    private final Set<String> vAllJavaFiles=new HashSet();
    private final static Map<File,File> mapJavafile2Classfile=new HashMap();
    private final static Map<File,Long> mapWhenCompiled=new HashMap();
    private long myUpdateUI_classefiles_LM;
    private String shortClassName(String s0) {
        final String s=delSfx(".java",s0);
        return s!=null ? s.substring(1+s.lastIndexOf(File.separatorChar)) : null;
    }
    private ProxyClass assignNewProxy(File f) {
        if (f==null) return null;
        final String className=withoutDotSuffix(f.getName());
        ProxyClass proxyClass=ProxyClass.getProxyClass(className);
        if (proxyClass!=null) { ProxyClass.remove(proxyClass); proxyClass=null;}
        proxyClass=ProxyClass.getProxyClass(f,className);
        toBeSelected=proxyClass;
        sortClasses();
        jTree.updateKeepStateLater(0L, 33);
        toBeSelected=proxyClass;
        ChDelay.runAfterMS(EDT,threadSelect,999);
        return proxyClass;
    }
    private ProxyClass toBeSelected;
    /* <<< File <<< */
    /* ---------------------------------------- */
    /* >>> StrapEvent >>> */
    private void dispatchStrapEvent(Class c) {
        if (c==null) return;
        final int type=
            ProteinProteinValue.class.isAssignableFrom(c) || SequenceDisSimilarity.class.isAssignableFrom(c) ? StrapEvent.PROTEIN_PROTEIN_DISTANCE_CHANGED:
            ValueOfAlignPosition.class.isAssignableFrom(c) ? StrapEvent.VALUE_OF_ALIGN_POSITION_CHANGED:
            ValueOfProtein.class.isAssignableFrom(c) ? StrapEvent.VALUE_OF_PROTEIN_CHANGED:
            ValueOfResidue.class.isAssignableFrom(c) ? StrapEvent.VALUE_OF_RESIDUE_CHANGED:
            ResidueSelection.class.isAssignableFrom(c) ? StrapEvent.RESIDUE_SELECTION_CHANGED: 0;
        if (type>0) {
            ChDelay.runAfterMS(EDT,new StrapEvent(this,type),999);
            ChDelay.runAfterMS(EDT,new StrapEvent(this,type),2999);
        }
        ChDelay.runAfterMS(EDT,new StrapEvent(this,StrapEvent.PLUGIN_CHANGED),99);
    }
    /* <<< StrapEvent  <<< */
    /* ---------------------------------------- */
    /* >>>Threading  >>> */
    private final Runnable threadSelect=ChDelay.newThread(this,"SELECT",null);
    private boolean wasPainted, isRunning=true;
    public boolean paintHook(JComponent component, Graphics g, boolean after) {
        if (!wasPainted) {
            wasPainted=true;
            run("UPDATE",null);
            ChDelay.newThread(StrapHotswap.this,"CK",null).start();
            jTree.expandPath(jTree.getPathForRow(2));
        }
        return true;
    }
    public Object run(String id,Object arg) {
        final boolean isWin=isWindows();
        if (id=="UPDATE") {
            final long time=System.currentTimeMillis();
            //ProxyClass.setClassLoader(ChClassLoader1.getInstance(new File[]{dirHotswap()}));
            vAllJavaFiles.clear();
            final File ff[]=dirHotswap().listFiles();
            final boolean isCompiler=togCompiler.s();
            if (ff==null) return null;
            for(File f : ff) {
                if (!isCompiler || !f.getName().endsWith(".java") || !is(LETTR,f.getName(), 0)) continue;
                File fClass=mapJavafile2Classfile.get(f);
                if (fClass==null) mapJavafile2Classfile.put(f,fClass=file(delSfx("java",f)+"class"));
                Long whenCompiled=mapWhenCompiled.get(f);
                if (whenCompiled==null) whenCompiled=sze(fClass)>0 ? new Long(fClass.lastModified()) : null;
                if (atol(whenCompiled)< f.lastModified()) {
                    new ChCompilerJavacPROXY().compile(classpth(true, dirHotswap()), f);
                    tabPane.setSelectedIndex(1);
                    mapWhenCompiled.put(f,new Long(System.currentTimeMillis()));
                }
                ChDelay.afterMS(EDT,ChDelay.BG_COLOR,ChCompilerJavacPROXY.log(),111, C(0xFFffFF));
            }
            for(File f : ff) {
                final File fClass=f.getName().endsWith(".java") ? mapJavafile2Classfile.get(f) : null;
                if (fClass==null) continue;
                final String className=shortClassName(f.getName());
                if (null==ProxyClass.getProxyClass(className))  assignNewProxy(f);
                vAllJavaFiles.add(className);
                if (fClass.lastModified()>myUpdateUI_classefiles_LM) {
                    final ProxyClass pc=ProxyClass.getProxyClass(className);
                    if (pc!=null) {
                        pc.hotswap();
                        if (pc!=null) {
                            final StringBuffer sb=new StringBuffer();
                            for(Class i:pc.getProxyInterfaces()) { String sn=i.getName(); sb.append(' ').append(sn.substring(sn.lastIndexOf('.')+1));}
                            final String s=sb.toString();
                            if (!s.equals(mapFile_interfaces.get(f)))  {
                                ChCompilerJavacPROXY.log(className+(" "+ANSI_YELLOW+"implements "+ANSI_RESET)+s+"\n");
                                mapFile_interfaces.put(f,s);
                                ProxyClass.getProxyClass(f,className);
                            }
                        }
                        dispatchStrapEvent(pc.getClassInstance());
                    } else ProxyClass.getProxyClass(f,className);
                }
            }
            for(ProxyClass pc : ProxyClass.getProxyClasses()) {
                if (!vAllJavaFiles.contains(pc.getName())) ProxyClass.remove(pc);
            }
            sortClasses();
            jTree.updateKeepStateLater(0L, 333);
            myUpdateUI_classefiles_LM=time;
        }
        if (id=="SELECT") {
            final javax.swing.tree.TreePath tp=toBeSelected!=null ? getTreePath(toBeSelected,jTree) : null;
            if (tp!=null) {
                jTree.clearSelection();
                jTree.addSelectionPath(tp);
                selectionChanged(toBeSelected);
            }
        }
        if (id=="EXPAND")  jTree.expandPath(getTreePath(arg,jTree));
        if (id=="CK"){
            long dirPlugins_lastModified=0;
            for(int count=0; isRunning; count++) {
                sleep(isWin ? 444 : 44);
                final long lm=dirHotswap().lastModified();
                if ((count&15)!=0 && lm==dirPlugins_lastModified) continue;
                dirPlugins_lastModified=lm;
                final boolean isCompiler=togCompiler.s();
                final File ff[]=dirHotswap().listFiles();
                if (ff==null) continue;
                long sum=0;
                for(File f : ff) {
                    if (f.getName().endsWith(isCompiler ? ".java" : ".class")) {
                        sum+=f.lastModified();
                    }
                }
                if (lastModified!=sum) {
                    lastModified=sum;
                    ChDelay.runAfterMS(EDT,this,"UPDATE",3);
                }
            }
        }
        return null;
    }
    /* <<< Threading <<< */
    /* ---------------------------------------- */
    /* >>> create >>> */
    private void create(final Class interFace) {
        if (interFace==null) return;
        final String ifName=shrtClasNam(interFace);
        String name="";
        for(int i=0;i<99;i++) {
            name="Test_"+ifName+"_"+i;
            if (!className2SourceFile(name).exists()) break;
        }
        ChDelay.runAfterMS(EDT, ChDelay.newThread(this,"EXPAND",StrapPlugins.allClassesV(interFace)),444);
        final String source;
        {
            String s=readStrg(getClass().getResourceAsStream("hotswapExamples/"+ifName+"Example.java"));
            if (sze(s)==0){StrapAlign.errorMsg("No example for "+shrtClasNam(interFace)+" available."); return;}
            source=s.replace('\t',' ').replaceAll(ifName+"Example",name);
        }
        boolean isAppropriate;
        do {
            String MSG=
                "Creating class of type "+ifName+".<br><br>"+
                "Enter a name for the new class.<br>"+
                "It should start with a capital letter and<br>"+
                "should contain only letters, digits and underscore<br>";
            name=ChMsg.input(ChMsg.AT_CLICK,MSG,30,name);
            if (name==null) return;
            isAppropriate=sze(name)>1 &&  is(UPPR,name,0);
            for(int i=sze(name);--i>=0;) {
                if (!is(LETTR_DIGT_US,name,i)) isAppropriate=false;
            }
        } while(!isAppropriate);
        final File f=className2SourceFile(name);
        if (f.exists()){
            StrapAlign.errorMsg("The file "+f+" already exists.");
            return;
        }
        wrte(f,source);
        if (!f.exists()){
            StrapAlign.errorMsg("Could not write  file "+f);
            return;
        }
        new Thread(new ChExec(0).setCommandsAndArgument(getCustomSettings(Customize.javaSourceEditor),f.getAbsolutePath())).start();
    }
    /* <<< create <<< */
    /* ---------------------------------------- */
    /* >>> GUI  >>> */
    private final ChButton
        butApply=new ChButton(BUT_LAB_Apply).li(LI)
        .tt("The activated class is loaded and instantiated.<br>Depending on the interfaces the class implements<br>an appropriate action is performed.<br>If the CTRL key is pressed simultaneously the class is discarded and reloaded so that recent changes take effect."),
        butNew=new ChButton(BUT_LAB_New).li(LI).tt("creates a new plugin implementing the selected interface"),
        butEdit=new ChButton(BUT_LAB_Edit).li(LI)
        .tt("loads the java source code into the editor<br>Unless you hold the CTRL key a second non editable view is opened.");
    private final ChButton togCompiler=toggl("Automatic compilation on/off").tt("Turn off if you use another compiler.").s(true).li(LI);
    private void enableDisable(){
        butEdit.setEnabled(this.theFile!=null||this.theClass!=null || this.proxyClass!=null || this.theInterface!=null);
        butNew.setEnabled(theInterface!=null);
        butApply.setEnabled(theClass!=null || proxyClass!=null);
    }
    /* <<< GUI <<< */
    /* ---------------------------------------- */
    /* >>> Selection  >>> */
    private File theFile;
    private Class theClass,theInterface;
    private ProxyClass proxyClass;
    private Object selection;
    private void selectionChanged(Object o) {
        this.theFile=null;
        this.theClass=this.theInterface=null;
        this.proxyClass=deref(o, ProxyClass.class);
        if (this.proxyClass!=null) this.theFile=this.proxyClass.getSourcefile();

        if (o instanceof Class) {
            final Class c=(Class)o;
            if (c.isInterface()) this.theInterface=c;
            else this.theFile=className2SourceFile(shrtClasNam(this.theClass=c));
        }
        if (o instanceof String) this.theFile=className2SourceFile((String)o);
        if (sze(this.theFile)==0) this.theFile=null;
        enableDisable();
        toBeSelected=null;
    }
    /* <<< Selection <<< */
    /* ---------------------------------------- */
    /* >>> Sort Classes >>> */
    private final List
        vNO_INTERFACE=new UniqueList(Object.class).t("Classes not implementing a proper interface").i(IC_BUG),
        vNOT_COMPILED=new UniqueList(Object.class).t("Java files failed to compile").i(IC_BUG);
    private final Map<File,String> mapFile_interfaces=new HashMap();
    private void sortClasses() {
        final Class II[]=StrapPlugins.interfaces();
        for(int i=II.length; --i>=0;) {
            final List l=StrapPlugins.allClassesV(i);
            for(int j=sze(l);--j>=0;) {
                final Object o=l.get(j);
                if (o instanceof ProxyClass) l.remove(o);
            }
        }
        vNOT_COMPILED.clear();
        vNO_INTERFACE.clear();
        for(String f: dirHotswap().list())
            if (f.endsWith(".java")  && is(LETTR,f,0))
                vNOT_COMPILED.add(withoutDotSuffix(f));
        for(ProxyClass pc : ProxyClass.getProxyClasses()) {
            final Class c=pc.getClassInstance();
            if (c!=null) {
                vNOT_COMPILED.remove(pc.getName());
                boolean hasInterface=false;
                for(int i=0;i<II.length; i++) {
                    final List l=StrapPlugins.allClassesV(i);
                    if (II[i].isAssignableFrom(c)) {
                        hasInterface=true;
                        if (!l.contains(pc)) { l.add(0,pc); }
                    } else if (l.contains(pc)) { l.remove(pc); }
                }
                if (!hasInterface && !vNO_INTERFACE.contains(pc)) vNO_INTERFACE.add(pc);
                else if (vNO_INTERFACE.contains(pc)) vNO_INTERFACE.remove(pc);
            } else vNOT_COMPILED.add(pc.getName());
        }
        StrapPlugins.setChanged();
        UIupdateList.update(UIupdateList.PLUGINS_ADDED_OR_REMOVED);
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Implemented Interfaces >>> */
    public JComponent getPanel(int options) {return tabPane;}
    public void dispose() { isRunning=false;}
    /* <<< Implemented Interfaces <<< */
    /* ---------------------------------------- */
    /* >>> static utils >>> */
}
