package charite.christo.strap;
import charite.christo.*;
import charite.christo.protein.*;
import charite.christo.interfaces.*;
import java.awt.*;
import java.net.URL;
import java.io.File;
import java.awt.event.*;
import java.util.*;
import java.util.List;
import javax.swing.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
import static charite.christo.strap.StrapAlign.*;

public class StrapScriptInterpreter implements ChRunnable {
    public final static String
    SCRIPT_use_3D_viewer="use_3D_viewer ",
        SCRIPT_to_3D="to_3D ",
        SCRIPT_select_3D=Protein3dUtils.SCRIPT_select_3D,
        SCRIPT_open_3D="open_3D ",
        SCRIPT_close_3D="close_3D ",
        SCRIPT_to_front_3D="to_front_3D ",
        SCRIPT_aa_sequence="aa_sequence ",
        SCRIPT_to_structure_viewer="to_structure_viewer ",
        SCRIPT_accession_id="accession_id ",
        SCRIPT_add_annotation="add_annotation ",
        SCRIPT_add_xref="add_xref ",
        SCRIPT_align="align ",
        SCRIPT_below_row="below_row ",
        SCRIPT_box="box ",
        SCRIPT_cds="cds ",
        SCRIPT_close="close ",
        SCRIPT_delete="delete ",
        SCRIPT_wire="wire ",
        SCRIPT_close_wire="close_wire ",
        SCRIPT_cursor="cursor ",
        SCRIPT_deiconify="deiconify",
        SCRIPT_gaps="gaps ",
        SCRIPT_hide="hide ",
        SCRIPT_icon="icon ",
        SCRIPT_iconify="iconify",
        SCRIPT_load="load ",
        SCRIPT_sleep="sleep ",
        SCRIPT_load_from_databases="load_from_databases ",
        SCRIPT_new_nucleotide_selection="new_nucleotide_selection ",
        SCRIPT_new_aminoacid_selection="new_selection ",
        SCRIPT_plugin="plugin ",
        SCRIPT_remove_xref="remove_xref ",
        SCRIPT_rotate_translate="rotate_translate ",
        SCRIPT_scroll_to="scroll_to ",
        SCRIPT_select="select ",
        SCRIPT_seqvista="seqvista ",
        SCRIPT_jalview="jalview ",
        SCRIPT_spice="spice ",
        SCRIPT_set_annotation="set_annotation ",
        SCRIPT_superimpose="superimpose ",
        SCRIPT_to_row="to_row ",
        SCRIPT_tree="tree ",
        SCRIPT_unhide="unhide ",
        SCRIPT_unselect="unselect ",
        SCRIPT_infer_xyz="project_coordinates ",
        SCRIPT_das="DAS_features ",
        SCRIPT_toFront="STRAP_to_front",
        SCRIPT_src="source_code ",
        SCRIPT_msec="mseconds",
        SCRIPT_exit="exit",
        SCRIPT_job_id="job_id",
        SCRIPT_wait_for="wait_for ",
        ANY_VIEWER="ANY_VIEWER", ANY_WIRE="ANY_WIRE";

    private String view3D_ID, nextJobID;
    private final Set<String> setJobs=new HashSet();
    private Protein[] view3D_pp;

    private static String s_view3D_ID;
    private static Protein[] s_view3D_pp;

    public Thread thread_interpretScript(CharSequence script) {
        return threadM("interpretScript", this, new Object[]{toByteArray(script)});
    }

    public void interpretScript(ByteArray script) {
        if (sze(script)==0) return;
        script.replace("new_aminoacid_selection ",SCRIPT_new_aminoacid_selection);
        if (!isEDT()) {
            inEDT(thread_interpretScript(script));
            return;
        }
        final int ends[]=script.getLineEnds();
        final byte[] T=script.getBytes();
        final ChTokenizer TOK=new ChTokenizer();
        final StrapAlign A=StrapAlign.getInstance();
        if (getAlignmentPanelWithFocus()==null) A.newAlignmentPanel();
        final StrapView view=getAlignmentPanelWithFocus();
        final String WNA="Error: wrong number of arguments ";
        String selectTablePanel=null;
        for(int iL=0;iL<ends.length; iL++) {
            progressBarTurn(4);
            final StrapProtein ppVis[]=getVisibleProteins();
            final int B=nxtE(NO_SPC,T,iL==0 ? script.getBegin() : ends[iL-1]+1,ends[iL]),
                E=prev(NO_SPC,T, ends[iL]-1,B-1)+1;
            if (E-B<2 || T[B]=='#') continue;
            final String cmd;
            final boolean is3DCmd;
            {
                String c=null;
                boolean is3D=false;
                for(String s : allScriptCommands()) {
                    if (B+s.length()<=E && equls(s,T,B)) { c=s; break; }
                }
                for(String s : Protein3dUtils.allCommands()) {
                    if (B+s.length()<=E && equls(s,T,B)) { c=s; is3D=true; break; }
                }
                is3DCmd=is3D;
                cmd=c;
            }
            if (cmd==SCRIPT_exit) break;
            final int spc=strchr(STRSTR_E,' ',T,B,E);
            final int noSpc=spc<0 ? E : nxtE(NO_SPC, T,spc,E);
            TOK.setText(T,noSpc,E).setDelimiters(charClassOneChar(','));

            final boolean parameter_pp=
                cmd==SCRIPT_infer_xyz || cmd==SCRIPT_gaps || cmd==SCRIPT_cds || cmd==SCRIPT_to_row ||
                cmd==SCRIPT_below_row || cmd==SCRIPT_rotate_translate || cmd==SCRIPT_das ||
                cmd==SCRIPT_add_xref || cmd==SCRIPT_accession_id || cmd==SCRIPT_remove_xref,
                only_pp=
                cmd==SCRIPT_load_from_databases ||
                cmd==SCRIPT_wire || cmd==SCRIPT_close_wire || cmd==SCRIPT_hide || cmd==SCRIPT_unhide || cmd==SCRIPT_tree ||
                cmd==SCRIPT_align  || cmd==SCRIPT_superimpose ||
                cmd==SCRIPT_seqvista || cmd==SCRIPT_jalview || cmd==SCRIPT_spice,
                parameter_pp_aa= cmd==SCRIPT_icon,
                only_pp_aa= cmd==SCRIPT_close || cmd==SCRIPT_delete  || cmd==SCRIPT_select || cmd==SCRIPT_unselect ||
                cmd==SCRIPT_scroll_to || cmd==SCRIPT_box || cmd==SCRIPT_cursor;

            final boolean parameter_p_or_new=cmd==SCRIPT_aa_sequence;

            CharSequence ensp=null, ensa=null, error=null;

            int evt=-1, evtLater=-1;
            int top=MAX_INT, bottom=-1, left=MAX_INT, right=-1;
            Object evtObj[]=null;

            final ByteArray proteinList;
            final String para;
            if (parameter_pp_aa || parameter_pp || parameter_p_or_new) {
                final int lastKomma=strchr(STRSTR_PREV,',',T,E,noSpc);
                if (lastKomma<0) error="Komma missing";
                para=toStrg(T,noSpc,lastKomma).trim();
                proteinList=new ByteArray(T,lastKomma+1, E).delBlanksL();
            } else if (only_pp_aa || only_pp) {
                proteinList=new ByteArray(T, noSpc, E).delBlanksL();
                para=null;
            } else { proteinList=null; para=new ByteArray(T, noSpc, E).delBlanksL().newString(); }
            final boolean paraNull=sze(para)==0 || "NULL".equals(para);
            /*
            //putln(ANSI_YELLOW+"cmd="+ANSI_RESET+cmd+"<\n"+ANSI_YELLOW+"para="+ANSI_RESET+para+"\n"+ANSI_YELLOW+"proteinList="+ANSI_RESET+proteinList);
            */
            final boolean timeConsuming=cmd==SCRIPT_infer_xyz||cmd==SCRIPT_align||cmd==SCRIPT_load
                ||cmd==SCRIPT_load_from_databases||cmd==SCRIPT_superimpose||cmd==SCRIPT_sleep;
            final String nJobID=this.nextJobID;
            final Runnable
                threadJobFinished=timeConsuming&&sze(nJobID)>0? threadM("remove", setJobs, new Object[]{nJobID}) : null,
                continueWhenFinished=threadJobFinished!=null?threadJobFinished :
                cmd==SCRIPT_wait_for||timeConsuming ? thread_interpretScript(new ByteArray(T,E+1,script.getEnd()))
                : null;

            this.nextJobID=null;
            if (timeConsuming) {
                if (nJobID!=null) setJobs.add(nJobID);
            }
            if (error!=null) {
            } else if (cmd==SCRIPT_wait_for){
                Object ids[]=splitTokns(0L, para);
                if (ids.length==0) ids=setJobs.toArray();
                newThrd(ChDelay.newThread(this,"WAIT_JOB",ids), continueWhenFinished).start();
                return;
            } else if (cmd==SCRIPT_job_id) {
                this.nextJobID=TOK.nextToken() ? TOK.asString() : null;
            }

            else if (cmd==SCRIPT_sleep) {
                newThrd(threadSleep(atoi(para)), continueWhenFinished).start();
                return;
            } else if (cmd==SCRIPT_use_3D_viewer) { StrapAlign.setDefaultC(ProteinViewer.class, StrapPlugins.mapS2L(para));
            } else if (cmd==SCRIPT_iconify ||cmd==SCRIPT_deiconify) { A.getFrame().setExtendedState(cmd==SCRIPT_deiconify ? JFrame.NORMAL : JFrame.ICONIFIED);
            } else if (cmd==SCRIPT_toFront) {
                setWndwStateT('F', -1, -1, A.getFrame().getTitle());
                A.getFrame().shw(0).setExtendedState(JFrame.NORMAL);
                final StrapView v=getAlignmentPanelWithFocus();
                if (v!=null) v.getAlignmentPane().requestFocus();
            } else if (cmd==SCRIPT_msec && is(DIGT, T, noSpc)) {
                final int ms=ChUtils.atoi(T,noSpc);
                if (ms>0) {
                    inEDT(thread_interpretScript(new ByteArray(T,nxtE(SPC,T,noSpc,E),E)));
                }
            } else if (T[B]=='*') A.dispatch(new StrapEvent(StrapAlign.class,StrapEvent.SCRIPT_LINE).setParameters(new String[]{script.newString(B,E)}));
            else if (cmd==SCRIPT_load) {
                final String[] ss=splitTokns(0L, T,noSpc,E, charClass(SPC));
                newThrd(
                        thread_downloadProteins(ss, oo(StrapView.KEY_AP),(List)null,(List)null, OPTION_PROCESS_LOADED_PROTS),
                        //new StrapEvent(A,StrapEvent.PROTEINS_SHOWN),
                        continueWhenFinished!=null ? continueWhenFinished:threadJobFinished).start();
                if (threadJobFinished==null) return;
            } else if (parameter_p_or_new) {
                final StrapProtein pp[]=getProteinsWithRegex(proteinList,ppVis);
                if (cmd==SCRIPT_aa_sequence) {
                    final String gapped=toStrgTrim(para);
                    if (sze(gapped)>0 && sze(proteinList)>0) {
                        final StrapProtein prots[];
                        if (sze(pp)>0) prots=pp;
                        else {
                            prots=new StrapProtein[]{new StrapProtein(A)};
                            prots[0].setName(toStrgTrim(rplcChars(charClass(NO_FILENM),'_', toStrgTrim(proteinList))));
                            A.setIsInAlignment(true,prots);
                            dispatchLater(StrapEvent.PROTEINS_ADDED,0);
                        }
                        for(StrapProtein p : prots) {
                            p.setResidueType( onlyLetters(gapped));
                            p.setTranslatedStrand(null);
                            if (gapped.indexOf('-')>=0) p.inferGapsFromGappedSequence(gapped.getBytes());
                            evt=StrapEvent.ALIGNMENT_CHANGED;
                        }
                    } else error=WNA;
                }
            } else if (parameter_pp_aa || only_pp_aa) {
                if (sze(proteinList)==0 || parameter_pp_aa && sze(para)==0) error=WNA;
                else if ("*".equals(para) && cmd==SCRIPT_unselect) {
                    getSelectedObjectsV().clear();
                    evt=StrapEvent.OBJECTS_SELECTED;
                } else {
                    final char alsoAnonymSel=cmd==SCRIPT_scroll_to || cmd==SCRIPT_box || cmd==SCRIPT_cursor ?'A':'a';
                    final Object[] pp_aa=getProteinsAndAnnotationsWithRegex(proteinList,ppVis, alsoAnonymSel);
                    for(Object o : pp_aa) {
                        final ResidueAnnotation a=deref(o, ResidueAnnotation.class);
                        final StrapProtein p=deref(o, StrapProtein.class);
                        final StrapProtein prot= p!=null ? p : a!=null ? a.getProtein() : null;
                        if (cmd==SCRIPT_close) {
                            if (p!=null) {
                                A.removeProteins(false,p);
                                evt=StrapEvent.PROTEINS_KILLED;
                            } else if (a!=null) {
                                a.dispose();
                                evt=StrapEvent.RESIDUE_SELECTION_CHANGED;
                            }
                        } if (cmd==SCRIPT_delete) {
                            if (a!=null) { a.dispose(); evt=StrapEvent.RESIDUE_SELECTION_CHANGED; }
                        }  else if (cmd==SCRIPT_icon) {
                            if (p!=null) {
                                p.setIconImage(paraNull?null:para);
                                evt=StrapEvent.PROTEIN_RENAMED;
                            } if (a!=null) {
                                a.setValue(ResidueAnnotation.BG_IMAGE,paraNull?null:para);
                                evtLater=StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR;
                            }
                        } else if (cmd==SCRIPT_select || cmd==SCRIPT_unselect) {
                            if (cmd==SCRIPT_unselect) getSelectedObjectsV().remove(o); else getSelectedObjectsV().add(o);
                            evt=StrapEvent.OBJECTS_SELECTED;
                        } else if (cmd==SCRIPT_scroll_to || cmd==SCRIPT_box) {
                            final int row=prot!=null ? idxOf(prot,ppVis) : -1;
                            if (row>=0) {
                                final int nR=prot.countResidues();
                                top=mini(top,row);
                                bottom=maxi(bottom,row+1);
                                int a0=0, a1=nR-1;
                                if (a!=null) {
                                    final int[] range=a.getRange();
                                    a0=range[0];
                                    a1=range[1];
                                }
                                if (a0>=0 && a1>=0) {
                                    left=mini(left,prot.getResidueColumn(mini(nR-1,a0)));
                                    right=maxi(right,prot.getResidueColumn(mini(nR-1,a1)));
                                }
                            }
                        } else if (cmd==SCRIPT_cursor && a!=null) {
                            if (view!=null) {
                                view.setCursorToResidue(prot,firstTrue(a.getSelectedAminoacids())+a.getSelectedAminoacidsOffset());
                                view.scrollCursorToVisible();
                            }
                        }
                    }
                    if (top<(1<<30) && left<(1<<30) && left<right && top<bottom) {
                        if (cmd==SCRIPT_scroll_to){
                            final int
                                w=view.charB().width,
                                h=view.charB().height+view.getSpaceBetweenRows(),
                                x1=left*w,  y1=top*h,
                                x2=right*w, y2=bottom*h;
                            final JComponent pane=view.getAlignmentPane();
                            pane.invalidate();
                            final Rectangle r=new Rectangle(x1,y1,x2-x1-EX, y2-y1+2*EX);
                            view.scrRectToVis(r,true);
                            try{
                                final Graphics g=pane.getGraphics();
                                g.setColor(C(0xffff,60));
                                g.fillRect(x1,y1, x2-x1, y2-y1);
                            } catch(Exception e){}
                        } else if (cmd==SCRIPT_box) view.setRectRubberBand(left,top, right-left, bottom-top);
                    }
                }
            } else if (cmd==SCRIPT_new_aminoacid_selection || cmd==SCRIPT_new_nucleotide_selection) {
                final String
                    aName_slash_protein=TOK.nextAsStringTrim(),
                    protein=delFromChar(aName_slash_protein,'/'),
                    aName=lastPathCompnt(aName_slash_protein),
                    grp=TOK.nextAsStringTrim(),
                    resi=TOK.nextToken() ? script.newString(TOK.getFrom(),E) : null;
                if (resi!=null && grp!=null && aName_slash_protein!=aName && aName.length()>0) {
                    for(Object o : getProteinsAndAnnotationsWithRegex(protein,ppVis, 'p')) {
                        final StrapProtein p=deref(o,StrapProtein.class);
                        if (p==null) continue;
                        final ResidueAnnotation a=new ResidueAnnotation(p);
                        p.addResidueSelection(a);
                        a.addEnS(ResidueAnnotation.NAME, aName);
                        a.addE(ResidueAnnotation.ENTRY_NOT_SAVED, ResidueAnnotation.POS, resi, cmd==SCRIPT_new_aminoacid_selection, MAX_INT);
                        a.addEnS(ResidueAnnotation.GROUP, grp);
                        for(ResidueAnnotation b : p.getResidueAnnotations()) {
                            if (a!=b && b.getFeatureName()==null && aName.equals(b.getName()) &&
                                a.getSelectedAminoacidsOffset()==b.getSelectedAminoacidsOffset() &&
                                Arrays.equals(a.getSelectedAminoacids(), b.getSelectedAminoacids())) p.removeResidueSelection(b);
                        }
                        evt=StrapEvent.RESIDUE_SELECTION_ADDED;
                    }
                } else error=WNA;
            } else if (parameter_pp || only_pp) {
                final StrapProtein[] pp=getProteinsWithRegex(proteinList,ppVis);
                if (sze(proteinList)==0 || parameter_pp &&  sze(para)==0)  error="";
                else if (pp.length==0) ensp=proteinList;
                else {
                    if (pp.length==0) ensp=proteinList;
                    for(StrapProtein p : pp) {
                        if (cmd==SCRIPT_gaps) {
                            p.inferGapsFromGappedSequence(para.getBytes());
                            evtLater=StrapEvent.ALIGNMENT_CHANGED;
                        } else if (cmd==SCRIPT_cds) {
                            final String s=cntainsOnly(DIGT,para,0,99) ? get(atoi(para)-1,p.getCDS()) : para;
                            p.applyCDS_expression(s);
                            evt=StrapEvent.NUCL_TRANSLATION_CHANGED;
                        } else if ( cmd==SCRIPT_to_row || cmd==SCRIPT_below_row) {
                            int dst=cntainsOnly(DIGT,para) ? atoi(para)-1 : idxOf(StrapProtein.getProteinWithName(para,ppVis), ppVis);
                            if (dst>=0) {
                                if (cmd==SCRIPT_below_row)dst++;
                                final int src=idxOf(p,ppVis);
                                if (src>=0 && dst>=0 && src!=dst) { view.moveLines(src, mini(view.countRows()-1, dst), 1);  }
                                evt=StrapEvent.ORDER_OF_PROTEINS_CHANGED;
                            } else error="Wrong destination row";
                        } else if (cmd==SCRIPT_rotate_translate) {
                            final Matrix3D m3d=paraNull ? null : new Matrix3D();
                            if (paraNull || m3d.parsePlain(para.getBytes(),0, MAX_INT)) { p.setRotationAndTranslation(m3d); evt=StrapEvent.PROTEIN_3D_MOVED; evtObj=pp;}
                            else error="Error parsing matrix ";
                        } else if (cmd==SCRIPT_spice) SequenceFeatures.toSpiceViewer(p);
                        else if (cmd==SCRIPT_close_wire) {
                            V3dUtils.disposeV( derefArray(p.getProteinViewers(),Protein3d.PView.class));
                        } else if (cmd==SCRIPT_add_xref || cmd==SCRIPT_accession_id || cmd==SCRIPT_remove_xref ) {
                            final String xref=paraNull?null:para;
                            if (cmd==SCRIPT_accession_id) p.setAccessionID(xref);
                            else if (cmd==SCRIPT_add_xref) p.addSequenceRef(xref);
                            else if (cmd==SCRIPT_remove_xref) p.removeSequenceRef(xref);
                            evtLater=StrapEvent.PROTEIN_XREF_CHANGED;
                        } else if (cmd==SCRIPT_seqvista) {
                            Hyperrefs.toSeqvista(new File[]{p.getFile()});
                        }
                    }/*for p*/
                    if (cmd==SCRIPT_infer_xyz && pp.length>0) {
                        final String ids[]="AUTO".equals(para) ? StrapProteinUtils.INFER_COORDINATES_AUTO : toStrgArryO(para);
                        final Thread t1=StrapProteinUtils.threadInferCoordinates(pp, ids, ChLog.NO_LOG, StrapProteinUtils.INFER_3D_LOG);
                        newThrd(t1,continueWhenFinished).start();
                        if (threadJobFinished==null) return;
                    } else if (cmd==SCRIPT_das) { StrapDAS.thread_load(0L, splitTokns(0L, para), pp, (ChRunnable)A).start();
                    } else if (cmd==SCRIPT_wire) { A.webAlignment_3D(pp, null,true);
                    } else if (cmd==SCRIPT_hide)  { if (A.setIsInAlignment(false,pp)) evt=StrapEvent.PROTEINS_HIDDEN;
                    } else if (cmd==SCRIPT_unhide){ if (A.setIsInAlignment(true,pp))  evt=StrapEvent.PROTEINS_SHOWN;
                    } else if (cmd==SCRIPT_align) {
                        newThrd(StrapProteinUtils.threadAlignProteins((SequenceAligner)null,Arrays.asList(pp),CAN_BE_STOPPED, 0L), continueWhenFinished).start();
                        if (threadJobFinished==null) return;
                    } else if (cmd==SCRIPT_load_from_databases) {
                        newThrd(StrapProteinUtils.threadDownloadOriginalProteins(pp,0L), continueWhenFinished).start();
                        if (threadJobFinished==null) return;
                    } else if (cmd==SCRIPT_jalview) { new Export2Jalview().launchJalview(pp,Export2Jalview.SECSTRU);
                    } else if (cmd==SCRIPT_superimpose) {
                        newThrd(StrapProteinUtils.threadSuperimposeProteins(Arrays.asList(pp), CAN_BE_STOPPED, StrapProteinUtils.SUPERIMP_EVENT), continueWhenFinished).start();
                        if (threadJobFinished==null) return;
                    } else if (cmd==SCRIPT_tree) {
                        final String nn[]=new String[pp.length];
                        final byte[][] gg=new byte[pp.length][];
                        for(int i=pp.length; --i>=0;) {
                            final String n=pp[i].getName();
                            if (0<=idxOfStrg(n,nn)) continue;
                            nn[i]=n;
                            gg[i]=pp[i].getGappedSequenceExactLength();
                        }
                        final PhylogeneticTree tree=
                            HotswapUtil.instance(CLASS_PhylogeneticTree_TreeTop/*getDefaultPhylogeneticTree()*/, PhylogeneticTree.class);
                        if (tree!=null) {
                            tree.setNamesIconsAndGappedSequences(rmNullS(nn), null, rmNullA(gg,byte[].class) );
                            tree.drawTree();
                        }
                        setWndwStateLaterT('F',444, charite.christo.strap.extensions.PhylogeneticTree_ATVPROXY.FRAME_TITLE);
                    }
                }
            } else if (cmd==SCRIPT_close_3D || cmd==SCRIPT_open_3D || cmd==SCRIPT_select_3D || cmd==SCRIPT_to_front_3D) {
                final int komma;
                {
                    int i, countQ=0;
                    for( i=noSpc; i<E; i++) {
                        if (T[i]=='"') countQ++;
                        if (countQ%2==0 && T[i]==',') break;
                    }
                    komma=i;
                }
                final String script_vID=toStrg(T,noSpc,komma).trim();
                final String pp_aa_ss=script.newString(komma+1, E);
                final Class c=name2class(getDefault(ProteinViewer.class));
                final int lastSpc=prev(SPC, script_vID, MAX_INT,-1);
                final String viewId=toStrgIntrn(script_vID.substring(lastSpc+1));

                if (sze(viewId)==0) error="3D-Viewer-ID not defined";
                else if (c==null) error="No such 3D-viewer Java class  "+getDefault(ProteinViewer.class);
                else {
                    final Object pp_aa[]=getProteinsAndAnnotationsWithRegex(pp_aa_ss,ppVis,'A');
                    final ResidueAnnotation aa[]=derefArray(pp_aa,ResidueAnnotation.class);
                    final ProteinViewer
                      pv=deref(V3dUtils.mapProteinViewers().get(viewId),ProteinViewer.class),
                        vv[]=Protein3dUtils.getViewersSharingView(pv);
                    final Protein[] pp;
                    if (sze(pp_aa_ss)==0) pp=Protein3dUtils.getProteinsInView(vv);
                    else {
                        final List vP=new ArrayList();
                        for (Object o : pp_aa) adUniq( o instanceof Protein ? o : o instanceof ResidueSelection ? ((ResidueSelection)o).getProtein() : null, vP);
                        pp=toArry(vP, Protein.class);
                    }

                    if (cmd==SCRIPT_close_3D) {
                        ProteinViewer vRemaining=null;
                        for(ProteinViewer v : Protein3dUtils.getViewersSharingView(pv)) {
                            if (sze(pp_aa_ss)==0 || cntains(v.getProtein(),pp)) V3dUtils.disposeV(v);
                            else vRemaining=v;
                        }
                        V3dUtils.mapProteinViewers().put(viewId,weakRef(vRemaining));
                        //ChDelay.repaint(Protein3dUtils.viewerCanvas(vRemaining),10);
                    } else if (cmd==SCRIPT_open_3D) {
                        final Protein[] ppLoaded=Protein3dUtils.getProteinsInView(pv);
                        for(int i=pp.length; --i>=0;) if (cntains(pp[i],ppLoaded)) pp[i]=null;
                        if (countNotNull(pp)>0) {
                            final ProteinViewer pvNew=V3dUtils.open3dViewer(0, pp, pv!=null?pv:c);
                            if (pvNew==null) error="Could not instantiate "+c;
                            V3dUtils.mapProteinViewers().put(viewId,weakRef(pvNew));
                        }
                    }
                    if(cmd==SCRIPT_select_3D || cmd==SCRIPT_open_3D) {
                        s_view3D_pp=this.view3D_pp=pp;
                        s_view3D_ID=this.view3D_ID=viewId;
                        if (sze(aa)>0) { //!!!
                          for (ProteinViewer v : vv) V3dUtils.selectedAAto3D_single(0L, aa, v, "unnamed");

                        }
                    }
                    if (cmd!=SCRIPT_close_3D) ChDelay.runAfterMS(EDT,V3dUtils.runViewerToFront(pv),99);
                }
            } else if(is3DCmd || cmd==SCRIPT_to_3D) {

                Protein pp[]=this.view3D_pp;
                if (pp==null) pp=s_view3D_pp;
                String viewId=this.view3D_ID;
                if (viewId==null) viewId=s_view3D_ID;
                final ProteinViewer
                  pv=deref(V3dUtils.mapProteinViewers().get(viewId),ProteinViewer.class),
                    vv[]=Protein3dUtils.getViewersSharingView(pv);

                final String txt;
                if (cmd==SCRIPT_to_3D) {
                    final String pvNames=TOK.nextAsStringTrim();
                    txt=strstr(STRSTR_IC|STRSTR_w,niceShrtClassNam(pv),pvNames)<0?null: TOK.remainingAsString(true);
                } else txt=script.newString(B,E);
                if (sze(txt)>0) {
                    for(int iP=0; iP<sze(pp); iP++) {
                        if (pp[iP]==null) continue;
                        boolean sent=false;

                        for(ProteinViewer v : pp[iP].getProteinViewers()) {
                            if (ANY_WIRE==viewId && v instanceof Protein3d.PView ||
                                (ANY_VIEWER==viewId || "*"==viewId) && !(v instanceof Protein3d.PView) ||
                                viewId!=null && viewId.equalsIgnoreCase(niceShrtClassNam(v)) ||
                                cntains(v,vv)
                                ) {

                                V3dUtils.sendCommand(ProteinViewer.INTERPRET_NO_MSG_DIALOGS, txt, v);
                                sent=true;
                            }
                        }
                        if (!sent) error="No 3D-viewer for id \""+viewId+"\"";
                    }
                }
            } else if (cmd==SCRIPT_add_annotation || cmd==SCRIPT_set_annotation || cmd==SCRIPT_to_structure_viewer) {
                final boolean edit=cmd==SCRIPT_add_annotation || cmd==SCRIPT_set_annotation;
                if (TOK.nextToken()) {
                    final ChTokenizer tok=new ChTokenizer().setText(T,TOK.getFrom(),TOK.getTo());
                    final String
                        key=edit ? StrapPlugins.mapS2L(TOK.nextAsStringTrim()) : null,
                        value=TOK.nextToken() ? script.newString(TOK.trim().getFrom(),E) : null;
                    while(tok.nextToken()) {
                        for(Object o : getProteinsAndAnnotationsWithRegex(tok.asString(),ppVis,'a')) {
                            if (!(o instanceof ResidueAnnotation)) {ensa=tok.asString(); continue; }
                            final ResidueAnnotation a=(ResidueAnnotation)o;
                            if (edit) {
                                if (key==null || value==null) error=WNA;
                                if (ResidueAnnotation.COLOR.equals(key)) a.setColor(str2color(value));
                                else {
                                    if (cmd==SCRIPT_add_annotation) a.addEnS(key,value);
                                    else a.setValue(key,"NULL".equals(value)?null:value);
                                }
                                evt=StrapEvent.RESIDUE_SELECTION_CHANGED;
                            }
                        }
                    }
                } else error=WNA;
            } else if (cmd==SCRIPT_plugin) {
                A.getFrame().setExtendedState(JFrame.NORMAL);
                A.getFrame().toFront();
                final String className=TOK.nextAsStringTrim(), sUrl=TOK.nextAsString(), delete[]=splitTokns(0L, TOK.nextAsString());
                final URL u=url(sUrl);
                if (u==null) error="No URL: "+sUrl;
                if (sze(className)>0 && u!=null) {
                    for(int i=sze(delete); --i>=0;) delFile(StrapPlugin.jarFile(delete[i]));
                    final File f=StrapPlugin.jarFile(u);
                    InteractiveDownload.downloadFiles(new URL[]{u}, new File[]{f});
                    ChDelay.newThread(A,"CHECK_PLUGIN_UPTODATE",new URL[]{u}).start();
                    java.util.zip.ZipFile zf=null;
                    try {zf=new java.util.zip.ZipFile(f);} catch(Exception ex){}
                    if (zf==null) error="Cann not open Jar-file "+f;
                    else {
                        final StrapPlugin plugin=new StrapPlugin(className, "", zf);
                        final Class c=plugin.getClazz(true);
                        if (c!=null) StrapPlugins.applyPlugin(c);
                        else error="Unable to load class :"+className;
                    }
                }
            } else if (cmd==SCRIPT_src) {
                if (TOK.nextToken())  {
                    final String cn=TOK.trim().asString();
                    try { Class.forName(cn); } catch(Exception ex) { error=ex.toString();}
                    showJavaSource(cn,NO_STRING);
                } else error="";
            } else error="No such command \""+cmd+"\"";
            if (evt>=0) A.dispatch(new StrapEvent(StrapAlign.class,evt).setParameters(evtObj));
            if (evtLater>=0) dispatchLater(evtLater,111);
            boolean nl=false;
            if (sze(ensp)>0 && !(sze(ensp)==1 && '*'==chrAt(0,ensp))) {
                nl=true; log("Error: no such protein \""+ensp+"\"\n"); }
            if (sze(ensa)>0 && !(sze(ensa)==1 && '*'==chrAt(0,ensa))) { nl=true; log("Error: no such residue selection \""+ensa+"\"\n"); }
            if (error!=null) {
                nl=true; log(script.newString(B,E)+" "+error+"\n");
                HelpCommands hlp=HelpCommands.getInstance(_CCS.replace('.','/')+"/scriptHelp.rsc");
                log(hlp.get(HelpCommands.ALL, cmd));
            }

            if (nl) ChFrame.frame("Script Log",log("\n"),ChFrame.SCROLLPANE|CLOSE_ON_CtrlW).shw(0);
        }
        {
            if (selectTablePanel!=null) getDialogPanel().setSelectedComponent(A.getTablePanel(selectTablePanel));
            setWndwState('F',A.getFrame());
        }
    }
    /* <<< interpretScript <<< */
    /* ---------------------------------------- */
    /* >>> Log >>> */
    private static ChTextArea _log;
    public static ChTextArea log(String s) {
        if (_log==null) _log=new ChTextArea("");
        if (sze(s)>0) _log.appndS(s);
        return _log;
    }
    /* <<< Log <<< */
    /* ---------------------------------------- */
    /* >>> Thread >>> */
    public Object run(String id,Object arg) {
        final Object argv[]=arg instanceof Object[] ? (Object[]) arg : null;
        if (id=="WAIT_JOB") {
            while(sze(setJobs)>0) {
                sleep(111);
                boolean running=false;
                for(Object s : argv) if (setJobs.contains(s)) { running=true; break;}
                if (!running) break;
            }
        }
        return null;
    }

    /* <<< Thread <<< */
    /* ---------------------------------------- */
    /* >>> static Utils >>> */
    private static String[] _allCmds;
    public static String[] allScriptCommands() {
        if (_allCmds==null) _allCmds=getFinalStaticStrings("SCRIPT_", StrapScriptInterpreter.class);
        return _allCmds;
    }

}