package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import java.awt.*;
import java.awt.event.*;
import java.io.File;
import java.util.*;
import java.util.List;
import javax.swing.*;
import static charite.christo.ChConstants.*;
import static charite.christo.protein.ProteinModiCount.*;
import static charite.christo.ChUtils.*;
import static java.awt.event.MouseEvent.*;

/**HELP

<b>Context menus in STRAP:</b>

The term WIKI:Context_menu is commonly used for menus which pop up when right
clicking (<b>right mouse button</b>) an item in a graphical user interface,
offering a list of options which vary depending on the item selected.
Whatch movie <i>STRING:ChConstants#MOVIE_Context_Menu</i>
<br><br>
On Macintosh computers the right  mouse button is often deactivated.
It can be activated in the Settings panel of the Mac.
Alternatively, the right mouse button can be simulated by  ALT+CTRL+left-click.
<br><br>
Context menus are available for many different object types.
For a complete list click <i>BUTTON:StrapAlign#button(CMD_ALL_POPUP)!</i> in the help-menu of the menu-bar.

<br><b>Selecting single items:</b>
List items are selected simply by left-click the respective node.
Single residue selections in the alignment panel can be selected by clicking with the CTRL key.

<br><br><b>Selecting more than one item:</b>
Selecting more than one list item requires the <b>SHIFT</b> and <b>CTRL</b> keys.
The CTRL key is located at the lower left of the key board and is sometimes labeled  <b>STRG</b> and
the SHIFT key is sometimes termed "UMSCHALT".

By dragging a rectangular region in the alignment, all contained
residue selections are selected. With the SHIFT or CTRL key the
union-set or cut-set is formed with the set of selections inside the
rectangular region.

The tool-button  of the tree-panel allows to select items according to <b>text matches</b>.

<br><br><b>Frequently used menu items</b> can be dragged out the menu and placed on the desktop.
Instead of clicking a menu-item the user drags the menu-item to the desktop.
Being on the desktop, it can now be accessed much easier.
When the menu-item on the desktop is not needed any more, it can be
deleted.

<br><br><b>The tree view</b> is located at the left of the application.
It is usually  hidden and can be opened by dragging the vertical divider bar.
The tree contains all loaded proteins and their child objects.

@author Christoph Gille
*/
public class StrapTree extends ChJTree implements ChRunnable, ChProcessEvent  {
    /* >>> Instance >>> */
    private boolean isInitialized;
    private final StrapAlign align;
    public StrapTree(StrapAlign a) {
        super(ChJTable.DRAG_ENABLED|ChJTable.DEFAULT_RENDERER,null);
        align=a;
        pcp(KEY_SCROLL_INCR_V,intObjct(32),this);
        this.setRootVisible(false);
        this.setShowsRootHandles(true);
        //TabItemTipIcon.set("",">","","",this);
        addActLi(LI,this);
    }
    /* <<< Instance <<< */
    /* ---------------------------------------- */
    /* >>> Toolbar  >>> */
    private static AbstractButton[] radioGrp;
    final Set vMatching=new HashSet();
    private static int idxRadioSearch;
    public static int idxRadioSearch() { return idxRadioSearch;}
     JComponent toolbar() {
        final DialogStringMatch d=new DialogStringMatch(0L, this, vMatching, "StrapTree");
        radioGrp=radioGrp(new String[]{"Protein names","Protein header", "Annotations"},0, null);
        pcp(KEY_NORTH_PANEL,pnl("Search in ",radioGrp),d);
        addActLi(LI,d);
        return hBL( smallHelpBut(StrapTree.class),
                   d.button("Search").i(IC_SEARCH),
                   new FreeMemoryBar().start()
                   );
    }
    public void setMatchingItems(Object[] oo) {
        adAllUniq(oo, StrapAlign.getSelectedObjectsV());
        StrapAlign.dispatchLater(StrapEvent.OBJECTS_SELECTED,99);
    }

    /* <<< Instance <<< */
    /* ---------------------------------------- */
    /* >>> Annotation Group >>> */
    private void renameGroupOfAnnotations(UniqueList<ResidueAnnotation> l) {
        final String oldName=l.getText();
        final String str=ChMsg.input(ChMsg.AT_CLICK,"Please enter the new name for this group of annotations",30,oldName);
        if (str!=null) {
            final String newName=str.trim();
            if (!"".equals(newName)) {
                for(ResidueAnnotation s: l.asArray()) {
                    if (oldName.equals(s.getValue(ResidueAnnotation.GROUP)))
                        s.setValue(ResidueAnnotation.GROUP,newName);
                }
                l.t(newName);
            }
        }
        StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
    }
    /* <<< Annotation Group  <<< */
    /* ---------------------------------------- */
    /* >>> Repaint  >>> */
    public void paintComponent(Graphics g1) {
        if (!isInitialized) {
            isInitialized=true;
            ChDelay.invokeAndWait(this,"I",null);
        }
        try { super.paintComponent(g1); } catch(Throwable e){ stckTrc(e);}
    }
    private static int update_modiCount,repaint_modiCount ;

    void maybeUpdate() {
        final int modi[]=Protein.MODI_COUNTS_ALL;
        if(!isInitialized) return;
        {
            final int mc=modi[MC_RESIDUE_ANNOTATION_GROUPS] + modi[MCA_PROTEINS_V] + modi[MC_RESIDUE_SELECTIONS] +
                modi[MC_PROTEIN_USER_OBJECTS_V] + modi[MC_PROTEIN_VIEWERS_V];
            if (update_modiCount!=mc) {
                update_modiCount=mc;
                updateKeepStateLater(0L, align.getProteins().length/10);
            }
        }
        {
            final int mc= modi[MC_PROTEIN_LABELS]+ modi[MC_RESIDUE_SELECTION_LABELS];
            if (repaint_modiCount!=mc) {
                repaint_modiCount=mc;
                repaint();
            }
        }
    }
    /* <<< Repaint  <<< */
    /* ---------------------------------------- */
    /* >>> AWTEvent  >>> */
    final ChEventAdapter LI=new ChEventAdapter(this);
    public AWTEvent processEvent(final JComponent component, final AWTEvent ev) {
        final Object q=ev.getSource();
        final String cmd=actionCommand(ev);
        final UniqueList vSel=StrapAlign.getSelectedObjectsV();
        final int vSel_mc=vSel.getModificationCount();
        final int id=ev.getID();
        final int kcode=keyCode(ev);
        if (cmd==DialogStringMatch.ACTION_BEFORE) {
            StrapTree.idxRadioSearch=radioGrpIdx(radioGrp);
        }
        if (cmd==DialogStringMatch.ACTION_FINISHED) {
            for(Object o : vMatching.toArray()) vSel.add(o);
            StrapTree.idxRadioSearch=-1;
        }
        if (isPopupTrigger(ev) && id==MOUSE_PRESSED) {
            final Object oMouse=objectAt(ev);
            if (oMouse instanceof UniqueList && ((UniqueList)oMouse).getClazz()==ResidueAnnotation.class) renameGroupOfAnnotations((UniqueList)oMouse);
            else {
                final char menuType=ContextObjects.menuID(oMouse);
                if (menuType!=0) ContextObjects.showContextMenu(menuType, oo(oMouse), getSelValues(q));
            }
        }
        if (id==MOUSE_ENTERED || id==MOUSE_EXITED) maybeUpdate();

        if (id==MouseEvent.MOUSE_CLICKED && clckCount(ev)>1) {
            final Object o=objectAt(ev);
            if (o instanceof ResidueAnnotation) StrapAlign.editAnnotation((ResidueAnnotation)o);
            if (o instanceof UniqueList && ((UniqueList)o).getClazz()==ResidueAnnotation.class) renameGroupOfAnnotations((UniqueList<ResidueAnnotation>)o);
            if (component!=null) component.requestFocus();
        }
        if (component!=null && (id==KeyEvent.KEY_PRESSED && (kcode==KeyEvent.VK_BACK_SPACE || kcode==KeyEvent.VK_DELETE))) {
            boolean protsKilled=false, selResidues=false;
            final Class cc[]=new Class[]{StrapProtein.class, UniqueList.class, ResidueAnnotation.class, ResidueSelection.class, HeteroCompound.class};
            final ChJList jList=q instanceof JList ? (ChJList)q : null;
            final Object[] ooJL=getSelValues(jList);

            if (gcp(StrapAlign.OBJECT_JLIST,jList)!=null && 0==(jList.getOptions()&ChJTable.CHJTABLE_BACKSPACE_DEACTIVATED)) {
                if (ChMsg.yesNo("Remove the selected items from this list?")) {
                    for(int i=sze(ooJL); --i>=0;) jList.getList().remove(ooJL[i]);
                    jList.repaint();
                }
            } else {
                nxtClass:
                for(Class c : cc) {
                    for(boolean del : FALSE_TRUE) {
                        int count=0;
                        final Object[] ooSel=jList!=null ? ooJL : vSel.asArray();
                        if (sze(ooSel)==0) break;
                        for(Object o : ooSel) {
                            for(Class c2 : cc) if (!isAssignblFrm(c,o) || c!=c2 && isAssignblFrm(c,c2) && isAssignblFrm(c2,o)) continue nxtClass;
                            final StrapProtein prot=o instanceof StrapProtein ? (StrapProtein)o : null;
                            final ResidueSelection selection= o instanceof ResidueSelection ? (ResidueSelection)o : null;
                            final HeteroCompound hetero= o instanceof HeteroCompound ? (HeteroCompound)o : null;
                            final UniqueList<ResidueSelection> g=o instanceof UniqueList ? (UniqueList)o : null;
                            if (g!=null && !isAssignblFrm(ResidueSelection.class, g.getClazz())) continue;
                            count++;
                            if (del) {
                                if (hetero!=null) {
                                    for(StrapProtein p : align.getProteins()) p.removeHeteroCompound(hetero);
                                    protsKilled=true;
                                }
                                if (prot!=null) {
                                    align.removeProteins(false,prot);
                                    for(ResidueSelection  s:prot.getAllResidueSelections()) vSel.remove(s);
                                    //for(List group: prot.getResidueAnnotationsByGroup()) vSel.removeAll(group);
                                    protsKilled=true;
                                }
                                final ResidueSelection[] ss= selection!=null ? new ResidueSelection[]{selection} : g!=null ? g.asArray() : null;
                                if (sze(ss)>0) {
                                    for(ResidueSelection s : ss) {
                                        final Protein p=s.getProtein();
                                        if (p!=null) p.removeResidueSelection(s);
                                        if (s instanceof Disposable) ((Disposable)s).dispose();
                                        vSel.remove(s);
                                    }
                                }
                                vSel.remove(o);
                                selResidues=true;
                            }
                        }
                        final String n= c==HeteroCompound.class ? "hetero compounds/nucleotides" : c==StrapProtein.class ? "protein" :  c==ResidueAnnotation.class ? "residue annotation" :  c==ResidueSelection.class ? "residue selection" : c==UniqueList.class?"residue selection group" :  "";
                        if (count>0 && !del && !ChMsg.yesNo("Delete "+count+" "+n+(count>1?"s":""))) break;
                    }
                }
                ChDelay.repaint(component,333);
            }
            if (jList!=null) jList.setSelOO(ooJL);
            if (protsKilled) StrapAlign.dispatchLater(StrapEvent.PROTEINS_KILLED,1);
            if (selResidues) StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,1);
        }

        if (cmd==ACTION_SELECTION_CHANGED) {
            if (gcp(KEY_IGNORE_SELECTION_EVT,this)==null && StrapView.alreadyPaintedStatic()) {
                vSel.clear();
                vSel.rmOptions(UniqueList.UNIQUE);
                adAll(getSelValues(this),vSel);
                vSel.addOptions(UniqueList.UNIQUE);
            }
        }
        if (isInitialized && vSel_mc!=vSel.getModificationCount())  StrapAlign.dispatchLater(StrapEvent.OBJECTS_SELECTED,1);
        return ev;
    }

    /* <<< Repaint <<< */
    /* ---------------------------------------- */
    /* >>> Thread >>> */
    public Object run(String id, Object arg) {
        if (id==PROVIDE_JTREE_CHILDS) {
            final StrapView pan=align.getAlignmentPanelWithFocus();
            return pan!=null ? pan.getVisibleProteins() : align.vProteins;
        }
        if (id==DialogStringMatch.PROVIDE_HAYSTACK || id==PROVIDE_WORD_COMPLETION_LIST) {
            return align.getProteins();
        }
        if (id=="I") {
            setRowHeight(ICON_HEIGHT);
            setModel(new ChTreeModel(0L,this));
            rtt(this);
            addMoliRefreshTip(this);
            addMouseListener(LI);
            addActLi(LI,this);
            setDefaultPopupListener(this);
        }
        return null;
    }
    /* <<< Thread <<< */
    /* ---------------------------------------- */
    /* >>> StrapEvent >>> */
    private void repaintNode(Object o) {
        final javax.swing.tree.TreePath tp=o==null ? null : getTreePath(o,this);
        if (tp!=null) repaint(getPathBounds(tp));
    }
    private Protein proteinAtCursor;
    /* <<< StrapEvent <<< */
    /* ---------------------------------------- */
    /* >>> Selection >>> */
    /*
    @Override public void clearSelection() {
        setSelectionPaths(new TreePath[0]);
        valueChanged((TreeSelectionEvent)null);
        StrapAlign.getSelectedObjectsV().clear();
    }
    */
    /* <<< Selection <<< */
    /* ---------------------------------------- */

}
