package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.strap.interfaces.*;
import charite.christo.protein.*;
import java.awt.*;
import java.io.File;
import java.util.*;
import javax.swing.*;
import javax.swing.border.TitledBorder;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
import static charite.christo.strap.StrapUtils.*;
/*
   @author Christoph Gille
*/
public final class Texshade  extends AbstractDialogJTabbedPane implements Disposable,NeedsProteins,ChRunnable, StrapListener,java.awt.event.ActionListener  {
    private final static String
        TEXSHADE="texshade120", VMARGIN="vmargin",
        helpBeginEnd="The latex source is <br>built from <ul><li>a constant beginning part</li><li>a variable middle part</li><li>a constant end part</li><ul>",
        NON_TOP_BOTTOM[]={"non","ttop","top","bottom","bbottom"};
    public final static String
        SAVE="texshade/",
        DIR=STRAPTMP+"/texshade/",
        TEX_FILE=DIR+"alignm.tex", MSF_FILE=DIR+"alignm.msf", PDF_FILE=DIR+"alignm.pdf",
        PDF_COPY_FILE=DIR+"align.pdf",DVI_FILE=DIR+"alignm.dvi",PS_FILE=DIR+"alignm.dvi.ps",PS_POSTER_FILE=DIR+"poster.ps",LOG_FILE=DIR+"alignm.log",AUX_FILE=DIR+"alignm.aux",
        CB_LAB_Postscript="Generate PostScript instead of PDF",
        BUT_LAB_Source="Make LaTeX and Multiple Sequence file",
        BUT_LAB_RunTex="Make PDF",
        BUT_LAB_EditSource="edit tex",
        BUT_LAB_Pdf="PDF";
    private static AbstractButton tt( AbstractButton b, String tt) { b.setToolTipText(addHtmlTagsAsStrg(tt)); return b;}
    private AbstractButton cb(String t) { return cb(false, t, null);}
    private AbstractButton cb(String t, String saveInFile) { return cb(false, t, saveInFile);}
    private AbstractButton cbOn(String t, String saveInFile) { return cb(true, t, saveInFile);}
    private AbstractButton cb(boolean set, String t, String saveInFile) {
        final ChButton tog=toggl(t).s(set);
        if (saveInFile!=null) tog.save(Texshade.class,saveInFile);
        final AbstractButton cb=tog.cb();
        cb.addActionListener(this);
        return cb;
    }
    private final AbstractButton
        butRunTex=new ChButton(BUT_LAB_RunTex).li(this).tt(
            "Processes the .tex file and creates the .pdf file.<br>"+
            "This button is active if "+TEX_FILE+" exists and no <br>other tex-process is running.<br>"+
            "LaTex is run twice unless the shift-key is hold down."),
        butDvi=new ChButton("dvi").actionExecuteProgram(Customize.customize(Customize.dviViewer),file(DVI_FILE)).tt("View DVI file"),
        butDviPs=new ChButton("dvips").actionExecuteProgram(Customize.customize(Customize.dviPsConverter),file(DVI_FILE)).tt("Convert dvi to postscript"),
        butPS=new ChButton("PS").actionViewFile(file(PS_FILE)),
        butPosterViewPS=new ChButton("view PS-file").actionExecuteProgram(Customize.customize(Customize.psViewer),file(PS_POSTER_FILE)),
        butPdf=new ChButton(BUT_LAB_Pdf).tt("View PDF-file"),
        butPoster=new ChButton("Run Poster on<br>"+PS_FILE).li(this),
        cbNumAsPdb=cb("PDB residue numbers. (Currently, the initial number is set only)","cbNumAsPdb"),
        cbCustomSize=cb("Custom size","cbCustomSize"),
        cbStdout=cb("Show output","cbStdout"),
        cbLandscape=cbOn("Landscape","cbLandscape"),
        cbNoExtension=tt(cbOn("Suppress file suffix in protein file names","noExtension"),"protein1.swiss would be written as protein1"),
        cbPostscript=tt(cb(CB_LAB_Postscript),"PDF-LaTeX produces PDF directly. <br>LaTeX generates a DVI <br>which can be converted to PostScript"),
        cbSubstituteNames=tt(cb("Apply substitutions of protein names","displayedProteinName"),"You may substitute the <br>names of the proteins <br>by any latex expression"),
        cbSeparationLine=cb("Additional space after alignment","cbSeparationLine"),
        cbUpperCase=cb("Use upper case only","cbUpperCase"),
        cbHideLegend=cb("Hide legend","cbHideLegend"),
        cbFingerprint=cb("Fingerprint","cbFingerprint"),
        cbHideConsensus=cb("Hide consensus","cbHideConsensus"),
        cbDofrequencycorrection=cb("\\dofrequencycorrection","cbDofrequencycorrection"),
        cbHideSeqs=cb("\\hideseqs","cbHideSeqs");
    private final JComponent panPoster=hB();
    private JComponent panViewers;
    private final ChCombo
        choiceConsensus=new ChCombo("Gray","BlueRed","RedBlue","GreenRed","RedGreen","ColdHot").save(Texshade.class,"choiceConsensus"),
        choiceConsensusLocation=new ChCombo(NON_TOP_BOTTOM).save(Texshade.class,"choiceConsensusLocation"),
        choicePosterSize=new ChCombo("same","A0","A1","A2","A4").save(Texshade.class,"choicePosterSize"),
        choiceShading=new ChCombo("none","functional","identical","similar","diverse").s(1).save(Texshade.class,"choiceShading"),
        choiceShowNames=new ChCombo("left","right","none").save(Texshade.class,"choiceShowNames"),
        choiceEmph=new ChCombo("bf","md","up","it","sl","rm","sf","tt").save(Texshade.class,"choiceEmph"),
        choiceShowNumbering=new ChCombo("left","right","none").save(Texshade.class,"choiceShowNumbering"),
        choiceSequenceLogo=new ChCombo(NON_TOP_BOTTOM).save(Texshade.class,"choiceSequenceLogo"),
        choiceRuler=new ChCombo(NON_TOP_BOTTOM).save(Texshade.class,"choiceRuler"),
        choiceSize=new ChCombo(
        "A0","A1","A2","A3","A4","A5","A6","A7","A8","A9",
        "B0","B1","B2","B3","B4","B5","B6","B7","B8","B9",
        "C0","C1","C2","C3","C4","C5","C6","C7","C8","C9",
        "USletter", "USlegal", "USexecutive").s(4).save(Texshade.class,"choiceRuler"),
        choiceSubfamilyLocation=new ChCombo(NON_TOP_BOTTOM).li(this).save(Texshade.class,"choiceSubfamilyLocation");
    private final ChTextField
        tfBargraphstretch=new ChTextField(" 1.0 ").ct(Float.class).saveInFile(SAVE+"bargraphstretch"),
        tfVblockspace=new ChTextField("10").cols(7,true,true).ct(Float.class).saveInFile(SAVE+"vblockspace"),
        tfMarginL=new ChTextField("0").cols(5,true,true).saveInFile(SAVE+"marginL"),
        tfMarginR=new ChTextField("0").cols(5,true,true).ct(Integer.class).saveInFile(SAVE+"marginR"),
        tfMarginT=new ChTextField("5").cols(5,true,true).saveInFile(SAVE+"marginT"),
        tfMarginB=new ChTextField("0").cols(5,true,true).ct(Integer.class).saveInFile(SAVE+"marginB"),
        tfResiduesperline=new ChTextField("auto").saveInFile(SAVE+"residuesperline"),
        tfPaperheight=new ChTextField(" auto ").cols(6,true,true).ct(Integer.class).saveInFile(SAVE+"paperheight"),
        tfCustomWidth=new ChTextField("10").cols(6,true,true).ct(Integer.class).saveInFile(SAVE+"customWidth"),
        tfCustomHeight=new ChTextField("10").cols(6,true,true).ct(Integer.class).saveInFile(SAVE+"customHeight"),
        tfRulerstep=new ChTextField("10").cols(5,true,true).ct(Float.class).saveInFile(SAVE+"rulerstep"),
        tfSubfamilyName=new ChTextField().cols(30,true,false).saveInFile(SAVE+"subfamilyName");
    // ----------------------------------------
    private final TexshadeGraph[] resGraph=new TexshadeGraph[4], consGraph=new TexshadeGraph[4];
    private final TexshadeSecStructure[] secGraph=new TexshadeSecStructure[4];
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Constructor >>> */

    public Texshade() {
        setSettings(this,
                    customizeTexshadeBegin,customizeTexshadeEnd,
                    Customize.getCustomizableForClass(Texshade.class, Customize.CLASS_Example),
                    Customize.customize(Customize.latex),
                    Customize.customize(Customize.pdflatex),
                    Customize.customize(Customize.latexEditors),
                    Customize.customize(Customize.pdfViewer));
        mkdrsErr(file(DIR));
        for(int loc=4; --loc>=0;) {
            final String s=NON_TOP_BOTTOM[loc+1];
            consGraph[loc]=new TexshadeGraph(s, ValueOfAlignPosition.class);
            resGraph[loc]=new TexshadeGraph(s, ValueOfResidue.class);
            secGraph[loc]=new TexshadeSecStructure(s);
        }

        proteinChoiceRuler=new ProteinCombo(align,0);
        panViewers=hBL("3 ",butPdf);
        for(String s : new String[]{TEX_FILE,PDF_FILE,LOG_FILE,AUX_FILE,MSF_FILE,DVI_FILE,PS_FILE}) delFile(file(s));
        proteins=null;
        myDoLayout();
        ChThread.callEvery(1111,runUpdateUI,"Texshade",EDT);
    }
    @Override public void dispose() {
        ChThread.callEvery(-1,runUpdateUI,"",0);
        final ChExec ex=this.myExec;
        if (ex!=null) ex.dispose();
        super.dispose();
    }
    /* <<< Constructor <<< */
    /* ---------------------------------------- */
    /* >>> Layout >>> */
    @Override public Dimension getMinimumSize() { return dim(0,0);}

    private void myDoLayout() {
        final String INFO_LOCATION="The locations \"top and \"ttop\" are above and \"bottom\" and \"bbottom\" below the alignment.";
        {
            Object
                msgRestore=
                "RESTORING THE DEFAULT SETTINGS\n\n"+
                "\nAfter closing STRAP, You can manually delete lines in the property file "+file(PROPERTY_FILE)+
                "\n\nFurther you can delete Files in the folder "+file(ChTextComponents.DIR_SAVED_TEXT+SAVE)+
                "\n\nOther settings are modified/reset using the settings tool button at the top.",
                p=vBhB(
                          StrapAlign.headLine(this),
                          "TeXshading needs 3 steps:",
                          hBL("1 ",new ChButton(BUT_LAB_Source).li(this).tt("Generates the latex source and the multiple sequence file of the proteins.")),
                          hBL("     ","<sup>(From the selected proteins or from a rectangular region)</sup>"),
                          hBL("2 ",butRunTex," ",cbStdout),
                          panViewers,
                          panPoster,
                          "Note: Click buttons with CTRL for more options.",
                          " ",
                          " ",
                          hB(" Restore default settings ",ChButton.actionViewMessage("Restore",msgRestore))
                          );
            adMainTab(pnl(p), this,null);
        }
        {
            final Object
                h="For PDB-files, helices are drawn <font bgcolor=#ff3333> red </font> and sheets <font bgcolor=#ffff00>yellow</font>. "+
                "The secondary structure information of one PDB-protein can be drawn over the alignment.",
                p=vBhB(
                       INFO_LOCATION,
                       hB("<h2>Secondary structure</h2>","#",smallHelpBut(h)),
                       " ",
                       secGraph[0].getPanel(), "#JS",
                       secGraph[1].getPanel(), "#JS",
                       secGraph[2].getPanel(), "#JS",
                       secGraph[3].getPanel()
                       );
            adTab("",iicon(IC_STRUCTURE),pnl(p), "", (JTabbedPane)this);
        }
        {
            final Object
                h=
                "By default the file names of the proteins are written in the PDF-output. "+
                "Alternatively,  different names can be defined using  simple text or  complex LaTeX expressions.<br><br>"+
                "Please not that <b>underscore</b> \"_\" has a special meaning in latex and needs to be quoted.",
                p=vBhB(
                       hB("<h2>Substitute protein name</h2>","#",smallHelpBut(h)),
                       cbSubstituteNames,
                       cbNoExtension
                       );
            adTab("Row Header",null, pnl(p), "", (JTabbedPane)this);
        }
        {
            final Object
                hBar="<i>PACKAGE:charite.christo.strap.interfaces.</i>"+
                "Bar charts can be drawn at four different locations: ttop, top,bottom, bbottom<br>"+
                "The values are calculated by instances of the interface <i>ValueOfAlignPosition</i> / <i>ValueOfAlignPosition</i> "+
                "with the specified aligned proteins.<br>",
                p1=vBhB(
                        hB("<h2>Plotting values of residues</h2>","#",smallHelpBut(hBar)),
                        INFO_LOCATION,
                        hBL("Bargraphstretch=",new ChTextField(tfBargraphstretch.getDocument())),
                        " ",
                        resGraph[0],"#JS",
                        resGraph[1],"#JS",
                        resGraph[2],"#JS",
                        resGraph[3]
                        ),
                p2=vBhB(
                        hB("<h2>Plotting values of alignment positions (\"consensus\")</h2>","#",smallHelpBut(hBar)),
                        INFO_LOCATION,
                        hBL("Bargraphstretch=",tfBargraphstretch),
                        " ",
                        consGraph[0],"#JS",
                        consGraph[1],"#JS",
                        consGraph[2],"#JS",
                        consGraph[3]
                        );
            adTab(null,iicon(IC_PLOT),pnl(p1), "", (JTabbedPane)this);
            adTab(null,iicon(IC_PLOT),pnl(p2), "", (JTabbedPane)this);
        }
        {
            final Object
                h=
                "The output format can be altered.<br><br>"+
                "<b>Postscript:</b> In contrast to PDF viewers  PostScript viewers usually do not recognize page formats larger than A4."+
                "<h2>Printing over several sheets of paper</h2>"+
                "The PostScript utility \"poster\" allows to print the alignment horizontally over several sheets of paper<br>"+
                "Poster (http://www.geocities.com/SiliconValley/5682/poster.html) is a small utility for making a poster -- a large printed image -- from a PostScript document.<br>"+
                "Postscript output must be activated for the alignment since poster does not work with  PDF.<br>"+
                "The respecive check-box becomes accessible by pressing the buttons in the main tab with CTRL.<br>",
                p=vBhB(
                       hBL("<h2>Page layout</h2>","#",smallHelpBut(h)),
                       hBL("Specify the paper format.",choiceSize," ",cbLandscape),
                       "A4 is normal paper size in Europe, A2 has twice the  width and A0 is 4 times larger.",
                       hBL("Left margin=",tfMarginL,"mm"),
                       hBL("Right margin=",tfMarginR,"mm"),
                       hBL("Top margin=",tfMarginT,"mm"),
                       hBL("Bottom margin=",tfMarginB,"mm"),
                       hBL("Residues per Line=",tfResiduesperline),
                       hBL(cbCustomSize," Width=",tfCustomWidth,"mm  Height=",tfCustomHeight,"mm")
                       );
            adTab("Page Layout",null,pnl(p), "", (JTabbedPane)this);
        }
        {
            Object
                p=vBhB(
                       "<h2>Appearance of the Alignment</h2>",
                       hBL("Shading Mode",choiceShading),
                       hBL("Protein name",choiceShowNames),
                       hBL("Residue numbers",choiceShowNumbering),
                       cbNumAsPdb,

                       hBL("Vertical block space ",tfVblockspace,"mm"),
                       hBL(cbFingerprint,cbHideConsensus,cbHideLegend),
                       hBL("Consensus",choiceConsensus,choiceConsensusLocation),
                       hBL("Style for emphblock and emphregion: ",choiceEmph),
                       cbSeparationLine,
                       cbUpperCase
                       );
            adTab("Appearance",null,pnl(p), "", (JTabbedPane)this);
        }
        {
            subfamily_listProteins=new ProteinList(align,0);
            Object h="A sequence logo visualizes the degree of conservation and the prevalence of amino acid letters at each alignment position.";
            JComponent p=vBhB(
                              hB("<h2>Simple Sequence Logo</h2>","#",smallHelpBut(h)),
                              hBL("Location ",choiceSequenceLogo),
                              hBL(cbHideSeqs),
                              hBL(cbDofrequencycorrection),
                              hBL("Ruler: ",choiceRuler," step:",tfRulerstep,proteinChoiceRuler),
                              " ","#JS",
                              "<h2>Advanced Subfamily Logo</h2>",
                              hBL(dim(1,ICON_HEIGHT),"Location:",pnl(choiceSubfamilyLocation), "Name: ",pnl(tfSubfamilyName)),
                              "##",
                              subfamily_listProteins.scrollPane()
                              );
            adTab("Sequ Logo",null,p, "", (JTabbedPane)this);
        }
        String empty="You can type addition LaTeX/TeXshade commands";
        pcp(KEY_IF_EMPTY,empty,taBefore);
        pcp(KEY_IF_EMPTY,empty,taIn);
        taIn.saveInFile(SAVE+"within.txt");
        taBefore.saveInFile(SAVE+"before.txt");
    }
    /* <<< Layout <<< */
    /* ---------------------------------------- */
    /* >>> Proteins >>> */
    private static volatile StrapProtein proteins[];
    public static StrapProtein[] proteins(){ return proteins;}
    public void setProteins(charite.christo.protein.Protein... pp) { proteins=spp(pp);}
    private final ProteinCombo proteinChoiceRuler;
    /* <<< Proteins <<< */
    /* ---------------------------------------- */
    /* >>> Src and MSF >>> */
    private void makeMsfAndSource(final StrapProtein pp[], final int wideColFrom, final int wideColTo) {
        if (pp==null || pp.length==0) return;
        this.proteins=pp;
        for(String s : new String[]{MSF_FILE,TEX_FILE,PDF_FILE,LOG_FILE,AUX_FILE,DVI_FILE,PS_FILE}) delFile(file(s));;
        final long mode=ExportAlignment.MSF| (cbNoExtension.isSelected() ? ExportAlignment.FILE_SUFFIX : 0);
        final Range columnRange=new Range(wideColFrom,wideColTo);
        wrte(file(MSF_FILE),ExportAlignment.export(pp,mode, columnRange,60,null,0f, null));

        Protein.setResidueIndexForExportAsPDB(isSelctd(cbNumAsPdb));
        final StringBuffer tex=joinLines(customizeLatexBegin.getSettings()).append("\n");
        final String texshadePoster,texshadePosterPreamble;
        {
            atoi_debug=false;
            final int i=atoi(tfResiduesperline);
            atoi_debug=true;
            texshadePoster= i==0 ? "" : "\\residuesperline*{"+i+"}\n";
            final String landscape= cbLandscape.isSelected() ? "[landscape]" : "";
            String size="\\setpapersize"+landscape+"{"+choiceSize+"}\n";
            if (cbCustomSize.isSelected()) size="\\setpapersize{custom}{"+tfCustomWidth+"mm}{"+tfCustomHeight+"mm}\n";
            String  s=size+"\\setmargrb{"+tfMarginL+"mm}{"+tfMarginT+"mm}{"+tfMarginR+"mm}{"+tfMarginB+"mm}\n";
            final float height=(float)atof(tfPaperheight,0);
            if (!Float.isNaN(height)) s+="\\setlength{\\paperheight}{"+height+"mm}\n";
            texshadePosterPreamble=s;
        }
        InteractiveDownload.downloadFilesAndUnzip(Web.urls(URL_STRAP_DATAFILES,TEXSHADE+".sty.gz "+VMARGIN+".sty.gz"),dirDataFiles());
        for(String p:new String[]{TEXSHADE,VMARGIN}){
            File fFrom=file(dirDataFiles()+"/"+p+".sty"), fTo=file(DIR+p+".sty");
            if (sze(fFrom)!=sze(fTo) || sze(fFrom)==0) cpy(fFrom,fTo);
        }
        for(String s : new String[]{TEX_FILE,PDF_FILE,LOG_FILE,AUX_FILE,PS_FILE,DVI_FILE}) delFile(file(s));
        if (cbPostscript.isSelected()) rplcSb(0L,"\\documentclass[pdftex]","\\documentclass",tex);
        /* --- setends --- */
        //final String setends;
        tex.append(
                   texshadePosterPreamble+
                   "\\begin{document}\n"+
                   taBefore+
                   joinLines(customizeTexshadeBegin.getSettings())+
                   taIn+
                   (choiceShowNames.i()!=2 ? "\n\\shownames{"+choiceShowNames+"}\n" : "\\hidenames\n")+
                   "\\emphdefault{"+choiceEmph+"}\n"+
                   (choiceShowNumbering.i()!=2 ? "\n\\shownumbering{"+choiceShowNumbering+"}\n" : "\\hidenumbering\n")+
                   "\\vblockspace{"+tfVblockspace+"mm}\n"+
                   (cbSeparationLine.isSelected()? "\\separationline{"+pp.length+"}\\n" :"")+
                   (cbHideLegend.isSelected() ? "\\hidelegend\n" : "\\showlegend\n")+
                   (cbHideConsensus.isSelected() ? "\\hideconsensus\n" : "")+
                   (choiceConsensusLocation.i()>0 ?
                    "\\defconsensus{{$\\bullet$}}{{$\\bullet$}}{{$\\bullet$}} "+
                    "\\showconsensus["+choiceConsensus+"]{"+choiceConsensusLocation+"}" : "")+
                   (cbFingerprint.isSelected() ? "\\fingerprint{"+maxi(30,(columnRange.getTo()-columnRange.getFrom()+2))+"}\n" :"")+
                   texshadePoster+
                   nameseq()+
                   "\\bargraphstretch{"+tfBargraphstretch+"}\n"
                   );
        tex.append("\n% The setlength commands do not change the output, but may enhance speed \n");
        for(int iP=0;iP<pp.length;iP++) {
            final StrapProtein p=pp[iP];
            int i0=p.getResidueIndexForExport(p.column2nextIndex(wideColFrom));  // Bei Erhoehung rutscht alles nach r.
            if (i0!=1) tex.append("\\startnumber{"+(iP+1)+"}{"+i0+"}\n");
            if (!cbUpperCase.isSelected()) {
                final byte seq[]=p.getResidueType();
                boolean lc[]=null;
                for(int i=seq.length; --i>=0;) {
                    if (is(LOWR,seq,i)) (lc==null ? lc=new boolean[i+1] : lc)[i]=true;
                }
                if (lc!=null) tex.append("\\lowerregion{"+(iP+1)+"}{"+SelectionVariables.texshadeRange(lc,0,p)+"}\n");
            }
            for(ResidueAnnotation s:p.getResidueAnnotations()) {
                if (countTrue(s.getSelectedAminoacids())==0) continue;
                for(ResidueAnnotation.Entry e: s.getEntries()) {
                    if (e.isEnabled() && CLASS_Texshade.equals(e.getKey())) {
                        final boolean bb[]=s.getSelectedAminoacids();
                        final int offset=s.getSelectedAminoacidsOffset();
                        tex.append((Object)SelectionVariables.replaceBB(e.getValue(), bb,offset,s.getName(),p,proteins))
                            .append(" % ").append(p).append('\n');
                    }
                }
            }
        }
        tex.append("\n");
        if (onlyOneAnnotation==null) {
            final String s=StrapView.getShadingColorsName();
            tex.append("\\threshold{"+StrapView.sliderSimilarity().getValue()+"}\n");
            final int i=choiceShading.i();
            if (i==1 && ("charge".equals(s) || "hydropathy".equals(s) || "chemical".equals(s))) tex.append( "\n\\shadingmode["+s+"]{functional}\n");
            else if (i>0) tex.append("\n\\shadingmode{"+choiceShading+"}\n");
            if (choiceSequenceLogo.i()!=0 || choiceSubfamilyLocation.i()!=0) {
                if (cbHideSeqs.isSelected()) tex.append("\\hideseqs\n");
                if (cbDofrequencycorrection.isSelected()) tex.append("\\dofrequencycorrection\n");
            }
            if (choiceSequenceLogo.i()!=0) tex.append("\\showsequencelogo{"+choiceSequenceLogo+"} \\showlogoscale{leftright}\n");

            for(int loc=4;--loc>=0; ) {
                resGraph[loc].makeTex(tex,wideColFrom,wideColTo);
                consGraph[loc].makeTex(tex,wideColFrom,wideColTo);
                if (ProteinList.isEnabled(pp, ProteinList.SECSTRU)) secGraph[loc].makeTex(tex);
            }
            DialogPredictTexshade.getTexts(tex,pp);
        }
        if (choiceRuler.i()>0) {
            final StrapProtein p= (StrapProtein)proteinChoiceRuler.getProtein();
            final int idx=idxOf(p,pp);
            if (idx>=0) tex.append("\\showruler{"+choiceRuler+"}{"+(idx+1)+"}\n\\rulersteps{"+tfRulerstep+"}\n");
        }
        subfamily_makeTex(tex);
        join(customizeTexshadeEnd.getSettings(),"\n",tex.append('\n'));
        tex.append("\n\\end{document}\n");
        wrte(file(TEX_FILE),tex);
    }
    /* <<< Src and MSF <<< */
    /* ---------------------------------------- */
    /* >>> Error >>> */
    private static void errorWindows() {
        JComponent msg=
            vBhB(
                 "Cannot launch LaTex. There are two possible reasons:<ol>"+
                 "<li>LaTex is not yet installed.</li>"+
                 "<li>The program path variable is not correctly set.</li></ol><br><br>",
                 "Solution: ",
                 hBL("Either ", Microsoft.newCygwinButton(Microsoft.CYGWIN_LATEX).actionClose(CLOSE_PARENT_WINDOW)),
                 " ",
                 hBL("Or instal  ","WIKI:Miktex")
            );
    error(0L,"notLaunchLatex",msg);
}
    /* <<< Error <<< */
    /* ---------------------------------------- */
    /* >>> Exec >>> */
    public void compute(){
        makeMsfAndSource(proteins,0,MAX_INT);
        ChDelay.newThread(this,"RUN",null).run();
        viewFile(filePdfCopied,0);
    }
    private ChExec myExec;
    private final static ChFrame frameExec=new ChFrame("Texshade messages");
    /* <<< Exec <<< */
    /* ---------------------------------------- */
    /* >>> Threading >>> */
    private static String onlyOneAnnotation;
    public Object run(String id,Object arg)  {
        if (id=="ENABLE_DISABLE") {
            final boolean finished=sze(filePdfCopied)>0;
            butPdf.repaint(); /* Workaround ??? */
            setEnabld(finished, butPdf);
            final ChExec ex=this.myExec;
            setEnabld(sze(file(TEX_FILE))>0 && sze(file(MSF_FILE))>0 && (ex==null ||!ex.isRunning()), butRunTex);
            setEnabld(sze(file(DVI_FILE))>0, butDvi);
            setEnabld(sze(file(DVI_FILE))>0, butDviPs);
            setEnabld(sze(file(PS_FILE))>0, butPS);
            setEnabld(sze(file(PS_FILE))>0, butPoster);
            setEnabld(sze(file(PS_POSTER_FILE))>0, butPosterViewPS);
        }
        if (id=="RUN") {
            for (String s : new String[]{PDF_FILE,AUX_FILE,LOG_FILE,PS_FILE,DVI_FILE}) delFile(file(s));
            final int nPass=arg!=null ? 1 : 2;
            for(int pass=nPass; --pass>=0;) {
                final ChExec ex=this.myExec=new ChExec(ChExec.STDOUT | ChExec.STDERR | ChExec.IGNORE_ERROR | (isSelctd(cbStdout)?ChExec.STDOUT_IN_TEXT_BOX:0))
                    .showStdoutAndStderr(frameExec).dir(file(DIR));
                if (pass==0) addActLi(this,ex);
                ex.setCommandsAndArgument(getCustomSettings(cbPostscript.isSelected() ? Customize.latex : Customize.pdflatex),file(TEX_FILE).getName());
                ex.run();
            }
            copyPdf();
            if (0==sze(file(cbPostscript.isSelected() ? DVI_FILE : PDF_FILE))) ChDelay.invokeAndWait(this,"ERROR",null);
        }
        if (id==ResSelUtils.RUN_TEX_ONE_PROTEIN) {
            onlyOneAnnotation=toStrg(arg);
            compute();
            onlyOneAnnotation=null;
        }
        if (id=="ERROR") {
            final String sLog=readStrg(file(LOG_FILE));
            if (sLog==null) {
                if (isWindows()) {errorWindows();}
                else {
                    error(0L,null,
                          vBhB(
                                "<h2>Failed to run LaTeX</h2>",
                                getCtrlPanel(myExec),
                                "</pre>In case LaTeX is installed please check the binary path.<br>"+
                                "In case LaTeX is not yet installed install it now:<br>"+
                                "Depending on the system  the software group might be \"type setting\" or \"text\".<br>"+
                                "The package name might be \"tetex-bin\" or \"tetex\"."+
                                "After successful installation the shell command \"pdflatex\" should exist."+
                                ""
                                )
                          );
                }
            } else {
                if (sLog.indexOf("capacity exceeded")<0) {
                    error(0L,"LaTeX no output",
                          vBhB("<h2>No output written by LaTeX.</h2>\n\n",
                               isWindows() ? Microsoft.newCygwinButton(Microsoft.CYGWIN_LATEX) : null)
                          );
                } else {
                    final JComponent
                        panBut=isWindows() && sze(file(dirCygwin()+Insecure.TEXMF_DOT_CONF))>0 ?
                        new ChButton("INCREASE_MEM").t("Let STRAP solve the problem").li(this) : null,
                        msg=vBhB(
                                  "LaTeX reports that the memory  capacity exceeded ",
                                  panBut,
                                  "<br>Please use less proteins or configure the  memory limits in the LaTeX configuration file "+
                                  "<pre> texmf.cnf </pre>"+
                                  "Before modifying this config file you should keep a copy of the original form.<br>"+
                                  "On UNIX/LINUX this file belongs to root and you must login as  superuser.<br>"+
                                  "Increase (multiply by 10 or 100) all  memory values that have two or more zero digits at the end. <br>"+
                                  "Do not change the comment line which are those preceed  by a percent sign<br>"+
                                  "Then run as superuser "+
                                  "<pre class=\"terminal\">   fmtutil --byfmt=pdflatex </pre>"+
                                  "and tex the alignment again. <br>"+
                                  "More details are found in the  help text."
                                  );
                    error(0L,"LaTeX ERROR",msg);
                }
            }
        }
        return null;
    }
    private File filePdfCopied;
    private int filePdfCopiedCount=0;
    private void copyPdf() {
        ((ChButton) butPdf).actionViewFile(filePdfCopied=file(STRAPOUT+"/align"+filePdfCopiedCount++ +".pdf"));
        setEnabld(true, butPdf);
        cpy(file(PDF_FILE),filePdfCopied);
    }
    private final Runnable runUpdateUI=ChDelay.newThread(this,"ENABLE_DISABLE",null);
    private String nameseq() {
        final StrapProtein pp[]=proteins;
        if (pp==null) return "";
        final StringBuffer sb=new StringBuffer();
        for(int iP=0;iP<pp.length;iP++) {
            final String name=pp[iP].getName();
            final String newName= !cbSubstituteNames.isSelected() ? null : readStrg(DIR_ANNOTATIONS+"/"+pp[iP]+".texName");
            if (nxt(NO_SPC,newName)>=0) sb.append("\\nameseq{").append(iP+1).append("}{").append(newName).append("} % ").append(name).append("\n");
        }
        return sb.toString();
    }
    /* <<< nameseq <<< */
    /* ---------------------------------------- */
    /* >>> Customize >>> */
    private final ChTextArea taBefore=new ChTextArea("\n\n"), taIn=new ChTextArea("\n\n"), taSourceFiles=new ChTextArea("").ro();
    private final static String LATEXBEGIN[]={
        "\\documentclass[pdftex]{letter}  \n"+
        "% NOTE: the above line is suitable for the command pdflatex but not for the command latex .\n"+
        "% when 'latex' is selected the  String '\\documentclass[pdftex,' is automatically replaced by '\\documentclass[' to get rid of the pdflatex-option which would cause an error \n"+
        "\\usepackage{"+TEXSHADE+"}\n\\usepackage{"+VMARGIN+"}\n"+
        "\\usepackage{color}\n"+
        "\\usepackage{hyperref}\n"+
        "\n\n"};
    private static Customize
        customizeLatexBegin=new Customize("latexBegin",LATEXBEGIN,helpBeginEnd+"\nThis is the beginning part.",Customize.LATEXBEGIN),
        customizeTexshadeEnd=new Customize("texshadeEnd",new String[] {"\\end{texshade}\n"}, helpBeginEnd+"\nThis is the end part.",Customize.LATEXEND),
        customizeTexshadeBegin=new Customize("texshadeBegin",new String[] {
                "\\pagestyle{empty}\n"+
                "\\begin{texshade}{"+file(MSF_FILE).getName()+"}\n"+
                "\\seqtype{P}\n"+
                "\\gapchar{.}\n"
            },"\nThis is the beginning part for texshade.",Customize.LATEXBEGIN);
    public static String getCustomizableExamples() {
        return
            "\\feature{top}{$TEX_PROTEIN}{$TEX_RESIDUES}{box[Red]}{peptide}\n"+
            "\\feature{ttop}{$TEX_PROTEIN}{$TEX_RESIDUES}{box[Red]}{peptide}\n"+
            "\\feature{bottom}{$TEX_PROTEIN}{$TEX_RESIDUES}{fill:$\\uparrow$}{active site}\n"+
            "\\feature{top}{$TEX_PROTEIN}{$TEX_RESIDUES}{fill:$\\downarrow$}{active site}\n"+
            "\\feature{bottom}{$TEX_PROTEIN}{$TEX_RESIDUES}{fill:$\\uparrow$}{$$FIRST_A_{$FIRST_INDEX}$ - $$LAST_A_{$LAST_INDEX}$}\n"+
            "\\feature{bottom}{$TEX_PROTEIN}{$TEX_RESIDUES}{fill:$\\uparrow$}{$$FIRST_A_{$FIRST_NUMBER}$ - $$LAST_A_{$LAST_NUMBER}$}\n"+
            "\\feature{bottom}{$TEX_PROTEIN}{$TEX_RESIDUES}{fill:$\\uparrow$}{$$FIRST_A_{$FIRST_NUMBER:$CHAIN}$ - $$LAST_A_{$LAST_NUMBER:$CHAIN}$}\n"+
            "\\frameblock{$TEX_PROTEIN}{$TEX_RESIDUES}{Red[2pt]}\n"+
            "\\emphregion{$TEX_PROTEIN}{$TEX_RESIDUES}\n"+
            "\\emphblock{$TEX_PROTEIN}{$TEX_RESIDUES}\n"+
            "\\shaderegion{$TEX_PROTEIN}{$TEX_RESIDUES}{Red}{Green}\n"+
            "\\shadeblock{$TEX_PROTEIN}{$TEX_RESIDUES}{Red}{Green}\n"+
            "\\feature{bottom}{$TEX_PROTEIN}{$TEX_RESIDUES}{fill:$\\uparrow$}{$$FIRST_Aaa_{$FIRST_INDEX}$}\n"+
            "\\feature{bbottom}{$TEX_PROTEIN}{$TEX_RESIDUES}{fill:X}{cleavage}\n"+
            "\\fingerprint{300}\n"+
            "\\separationline{$TEX_PROTEIN}";
    }
    /* <<< Customize <<< */
    /* ---------------------------------------- */
    /* >>> AWTEvent  >>> */
    public void handleEvent(StrapEvent ev) { if (ev.getType()==StrapEvent.PROTEINS_KILLED) proteins=null; }
    public void actionPerformed(java.awt.event.ActionEvent ev) {
        StrapAlign.errorMsg("");
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        final boolean isSelected=isSelctd(q);
        final boolean ctrl=isCtrl(ev), shift=isShift(ev);

        if (q==myExec && cmd==ChExec.ACTION_ERROR && isWindows()) errorWindows();
        if (q==choiceSubfamilyLocation) tfSubfamilyName.setEnabled(choiceSubfamilyLocation.i()>0);
        if (q==butPoster) {
            Object size=choiceSize.getSelectedItem(),posterSize=choicePosterSize.getSelectedItem();
            if (posterSize=="same") posterSize=size;
            ChExec ex=new ChExec(isSelctd(cbStdout)?ChExec.LOG:0).dir(file(DIR))
                .setCommandsAndArgument(getCustomSettings(Customize.psPoster)," -i"+size+" -p"+posterSize+"  -o "+PS_POSTER_FILE+" "+file(PS_FILE).getName());
            if (ctrl) ex.log(Customize.customize(Customize.psPoster));
            new Thread(ex).start();
        }
        if (q==cbCustomSize) {
            choiceSize.setEnabled(!isSelected);
            tfCustomHeight.setEnabled(isSelected);
            tfCustomWidth.setEnabled(isSelected);
        }
        if (q==cbPostscript) {
            panPoster.removeAll();
            if (isSelected) {
                panPoster.add(pnl(choicePosterSize,butPoster, butPosterViewPS, ChButton.actionOpenURL("http://www.geocities.com/SiliconValley/5682/poster.html").t("Web page<br>of poster")));
            }
            panPoster.setBorder(isSelected ? new TitledBorder("Scale and tile alignment output to print on multiple pages:"):null);
            final JComponent p=this.panViewers;
            p.remove(butDvi); p.remove(butDviPs); p.remove(butPS); p.remove(butPdf);
            if (isSelected) { p.add(butDvi,1); p.add(butDviPs,2); p.add(butPS,3); } else p.add(butPdf,1);
            panViewers.revalidate();
            delFile(file(TEX_FILE));
            setTxt(isSelected ? "Make PS":BUT_LAB_RunTex, butRunTex);
        }
        if (cmd==BUT_LAB_Source) {
            final StrapProtein pp[]=StrapView.getProteinsInRectangleForAlignment(1);
            if (sze(pp)==0) { StrapAlign.errorMsg("No proteins"); return;}
            final Rectangle r=StrapView.getRectangle(null);
            makeMsfAndSource(pp, r!=null ? r.x : 0, r!=null ? r.x+r.width-1 : MAX_INT);
            taSourceFiles.setText("Generated Multiple Sequence file: \n "+ filepathUnixStyle(file(MSF_FILE))+"\n\n Genereated LaTeX-file:\n "+filepathUnixStyle(file(TEX_FILE))+"\n");
            taSourceFiles.tools().underlineRefs(0);
            if (ctrl) {
                if (frameSourceFiles==null)  {
                    final JComponent
                        pC=vBhB("#TBem Customize Source",
                                 hB(customizeLatexBegin.newButton().t("preamble")," ",customizeTexshadeBegin.newButton().t("\\begin{texshade}texshade")," ",customizeTexshadeEnd.newButton().t("end")),
                                 " ", taBefore,
                                 " ", taIn
                                 ),
                        pnl=vBhB(
                                 cbPostscript,
                                 pC,
                                 hB("#EBem",taSourceFiles)
                                 );
                    frameSourceFiles=new ChFrame("Texshade").ad(pnl).size(444,444).shw(CLOSE_ON_CtrlW);
                }
                taBefore.setBorder(new TitledBorder("LaTeX code between \\begin{document} and \\begin{texshade}"));
                taIn.setBorder(new TitledBorder("LaTeX code between  \\begin{texshade} and \\end{texshade}"));
                taSourceFiles.setBorder(new TitledBorder("Generated files"));
                frameSourceFiles.size(500,500).shw(0);
            }
        }
        if (q==butRunTex) {
            final ChExec ex=this.myExec;
            if (ctrl && ex!=null) ex.log(Customize.customize(cbPostscript.isSelected() ? Customize.latex : Customize.pdflatex));
            filePdfCopied=null;
            ChDelay.newThread(this,"RUN", shift?"":null).start();
        }
        if (q==cbSubstituteNames && isSelected)  showTexshadeNameseq();
        if (cmd=="INCREASE_MEM") {
            final File fTexmf=file(dirCygwin()+Insecure.TEXMF_DOT_CONF),fOri=file(fTexmf+".original");
            if (sze(fOri)==0) cpy(fTexmf,fOri);
            final String ss[]=readLines(fTexmf);
            if (ss==null) return;
            for(int i=ss.length; --i>=0;) {
                final String s=ss[i].trim();
                if (s.startsWith("%")) continue;
                if (s.endsWith("00")) { ss[i]=s+"00"; continue;}
                int idx00=s.indexOf("00 ");
                if (idx00<0) idx00=s.indexOf("00\t");
                if (idx00<0) continue;
                ss[i]=s.substring(0,idx00+2)+"00"+s.substring(idx00+2);
            }
            wrte(fTexmf,joinLines(ss)+"\n\n Modified by STRAP\n");
            new Thread(new ChExec(ChExec.LOG).setCommandLine("CYGWINSH fmtutil --byfmt=pdflatex\nfmtutil --byfmt=latex")).start();
            final Object msg=vBpnl(pnl("I modified the file ",fOri," ==> ", fTexmf),"Please wait for the fmtutil to finish and then try again.");
            ChMsg.msgDialog(0L, vBpnl("fmtutil",msg));
        }
    }
    /* <<< AWTEvent <<< */
    /* ---------------------------------------- */
    /* >>> subfamilyLogo >>> */
    private StringBuffer subfamily_makeTex(StringBuffer sb) {
        if (choiceSubfamilyLocation.i()==0) return sb;
        StrapProtein pp[]=subfamily_listProteins.getSelectedProteins();
        if (pp==null || pp.length==0) pp=proteins();
        String sep="";
        sb.append("\n\\setsubfamily{");
        for(StrapProtein p : pp) {
            int idx=idxOf(p,proteins());
            if (idx<0) continue;
            sb.append(sep).append(idx+1);
            sep=",";
        }
        sb.append("}\n").append("\\showsubfamilylogo{")
            .append(choiceSubfamilyLocation).append("}\n")
            .append("\\showlogoscale{leftright}\n")
            .append("\\namesubfamilylogo[others]{")
            .append(tfSubfamilyName).append("}\n\n");
        return sb;
    }
    private ProteinList subfamily_listProteins;
    {
        tfSubfamilyName.setEnabled(false);
        pcp(KEY_IF_EMPTY,"enter the name of the logo",tfSubfamilyName);
    }
    /* --- NameSeq --- */
    private ChFrame frameSourceFiles;
    private static Container _panProtNam;
    private void showTexshadeNameseq() {
        StrapProtein pp[]=align.getSelectedProteins();
        if (pp.length==0) pp=align.getVisibleProteins();
        if (pp.length==0) return;
        if (_panProtNam==null) _panProtNam=vBhB();
        else _panProtNam.removeAll();

        final String info="<h2>Change displayed protein names</h2>"+
            "The typed name must follow the syntax of LaTeX.<br>"+
            "E.g. an unquoted underscore '_' would lead to a LaTeX-error.<br>";
        _panProtNam.add(hBL(info));

        for(StrapProtein p : pp) {
            final String key="TS$$PPN";
            Container pan=gcp(key,p,Container.class);
            if (pan==null) {
                final ChTextField tf=new ChTextField();
                tf.tools().saveInFile(file(DIR_ANNOTATIONS+"/"+p+".texName"));
                pan=hB(p.getName()," ", tf);
                pan.setPreferredSize(dim(80*EM,ICON_HEIGHT+2));
                pcp(key,pan,p);
            }
            _panProtNam.add(pan);
        }
        ChFrame.frame("Substitute names",_panProtNam,ChFrame.SCROLLPANE).shw(ChFrame.AT_CLICK);
    }
    /* <<< subfamilyLogo <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */
}
/*
\lowerregion und \lowerblock

\allmatchresidues{Black}{White}{upper}{}
\nomatchresidues{Black}{White}{upper}{}
\conservedresidues{Black}{White}{upper}{}
\similarresidues{Black}{White}{upper}{}
*/
