package charite.christo.strap;
import charite.christo.*;
import charite.christo.protein.*;
import javax.swing.JComponent;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
class TexshadeSecStructure  {
    private final String where;
    private final ChButton cbSkip,toggle;
    private JComponent panel;
    private ProteinCombo choiceProtein;
    private final ChCombo choiceSheetBox;
    private final ChTextField tf_Helix1, tf_Sheet1, tf_Helix2, tf_Sheet2;
    TexshadeSecStructure(String where) {
        this.where=where;
        final String save=Texshade.SAVE+where+" ";
        tf_Helix1=new ChTextField().cols(10,true,true).saveInFile(save+"tf_Helix1");
        tf_Sheet1=new ChTextField().cols(10,true,true).saveInFile(save+"tf_Sheet1");
        tf_Helix2=new ChTextField().cols(10,true,true).saveInFile(save+"tf_Helix2");
        tf_Sheet2=new ChTextField().cols(10,true,true).saveInFile(save+"tf_Sheet2");
        cbSkip=toggl("skip sequence").save(TexshadeSecStructure.class,"skip_"+where)
            .tt("You may not want to show the protein<br>which texshade retrieves the <br>secondary structure information from.");
        toggle=toggl().s(where=="top").save(TexshadeSecStructure.class, "enable_"+where);
        choiceSheetBox=new ChCombo("box","---","===","-->","'->","<-|","<=>",",-,","|=|","helix").s("---").save(TexshadeSecStructure.class,"choiceSheetBox_"+where);
    }
    JComponent getPanel() {
        if (panel==null) {
            choiceProtein=new ProteinCombo(StrapAlign.getInstance(),ProteinList.SECSTRU);
            choiceProtein.save(TexshadeSecStructure.class,"choiceProtein_"+where);
            final JComponent
                pan0=hBL(choiceProtein, " ",cbSkip.cb()),
                pan1=hBL("\\feature{",where,"}{ ... }{helix[RED]",tf_Helix1,"}{",tf_Helix2,"}"),
                pan2=hBL("\\feature{",where,"}{ ... }{",choiceSheetBox,"[YELLOW]:",tf_Sheet1,"}{",tf_Sheet2,"}");
            panel=vBhB(hBL(toggle,"Location : ",where,pan0),  pan1, pan2);
            toggle.actionCollapse(false,new Object[]{pan1,pan2,choiceProtein, pan0});
        }
        return panel;
    }
       StringBuffer makeTex(StringBuffer sb) {
        sb.append("% secondary structure \n");
        StrapProtein protein=choiceProtein.getProtein();
        if (!toggle.isSelected() || protein==null) return sb;

        if (protein.getResidueSecStrType()==null) {
            final StrapProtein pp[]=Texshade.proteins();
            if (pp!=null)
                for(StrapProtein p:pp)
                    if (p.getResidueSecStrType()!=null) {
                        choiceProtein.s(protein=p);
                        break;
                    }
        }

        final int i=idxOf(protein,Texshade.proteins());
        if (i<0) { error("Secondary structure of "+protein+" will not be included, <br>because the protein is not selected."); return sb; }
        final byte ss[]=protein.getResidueSecStrType();
        if (ss==null) return sb.append("% TexshadeSecStructure: ").append(protein).append(" has no secondary structures \n");
        //final int iFrom=protein.column2nextIndex(colFrom);
        //final int iTo=protein.column2thisOrPreviousIndex(colTo)+1;
        texshade(i+1,ss, protein/*,iFrom,  iTo*/,sb);
        if (cbSkip.isSelected()) sb.append("\\hideseq{"+(i+1)+"} % "+protein+"\n");
        return sb;
    }
    private StringBuffer texshade(int num, byte ss[],StrapProtein p,/*int iFrom, int iTo,*/StringBuffer sb) {
        final boolean[] sheet=new boolean[ss.length], helix=new boolean[ss.length];
        boolean hasSheet=false,hasHelix=false;
        //for(int i=iFrom;i<ss.length && i<iTo;i++) {
        for(int i=0;i<ss.length;i++) {
            switch(ss[i]|32) {
            case 'h':hasHelix=helix[i/*-iFrom*/]=true;break;
            case 'e':hasSheet=sheet[i/*-iFrom*/]=true;break;
            }
        }
        if (hasHelix) sb.append("\\feature{"+where+"}{"+num+"}{"+SelectionVariables.texshadeRange(helix,0,p)+"}{helix[Red]:"+tf_Helix1+"}{"+tf_Helix2+"}  % "+p+"\n");
        if (hasSheet) sb.append("\\feature{"+where+"}{"+num+"}{"+SelectionVariables.texshadeRange(sheet,0,p)+"}{"+choiceSheetBox+"[Yellow]:"+tf_Sheet1+"}{"+tf_Sheet2+"}  % "+p+"\n");
        return sb;
    }

}
