package charite.christo.strap;
import charite.christo.protein.*;
import charite.christo.*;
import charite.christo.interfaces.*;
import java.awt.*;
import java.io.File;
import java.util.*;
import java.util.List;
import javax.swing.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
import static charite.christo.protein.ProteinViewer.*;
import static charite.christo.protein.Protein3dUtils.*;
import java.awt.event.*;
import static java.awt.event.MouseEvent.*;
import static java.awt.event.KeyEvent.*;
import static charite.christo.strap.StrapUtils.*;
/**HELP
   <br><br><b>Menu-bar: </b>

   The embedded 3D-views that are shown within the STRAP application
   frame come along with a menu-bar.

   This means that if the user clicks inside the 3D-view, a previous
   active menu-bar disappears and the menu-bar of the 3D-view is shown
   instead.

   Initially, this is the generic menu-bar that is common to all different 3D-views in Strap.

   Some 3D-programs have their own menu bar.
   In this case, the user can swap between the native and the generic menu-bar.

   To get back the menu-bar of the STRAP-alignment, the user needs to click into the alignment panel
   hitting a  protein that is not displayed three-dimensionally or into the scroll-bar of the  alignment panel.

   <br><br><b>Undock the viewer pane: </b>

   3D-viewers that are located inside the STRAP-application frame can be moved outside (undocking).
   For this purpose the title header needs to be dragged.

   <br><br><b>Always floating on top of other windows</b>

   The Application frame of 3D-viewers written in Java can be set above all other windows: Type ctrl-T.

   <br><br><b>Undock menu items: </b>

   The user may want to change the style (e.g. spheres, sticks or surface) for many different groups of amino
   acids or atoms.
   Opening the style-menu each time, is tedious.

   For convenience, menu items can be undocked by mouse-drag.
   They are then located on the desktop and can be used directly.

   <br><br><b>Picking Atoms in the 3D-View: </b>

   When atoms are clicked in the 3D-view, Strap will display a
   toolbar with buttons that act upon the clicked amino acid.

   One button is used to set the alignment cursor on the clicked residues.
   Two further buttons are used to add the clicked amino acid to a residue selection or to remove the amino acid.

   <br><br><b>Movements of the alignment cursor: </b>

   The alignment cursor position is passed to the 3D viewer.
   This can be used to change the display style of a particular amino acid.

   <br><br><b>Selecting amino acids in the alignment panel: </b>
   Using the mouse or using arrow-left/ arrow-right keys together with shift, a number of amino acids can be selected
   in STRAP. This selection is passed to the 3D viewer.
Alternatively, a residue selection can be dragged with the mouse into the 3D-view.

   <br><br><b>Adding proteins: </b>

   Proteins can be added by WIKI:Drag_and_drop.
   See <i>STRING:ChConstants#MOVIE_Drag_to_another_STRAP</i>

   This is not yet supplorted for JMol.

   <br><br><b>Displaying Sequence Features: </b>

   Sequence Features can highlighted in the 3D-view using  WIKI:Drag_and_drop.
   See <i>STRING:ChConstants#MOVIE_Sequence_Features_in_3D</i>

   @author Christoph Gille

*/
public class V3dUtils implements ChRunnable, ChProcessEvent, TooltipProvider {
    public final static int OPEN_VERBOSE=1, OPEN_ASK=1<<1, OPEN_SEND_ALL_ANNOTATIONS=1<<2, OPEN_NOT_ADD_PANEL=1<<3,
        OPEN_NARROW_OTHERS=1<<4, OPEN_ADD_TAB=1<<5, OPEN_INIT_SCRIPT=1<<6;
    final static String
    TOG_CURSOR_SELECT="Select alignment cursor position in 3D",
        BUT_DEFAULT_VIEWER="   \u25baFull featured 3D-view",
        BUT_NATIV_MENUS="PVP$$NM",
        BUT_SURF_OBJ="SURF_OBJ",
        BUT_COLOR="COLOR",
        BUT_LABEL="LABEL",
        KEY_WHICH_MENU="PVU$$TN",
        KEY_PROTEIN_VIEWER="PV$$KPV",
        KOPT_IS_PREVIEW="PV$$KIP",
        INFO_HETERO_NOT_SELECTABLE="No effect for simple backbone view";
    private final static String ACTION_POPUP_PV="V3D$$appv";
    final static List vStyleButton=new ArrayList();
    private final static Object KEY_PV_CMD=new Object();
    private final static ProteinViewer[] ONE_VIEWER={null,null, null};
    /* ---------------------------------------- */
    /* >>> Instance, Constructor >>> */
    private final ChEventAdapter LI=new ChEventAdapter(this);
    static ChEventAdapter li() { return instance().LI; }
    private final ContextObjects contextObjects=new ContextObjects(false,ONE_VIEWER);
    private static V3dUtils _inst;
    static V3dUtils instance() {  if (_inst==null) _inst=new V3dUtils();  return _inst;  }
    /* <<< Instance <<< */
    /* ---------------------------------------- */
    /* >>> open3dViewer >>> */
    public static ProteinViewer open3dViewer(long options, Protein pp[],Object clazzOrViewer) {
        final Class clazz=HotswapUtil.getClass(clazzOrViewer);
        final ProteinViewer inSameView=deref(clazzOrViewer,ProteinViewer.class);
        if (clazz==null || sze(pp)==0 ||
            (options&OPEN_ASK)!=0 && !ChMsg.yesNo("Load "+ (pp.length==1 ? "the protein "+pp[0] : pp.length+" proteins ")+"\n into the protein viewer?")) return null;
        if ((options&OPEN_VERBOSE)!=0) setNextProcessHasLogPanel(true);
        ProteinViewer oneViewer=null;
        nextProtein:
        for(Protein p : pp) {
            if (p==null || (p.getResidueCalpha()==null && p.getDnaAndRnaAndHeteros().length==0)) continue;
            if (getViewerForProtein(p, inSameView)!=null) continue;
            final ProteinViewer pv=HotswapUtil.instance(clazz,ProteinViewer.class,true);
            if (pv==null) {
                error("Error in V3dUtils: cannot instantiate "+clazz);
                return null;
            }
            final long properties=flags(pv);
            if ((properties&PROPERTY_EXTERNAL_PROCESS)!=0) {
                StrapAlign.drawMessage("To see details of the external process,\n"+
                                       "hold ctrl when pressing the button to start the 3D viewer.");
            }
            final boolean isWire=clazz==Protein3d.PView.class;
            if (pv.setProtein(options, p, oneViewer!=null ? oneViewer:inSameView)) {
                transform3D(p.getRotationAndTranslation(), pv);
                if (0!=(options&OPEN_INIT_SCRIPT)) {
                    setSelAminos(0L, toSelectionT("$FIRST-$LAST",p), "",  pv);
                    for(String s : getCustomSettings(clazz, Customize.CLASS_InitCommands)) sendCommand(0L, s,pv);
                }
                if (!isWire) {
                    if (0!=(options&OPEN_SEND_ALL_ANNOTATIONS)) residueSelectionsTo3D(RSto3D_COMMANDS|RSto3D_COLOR,((StrapProtein)p).getResidueAnnotations(), new ProteinViewer[]{pv});
                    new StrapEvent(Dialog3DViewer.class,StrapEvent.PROTEIN_VIEWER_LAUNCHED).setParameter(pv).run();
                }
                if (isWire || p.getFileMayBeWithSideChains()!=null)  p.addProteinViewer(pv);
                if (oneViewer==null) oneViewer=pv;
            }
            final Component pan=viewerCanvas(pv);
            addCP(KEY_DISPOSABLE, pv, pan,5);
            final Object msg=newCNSEW(
                                      "If you came here because you intended to drag a protein with the mouse:<br>"+
                                      "This is not the right way to drag a protein.<br>"+
                                      "You can start dragging a protein  for example from the respective row header of the sequence alignment.<br>"+
                                      "The row-header displays the protein names. Here you can initiater Drag-and-Drop.",
                                      null,pnl(WATCH_MOVIE+MOVIE_Export_Proteins),null,null);
            pcp(KEY_UNDOCK_MESSAGE, msg, pv);
            pcp(KEY_UNDOCK_MESSAGE, msg, pan);
        }
        if (oneViewer!=null) {
            setPV(PV_FOCUSED,oneViewer);
            if (0!=(options&OPEN_INIT_SCRIPT)) {
                for(String s : getCustomSettings(clazz, Customize.CLASS_LateCommands)) sendCommand(0L, s,oneViewer);
            }
            if (supports(COMMANDcenter_amino, oneViewer)) sendCommand(0L, COMMANDcenter_amino, oneViewer);
            final Component panel=viewerCanvas(oneViewer);
            if (panel!=null) {
                pcp(KOPT_UNDOCKABLE, "", panel);
                final Object ref=weakRef(oneViewer);
                disposeOnCtrlW(0L, panel, getViewersSharingView(oneViewer));
                pcp(KEY_PROTEIN_VIEWER, ref, panel);
                for(Component child : childs(panel, Component.class)) {
                    pcp(KEY_PROTEIN_VIEWER, ref, child);

                    StrapAlign.newDropTarget(child,false);
                    li().addTo("mMk",child);
                    ProteinViewer.mapViewer.put(child,ref);
                    disposeOnCtrlW(0L, child, getViewersSharingView(oneViewer));
                }
                li().addTo("Mw",panel);
                shiftTooltip(panel);
                pcp(ChRunnable.COLUMN_TITLE, ref, panel);
                if ((options&OPEN_NOT_ADD_PANEL)==0) viewerToFront(options|OPEN_ADD_TAB, oneViewer);
                evLstnr(KLI_ALWAYS_ON_TOP).addTo("k",panel);
            } else {
                final Map m=getSharedHashMap(oneViewer);
                if (m.get(ACTION_POPUP_PV)==null) {
                    final ChButton b=new ChButton(ACTION_POPUP_PV).li(li())
                        .t(niceShrtClassNam(oneViewer)+" \u25bc")
                        .cp(KEY_PROTEIN_VIEWER, weakRef(oneViewer))
                        .i(TabItemTipIcon.icn(oneViewer));
                    undock_addOrRm(true,b);
                    m.put(ACTION_POPUP_PV,b);
                }
            }
            maySetFrameTitle(oneViewer);
        }
        return oneViewer;
    }
    /* <<<  open3dViewer <<< */
    /* ---------------------------------------- */
    /* >>> get or open new >>> */
    static ProteinViewer[] getViewersOfProteinAsk(Protein p, Class clazz) {
        if (p==null) return ProteinViewer.NONE;
        final ProteinViewer vv0[]=p.getProteinViewers();
        ProteinViewer vv[]=vv0;
        for(int i=sze(vv);--i>=0;) {
            if (clazz==null && vv[i] instanceof Protein3d.PView  || !isAssignblFrm(clazz,vv[i])) {
                if (vv==vv0) vv=vv0.clone();
                vv[i]=null;
            }
        }
        vv=rmNullA(vv,ProteinViewer.class);
        if (vv.length==0) {
            final ProteinViewer v=open3dViewer(OPEN_ASK|OPEN_SEND_ALL_ANNOTATIONS, new Protein[]{p},clazz);
            if (v!=null) vv=new ProteinViewer[]{v};
        }
        return vv;
    }
    /* <<< get or open new <<< */
    /* ---------------------------------------- */
    /* >>> Disposable >>> */
    private static Map _mPV;
    static Map<String,Object> mapProteinViewers() {
        if (_mPV==null) _mPV=new ChMap(0L,String.class, java.lang.ref.Reference.class,9);
        return _mPV;
    }
    static void disposeV(ProteinViewer...vv) {
        final String key="V3D$$dis";
        for(ProteinViewer pv : vv) {
            final Protein p=p(pv);
            if (pv==null || gcp(key,pv)!=null) continue;
            pcp(key,"", pv);
            assertTrue(AbstractProxy.getProxy(pv,ProteinViewer.class)==null);
            final Object panel=viewerCanvas(pv), sameView[]=getViewersSharingView(pv);
            if (sameView.length==0 || (sameView.length==1 && sameView[0]==pv))  rmFromParent(panel);
            for(Object surface : oo(pv.getProperty(GET_SURFACEOBJECTS))) {
                sendCommand(0L, COMMANDobject_delete+" "+surface, pv);
            }
            pv.dispose();
            if (p!=null) p.removeProteinViewer(pv);
            for(Map.Entry<String,Object> e : entryArry(mapProteinViewers())) {
                if (deref(e.getValue())==pv) mapProteinViewers().remove(e.getKey());
            }
            maySetFrameTitle(pv);
            final ProteinViewer vvSame[]=getViewersSharingView(pv);
            if (sze(vvSame)==0 || vvSame[0]==pv) {
                undock_addOrRm(false, getSharedHashMap(pv).get(ACTION_POPUP_PV));
            }
        }
        StrapAlign.setMenuBar(null,0,0);
        setPV(PV_FOCUSED,null);
    }
    /* <<< Disposable  <<< */
    /* ---------------------------------------- */
    /* >>> ChRunnable >>> */
    public Object run(String id, Object arg) {
        final Object argv[]=arg instanceof Object[] ? (Object[]) arg : null;
        if (id=="SAVE_TB") {
            final int type=atoi(arg);
            final StringBuffer sb=new StringBuffer(999);
            for(ChTextField tf : toArry(TEXT_TF[type], ChTextField.class)) sb.append(tf).append('\n');
            wrte(fileTB(type), sb);
        }
        if (id=="viewerToFront") viewerToFront(atol(argv[1]), (ProteinViewer)argv[0] );
        if (id==ChScriptPanel.RUN_SCRIPT) {
            sendCommand(0L, toStrg(argv[1]), getPV(PV_FOCUSED));
        }
        return null;
    }
    private static ChFrame _fScript;
    private static ChScriptPanel _scriptPanel;
    static void showScriptPanel() {
        if (_scriptPanel==null) {
            final Object
                wc[]={
                allCommands(),
                dirWorking(),
                StrapAlign.getInstance().vProteins,
                //splitTokns(0L, "color red orange brown yellow green blue cyan magenta purple white lightgrey grey darkgrey black "),
            };
            _scriptPanel=new ChScriptPanel(0L, "3D view", wc, null, V3dUtils.instance());
        }
        _fScript=_scriptPanel.showFrame("3D-Script-panel ");
    }
    /* <<< ChRunnable <<< */
    /* ---------------------------------------- */
    /* >>> Style Menu >>> */
    private static Map<String,ChButton> mapBC=new HashMap();
    static ChButton butColor(long options, Color color, String command, ProteinViewer pv) {
        ChButton bc=pv==null?mapBC.get(command) : null;
        if (bc==null) {
            bc= new ButColor(options,color,li()).cp(KEY_PV_CMD, command);
            if (pv==null) mapBC.put(command,bc);
        }
        return bc;
    }
    /* <<< Style menu <<< */
    /* ---------------------------------------- */
    /* >>> MenuBar >>> */
    private static ChLabel labIcon;
    private static void viewMenuBar(ProteinViewer pv) {
        final Object panel=viewerCanvas(pv);
        final boolean isNative=gcp(KEY_WHICH_MENU, panel)==BUT_NATIV_MENUS;
        final ChFrame f=deref(parentWndw(panel),ChFrame.class), fMain=StrapAlign.getFrame();
        if (f==null) return;
        if (gcp(KEY_ADD_TO_NATIVE_MENU, pv)==null) pcp(KEY_ADD_TO_NATIVE_MENU, jMenu(0L, pvp(pv).m('b'),""), panel);
        final JComponent[] mm=
            isNative ? getNativeMenus(pv) :
            (f==fMain ? pvp() : pvp(pv)).getGenericMenus(false);
        final int fg=isNative?0xFF00 : 0xFFffFF;
        if (fMain==f) {
            if (labIcon==null) labIcon=new ChLabel("");
            labIcon.i(TabItemTipIcon.imageIcon(pv));
            pcp(KEY_BEFORE_MENU_ITEM, labIcon, get(0,mm));
            StrapAlign.setMenuBar(mm, fg,0);
            setPV(PV_MENUBAR,pv);
        } else {
            JMenuBar mb=f.getJMenuBar();
            if (mb==null) {
                f.setJMenuBar(mb=new ChJMenuBar());
                pcp(KEY_BACKGROUND,C(0), mb);
            }
            if (!arraysEqul(JMenu.class, mm,mb.getComponents())) {
                mb.removeAll();
                toMenuBar(0L,mm,mb);
                mm[0].revalidate();
            }
            setFgBgMb(mb,fg,0);
            ChDelay.revalidate(mb,333);
        }
    }
    static ProteinViewerPopup pvp() { return instance().contextObjects.proteinViewerMenu(); }
    private static ProteinViewerPopup pvp(ProteinViewer pv) {
        if (pv==null) return null;
        ProteinViewerPopup pvp=null;
        final ProteinViewer vv[]=pv.getViewersSharingView();
        for(int i=sze(vv); pvp==null && --i>=-1;) {
            pvp=gcp(KEY_GET_MENUBAR, i<0?pv:vv[i], ProteinViewerPopup.class);
        }
        if (pvp==null) pvp=new ContextObjects(true, new Object[]{weakRef(pv)}).proteinViewerMenu();
        for(int i=sze(vv); --i>=-2;) pcp(KEY_GET_MENUBAR, pvp, i<-1?viewerCanvas(pv): i<0?pv : vv[i]);
        return pvp;
    }
    /* <<< MenuBar <<< */
    /* ---------------------------------------- */
    /* >>> Set Title >>> */
    private static void maySetFrameTitle(ProteinViewer v) {
        final JFrame f=deref(parentWndw( viewerCanvas(v)), JFrame.class);
        if (f!=null && f!=StrapAlign.getFrame()) setTitl(titleFor3dViewers(getProteinsInView(v)), f);
    }
    private static void maySetTitleStyleMenu(ProteinViewer pv, boolean setFocused) {
        final Protein p=p(pv);
        if (p==null) return;
        final boolean isP3d=pv instanceof Protein3d.PView;
        final String title;
        if (isP3d) title="";
        else {
            Selection3D atoms[]=getSelA(pv);
            if (atoms==null) atoms=getSelA(pv);
            final String pn=delSfx(".pdb",delSfx(".ent",toStrg(p)));
            title= atoms==null ? "Style" : "Style "+pn+" "+partDotDotDot(selectionToText(GENERIC, atoms), sze(pn)+20);
        }
        setEnabld(isP3d, pvp(pv).b(BUT_DEFAULT_VIEWER));
        setEnabld(isP3d, pvp().b(BUT_DEFAULT_VIEWER));
        for(int i=sze(vStyleButton); --i>=0;) {
            final AbstractButton b=get(i,vStyleButton,AbstractButton.class);
            if (b==null) vStyleButton.remove(i);
            else {
                setEnabld(!isP3d, b);
                setTxt(title,b);
            }
        }
        if (pv!=null && setFocused) {
            final ProteinViewer pv0=getPV(PV_FOCUSED);
            if (pv0!=null && getSharedHashMap(pv0)==getSharedHashMap(pv)) setPV(PV_FOCUSED,pv);
        }
        setTitl(p+" @ "+niceShrtClassNam(pv), _fScript);
    }
    /* <<< Set Title <<< */
    /* ---------------------------------------- */
    /* >>> Picked atoms >>> */
    private final static JComponent[] getComp=new JComponent[3];
    final static int GET_ANIM_SPEED=1;
    static JComponent getJComp(int id) {
        JComponent c=getComp[id];
        if (c==null) {
            if (id==GET_ANIM_SPEED) c=new ChTextField("  6.0");
            getComp[id]=c;
        }
        return c;
    }
    public final static int PV_FOCUSED=0, PV_MENUBAR=1,    PV_FOCUSED_NOT_MB=1000;
    static ProteinViewer getPV(int type) {
        final ProteinViewer pv=ONE_VIEWER[type];
        return isActive(pv) ? pv : null;
    }
    static void setPV(int type0, ProteinViewer v0) {
        final int type=type0==PV_FOCUSED_NOT_MB ? PV_FOCUSED : type0;
        final ProteinViewer v=isActive(v0)?v0:null;
        final boolean changed=ONE_VIEWER[type]!=v;
        ONE_VIEWER[type]=v;
        if (type==PV_FOCUSED) {
            enableDisable();
            if (v==null) StrapAlign.setToolpane(null);
            else {
                if (_toolpane!=null) {
                    final Component cc[]=_toolpane.getComponents();
                    for(int i=-1; i<cc.length; i++) pcp(KEY_DROP_TARGET_REDIRECT,weakRef(v),i<0?_toolpane:cc[i]);
                }
                pcp(ChRunnable.COLUMN_TITLE, weakRef(v), viewerCanvas(v));
                maySetTitleStyleMenu(v, false);
            }
            if (type0!=PV_FOCUSED_NOT_MB) viewMenuBar(v);
            if (changed) {
                final Selection3D[] ss=gcp(KEY_SEL_ATOM, v, Selection3D[].class);
                setSelAminos(0L, ss,"",v);
            }
        }
    }
    private static ProteinLabel labPicked;
    private static JComponent _toolpane, labPickedAA;
    private static Selection3D aaPicked;
    public static JComponent getToolpane() {
        if (_toolpane==null) {
            setFG(0xFFffFF, labPicked=new ProteinLabel(ProteinLabel.ALWAYS_CONTEXT_MENU,null));
            setFG(0xFFffFF, labPickedAA=new ChLabel());
            final Object[] bb={null,null,null};
            for(int i=3; --i>=0;) {
                final ChButton b=new ChButton(i==0?"SET_C":i==1?"ADD_S":"RM_S", ChButton.DISABLED)
                    .t(i==0?"Set alignment cursor":i==1?"Add to selection":"Remove from selection");
                vEnableDisable.add(bb[i]=b);
                li().addTo("M", setUndockble(b.li(li())));
            }
            final Dimension d=StrapView.getToolbars().getPreferredSize();
            _toolpane=pnl(KOPT_NOT_INDICATE_TOO_SMALL, d.height>1 ? d : null, C(0), labPicked, labPickedAA, bb[0], bb[1], bb[2]);
            StrapAlign.newDropTarget(_toolpane,true);
        }
        return _toolpane;
    }
  private static int pickCount;
    private static void aminoacidPicked(Selection3D at, ProteinViewer pv0) {
        getToolpane();
        final ProteinViewer
            proxy=AbstractProxy.getProxy(pv0, ProteinViewer.class),
            pv=proxy!=null ? proxy : pv0;
        final Protein p=p(pv);
        if (p==null || at==null) return;
        final int
            iA=selectionToI(false, at, p),
            maskCont=atoi(pv.getProperty(GET_AWTMASK_CONT_SELECTION)),
            maskDisc=atoi(pv.getProperty(GET_AWTMASK_DISC_SELECTION));
        final String
            KEY_last="V3DU$$lc",
            cn=niceShrtClassNam(pv),
            KOPT_INFO="V3DU$$ctrlShift";
        Selection3D[] atoms={aaPicked=at};
        final Selection3D lastSel=gcp(KEY_last, pv, Selection3D.class);
        pcp(KEY_last, at, pv);
        if ((maskCont|maskDisc)!=0 && lastSel!=null && at.getChain()==lastSel.getChain()) {
            final int frst=at.getFirst(), last=at.getLast();
            if (maskCont==(awtModi&MASK_KEYMODIFIERS)) {
                final int f=mini(frst, lastSel.getFirst()), t=maxi(last, lastSel.getLast());
                atoms=addOrRemove('+', getSelA(pv), f, t, at.getInsertion(), at.getChain());
            }
            if (maskDisc==(awtModi&MASK_KEYMODIFIERS)) {
                atoms=addOrRemove('!', getSelA(pv), frst, last, at.getInsertion(), at.getChain());
            }
        }
        if ((maskCont|maskDisc)!=0 && null==gcp(KOPT_INFO,cn) && pickCount++>4) {
            pcp(KOPT_INFO,"", cn);
            final String
                msg="<h2>Continuous and discontinuous selections in "+cn+"</h2>"+
                "For selecting more than one amino acid you need to hold down the following modifier keys while clicking: <br>"+
                "For continuous selection (from last to current amino acid): "+KeyEvent.getKeyModifiersText(maskCont)+"<br>"+
                "For discontinuous selection (toggle selected on/off): "+KeyEvent.getKeyModifiersText(maskDisc)+"<br>";
            ChMsg.savedMsg(msg, cn+"ShiftCtrl3");
        }
        setSelAminos(0L, atoms, SELECTION_PICKED,pv);
        {
            labPicked.setProtein(p);
            setTxt(partDotDotDot(selectionToText(GENERIC,atoms),30), labPickedAA);
            labPickedAA.setOpaque(true);
            setBG(0xFFffFF,labPickedAA);
            setFG(0,labPickedAA);
            ChDelay.fgBg(true,labPickedAA,C(0),444);
            ChDelay.fgBg(false,labPickedAA,C(0xFFffFF),444);
            StrapAlign.startAnimatePosition(0L, p, iA);
            setPV(PV_FOCUSED,pv);
            StrapAlign.setToolpane(getToolpane());
            final AWTEvent ev=mousePressedEvent;
            final StrapView view=StrapAlign.getAlignmentPanelWithFocus();
            final Component
                canvas=deref(pv.getProperty(GET_CANVAS), Component.class),
                pCenter=view==null ? null:view.getAlignmentPane();
            if (pCenter!=null && ev!=null && pv.getProperty(GET_CANVAS)==ev.getSource()) {
                final int
                    row=view.findRow(p),
                    col=((StrapProtein)p).getResidueColumn(iA);
                final float speed=(float)atof(toStrg(getJComp(GET_ANIM_SPEED)));
                if (row>=0 && col>=0 && speed>1E-8) {
                    final Point
                        pSrc= new Point(x(ev),y(ev)),
                        pDest=new Point((view.col2x(col)+view.col2x(col+1))/2,  (view.row2y(row)+view.row2y(row+1))/2);
                    final String s=new ByteArray(9).appendSomeBytes(p.getResidueName32(iA),4).newString();
                    new Thread(new ChSprite(ChSprite.SPEED_SINUS, canvas, pSrc, view.getAlignmentPane(), pDest, s, speed)).start();
                }
            }
        }
    }
    /* <<< Picked atoms <<< */
    /* ---------------------------------------- */
    /* >>> AWTEvent >>> */
    private static int awtModi;
    final static List vEnableDisable=new ArrayList();
    public static void enableDisable() {
        ProteinViewer pv=getPV(PV_FOCUSED);
        if (!isActive(pv)) pv=null;
        final Object refPV=weakRef(pv);
        for(int i=sze(vEnableDisable); --i>=0;) {
            final Object c=get(i,vEnableDisable);
            final String cmd=gcps(KEY_ACTION_COMMAND,c);
            if (cntains(cmd, Protein3dUtils.allCommands())) setEnabld(supports((String)cmd,pv), pv);
            final Object sze=
                cmd=="UNDEL_O" ? V3dListModel.list(V3dListModel.DELETED_OBJECTS, pv, false) :
                cmd=="UPDATE"|| cmd==COMMANDchange_object_color || cmd=="CH_COLOR" || cmd==COMMANDobject_delete ?
                V3dListModel.jList(V3dListModel.SURFACE_OBJECTS, false,0) :
                cmd=="UNDEL_OBJ"  ? V3dListModel.jList(V3dListModel.DELETED_OBJECTS, true,0) :
                "";
            if (sze!="") setEnabld(sze(sze instanceof ChJList ? ((ChJList)sze).getSelectedValues() : sze)>0,c);
            if (cmd=="SET_C" || cmd=="ADD_S" || cmd=="RM_S") {
                final int iA=selectionToI(false, aaPicked, p(getPV(PV_FOCUSED)));
                setEnabld(iA<0?false : cmd=="ADD_S"||(cmd=="RM_S")==ResSelUtils.isSelectedAA(iA, residueSelectionsF(false)), c);
            }
            if (c instanceof ChJList) pcp(KEY_PROTEIN_VIEWER, refPV, c);
            if (c instanceof ChLabel) {
                if (getTxt(c)==INFO_HETERO_NOT_SELECTABLE) setEnabld(pv instanceof Protein3d.PView,c);
            }
            revalAndRepaintC(c);
        }
    }
    private static boolean checkResiduesSelected(ProteinViewer pv, boolean setTrue) {
        final Protein p=p(pv);
        if (p==null) return false;
        Selection3D atoms[]=getSelA(pv);
        if (sze(atoms)>0) return true;
        final String MSG_NO_RES=
            "Warning: No residue selected.<br>"+
            "Residues can be selected by clicking into the 3D-model or "+
            "by selecting positions in the alignment pane.<br><br>"+
            "Continue to process all amino acids of the protein?";
        if (ChMsg.yesNo(MSG_NO_RES)) {
            if (setTrue) setSelAminos(0L, toSelectionI(0,MAX_INT,p), "",pv);
            return true;
        }
        return false;
    }

    private final static AbstractResidueSelection residueSelectionsF(boolean create) {
        final ProteinViewer pvF=getPV(PV_FOCUSED);
        final String name=pvF==null ? null : pvF instanceof Protein3d.PView ? "3D-Backbone" :  niceShrtClassNam(pvF);
        final Protein pF=p(pvF);
        if (name==null) return null;
        AbstractResidueSelection s=pF==null?null:findWithName(name, pF.getResidueSelections(),AbstractResidueSelection.class);
        if (s==null && create) {
            s=new AbstractResidueSelection();
            s.setName(name);
            s.setProtein(pF);
            pF.addResidueSelection(s);
            s.setVisibleWhere(0xffffffff);
        }
        return s;
    }

    private static AWTEvent mousePressedEvent;
    private static long whenMouseViewer;
    private static ProteinViewer getViewerForP(Protein p) {
        if (p==null) return null;
        final ProteinViewer pvF=getPV(PV_FOCUSED);
        if (p(pvF)==p) return pvF;
        for(ProteinViewer pv : getViewersSharingView(pvF)) {
            if (p(pv)==p) return pv;
        }
        return get(0, p.getProteinViewers(), ProteinViewer.class);
    }
    public AWTEvent processEvent(JComponent component, AWTEvent ev) {
        final int id=ev.getID(), kcode=keyCode(ev), modi=modifiers(ev);
        final Object q=ev.getSource();
        if (q==null) return ev;
        final String cmd=actionCommand(ev);
        final boolean ctrl=0!=(modi&CTRL_MASK), shift=0!=(modi&SHIFT_MASK), isContext=isPopupTrigger(ev);
        ChFrame.setAotEv(ev);
        mayClosePopupMenu(ev);
        if (cmd==BUT_COLOR) openColorDialog();
        final ProteinViewer pvF=getPV(PV_FOCUSED);
        if (q instanceof ChJList && id==MOUSE_CLICKED) {
            final List<Selection3D> v=new ArrayList();
            Protein aProtein=null;
            for(Object o : getSelValues(q)) {
                if (o instanceof Protein) aProtein=(Protein)o;
                final HeteroCompound het=deref(o,HeteroCompound.class);
                if (het!=null) {
                    final char chain= het.getCompoundChain();
                    if (het.isNucleotideChain()) {
                        final int rn[]=het.getNucleotideNumber();
                        if (sze(rn)>0) v.add(new Selection3D(0L, rn[0], rn[rn.length-1], (char)0,  chain, null));
                    } else {
                        final int rn=het.getCompoundNumber();
                        v.add(new Selection3D(0L, rn, rn, (char)0,  chain, null));
                    }
                }
                final ResidueSelection s=deref(o,ResidueSelection.class);
                if (s!=null) {
                    putln("s="+s+"  "+boolToTxt(s.getSelectedAminoacids()));
                    for(Selection3D s3 : toSelectionS(s)) {
                        putln("s3="+selectionToText(GENERIC, new Selection3D[]{s3}));
                        v.add(s3);
                    }
                }
            }
            if (aProtein!=null) setPV(PV_FOCUSED, getViewerForP(aProtein));
            else {
                setPV(PV_FOCUSED, pvF);
                setSelAminos(0L, toArry(v,Selection3D.class), "", pvF);
            }
        }
        final Protein pF=p(pvF);
        if (cmd==ACTION_SELECTION_CHANGED) enableDisable();
        if (id==MOUSE_PRESSED) mousePressedEvent=ev;
        if (pF!=null && isActive(pvF)) {
            if (id==KEY_PRESSED) {
                if (kcode==VK_W && ctrl) disposeV(getViewersSharingView(pvF));
                if (kcode==VK_V && shift && ctrl) {
                    if (atoi(gcp(KEY_SEL_DIALOG_TYPE,q))==SELECT_AMINO && q instanceof ChTextField) {
                        if (kcode==VK_W && ctrl) disposeV(getViewersSharingView(pvF));
                        final ProteinViewer pv=pvF;
                        Selection3D[] sel=getSelA(pv);
                        if (sel!=null) {
                            changeAtoms(true, sel=sel.clone(), "");
                            String ins=" "+selectionToText(GENERIC, sel)+" ", pn=toStrg(pv.getProtein());
                            if (strstr(STRSTR_w, pn,toStrg(q))<0) ins=pn+ins;
                            ChTextComponents.insertAtCaret(0L, ins, (JComponent)q);
                        }
                    }
                    return null;
                }
            }
            final Color color=cmd==ACTION_COLOR_CHANGED &&  q instanceof ButColor ? ((ButColor)q).getColor() :null;
            final String hex=toHex(color,true);
            if (color!=null) {
                if (q!=_color_butColor) {
                    final String script=gcps(KEY_PV_CMD,q);
                    final String lines=rplcToStrg("COLOR", hex, script);
                    if (script!=null && (script.startsWith(COMMANDchange_object_color) || checkResiduesSelected(pvF, true))) {
                        boolean success=false;
                        for(String line : splitLnes(0L, lines)) success|=sendCommand(0L, line, pvF);
                        if (success) ChDelay.repaint(viewerCanvas(pvF),99);
                    }
                } else {
                    final String cmdColor=COMMANDcolor+" #"+hex;
                    final ChJList jlProteins=V3dListModel.jList(V3dListModel.PROTEINS, false,0);
                    final Object entire[]=oo(prefHeight(jlProteins)>0 ? jlProteins.getSelectedValues() : null);
                    for(Object p : entire) {
                        final ProteinViewer pv=getViewerForProtein((Protein)p, pvF);
                        if (pv==null) continue;
                        setSelAminos(0L, toSelectionBB((boolean[])null,(Protein)p), "", pv);
                        sendCommand(0L, cmdColor, pv);
                        if (0!=(PROPERTY_RIBBON_COLOR_NO_CHANGE&flags(pv)) && isSelctd(cbUpdateSurfaceColors)) updateSurfacesP(pv,true);
                        pcp(KEY_COLOR,color,pv); /* For ChRenderer */
                        pcp(KEY_COLOR,color,p);
                    }
                    if (entire.length==0) {
                        if (checkResiduesSelected(pvF,true)) {
                            sendCommand(0L, cmdColor, pvF);
                            if (isSelctd(cbUpdateSurfaceColors)) updateSurfacesP(pvF, false);
                        }
                    }
                    for(Object o : entire) {
                        final Protein p=deref(o, Protein.class);
                        final ProteinViewer pv=getViewerForProtein(p, pvF);
                        if (pv!=null) {
                            setSelAminos(0L, toSelectionT(" $ALL ",p), "", pv);
                            sendCommand(0L, cmdColor, pv);
                            if (isSelctd(cbUpdateSurfaceColors)) updateSurfacesP(pv, false);
                        }
                    }
                    return null;
                }
            }
            /* Surface Objects */
            final ChJList jlSurf=V3dListModel.jList(V3dListModel.SURFACE_OBJECTS, false,0);
            final Object[] surfOO=jlSurf==null?null:jlSurf.getSelectedValues();
            Object toSelection= q==jlSurf && id==MOUSE_PRESSED ? get(0,surfOO) : null;
            if (sze(surfOO)>0) {
                if (gcp(KEY_ACTION_COMMAND,q)=="CH_COLOR" && color!=null) {
                    for(Object o : surfOO) sendCommand(0L, COMMANDchange_object_color+" "+o+" #"+hex, pvF);
                }
                if (cmd==COMMANDobject_delete || cmd=="UPDATE" || cmd==COMMANDchange_object_color) {
                    for(Object obj : surfOO) {
                        final String name=toStrg(obj);
                        if (cmd==COMMANDobject_delete) {
                            sendCommand(0L, COMMANDobject_delete+" "+name, pvF);
                            adUniq(name,   V3dListModel.list(V3dListModel.DELETED_OBJECTS, pvF, true));
                            enableDisable();
                        }
                        if (cmd==COMMANDchange_object_color) updateSurface(name, pvF, true, false);
                        if (cmd=="UPDATE") updateSurface(name, pvF, false, false);
                        toSelection=name;
                    }
                    revalAndRepaintC(jlSurf);
                }
                if (sze(toSelection)>0) updateSurface(toStrg(toSelection), pvF, false, true);
            }
            if (cmd=="UNDEL_O")  undelObj(true, pvF);
            if (cmd=="UNDEL_OBJ")undelObj(false,pvF);
            if (cmd==BUT_SURF_OBJ) ChFrame.frame("", pnlSurfaceObjects(), ChFrame.PACK).shw(ChFrame.AT_CLICK);
        }/* pvF */
        /* Atom selection */
        if (cmd=="DIA_ATOM") openDialogSelect(SELECT_ATOMS_G);
        if (cmd==BUT_LABEL)    openDialogSelect(SELECT_LABEL);
        if (cmd=="TF" || gcp(KEYS_DIALOG_SEL,q)!=null && cmd==ACTION_ENTER) {
            final Object tf=gcp(KEYS_DIALOG_SEL,q);
            final int t=atoi(gcp(KEY_SEL_DIALOG_TYPE,tf));
            final String txt=toStrg(tf);
            if (t==SELECT_ATOMS_G || t==SELECT_ATOMS) setSelAtom(SELECT_ATOMS_G==t, txt, pvF);
            if (t==SELECT_AMINO) {
                ProteinViewer pvS=pvF;
                final String tt[]=splitTokns(0L, txt);
                for(int i=tt.length; --i>=0;) {
                    Protein p0=StrapProtein.getProteinWithName(tt[i], StrapAlign.getInstance().getProteins());
                    final ProteinViewer pv0=getViewerForProtein(p0,pvF);
                    if (pv0!=null) {
                        setPV(PV_FOCUSED, pvS=pv0);
                        tt[i]=null;
                    }
                }
                if (pvS!=null) setSelAminos(0L, toSelectionT(joinAsStrg(tt," "), pvS.getProtein()), "Saved", pvS);
            }
            if (t==SELECT_LABEL) sendCommand(0L, COMMANDlabel+" "+(sze(txt)==0?"off":txt), pvF);
        }
        if (q instanceof ChTextField && gcp(KEYS_DIALOG_SEL,q)!=null) {
            final JComponent parent= (JComponent) ((JComponent)((JComponent)q).getParent()).getParent();
            newRow(atoi(gcp(KEY_SEL_DIALOG_TYPE,q)), parent);
        }
        if (cmd==ACTION_POPUP_PV) {
            final ProteinViewer pv=gcp(KEY_PROTEIN_VIEWER,q, ProteinViewer.class);
            if (pv!=null) shwPopupMenu(pvp(pv).menu());
        }
        awtModi=gcp(KEY_EVT_MODI,q)!=null? atoi(gcp(KEY_EVT_MODI,q)):modi;
        if (id==MOUSE_ENTERED && q instanceof ChJList && gcp(ProteinViewer.class,q)!=null) {
            final ChJList v=(ChJList)q;
            final Protein pp[]=getProteinsInView(gcp(ProteinViewer.class,q,ProteinViewer.class));
            if (sze(v)!=pp.length || !cntainsAll(v,pp)) {
                v.setData(pp);
                v.revalidate();
            }
        }
        if ( (id==MOUSE_PRESSED || id==MOUSE_EXITED || id==MOUSE_ENTERED || id==MOUSE_MOVED) && q!=null) {
            final ProteinViewer pvRep=gcp(KEY_PROTEIN_VIEWER, q, ProteinViewer.class);
            final ProteinViewer pv=cntains(pvRep, getViewersSharingView(pvF)) ? pvF : pvRep;
            if (isActive(pv)) {
                whenMouseViewer=System.currentTimeMillis();
                if (id==MOUSE_PRESSED) {
                    setPV( pv.getProperty(GET_CANVAS)==q ? PV_FOCUSED : PV_FOCUSED_NOT_MB, pv);
                    StrapAlign.setToolpane(getToolpane());
                    return ev;
                }
            }
        }
        {
            final int iA=selectionToI(false, aaPicked, pF);
            if ((cmd=="ADD_S" || cmd=="RM_S" || cmd=="SET_C") && iA>=0) {
                final AbstractResidueSelection s=residueSelectionsF(cmd=="ADD_S");
                if (q instanceof ChButton && isContext) StrapAlign.showContextMenu('S',oo(s));
                if (cmd=="SET_C") StrapAlign.setCursorToResidue(pF, iA);
                else if (s!=null) {
                    ResSelUtils.setSelectedAA(cmd=="ADD_S", iA, s, true);
                    final boolean bb[]=s.getSelectedAminoacids();
                    if (firstTrue(bb)<0) pF.removeResidueSelection(s);
                    setSelAminos(0L, toSelectionS(s), "", pvF);
                    setPV(PV_FOCUSED,pvF);
                    pF.removeResidueSelection(findWithName(ResidueSelection.CURSOR, pF.getResidueSelections(),AbstractResidueSelection.class));
                    StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,10);
                }
                StrapAlign.startAnimatePosition(StrapAlign.ANIM_SET_CURS_SCROLL_TO_VISIBLE, pF, iA);
                enableDisable();
            }
        }
        if (isActive(pvF)) {
            final int n=scaleUpDown(ev);
            if (n!=0) sendCommand(0L, COMMANDzoom+" "+(100-n*10), pvF);
            else {
                final int rotZ=3*wheelRotation(ev);
                if (rotZ!=0)sendCommand(0L, COMMANDrotate+" z"+" "+rotZ, pvF);
            }
        }
        return ev;
    }
    /* <<< AWTEvent <<< */
    /* ---------------------------------------- */
    /* >>> StrapEvent >>> */
    static boolean handleEvent(StrapAlign align, StrapEvent ev) {
        final int id=ev.getType();
        final Object q=ev.getSource(), para[]=ev.getParameters();
        final Object oo[]=ev.getParameters();
        //final StrapAlign align=StrapAlign.getInstance();
        if (id==StrapEvent.PROTEIN_VIEWER_CLOSED) {

            disposeV(deref(q, ProteinViewer.class));
        }
        if (id==StrapEvent.PROTEIN_VIEWER_PICKED) aminoacidPicked((Selection3D)oo[0] , (ProteinViewer)q);
        if (id==StrapEvent.BACKGROUND_CHANGED) {
            final String cmd=COMMANDbackground+" #"+(isWhiteBG()?"FFffFF":"000000");
            for(StrapProtein p : align.getProteins()) {
                for(ProteinViewer pv : p.getProteinViewers()) sendCommand(0L, cmd, pv);
            }
        }
        if (id==StrapEvent.PROTEIN_VIEWER_SURFACES_CHANGED) {
            revalAndRepaintC(V3dListModel.jList(V3dListModel.SURFACE_OBJECTS, false,0));
        }
        if (id==StrapEvent.PROTEIN_3D_MOVED) {
            for(int i=sze(para); --i>=0;) {
                final Protein p=deref(para[i],Protein.class);
                if (p==null) continue;
                Matrix3D m3d=p.getRotationAndTranslation();
                if (m3d!=null && m3d.isUnit()) m3d=null;
                for(ProteinViewer v : p.getProteinViewers()) {
                    if (gcp(KOPT_IS_PREVIEW,v)==null) transform3D(m3d,v);
                }
            }
        }
        if (id==StrapEvent.CURSOR_CHANGED_PROTEIN || id==StrapEvent.CURSOR_MOVED_WITHIN_PROTEIN || id==StrapEvent.CURSOR_MOVED_WITHIN_PROTEIN_DELAYED || id==StrapEvent.CURSOR_CLICKED) {
            if (isSelctd(pvp().b(TOG_CURSOR_SELECT))) {
                final Protein pCursor=align.getProteinAtCursor();
                final int iA=align.getIndexOfAminoAcidAtCursor();
                final ProteinViewer[] vvFocused=getViewersSharingView(getPV(PV_FOCUSED));
                boolean setMb=false;
                if (pCursor!=null && iA>=0) {
                    for(ProteinViewer pv : pCursor.getProteinViewers()) {
                        final boolean delay= 0!=(flags(pv) & PROPERTY_SEQUENCE_CURSOR_DELAYED);
                        if (id==StrapEvent.CURSOR_MOVED_WITHIN_PROTEIN_DELAYED && !delay) continue;
                        if (id==StrapEvent.CURSOR_MOVED_WITHIN_PROTEIN && delay) continue;
                        setSelAminosI(0L, iA, SELECTION_CURSOR, pv);
                        if (!setMb && !StrapAlign.isAlignmentMenuBar() && pv!=getPV(PV_FOCUSED) ) {
                            maySetTitleStyleMenu(pv, false);
                            setMb=true;
                        }
                    }
                }
            }
        }
        if (id==StrapEvent.RESIDUE_SELECTION_CHANGED || id==StrapEvent.RESIDUE_SELECTION_ADDED || id==StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR ||
            (id&StrapEvent.FLAG_ATOM_COORDINATES_CHANGED)!=0 || id==StrapEvent.HETERO_OR_NUCLEOTIDE_STRUCTURE_ADDED_OR_REMOVED /*|| id==StrapEvent.BIOMOLECULES_SPECIFIED*/) {
            for(StrapProtein p : align.getProteins()) {
                for(ProteinViewer pv : p.getProteinViewers()) {
                    p.getSelectedAminoacids();
                    /* TODO nur die geaenderten */
                    if (pv instanceof StrapListener) ((StrapListener)pv).handleEvent(ev);
                    if (pv instanceof Protein3d.PView) {
                        final Protein3d p3d=(Protein3d)viewerCanvas(pv);
                        if (id==StrapEvent.RESIDUE_SELECTION_CHANGED)  { p3d.enableDisable(); ChDelay.repaint(p3d,200);}
                        if (id==StrapEvent.RESIDUE_SELECTION_ADDED || id==StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR ||
                            (id&StrapEvent.FLAG_ATOM_COORDINATES_CHANGED)!=0 || id==StrapEvent.HETERO_OR_NUCLEOTIDE_STRUCTURE_ADDED_OR_REMOVED)
                            ChDelay.repaint(p3d,11);
                        //if (id==StrapEvent.BIOMOLECULES_SPECIFIED)  p3d.run(Protein3d.KEY_BIO_MOLECULES,ev.getParameters());
                    }
                }
            }
            for(int vers01=2; --vers01>=0;) {
                for(int type : V3dListModel.TYPES_SEL) {
                    revalAndRepaintC(V3dListModel.jList(type,false,vers01));
                }
            }

        }
        return true;
    }
    /* <<< StrapEvent <<< */
    /* ---------------------------------------- */
    /* >>> Selection >>> */
    private final static Object KEY_SEL_ATOM=new Object(), KEY_SEL_ID=new Object();
    public static Selection3D[] getSelA(ProteinViewer pv) {
        Selection3D[] ss=gcp(KEY_SEL_ATOM, pv, Selection3D[].class);
        if (ss==null) ss=gcp(KEY_SEL_ATOM, getSharedHashMap(pv), Selection3D[].class);
        return ss==null?Selection3D.NONE:ss;
    }
    private static void setSelAminosI(long options, int iA, String id, ProteinViewer pv) {
        final Protein p=p(pv);
        if (p==null) return;
        if (pv instanceof Protein3d.PView && id==ProteinViewer.SELECTION_CURSOR) {
            final Protein3d p3d=(Protein3d)pv.getProperty(GET_CANVAS);
            if (p3d!=null) p3d.highlightAminoAcid(p,iA);
        } else setSelAminos(options, toSelectionI(iA,iA, p), id,  pv);
    }
    static void setSelAminos(long options0, Selection3D atoms[], String id, ProteinViewer pv) {
        final Protein p=p(pv);
        if (p==null || !supports(COMMANDselect,pv)) return;
        if (pv instanceof Protein3d.PView) ((Protein3d.PView)pv).setSelectedResidues(selectionToBoolean(atoms, p));
        final long time=System.currentTimeMillis();
        final long options=options0 | ( (time-whenMouseViewer>999) ? SET_SELECTED_RESIDUES_HIGHLIGHT :0);
        if (sze(id)>0 && supports(COMMANDselection_name, pv)) pv.interpret(0L, COMMANDselection_name+" "+id);
        pv.interpret(options, COMMANDselect+" "+selectionToText(GENERIC, atoms));
        pcp(KEY_SEL_ID, id,  pv);
        pcp(KEY_SEL_ATOM, atoms,  pv);
        maySetTitleStyleMenu(pv, false);
    }
    private static void setSelAtom(boolean isGeneric, String atomType, ProteinViewer pv) {
        final Protein p=p(pv);
        if (p==null) return;
        final Selection3D[] atoms=getSelA(pv);
        if (countNotNull(atoms)==0) return;
        changeAtoms(isGeneric, atoms,atomType);
        setSelAminos(0L, atoms, gcps(KEY_SEL_ID,pv), pv);
    }
    static boolean[] selectedAAto3D_single(long options, ResidueSelection ss[], ProteinViewer viewer, String selName) {
        final Protein p=p(viewer);
        if (p==null) return null;
        final boolean selected[]=ResSelUtils.mapPositions(ss,p);
        if (firstTrue(selected)<0) return null;
        final float[] xyz=p.getResidueCalpha();
        final int countTrue=countTrue(selected);
        int count=0;
        for(int i=mini(sze(xyz)/3, sze(selected)); --i>=0;) {
            if (selected[i] && !Float.isNaN(xyz[i*3])) count++;
        }
        if (countTrue>0) {
            final ProteinViewer vv[]=viewer!=null ? new ProteinViewer[]{viewer} : getViewersOfProteinAsk(p, (Class)null);
            if (vv.length>0) {
                boolean needsName=false;
                for(ProteinViewer v : vv) needsName|= 0!=(PROPERTY_NAMED_SELECTIONS&flags(v));
                String name=null;
                if (needsName) {
                    final Object
                        msg=vBhB("Protein "+p+"<br>Selected aminocids: ",
                                 scrllpn(0L, monospc(Protein.selectedPositionsToText(selected,0,p)), dim(1,3*EX)),
                                 "What should the selection be named in the protein viewer?",
                                 (count==0 ? "<br>Problem: None of the selected amino acids has coordinates.<br>You may not see the selection.":null)
                                 );
                    name=
                        selName!=null ? selName :
                        0!=(options&RSto3D_GENERIC_NAME) ? toStrg(boolToTxt(selected,1,"__","_",null)) :
                        0!=(PROPERTY_NAMED_SELECTIONS&flags(viewer)|(options&RSto3D_ASK_NAME)) ? ChMsg.input(0L,msg,20,"my_selection") :
                        "Unnamed";
                }
                ProteinViewer pvSel=null;
                if (!needsName || sze(name)>=0) {
                    for(ProteinViewer v : vv) {
                        setSelAminos(0L, toSelectionBB(selected, p), name,v);
                        pvSel=v;
                    }
                }
                if (pvSel!=null) setPV(PV_FOCUSED,pvSel);
            }
        }
        return selected;
    }
    public final static long
        RSto3D_COMMANDS=1L<<11, RSto3D_COLOR=1L<<2, RSto3D_AS_ONE_SINGLE=1L<<3,
        RSto3D_SAME_PROTEIN=1L<<4,
        RSto3D_ASK_NAME=1L<<5, RSto3D_REPORT_ERROR=1L<<6, RSto3D_NOT_SAME_WIRE=1L<<7,
        RSto3D_NO_WIRE=1L<<8,  RSto3D_GENERIC_NAME=1L<<9, RSto3D_SPECIFIC_ATOMS=1L<<10,
        RSto3D_SET_FOCUSED_PV=1<<11;
    public static boolean residueSelectionsTo3D(long options, ResidueSelection ss[], ProteinViewer[] vv) {
        int count=0;
        String error=null;
        if (0==(options&RSto3D_AS_ONE_SINGLE)) {
            for(ResidueSelection s:ss) {
                final Protein prot=sp(s);
                if (prot==null) continue;
                final ResidueAnnotation a=deref(s,ResidueAnnotation.class);
                Selection3D atoms[]=null;
                ProteinViewer pvSel=null;
                for(ProteinViewer v : vv!=null ? vv : prot.getProteinViewers()) {
                    final Protein p=p(v);
                    if (0!=(RSto3D_SAME_PROTEIN&options) && prot!=p) continue;
                    final String id=s instanceof ResidueAnnotation ?  ((ResidueAnnotation)s).getID() : new ByteArray(99).filter(LETTR_DIGT,s,0,MAX_INT,'_').newString();
                    final boolean bb[]=ResSelUtils.mapPositions(new ResidueSelection[]{s}, p);
                    if (firstTrue(bb)<0) { error="The selected residues do not map to the protein "+p; continue;}
                    final String pos=a==null?null:a.getValue(ResidueAnnotation.POS);
                    if (atoms==null && pos!=null && pos.indexOf('.')>0) atoms=toSelectionT(true, pos,p);
                    if (atoms==null) atoms=toSelectionBB(bb, p);
                    setSelAminos(0L, atoms, id, v);
                    pvSel=v;
                    if (v instanceof Protein3d.PView) continue;
                    if ( (RSto3D_COMMANDS&options)!=0 && a!=null) {
                        /* Color of a */
                        boolean hasColorCmd=false, hasViewerCommand=false;
                        for(ResidueAnnotation.Entry e : a.getEntries()) {
                            final String k=e.getKey(), val=e.getValue();
                            hasColorCmd= hasColorCmd || k==ResidueAnnotation.VIEW3D && val.startsWith(COMMANDcolor);
                            hasViewerCommand|=k==ResidueAnnotation.VIEW3D;
                        }
                        if (!hasColorCmd && 0!=(RSto3D_COLOR&options) && hasViewerCommand) {
                            final String hex=toHex(colr(s),true);
                            if (hex!=null) {
                                for(int iC=2; --iC>=0;) {
                                    final String cmd=(iC==1?COMMANDcolor : COMMANDsurface_color) + " #"+hex;
                                    sendCommand(INTERPRET_NO_MSG_DIALOGS, cmd, v);
                                }
                            }
                        }
                        /* Command */
                        final String cn=clazzName(v), scn=shrtClasNam(v);
                        for(ResidueAnnotation.Entry e : a.getEntries()) {
                            final String k=e.getKey(), val=e.getValue();
                            if (!e.isEnabled() || sze(val)==0) continue;
                            if (k==COMMANDselect && supports(COMMANDselect,v)) {
                                if (chrAt(0,val)=='.') changeAtoms(true, atoms, val); else atoms=toSelectionT(true, val, p);
                            }
                            if (k==ResidueAnnotation.ATOMS) changeAtoms(true, atoms, val);
                            if (ResidueAnnotation.VIEW3D==k && (supports(val,v) || k.equals(cn) || k.equals(scn))) {
                                setSelAminos(0L, atoms, id, v);
                                sendCommand(INTERPRET_NO_MSG_DIALOGS, SelectionVariables.replace(val,a) , v);
                            }
                        }
                    }
                    count++;
                }
                if (pvSel!=null && 0!=(options&RSto3D_SET_FOCUSED_PV)) setPV(PV_FOCUSED, pvSel);
            }
        }
        if (count!=1) {
            for (ProteinViewer v  : vv!=null ? vv : getViewers(StrapProtein.getProteinsOfObjects(ss),null)) {
                final boolean isP3d=v instanceof Protein3d.PView;
                if (!isActive(v) || isP3d && 0!=(options&RSto3D_NO_WIRE)) continue;
                ResidueSelection ss2[]=ss;
                if (0!=(options&RSto3D_NOT_SAME_WIRE) && isP3d) {
                    for(int i=sze(ss); --i>=0;) {
                        final ResidueSelection s=ss[i];
                        if (sp(s)==sp(v)) {
                            (ss2!=ss ? ss2=ss.clone() : ss2 )[i]=null;
                            if (s instanceof VisibleIn123) {
                                final int
                                    w0=((VisibleIn123)s).getVisibleWhere(),
                                    w1=w0|VisibleIn123.STRUCTURE;
                                if (w0!=w1) {
                                    ((VisibleIn123)s).setVisibleWhere(w1);
                                    StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,99);
                                }
                            }
                        }
                    }
                }
                final String name=0!=(options&RSto3D_AS_ONE_SINGLE)? null:"Unnamed";
                count+=countTrue(selectedAAto3D_single(options, rmNullA(ss2,ResidueSelection.class), v, name));
            }
        }
        if (count==0 && 0!=(options&RSto3D_REPORT_ERROR)) error(error!=null?error:"No residue selection copied to any 3D-viewer.");
        return count>0;
    }
    private static Object msgDnD;
    public static void residueSelectionsTo3D_I(long option, ResidueSelection ss[], ProteinViewer vv[]) {
        AbstractButton cbOne=null, cbCmd=null, cbCol=null;
        boolean hasPvNonWire=false;
        int hasPv=0;
        for(ResidueSelection s : ss) {
            final Protein p=sp(s);
            if (p==null) continue;
            for(ProteinViewer v : vv!=null ? vv : p.getProteinViewers()) {
                final boolean isWire=v instanceof Protein3d.PView;
                if (0!=(option&RSto3D_NO_WIRE) && isWire) continue;
                hasPv++;
                if (v instanceof Protein3d.PView) continue;
                hasPvNonWire=true;
                if (cbCol==null) cbCol=cbox("Use the color of the amino acid selection");
                if (s instanceof ResidueAnnotation) {
                    final String cn=clazzName(v), scn=shrtClasNam(v);
                    for(ResidueAnnotation.Entry e : ((ResidueAnnotation)s).getEntries()) {
                        if (!e.isEnabled()) continue;
                        final String k=e.getKey(), val=e.getValue();
                        if (cbCmd==null && (k==ResidueAnnotation.VIEW3D && supports(val,v) || k.equals(cn) || k.equals(scn)))
                            cbCmd=cbox("Send 3D-Viewer  commands contained in residue annotation");
                    }
                }
            }
        }
        if (hasPv==0)  return;
        cbOne=ss.length>1 && hasPvNonWire ? cbox("Combine all "+ss.length+" amino acid selection  to one single amino acid selection to be passed to the 3D-View") : null;
        if (cbOne!=null && cbCol!=null) radioGrp(0, cbOne,cbCol);
        if (cbOne!=null && cbCmd!=null) radioGrp(0, cbOne,cbCmd);
        if ( !ChMsg.yesNo(vB("Sending "+ss.length+" residue selections to 3D-View",cbOne,cbCmd, cbCol))) return;
        final long opt=option
            | RSto3D_REPORT_ERROR
            | RSto3D_SET_FOCUSED_PV
            | (isSelctd(cbCmd) ? RSto3D_COMMANDS : 0)
            | (isSelctd(cbOne) ? RSto3D_AS_ONE_SINGLE|RSto3D_ASK_NAME : 0)
            | (isSelctd(cbCol) ? RSto3D_COLOR : 0);
        if (!residueSelectionsTo3D(opt,ss, vv)) {
            if (msgDnD==null) {
                msgDnD=vB(
                          "Conveniently, Drag-and-drop can be used to get a residue-selection into a 3D-viewer.",
                          "This works for all Java-based 3D-Viewers like Astex, but not for native ones like Pymol.",
                          " ",
                          "The protein of the residue-selection may be different from the one in the 3D-viewer.",
                          " ",
                          WATCH_MOVIE+MOVIE_Sequence_Features_in_3D
                          );
                ChMsg.msgDialog(0L, msgDnD);
            }
        }
    }
    /* <<< ResidueSelection <<< */
    /* ---------------------------------------- */
    /* >>> Atom Selection >>> */
    final static int SELECT_ATOMS_G=1, SELECT_ATOMS=2, SELECT_AMINO=3, SELECT_LABEL=4;
    private static Object KEY_SEL_DIALOG_TYPE=new Object(), KEYS_DIALOG_SEL[]={null, new Object(),new Object(),new Object(), new Object()};
    private static ChTextField newRow(int t, JComponent parent) {
      for(ChTextField tf : childs(parent, ChTextField.class)) if (sze(toStrg(tf))==0) return tf;
        final ChTextField
            tf=new ChTextField().cols(40,true,false).li(li())
            .cp(KEY_SEL_DIALOG_TYPE, intObjct(t));
        final JComponent
            bClear=tf.tools().newClearButton().t("C").tt("Clear"),
            b=new ChButton("TF").cp(KEYS_DIALOG_SEL, weakRef(tf)).t(t==SELECT_LABEL?"Set label": "Apply").li(li()),
            row=newCNSEW(tf, null,null,b, bClear);
        if (t==SELECT_AMINO) {
            li().addTo("k",tf);
            tf.tools().enableWordCompletion(StrapAlign.getInstance().vProteins)
                .highlightOccurrence(StrapAlign.getInstance(), charClass(SPC), charClass(SPC), HIGHLIGHT_UPDATE_IF_TEXT_CHANGES, ChTextComponents.COLOR_U_GRAY);
        }
        (TEXT_TF[t]==null ? TEXT_TF[t]=new ArrayList() : TEXT_TF[t]).add(tf);
        pcp(KEYS_DIALOG_SEL, weakRef(tf), tf);
        parent.add(row);
        closeOnCtrlW(true,row, CLOSE_PARENT_WINDOW);
        packW(parent);
        return tf;
    }
    private static JComponent[] TEXT_DIALOGS=new JComponent[5];
    private static List<ChTextField> TEXT_TF[]=new List[5];
    static void openDialogSelect(int t) {
        JComponent dialog=TEXT_DIALOGS[t];
        if (dialog==null) {
            final StringBuffer sb=new StringBuffer(" Specify atom types using the ");
            if (t==SELECT_ATOMS) {
                sb.append("specific syntax of the particular 3D-viewer. <ul>");
                for(Object o :  StrapPlugins.allClassesV(ProteinViewer.class).asArray()) {
                    final ProteinViewer v=HotswapUtil.instance(o, ProteinViewer.class, false);
                    final Object example=v==null?null:v.getProperty(GET_ATOM_SELECTION_EXAMPLE);
                    if (example!=null) sb.append("<li>"+niceShrtClassNam(v)+"<pre>"+example+"</pre></li>");
                }
                sb.append("</ul>");
            }
            final Object
                info=
                t==SELECT_LABEL ? "A label text attached to the selected atoms." :
                t==SELECT_AMINO ?
                "An amino acid / atom selection like \"10-20,44\".<br>"+
                "Example giving the peptide chain:<pre class=\"data\">  30:C</pre>"+
                "Example giving the insertion code:<pre class=\"data\">  30A:C</pre>"+
                "Example for atoms C-Alpha and C-Beta:<pre class=\"data\">  30:C.CA.CB</pre>"+
                "Example for any carbon atom:<pre class=\"data\">  30:C.C*</pre>"+
                "Example of selecting residues by name:<pre class=\"data\">  \"GLY\"</pre>"+
                "Example for hetero compounds:<pre class=\"data\">  \"FAD\"</pre>"+
                "<br>To refere to a specific protein just type the protein name flanked by white space.<br>"+
                "For convenience use tab-key completion for protein names.<br>"+
                "<br>Inserting the current selection: Type Shift-Ctrl-V"
                :
                t==SELECT_ATOMS ? sb :
                sb+"generic (same for all 3D-programs) syntax.<br>"+
                "Examples: "+
                "Example one single atom: <pre class=\"data\"> .CA</pre>"+
                "Example main chain atoms: <pre class=\"data\"> .C.CA.CB.N.O</pre>"+
                "Example all nitrogen using asterisk: <pre class=\"data\"> .N*</pre>"+
                "Hint: ";
            final JComponent
                pCenter=vB(),
                panInfo=pnl(info+"<br>"+ChFrame.hintAOT(true)),
                pTop=vBhB(
                          hBL(t==SELECT_ATOMS?C(0xFF00):t==SELECT_AMINO?C(0xFF):null,
                              "Info", toggl().actionCollapse(false,panInfo).actionPack(),
                              " ",
                              t!=SELECT_LABEL? null: new ChButton("DIA_ATOM").li(li()).t("Specify atom ...")
                              ),
                          panInfo
                          );
            TEXT_DIALOGS[t]=dialog=newCNSEW(pCenter, pTop, null,null,null);
            final String ss[]=readLines(fileTB(t));
            if (ss!=null) for(String s : ss) if (nxt(NO_SPC,s)>=0) newRow(t,pCenter).t(s);
            addShutdownHook1(ChDelay.newThread(instance(),"SAVE_TB", intObjct(t)));
            newRow(t, pCenter);
        }
        ChFrame.frame(t==SELECT_LABEL ? "Label text" : t==SELECT_AMINO?"Aminos ":"Atoms ", dialog, 0)
            .shw(ChFrame.AT_CLICK|ChFrame.PACK);
    }
    final static File fileTB(int t) { return file(STRAPTMP+"/3D_sel_textbox"+t+".txt"); }
    /* <<< Atom Selection <<< */
    /* ---------------------------------------- */
    /* >>> View >>> */
    private static Map<ProteinViewer,Runnable> mapV2F=new WeakHashMap();
    static Runnable runViewerToFront(ProteinViewer pv) {
        Runnable r=mapV2F.get(pv);
        if (r==null) mapV2F.put(pv,r=ChDelay.newThread(instance(), "viewerToFront",new Object[]{pv,longObjct(OPEN_ADD_TAB)}));;
        return r;
    }
    public static void viewerToFront(long options, ProteinViewer pv) {
        if (pv==null) return;
        if (!isEDT()) inEDT(runViewerToFront(pv));
        else {
            final StrapAlign align=StrapAlign.getInstance();
            final String cn=clazzName(pv);
            final Component panel=viewerCanvas(pv);
            if (panel!=null) {
                final ChTableLayout tl=align.getTablePanel(cn);
                //tl.getJTable(). setIntercellSpacing(dim(10,10));
                final Window w=parentWndw(panel);
                if (tl!=null && cntains(panel,tl.getAllComponents())) StrapAlign.addDialog(tl);
                else if (gcp(KEY_detach_component,w)!=null) setWndwState('F',w);
                else if ((options&OPEN_ADD_TAB)!=0) tl.addColumn(panel);
                if ((options&OPEN_NARROW_OTHERS)!=0 && tl!=null) tl.narrowAllOthers(panel, ICON_HEIGHT);
            }
        }
    }
    /* <<< View <<< */
    /* ---------------------------------------- */
    /* >>> Script >>> */
    private final static long SEND_NO_STORE=1L<<32;
    public static boolean sendCommand(long options, CharSequence command0, ProteinViewer pv) {
        final Protein p=p(pv);
        String command=toStrgTrim(command0);
        if (p==null || sze(command)==0) return false;
        if (command.indexOf('\n')>0) {
            boolean succ=true;
            for(String line:splitLnes(SPLIT_TRIM, command)) {
                if (!sendCommand(options, line, pv)) succ=false;
            }
            return succ;
        }
        final boolean sup=supports(command,pv);
        final String word0=toStrgIntrn(wordAt(nxt(NO_SPC,command),command));
        if (idxOf(word0,allCommands())<0 || sup) {
            final Selection3D atoms[]=getSelA(pv);
            final String selID=gcps(KEY_SEL_ID,pv);
            if (word0==COMMANDlabel) {
                final boolean bb[]=selectionToBoolean(atoms,p);
                command=toStrg(SelectionVariables.replaceBB(command, bb, 0, selID, p, null));
            }
            pv.interpret(options,command);
            final String objectName=(String)pv.getProperty(GET_LAST_CREATED_OBJECT);
            final Map map=getSharedHashMap(pv);
            final boolean store=0==(options&SEND_NO_STORE);
            if (objectName!=null) {
                final Color color=gcp(KEY_SURF_COLOR, pv,Color.class);
                if (store) {
                    pcp2(false, KEY_SURF_ATOMS,  objectName, atoms, map);
                    pcp2(false, KEY_SURF_CMD,    objectName, command,     map);
                }
                pcp2(true, KEY_SURF_COLOR, objectName, color, map);
                pcp(KEY_COLOR, color, objectName);
                pcp(KEY_SURF_COLOR, pv, null);
            }
            if (sup) {
                final long rgb=rgbInScript(command);
                final Color color=rgb==-1 ? null : new Color((int)rgb,true);
                if (word0==COMMANDchange_object_color) {
                    final String thisObj=objectIdInScript(command);
                    pcp(KEY_COLOR, color, thisObj);
                    pcp2(false, KEY_SURF_COLOR, thisObj, color, map);
                    ChDelay.repaint(V3dListModel.jList(V3dListModel.SURFACE_OBJECTS,false,0),222);
                    ChDelay.repaint(V3dListModel.jList(V3dListModel.DELETED_OBJECTS,false,0),222);
                }
                if (word0==COMMANDsurface_color) pcp(KEY_SURF_COLOR, color, pv);
                if (word0==COMMANDcolor && 0!=(PROPERTY_RIBBON_COLOR_NO_CHANGE&flags(pv))
                    && color!=null && isSelctd(cbUpdateSurfaceColors)) updateSurfacesP(pv, false);
            }
            return true;
        }
        return false;
    }
    /* <<< Script <<< */
    /* ---------------------------------------- */
    /* >>> Surfaces >>> */
    private final static Object KEY_SURF_CMD=new Object(), KEY_SURF_COLOR=new Object(), KEY_SURF_ATOMS=new Object();
    private static void updateSurfacesP(ProteinViewer pv, boolean updateColor) {
        final String selID=gcps(KEY_SEL_ID, pv);
        final Selection3D[] atomType=getSelA(pv);
        final List v=(List)pv.getProperty(GET_SURFACEOBJECTS);
        int count=0;
        for(int i=sze(v); --i>=0;) {
            final String obj=get(i,v, String.class);
            updateSurface(obj, pv, updateColor, false);
            count++;
        }
        if (count>0) setSelAminos(0L, atomType , selID,  pv);
    }
    private static void updateSurface(String obj, ProteinViewer pv, boolean updateColor, boolean selectOnly) {
        final List vObjects=pv==null ? null : (List) pv.getProperty(GET_SURFACEOBJECTS);
        if (vObjects==null || obj==null) return;
        final Map map=getSharedHashMap(pv);
        final String script=gcp2(KEY_SURF_CMD, obj, map, String.class);
        final Selection3D[] atoms=gcp2(KEY_SURF_ATOMS, obj, map, Selection3D[].class);
        final Color color=gcp2(KEY_SURF_COLOR, obj, map, Color.class);
        if (selectOnly) setSelAminos(0L, atoms, "",  pv);
        else if (script!=null) {
            sendCommand(0L, COMMANDobject_delete+" "+obj,  pv);
            setSelAminos(0L, atoms, "",  pv);
            sendCommand(0L, script, pv );
            if (color!=null && updateColor) sendCommand(0L, COMMANDchange_object_color+" "+obj+" #"+toHex(color,true),  pv);
        }
    }
    private static JComponent _colorDia, _sfc_dia;
    private static ChButton cbUpdateSurfaceColors;
    private static ButColor _color_butColor;
    public static void openColorDialog() {
        if (_colorDia==null) {
            final ChJList jl=V3dListModel.jList(V3dListModel.PROTEINS, true,0);
            if (cbUpdateSurfaceColors==null) cbUpdateSurfaceColors=toggl("Also update colors of ribbons");
            final Object
                pNorth=hBL("To change the color of the entire protein you need to select a protein:"),
                pEntireProt=newCNSEW(scrllpn(0L,jl, dim(EX,4*ICON_HEIGHT)),pNorth,null,null,null),
                pSouth=vBhB(
                            cbUpdateSurfaceColors.cb(),
                            " ",
                            pEntireProt,
                            hBL(toggl().actionCollapse(false,new Object[]{jl, pEntireProt}),"Change color of the entire protein"),
                            pEntireProt
                            );
            (_color_butColor=new ButColor(0L, C(0xFF00), li())).li(li());
            _colorDia=newCNSEW(_color_butColor.getDialogPanel(),null,pSouth,null,null);
            enableDisable();
        }
        ChFrame.frame("Color", _colorDia, ChFrame.PACK).shw(ChFrame.AT_CLICK);
    }
    public static JComponent pnlSurfaceObjects() {
        if (_sfc_dia==null) {
            final Object
                ttU="The surface object is disposed and created de-nove.<br>"+
                "This might be necessary if the protein has moved in 3D or to adjust the color",
                bb[]=new Object[10],
                pSouth=vBhB(
                            hBL(
                                bb[0]=new ChButton("UPDATE").li(li()).t("Update ribbon or surface").i(IC_BLANK).tt(ttU),
                                bb[1]=new ChButton(COMMANDchange_object_color).li(li()).t("Update and set color").i(IC_BLANK).tt(ttU)
                                ),
                            bb[2]=butColor(ButColor.TRANSPARENCY_CHANGE_EVT, C(0xFF00), null, null).li(li())
                            .cp(KEY_ACTION_COMMAND, "CH_COLOR").t("Change color ...")
                            .cp(KEY_FRAME_TITLE, "Surface color "),
                            hBL(
                                bb[3]=new ChButton(COMMANDobject_delete).li(li()).t("Delete ribbon or surface").i(IC_KILL),
                                bb[4]=new ChButton("UNDEL_O").t("Rescue deleted").li(li()).i(IC_HAPPY).cp(KOPT_HIDE_IF_DISABLED,"")
                                )
                            );
            adAll(bb,vEnableDisable);
            _sfc_dia=newCNSEW(scrllpn(0L,V3dListModel.jList(V3dListModel.SURFACE_OBJECTS,true,0), dim(222,99)),"This is a list of ribbons and surfaces",pSouth,null,null);
            enableDisable();
        }
        return _sfc_dia;
    }
    public static void transform3D(Matrix3D m3d, ProteinViewer pv) {
        if (supports(COMMANDset_rotation_translation,pv)) {
            sendCommand(0L, COMMANDset_rotation_translation+" "+(m3d==null?null:m3d.toText(0,"")), pv);
            final long flags=flags(pv);
            if (0!=(flags&PROPERTY_NEEDS_UPDATE_SURFACES_WHEN_ROTATED)) updateSurfacesP(pv,true);
            if (supports(COMMANDcenter_amino,pv)) sendCommand(0L,COMMANDcenter_amino, pv);
        }
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> ToolTip >>> */
    public String getToolTipText(JComponent c, AWTEvent ev) {
        final ProteinViewer pvF=getPV(PV_FOCUSED);
        final Map mapF=getSharedHashMap(pvF);
        if (c instanceof ChJList && mapF!=null) {
            final Object obj=objectAt(ev);
            if (obj instanceof String) {
                return
                    "<pre>"+
                    "\n  Cmd: "+gcp2(KEY_SURF_CMD, obj, mapF, String.class)+
                    "\natoms: "+selectionToText(GENERIC, gcp2(KEY_SURF_ATOMS, obj, mapF, Selection3D[].class))+
                    "\nColor: "+toHex(gcp2(KEY_SURF_COLOR, obj, mapF, Color.class), true)+
                    "</pre>";
            }
        }
        return null;
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Undelete >>> */
    private static  JComponent _undelObj;
    private static void undelObj(boolean showOnly, ProteinViewer pv) {
        final Map map=getSharedHashMap(pv);
        if (map==null) return;
        final ChJList jl=V3dListModel.jList(V3dListModel.DELETED_OBJECTS,true,0);
        if (showOnly) {
            if (_undelObj==null) {
                final Object
                    b=new ChButton("UNDEL_OBJ").t("Undelete").li(li()),
                    sp=scrllpn(0L, jl, dim(99,99));
                _undelObj=newCNSEW(sp,null,pnl(b),null,null);
                vEnableDisable.add(b);
            }
            ChFrame.frame("Deleted",_undelObj,0L).shw(ChFrame.AT_CLICK);
        } else {
            for(Object o : oo(jl.getSelectedValues())) {
                V3dListModel.list(V3dListModel.DELETED_OBJECTS, pv, true).remove(o);
                updateSurface(toStrg(o), pv, true, false);
            }
            enableDisable();
        }
    }
}