package charite.christo.strap.extensions;
import java.awt.*;
import javax.swing.*;
import charite.christo.strap.*;
import charite.christo.*;
import charite.christo.protein.*;
import charite.christo.interfaces.*;
import charite.christo.strap.interfaces.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;

public class AbstractAlignAndCompare implements NeedsSharedInstance, ProteinProteinValue,HasSharedControlPanel, HasControlPanel, Disposable, StrapListener, HasModiCount {
    private char mode='D';
    protected void setMode(char m) { this.mode=m;}

    @Override protected void finalize() throws Throwable{
        logFinalize(ANSI_GREEN+" IndividualSequenceDisSimilarity ","");
        super.finalize();
    }
    private int modificationCount, modiCountPara;

    private StrapProtein p0,p1;
    public boolean setProtein(Protein prot0) { if (p0!=prot0) { p0=(StrapProtein)prot0; modiCountPara++; } return true;}
    public void setOtherProtein(StrapProtein prot1) { if (p1!=prot1) { p1=prot1; modiCountPara++;}}
    public StrapProtein getProtein(){ return p0;}
    public StrapProtein getOtherProtein(){ return p1;}

    private ChCombo comboClassAligner_;
    public ChCombo comboClass() {
        if (p0!=null) {
            return getShared().comboClass();
        }
        if (comboClassAligner_==null) {
            comboClassAligner_=StrapAlign.classChoice(mode=='R' ? Superimpose3D.class : SequenceAligner.class);
            if (mode=='D') comboClassAligner_.s(CLASS_MultipleAlignerClustalW).classRenderer(SequenceAligner.class);
            if (mode=='R') comboClassAligner_.classRenderer(Superimpose3D.class);
            if (mode=='S') comboClassAligner_.s(CLASS_MultipleAlignerClustalW).classRenderer(HasScore.class);
        }
        return comboClassAligner_;
    }

    public Object getOtherClass() {  return (mode=='A' ? comboClassDis() : comboClass()).getSelectedItem();}
    public void setOtherClass(Object o) { (mode=='A' ? comboClassDis() : comboClass()).s(o); }

    /* SequenceDisSimilarity */
    private int comboClassDis_modi;
    private SequenceDisSimilarity sequenceDis_;
    private SequenceDisSimilarity sequenceDis() {
        if (p0!=null) return getShared().sequenceDis();
        final int modi=comboClassDis().getModificationCount();
        if (sequenceDis_==null || comboClassDis_modi!=modi) {
            comboClassDis_modi=modi;
            modiCountPara++;
            sequenceDis_=HotswapUtil.instance(comboClassDis(),SequenceDisSimilarity.class,true);
        }
        return sequenceDis_;
    }
    private ChCombo comboClassDis_;
    private ChCombo comboClassDis() {
        if (p0!=null) return getShared().comboClassDis();
        if (comboClassDis_==null) {
            comboClassDis_=StrapAlign.classChoice(SequenceDisSimilarity.class)
                .classRenderer(SequenceDisSimilarity.class).s(charite.christo.protein.CrudeSequenceDisSimilarity.class);
        }
        return comboClassDis_;
    }

    private AbstractAlignAndCompare sharedInstance_;
    public void setSharedInstance(Object i) { this.sharedInstance_=(AbstractAlignAndCompare)i;}
    public Object getSharedInstance() { return this.sharedInstance_;}
    private static AbstractAlignAndCompare staticShared[];

    private AbstractAlignAndCompare getShared() {
        if (this.sharedInstance_!=null) return  (AbstractAlignAndCompare)this.sharedInstance_;
        if (staticShared==null) staticShared=new AbstractAlignAndCompare['Z'];
        if (staticShared[this.mode]==null) staticShared[this.mode]=new AbstractAlignAndCompare();
        return staticShared[mode];
    }

    private Container ctrlPnl;
    public Container getSharedControlPanel() {
        if (p0!=null) return getSharedCtrlPan(getShared(), false);
        if (ctrlPnl==null) {
            switch(mode) {
            case 'D':
                ctrlPnl=vBhB(
                           "<h2>Sequence Dis-similarity of 2 Sequences<h2>",
                           "Class to compute alignment of two sequences",
                           pnl(comboClass().panel()),
                           " ",
                           "Class to measure the distance between two aligned sequences",
                           pnl(comboClassDis().panel())
                           );
                break;
            case 'R':
                ctrlPnl=vBhB("What superposition method ? ", comboClass().panel());
                break;
            case 'S':
                ctrlPnl=vBhB(
                           "<h2>Alignment score of two proteins<h2>",
                           "You can chose a class to align two sequences. It must implement <i>HasScore</i>.",
                           pnl(comboClass().panel())
                           );
                break;
            case 'A':
                ctrlPnl=vBhB(
                           "<h2>Sequences-Dissimilarity<h2>",
                           "The alignment is taken from the working alignment.",
                           "You can chose a class to measure the distance between two aligned sequences",
                           pnl(comboClassDis().panel())
                           );
                break;
            }
        }
        return ctrlPnl;
    }

    public Container getControlPanel() { return getCtrlPanel(this._calculator);}

    private int from=0,to=MAX_INT;
    public void setColumnRange(int f,int t) { if (from!=f || to!=t) { from=f; to=t; modiCountPara++; }}

//     private SequenceAligner shared_sa;
//     public SequenceAligner shared_sa() {
//         if (p0!=null) return getShared().shared_sa();
//         if (shared_sa==null) shared_sa=StrapCombo.instanceSA(comboClass(),2,null);
//         return shared_sa;
//     }

//     private Superimpose3D shared_s3d;

    private transient StrapProtein otherProt_last;
    private double value=Double.NaN;
    public double getValue() {
        if (mode=='A') {
            final StrapProtein p0=getProtein(), p1=getOtherProtein();
            if (p0==null||p1==null) return Double.NaN;
            final byte[] gapped0=p0.getGappedSequenceTS();
            final byte[] gapped1=p1.getGappedSequenceTS();
            final int m=
                p0.getModificationCount(ProteinModiCount.MC_GAPPED_SEQUENCE)+
                p1.getModificationCount(ProteinModiCount.MC_GAPPED_SEQUENCE)+modiCountPara;
            if (modificationCount!=m || otherProt_last!=p1) {
                modificationCount=m;
                otherProt_last=p1;
                final int max=maxi(gapped0.length, gapped1.length, to);
                int maxCol=0;
                while(maxCol<max && gapped0[maxCol]!=0 && gapped1[maxCol]!=0) maxCol++;

                SequenceDisSimilarity dis=sequenceDis();
                this.value= dis!=null && p0!=null && p1!=null ? dis.getValue(gapped0, gapped1,from, maxCol) : Double.NaN;
            }
        }
        return this.value;
    }

    private Object _calculator;
    public void compute(){
        if (mode=='A') return;
        final StrapProtein p0=getProtein(), p1=getOtherProtein();
        if (p0==null||p1==null || mode=='R' && ! (p0.hasCalpha() && p1.hasCalpha()) ) { this.value=Double.NaN; return; }
        final byte[]
            seq0=p0.getResidueType(),
            seq1=p1.getResidueType();

        SequenceAligner sa=null;
        Superimpose3D s3d=null;
        final Object calculator=_calculator=
            (mode=='D' || mode=='S')  ? sa=StrapPlugins.instanceSA(comboClass(),2,true) :
            (mode=='R') ? s3d=HotswapUtil.instance(comboClass(),Superimpose3D.class,true) :
            null;
        if (isInstncOf(NeedsProteins.class, calculator)) ((NeedsProteins)calculator).setProteins(p0,p1);
        if (isInstncOf(Superimpose3D.class, calculator)) {
            Superimpose3D sa3d=(Superimpose3D)calculator;
            boolean subset=from>0 || to!=MAX_INT;
            sa3d.setReferenceProtein(!subset ? p0 : DialogSuperimpose3D.pSubset(p0,p0.getResiduesInAlignmentcolumnrangeAsBoolean(from,to)));
            sa3d.setMobileProtein   (!subset ? p1 : DialogSuperimpose3D.pSubset(p1,p1.getResiduesInAlignmentcolumnrangeAsBoolean(from,to)));
        }
        final int m=p0.getModificationCount(ProteinModiCount.MC_RESIDUE_TYPE)+p1.getModificationCount(ProteinModiCount.MC_RESIDUE_TYPE)+comboClass().getModificationCount()+modiCountPara;
        if (modificationCount!=m) {
            modificationCount=m;
            if ( (mode=='D' || mode=='S') && sa!=null) {
                //byte ss[][]=null;
                sa.setSequences(seq0,seq1);
                sa.compute();
                if (mode=='D') {
                    final SequenceDisSimilarity sequenceDis=sequenceDis();
                    final byte[][] ss=sa.getAlignedSequences();
                    this.value=sequenceDis!=null && ss!=null && ss.length>=2 ? sequenceDis.getValue(ss[0],ss[1],0,MAX_INT) : Double.NaN;
                } else if (isAssignblFrm(HasScore.class,sa)) this.value=((HasScore)sa).getScore();
            }
            if (mode=='R' && s3d!=null) {
                s3d.compute();
                Superimpose3D.Result r=s3d.getResult();
                this.value=r!=null ? r.getRmsd() : Double.NaN;
            }
        }
    }
    public int getModificationCount() { return modificationCount;}
    public void dispose() {
        Object o=_calculator;
        if (o!=null && isInstncOf(Disposable.class,o)) ((Disposable)o).dispose();
    }

    public void handleEvent(StrapEvent ev) {
        final int t=ev.getType();
        if (
            (mode=='D' || mode=='S') && (t&StrapEvent.FLAG_RESIDUE_TYPES_CHANGED)!=0 ||
            mode=='R' && (t&StrapEvent.FLAG_ATOM_COORDINATES_CHANGED)!=0 ||
            mode=='A' && (t&StrapEvent.FLAG_ALIGNMENT_CHANGED)!=0)
            StrapAlign.dispatchLater(StrapEvent.PROTEIN_PROTEIN_DISTANCE_CHANGED,222);
    }
}
