package charite.christo.strap.extensions;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.strap.*;
import charite.christo.protein.*;
import charite.christo.strap.interfaces.*;
import java.awt.event.*;
/**HELP
A profile of the sequence variability of the multiple sequence alignment is displayed.
The variability is calculated with the formula by Shannon.
Gaps are treated as a 21th amino acid.
*/
public class AlignmentEntropy implements ValueOfAlignPosition,StrapListener,HasSharedControlPanel,ActionListener, HasModiCount{
    private StrapProtein pp[];
    private static ActionListener inst;
    public void actionPerformed(ActionEvent ev) {
        StrapAlign.dispatchLater(StrapEvent.VALUE_OF_ALIGN_POSITION_CHANGED,111);
    }

    public void setProteins(StrapProtein... proteins) { pp=proteins; lastValues=null;   }
    public StrapProtein[] getProteins() {return pp;}
    public void handleEvent(StrapEvent ev) {
        if (ev.getType()==StrapEvent.ALIGNMENT_CHANGED)
            StrapAlign.dispatchLater(StrapEvent.VALUE_OF_ALIGN_POSITION_CHANGED,111);
    }
    /**
       Avoid calculating by caching the value;
    */
    private double[] lastValues;
    private int[] wieOft;
    public int getModificationCount() { return modiCount;}
    private int modiCount;
    public double[] getValues() {return getValues(combo==null ? 0 : combo.i());}
    public double[] getValues(int mode) {
        if (pp==null || pp.length==0) return null;
        int mc=(combo!=null ? combo.getModificationCount() : 0);
        for(StrapProtein p : pp) {
            if (p!=null)  mc+=p.getModificationCount(ProteinModiCount.MC_GAPPED_SEQUENCE);
        }

        if (this.lastValues==null ||  this.modiCount!=mc) {
            this.modiCount=mc;
            final int posMax=StrapProtein.getMaxColumn(pp);
            final double
                entropyMax=Math.log(pp.length)*pp.length,
                entropy[]=new double[posMax+1];
            if (wieOft==null) wieOft=new int[255]; else java.util.Arrays.fill(wieOft,0);
            double sumEntropy=0;
            for(int iPos=0;iPos<posMax;iPos++) {
                java.util.Arrays.fill(wieOft,0);
                for(int iP=0;iP<pp.length;iP++) {
                    final StrapProtein p=pp[iP];
                    final int iA=p.column2index(iPos);
                    final byte c= iA<0 ? 32 : (byte)(p.getResidueType(iA)&~32);
                    ++wieOft[c];
                }
                int insgesamt=0;
                for(int c='A'; c<='Z'; c++) insgesamt+=wieOft[c];
                double entro=0;
                if (insgesamt>0) {
                    for(int c='A'; c<='Z'; c++)
                        if (wieOft[c]>0 && (mode==GAP_IS_21TH_AMINO_ACID || c!=' '))  {
                            final double p=wieOft[c]/(double)insgesamt;
                            entro-=p*Math.log(p);
                        }
                    /* A gap is like a new amino acid occurring once */
                    if (mode==GAP_IS_A_DIFFERENT_AMINO_ACID && wieOft[' ']>0) { double p=1f/insgesamt; entro-=wieOft[' ']*p*Math.log(p);}
                }
                if (mode==SKIP_GAPS && wieOft[' ']>0) entropy[iPos]=Double.NaN;
                else {
                    entropy[iPos]=entro/entropyMax;
                    sumEntropy+=entro;
                }
            }

            this.lastValues=entropy;
        }
        return lastValues;
    }
    public final static int SKIP_GAPS=0, GAP_IS_21TH_AMINO_ACID=1, GAP_IS_A_DIFFERENT_AMINO_ACID=2;
    private static ChCombo combo;
    private static java.awt.Container controlPanel_;
    public java.awt.Container getSharedControlPanel(){
        if (controlPanel_==null) {
            combo=new ChCombo("skip columns with gaps","a gap is the 21st amino acid","a gap is always a different amino acid").li(inst=new AlignmentEntropy());
            controlPanel_=ChUtils.vB("How should gaps be treated ?",combo);
        }
        return controlPanel_;
    }
}
