package charite.christo.strap.extensions;
import static charite.christo.ChUtils.*;
import charite.christo.protein.*;
import charite.christo.strap.*;
import charite.christo.*;
import charite.christo.interfaces.*;
import java.io.File;
import static charite.christo.ChUtils.*;

/**HELP
Align-m is an accurate and highly versatile multiple alignment program.
<br>
Author: Ivo Van Walle
<br>

http://bioinformatics.vub.ac.be/software/software.html
<pre>
<b>Copyright notice</b>

Align-m Multiple Sequence Alignment
Copyright (c) 2003-2005 AlgoNomics NV
Author: Ivo Van Walle (ivwalle@vub.ac.be)
All Rights Reserved

Permission to use, copy, and distribute any part of this Align-m
software for educational, research and non-profit purposes, without
fee, and without a written agreement is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies.

Those desiring to incorporate this Align-m Software into commercial products or use for commercial purposes should contact
AlgoNomics NV
Technologiepark 4
9052 Gent-Zwijnaarde
Belgium

Ph : +32 (0) 9 241 1100
FAX: +32 (0) 9 241 1102
Web: www.algonomics.com

IN NO EVENT SHALL ALGONOMICS BE LIABLE TO ANY PARTY FOR DIRECT,
INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
LOST PROFITS, ARISING OUT OF THE USE OF THIS ALIGN-M SOFTWARE, EVEN IF
ALGONOMICS HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE ALIGN-M SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND
ALGONOMICS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS. ALGONOMICS MAKES NO REPRESENTATIONS
AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR EXPRESS,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE
OF THE ALIGN-M SOFTWARE WILL NOT INFRINGE ANY PATENT, TRADEMARK OR
OTHER RIGHTS.
</pre>
*/
public final class MultipleAlignerAlign_m extends AbstractSuperposition implements SequenceAligner, HasPrgParas {
    private final static String PARAMERS[][]={
            {"-v","[x]Verbose comments during execution"},
            {"-s2p_guided_aln","[x]Substitution matrix name. For protein sequences"},
            {"-s2p_m","Width of ungapped segment to average over. default: BLOSUM62"},
            {"-s2p_w", "nsigma max. Maximum number of sigma that a new set of columns may score on average below the first one. default: 15"},
            {"-s2p_minscore","minscore. Minimum score that a new set of columns may score on average (supersedes -s2p nsigma max) .default: 0.0"},
            {"-s2p_go","Gap opening penalty for guided alignments Gap extension penalty for guided. default: 12.0"}
    };

    public MultipleAlignerAlign_m() {
        getPrgParas().defineGUI(PARAMERS, FURTHER_PARAMETERS);
    }

    @Override public Superimpose3D.Result computeResult(byte[][] seqs, Protein[] proteinsNotNeeded) {
        final String n="align_m";
        final AbstractExecutable e=new AbstractExecutable(n);
        setExecutable(e, (String[])null, (String)null, (String[])null);
        if (!Insecure.EXEC_ALLOWED) return null;
        final File dirTmp=dirTemp(), mfaInfile=mfaInFile(seqs), fExe=file(e.dirBinaries(),"/"+n+winDotExe());
        final byte[][] sequences=getSequences();
        if (sequences.length<3) { StrapAlign.errorMsg("At least 3 sequences required for Align-m"); return null;}

        e.exec().setCommandLineV(fExe, getPrgParas().asStringArray(0L), "-i", mfaInfile, "-o", "output.fa")
            .dir(dirTmp).setOptions(ChExec.STDOUT | ChExec.STDERR | ChExec.LOG | ChExec.PLEASE_INSTALL_MANUALLY)
                .run();
        return parse(seqs.length, readBytes(file(dirTmp,"output.fa")), null, Superimpose3D.PARSE_SCORE|Superimpose3D.PARSE_MSA_START_S);
    }
}
