package charite.christo.strap.extensions;
import static charite.christo.ChUtils.*;
import charite.christo.protein.*;
import charite.christo.*;
import charite.christo.strap.*;
import charite.christo.interfaces.*;
import java.io.File;
/**HELP
<b>Author:</b> A.S. Schwartz, E.W. Myers and L. Pachter
<br>
WIKI:AMAP
*/
public final class MultipleAlignerAmap extends AbstractSuperposition implements SequenceAlignerSorting, HasPrgParas {
    private final static String PARAMERS[][]={
        {"--consistency","use 0 <= REPS <= 5 (default: 0) passes of consistency transformation"},
        {"--iterative-refinement","use 0 <= REPS <= 1000 (default: 100) passes of iterative-refinement"},
        {"--pre-training","use 0 <= REPS <= 20 (default: 0) rounds of pretraining"},
        {"-pairs","[x]generate all-pairs pairwise alignments"},
        {"-viterbi","[x]use Viterbi algorithm to generate all pairs (automatically enables -pairs)"},
        {"--verbose","[x]report progress while aligning (default: off)"},
        {"-annot","write annotation for multiple alignment to FILENAME"},
        {"--train","compute EM transition probabilities, store in FILENAME (default: no training)"},
        {"--emissions","also reestimate emission probabilities (default: off)"},
        {"--paramfile","read parameters from FILENAME (default: )"},
        {"--gap-factor","use GF as the gap-factor parameter, set to 0 for best sensitivity, higher values for better specificity (default: 0.5)"}
    };
    public MultipleAlignerAmap() {
        getPrgParas().defineGUI(PARAMERS, FURTHER_PARAMETERS);
    }

    @Override public Superimpose3D.Result computeResult(byte[][] seqs, Protein[] proteinsNotNeeded) {
        final String n="amap_2_1",
            instScript=
            "cd amap-align/align\n"+
            "C_PLUS_PLUS_COMPILER MNO_CYGWIN_FLAG -O3  -W -Wall -pedantic -DNDEBUG -DNumInsertStates=1 -DVERSION='\"AMAP.2.1\"' -funroll-loops -lm -o ../../"+n+".exe Amap.cc",
            srcUrls[]={"http://baboon.math.berkeley.edu/amap/download/amap.2.1.tar.gz"}; /* WATCH_URL */
        final AbstractExecutable e=new AbstractExecutable(n);
        setExecutable(e, (String[])null, instScript, (String[])null);
        if (!Insecure.EXEC_ALLOWED) return null;
        final File mfaInfile=mfaInFile(seqs);

        e.exec()
            .setCommandLineV(e.fileExecutable(), getPrgParas().asStringArray(EQUALS_SIGN) ,"-a", mfaInfile)
            .dir(dirTemp())
            .setOptions(ChExec.STDOUT | ChExec.STDERR)
            .run();
        return parse(seqs.length, e.exec().getStdoutAsByteArray(), null, Superimpose3D.PARSE_SCORE|Superimpose3D.PARSE_MSA_START_S);
    }
}
