package charite.christo.strap.extensions;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import static charite.christo.ChUtils.*;
/**HELP
ClustalW is a widely applied program to generate multiple sequence
alignments automatically.
<br>
<b>Publication:</b> PUBMED:12166070 <br>
<b>Home: </b> http://www.clustal.org/ <br>
<br><br><b>NT-sequence:</b>
ClustalW treats the letters as nucleotides if 85% or more of the characters in the sequence are A,C,G,T,U or N.
See UBIC:clustalw

<b>Related sites:</b>http://www.clustal.org/
*/
public final  class MultipleAlignerClustalW extends charite.christo.strap.AbstractSuperposition implements HasPrgParas, SequenceAlignerTakesProfile, SequenceAlignerSorting, HasScore {
    private final static String PARAMERS[][]={
        {"-GAPOPEN","gap opening penalty"},
        {"-GAPEXT","gap extension penalty"},
        {"-MATRIX","Protein weight matrix=BLOSUM, PAM, GONNET, ID or filename"},
    };
    public MultipleAlignerClustalW() {
        getPrgParas().defineGUI(PARAMERS, FURTHER_PARAMETERS);
    }
    /* ---------------------------------------- */
    /* >>> Evolutionary Tree, used by  >>> */

    private String outputTree;   /** option -OUTPUTTREE=nj */
    //public void setOutputTree(String tree) { outputTree=tree; }

    public java.io.File computePhFile(byte[][] alignment, String tree) {
        if (alignment==null) return null;
        outputTree=tree;
        computeResult(alignment, null);
        return file(dirTemp(),withoutLastComponent(MFA_FILE,'.')+".ph");

    }
    //public File getPhFile() { return file(withoutLastComponent(MFA_FILE,'.')+".ph");}
    /* <<< Evolutionary <<< */
    /* ---------------------------------------- */
    /* >>> Compute >>> */

    @Override public Superimpose3D.Result computeResult(byte[][] seqs, Protein[] proteinsNotNeeded) {
        final String n="clustalw_2",
            instScript="sh configure\n"+
            "make\n"+
            "cp src/clustalw2"+winDotExe()+("  "+n+".exe\n");
        final AbstractExecutable e=new AbstractExecutable(n);
        setExecutable(e, (String[])null, instScript,AbstractExecutable.DEFAULT_BINARY);
        if (!Insecure.EXEC_ALLOWED) return null;
        final java.io.File  fileOut=file(dirTemp(),"output.msf"), mfaInfile=mfaInFile(seqs), fileExe=e.fileExecutable();
        if (outputTree!=null) wrte(mfaInfile,ChAlignUtils.toGappedFasta(seqs));
        final ChExec ex=e.exec();
        final int offset=seqs.length;
        final byte[][][] profiles=getProfiles();
        if (profiles!=null && profiles.length>0 && profiles[0]!=null) {
            wrte(file(dirTemp(),"profile1"),ChAlignUtils.toGappedMSF(profiles[0],"p",offset,true,(StringBuffer)null));
            ex.setCommandLineV(fileExe, getPrgParas().asStringArray(EQUALS_SIGN),"-PROFILE1=profile1","-sequences","-PROFILE2="+MFA_FILE,"-OUTFILE=output.msf");
        }
        else if (outputTree!=null) ex.setCommandLineV(fileExe,"-tree","-INFILE="+MFA_FILE,"-OUTPUTTREE="+outputTree );
        else {
            ex.setCommandLineV(fileExe,getPrgParas().asStringArray(EQUALS_SIGN),"-INFILE="+MFA_FILE,"-OUTFILE=output.msf" );
            //putln("getParametersAsStringArray()=",getPrgParas().asStringArray(EQUALS_SIGN));
            //putln("map=",getPrgParas().getMapVariables());
        }
        ex.dir(dirTemp()).setOptions(ChExec.LOG | ChExec.STDOUT | ChExec.STDERR | ChExec.CYGWIN_DLL);
        if (e.isStopped()) return null;
        ex.run();
        if (outputTree!=null) return null;
        final Superimpose3D.Result result=parse(seqs.length,ex.getStdoutAsByteArray(), null, Superimpose3D.PARSE_SCORE);
        return parse(seqs.length, readBytes(fileOut), result, Superimpose3D.PARSE_MSA_START_S);
    }

    @Override public int getMaxNumberOfProfiles() { return 1;}

}
/*
EBI (apc) (SUP#552193)
 */