package charite.christo.strap.extensions;
import static charite.christo.ChUtils.*;
import charite.christo.protein.*;
import charite.christo.*;
import charite.christo.strap.*;
import charite.christo.interfaces.*;
import charite.christo.strap.interfaces.*;

import static charite.christo.ChUtils.*;
/**HELP
Home page <i>HOME_PAGE:MultipleAlignerKalign</i>
Author: Timo Lassmann
Publication: PUBMED:16343337
*/
// <timolassmann@gmail.com> lukas.kall@ki.se
public final class MultipleAlignerKalign extends AbstractSuperposition implements HasPrgParas, SequenceAlignerSorting {
    private final static String PARAMERS[][]={
        {"--gapopen"," Gap open penalty"},
        {"--gapextension","Gap extension penalty"},
        {" --terminal_gap_extension_penalty","Terminal gap penalties"},
        {"--matrix_bonus","A constant added to the substitution matrix."}
    };
    public MultipleAlignerKalign() {
        getPrgParas().defineGUI(PARAMERS, FURTHER_PARAMETERS);
    }

    @Override public Superimpose3D.Result computeResult(byte[][] seqs, Protein[] proteinsNotNeeded) {
        final String n="kalign",
            instScript=
            "C_COMPILER --version\n"+
            "cd "+n+"\n"+
            "sh configure\n"+
            "make\n"+
            "mv kalign"+winDotExe()+(" ../"+n+".exe\n"),
            srcUrls[]={"http://msa.cgb.ki.se/downloads/kalign/current.tar.gz"}; /* WATCH_URL */
        final AbstractExecutable e=new AbstractExecutable(n);
        setExecutable(e, srcUrls, instScript, (String[])null);
        if (!Insecure.EXEC_ALLOWED) return null;
        final java.io.File dirTmp=dirTemp(), fileOut=file(dirTmp,"output.fa"), mfaInfile=mfaInFile(seqs);

        e.exec().setCommandLineV(
                                 // isWindows() ?"CYGWINSH" : null,
                                 e.fileExecutable(), getPrgParas().asStringArray(EQUALS_SIGN),
                                 "-sort","tree","--quiet",
                                 "--infile",mfaInfile, "-o", fileOut)
            .dir(e.dirTemp()).setOptions(ChExec.STDOUT | ChExec.STDERR | ChExec.LOG | ChExec.CYGWIN_DLL).run();
        return parse(seqs.length, readBytes(fileOut), null, Superimpose3D.PARSE_SCORE|Superimpose3D.PARSE_MSA_START_S);
    }
}
