package charite.christo.strap.extensions;
import charite.christo.strap.interfaces.*;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.strap.*;
import charite.christo.protein.Superimpose3D;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
public class PairAlignerJAligner extends AbstractSuperposition implements HasPrgParas, SequenceAligner, HasScore  {

    private static String
        MATRICES[]={
        "BLOSUM100","BLOSUM30","BLOSUM35","BLOSUM40","BLOSUM45","BLOSUM50","BLOSUM55","BLOSUM60","BLOSUM62","BLOSUM65","BLOSUM70",
        "BLOSUM75","BLOSUM80","BLOSUM85","BLOSUM90","BLOSUMN","DAYHOFF","EDNAFULL","GONNET","IDENTITY","MATCH",
        "PAM10","PAM100","PAM110","PAM120","PAM130","PAM140","PAM150","PAM160","PAM170","PAM180","PAM190","PAM20","PAM200","PAM210","PAM220","PAM230","PAM240","PAM250","PAM260","PAM270","PAM280","PAM290","PAM30","PAM300","PAM310","PAM320","PAM330","PAM340","PAM350","PAM360","PAM370","PAM380","PAM390","PAM40","PAM400","PAM410","PAM420","PAM430","PAM440","PAM450","PAM460","PAM470","PAM480","PAM490","PAM50","PAM500","PAM60","PAM70","PAM80","PAM90"},

        PARAMERS[][]={
            {"open",   "f gap open panelty","10.5"},
            {"extend", "f gap extension panelty","0.5"},
            {"matrix", "{"+joinAsStrg(MATRICES,",")+"} similarity matrix", "BLOSUM62" }
        };

    public PairAlignerJAligner() {
        getPrgParas().defineGUI(PARAMERS,0L);
    }

    public final long getPropertyFlags() { return PROPERTY_ONLY_TWO_SEQUENCES|PROPERTY_LOCAL_ALIGNMENT;}
    private static jaligner.matrix.Matrix matrix;
    private static String matrixName;

    public Superimpose3D.Result computeResult(byte[][] ss, charite.christo.protein.Protein[] ppNotNeeded) {
        final java.util.Map<String,String>map=getPrgParas().getMapVariables();
        String mxName=map!=null ? map.get("matrix") : null;
        if (mxName==null) mxName="BLOSUM62";
        if (matrixName!=mxName) {
            try {
                matrix=jaligner.matrix.MatrixLoader.load(mxName);
                matrixName=mxName;
            } catch(Exception e){ putln("caught in PairAlignerJAligner mx="+mxName+" "+e); }
        }
        if (matrix==null) { putln("Error in PairAlignerJAligner: could not load matrix ",mxName," matrix"); return null;}

        float open=map!=null ?   (float)atof(map.get("open")) : Float.NaN;
        float extend=map!=null ? (float)atof(map.get("extend")) : Float.NaN;
        if (Float.isNaN(open) || open<0) open=10.5f;
        if (Float.isNaN(extend) || extend<0) extend=0.5f;
        jaligner.Sequence seq0=null,seq1=null;
        try {
            seq0=jaligner.util.SequenceParser.parse(new String(ss[0]));
            seq1=jaligner.util.SequenceParser.parse(new String(ss[1]));
        }catch(Exception e){
            putln(ANSI_RED+"Caught in PairAlignerJAligner"+ANSI_RESET);
            putln("\nss[0]=", ss[0],"\nss[1]=", ss[1]);
            stckTrc(e);
            return null;
        }
        final jaligner.Alignment a=jaligner.SmithWatermanGotoh.align(seq0,seq1,matrix,open,extend);
        final char[] c1=a.getSequence1(),c2=a.getSequence2();
        final byte[] b1=new byte[c1.length], b2=new byte[c2.length];
        for(int i=c1.length; --i>=0;) b1[i]=(byte)c1[i];
        for(int i=c2.length; --i>=0;) b2[i]=(byte)c2[i];
        final Superimpose3D.Result result=new Superimpose3D.Result(2);
        result.setGappedSequences(new byte[][]{b1,b2});
        result.setScore(a.getScore());

        return result;
    }

}
