package charite.christo.strap.extensions;
import neobio.alignment.*;
import charite.christo.*;
import charite.christo.interfaces.*;
import static charite.christo.protein.Superimpose3D.Result;
import java.io.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static charite.christo.strap.extensions.PairAlignerNeoBioPROXY.*;
public class PairAlignerNeoBio extends  charite.christo.strap.AbstractSuperposition implements SequenceAligner, HasScore {
    private final static String
    ALGOROTHMS[]={ LOCAL_SMITHWATERMAN, GLOBAL_NEEDLEMANWUNSCH, GLOBAL_CROCHEMORELANDAUZIVUKELSON,LOCAL_CROCHEMORELANDAUZIVUKELSON};
  /*
        PARAMERS[][]={
            {PARA_ALGORITHM,"{"+joinAsStrg(ALGOROTHMS,",")+"} the algorithm"},
            {PARA_MATCH,"i ","1"},
            {PARA_MISMATCH,"i ", "-1"},
            {PARA_GAP,"i ","-1"},
        };
  */
    public PairAlignerNeoBio() {
        setImplementationID("3");
    }
    public final long getPropertyFlags() { return PROPERTY_ONLY_TWO_SEQUENCES|PROPERTY_LOCAL_ALIGNMENT;}

    @Override public Result computeResult(byte[][] ss, charite.christo.protein.Protein[] notNeeded) {
        PairwiseAlignment alignment;
        if (ss==null || ss.length<2) return null;
        //putln("PairAlignerNeoBio ss0=",nxt(NO_LETTR,ss[0]));
        //putln("PairAlignerNeoBio ss1=",nxt(NO_LETTR,ss[1]));

        final java.util.Map<String,String> mapPara=getPrgParas().getMapVariables();
        int iAlg=mapPara==null ? 0 : idxOfStrg(mapPara.get(PARA_ALGORITHM), ALGOROTHMS);
        final PairwiseAlignmentAlgorithm alg=
            iAlg<=0 ? new SmithWaterman():
            iAlg==1 ? new NeedlemanWunsch():
            iAlg==2 ? new CrochemoreLandauZivUkelsonGlobalAlignment():
            new CrochemoreLandauZivUkelsonLocalAlignment();
        int match=1,mismatch=-1,gap=-1;
        if (mapPara!=null) {
            match=atoiOrDefault(mapPara.get(PARA_MATCH), match);
            mismatch=atoiOrDefault(mapPara.get(PARA_MISMATCH), mismatch);
            gap=atoiOrDefault(mapPara.get(PARA_GAP), gap);
        }
        InputStreamReader r0=null,r1=null;
        try{
            r0=new InputStreamReader(new ByteArrayInputStream(ss[0]));
            r1=new InputStreamReader(new ByteArrayInputStream(ss[1]));
            alg.loadSequences(r0,r1);
            alg.setScoringScheme(new BasicScoringScheme(match,mismatch,gap));
            //putln("match="+match+" mismatch="+mismatch+" gap="+gap);
            alignment=alg.getPairwiseAlignment();
        } catch(Exception e) {
            putln(RED_CAUGHT_IN+"PairAlignerNeoBio:",e,"\n");
            putln(ss[0], "\n", ss[1]);
            return null;
        }
        closeStrm(r0);
        closeStrm(r1);
        final Result result=new Result(2);
        //putln("PairAlignerNeoBio getGappedSequence1="+alignment.getGappedSequence1());
        //putln("PairAlignerNeoBio getGappedSequence2="+alignment.getGappedSequence2());

        result.setGappedSequences(new byte[][]{alignment.getGappedSequence1().getBytes(),alignment.getGappedSequence2().getBytes()});
        result.setScore(alignment.getScore());
        return result;
    }
    /*
    private static JComboBox comboAlgorythm;
    private static ChTextField tfMatch,tfMismatch,tfGap;
    private static Container controlPanel;
    public Container getSharedControlPanel() {
        if (controlPanel==null) {
            controlPanel=vB(
                            comboAlgorythm=new JComboBox(new String[]{"Global NeedlemanWunsch","Local SmithWaterman","Global CrochemoreLandauZivUkelson","Local CrochemoreLandauZivUkelsony"}),
                            hB("match",tfMatch=new ChTextField("  1").ct(Integer.class)),
                            hB("mismatch",tfMismatch=new ChTextField("  -1").ct(Integer.class)),
                            hB("gap",tfGap=new ChTextField("  -1").ct(Integer.class))
                            );
        }
        return controlPanel;
    }
    */

}
