package charite.christo.strap.extensions;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import java.io.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP
The purpose of this class is to assign numeric values to residues in
proteins.  STRAP expects a plain text file containing the data.  These
data file can be typed by hand using a text program or may be
generated by other programs or scripts.  When the contents of the data
file[s] is changed the plot is updated automatically.  The numbers in
the file must be separated by space (space, line break, tab) or
colon. The decimal point is a dot and not a colon.  Undefined values
should be coded as NaN ("Not A Number").  The file name is specified in the
control panel.
*/
public final class ResidueValues_from_textfile implements charite.christo.strap.interfaces.ValueOfResidue, HasControlPanel, Disposable, java.awt.event.ActionListener, Runnable, HasModiCount {
    private final ChTextField tfFile=new ChTextField().ro();
    private final ChCombo comboDataset=new ChCombo("0","1","2","3","4","5","6","7","8","9","10","11","12").li(this);
    private boolean isRunning=true, isStarted;
    private File file_;
    private Protein file_lastProt;
    private int file_lastDataset;
    private Protein protein;
    private double[] values;
    public ResidueValues_from_textfile() {
        setSettings(this, Customize.fileBrowsers, Customize.textEditors);
    }
    public void dispose() { isRunning=false; }
    /* <<< Instance  <<< */
    /* ---------------------------------------- */
    /* >>> Protein >>> */
    public boolean setProtein(Protein p) {
        //putln(ANSI_GREEN+"ResidueValues_from_textfile"+ANSI_RESET+" setProtein "+p);
        if (protein!=p) {
            values=null;
            protein=p;
            updateTextFields();
            if (!isStarted) {
                isStarted=true;
                new Thread(this).start();
            }
        }
        return true;
    }
    public Protein getProtein() { return protein;}
    /* <<< Protein <<< */
    /* ---------------------------------------- */
    /* >>>  File >>> */
    private void updateTextFields() {
        File f=f();
        tfFile.t(f==null?"":f+(f.exists() ? "  "+sze(f)+" Byte" : "  not existing"))
            .tools().underlineRefs(UNDERLINE_REFS);
    }
    private File f() {
        final Protein p=getProtein();
        final int d=comboDataset.i();
        if (p==null) return null;
        if (file_==null || file_lastDataset!=d || file_lastProt!=p) {
            file_=file("residueData/"+p+"."+d);
            file_lastProt=p;
            file_lastDataset=d;
            mkParentDrs(file_);
        }
        return file_;
    }
    /* <<< File <<< */
    /* ---------------------------------------- */
    /* >>> AWTEvent >>> */
    public void actionPerformed(java.awt.event.ActionEvent ev) {
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        if (q==comboDataset){
            values=null;
            new StrapEvent(ResidueValues_from_textfile.this,StrapEvent.VALUE_OF_RESIDUE_CHANGED).run();
            updateTextFields();
        }
        if (cmd=="EDIT" && f()!=null) edFile(-1, f(), modifiers(ev));
    }

    /* <<< AWTEvent <<< */
    /* ---------------------------------------- */
    /* >>> ControlPanel >>> */
    private java.awt.Container controlPanel;
    public java.awt.Container getControlPanel() {
        if (controlPanel==null) {
            controlPanel=vBhB(
                 hBL("dataset: ",comboDataset),
                 newCNSEW(tfFile,null,null, new ChButton("EDIT").t("Edit data file").li(this), "data file: ")
                 );
        }
        return controlPanel;
    }
    /* <<< ControlPanel <<< */
    /* ---------------------------------------- */
    /* >>> getValues >>> */
    public int getModificationCount() { return modiCount;}
    private int modiCount;

    public double[] getValues() {
        final Protein p=getProtein();
        if (p==null) return null;
        final File fData=f();
        if (fData==null || !fData.exists()) {
            //error("file "+fData+" does not exist.\n<br>This file should contain a float value for each residue.\n");
            return null;
        }
        final long lastModi=fData.lastModified();
        if (values==null || lastModifiedFile!=lastModi) {
            modiCount++;
            lastModifiedFile=lastModi;
            values=readValues(fData,p.countResidues());
        }
        return values;
    }
    private double[] readValues(File fData,int nResidues) {
        double vv[]=new double[nResidues];
        for(int i=vv.length;--i>=0;) vv[i]=Double.NaN;
        InputStream is=null;
        try {
            is=new FileInputStream(fData);
            final StreamTokenizer tokenizer=new StreamTokenizer(is);
            for(int type,iAa=0; (type=tokenizer.nextToken())!=StreamTokenizer.TT_EOF && iAa<nResidues;) {
                if (type==StreamTokenizer.TT_NUMBER) {
                    vv[iAa++]=tokenizer.nval;
                }
            }
        } catch(IOException e){ return null;}
        closeStrm(is);
        return vv;
    }
    /* <<< getValues <<< */
    /* ---------------------------------------- */
    /* >>> Threading >>> */
    private long lastModifiedFile;
    public void run() {
        while(isRunning) {
            ChUtils.sleep(111);
            final File f=f();
            if (f==null) continue;
            final long lastModi=f.lastModified();
            if (lastModifiedFile!=lastModi) {
                lastModifiedFile=lastModi;
                values=null;
                new StrapEvent(ResidueValues_from_textfile.this,StrapEvent.VALUE_OF_RESIDUE_CHANGED).run();
                updateTextFields();
            }
        }
    }
    /* <<< Threading <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */
}
