package charite.christo.strap.science;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.protein.*;
import charite.christo.strap.*;
import charite.christo.strap.extensions.*;
import java.awt.*;

import java.io.File;
import java.util.*;

import javax.swing.*;
import org.jfree.chart.*;
import org.jfree.chart.entity.*;
import org.jfree.chart.plot.*;
import org.jfree.data.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;

public class Flavine extends AbstractAlignmentProject implements HasControlPanel, java.awt.event.ActionListener, ChartMouseListener{
    /* --- archives to download --- */
    {FlavinePROXY.setInstance(this); }
    public Flavine() {
        final String URL_BASE=URL_STRAP+"science/flavine/";
        setURLs(url(URL_BASE+"FlavineAlignments.zip.gz"),url(URL_BASE+"FlavineStructures.zip.gz"));
        run();
    }
    @Override public File proteinName2file(String name) {
        return  file(DIR_PROTEINS+ (sze(name)>6 ? name.substring(0,6): name) +".ent");
    }
    /* --- control panel --- */
    private Container controlPanel;
    public Container getControlPanel() {
        if (controlPanel==null) {
            final String cn=Flavine.class.getName()+"Preparation";
            Object o=invokeMthd("getPanel", cn);
            controlPanel= o instanceof JComponent ? (JComponent)o : pnl(cn+" not available");
        }
        return controlPanel;
    }
    private final AbstractButton cbTitle=cbox((String)null);
    private final static String HALF_LABELS[]={"1D","3D","3D","1d","3d","EC"};
    private final static Object KEY_ALIGNMENT=new Object();
    @Override public void actionPerformed(java.awt.event.ActionEvent ev) {
        StrapAlign.errorMsg("");
        super.actionPerformed(ev);
        final String cmd=ev.getActionCommand();
        final Object q=ev.getSource();
        final JComponent jc=q instanceof JComponent ? (JComponent)q:null;
        final Alignment alignment=(Alignment)gcp(KEY_ALIGNMENT,q);
        if (alignment!=null) {
            int col1=-1,col2=-1,i=0;
            for(String hl:HALF_LABELS) {
                if (cmd.startsWith(hl)) col1=i;
                if (cmd.endsWith(hl)) col2=i;
                i++;
            }
            if (col1>=0 && col2>=0) {
                jc.setBackground(jc.getBackground().darker());
                final File file=alignment.fChartData();
                final String lines[]=readLines(file);
                if (lines==null) {  StrapAlign.errorMsg("Could not read file "+file); return;}
                final int n=lines.length-1; // 1st line is header
                final float data[][]=new float[n][];
                final String c0[]=new String[n], c1[]=new String[n];
                for(int j=0;j<n;j++) {
                    final StringTokenizer t=new StringTokenizer(lines[j+1]);
                    if (t.countTokens()<6) continue;
                    c0[j]=t.nextToken();
                    c1[j]=t.nextToken();
                    data[j]=new float[]{f(t),f(t),f(t),f(t),f(t),f(t)};
                }
                String name=withoutDotSuffix(alignment.getInfo(NAME));
                showChart(alignment,  name, file, data, c0, c1, col1,col2);
            }
            //            final String ss[]=alignment.getProteinNames(); // ???
        }
        if (cmd=="SELECT") {
            for(int i=0;i<2;i++) StrapAlign.getSelectedObjectsV().add(StrapProtein.getProteinWithName(i==0 ? clickedC0 : clickedC1, align.getProteins()));
            new StrapEvent(this,StrapEvent.OBJECTS_SELECTED).setParameter(Protein.class).run();
        }
        if (cmd=="SUP_G" || cmd=="SUP_R") showSuperposition(clickedC0,clickedC1,  cmd=="SUP_R");
        if (cmd=="LOAD") { align.loadTheProteinsInList( proteinName2file(clickedC0)+" "+proteinName2file(clickedC1),StrapAlign.OPTION_PROCESS_LOADED_PROTS|StrapAlign.OPTION_EVENT_PROTEINS_ADDED); }
    }
    @Override public void alignmentInstantiatedHook(Alignment ali) {
        String ss[]=null;
        int count=0;
        while(ss==null) {
            if (count>0) { ss=new String[count]; count=0;}
            final boolean[] PLOTED_CPLUMN={true,false,true,true,true,true,true,true,true,true};
            for(int i=0;i<HALF_LABELS.length;i++)
                for(int j=i;j<HALF_LABELS.length;j++)
                    if (i!=j && PLOTED_CPLUMN[i] && PLOTED_CPLUMN[j]) {
                        if (ss!=null) ss[count]=HALF_LABELS[i]+HALF_LABELS[j];
                        count++;
                    }
        }
        final String[] BUT_LABELS=ss;
        final JComponent pBut=pnl(new GridLayout(2,(2+3+BUT_LABELS.length)/3));
        for(String s:BUT_LABELS) {
          pBut.add(new ChButton(s).bg(0xFFffFF).fg(s.indexOf('d')>0 ? 0xFF : 0xFF00FF).li(this));
        }
        for(Component c:pBut.getComponents()) pcp(KEY_ALIGNMENT,ali,c);
        ali.setAdditionalComponent(pBut,SwingConstants.EAST);
        ali.setAdditionalComponent(new ChButton("",ChButton.ICON_SIZE).actionEditFile(ali.fChartData()).t("As table"),SwingConstants.WEST);

    }

    private static float f(StringTokenizer t){  return (float) atof(t.nextToken(),0);}
    private final WeakHashMap<JFreeChart,Object[]> map_Chart_components=new WeakHashMap();
    private void showChart(Alignment alignment,
                           String name, File file,
                           float data[][],
                           String c0[], String c1[],
                           int col1,int col2) {
        final XYSeries series = new XYSeries("",false,true);
        final String
            title=alignment.getInfo(TITLE),
            title2=name+":     "+HALF_LABELS[col1].replaceAll("D","D ")+" - "+HALF_LABELS[col2].replaceAll("D","D "),
            AXIS_LABELS[]={
                "global sequ identity [%]",
                "global 3D-similar",
                "root mean square distance (C-alpha)",
                "sequ similar around flavine",
                "3D dissimilar of flavine pocket",
                "Enzyme Classification",
            };
        final double ff1[]=new double[data.length],ff2[]=new double[data.length];
        for(int i=0;i<data.length;i++) {
            ff1[i]=data[i][col1];
            ff2[i]=data[i][col2];
            if (!Double.isNaN(ff1[i]) && !Double.isNaN(ff2[i])) series.add(ff1[i], ff2[i]);
        }
        final ChTextField tfClicked=new ChTextField("picked dot: pick a dot with the mouse");
        final XYDataset xyDataset = new XYSeriesCollection(series);
        final JFreeChart chart=ChartFactory.createScatterPlot(null,AXIS_LABELS[col1],AXIS_LABELS[col2],xyDataset,PlotOrientation.VERTICAL,true ,true,false);
        final ChartPanel panel=new ChartPanel(chart);
        panel.addChartMouseListener(this);
        final JMenu menuApply=new ChJMenu("view");
        menuApply.setEnabled(false);
        final JMenuBar menuBar=new JMenuBar();
        menuBar.add(menuApply);
        final String
            TT_OVL="take two proteins and superimpose and display",
            TT_OVL_R=TT_OVL+"<br>Only the alignable core region is taken.";

        menuApply.add(new ChButton("LOAD").li(this).mi("load proteins into alignment pane"));
        menuApply.add(new ChButton("SELECT").li(this).mi("select both proteins in the alignment"));
        menuApply.add(new ChButton("SUP_G").li(this).tt(TT_OVL).i(IC_SUPERIMPOSE).mi("superimpose entire chains"));
        menuApply.add(new ChButton("SUP_R").li(this).tt(TT_OVL_R).i(IC_SUPERIMPOSE).mi("superimpose only core regions"));
        menuApply.add(new ChButton().actionEditFile(file).mi("show data table "));
        final JComponent taTitle=hB(monospc(title));
        final ChButton tog=toggl().actionCollapse(false,taTitle);
        tog.setModel(cbTitle.getModel());
        final JComponent pTitle=vB(taTitle,newCNSEW(tfClicked,null,null,pnl(menuBar),tog ));
        new ChFrame(title2)
            .ad(newCNSEW(panel,pTitle,"Spearmans roh="+SpearmansRangCorrelation.calcRoh(ff1,ff2),null,null))
            .size(360,280).i(IC_PLOT).shw(ChFrame.STAGGER|ChFrame.DISPOSE_ON_CLOSE);
        map_Chart_components.put(chart,new Object[]{c0,c1,tfClicked,menuApply});
    }
    public void chartMouseClicked(ChartMouseEvent ev) {
        final ChartEntity entity=ev.getEntity();
        JFreeChart chart=ev.getChart();
        if (entity==null) return;
        final int i=((XYItemEntity)entity).getItem();
        final Object[] oo=map_Chart_components.get(chart);
        clickedC0=((String[])oo[0])[i];
        clickedC1=((String[])oo[1])[i];
        final JTextField tf=(JTextField)oo[2];
        final JMenu menuApply=(JMenu)oo[3];
        tf.setText("clicked: "+clickedC0+" "+clickedC1);
        menuApply.setEnabled(true);
    }
    public void chartMouseMoved(ChartMouseEvent ev) {}
    private String clickedC0,clickedC1;
    private final void showSuperposition(String pn1,String pn2, boolean isLocal) {
        final StrapProtein p1=StrapProtein.newInstance(proteinName2file(pn1));
        final StrapProtein p2=StrapProtein.newInstance(proteinName2file(pn2));
        if (p1==null || p2==null) { putln("Error showSuperposition( p1="+pn1+" p2="+pn2+")"); return;}
        p1.setFileWithSideChainsObtainableFromPDB(true);
        p2.setFileWithSideChainsObtainableFromPDB(true);
        p1.setName(pn1);
        p2.setName(pn2);
        if (isLocal) {
            for(StrapProtein p : new StrapProtein[]{p1,p2}) {
                StrapAlign.readAttributes("A",p, (File)null);
                int idxFrom=maxi(0,ResSelUtils.getResidueindexOfAnnotation(p,"b"));
                int idxTo=ResSelUtils.getResidueindexOfAnnotation(p,"e");
                if (idxTo<0) idxTo=p.countResidues()-1;
                if (idxFrom>idxTo) idxFrom=0;
                p.setResidueSubset(idxFrom+"-"+idxTo);
            }
        }
        StrapProteinUtils.viewComparison(p1,p2,Superimpose_CEPROXY.class);
    }
    @Override public StrapProtein[] loadProteins(String name, String proteinNames[]){
        putln(" proteinNames=");putln(proteinNames);

        final UniqueList<StrapProtein> vP=new UniqueList(StrapProtein.class);
        for(String fn:proteinNames) {
            if (sze(fn)<6) continue;
            final File f=proteinName2file(fn);
            StrapProtein p=StrapProtein.getProteinWithName(fn,align.getProteins());
            if (p==null) p=StrapProtein.newInstance(f);
            if (p!=null) {
                p.setName(fn);
                p.setAlignment(align);
                p.setFileWithSideChainsObtainableFromPDB(true);
                StrapAlign.readAttributes("3G",p,(File)null);
                final int us=fn.lastIndexOf('_');
                final String hetero=atoi(fn,7)+":"+ (us+1<sze(fn) ? fn.substring(us+1,us+2) : "");
                pcp(Distance3DToHetero.KEY_DEFAULT_HETERO, hetero, p);
                vP.add(p);
            } else putln("no such protein "+f);
        }

        final StrapProtein pp[]=vP.asArray();
        final String fBase=withoutDotSuffix("info/"+name);
        mkParentDrs(file("info"));
        align.addProteins(0,pp);
        for(int t=3;--t>=0;) {
            final StringBuffer sb=new StringBuffer(999);
            String suffix=null;
            for(StrapProtein p:pp) {
                if (p==null) continue;
                String s=null;
                switch(t) {
                case 0:
                    s=joinAsStrg(p.getEC()," ");
                    suffix=".EC";
                    break;
                case 1:
                    s=p.getCompound();
                    if (s!=null) {
                        int i=s.indexOf("MOLECULE:");
                        int cr=s.indexOf('\n',i);
                        if (i>0 && cr>0) s=s.substring(i+10,cr);
                    }
                    suffix=".DEF";
                    break;
                case 2:
                    s=p.getTitle();
                    if (s!=null) s=s.replace('\n',' ');
                    suffix=".TITLE";
                }
                s=delSfx(",",s);
                s=delSfx(";",s);
                if (sze(s)>0 && sb.indexOf(s)<0) sb.append(s.replace('\n',' ')).append("\n");
            }
            if (sze(sb)>0) wrte(file(fBase+suffix),sb);
        }
        align.dispatchLater(StrapEvent.PROTEINS_ADDED,111);
        if (align.getAlignmentPanels().length==0) align.newAlignmentPanel();

        return pp;
    }
}
